theory Card_Mode
imports Main
begin

text {*
  The concrete card mode. Degradations of the hash region and the PTR
  have special modes associated with them. This is invisible from the
  outside, but internally we need to distinguish these modes.  
*}

datatype card_mode = CFail | CDegradedH | CDegradedP | CDegraded | CNormal

instantiation card_mode :: "{bot, top, linorder}"
begin

fun less_eq_card_mode :: "card_mode \<Rightarrow> card_mode \<Rightarrow> bool" where
  "less_eq_card_mode CFail _ = True"
  | "less_eq_card_mode _ CFail = False"
  | "less_eq_card_mode CDegradedH _ = True" 
  | "less_eq_card_mode _ CDegradedH = False" 
  | "less_eq_card_mode CDegradedP _ = True" 
  | "less_eq_card_mode _ CDegradedP = False" 
  | "less_eq_card_mode CDegraded _ = True" 
  | "less_eq_card_mode _ CDegraded = False" 
  | "less_eq_card_mode CNormal CNormal = True" 

definition bot_card_mode :: card_mode where
[simp]: "bot_card_mode \<equiv> CFail"

definition top_card_mode :: card_mode where
[simp]: "top_card_mode \<equiv> CNormal"

definition less_card_mode :: "card_mode \<Rightarrow> card_mode \<Rightarrow> bool" where
[simp]: "less_card_mode x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: card_mode
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: card_mode
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: card_mode
  show "bot \<le> a"
    by(simp)
next
  fix a :: card_mode
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

text {*
  The abstract card mode. All degradations are treated uniformly,
  i.e. we do not distinguish e.g. @{term "Degraded"} and @{term
  "DegradedH"}. Used in the more abstract models, where we do away
  with the explicit representations of the admin regions.
*}

datatype abs_card_mode = ACFail | ACDegraded | ACNormal

instantiation abs_card_mode :: "{bot, top, linorder}"
begin

fun less_eq_abs_card_mode :: "abs_card_mode \<Rightarrow> abs_card_mode \<Rightarrow> bool" where
  "less_eq_abs_card_mode ACFail _ = True"
  | "less_eq_abs_card_mode _ ACFail = False"
  | "less_eq_abs_card_mode ACDegraded _ = True" 
  | "less_eq_abs_card_mode _ ACDegraded = False" 
  | "less_eq_abs_card_mode ACNormal ACNormal = True" 

definition bot_abs_card_mode :: abs_card_mode where
[simp]: "bot_abs_card_mode \<equiv> ACFail"

definition top_abs_card_mode :: abs_card_mode where
[simp]: "top_abs_card_mode \<equiv> ACNormal"

definition less_abs_card_mode :: "abs_card_mode \<Rightarrow> abs_card_mode \<Rightarrow> bool" where
[simp]: "less_abs_card_mode x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: abs_card_mode
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: abs_card_mode
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: abs_card_mode
  show "bot \<le> a"
    by(simp)
next
  fix a :: abs_card_mode
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

end