theory Cartesian_Equipotent
imports Main "~~/src/HOL/ex/Set_Theory"
begin

class cartesian_equipotent =
  fixes square_map :: "'a \<Rightarrow> ('a \<times> 'a)"
  assumes map_surj: "surj square_map"

lemma inj_to_square: "\<exists>f :: 'a \<Rightarrow> ('a \<times> 'a). inj f"
  by(rule_tac x="\<lambda>y. (y, y)" in exI, rule injI, auto)

lemma surj_implies_backward_inv:
  assumes f_surj: "\<exists>f :: 'a \<Rightarrow> 'b. surj f" 
  shows "\<exists>g :: 'b \<Rightarrow> 'a. inj g"
proof-
  from f_surj obtain f :: "'a \<Rightarrow> 'b" where "surj f" ..
  thus "\<exists>g :: 'b \<Rightarrow> 'a. inj g"
    by(rule_tac x="inv f" in exI, rule surj_imp_inj_inv)
qed

lemma (in cartesian_equipotent) map_inj:
  "\<exists>g :: ('a \<times> 'a) \<Rightarrow> 'a. inj g"
  apply(rule surj_implies_backward_inv)
  apply(metis map_surj)
  done

lemma (in cartesian_equipotent) map_bij': 
  shows "\<exists>f :: ('a \<Rightarrow> ('a \<times> 'a)). inj f \<and> surj f"
proof-
  from inj_to_square obtain f :: "'a \<Rightarrow> ('a \<times> 'a)" where forward: "inj f" ..
  from map_inj obtain g :: "'a \<times> 'a \<Rightarrow> 'a" where backward: "inj g" ..
  show ?thesis using forward backward
    by(rule Schroeder_Bernstein)
qed

lemma (in cartesian_equipotent) map_bij: 
  "\<exists>f :: 'a \<Rightarrow> ('a \<times> 'a). bij f"
proof-
  show ?thesis using map_bij'
    by(blast intro!: bijI)
qed

lemma (in cartesian_equipotent) map_bij_rev:
  "\<exists>f :: ('a \<times> 'a) \<Rightarrow> 'a. bij f"
proof-
  from map_bij obtain f :: "'a \<Rightarrow> ('a \<times> 'a)" where "bij f" ..
  thus ?thesis
    by(rule_tac x="inv f" in exI) (blast intro: bij_imp_bij_inv)
qed

lemma card_product:  "\<lbrakk> finite A; finite B \<rbrakk> \<Longrightarrow> card (A \<times> B)  = card A * card B"
  by(induct set: finite, auto)

lemma card_product_types: "\<lbrakk> finite (UNIV :: 'a set); finite (UNIV :: 'b set) \<rbrakk> \<Longrightarrow>
  card (UNIV :: ('a \<times> 'b)set) = card (UNIV :: 'a set) * card (UNIV :: 'b set)"
  by(drule card_product, auto)

(* This is *NOT* provable, since infinite sets have a card of 0, but finite sets (which
    could be the codomain) have card \<ge> 0.

lemma surj_image: "surj (f :: 'a \<Rightarrow> 'b) \<Longrightarrow>  card (UNIV :: 'b set) \<le> card (UNIV :: 'a set)"

*)

text {* Since @{term "UNIV \<noteq> {}"} always holds, we can assume that card = 1 below *}

lemma (in cartesian_equipotent) singleton_or_infinite:
  shows "\<not> finite (UNIV :: 'a set) \<or> card (UNIV :: 'a set) = 1"
proof(rule disj_not1[THEN iffD2, OF impI])
  assume finiteness: "finite (UNIV :: 'a set)"
  have card_eq: "card (UNIV :: 'a set) = card (UNIV :: ('a \<times> 'a)set)" using map_bij
    by(auto dest!: bij_betw_same_card)
  note card_prod' =  card_product_types[OF finiteness finiteness]
  hence "card (UNIV :: 'a set) = 0 \<or> card (UNIV :: 'a set) = 1" using card_eq
    by(auto)
  thus "card (UNIV :: 'a set) = 1" using finiteness
    by(auto)
qed

definition inj2 :: "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> bool" where
  "inj2 f \<equiv> \<forall>x x' y y'. x \<noteq> x' \<or> y \<noteq> y' \<longrightarrow> f x y \<noteq> f x' y'"

lemma inj2D: "inj2 f \<Longrightarrow> (x' \<noteq> x \<longrightarrow> f x y \<noteq> f x' y') \<and> (b \<noteq> b' \<longrightarrow> f a b \<noteq> f a' b')"
  by(auto simp add: inj2_def)

lemma inj2I: "inj f \<Longrightarrow> inj2 (curry f)"
  by(auto simp add: curry_def inj2_def dest: injD)

lemma inj2_foldr:
  "\<lbrakk> inj2 f; l2 \<noteq> l1; length l2 = length l1 \<rbrakk> \<Longrightarrow> foldr f l1 a \<noteq> foldr f l2 b"
proof(induct l1 arbitrary: l2 a b)
  case Nil
  thus ?case
    by simp
next
  case (Cons x xs) note Asm=Cons
  show ?case
  proof(cases l2)
    case Nil
    thus ?thesis using Asm(4)
      by simp
  next
    case (Cons y ys)
    have diff: "x \<noteq> y \<or> xs \<noteq> ys" using Cons Asm(3)
      by(auto)
    show ?thesis using Cons Asm
      by(case_tac "a=b", auto simp add: inj2D)
  qed
qed

lemma inj_foldr:
  "\<lbrakk> inj f; l2 \<noteq> l1; length l2 = length l1 \<rbrakk> \<Longrightarrow>
    foldr (curry f) l1 a \<noteq> foldr (curry f) l2 b"
  by(rule inj2_foldr, drule inj2I, simp_all)

lemma length_map2: "length (map f l) = length (map g l)"
  by(induct l, auto)

lemma (in cartesian_equipotent) inj_from_finite_maps:
  "\<exists>f :: ('b :: finite \<Rightarrow> 'a) \<Rightarrow> 'a. inj f"
proof-
  have "\<exists>l. fold_graph Cons [] (UNIV :: 'b set) l"
    by(auto intro!: finite_imp_fold_graph)
  from this obtain l :: "'b list" where l_def: "fold_graph Cons [] (UNIV :: 'b set) l" ..
  have l_UNIV: "set l = UNIV"
  proof
    show "set l \<subseteq> UNIV" 
      by simp
  next
    show "UNIV \<subseteq> set l" using l_def
      by(induct set: fold_graph) auto
  qed

  {
    fix f f'
    have "f \<noteq> f' \<Longrightarrow> \<exists>x \<in> set l. f x \<noteq> f' x"
    proof(rule ccontr)
      assume assm1: "\<not> (\<exists>x \<in> set l. f x \<noteq> f' x)" and assm2: "f \<noteq> f'"
      have "\<forall>x. f x = f' x" using assm1 l_UNIV
        by auto
      from this[rule_format] have "f = f'"
        by(rule ext)
      thus False using assm2 
        by simp
    qed
    hence "f \<noteq> f' \<Longrightarrow> map f l \<noteq> map f' l"
      by(auto intro!: map_eq_conv[THEN Not_eq_iff[THEN sym, THEN iffD1], THEN iffD2])
  } note diff=this

  fix start :: 'a
  from map_inj obtain h :: "('a \<times> 'a) \<Rightarrow> 'a" where inj_h: "inj h" ..
  have "inj (\<lambda>g. foldr (curry h) (map g l) start)"
  proof(rule injI[OF ccontr, rule_format])
    fix f f' :: "'b \<Rightarrow> 'a"
    assume not_eq: "f \<noteq> f'"
      and eq_fold: "foldr (curry h) (map f l) start = foldr (curry h) (map f' l) start"
    from not_eq have "map f l \<noteq> map f' l"
      by(rule diff)
    hence maps_diff: "map f' l \<noteq> map f l" ..
    have "foldr (curry h) (map f l) start \<noteq> foldr (curry h) (map f' l) start" using maps_diff inj_h
      apply -
      apply(rule inj_foldr)
      apply(simp_all)
      done
    thus False using eq_fold ..
  qed
  thus ?thesis
    by(blast)
qed

lemma (in cartesian_equipotent) inj_to_finite_maps:
  "\<exists>f :: 'a \<Rightarrow> ('b :: finite \<Rightarrow> 'a). inj f"
  by(rule_tac x="\<lambda>x y. x" in exI) (auto intro!: injI dest!: fun_cong)

lemma (in cartesian_equipotent) bij_finite_maps': 
  "\<exists>f :: 'a \<Rightarrow> ('b :: finite \<Rightarrow> 'a). inj f \<and> surj f"
proof-
  from inj_from_finite_maps obtain f :: "('b :: finite \<Rightarrow> 'a) \<Rightarrow> 'a" where "inj f " ..
  moreover from inj_to_finite_maps obtain g :: "'a \<Rightarrow> ('b :: finite \<Rightarrow> 'a)" where "inj g" ..
  ultimately show ?thesis
    by(auto intro!: Schroeder_Bernstein) (* TODO: again, rule doesn't work here? *)
qed

lemma (in cartesian_equipotent) bij_finite_maps:
  "\<exists>f :: 'a \<Rightarrow> ('b :: finite \<Rightarrow> 'a). bij f"
  by(simp add: bij_finite_maps' bij_def)

lemma (in cartesian_equipotent) bij_finite_maps_rev:
  "\<exists>f :: ('b :: finite \<Rightarrow> 'a) \<Rightarrow> 'a. bij f"
proof-
  from bij_finite_maps obtain f :: "'a \<Rightarrow> ('b :: finite \<Rightarrow> 'a)" where "bij f" ..
  thus ?thesis
    by(auto intro!: bij_imp_bij_inv)
qed

lemma unit_functions: "f x = ((), ())"
  by(case_tac "f x", auto)

instantiation unit :: cartesian_equipotent
begin
definition "square_map \<equiv> \<lambda>x. ((), ())"
instance proof
  show "surj (square_map :: unit \<Rightarrow> unit \<times> unit)"
    by(auto intro!: range_eqI unit_functions[THEN sym])
qed
end

type_synonym bit_vector = "bool list"

term "[True, False]"

class has_two =
  fixes has_two_a :: 'a and has_two_b :: 'a
  assumes diff_vals: "has_two_a \<noteq> has_two_b"

instantiation bool :: has_two 
begin
definition "has_two_a \<equiv> True"
definition "has_two_b \<equiv> False"
instance proof qed (simp add: has_two_a_bool_def has_two_b_bool_def)
end

term "repeat"

lemma list_prefix_eqD: "\<lbrakk>xs = ys;  xs@zs = ys@zs'\<rbrakk> \<Longrightarrow> zs = zs'"
  by auto

lemma prefix_diff: "\<lbrakk> x \<noteq> y; n1 \<noteq> n2 \<rbrakk> \<Longrightarrow> replicate n1 x @ y # l1 \<noteq> replicate n2 x @ y # l2"
  apply(induct n1 arbitrary: n2 l1 l2)
  apply(case_tac[!] n2)
  apply(auto)
  done

definition 
  "encode_list_pair x y \<equiv>
    \<lambda>(l, l'). replicate (length l) x @ [y] @ replicate (length l') x @ [y] @ l @ l'"

lemma
  inj_encode: "x \<noteq> y \<Longrightarrow> inj (encode_list_pair x y)"
proof(rule injI, rule ccontr)
  fix p1 p2
  assume diff: "p1 \<noteq> p2"
    and diff_x_y: "x \<noteq> y"
    and enc_eq: "encode_list_pair x y p1 = encode_list_pair x y p2"
  obtain l1 l2 where p1_exp: "p1 = (l1, l2)" by (case_tac p1, auto)
  obtain l3 l4 where p2_exp: "p2 = (l3, l4)" by (case_tac p2, auto)
  from p1_exp p2_exp diff have diff_lists: "l1 \<noteq> l3 \<or> l2 \<noteq> l4" by(auto)
  have "encode_list_pair x y (l1, l2) \<noteq> encode_list_pair x y (l3, l4)" 
    using diff_lists diff_x_y
  proof(cases "l1 = l3")
    case True note eq=this
    hence diff_2_4: "l2 \<noteq> l4" using diff_lists by auto
    thus ?thesis using eq diff_x_y
      by(case_tac "length l2 = length l4") (auto simp add: encode_list_pair_def dest: prefix_diff)
  next
    case False note neq = this
    thus ?thesis using diff_x_y
      apply(case_tac "length l1 = length l3", case_tac "length l2 = length l4")
      apply(auto simp add: encode_list_pair_def dest: prefix_diff)
      done
  qed

  thus False using enc_eq p1_exp p2_exp
    by simp
qed

consts two_diffs :: "'a :: has_two \<times> 'a"
specification(two_diffs) diff_two_diffs: "fst two_diffs \<noteq> snd two_diffs"
  by(insert diff_vals) (auto)
  
instantiation list :: (has_two) cartesian_equipotent
begin
definition "square_map \<equiv> inv (encode_list_pair (fst two_diffs) (snd two_diffs))"
instance proof
  show "surj (square_map :: ('a :: has_two) list \<Rightarrow> 'a list \<times> 'a list)"
    by(unfold square_map_list_def) (
      blast intro!: inj_imp_surj_inv inj_encode diff_two_diffs)
qed
end

end
