theory Full_Ref
imports 
  "abs1/Abs1_SpecRef"
  "abs2/Abs2_SpecRef"
  "abs3/Abs3_SpecRef"
  "abs4/Abs4_SpecRef"
  "abs5/Abs5_SpecRef"
  "abs6/Abs6_SpecRef"
begin

definition full_ref_rel where
  "full_ref_rel \<equiv>
    abs6_abs5_strel
    O (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    O (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    O abs3_abs2_strel
    O abs2_abs1_strel
    O abs1_conc_strel
  "

definition full_pi_composition where
  "full_pi_composition \<equiv> id o abs4_abs5_pi o abs3_abs4_pi o abs2_abs3_pi o abs1_abs2_pi
    o pmm_abs1_pi"

lemmas refinement_lemmas =
  pmm_refines_abs1
  abs1_refines_abs2
  abs2_refines_abs3
  abs3_refines_abs4
  abs4_refines_abs5
  abs5_refines_abs6

theorem concrete_refines_abs6:
  "refines full_ref_rel full_pi_composition abs6_spec pmm_spec"
  apply(simp only: full_ref_rel_def full_pi_composition_def comp_assoc)
  apply(intro refinement_transitive, (rule refinement_lemmas)+)
  done

lemmas implementation_lemmas =
  pmm_implements_abs1
  abs1_implements_abs2
  abs2_implements_abs3
  abs3_implements_abs4
  abs4_implements_abs5
  abs5_implements_abs6


theorem concrete_implements_abs6:
  "implements full_pi_composition abs6_spec pmm_spec"
  by(unfold full_pi_composition_def, intro implements_trans, (rule implementation_lemmas)+)

lemma update_full_ref_N:
  "sim_norm full_ref_rel
    (<tryR>
        Abs6_UpdateDefs.update upd_f
     <catchR>
        Abs6_StartupDefs.startup)
    (<tryR>
        UpdateDefs.update upd_f
     <catchR>
        rec_tryR StartupDefs.startup)
  "
  apply(unfold full_ref_rel_def)
  apply(auto intro!: sim_norm_compose abs6_update_full_ref_N abs5_update_full_ref_N
    abs4_update_full_ref_N Abs3_Ref.update_full_ref Abs2_Ref.update_full_ref 
    Abs1_Ref.update_full_ref[THEN sim_imp_sim_norm])
  done

lemma fetch_full_ref_N:
  "sim_norm full_ref_rel
    (<tryR>
        Abs6_FetchDefs.fetch ind
     <catchR>
        Abs6_StartupDefs.startup)
    (<tryR>
        FetchDefs.fetch ind
     <catchR>
        rec_tryR StartupDefs.startup)
  "
  apply(unfold full_ref_rel_def)
  apply(auto intro!: sim_norm_compose abs6_fetch_full_ref_N abs5_fetch_full_ref_N
    abs4_fetch_full_ref_N Abs3_Ref.fetch_full_ref Abs2_Ref.fetch_full_ref 
    Abs1_Ref.fetch_full_ref[THEN sim_imp_sim_norm])
  done

lemma startup_full_ref_N:
  "sim_norm full_ref_rel
    Abs6_StartupDefs.startup
    (<tryR>
        StartupDefs.startup
     <catchR>
        rec_tryR StartupDefs.startup)
  "
  apply(unfold full_ref_rel_def)
  apply(auto intro!: sim_norm_compose abs6_startup_full_ref_N abs5_startup_full_ref_N
    abs4_startup_full_ref_N abs3_startup_full_ref_N abs2_startup_full_ref_N 
    abs1_startup_full_ref_N[THEN sim_imp_sim_norm])
  done

end