theory HoareLogic
imports EvaluationLogicR Projections Composition
begin

definition
  Hoare :: "'s assn \<Rightarrow> ('a, 's)nds_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>>/ _\<rbrace>)" [0, 0, 0] 90) where
  "Hoare P m Q \<equiv> P \<subseteq> Box m Q"

definition
  HoareN :: "'s assn \<Rightarrow> ('a, 's)r_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>>R/ _\<rbrace>)" [0, 0, 0] 90) where
  "HoareN P m Q \<equiv> P \<subseteq> [R x \<leftarrow> m](Q x)"

definition
  HoareS :: "'s assn \<Rightarrow> ('s \<Rightarrow> 's nd_monad) \<Rightarrow> 's assn \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>S/ _\<rbrace>)" [0, 0, 0] 90) where
  "HoareS P m Q \<equiv> P \<subseteq> [S m] Q"

definition
  HoareR :: "'s assn \<Rightarrow> ('a, 's)r_monad \<Rightarrow> 's assn \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>@R/ _\<rbrace>)" [0, 0, 0] 90) where
  "HoareR P m Q \<equiv> P \<subseteq> [@R m] Q"


definition
  HoareNN :: "'s assn \<Rightarrow> ('a, 'e, 's)re_monad \<Rightarrow> ('a \<Rightarrow> 's assn) \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>>N/ _\<rbrace>)" [0, 0, 0] 90) where
  "HoareNN P m Q \<equiv> P \<subseteq> [x \<leftarrow> m](Q x)"

definition
  HoareNE :: "'s assn \<Rightarrow> ('a, 'e, 's) re_monad \<Rightarrow> ('e \<Rightarrow> 's assn) \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>@N/ _\<rbrace>)" [0, 0, 0] 90) where
  "HoareNE P m Q \<equiv> P \<subseteq> [@ x\<leftarrow> m](Q x)"

definition
  Hoare_all :: "'s assn \<Rightarrow>  ('a, 'e, 's)re_monad \<Rightarrow> 's assn \<Rightarrow> bool"
           ("(3\<lbrace>_\<rbrace> _ \<lbrace>&/ _\<rbrace>)" [0, 0, 0] 90) where
  "Hoare_all P m Q \<equiv> HoareNN P m (\<lambda>_. Q) \<and> HoareNE P m (\<lambda>_. Q) \<and> HoareR P m Q"

lemmas Hoare_defs = Hoare_def HoareN_def HoareR_def HoareNN_def HoareNE_def 

lemma HoareN_alt:
  "HoareN P m Q = Hoare P (m\<Down>N) Q"
  by(auto simp add: HoareN_def Hoare_def project_defs Box_defs Box_def)

lemma HoareNN_alt:
  "HoareNN P m Q = Hoare P (m\<Down>NN) Q"
  by(auto simp add: HoareNN_def Hoare_def project_defs Box_defs Box_def split: sum.split_asm sum.split)

lemma HoareNE_alt:
  "HoareNE P m Q = Hoare P (m\<Down>NE) Q"
  by(auto simp add: HoareNE_def Hoare_def project_defs Box_defs Box_def split: sum.split_asm sum.split)

lemma HoareR_alt:
  "HoareR P m Q = Hoare P (\<lambda>s. (\<lambda>x. (undefined, x)) ` (m\<Down>R) s) (\<lambda>_. Q)"
  by(auto simp add: HoareR_def Hoare_def project_defs Box_defs Box_def split: sum.split_asm sum.split)

lemma HoareR_alt2:
  "HoareR P m Q = HoareS P (m\<Down>R) Q"
  by(auto simp add: HoareR_def HoareS_def project_defs Box_defs BoxS_def split: sum.split_asm sum.split)

lemmas Hoare_alt = HoareN_alt HoareNN_alt HoareNE_alt

(******************************************************************************)
subsection {* Ex falsum and trivial post-condition *}
(******************************************************************************)

lemma ex_falsum: "\<lbrace>false\<rbrace> m \<lbrace>> Q\<rbrace>"
  by (simp add: Hoare_defs)

lemma ex_falsumN: "\<lbrace>false\<rbrace> m \<lbrace>>R Q\<rbrace>"
  by (simp add: Hoare_defs)

lemma ex_falsumR: "\<lbrace>false\<rbrace> m \<lbrace>@R Q\<rbrace>"
  by (simp add: Hoare_defs)

lemma ex_falsumNN: "\<lbrace>false\<rbrace> m \<lbrace>>N Q\<rbrace>"
  by (simp add: Hoare_defs)

lemma ex_falsumNE: "\<lbrace>false\<rbrace> m \<lbrace>@N Q\<rbrace>"
  by (simp add: Hoare_defs)

lemma true_post: "\<lbrace>P\<rbrace> m \<lbrace>> K true\<rbrace>"
  by(simp add: Hoare_defs Box_true K_def)

lemma true_postN: "\<lbrace>P\<rbrace> m \<lbrace>>R K true\<rbrace>"
  by(simp add: Hoare_defs BoxN_true K_def)

lemma true_postR: "\<lbrace>P\<rbrace> m \<lbrace>@R true\<rbrace>"
  by(simp add: Hoare_defs BoxR_true)

lemma true_postNN: "\<lbrace>P\<rbrace> m \<lbrace>>N K true\<rbrace>"
  by(simp add: Hoare_defs BoxNN_true K_def)

lemma true_postNE: "\<lbrace>P\<rbrace> m \<lbrace>@N K true\<rbrace>"
  by(simp add: Hoare_defs BoxNE_true K_def)

(******************************************************************************)
subsection {* Consequence rules *}
(******************************************************************************)

lemma conseq_rule:
  "\<lbrakk> P \<subseteq> Q; \<lbrace>Q\<rbrace> m \<lbrace>> R\<rbrace>; \<forall>x. R x \<subseteq> S x \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>> S\<rbrace>"
  apply (unfold Hoare_def)
  apply (subgoal_tac "[N x \<leftarrow> m] (R x) \<subseteq> [N x \<leftarrow> m] (S x)", simp)
  apply (intro Box_monotone, simp)
  done
 
                                    
lemma conseqN_rule:
  "\<lbrakk> P \<subseteq> Q; \<lbrace>Q\<rbrace> m \<lbrace>>R R\<rbrace>; \<forall>x. R x \<subseteq> S x \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>>R S\<rbrace>"
  apply (unfold HoareN_def)
  apply (subgoal_tac "[R x \<leftarrow> m] (R x) \<subseteq> [R x \<leftarrow> m] (S x)", simp)
  apply (intro BoxN_monotone, simp)
  done
                                    
lemma conseqR_rule:
  "\<lbrakk> P \<subseteq> Q; \<lbrace>Q\<rbrace> m \<lbrace>@R R\<rbrace>; R \<subseteq> S \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@R S\<rbrace>"
  apply (unfold HoareR_def)
  apply (subgoal_tac "[@R m]R <= [@R m]S", simp)
  apply (intro BoxR_monotone, simp)
  done

lemma conseqNN_rule:
  "\<lbrakk> P \<subseteq> Q; \<lbrace>Q\<rbrace> m \<lbrace>>N R\<rbrace>; \<forall>x. R x \<subseteq> S x \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>>N S\<rbrace>"
  apply (unfold HoareNN_def)
  apply (subgoal_tac "[ x \<leftarrow> m] (R x) \<subseteq> [ x \<leftarrow> m] (S x)", simp)
  apply (intro BoxNN_monotone, simp)
  done

lemma conseqNE_rule:
  "\<lbrakk> P \<subseteq> Q; \<lbrace>Q\<rbrace> m \<lbrace>@N R\<rbrace>; \<forall>x. R x \<subseteq> S x \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@N S\<rbrace>"
  apply (unfold HoareNE_def)
  apply (subgoal_tac "[@ x \<leftarrow> m] (R x) \<subseteq> [@ x \<leftarrow> m] (S x)", simp)
  apply (intro BoxNE_monotone, simp)
  done
 
(******************************************************************************)
subsection {* Rules related to monad *}
(******************************************************************************)

lemma return_rule: "\<lbrace>P x\<rbrace> returnN x \<lbrace>> P\<rbrace>"
  by (simp add: Hoare_def Box_return)

lemma returnN_rule: "\<lbrace>P x\<rbrace> returnR x \<lbrace>>R P\<rbrace>"
  by (simp add: HoareN_def BoxN_return)

lemma returnR_rule: "\<lbrace>P\<rbrace> returnR x \<lbrace>@R Q\<rbrace>"
  by (simp add: HoareR_def BoxR_return)

lemma returnNN_rule: "\<lbrace>P x\<rbrace> return x \<lbrace>>N P\<rbrace>"
  by (simp add: HoareNN_def BoxNN_return)

lemma returNEN_rule: "\<lbrace>P x\<rbrace> return x \<lbrace>@N P\<rbrace>"
  by (simp add: HoareNE_def BoxNE_return)

lemma bindN_ruleN: 
  "\<lbrakk> \<lbrace>P\<rbrace> m1 \<lbrace>> Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> m2 x \<lbrace>> R\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> (m1 \<guillemotright>=N m2) \<lbrace>> R\<rbrace>"
  apply (unfold Hoare_def)
  apply (subgoal_tac "Box m1 Q \<subseteq> [N x \<leftarrow> m1][N y \<leftarrow> m2 x] (R y)")
  apply(auto intro!: order_trans Box_monotone simp add: Box_bind)
  done

lemma bindR_ruleN: 
  "\<lbrakk> \<lbrace>P\<rbrace> m1 \<lbrace>>R Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> m2 x \<lbrace>>R R\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> (m1 \<guillemotright>=R m2) \<lbrace>>R R\<rbrace>"
  apply (unfold HoareN_def)
  apply (subgoal_tac "BoxN m1 Q \<subseteq> [R x \<leftarrow> m1][R y \<leftarrow> m2 x] (R y)")
  apply (rule order_trans, simp_all add: BoxN_bind)
  apply (rule BoxN_monotone, simp)
  done

lemma bindR_ruleR:  
  "\<lbrakk> \<lbrace>P\<rbrace> p \<lbrace>@R R\<rbrace>;
      \<lbrace>P\<rbrace> p \<lbrace>>R Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> q x \<lbrace>@R R\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> (p \<guillemotright>=R q) \<lbrace>@R R\<rbrace>"
  apply (simp only: HoareN_def HoareR_def BoxR_bind)
  apply (rule Int_greatest, simp_all)
  apply (subgoal_tac "[R x \<leftarrow> p] Q x \<subseteq> [R x \<leftarrow> p][@R q x] R", simp)
  apply (rule BoxN_monotone, simp)
  done

lemma bind_ruleNN:
  "\<lbrakk> \<lbrace>P\<rbrace> m1 \<lbrace>>N Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> m2 x \<lbrace>>N R\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> (m1 \<guillemotright>= m2) \<lbrace>>N R\<rbrace>"
  apply (unfold HoareNN_def)
  apply (subgoal_tac "BoxNN m1 Q \<subseteq> [ x \<leftarrow> m1][ y \<leftarrow> m2 x] (R y)")
  apply (rule order_trans, simp_all add: BoxNN_bind)
  apply (rule BoxNN_monotone, simp)
  done

lemma bind_ruleNE:
  "\<lbrakk> \<lbrace>P\<rbrace> m1 \<lbrace>@N R\<rbrace>;
    \<lbrace>P\<rbrace> m1 \<lbrace>>N Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> m2 x \<lbrace>@N R\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> (m1 \<guillemotright>= m2) \<lbrace>@N R\<rbrace>"
  apply (unfold HoareNE_def HoareNN_def BoxNE_bind)
  apply (rule Int_greatest, simp_all)
  apply (subgoal_tac "[ x \<leftarrow> m1] Q x \<subseteq> [ x \<leftarrow> m1][@ y \<leftarrow> m2 x] R y", simp)
  apply (rule BoxNN_monotone, simp)
  done

lemma bind_inv_same:
  "\<lbrakk> \<lbrace>P\<rbrace> (m1 :: ('a, 'e, 's)re_monad) \<lbrace>> K (Q :: 's set)\<rbrace>; \<forall>x. \<lbrace>Q\<rbrace> m2 x \<lbrace>> K Q\<rbrace> \<rbrakk>
  \<Longrightarrow> \<lbrace>P\<rbrace> bind m1 m2 \<lbrace>> K Q\<rbrace>"
  apply(unfold Hoare_def)
  apply(auto simp add: Box_def unfold_monad split: option.split_asm sum.split_asm)
  apply(auto dest!: set_mp)
  done

lemma bindS_inv_same:
"\<lbrakk>\<lbrace>P\<rbrace> m1 \<lbrace>> K Q\<rbrace>; \<lbrace>Q\<rbrace> m2 \<lbrace>> K Q\<rbrace>\<rbrakk> \<Longrightarrow>
  \<lbrace>P\<rbrace> (\<lambda>s. (m1\<Down>R) s \<guillemotright>=S m2) \<lbrace>> K Q\<rbrace>"
  by(simp add: Hoare_def Box_def unfold_monad project_defs, blast)
 
lemmas bind_rules =
  bind_inv_same
  bindN_ruleN
  bindR_ruleN
  bindR_ruleR
  bind_ruleNN
  bind_ruleNE

lemma try_catch_inv_same:
  "\<lbrakk> \<lbrace>P\<rbrace> (m1 :: ('a, 'e, 's)re_monad) \<lbrace>> K (Q :: 's set)\<rbrace>; \<forall>e. \<lbrace>Q\<rbrace> m2 e \<lbrace>> K Q\<rbrace> \<rbrakk>
  \<Longrightarrow> \<lbrace>P\<rbrace> try_catch m1 m2 \<lbrace>> K Q\<rbrace>"
  apply(unfold Hoare_def)
  apply(auto simp add: Box_def unfold_monad split: option.split_asm sum.split_asm)
  apply(auto dest!: set_mp)
  done

lemma choice_inv:
  "\<lbrace>P\<rbrace> m1 [+] m2 \<lbrace>> Q\<rbrace> = (\<lbrace>P\<rbrace> m1 \<lbrace>> Q\<rbrace> \<and> \<lbrace>P\<rbrace> m2 \<lbrace>> Q\<rbrace>)"
  by(auto simp add: Hoare_def Box_def choice_def)
   
lemma returnN_rule': "\<lbrace>P\<rbrace> returnN x \<lbrace>> K P\<rbrace>"
  by (simp add: Hoare_def Box_return K_def)

lemma returnR_rule': "\<lbrace>P\<rbrace> returnR x \<lbrace>> K P\<rbrace>"
  by (simp add: returnR_def returnN_rule')

lemma return_rule': "\<lbrace>P\<rbrace> return x \<lbrace>> K P\<rbrace>"
  by (simp add: return_def returnR_rule')

lemma throw_rule': "\<lbrace>P\<rbrace> throw x \<lbrace>> K P\<rbrace>"
  by (simp add: throw_def returnR_rule')

lemma compose_inv:
  "\<lbrakk>
    \<lbrace>I\<rbrace> mn \<lbrace>> K I\<rbrace>; \<lbrace>I\<rbrace> me \<lbrace>> K I\<rbrace>; \<lbrace>I\<rbrace> mr \<lbrace>> K I\<rbrace>
  \<rbrakk> \<Longrightarrow>
    \<lbrace>I\<rbrace> compose mn me (mr\<Down>R) \<lbrace>> K I\<rbrace>"
  by(auto simp only: Hoare_def Box_def K_def compose_def project_defs)

lemma Hoare_all_Hoare:
  "\<lbrace>I\<rbrace> m \<lbrace>& I\<rbrace> = \<lbrace>I\<rbrace> m \<lbrace>> K I\<rbrace>"
  apply(auto simp add: K_def Hoare_all_def Hoare_defs Box_def Box_defs
    split: option.split option.split_asm sum.split sum.split_asm)
  apply(case_tac a)
  apply(auto)
  apply(case_tac aa)
  apply(auto)
  done

lemma K_Hoare_to_NN: "\<lbrace> I \<rbrace> m \<lbrace>> K I \<rbrace> \<Longrightarrow> \<lbrace> I \<rbrace> (m\<Down>NN) \<lbrace>> K I\<rbrace>"
  by(auto simp add: project_defs Hoare_def Box_def K_def)

lemma K_Hoare_to_NE: "\<lbrace> I \<rbrace> m \<lbrace>> K I \<rbrace> \<Longrightarrow> \<lbrace> I \<rbrace> (m\<Down>NE) \<lbrace>> K I\<rbrace>"
  by(auto simp add: project_defs Hoare_def Box_def K_def)

lemma K_Hoare_to_R: "\<lbrace> I \<rbrace> m \<lbrace>> K I \<rbrace> \<Longrightarrow> \<lbrace> I \<rbrace> (m\<Down>R) \<lbrace>S I\<rbrace>"
  by(auto simp add: project_defs Hoare_def Box_def HoareS_def BoxS_def K_def)

lemma decompose_HoareN:
  "\<lbrace>P\<rbrace> m \<lbrace>>R K Q\<rbrace> = (\<lbrace>P\<rbrace> m \<lbrace>>N K Q\<rbrace> \<and> \<lbrace>P\<rbrace> m \<lbrace>@N K Q\<rbrace>)"
  apply(auto simp add: Hoare_alt Hoare_def Box_def project_defs K_def)
  apply(case_tac a, auto)
  done

lemma Hoare_choice:
  "\<lbrace>P\<rbrace> m1 [+] m2 \<lbrace>>Q\<rbrace> = (\<lbrace>P\<rbrace> m1 \<lbrace>>Q\<rbrace> \<and> \<lbrace>P\<rbrace> m2 \<lbrace>>Q\<rbrace>)"
  by(auto simp add: Hoare_def Box_choice)

lemma Box_empty:
  "Box empty P = UNIV"
  by(auto simp add: Box_def)

lemma Hoare_empty:
  "\<lbrace>P\<rbrace> empty \<lbrace>>Q\<rbrace>"
  by(auto simp add: Hoare_def Box_empty)

lemma Hoare_if:
  "\<lbrakk> cond \<Longrightarrow> \<lbrace>P\<rbrace> mt \<lbrace>>Q\<rbrace>; \<not>cond \<Longrightarrow> \<lbrace>P\<rbrace> mf \<lbrace>>Q\<rbrace> \<rbrakk>
  \<Longrightarrow> \<lbrace>P\<rbrace> if cond then mt else mf \<lbrace>>Q\<rbrace>"
  by(auto)

lemma HoareNE_if:
  "\<lbrakk> cond \<Longrightarrow> \<lbrace>P\<rbrace> mt \<lbrace>@N Q\<rbrace>; \<not>cond \<Longrightarrow> \<lbrace>P\<rbrace> mf \<lbrace>@N Q\<rbrace> \<rbrakk>
  \<Longrightarrow> \<lbrace>P\<rbrace> if cond then mt else mf \<lbrace>@N Q\<rbrace>"
  by(auto)

lemma Hoare_returnN:
  "\<lbrace>P x\<rbrace> returnN x \<lbrace>> Q\<rbrace> = (P x \<subseteq> Q x)"
  by(auto simp add: Hoare_def Box_def unfold_monad)
  
lemma Hoare_returnN':
  "\<lbrace>P\<rbrace> returnN x \<lbrace>> Q\<rbrace> = (P \<subseteq> Q x)"
  by(auto simp add: Hoare_def Box_def unfold_monad)

lemma HoareNN_return:
  "\<lbrace>P x\<rbrace> return x \<lbrace>>N Q\<rbrace> = (P x \<subseteq> Q x)"
  by(auto simp add: HoareNN_alt Hoare_returnN project_simps)

lemma HoareNN_return':
  "\<lbrace>P\<rbrace> return x \<lbrace>>N Q\<rbrace> = (P \<subseteq> Q x)"
  by(auto simp add: HoareNN_alt Hoare_returnN' project_simps)

lemma HoareNN_throw:
  "\<lbrace>P\<rbrace> throw e \<lbrace>>N Q\<rbrace>"
  by(auto simp add: HoareNN_alt project_simps Hoare_empty)

lemma HoareNE_throw':
  "\<lbrace>P\<rbrace> throw e \<lbrace>@N Q\<rbrace> = (P \<subseteq> Q e)"
  by(auto simp add: HoareNE_alt Hoare_returnN' project_simps)

lemma HoareNE_throw:
  "\<lbrace>P e\<rbrace> throw e \<lbrace>@N Q\<rbrace> = (P e \<subseteq> Q e)"
  by(auto simp add: HoareNE_alt Hoare_returnN project_simps)

lemma HoareNE_return:
  "\<lbrace>P\<rbrace> return x \<lbrace>@N Q\<rbrace>"
  by(auto simp add: HoareNE_alt project_simps Hoare_empty)

lemma bind_ruleNE_same:
  "\<lbrakk>\<lbrace>P\<rbrace> m1 \<lbrace>@N R\<rbrace>; \<lbrace>P\<rbrace> m1 \<lbrace>>N K P\<rbrace>; \<forall>x. \<lbrace>P\<rbrace> m2 x \<lbrace>@N R\<rbrace>\<rbrakk>
    \<Longrightarrow> \<lbrace>P\<rbrace> m1 \<guillemotright>= m2 \<lbrace>@N R\<rbrace>"
  by(auto simp add: K_def intro!: bind_ruleNE)

lemma bind_ruleNN_same:
  "\<lbrakk>\<lbrace>P\<rbrace> m1 \<lbrace>>N K P\<rbrace>; \<forall>x. \<lbrace>P\<rbrace> m2 x \<lbrace>>N R\<rbrace>\<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m1 \<guillemotright>= m2 \<lbrace>>N R\<rbrace>"
  by(auto simp add: K_def intro!: bind_ruleNN)
  
lemma try_catch_ruleNE:
   "\<lbrakk>\<lbrace>P\<rbrace> m \<lbrace>@N Q\<rbrace>; \<forall>x. \<lbrace>Q x\<rbrace> h x \<lbrace>@N R\<rbrace>\<rbrakk>
    \<Longrightarrow> \<lbrace>P\<rbrace> try_catch m h \<lbrace>@N R\<rbrace>
  "
  by(auto simp add: HoareNE_alt project_simps intro!: bindN_ruleN)

lemma try_catch_ruleNN:
  "\<lbrakk>\<lbrace>P\<rbrace> m \<lbrace>>N R\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>@N Q\<rbrace>; \<forall>e. \<lbrace>Q e\<rbrace> h e \<lbrace>>N R\<rbrace>\<rbrakk>
    \<Longrightarrow> \<lbrace>P\<rbrace> try_catch m h \<lbrace>>N R\<rbrace>"
  by(auto simp only: HoareNE_alt HoareNN_alt project_simps Hoare_choice K_def
    intro!: bindN_ruleN)

lemma try_catch_ruleNN_same:
  "\<lbrakk>\<lbrace>P\<rbrace> m \<lbrace>>N R\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>@N K P\<rbrace>; \<forall>e. \<lbrace>P\<rbrace> h e \<lbrace>>N R\<rbrace>\<rbrakk>
    \<Longrightarrow> \<lbrace>P\<rbrace> try_catch m h \<lbrace>>N R\<rbrace>"
  by(auto intro!: try_catch_ruleNN simp add: K_def)

lemma try_catch_ruleNE_same:
  "\<lbrakk>\<lbrace>P\<rbrace> m \<lbrace>@N R\<rbrace>; \<forall>x. \<lbrace>R x\<rbrace> h x \<lbrace>@N R\<rbrace>\<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> <try> m <catch> h \<lbrace>@N R\<rbrace>"
  by(auto simp add: K_def intro!: try_catch_ruleNE)

lemma list_set_induct':
  "P [] \<longrightarrow> (\<forall>x xs. x \<in> set l \<longrightarrow> set xs \<subseteq> set l \<longrightarrow> P xs \<longrightarrow> P (x # xs)) \<longrightarrow> P l"
  apply(induct l rule: list.induct)
  apply(auto)
  done

lemma list_distinct_set_induct':
  "distinct l
  \<longrightarrow> P []
  \<longrightarrow> (\<forall>x xs. x \<in> set l \<longrightarrow> x \<notin> set xs \<longrightarrow> set xs \<subseteq> set l \<longrightarrow> P xs \<longrightarrow> P (x # xs))
  \<longrightarrow> P l"
  apply(induct l rule: list.induct)
  apply(auto)
  done

lemma list_set_induct [case_names Nil Cons]:
  "\<lbrakk> P []; \<And>x xs. \<lbrakk> x \<in> set l; set xs \<subseteq> set l; P xs \<rbrakk> \<Longrightarrow> P (x#xs) \<rbrakk> \<Longrightarrow> P l"
  by(auto intro: list_set_induct'[rule_format])

lemma list_distinct_set_induct [case_names d Nil Cons]:
  "\<lbrakk> distinct l; P []; \<And>x xs. \<lbrakk> x \<in> set l; x \<notin> set xs; set xs \<subseteq> set l;  P xs \<rbrakk> \<Longrightarrow> P (x#xs) \<rbrakk> \<Longrightarrow> P l"
  by(auto intro: list_distinct_set_induct'[rule_format])

lemma ballE':
  "\<lbrakk> \<forall>x \<in> A. P x; x \<in> A \<rbrakk> \<Longrightarrow> P x"
  by(auto)

lemma HoareNN_preserves_Inter:
  assumes
  preserves: "\<forall>x'. x' \<in> A \<longrightarrow> \<lbrace> P x' \<rbrace> m x \<lbrace>>N K (P x')\<rbrace>"
  and not_elem: "x \<notin> A"
  and subset: "A \<subseteq> B"
  shows "\<lbrace>\<Inter>x' \<in> B. P x'\<rbrace> m x \<lbrace>>N K (\<Inter>x' \<in> A. P x')\<rbrace>"
proof-
  show ?thesis using assms
    apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def K_beta projectNN_def)
    apply(smt mem_Collect_eq set_mp)
    done
qed

lemma mapM'_preserves:
  assumes
  preserves: "\<forall>x'. x' \<in> set xs \<longrightarrow> \<lbrace> P x \<rbrace> m x' \<lbrace>>N K (P x)\<rbrace>"
  and not_elem: "x \<notin> set xs"
  shows "\<lbrace>P x\<rbrace> mapM' m xs \<lbrace>>N K (P x)\<rbrace>"
proof(induct rule: list_set_induct)
  case Nil
  show "\<lbrace>P x\<rbrace> mapM' m [] \<lbrace>>N K (P x)\<rbrace>"
    by(simp add: HoareNN_alt Hoare_def Box_def projectNN_def unfold_monad)
next
  case (Cons y ys)
  show "\<lbrace>P x\<rbrace> mapM' m (y # ys) \<lbrace>>N K (P x)\<rbrace>"
  proof(simp only: mapM'.simps, rule bind_ruleNN_same)
    show "\<lbrace>P x\<rbrace> m y \<lbrace>>N K (P x)\<rbrace>" 
      by(rule preserves[rule_format, OF Cons(1)])
  next
    show "\<forall>xa. \<lbrace>P x\<rbrace> mapM' m ys \<lbrace>>N K (P x)\<rbrace>" using Cons
      by auto
  qed
qed

lemma HoareNN_join_post:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>>N Q\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>>N Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>>N \<lambda>x. Q x \<inter> Q' x\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def)

lemma HoareNN_join:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>>N Q\<rbrace>; \<lbrace>P'\<rbrace> m \<lbrace>>N Q'\<rbrace> \<rbrakk> \<Longrightarrow>
    \<lbrace>P \<inter> P'\<rbrace> m \<lbrace>>N \<lambda>v. Q v \<inter> Q' v\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def)

lemma HoareNE_join:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>@N Q\<rbrace>; \<lbrace>P'\<rbrace> m \<lbrace>@N Q'\<rbrace> \<rbrakk> \<Longrightarrow>
    \<lbrace>P \<inter> P'\<rbrace> m \<lbrace>@N \<lambda>v. Q v \<inter> Q' v\<rbrace>"
  by(auto simp add: HoareNE_alt Hoare_def Box_def)

lemma mapM'_ruleNN:
  assumes
  establishesQ: "\<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>>N K (Q x)\<rbrace>"
  and preservesP: "\<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>>N K P\<rbrace>"
  and preservesQ: "\<forall>x x'. x \<in> set l \<and> x' \<in> set l \<and> x \<noteq> x' \<longrightarrow> \<lbrace>Q x'\<rbrace> m x \<lbrace>>N K (Q x')\<rbrace>"
  shows "\<lbrace>P\<rbrace> mapM' m l \<lbrace>>N K (\<Inter>x'\<in> set l. Q x') \<rbrace>"
proof(induct l rule: list_set_induct)
  show "\<lbrace>P\<rbrace> mapM' m [] \<lbrace>>N K (\<Inter>x'\<in>set []. Q x')\<rbrace>"
    by(auto simp add: HoareNN_alt Hoare_def Box_def unfold_monad)
next
  case (Cons x xs)
  have "\<lbrace>P\<rbrace> m x \<lbrace>>N K P\<rbrace>" using `x \<in> set l` preservesP by blast
  moreover have "\<lbrace>P\<rbrace> m x \<lbrace>>N K (Q x)\<rbrace>" using `x \<in> set l` establishesQ by blast
  ultimately have PandQ: "\<lbrace>P\<rbrace> m x \<lbrace>>N K (P \<inter> Q x)\<rbrace>"
    by(auto simp add: K_def intro!: HoareNN_join_post)
  show "\<lbrace>P\<rbrace> mapM' m (x # xs) \<lbrace>>N K (\<Inter>x'\<in>set (x # xs). Q x')\<rbrace>"
  proof(subst mapM'.simps, intro bind_ruleNN[OF PandQ] allI, subst K_beta)
    show "\<lbrace>P \<inter> Q x\<rbrace> mapM' m xs \<lbrace>>N K (\<Inter>x' \<in> set (x # xs). Q x')\<rbrace>"
    proof(cases "x \<in> set xs")
      case True
      have "\<lbrace>P\<rbrace> mapM' m xs \<lbrace>>N K (\<Inter>x' \<in> set (x # xs). Q x')\<rbrace>"
        by (metis Cons(3) True remdups.simps(2) set_remdups)
      thus ?thesis by (blast intro: conseqNN_rule)
    next
      case False
      have "\<lbrace>Q x\<rbrace> mapM' m xs \<lbrace>>N K (Q x)\<rbrace>"
        by (metis (no_types) Cons(1) Cons(2) False mapM'_preserves preservesQ set_rev_mp)
      thus ?thesis using Cons
        apply(clarsimp simp add: K_def HoareNN_alt Hoare_def Box_def project_defs)
        apply(smt CollectD set_mp)
        done
    qed
  qed
qed
  
lemma mapM'_distinct_ruleNN:
  assumes
  distinct: "distinct l"
  and establishes: "\<forall>x \<in> set l. \<lbrace>P x\<rbrace> m x \<lbrace>>N K (Q x)\<rbrace>"
  and preservesP: "\<forall>x x'. x \<in> set l \<and> x' \<in> set l \<and> x \<noteq> x' \<longrightarrow> \<lbrace> P x \<rbrace> m x' \<lbrace>>N K (P x)\<rbrace>"
  and preservesQ: "\<forall>x x'. x \<in> set l \<and> x' \<in> set l \<and> x \<noteq> x' \<longrightarrow> \<lbrace> Q x \<rbrace> m x' \<lbrace>>N K (Q x)\<rbrace>"
  shows "\<lbrace>\<Inter>x'\<in> set l. P x'\<rbrace> mapM' m l \<lbrace>>N K (\<Inter>x'\<in> set l. Q x') \<rbrace>"
proof(induct l rule: list_distinct_set_induct)
  show "distinct l" using distinct .
next
  case Nil
  show "\<lbrace>\<Inter>x' \<in> set []. P x'\<rbrace> mapM' m [] \<lbrace>>N K (\<Inter>x' \<in> set []. Q x')\<rbrace>"
    by(auto simp add: HoareNN_alt Hoare_def Box_def unfold_monad)
next
  case (Cons x xs)
  have Qx: "\<lbrace>P x\<rbrace> m x \<lbrace>>N K (Q x)\<rbrace>" using Cons establishes
    by(blast)
  show "\<lbrace>\<Inter>x'\<in>set(x # xs). P x'\<rbrace> mapM' m (x # xs) \<lbrace>>N K (\<Inter>x'\<in>set (x#xs). Q x')\<rbrace>"
  proof(simp only: mapM'.simps)
    have Qx: "\<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<guillemotright>= (\<lambda>_. mapM' m xs) \<lbrace>>N K (Q x)\<rbrace>"
    proof(intro bind_ruleNN allI ballI)
      show "\<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<lbrace>>N K (Q x)\<rbrace>"
        by(auto intro!: conseqNN_rule[OF _ Qx])
    next
      fix x'
      show "\<And>x'. \<lbrace>K (Q x) x'\<rbrace> mapM' m xs \<lbrace>>N K (Q x)\<rbrace>"
        apply(simp only: K_beta)
        apply(intro mapM'_preserves Cons preservesQ)
        apply (metis Cons(1) Cons(2) Cons(3) preservesQ set_mp)
        done
    qed
    have "\<forall>x''\<in>set xs. \<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<guillemotright>= (\<lambda>_. mapM' m xs) \<lbrace>>N K (Q x'')\<rbrace>"
    proof(intro bind_ruleNN allI ballI)
      show "\<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<lbrace>>N K (\<Inter>x\<in>set xs. P x)\<rbrace>"
        apply(intro HoareNN_preserves_Inter Cons preservesP, auto)
        apply (metis Cons(1) Cons(2) Cons(3) preservesP set_mp)
        done
    next
      fix z x'
      assume "x' \<in> set xs"
      hence post: "\<lbrace>\<Inter>x'\<in>set xs. P x'\<rbrace> mapM' m xs \<lbrace>>N K (\<Inter>x'\<in>set xs. Q x')\<rbrace>" using Cons
        by blast
      have "\<forall>x. K (\<Inter>x'\<in>set xs. Q x') x \<subseteq> K (Q x') x" using `x' \<in> set xs`
        by(auto simp add: K_beta)
      note conseq = conseqNN_rule[OF _ _ this]
      hence "\<lbrace>\<Inter>x'\<in>set xs. P x'\<rbrace> mapM' m xs \<lbrace>>N K (Q x')\<rbrace>" using conseq[OF order_refl post]
        by auto
      thus "\<lbrace>K (\<Inter>x\<in>set xs. P x) z\<rbrace> mapM' m xs \<lbrace>>N K (Q x')\<rbrace>"
        by(simp only: K_beta)
    qed
    hence Qxs: "\<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<guillemotright>= (\<lambda>_. mapM' m xs) \<lbrace>>N K (\<Inter>x' \<in> set xs. Q x')\<rbrace>"
      by(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
    show "\<lbrace>\<Inter>x\<in>set (x # xs). P x\<rbrace> m x \<guillemotright>= (\<lambda>_. mapM' m xs) \<lbrace>>N K (\<Inter>x' \<in> set (x#xs). Q x')\<rbrace>"
      using HoareNN_join_post[OF Qx Qxs]
      by(auto simp add: K_def)
  qed
qed

lemma mapM'_ruleNE:
  "\<lbrakk> \<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>@N Q\<rbrace>; \<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>>N K P\<rbrace> \<rbrakk> \<Longrightarrow> 
    \<lbrace>P\<rbrace> mapM' m l \<lbrace>@N Q\<rbrace>"
  apply(induct rule: list_set_induct)
  apply(auto intro!: HoareNE_return bind_ruleNE simp add: K_def)
  done

lemma mapM'_invNN:
  "\<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>>N K P\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> mapM' m l \<lbrace>>N K P \<rbrace>"
  apply(induct rule: list_set_induct)
  apply(auto simp add: HoareNN_return' K_beta intro!: bind_ruleNN_same)
  done

lemma mapM'_rule:
  "\<forall>x \<in> set l. \<lbrace>P\<rbrace> m x \<lbrace>> K P\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> mapM' m l \<lbrace>> K P \<rbrace>"
  apply(induct rule: list_set_induct)
  apply(auto simp add: K_beta intro!: bind_inv_same return_rule')
  done

lemma HoareNN_choice: "\<lbrace>P\<rbrace> m1 [+] m2 \<lbrace>>N Q\<rbrace> = (\<lbrace>P\<rbrace> m1 \<lbrace>>N Q\<rbrace> \<and> \<lbrace>P\<rbrace> m2 \<lbrace>>N Q\<rbrace>)"
  by(clarsimp simp add: HoareNN_alt project_simps choice_inv)
 
lemma HoareNE_choice: "\<lbrace>P\<rbrace> m1 [+] m2 \<lbrace>@N Q\<rbrace> = (\<lbrace>P\<rbrace> m1 \<lbrace>@N Q\<rbrace> \<and> \<lbrace>P\<rbrace> m2 \<lbrace>@N Q\<rbrace>)"
  by(clarsimp simp add: HoareNE_alt project_simps choice_inv)
 
lemma HoareNN_Int:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>>N K P'\<rbrace>; \<lbrace>Q\<rbrace> m \<lbrace>>N K Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P \<inter> Q\<rbrace> m \<lbrace>>N K (P' \<inter> Q')\<rbrace>"
  by(force simp add: HoareNN_alt Hoare_def Box_def K_def)

lemma chooseN_inv_NN:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>>N K P\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def chooseN_def unfold_monad choose_def project_defs)

lemma chooseN_ruleNN:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>>N \<lambda>x. if x \<in> A then P else {}\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def chooseN_def unfold_monad choose_def project_defs)

lemma chooseN_ruleNE:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>@N Q\<rbrace>"
  by(auto simp add: HoareNE_alt Hoare_def Box_def chooseN_def choose_def project_defs unfold_monad)

lemma chooseN_inv:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>> K P\<rbrace>"
  by(auto simp add: Hoare_def Box_def chooseN_def choose_def unfold_monad)

end

