theory Infinite
imports Main
begin

class infinite =
  assumes infinite_UNIV: "\<not> finite (UNIV :: 'a set)"
begin
lemma infinite[simp]: "\<not> finite (UNIV :: 'a set)"
  by(rule infinite_UNIV)
end

lemma cons_inj: "inj (Cons x)"
  by(rule injI, auto)

lemma not_finite_list: "\<not> finite (UNIV :: 'a list set)"
proof
  fix x
  assume fin: "finite (UNIV :: 'a list set)"
  with finite_UNIV_inj_surj[of "\<lambda>l. x # l"] 
  show False
    by(auto simp add: cons_inj dest: surjD)
qed

instance list :: (type) infinite
proof qed (simp add: not_finite_list)

lemma inj_insert:
  "\<lbrakk> inj_on f A; x \<notin> A; f x \<notin> f ` A \<rbrakk> \<Longrightarrow> inj_on f (insert x A)"
  by (metis Diff_subset image_mono inj_on_insert subsetD)

lemma fin_to_inf_inj: "\<exists>f :: ('a :: finite \<Rightarrow> 'b :: infinite). inj f"
proof-
  fix h :: "'a \<Rightarrow> 'b"
    and some_el_a :: 'a
    and some_el_b :: 'b
  have finite_UNIV: "finite (UNIV :: 'a set)" by simp
  have infinite_UNIV: "\<not> finite (UNIV :: 'b set)" by simp
  def f \<equiv> "\<lambda>(x :: 'a) (h :: 'a \<Rightarrow> 'b, S :: 'a set). if h x \<notin> h ` S
                then (h, insert x S)
                else (h(x := (SOME y. y \<in> UNIV - h ` S)), insert x S)"
  {
    fix S res
    assume folded: "fold_graph f (h, {}) S res"
    hence "snd res = S"
      by(rule fold_graph.induct) (auto simp add: f_def split: split_if_asm)
  } note fold_set = this[rule_format]
  {
    fix A::"'a set" and g :: "'a \<Rightarrow> 'b"
    have "finite (f ` A)"
      by(auto)
    hence "\<exists>y. y \<notin> g ` A"
      by(auto intro!: ex_new_if_finite)
  } note choice_fresh=this
  {
    fix x y S
    assume "x \<in> S" "y \<notin> S"
    hence "x \<noteq> y"
      by(auto)
  } note diff_set_diff_el=this[rule_format]
  {
    fix S res
    assume folded: "fold_graph f (h, {}) S res"
    hence "inj_on (fst res) S"
    proof(induct rule: fold_graph.induct)
      case emptyI
      show ?case
        by simp
    next
      case insertI
      fix nx A res
      assume nx_nmem: "nx \<notin> A" and fold: "fold_graph f (h, {}) A res" 
        and inj: "inj_on (fst res) A"
      obtain g and S where res_rep: "res = (g, S)" by(case_tac res, auto)
      hence snd_A: "S = A" using fold
        by(auto dest!: fold_set)
      from choice_fresh[of g] obtain ny where ny_out: "ny \<notin> g ` A" by auto
      have inj2: "inj_on (fst (f nx res)) A" using nx_nmem inj res_rep
        by(auto simp add: f_def inj_on_def split: split_if_asm)
      have res_nmem: "fst (f nx res) nx \<notin> fst (f nx res) ` A" using nx_nmem snd_A res_rep ny_out
        by(auto simp add: f_def split: split_if_asm dest: someI)
      have "inj_on (fst (f nx (g, S))) (insert nx A)" using inj2 res_nmem
        by (metis inj_insert res_rep rev_image_eqI)
      thus "inj_on (fst (f nx res)) (insert nx A)" using res_rep 
        by simp
    qed
  } note inj=this[rule_format]

  from finite_UNIV[THEN finite_imp_fold_graph, of f "(h, {})"] obtain res where
    "fold_graph f (h, {}) UNIV res" by auto
  from inj[OF this] show ?thesis  by auto

qed

lemma infinite_has_two': "\<forall>cnt' :: 'a. cnt' = cnt \<Longrightarrow> finite (UNIV :: 'a set)"
proof-
  fix cnt :: 'a
  def U \<equiv> "UNIV :: 'a set" (* seems to be the only way to disambiguate UNIV? *)
  assume "\<forall>cnt'. cnt' = cnt"
  hence "U = insert cnt {}" using U_def
    by auto 
  hence "finite U"
    by(force intro!: Finite_Set.finite.insertI)
  thus ?thesis using U_def
    apply(auto)
    done
qed

lemma infinite_has_two: "\<exists>cnt' :: 'a :: infinite. cnt' \<noteq> cnt"
  apply(insert infinite_UNIV)
  apply(erule contrapos_np)
  apply(blast intro!: infinite_has_two')
  done

end
