theory Inst_State
imports Main
begin

datatype log_state = LDead | LDegraded | LOK

instantiation log_state :: "{bot, top, linorder}"
begin

fun less_eq_log_state :: "log_state \<Rightarrow> log_state \<Rightarrow> bool" where
  "less_eq_log_state LDead  _ = True"
  | "less_eq_log_state _ LDead = False"
  | "less_eq_log_state LDegraded _ = True" 
  | "less_eq_log_state _ LDegraded = False" 
  | "less_eq_log_state LOK LOK = True" 

definition bot_log_state :: log_state where
[simp]: "bot_log_state \<equiv> LDead"

definition top_log_state :: log_state where
[simp]: "top_log_state \<equiv> LOK"

definition less_log_state :: "log_state \<Rightarrow> log_state \<Rightarrow> bool" where
[simp]: "less_log_state x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: log_state
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: log_state
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: log_state
  show "bot \<le> a"
    by(simp)
next
  fix a :: log_state
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

datatype ec_state = EcDead | EcUnwr | EcBad | EcOK 

instantiation ec_state :: "{bot, top, linorder}"
begin

fun less_eq_ec_state :: "ec_state \<Rightarrow> ec_state \<Rightarrow> bool" where
  "less_eq_ec_state EcDead _ = True"
  | "less_eq_ec_state _ EcDead = False"
  | "less_eq_ec_state EcUnwr _ = True" 
  | "less_eq_ec_state _ EcUnwr = False" 
  | "less_eq_ec_state EcBad _ = True" 
  | "less_eq_ec_state _ EcBad = False" 
  | "less_eq_ec_state EcOK EcOK = True" 

definition bot_ec_state :: ec_state where
[simp]: "bot_ec_state \<equiv> EcDead"

definition top_ec_state :: ec_state where
[simp]: "top_ec_state \<equiv> EcOK"

definition less_ec_state :: "ec_state \<Rightarrow> ec_state \<Rightarrow> bool" where
[simp]: "less_ec_state x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: ec_state
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: ec_state
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: ec_state
  show "bot \<le> a"
    by(simp)
next
  fix a :: ec_state
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

lemma EcDead_bot: "EcDead = bot"
  by(simp)

lemma min_is_EcOK: "(min x y = EcOK) = (x = EcOK \<and> y = EcOK)"
  by(case_tac x, case_tac[!] y, auto simp add: min_def)

lemma less_eq_EcDead: "(x \<le> EcDead) = (x = EcDead)"
  by(simp only: EcDead_bot bot_unique)

lemma less_eq_EcUnwr: "(x \<le> EcUnwr) = (x = EcDead \<or> x = EcUnwr)"
  by(case_tac x, auto)

lemma less_eq_EcBad: "(x \<le> EcBad) = (x = EcDead \<or> x = EcUnwr \<or> x = EcBad)"
  by(case_tac x, auto)

lemma EcBad_less_eq: "(EcBad \<le> x) = (x = EcBad \<or> x = EcOK)"
  by(case_tac x, auto)

lemma EcOK_less_eq: "(EcOK \<le> x) = ( x = EcOK)"
  by(case_tac x, auto)

lemma exhaust_EcOK: "\<lbrakk> x \<noteq> EcDead; x \<noteq> EcUnwr; x \<noteq> EcBad \<rbrakk> \<Longrightarrow> x = EcOK"
  by(case_tac x, auto)

lemma exhaust_EcUnwr: "\<lbrakk> x \<noteq> EcDead; x \<noteq> EcBad; x \<noteq> EcOK \<rbrakk> \<Longrightarrow> x = EcUnwr"
  by(case_tac x, auto)

lemmas ec_state_simps = 
  min_is_EcOK
  less_eq_EcDead
  less_eq_EcUnwr
  EcBad_less_eq
  EcOK_less_eq
  less_eq_EcBad
  exhaust_EcOK
  exhaust_EcUnwr

fun combine_ec_states :: "ec_state \<Rightarrow> ec_state \<Rightarrow> log_state" where
  "combine_ec_states EcOK EcOK = LOK"
  | "combine_ec_states EcOK EcBad = LOK"
  | "combine_ec_states EcBad EcOK = LOK"
  | "combine_ec_states EcOK EcUnwr = LDegraded"
  | "combine_ec_states EcOK EcDead = LDegraded"
  | "combine_ec_states EcUnwr EcOK  = LDegraded"
  | "combine_ec_states EcDead EcOK  = LDegraded"
  | "combine_ec_states _ _  = LDead"

datatype hw_state = IDead | IUnwr | IOK 

instantiation hw_state :: "{bot, top, linorder}"
begin

fun less_eq_hw_state :: "hw_state \<Rightarrow> hw_state \<Rightarrow> bool" where
  "less_eq_hw_state IDead _ = True"
  | "less_eq_hw_state _ IDead = False"
  | "less_eq_hw_state IUnwr _ = True" 
  | "less_eq_hw_state _ IUnwr = False" 
  | "less_eq_hw_state IOK IOK = True" 

definition bot_hw_state :: hw_state where
[simp]: "bot_hw_state \<equiv> IDead"

definition top_hw_state :: hw_state where
[simp]: "top_hw_state \<equiv> IOK"

definition less_hw_state :: "hw_state \<Rightarrow> hw_state \<Rightarrow> bool" where
[simp]: "less_hw_state x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: hw_state
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: hw_state
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: hw_state
  show "bot \<le> a"
    by(simp)
next
  fix a :: hw_state
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

lemma IDead_bot: "IDead = bot"
  by(simp)

lemma min_is_IOK: "(min x y = IOK) = (x = IOK \<and> y = IOK)"
  by(case_tac x, case_tac[!] y, auto simp add: min_def)

lemma less_eq_IDead: "(x \<le> IDead) = (x = IDead)"
  by(simp only: IDead_bot bot_unique)

lemma less_eq_IUnwr: "(x \<le> IUnwr) = (x = IDead \<or> x = IUnwr)"
  by(case_tac x, auto)

lemma IOK_less_eq: "(IOK \<le> x) = ( x = IOK)"
  by(case_tac x, auto)

lemma exhaust_IOK: "\<lbrakk> x \<noteq> IDead; x \<noteq> IUnwr  \<rbrakk> \<Longrightarrow> x = IOK"
  by(case_tac x, auto)

lemma exhaust_IUnwr: "\<lbrakk> x \<noteq> IDead; x \<noteq> IOK \<rbrakk> \<Longrightarrow> x = IUnwr"
  by(case_tac x, auto)

text {* The flags are @{term "unwritable inst"} and @{term "dead inst"}, in that order *}
fun hw_state_by_flags :: "bool \<Rightarrow> bool \<Rightarrow> hw_state" where
  "hw_state_by_flags _ True = IDead"
  | "hw_state_by_flags True _ = IUnwr"
  | "hw_state_by_flags _ _ = IOK"

lemma hw_state_by_flags_simps:
  "(hw_state_by_flags fu fd = IDead) = fd"
  "(hw_state_by_flags fu fd = IUnwr) = (fu \<and> \<not> fd)"
  "(hw_state_by_flags fu fd = IOK) = (\<not>fu \<and> \<not> fd)"
  "(IDead = hw_state_by_flags fu fd) = fd"
  "(IUnwr = hw_state_by_flags fu fd) = (fu \<and> \<not> fd)"
  "(IOK = hw_state_by_flags fu fd) = (\<not>fu \<and> \<not> fd)"
  by(case_tac[!] fu, case_tac[!] fd, simp_all)

lemmas hw_state_simps = 
  min_is_IOK
  less_eq_IDead
  less_eq_IUnwr
  IOK_less_eq
  exhaust_IOK
  exhaust_IUnwr
  hw_state_by_flags_simps

datatype finst_state = FIDead | FIUnwr | FIUnwrOK | FIBad | FIOK 

instantiation finst_state :: "{bot, top, linorder}"
begin

fun less_eq_finst_state :: "finst_state \<Rightarrow> finst_state \<Rightarrow> bool" where
  "less_eq_finst_state FIDead _ = True"
  | "less_eq_finst_state _ FIDead = False"
  | "less_eq_finst_state FIUnwr _ = True" 
  | "less_eq_finst_state _ FIUnwr = False" 
  | "less_eq_finst_state FIUnwrOK _ = True" 
  | "less_eq_finst_state _ FIUnwrOK = False" 
  | "less_eq_finst_state FIBad _ = True" 
  | "less_eq_finst_state _ FIBad = False" 
  | "less_eq_finst_state FIOK FIOK = True" 

definition bot_finst_state :: finst_state where
[simp]: "bot_finst_state \<equiv> FIDead"

definition top_finst_state :: finst_state where
[simp]: "top_finst_state \<equiv> FIOK"

definition less_finst_state :: "finst_state \<Rightarrow> finst_state \<Rightarrow> bool" where
[simp]: "less_finst_state x y \<equiv> x \<le> y \<and> \<not> y \<le> x"

instance proof
  fix x :: finst_state
  show "x \<le> x"
    by(case_tac x, simp_all)
next
  fix x y z :: finst_state
  assume "x \<le> y" "y \<le> z"
  thus "x \<le> z"
    by(cases x, case_tac[!] y, case_tac[!] z, simp_all)
next
  fix a :: finst_state
  show "bot \<le> a"
    by(simp)
next
  fix a :: finst_state
  show "a \<le> top"
    by(case_tac a , simp_all)
qed(case_tac[!] x, case_tac [!] y, simp_all)

end

lemma FIDead_bot: "FIDead = bot"
  by(simp)

lemma min_is_FIOK: "(min x y = FIOK) = (x = FIOK \<and> y = FIOK)"
  by(case_tac x, case_tac[!] y, auto simp add: min_def)

lemma less_eq_FIDead: "(x \<le> FIDead) = (x = FIDead)"
  by(simp only: FIDead_bot bot_unique)

lemma less_eq_FIUnwr: "(x \<le> FIUnwr) = (x = FIDead \<or> x = FIUnwr)"
  by(case_tac x, auto)

lemma less_eq_FIBad: "(x \<le> FIBad) = (x = FIDead \<or> x = FIUnwr \<or> x = FIUnwrOK \<or> x = FIBad)"
  by(case_tac x, auto)

lemma FIBad_less_eq: "(FIBad \<le> x) = (x = FIBad \<or> x = FIOK)"
  by(case_tac x, auto)

lemma FIOK_less_eq: "(FIOK \<le> x) = ( x = FIOK)"
  by(case_tac x, auto)

lemma exhaust_FIOK: "\<lbrakk> x \<noteq> FIDead; x \<noteq> FIUnwr; x \<noteq> FIUnwrOK; x \<noteq> FIBad \<rbrakk> \<Longrightarrow> x = FIOK"
  by(case_tac x, auto)

lemma exhaust_FIUnwr: "\<lbrakk> x \<noteq> FIDead; x \<noteq> FIUnwrOK; x \<noteq> FIBad; x \<noteq> FIOK \<rbrakk> \<Longrightarrow> x = FIUnwr"
  by(case_tac x, auto)

lemmas finst_state_simps = 
  min_is_FIOK
  less_eq_FIDead
  less_eq_FIUnwr
  FIBad_less_eq
  FIOK_less_eq
  less_eq_FIBad
  exhaust_FIOK
  exhaust_FIUnwr


end