theory PersistencyLemmas
imports Persistency
begin

text {*

  Simplification rule #1: BoxR and volatile-memory-only computations

*}

lemma vi_BoxR: "vol_invP P \<Longrightarrow> P \<subseteq> [@R liftP vi] P"
  by(auto simp add: liftP_def BoxR_def vol_invP_alter)

text {* 

  Volatile-memory-only computation on the LHS
  of a bind.

*}

lemma vi1: "vol_invP P \<Longrightarrow> P \<inter> [R x \<leftarrow> liftP vi][@R choice (mp x) restart] P \<subseteq> 
  [@R doR { x \<leftarrow> liftP vi; choice (mp x) restart }] P"
  by(auto simp only: BoxR_bind vi_BoxR dest!: vi_BoxR)

text {*
  
  Computations that fail give rise to trivial conditions within the
  BoxR/N definitions, so we have to add a nofail assumption for the
  other side of the inclusion.

*}

lemma vi2: "\<lbrakk> vol_invP P; nofail vi \<rbrakk> \<Longrightarrow> 
  [@R doR { x \<leftarrow> liftP vi; choice (mp x) restart }] P
  \<subseteq>
  P \<inter> [R x \<leftarrow> liftP vi][@R choice (mp x) restart] P"
  apply(simp add: BoxR_bind)
  apply(auto simp add: liftP_def BoxN_def BoxR_def vol_invP_alter choice_def restart_def 
    returnN_def nofail_def)
  apply(rename_tac sv sp)
  apply(subgoal_tac "\<exists>x. x \<in> vi sv")
  defer
  apply(blast)
  apply(clarify)
  apply(rename_tac r sv')
  apply(case_tac r)
  apply(erule_tac x="sv'" in allE)
  back
  apply(erule_tac x="sp" in allE)
  back
  apply(drule mp)
  apply(force)
  apply(blast)
  apply(erule_tac x="a" in allE)
  apply(erule_tac x="sv'" in allE)
  back
  back
  apply(erule_tac x="sp" in allE)
  back
  apply(drule mp)
  apply(force)
  apply(erule_tac x="sv'" in allE)
  back
  back
  apply(erule_tac x="sp" in allE)
  back
  apply(clarsimp)
  apply(force)
  done

lemma "\<lbrakk> vol_invP P; nofail vi \<rbrakk> \<Longrightarrow> 
  [@R doR { x \<leftarrow> liftP vi; choice (mp x) restart }] P
  =  
  P \<inter> [R x \<leftarrow> liftP vi][@R choice (mp x) restart] P"
  apply(rule equalityI)
  apply(rule vi2)
  apply(simp)
  apply(simp)
  apply(rule vi1) 
  apply(simp)
  done


text {* Simplification rule for the RHS of bind *}
lemma vi_a1: "vol_invP P \<Longrightarrow> 
  [@R doR { x \<leftarrow> mp; restart}] P 
  \<subseteq> 
  [@R doR { x \<leftarrow> mp; choice (liftP (vi x)) restart }] P"
  apply(simp only: BoxR_bind)
  apply(force simp add: liftP_def BoxN_def BoxR_def vol_invP_alter choice_def restart_def returnN_def)
  done

lemma vi_a2: "\<lbrakk> vol_invP P \<rbrakk> \<Longrightarrow> 
  [@R doR { x \<leftarrow> mp; choice (liftP (vi x)) restart }] P
  \<subseteq>  
  [@R doR { x \<leftarrow> mp; restart}] P"
  apply(simp add: BoxR_bind)
  apply(auto simp add: liftP_def BoxN_def BoxR_def vol_invP_alter choice_def restart_def 
    returnN_def nofail_def)
  apply(blast)
  done



end
