theory Pmm_Refinement
imports PmmDefs Refinement HoareLogic Pmm_Tacs
begin

ML_file "ref_thms.ML" 
ML_file "debug_match.ML"

attribute_setup ref_thm =
  {* Attrib.add_del Ref_Thms.add Ref_Thms.del *}
  "refinement proofs"

definition bool_into_set :: "bool \<Rightarrow> 'a set" where
  "bool_into_set b \<equiv> {x. b}"

lemma push_sim_norm_ext_premise:
  "(P \<longrightarrow> sim_norm_ext R ma mc Sn Se) = (sim_norm_ext (R \<inter> bool_into_set P) ma mc Sn Se)"
  apply(auto simp add: bool_into_set_def intro!: sim_exts_empty_rel)
  done

lemma push_sim_ext_premise:
  "(P \<longrightarrow> sim_ext R ma mc Sn Se Sr) = (sim_ext (R \<inter> bool_into_set P) ma mc Sn Se Sr)"
  apply(auto simp add: bool_into_set_def intro!: sim_exts_empty_rel)
  done

lemmas push_sim_exts_premise = 
  push_sim_ext_premise
  push_sim_norm_ext_premise

ML_file "ref_tac.ML"

lemma parallel_exc_cases:
  "\<lbrakk> P ha1 hc1; 
    P ha2 hc2; 
    P ha3 hc3; 
    P ha4 hc4; 
    \<forall>e. P (ha7 e) (hc7 e); 
    P ha8 hc8; 
    P ha9 hc9 
  \<rbrakk> \<Longrightarrow>
  P
    (exceptions_case ha1 ha2 ha3 ha4 ha7 ha8 ha9 e)
    (exceptions_case hc1 hc2 hc3 hc4 hc7 hc8 hc9 e)
  "
  by(case_tac e, auto)

lemmas 
  simS_wpc_exceptions_caseI = parallel_exc_cases[where P="\<lambda>ma mc. R \<subseteq> simS_wpc ma mc S"] for R S

(* Alternatively: *)
thm parallel_exc_cases[where P="\<lambda>ma mc. R \<subseteq> simS_wpc ma mc S", standard]

lemmas 
  simN_wpc_exceptions_caseI = parallel_exc_cases[where P="\<lambda>ma mc. R \<subseteq> simN_wpc ma mc S"] for R S

lemmas sim_norm_exceptions_caseI = parallel_exc_cases[where P="sim_norm R"] for R
lemmas sim_normN_exceptions_caseI = parallel_exc_cases[where P="sim_normN R"] for R
lemmas sim_normE_exceptions_caseI = parallel_exc_cases[where P="sim_normE R"] for R
lemmas sim_exceptions_caseI = parallel_exc_cases[where P="sim R"] for R

lemmas sims_exceptions_caseI =
  sim_exceptions_caseI
  sim_norm_exceptions_caseI
  sim_normN_exceptions_caseI
  sim_normE_exceptions_caseI

lemmas sim_ext_exceptions_caseI = parallel_exc_cases[where 
  P="\<lambda>ma mc. sim_ext R ma mc Sn Se Sr"] for R Sn Se Sr
lemmas sim_norm_ext_exceptions_caseI = parallel_exc_cases[where 
  P="\<lambda>ma mc. sim_norm_ext R ma mc Sn Se"] for R Sn Se

lemma sim_norm_exceptions_case_bindS_I:
  "\<lbrakk>
    sim_norm R (\<lambda>s. (ha1\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc1\<Down>R) s \<guillemotright>=S mc);
    sim_norm R (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S mc);
    sim_norm R (\<lambda>s. (ha3\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc3\<Down>R) s \<guillemotright>=S mc);
    sim_norm R (\<lambda>s. (ha4\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc4\<Down>R) s \<guillemotright>=S mc);
    \<forall>a. sim_norm R (\<lambda>s. ((ha7 a)\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. ((hc7 a)\<Down>R) s \<guillemotright>=S mc);
    sim_norm R (\<lambda>s. (ha8\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc8\<Down>R) s \<guillemotright>=S mc);
    sim_norm R (\<lambda>s. (ha9\<Down>R) s \<guillemotright>=S ma) (\<lambda>s. (hc9\<Down>R) s \<guillemotright>=S mc)
  \<rbrakk> \<Longrightarrow>
    sim_norm R
    (\<lambda>s. ((exceptions_case ha1 ha2 ha3 ha4 ha7 ha8 ha9 e)\<Down>R) s \<guillemotright>=S ma)
    (\<lambda>s. ((exceptions_case hc1 hc2 hc3 hc4 hc7 hc8 hc9 e)\<Down>R) s \<guillemotright>=S mc)
  "
  by(case_tac e, simp_all)

(*
    This would yield a different conclusion:
lemmas sim_norm_exceptions_case_bindS_I = 
  parallel_exc_cases[where P="\<lambda>ha hc. 
        sim_norm R (\<lambda>s. ha s \<guillemotright>=S ma) (\<lambda>s. hc s \<guillemotright>=S mc)"] for R ma mc
*)

(*
lemmas sim_norm_ext_exceptions_case_bindS_I = 
  parallel_exc_cases[where P="\<lambda>ha hc. 
        sim_norm_ext R (\<lambda>s. ha s \<guillemotright>=S ma) (\<lambda>s. hc s \<guillemotright>=S mc) Sn Se"] for R Sn Se ma mc
    *)

lemmas exceptions_case_inv_same = exc_case_pred[where P="\<lambda>m. \<lbrace>P\<rbrace> m \<lbrace>> K P\<rbrace>"] for P

lemma sim_ext_return_diff:
  "sim_ext R (return a) (return b)
    Sn
    Se
    Sr = (R \<subseteq> (Sn a b))"
  by(auto simp add: sim_ext_def simN_wpc_empty_conc simS_wpc_empty_conc simN_wpc_returnN project_simps)

lemma sim_ext_throw_diff:
  "sim_ext R (throw a) (throw b)
    Sn
    Se
    Sr = (R \<subseteq> (Se a b))"
  by(auto simp add: sim_ext_def simN_wpc_empty_conc simS_wpc_empty_conc simN_wpc_returnN project_simps)

lemma sim_seqI:
  "\<lbrakk> sim R ma1 mc1; sim R ma2 mc2  \<rbrakk>
     \<Longrightarrow> sim R (ma1 \<guillemotright> ma2) (mc1 \<guillemotright> mc2)"
  by(rule sims_bindI, simp_all)

lemma sim_norm_seqI:
  "\<lbrakk> sim_norm R ma1 mc1; sim_norm R ma2 mc2 \<rbrakk>
     \<Longrightarrow> sim_norm R (ma1 \<guillemotright> ma2) (mc1 \<guillemotright> mc2)"
  by(rule sims_bindI, simp_all)

lemma sim_normN_seqI:
  "\<lbrakk> sim_normN R ma1 mc1; sim_normN R ma2 mc2 \<rbrakk>
     \<Longrightarrow> sim_normN R (ma1 \<guillemotright> ma2) (mc1 \<guillemotright> mc2)"
  by(rule sims_bindI, simp_all)

lemma sim_normE_seqI:
  "\<lbrakk> sim_normN R ma1 mc1; sim_normE R ma1 mc1; sim_normE R ma2 mc2 \<rbrakk>
     \<Longrightarrow> sim_normE R (ma1 \<guillemotright> ma2) (mc1 \<guillemotright> mc2)"
  by(rule sims_bindI, simp_all)

lemmas sim_if_I' = sim_if_I[rule_format]

lemma sim_if_no_cond:
  "\<lbrakk>sim R maT mcT; sim R maF mcF\<rbrakk> \<Longrightarrow>
    sim R (if cond then maT else maF) (if cond then mcT else mcF)"
  by(simp)

(* abs1-relevant lemmas
  TODO: weed out unnecessarry stuff

  *)
lemmas sim_ext_return_diffI = sim_ext_return_diff[THEN iffD2]
lemmas sim_ext_throw_diffI = sim_ext_throw_diff[THEN iffD2]

lemma sim_ext_return_diff':
  "sim_ext (Sn a b) (return a) (return b) Sn Se Sr"
  by(simp add: sim_ext_return_diff)

lemma sim_ext_throw_diff':
  "sim_ext (Se a b) (throw a) (throw b) Sn Se Sr"
  by(simp add: sim_ext_throw_diff)

lemma sim_sim_ext_bindI:
  "\<lbrakk> sim R ma1 mc1; \<forall>x. sim_ext R (ma2 x) (mc2 x) Sn (force_same_and R) R \<rbrakk> \<Longrightarrow>
    sim_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn (force_same_and R) R"
  by(auto simp add: force_same_and_def sim_def
    intro!: sim_ext_bindI[where Tn="force_same_and R"] sim_exts_empty_rel)

lemma sim_sim_ext_bindI':
  "\<lbrakk> sim R ma1 mc1; \<forall>x. sim_ext (R' x) (ma2 x) (mc2 x) Sn (force_same_and R) R; \<forall>x. R \<subseteq> R' x \<rbrakk> \<Longrightarrow>
    sim_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn (force_same_and R) R"
  apply(rule sim_sim_ext_bindI)
  apply(auto intro: sim_ext_mono)
  done

lemma simN_wpc_mono_concl:
  "\<lbrakk> R \<subseteq> simN_wpc ma mc S; \<forall>x y. S x y \<subseteq> S' x y \<rbrakk> \<Longrightarrow> R \<subseteq> simN_wpc ma mc S'"
  by(force simp add: simN_wpc_def)

lemma parallel_if_cases:
  "\<lbrakk> c \<Longrightarrow> P t t'; \<not>c \<Longrightarrow> P f f' \<rbrakk> \<Longrightarrow> P (if c then t else f) (if c then t' else f')"
  by(simp)

text {* Rules for dealing with the if construct *}
lemmas sim_ext_ifI = parallel_if_cases[where P="\<lambda>ma mc. sim_ext R ma mc Sn Se Sr"] for R Sn Se Sr

lemma parallel_if_cases_full:
  "\<lbrakk> c \<Longrightarrow> P t t'; \<not>c \<Longrightarrow> P f f'; c = c' \<rbrakk> \<Longrightarrow> P (if c then t else f) (if c' then t' else f')"
  by(simp)

lemmas sim_ext_ifI_full = parallel_if_cases_full[where
  P="\<lambda>ma mc. sim_ext R ma mc Sn Se Sr"] for R Sn Se Sr

lemmas sim_ifI_full = parallel_if_cases_full[where
  P="\<lambda>ma mc. sim R ma mc"] for R

lemmas sim_norm_ifI = parallel_if_cases[where P="\<lambda>ma mc. sim_norm R ma mc"] for R
lemmas sim_norm_ext_ifI = parallel_if_cases[where 
  P="\<lambda>ma mc. sim_norm_ext R ma mc Sn Se"] for R Sn Se

lemma sim_ext_ifI':
  "\<lbrakk>sim_ext (R \<inter> bool_into_set cond) maT mcT Sn Se Sr;
    sim_ext (R \<inter> bool_into_set (\<not>cond)) maF mcF Sn Se Sr\<rbrakk> \<Longrightarrow>
    sim_ext R
    (if cond then maT else maF)
    (if cond then mcT else mcF)
    Sn Se Sr"
  by(simp add: bool_into_set_def split: split_if_asm)

lemma sim_ext_ifI'':
  "\<lbrakk>sim_ext (R \<inter> bool_into_set cond) maT mcT Sn Se Sr;
    sim_ext (R' \<inter> bool_into_set (\<not>cond)) maF mcF Sn Se Sr; R'' \<subseteq> R \<inter> R'\<rbrakk> \<Longrightarrow>
    sim_ext R''
    (if cond then maT else maF)
    (if cond then mcT else mcF)
    Sn Se Sr"
  by(auto simp add: bool_into_set_def sim_exts_empty_rel split: split_if_asm intro: sim_ext_mono)

lemma sim_ext_bindI':
  "\<lbrakk> sim_ext R ma1 mc1 Tn Se Sr; \<forall>va vc. sim_ext (Tn' va vc) (ma2 va) (mc2 vc) Sn Se Sr;
    \<forall>va vc. Tn va vc \<subseteq> Tn' va vc\<rbrakk> \<Longrightarrow>
    sim_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se Sr"
proof(rule sim_ext_bindI[where Tn=Tn])
  assume
    "\<forall>va vc. sim_ext (Tn' va vc) (ma2 va) (mc2 vc) Sn Se Sr"
    "\<forall>va vc. Tn va vc \<subseteq> Tn' va vc"
  thus "\<forall>va vc. sim_ext (Tn va vc) (ma2 va) (mc2 vc) Sn Se Sr"
    by(force intro: sim_ext_mono)
qed(auto)

text {* General lemmas *}
lemma triv_precond: "(\<forall>x. x \<longrightarrow> P) = P"
  by blast

lemma Int_Collect:
  "A \<inter> {x. P x} = {x. x \<in> A \<and> P x}"
  "{x. P x} \<inter> A = {x. x \<in> A \<and> P x}"
  by(auto)

lemma subset_Collect:
  "(A \<subseteq> {x. P x}) = (\<forall>x. x \<in> A \<longrightarrow> P x)"
  by auto

lemma Int_subset_same':
  "(A \<subseteq> A \<inter> B) = (A \<subseteq> B)"
  "(A \<subseteq> B \<inter> A) = (A \<subseteq> B)"
  by simp_all

lemma imp_conj_same:
  "(x \<and> y \<longrightarrow> x \<and> z) = (x \<and> y \<longrightarrow> z)"
  "(x \<and> y \<longrightarrow> y \<and> z) = (x \<and> y \<longrightarrow> z)"
  "(x \<and> y \<longrightarrow> z \<and> x) = (x \<and> y \<longrightarrow> z)"
  "(x \<and> y \<longrightarrow> z \<and> y) = (x \<and> y \<longrightarrow> z)"
  "(x \<longrightarrow> x \<and> y) = (x \<longrightarrow> y)"
  "(x \<longrightarrow> y \<and> x) = (x \<longrightarrow> y)"
  by simp_all

lemma subset_bCollect: "(A \<subseteq> {x \<in> B. P x}) = (\<forall>x. x \<in> A \<longrightarrow> x \<in> B \<and> P x)"
  by(auto)

lemma imp_conj_simps:
  "((P \<longrightarrow> Q) \<and> (P \<longrightarrow> R)) = (P \<longrightarrow> Q \<and> R)"
  "((P \<longrightarrow> Q) \<and> ((P \<longrightarrow> R) \<and> S)) = ((P \<longrightarrow> Q \<and> R) \<and> S)"
  "((P \<longrightarrow> Q) \<and> (S \<and> (P \<longrightarrow> R))) = ((P \<longrightarrow> Q \<and> R) \<and> S)"
  "((P \<longrightarrow> Q) \<and> (S \<and> ((P \<longrightarrow> R) \<and> T))) = ((P \<longrightarrow> Q \<and> R) \<and> S \<and> T)"
  (* "(P \<longrightarrow> Q \<longrightarrow> R) = (P \<and> Q \<longrightarrow> R)" *)
  "((\<forall>x. P' x \<longrightarrow> Q' x) \<and> (\<forall>y. P' y \<longrightarrow> R' y)) = (\<forall>x. P' x \<longrightarrow> Q' x \<and> R' x)"
  "((\<forall>x y. P'' x y \<longrightarrow> Q'' x y) \<and> (\<forall>x y. P'' x y \<longrightarrow> R'' x y)) = (\<forall>x y. P'' x y \<longrightarrow> Q'' x y \<and> R'' x y)"
  "((\<forall>x y. P'' x y \<longrightarrow> Q'' x y) \<and> ((\<forall>x y. P'' x y \<longrightarrow> R'' x y) \<and> S)) =
        ((\<forall>x y. P'' x y \<longrightarrow> Q'' x y \<and> R'' x y) \<and> S)"
  by(auto)

lemma comm_same_alter:
  "(P \<and> (Q \<and> P)) = (P \<and> Q)"
  "(P \<and> (Q \<and> (P \<and> R))) = (P \<and> Q \<and> R)"
  by auto

lemma forall_conj_distrib: "(\<forall>x. P x \<and> Q x) = ((\<forall>x. P x) \<and> (\<forall>x. Q x))"
  by auto

lemma subset_refl_eq:
  "(A \<subseteq> A) = True"
  by(simp)

lemma Int_lower1_eq: "(A \<inter> B \<subseteq> A) = True"
  by(simp)

lemma Int_lower2_eq: "(A \<inter> B \<subseteq> B) = True"
  by(simp)

lemma set_equalityI_HOL:  "(A \<subseteq> B \<and> B \<subseteq> A) = (A = B)"
  by(auto)


text {* Rule variants for relational Hoare tuples *}
lemma sim_ext_same_relI:
  "
  sim_ext R (ma va) (mc va) Sn Se Sr \<Longrightarrow>
  sim_ext (force_same_and R va vc) (ma va) (mc vc) Sn Se Sr
  "
  by(auto simp add: force_same_and_def sim_exts_empty_rel)

lemma sim_ext_sim_bindI:
  "\<lbrakk> sim_ext R ma1 mc1 Sn (force_same_and R) R; \<forall>x y. sim R (ma2 x) (mc2 y); \<forall>x y. Sn x y \<subseteq> R  \<rbrakk> \<Longrightarrow>
    sim R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  apply(auto simp add: sim_def intro!: sim_ext_bindI)
  apply(rule_tac R=R in sim_ext_mono, auto)
  done

lemma sim_ext_try_catch_sameI:
  "\<lbrakk> sim_ext R ma mc Sn (force_same_and R') Sr; \<forall>e. sim_ext R' (ha e) (hc e) Sn Se Sr \<rbrakk> \<Longrightarrow>
    sim_ext R (try_catch ma ha) (try_catch mc hc) Sn Se Sr"
  apply(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set 
    del: subsetI)
  apply(auto simp add: force_same_and_def)
  done

lemma sim_ext_try_catchI':
  "\<lbrakk>sim_ext R ma mc Sn Te Sr;
    \<forall>ea ec. sim_ext (Te' ea ec) (ha ea) (hc ec) Sn Se Sr; \<forall>ea ec. Te ea ec \<subseteq> Te' ea ec\<rbrakk>
  \<Longrightarrow> sim_ext R (<try> ma <catch> ha) (<try> mc <catch> hc) Sn Se Sr"
  apply(rule sim_ext_try_catchI)
  apply(auto simp add: sim_ext_def intro: simN_wpc_mono_concl del: subsetI)
  done

lemma sim_ext_try_catch_sameI':
  "\<lbrakk>sim_ext R ma mc Sn (force_same_and R') Sr;
    \<forall>e. sim_ext (Te' e) (ha e) (hc e) Sn Se Sr; \<forall>e. R' \<subseteq> Te' e\<rbrakk>
  \<Longrightarrow> sim_ext R (<try> ma <catch> ha) (<try> mc <catch> hc) Sn Se Sr"
  apply(rule sim_ext_try_catch_sameI)
  apply(simp)
  apply(auto simp add: sim_ext_def del: subsetI)
  apply(blast)+
  done

lemma sim_ext_bind_sameI:
  "\<lbrakk>sim_ext R ma1 mc1 Tn Se Sr;
    \<forall>v. sim_ext (Tn' v) (ma2 v) (mc2 v) Sn Se Sr;
    \<forall>v. Tn v v \<subseteq> Tn' v; \<forall>va vc. va \<noteq> vc \<longrightarrow> Tn va vc = {} \<rbrakk>
    \<Longrightarrow> sim_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se Sr"
  apply(rule sim_ext_bindI')
  apply(auto)
  apply(case_tac "va = vc")
  apply(auto intro: sim_exts_empty_rel sim_ext_mono)
  done

lemma sim_2vars_bindI:
  "\<lbrakk>sim R ma1 mc1; \<forall>x y. sim R (ma2 x) (mc2 y)\<rbrakk> \<Longrightarrow>
    sim R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto intro: sim_bindI)

lemma sim_sim_ext_N_bindI:
  "\<lbrakk> sim_norm R ma1 mc1; \<forall>x. sim_norm_ext R (ma2 x) (mc2 x) Sn (force_same_and R) \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn (force_same_and R)"
  by(auto simp add: force_same_and_def sim_norm_def
    intro!: sim_norm_ext_bindI[where Tn="force_same_and R"] sim_exts_empty_rel)

lemma sim_sim_ext_N_bindI':
  "\<lbrakk> sim_norm R ma1 mc1; \<forall>x. sim_norm_ext (R' x) (ma2 x) (mc2 x) Sn (force_same_and R);
    \<forall>x. R \<subseteq> R' x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn (force_same_and R)"
  apply(rule sim_sim_ext_N_bindI)
  apply(auto intro: sim_norm_ext_mono)
  done

lemma sim_norm_ext_bind_sameI:
  "\<lbrakk>sim_norm_ext R ma1 mc1 Tn Se;
    \<forall>v. sim_norm_ext (Tn' v) (ma2 v) (mc2 v) Sn Se; \<forall>v. Tn v v \<subseteq> Tn' v;
    \<forall>va vc. va \<noteq> vc \<longrightarrow> Tn va vc = {}
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se"
  apply(rule sim_norm_ext_bindI, auto)
  apply(case_tac "va = vc")
  apply(auto intro: sim_exts_empty_rel sim_norm_ext_mono)
  done

lemma sim_norm_ext_try_catch_sameI:
  "\<lbrakk> sim_norm_ext R ma mc Sn Te; \<forall>e. sim_norm_ext (Te' e) (ha e) (hc e) Sn Se;
    \<forall>e. Te e e \<subseteq> Te' e; \<forall>ea ec. ea \<noteq> ec \<longrightarrow> Te ea ec = {}
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catch ma ha) (try_catch mc hc) Sn Se"
  apply(rule sim_norm_ext_try_catchI, auto)
  apply(case_tac "ea=ec")
  apply(auto intro: sim_exts_empty_rel sim_norm_ext_mono)
  done

lemma sim_ext_sim_N_bindI:
  "\<lbrakk> sim_norm_ext R ma1 mc1 Sn (force_same_and R);
    \<forall>x y. sim_norm R (ma2 x) (mc2 y); \<forall>x y. Sn x y \<subseteq> R  \<rbrakk> \<Longrightarrow>
    sim_norm R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  apply(auto simp add: sim_norm_def intro!: sim_norm_ext_bindI)
  apply(rule_tac R=R in sim_norm_ext_mono, auto)
  done

lemma sim_ext_sim_N_bind_sameI:
  "\<lbrakk> sim_norm_ext R ma1 mc1 (force_same_and R') Se;
    \<forall>x. sim_norm R' (ma2 x) (mc2 x);
    \<forall>x. R' \<subseteq> Se x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) (force_same_and R') Se"
  apply(auto simp add: sim_norm_def intro!: sim_norm_ext_bind_sameI intro: sim_norm_ext_mono)
  apply(auto simp add: sim_norm_ext_def forall_conj_distrib force_same_and_def
    intro: simN_wpc_monotone_rel simN_wpc_mono_concl subset_trans del: subsetI)
  done

lemma sim_norm_ext_sim_bind_sameI:
  "\<lbrakk> sim_norm_ext R ma1 mc1 (force_same_and R') Se;
    \<forall>x. sim R' (ma2 x) (mc2 x);
    \<forall>x. R' \<subseteq> Se x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) (force_same_and R') Se"
  by(auto intro!: sim_ext_sim_N_bind_sameI sim_imp_sim_norm)

lemma sim_ext_sim_N_try_catch_sameI:
  "\<lbrakk> sim_norm_ext R ma mc1 Sn (force_same_and R');
    \<forall>x. sim_norm R' (ma2 x) (mc2 x);
    \<forall>x. R' \<subseteq> Sn x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catch ma ma2) (try_catch mc1 mc2) Sn (force_same_and R')"
  apply(auto simp add: sim_norm_def intro!: sim_norm_ext_try_catch_sameI intro: sim_norm_ext_mono)
  apply(auto simp add: sim_norm_ext_def forall_conj_distrib force_same_and_def
    intro: simN_wpc_monotone_rel simN_wpc_mono_concl subset_trans del: subsetI)
  done

lemma sim_norm_ext_sim_try_catch_sameI:
  "\<lbrakk> sim_norm_ext R ma mc1 Sn (force_same_and R');
    \<forall>x. sim R' (ma2 x) (mc2 x);
    \<forall>x. R' \<subseteq> Sn x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catch ma ma2) (try_catch mc1 mc2) Sn (force_same_and R')"
  by(auto intro!: sim_ext_sim_N_try_catch_sameI sim_imp_sim_norm)

lemma sim_ext_imp_sim_norm_ext:
  "sim_ext R ma mc Sn Se Sr \<Longrightarrow> sim_norm_ext R ma mc Sn Se"
  by(auto simp add: sim_ext_def sim_norm_ext_def)

lemma sim_norm_ext_sim_ext_bind_sameI:
  "\<lbrakk> sim_norm_ext R ma1 mc1 Sn Se;
    \<forall>x. sim_ext (R' x) (ma2 x) (mc2 x) Sn' Se' Sr';
    \<forall>x y. x \<noteq> y \<longrightarrow> Sn x y = {};
    \<forall>x. Sn x x \<subseteq> R' x;
    \<forall>ea ec. Se' ea ec \<subseteq> Se ea ec
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn' Se
  "
  apply(rule sim_norm_ext_bind_sameI)
  apply(auto simp add: sim_ext_def sim_norm_ext_def forall_conj_distrib
    intro: subset_trans simN_wpc_mono_concl del: subsetI)
  done

lemma sim_norm_ext_sim_ext_try_catch_sameI:
  "\<lbrakk> sim_norm_ext R ma mc Sn Se;
    \<forall>x. sim_ext (R' x) (ha x) (hc x) Sn' Se' Sr';
    \<forall>x y. x \<noteq> y \<longrightarrow> Se x y = {};
    \<forall>x. Se x x \<subseteq> R' x;
    \<forall>va vc. Sn' va vc \<subseteq> Sn va vc
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catch ma ha) (try_catch mc hc) Sn Se'
  "
  apply(rule sim_norm_ext_try_catch_sameI)
  apply(auto simp add: sim_ext_def sim_norm_ext_def forall_conj_distrib
    intro: subset_trans simN_wpc_mono_concl del: subsetI)
  done


lemma sim_sim_ext_N_bindI'':
  "\<lbrakk> sim_norm R ma1 mc1; \<forall>x. sim_norm_ext (R' x) (ma2 x) (mc2 x) Sn Se;
    \<forall>x. R \<subseteq> R' x;
    \<forall>x. R \<subseteq> Se x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se"
  apply(simp only: sim_norm_def)
  apply(subgoal_tac "sim_norm_ext R ma1 mc1 (force_same_and R) Se")
  apply(erule sim_norm_ext_bind_sameI)
  apply(auto simp add: force_same_and_def sim_norm_ext_def forall_conj_distrib
    intro: simN_wpc_mono_concl del: subsetI)
  done

lemma sim_sim_norm_ext_bindI:
  "\<lbrakk> sim R ma1 mc1; \<forall>x. sim_norm_ext (R' x) (ma2 x) (mc2 x) Sn Se;
    \<forall>x. R \<subseteq> R' x;
    \<forall>x. R \<subseteq> Se x x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se"
  by(blast dest!: sim_imp_sim_norm intro: sim_sim_ext_N_bindI'')

lemma sim_norm_ext_mapM':
  "\<forall>x \<in> set l. sim_norm_ext R (ma x) (mc x) (force_same_and R) Se \<Longrightarrow>
    sim_norm_ext R (mapM' ma l) (mapM' mc l) (force_same_and R) Se"
  apply(induct l)
  apply(auto intro: sim_ext_imp_sim_norm_ext sim_ext_return)
  apply(auto simp add: force_same_and_def intro!: sim_norm_ext_bindI sim_exts_empty_rel)
  done

lemma force_same_and_diff:
  "(va \<noteq> vc \<longrightarrow> force_same_and R va vc = {}) = True"
  by(auto simp add: force_same_and_def)

lemma force_same_and_diff':
  "va \<noteq> vc \<Longrightarrow> force_same_and R va vc = {}"
  by(auto simp add: force_same_and_def)

lemma force_same_and_subset:
  "(force_same_and R x y \<subseteq> R) = True"
  by(auto simp add: force_same_and_def)

lemma sim_sim_norm_bindI:
  "\<lbrakk>sim R ma1 mc1; \<forall>x. sim_norm R (ma2 x) (mc2 x)\<rbrakk>
  \<Longrightarrow> sim_norm R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)
  "
  by(auto dest: sim_imp_sim_norm intro: sim_norm_bindI)

lemma sim_norm_sim_bindI:
  "\<lbrakk>sim_norm R ma1 mc1; \<forall>x. sim R (ma2 x) (mc2 x)\<rbrakk>
  \<Longrightarrow> sim_norm R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)
  "
  by(auto dest: sim_imp_sim_norm intro: sim_norm_bindI)

lemma sim_sim_norm_try_catchI:
  "\<lbrakk>sim R ma mc; \<forall>e. sim_norm R (ha e) (hc e)\<rbrakk>
  \<Longrightarrow> sim_norm R (<try> ma <catch> ha) (<try> mc <catch> hc)"
  by(auto dest: sim_imp_sim_norm intro: sim_norm_try_catchI)

lemma sim_norm_sim_try_catchI:
  "\<lbrakk>sim_norm R ma mc; \<forall>e. sim R (ha e) (hc e)\<rbrakk>
  \<Longrightarrow> sim_norm R (<try> ma <catch> ha) (<try> mc <catch> hc)"
  by(auto dest: sim_imp_sim_norm intro: sim_norm_try_catchI)

lemma simN_wpc_bindS:
  "\<lbrakk> R \<subseteq> simS_wpc ma mc R'; R' \<subseteq> simN_wpc ha hc S \<rbrakk> \<Longrightarrow>
  R \<subseteq> simN_wpc (\<lambda>s. ma s \<guillemotright>=S ha) (\<lambda>s. mc s \<guillemotright>=S hc) S
  "
  apply(auto simp add: simS_wpc_def simN_wpc_def unfold_monad split: split_if_asm)
  apply (smt split_conv subset_Collect)
  done

lemma sim_norm_bindS:
  "\<lbrakk> sim R ma mc; sim_norm R ha hc \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc)
  "
  by(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def
    intro!: simN_wpc_bindS
    del: subsetI)

thm simN_wpc_bindN_bindS

lemma sim_norm_bind_bindS:
  "\<lbrakk>
    sim_norm R ma1 mc1;
    sim_norm R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S hc);
    \<forall>x. sim_norm R (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. ((mc2 x)\<Down>R) s \<guillemotright>=S hc)
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S hc)
  "
  apply(auto simp add: bind_projectR choice_bindS_bindN intro!: sims_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  done

lemma sim_norm_try_catch_bindS:
  "\<lbrakk>
    sim_norm R ma mc;
    sim_norm R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S rhc);
    \<forall>e. sim_norm R (\<lambda>s. ((ha e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((hc e)\<Down>R) s \<guillemotright>=S rhc)
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((try_catch ma ha)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((try_catch mc hc)\<Down>R) s \<guillemotright>=S rhc)
  "
  apply(auto simp add: try_catch_projectR choice_bindS_bindN intro!: sims_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  done

lemma sim_norm_bind_bindS':
  "\<lbrakk>
    sim_norm_ext R ma1 mc1 Sn Se;
    sim_norm R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S hc);
    \<forall>x. sim_norm_ext R' (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. ((mc2 x)\<Down>R) s \<guillemotright>=S hc)
        (force_same_and R)
        (force_same_and R);
    \<forall>x y. x \<noteq> y \<longrightarrow> Sn x y = {};
    \<forall>x. Sn x x \<subseteq> R'
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S hc)
  "
  apply(auto simp add: bind_projectR choice_bindS_bindN intro!: sims_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def forall_conj_distrib
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  apply(case_tac[!] "x = y", auto intro: subset_trans del: subsetI)
  done

lemma simN_wpc_mapM_bindS_I:
  "\<lbrakk> \<forall>x \<in> set l. sim_normN R (ma x) (mc x);
     \<forall>x \<in> set l. R \<subseteq> simN_wpc (\<lambda>s. ((ma x)\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. ((mc x)\<Down>R) s \<guillemotright>=S hc) S
  \<rbrakk> \<Longrightarrow>
  R \<subseteq> simN_wpc
    (\<lambda>s. ((mapM' ma l)\<Down>R) s \<guillemotright>=S ha)
    (\<lambda>s. ((mapM' mc l)\<Down>R) s \<guillemotright>=S hc)
  S"
  apply(induct l)
  apply(auto simp add: sims_through_simN_simS simN_def project_simps
    simN_wpc_empty_conc first_stage empty_bindS
    intro!: simN_wpc_bindN_bindS simN_wpc_bindN_set simN_wpc_union_lockstep del: subsetI)
  apply(auto simp add: unfold_monad force_same_and_def simN_wpc_def)
  done

lemma sim_norm_mapM_bindS:
  "\<lbrakk> \<forall>x \<in> set l. sim_norm R (ma x) (mc x);
    \<forall>x \<in> set l. sim_norm R (\<lambda>s. ((ma x)\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. ((mc x)\<Down>R) s \<guillemotright>=S hc)
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((mapM' ma l)\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. ((mapM' mc l)\<Down>R) s \<guillemotright>=S hc)
  "
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def simN_def force_same_and_def
    intro!: simN_wpc_mapM_bindS_I
    del: subsetI)
  done

lemma sim_norm_if_bindS_I:
  "\<lbrakk> cond \<Longrightarrow> sim_norm R (\<lambda>s. (mat\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mct\<Down>R) s \<guillemotright>=S hc);
    \<not>cond \<Longrightarrow> sim_norm R (\<lambda>s. (maf\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mcf\<Down>R) s \<guillemotright>=S hc)
  \<rbrakk> \<Longrightarrow>
  sim_norm R
    (\<lambda>s. ((if cond then mat else maf)\<Down>R) s \<guillemotright>=S ha)
    (\<lambda>s. ((if cond then mct else mcf)\<Down>R) s \<guillemotright>=S hc)
  "
  by(simp)

lemma sim_norm_ext_if_bindS_I:
  "\<lbrakk> cond \<Longrightarrow> sim_norm_ext R (\<lambda>s. (mat\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mct\<Down>R) s \<guillemotright>=S hc) Sn Se;
    \<not>cond \<Longrightarrow> sim_norm_ext R (\<lambda>s. (maf\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mcf\<Down>R) s \<guillemotright>=S hc) Sn Se
  \<rbrakk> \<Longrightarrow>
  sim_norm_ext R
    (\<lambda>s. ((if cond then mat else maf)\<Down>R) s \<guillemotright>=S ha)
    (\<lambda>s. ((if cond then mct else mcf)\<Down>R) s \<guillemotright>=S hc)
    Sn Se
  "
  by(simp)

lemma sim_norm_ext_if_bindS_no_cond:
  "\<lbrakk> sim_norm_ext R (\<lambda>s. (mat\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mct\<Down>R) s \<guillemotright>=S hc) Sn Se;
     sim_norm_ext R (\<lambda>s. (maf\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mcf\<Down>R) s \<guillemotright>=S hc) Sn Se
  \<rbrakk> \<Longrightarrow>
  sim_norm_ext R
    (\<lambda>s. ((if cond then mat else maf)\<Down>R) s \<guillemotright>=S ha)
    (\<lambda>s. ((if cond then mct else mcf)\<Down>R) s \<guillemotright>=S hc)
    Sn Se
  "
  by(simp)

text {* Additional combinations of @{term "sim_norm"} and @{term
  "sim_norm_ext"} used for abs1/3 refinement *}

lemma sim_norm_ext_try_catch_bindS:
  "\<lbrakk>sim_norm_ext R ma mc Sn' Se';
    sim_norm_ext R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>e. sim_norm_ext (R' e)  (\<lambda>s. ((ha e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((hc e)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>x y. x \<noteq> y \<longrightarrow> Se' x y = {};
    \<forall>x. Se' x x \<subseteq> R' x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R
        (\<lambda>s. ((<try> ma <catch> ha)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((<try> mc <catch> hc)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se
  "
  apply(auto simp add: try_catch_projectR choice_bindS_bindN
    intro!: sim_norm_ext_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  apply(case_tac[!] "x=y")
  apply(auto intro: subset_trans del: subsetI)
  done

lemma sim_norm_try_catch_bindS':
  "\<lbrakk>sim_norm_ext R ma mc Sn Se;
    sim_norm R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S rhc);
    \<forall>e. sim_norm_ext (R' e)  (\<lambda>s. ((ha e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((hc e)\<Down>R) s \<guillemotright>=S rhc)
        (force_same_and R)
        (force_same_and R);
    \<forall>x y. x \<noteq> y \<longrightarrow> Se x y = {};
    \<forall>x. Se x x \<subseteq> R' x
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((<try> ma <catch> ha)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((<try> mc <catch> hc)\<Down>R) s \<guillemotright>=S rhc)
  "
  by(unfold sim_norm_def, rule sim_norm_ext_try_catch_bindS, assumption+)

lemma sim_norm_ext_bind_bindS:
  "\<lbrakk>sim_norm_ext R ma1 mc1 Sn' Se';
    sim_norm_ext R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>e. sim_norm_ext (R' e)  (\<lambda>s. ((ma2 e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc2 e)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>x y. x \<noteq> y \<longrightarrow> Sn' x y = {};
    \<forall>x. Sn' x x \<subseteq> R' x
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se
  "
  apply(auto simp add: bind_projectR choice_bindS_bindN
    intro!: sim_norm_ext_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  apply(case_tac[!] "x=y")
  apply(auto intro: subset_trans del: subsetI)
  done

lemma sim_norm_ext_bind_bindS':
  "\<lbrakk>sim_norm_ext R ma1 mc1 (force_same_and R') (force_same_and R');
    sim_norm_ext R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S rhc)
        (force_same_and R') (force_same_and R');
    \<forall>x. sim_norm R' (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc2 x)\<Down>R) s \<guillemotright>=S rhc)
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S rhc)
        (force_same_and R')
        (force_same_and R')
  "
  apply(unfold sim_norm_def)
  apply(rule sim_norm_ext_bind_bindS)
  apply(assumption+)
  apply(clarsimp_all simp add: force_same_and_diff force_same_and_same del: subsetI)
  done

lemma sim_norm_try_catch_bindS'':
  "\<lbrakk>sim_norm_ext R ma mc (force_same_and R') (force_same_and R');
    sim_norm_ext R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S rhc)
        (force_same_and R') (force_same_and R');
    \<forall>e. sim_norm R' (\<lambda>s. ((ha e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((hc e)\<Down>R) s \<guillemotright>=S rhc)
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R
        (\<lambda>s. ((<try> ma <catch> ha)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((<try> mc <catch> hc)\<Down>R) s \<guillemotright>=S rhc)
        (force_same_and R')
        (force_same_and R')
  "
  apply(auto simp add: try_catch_projectR choice_bindS_bindN sim_norm_def
    intro!: sim_norm_ext_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  done

lemma sim_norm_bindS'':
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; sim_norm R ha hc;
    Sr \<subseteq> R; \<forall>x. R \<subseteq> Se x x; \<forall>x. R \<subseteq> Sn x x
    \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc) Sn Se"
  apply(auto simp add: sim_norm_def sim_norm_ext_def sim_ext_def sim_def
    bindS_projectNN bindS_projectNE
    del: subsetI)
  apply(auto simp add: force_same_and_def intro!: simN_wpc_bindS del: subsetI)
  apply (smt Int_empty_left bot_least le_infI2 simN_wpc_mono_concl subset_trans)+
  done

lemma sim_ext_throw':
  "\<forall>x. R \<subseteq> Se x x \<Longrightarrow> sim_ext R (throw e) (throw e) Sn Se {}"
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def project_defs unfold_monad)

lemma sim_ext_return':
  "\<forall>x. R \<subseteq> Sn x x \<Longrightarrow> sim_ext R (return x) (return x) Sn Se {}"
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def project_defs unfold_monad)

lemma sim_norm_ext_return':
  "sim_norm_ext R (return x) (return x) (force_same_and R) Se"
  by(auto simp add: sim_norm_ext_def simN_wpc_def force_same_and_def project_defs unfold_monad)

lemma sim_norm_ext_throw':
  "sim_norm_ext R (throw e) (throw e) Sn (force_same_and R)"
  by(auto simp add: sim_norm_ext_def simN_wpc_def force_same_and_def project_defs unfold_monad)

lemma sim_norm_ext_bindS:
  "\<lbrakk> sim R ma mc; sim_norm_ext R ha hc Sn Se \<rbrakk>
    \<Longrightarrow> sim_norm_ext R
        (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc)
        Sn Se
  "
  by(auto simp add: sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def
    intro!: simN_wpc_bindS
    del: subsetI)

lemma sim_sim_ext_N_bind_bindS:
  "\<lbrakk> sim_norm R ma1 mc1;
    sim_norm_ext R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S rhc) Sn Se;
    \<forall>x. sim_norm_ext R' (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc2 x)\<Down>R) s \<guillemotright>=S rhc) Sn Se;
    R \<subseteq> R'
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S rhc) Sn Se"
  by(auto simp add: sim_norm_def force_same_and_def intro!: sim_norm_ext_bind_bindS)

lemma sim_sim_ext_N_try_catch_bindS:
  "\<lbrakk> sim_norm R ma1 mc1;
    sim_norm_ext R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S rhc) Sn Se;
    \<forall>e. sim_norm_ext R (\<lambda>s. ((ma2 e)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc2 e)\<Down>R) s \<guillemotright>=S rhc) Sn Se
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (\<lambda>s. ((try_catch ma1 ma2)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((try_catch mc1 mc2)\<Down>R) s \<guillemotright>=S rhc) Sn Se"
  by(auto simp add: sim_norm_def force_same_and_def intro!: sim_norm_ext_try_catch_bindS)



lemmas plain_sim_bind_rt =
  sim_bindI

lemmas sim_ext_bind_rt =
  plain_sim_bind_rt
  sim_sim_ext_bindI
  sim_sim_ext_bindI'
  sim_ext_sim_bindI
  sim_ext_bind_sameI
  sim_ext_bindI
  sim_ext_bindI'
  sim_2vars_bindI

lemmas sim_norm_bind_rt = 
  sim_ext_bind_rt
  sim_norm_bindI
  sim_norm_sim_bindI
  sim_norm_ext_bindI
  sim_norm_ext_bind_sameI
  sim_sim_ext_N_bindI
  sim_sim_ext_N_bindI'
  sim_sim_ext_N_bindI''
  sim_sim_norm_ext_bindI
  sim_ext_sim_N_bind_sameI
  sim_norm_ext_sim_bind_sameI
  sim_norm_ext_sim_ext_bind_sameI

lemmas plain_sim_try_catch_rt =
  sim_try_catchI

lemmas sim_ext_try_catch_rt =
  plain_sim_try_catch_rt
  sim_try_catchI
  sim_ext_try_catchI
  sim_ext_try_catchI'
  sim_ext_try_catch_sameI
  sim_ext_try_catch_sameI'

lemmas sim_norm_try_catch_rt =
  sim_ext_try_catch_rt
  sim_norm_try_catchI
  sim_ext_try_catch_rt
  sim_norm_ext_try_catch_sameI
  sim_norm_ext_sim_ext_try_catch_sameI
  sim_ext_sim_N_try_catch_sameI
  sim_norm_ext_sim_try_catch_sameI

lemmas plain_sim_exc_case_rt =
  sim_exceptions_caseI

lemmas sim_ext_exc_case_rt =
  plain_sim_exc_case_rt
  sim_ext_exceptions_caseI

lemmas sim_norm_exc_case_rt =
  sim_ext_exc_case_rt
  sim_norm_exceptions_caseI
  sim_norm_ext_exceptions_caseI
  sim_norm_exceptions_case_bindS_I

lemmas plain_sim_if_rt =
  sim_if_I'
  sim_if_no_cond

lemmas sim_ext_if_rt =
  plain_sim_if_rt
  sim_ext_ifI

lemmas sim_norm_if_rt =
  sim_ext_if_rt
  sim_norm_ifI
  sim_norm_ext_ifI
  sim_norm_if_bindS_I

lemmas plain_sim_if_full_rt =
  sim_ifI_full

lemmas sim_ext_if_full_rt =
  plain_sim_if_full_rt
  sim_ext_ifI_full

lemmas plain_sim_choice_rt =
  sim_choice_lockstep

lemmas sim_norm_choice_rt =
  plain_sim_choice_rt
  sim_norm_choice_lockstep

lemmas plain_mapM_rt =
  sim_mapM(1)

lemmas sim_ext_mapM_rt =
  plain_mapM_rt

lemmas sim_norm_mapM_rt =
  sim_ext_mapM_rt
  sim_mapM(2)
  sim_norm_ext_mapM'

lemmas bind_bindS_rt =
  sim_norm_bind_bindS
  sim_norm_bind_bindS'

lemmas try_catch_bindS_rt =
  sim_norm_try_catch_bindS

lemmas mapM_bindS_rt =
  sim_norm_mapM_bindS

lemmas plain_bindS_rt =
  sim_norm_bindS

declare
  sim_def[ref_sim_def]
  sim_norm_def[ref_sim_def]

declare
  UNIV_def[THEN sym, ref_simp]
  Set.empty_def[THEN sym, ref_simp]
  Int_UNIV_right[ref_simp]
  Int_UNIV_left[ref_simp]
  Int_empty_left[ref_simp]
  Int_empty_right[ref_simp]
  empty_subsetI[ref_simp]
  Int_Collects[ref_simp]
  Collect_subset[ref_simp]
  Set.Int_Collect[ref_simp]
  Pmm_Refinement.Int_Collect[ref_simp]
  mem_Collect_eq[ref_simp]
  subset_Collect[ref_simp]
  subset_bCollect[ref_simp]
  empty_iff[ref_simp]
  Int_iff[ref_simp]
  imp_conj_same[ref_simp]
  imp_conj_simps[ref_simp]
  conj_commute[ref_simp]
  conj_assoc[ref_simp]
  conj_left_absorb[ref_simp]
  iso_tuple_UNIV_I[ref_simp]
  HOL.simp_thms[ref_simp]
  HOL.ex_simps[ref_simp]
  HOL.True_implies_equals[ref_simp]
  Product_Type.split[ref_simp]
  admin_indx.simps[ref_simp]
  triv_precond[ref_simp]
  bool_into_set_def[ref_simp]
  indx.simps[ref_simp]
  comm_same_alter[ref_simp]
  force_same_and_diff[ref_simp]
  force_same_and_subset[ref_simp]
  force_same_and_same[ref_simp]
  if_True[ref_simp]
  subset_refl_eq[ref_simp]
  Int_lower1_eq[ref_simp]
  Int_lower2_eq[ref_simp]
  imp_ex[THEN sym, ref_simp]

lemma simN_wpc_add_inv':
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>> I'\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>> J'\<rbrace>;
    R \<inter> I \<times> J \<subseteq> simN_wpc ma mc S
  \<rbrakk> \<Longrightarrow>
    R \<inter> I \<times> J \<subseteq> simN_wpc ma mc (\<lambda>va vc. S va vc \<inter> (I' va \<times> J' vc))
  "
  apply(auto simp add: simN_wpc_def)
  apply(clarsimp_all simp add: Hoare_def Box_def)
  apply(drule_tac x=a in set_mp, simp)
  apply(drule_tac x=b in set_mp, simp)
  apply(clarsimp_all)
  apply (smt IntI SigmaI split_conv subset_Collect)
  done

lemma sim_normN_ext_add_inv':
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>>N I'\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>N J'\<rbrace>;
    sim_normN_ext (R \<inter> I \<times> J) ma mc Se
  \<rbrakk> \<Longrightarrow>
    sim_normN_ext (R \<inter> I \<times> J) ma mc (\<lambda>ea ec. Se ea ec \<inter> (I' ea \<times> J' ec))
  "
  by(auto simp only: sim_normN_ext_def HoareNN_alt intro!: simN_wpc_add_inv' del: subsetI)

lemma sim_normE_ext_add_inv':
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>@N I'\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>@N J'\<rbrace>;
    sim_normE_ext (R \<inter> I \<times> J) ma mc Se
  \<rbrakk> \<Longrightarrow>
    sim_normE_ext (R \<inter> I \<times> J) ma mc (\<lambda>ea ec. Se ea ec \<inter> (I' ea \<times> J' ec))
  "
  by(auto simp only: sim_normE_ext_def HoareNE_alt intro!: simN_wpc_add_inv' del: subsetI)

lemma sim_ext_add_NE_inv:
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>@N I'\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>@N J'\<rbrace>;
    sim_ext (R \<inter> I \<times> J) ma mc Sn Se Sr
  \<rbrakk> \<Longrightarrow>
    sim_ext (R \<inter> I \<times> J) ma mc Sn (\<lambda>ea ec. Se ea ec \<inter> (I' ea \<times> J' ec)) Sr
  "
  by(auto simp add: sim_ext_decompose intro: sim_normE_ext_add_inv')

lemma sim_ext_add_NN_inv:
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>>N I'\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>N J'\<rbrace>;
    sim_ext (R \<inter> I \<times> J) ma mc Sn Se Sr
  \<rbrakk> \<Longrightarrow>
    sim_ext (R \<inter> I \<times> J) ma mc (\<lambda>va vc. Sn va vc \<inter> (I' va \<times> J' vc)) Se Sr
  "
  by(auto simp add: sim_ext_decompose intro: sim_normN_ext_add_inv')

lemma sim_normN_ext_return:
  "sim_normN_ext R (return x) (return x) (force_same_and R)"
  by(auto simp add: sim_normN_ext_def simN_wpc_def project_defs unfold_monad force_same_and_def)

lemma sim_normN_ext_mapM':
  "\<forall>x \<in> set l. sim_normN_ext R (ma x) (mc x) (force_same_and R) \<Longrightarrow>
    sim_normN_ext R (mapM' ma l) (mapM' mc l) (force_same_and R)"
  apply(induct l)
  apply(auto intro: sim_normN_ext_return)
  apply(auto simp add: force_same_and_def
    intro!: sim_normN_ext_bindI sim_exts_empty_rel sim_normN_ext_return)
  done


lemma sim_normN_ext_mono_conc:
  "\<lbrakk>sim_normN_ext R ma mc Sn; \<forall>va vc. Sn va vc \<subseteq> Sn' va vc\<rbrakk> \<Longrightarrow> sim_normN_ext R ma mc Sn'"
  by(auto simp add: sim_normN_ext_def del: subsetI intro: simN_wpc_mono_concl)

lemma simN_wpc_join_conc_invs:
  "\<lbrakk> R \<inter> UNIV \<times> P \<subseteq> simN_wpc ma mc (\<lambda>va vc. Sn va vc \<inter> UNIV \<times> Q vc);
    R \<inter> UNIV \<times> P' \<subseteq> simN_wpc ma mc (\<lambda>va vc. Sn va vc \<inter> UNIV \<times> Q' vc) \<rbrakk> \<Longrightarrow>
  R \<inter> UNIV \<times> (P \<inter> P') \<subseteq> simN_wpc ma mc (\<lambda>va vc. Sn va vc \<inter> UNIV \<times> (Q vc \<inter> Q' vc))"
  apply(auto simp add: simN_wpc_def)
  by (smt IntI UNIV_I mem_Collect_eq mem_Sigma_iff split_conv subsetD)

lemma sim_ext_join_conc_invs:
  "\<lbrakk> sim_ext (R \<inter> UNIV \<times> P) ma mc (force_same_and (R \<inter> UNIV \<times> Q)) Se Sr;
    sim_ext (R \<inter> UNIV \<times> P') ma mc (force_same_and (R \<inter> UNIV \<times> Q')) Se Sr \<rbrakk> \<Longrightarrow>
  sim_ext (R \<inter> UNIV \<times> (P \<inter> P')) ma mc (force_same_and (R \<inter> UNIV \<times> (Q \<inter> Q'))) Se Sr"
  apply(simp only: sim_ext_def)
  apply(intro Int_greatest)
  apply(rule simN_wpc_mono_concl)
  apply(rule simN_wpc_join_conc_invs[where Sn="force_same_and R" and Q="K Q" and Q'="K Q'"])
  apply((clarsimp simp add: K_beta force_same_and_Int)+)[3]
  apply(blast)+
  done

lemma sim_ext_mapM'_distinct:
  assumes
  distinct: "distinct l"
  and refines:
    "\<forall>x \<in> set l. sim_ext (R \<inter> UNIV \<times> P x) (ma x) (mc x) (force_same_and R) Se Sr"
  and preservesP: "\<forall>x x'. x \<in> set l \<and> x' \<in> set l \<and> x \<noteq> x' \<longrightarrow> \<lbrace>P x\<rbrace> mc x' \<lbrace>>N K (P x)\<rbrace>"
  shows
    "sim_ext (R \<inter> UNIV \<times> (\<Inter>x \<in> set l. P x))
        (mapM' ma l)
        (mapM' mc l)
        (force_same_and R)
        Se Sr
    "
proof(induct l rule: list_distinct_set_induct)
  show "distinct l" using distinct .
next
  case Nil
  show ?case
    by(simp add: sim_ext_return)
next
  case (Cons x xs)
  note refx = ballE'[OF refines `x \<in> set l`]
  have refx_withPs:
    "sim_ext ((R \<inter> UNIV \<times> P x) \<inter> UNIV \<times> (\<Inter>x\<in>set xs. P x))
        (ma x) (mc x)
        (force_same_and (R \<inter> UNIV \<times> (\<Inter>x\<in>set xs. P x)))
        Se Sr
    "
    apply -
    apply(rule sim_ext_add_NN_inv[OF true_postNN, THEN sim_ext_mono_norm])
    apply(rule HoareNN_preserves_Inter[OF _ `x \<notin> set xs` order_refl])
    apply (metis Cons(1) Cons(2) Cons(3) preservesP set_rev_mp)
    apply(auto intro!: refx[THEN sim_ext_mono] simp add: force_same_and_def K_def)
    done
  hence refx_withPs:
    "sim_ext (R \<inter> UNIV \<times> (\<Inter>x\<in>set (x # xs). P x))
        (ma x) (mc x)
        (force_same_and (R \<inter> UNIV \<times> (\<Inter>x\<in>set xs. P x)))
        Se Sr
    "
    by(auto intro: sim_ext_mono)
  show "sim_ext (R \<inter> UNIV \<times> (\<Inter>x\<in>set (x # xs). P x))
    (mapM' ma (x # xs)) (mapM' mc (x # xs))
    (force_same_and R)
    Se Sr
    "
    apply -
    apply(simp only: mapM'.simps)
    apply(rule sim_ext_bindI[OF refx_withPs])
    apply(subst force_same_and_def)
    apply(auto intro!: Cons sim_exts_empty_rel)
    done
qed

end
