theory Pmm_Specification
imports Specification Refinement
begin

definition projectS :: "('a, 's)nds_monad \<Rightarrow> 's \<Rightarrow> 's set" ("(_\<Down>S)" [1000] 999) where
  "projectS m \<equiv> \<lambda>s. snd ` m s"

definition projectSN :: "('a, 's)r_monad \<Rightarrow> 's \<Rightarrow> 's set" ("(_\<Down>SN)" [1000] 999) where
  "projectSN \<equiv> projectS o projectN"

definition projectSNN :: "('a, 'e, 's)re_monad \<Rightarrow> 's \<Rightarrow> 's set" ("(_\<Down>SNN)" [1000] 999) where
  "projectSNN \<equiv> projectS o projectNN"

definition projectSNE :: "('a, 'e, 's)re_monad \<Rightarrow> 's \<Rightarrow> 's set" ("(_\<Down>SNE)" [1000] 999) where
  "projectSNE \<equiv> projectS o projectNE"

lemmas projectS_defs =
  projectS_def
  projectSN_def
  projectSNN_def
  projectSNE_def
  project_defs

definition set_fun_into_rel :: "('a \<Rightarrow> 'b set) \<Rightarrow> ('a \<times> 'b)set" where
  "set_fun_into_rel f \<equiv> {(x, y). y \<in> f x}"

definition try_rec_tryR where 
  "try_rec_tryR m h \<equiv> <tryR> m <catchR> rec_tryR h"

lemma sim_norm_init_inclusion:
  assumes
  ref: "sim_norm_ext R ma mc (force_same_and R') (force_same_and R')"
  and sub: "Sc \<subseteq> R `` Sa"
  shows "((set_fun_into_rel (projectSN mc)) `` Sc) \<subseteq> R' `` (set_fun_into_rel (projectSN ma) `` Sa)"
proof
  fix sc'
  assume "sc' \<in> set_fun_into_rel (mc\<Down>SN) `` Sc"
  hence "\<exists>v sc. sc \<in> Sc \<and> (Some v, sc') \<in> mc sc" (is "\<exists>v sc. ?P v sc")
    by(force simp add: set_fun_into_rel_def Image_def projectS_defs)
  then obtain v and sc where sc_in: "?P v sc" by auto
  hence "\<exists>sa. sa \<in> Sa \<and> (sa, sc) \<in> R" (is "\<exists>sa. ?P sa") using sub
    by(auto)
  then obtain sa where sa_in: "?P sa" ..
  from sc_in sa_in have "\<exists>sa'. sa' \<in> projectSN ma sa \<and> (sa', sc') \<in> R'" (is "\<exists>sa'. ?P sa'") using ref
    apply(case_tac v)
    apply(force simp add: sim_norm_ext_def simN_wpc_def projectS_defs force_same_and_def image_def 
      split: split_if_asm)+
    done
  then obtain sa' where sa'_def: "?P sa'" ..
  thus "sc' \<in> R' `` set_fun_into_rel (ma\<Down>SN) `` Sa" using sa_in
    by(auto simp add: set_fun_into_rel_def intro!: ImageI)
qed

lemma sim_norm_extD:
  "\<lbrakk> sim_norm_ext R ma mc Sn Se; (sa, sc) \<in> R; (Some vec, sc') \<in> mc sc \<rbrakk> \<Longrightarrow> 
  (\<exists>vc. vec = Inr vc \<and> (\<exists>va sa'. (Some (Inr va), sa') \<in> ma sa \<and> (sa', sc') \<in> Sn va vc))
  \<or> (\<exists>ec. vec = Inl ec \<and> (\<exists>ea sa'. (Some (Inl ea), sa') \<in> ma sa \<and> (sa', sc') \<in> Se ea ec))"
  apply(case_tac vec)
  apply(auto simp add: sim_norm_ext_def simN_wpc_def project_defs)
  done

lemma sim_norm_init_bind_inclusion:
  assumes
  ref: "sim_norm_ext R ma2 mc2 (force_same_and R') (force_same_and R')"
  and 
  subNN: "set_fun_into_rel (mc1\<Down>SNN) `` Sc \<subseteq> R `` (set_fun_into_rel (ma1\<Down>SNN) `` Sa)"
  and
  subNE: "set_fun_into_rel (mc1\<Down>SNE) `` Sc \<subseteq> R' `` (set_fun_into_rel (ma1\<Down>SNE) `` Sa)"
  shows "((set_fun_into_rel ((mc1 \<guillemotright> mc2)\<Down>SN)) `` Sc) 
        \<subseteq> R' `` (set_fun_into_rel ((ma1 \<guillemotright> ma2)\<Down>SN) `` Sa)"
proof
  fix sc''
  assume "sc'' \<in> set_fun_into_rel ((mc1 \<guillemotright> mc2)\<Down>SN) `` Sc"
  hence "\<exists>ve' sc. sc \<in> Sc \<and> (Some ve', sc'') \<in> (mc1 \<guillemotright> mc2) sc" (is "\<exists>ve' sc. ?P ve' sc")
    by(force simp add: set_fun_into_rel_def unfold_monad projectS_defs 
      split: option.split_asm sum.split_asm)
  then obtain ve' sc where res: "?P ve' sc" by auto
  hence "sc \<in> Sc" ..
  note refD=ref[THEN sim_norm_extD]
  show "sc'' \<in> R' `` set_fun_into_rel ((ma1 \<guillemotright> ma2)\<Down>SN) `` Sa"
  proof(cases ve')
    case (Inl e)
    hence 
      "(Some (Inl e), sc'') \<in> mc1 sc 
      \<or> (\<exists>v sc'. (Some (Inr v), sc') \<in> mc1 sc \<and> (Some (Inl e), sc'') \<in> mc2 sc')" using res
      by(auto simp add: bind_exc_decompose)
    thus ?thesis
    proof(elim disjE)
      assume "(Some (Inl e), sc'') \<in> mc1 sc"
      hence "\<exists>sa sa''. sa \<in> Sa \<and> sa'' \<in> ((ma1\<Down>SNE) sa) \<and> (sa'', sc'') \<in> R'" using subNE `sc \<in> Sc`
        by(force simp add: set_fun_into_rel_def projectS_defs image_def Image_def Collect_subset)
      thus ?thesis
        apply(auto simp add: set_fun_into_rel_def projectS_defs image_def Image_def)
        by (metis bind_exc_decompose)
    next
      assume 
        "\<exists>v sc'. (Some (Inr v), sc') \<in> mc1 sc \<and> (Some (Inl e), sc'') \<in> mc2 sc'" (is 
        "\<exists>v sc'. ?P v sc'")
      then obtain v sc' where res': "?P v sc'" by auto
      hence "\<exists>sa sa'. sa \<in> Sa \<and> sa' \<in> ((ma1\<Down>SNN) sa) \<and> (sa', sc') \<in> R" using subNN `sc \<in> Sc`
        by(force simp add: set_fun_into_rel_def projectS_defs image_def Image_def Collect_subset)
      thus ?thesis using res' refD
        apply(auto simp add: set_fun_into_rel_def projectS_defs image_def Image_def force_same_and_def
          split: split_if_asm)
        by (metis bind_exc_decompose equals0D sum.distinct(1))
    qed
  next
    case (Inr v')
    hence 
      "(\<exists>v sc'. (Some (Inr v), sc') \<in> mc1 sc \<and> (Some (Inr v'), sc'') \<in> mc2 sc')" (is 
      "\<exists>v sc'. ?P v sc'")
      using res
      by(auto simp add: bind_norm_decompose)
    then obtain v sc' where res': "?P v sc'" by auto
    hence "\<exists>sa sa'. sa \<in> Sa \<and> sa' \<in> ((ma1\<Down>SNN) sa) \<and> (sa', sc') \<in> R" using subNN `sc \<in> Sc`
      by(force simp add: set_fun_into_rel_def projectS_defs image_def Image_def Collect_subset)
    thus ?thesis using res' refD
      apply(auto simp add: set_fun_into_rel_def projectS_defs image_def Image_def force_same_and_def
        split: split_if_asm)
      by (metis bind_norm_decompose equals0D sum.distinct(1))
  qed
qed

lemma seq_lift_ex_abs_state:
  "\<lbrakk> (ba, bc) \<in> seq_lift R; sc \<in> set bc \<rbrakk> \<Longrightarrow> \<exists>sa \<in> set ba. (sa, sc) \<in> R"
  by(induct rule: seq_lift.induct, auto)

lemma spec_ref_ex_abs_state:
  assumes 
  reach: "sc \<in> reach Sc"
  and spec_ref: "refines R pi Sa Sc"
  shows "\<exists>sa \<in> reach Sa. (sa, sc) \<in> R"
proof-
  from reach obtain bc where bc_def: "bc \<in> beh Sc \<and> sc \<in> set bc"
    by(auto dest: reach_in_beh)
  then obtain ba where ba_def: "ba \<in> beh Sa \<and> (ba, bc) \<in> seq_lift R" using spec_ref
    by(auto simp add: refines_def dest!: behaviour_refinement)
  from this[THEN conjunct2] and bc_def have "\<exists>sa \<in> set ba. (sa, sc) \<in> R"
    apply -
    apply(elim conjE, rule seq_lift_ex_abs_state, assumption+)
    done
  thus ?thesis using ba_def[THEN conjunct1]
    apply -
    apply(erule bexE, drule beh_in_reach, assumption, rule bexI, assumption+)
    done
qed

lemma ref_HoareN_preservation:
  assumes 
  abs_Hoare: "\<lbrace>obs Sa -` P\<rbrace> ma \<lbrace>>R \<lambda>v. obs Sa -` Q v\<rbrace>"
  and spec_ref: "refines R pi Sa Sc"
  and comp_ref: "sim_norm R ma mc"
  shows "\<lbrace>(obs Sc -` (pi -` P)) \<inter> reach Sc\<rbrace> mc \<lbrace>>R \<lambda>v. obs Sc -` (pi -` Q v)\<rbrace>"
proof(subst HoareN_alt, unfold Hoare_def Box_def, intro subsetI, simp, elim conjE, intro allI impI)
  fix sc v sc'
  assume 
    pre: "pi (obs Sc sc) \<in> P"
    and reach: "sc \<in> reach Sc"
    and res: "(v, sc') \<in> (mc\<Down>N) sc"
  from reach and spec_ref obtain sa where inR: "(sa, sc) \<in> R"
    by(auto dest: spec_ref_ex_abs_state)
  then have "\<exists>sa'. (Some v, sa') \<in> ma sa \<and> (sa', sc') \<in> R" (is "\<exists>sa'. ?P sa'") 
    using comp_ref res
    apply(case_tac v)
    apply(auto simp add: sims_through_simN_simS simN_def simN_wpc_def project_defs force_same_and_def
      split: split_if_asm dest!: set_mp)
    done
  then obtain sa' where sa'_def: "(Some v, sa') \<in> ma sa \<and> (sa', sc') \<in> R" ..
  have obs_eq: "pi (obs Sc sc) = obs Sa sa \<and> pi (obs Sc sc') = obs Sa sa'" 
    using spec_ref inR conjunct2[OF sa'_def]
    by(auto simp add: refines_def obs_consistent_def)
  hence sa'_obs: "obs Sa sa' \<in> Q v" using abs_Hoare sa'_def pre
    by(auto simp add: HoareN_alt Hoare_def Box_def projectN_def vimage_def Collect_subset)
  thus "pi (obs Sc sc') \<in> Q v" using obs_eq
    by(simp)
qed

lemma sim_norm_ext_imp_PO_rhoare:
  assumes ref: "sim_norm_ext R ma mc (force_same_and R') (force_same_and R')"
  shows "{ R } set_fun_into_rel (ma\<Down>SN), set_fun_into_rel (mc\<Down>SN) {> R' }"
proof(clarsimp simp add: PO_rhoare_def relcomp_unfold)
  fix sa sc sc'
  assume assms: "(sa, sc) \<in> R" "(sc, sc') \<in> set_fun_into_rel (mc\<Down>SN)"
  hence "\<exists>v. (Some v, sc') \<in> mc sc" (is "\<exists>v. ?P v")
    by(auto simp add: set_fun_into_rel_def projectSN_def projectS_def project_defs)
  then obtain v where v_def: "?P v" ..
  hence "\<exists>sa'. sa' \<in> (ma\<Down>SN) sa \<and> (sa', sc') \<in> R'" using ref `(sa, sc) \<in> R`
    apply(case_tac v)
    apply(force simp add: sim_norm_ext_def simN_wpc_def projectSN_def projectS_def project_defs 
      force_same_and_def image_def
      split: split_if_asm)+
    done
  thus "\<exists>sa'. (sa, sa') \<in> set_fun_into_rel (ma\<Down>SN) \<and> (sa', sc') \<in> R'"
    by(auto simp add: set_fun_into_rel_def)
qed

lemma reach_TS_extend:
  "reach (TS.extend ts (spec.fields obs_f)) = reach ts"
proof(safe)
  fix s
  assume "s \<in> reach (TS.extend ts (spec.fields obs_f))"
  thus "s \<in> reach ts"
    apply(induct rule: reach.induct)
    apply(auto simp add: TS.extend_def spec.fields_def)
    done
next
  fix s
  assume "s \<in> reach ts" 
  thus "s \<in> reach (TS.extend ts (spec.fields obs_f))"
    apply(induct rule: reach.induct)
    apply(auto simp add: TS.extend_def spec.fields_def)
    by (metis TS.select_convs(1) TS.select_convs(2) reach.r_init rev_ImageI)
qed

lemma internal_to_external:
  "\<lbrakk> reach ts \<subseteq> I; obs_f_f ` I \<subseteq> J \<rbrakk> \<Longrightarrow>
    oreach (TS.extend ts (spec.fields obs_f_f)) \<subseteq> J"
  apply(auto simp add: oreach_def reach_TS_extend)
  apply(auto simp add: TS.extend_def spec.fields_def image_def)
  done

end
