theory Pmm_Tacs
  imports RMonad_Lemmas Misc_Lemmas PmmDefs Projections
begin

ML_file "pmm_library.ML" 
ML_file "pmm_tacs.ML"
declaration {* K Pmm_Tacs.ex_eq_simproc_setup *}

ML {*
  val TRY_SOLVE_ALL = TRYALL o SOLVED'
*}

ML {*
  fun clasimp_method' tac =
    Method.sections clasimp_modifiers >> K (SIMPLE_METHOD o tac);

*}

setup {*

  Method.setup @{binding clarsimp_all} (clasimp_method' (
    CHANGED_PROP o PARALLEL_GOALS o ALLGOALS o clarsimp_tac))
    "clarify simplified, all goals";

*}

ML {*
  fun iterative_clarsimp_tac ctxt =
  let
    val ss = simpset_of ctxt

    fun set_sdl sdl ss = simpset_of (Config.put simp_depth_limit sdl (Simplifier.the_context ss))

    val ss0 = set_sdl 0 ss
    val ss1 = set_sdl 1 ss

  in 
    distinct_subgoals_tac
    THEN
    (REPEAT
    (
        (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE}))))
        ORELSE
        (REPEAT1 (CHANGED (TRYALL (etac @{thm conjE}))))
        ORELSE
        (REPEAT1 (CHANGED (TRYALL (etac @{thm exE}))))
    ))
    THEN
    distinct_subgoals_tac
    THEN
    (IF_UNSOLVED ((PARALLEL_GOALS o ALLGOALS) (asm_full_simp_tac ss0)))
    THEN
    distinct_subgoals_tac
    THEN
    (IF_UNSOLVED ((PARALLEL_GOALS o ALLGOALS)(asm_full_simp_tac ss1)))
    THEN
    distinct_subgoals_tac
    THEN
    (IF_UNSOLVED ((PARALLEL_GOALS o ALLGOALS) (clarsimp_tac (Simplifier.the_context ss0))))
    THEN
    distinct_subgoals_tac
    THEN
    (IF_UNSOLVED ((PARALLEL_GOALS o ALLGOALS) (clarsimp_tac (Simplifier.the_context ss1))))
  end

*}

setup {*

  Method.setup @{binding iterative_simp} (clasimp_method' (CHANGED o iterative_clarsimp_tac))
    "clarify simplified, all goals";

*}

lemma exceptions_cases_f:
  "f (exceptions_case m1 m2 m3 m4 m7 m8 m9 e) =
    exceptions_case (f m1) (f m2) (f m3) (f m4) (\<lambda>v. f (m7 v)) (f m8) (f m9) e"
  by(auto split: exceptions.split)

lemmas projectR_exceptions_cases = exceptions_cases_f[where f=projectR]
lemmas projectN_exceptions_cases = exceptions_cases_f[where f=projectN]
lemmas projectNN_exceptions_cases = exceptions_cases_f[where f=projectNN]
lemmas projectNE_exceptions_cases = exceptions_cases_f[where f=projectNE]

lemma exceptions_case_choice:
  "exceptions_case (m1 [+] m1') m2 m3 m4 m7 m8 m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1' m2 m3 m4 m7 m8 m9 e)"
  "exceptions_case m1 (m2 [+] m2') m3 m4 m7 m8 m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2' m3 m4 m7 m8 m9 e)"
  "exceptions_case m1 m2 (m3 [+] m3') m4 m7 m8 m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2 m3' m4 m7 m8 m9 e)"
  "exceptions_case m1 m2 m3 (m4 [+] m4') m7 m8 m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2 m3 m4' m7 m8 m9 e)"
  "exceptions_case m1 m2 m3 m4 (\<lambda>v. m7 v [+] m7' v) m8 m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2 m3 m4 m7' m8 m9 e)"
  "exceptions_case m1 m2 m3 m4 m7 (m8 [+] m8') m9
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2 m3 m4 m7 m8' m9 e)"
  "exceptions_case m1 m2 m3 m4 m7 m8 (m9 [+] m9')
    = (\<lambda>e. exceptions_case m1 m2 m3 m4 m7 m8 m9 e
        [+] exceptions_case m1 m2 m3 m4 m7 m8 m9' e)"
  by(rule_tac[!] ext, case_tac[!] e, simp_all add: choice_absorb)

lemmas exceptions_case_choice' = exceptions_case_choice[THEN fun_cong]

lemma exceptions_case_cancel: "exceptions_case x x x x (\<lambda>_. x) x x = (\<lambda>_. x)"
  apply(rule ext)
  apply(simp add: K_def split: exceptions.split)
  done

definition throws_only :: "('a, 'e, 's)re_monad \<Rightarrow> 'e set \<Rightarrow> bool" where
  "throws_only m exs \<equiv> \<forall>e s s'. (Some (Inl e), s') \<in> m s \<longrightarrow> e \<in> exs"

definition doesnt_throw_raw where
  "doesnt_throw_raw m e \<equiv> (\<forall>s. e \<notin> fst ` m s)"

definition doesnt_throw where
  "doesnt_throw m e \<equiv> doesnt_throw_raw (m\<Down>NE) e"

text {*
  Looper - do not add to simpset directly! Using the simproc instead (which uses
  the lemma in turn)
*}

lemma doesnt_throw_raw_bindN:
  "doesnt_throw_raw m FailWrite \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case empty (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e)"
  "doesnt_throw_raw m FailLLRead \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) empty (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e)"
  "doesnt_throw_raw m FailRead \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) empty (m4 e) (m7 e) (m8 e) (m9 e) e)"
  "doesnt_throw_raw m FailMode \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) empty (m7 e) (m8 e) (m9 e) e)"
  "\<forall>c. doesnt_throw_raw m (Degraded c) \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e)  (\<lambda>_. empty) (m8 e) (m9 e) e)"
  "doesnt_throw_raw m InvalidInput \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) empty (m9 e) e)"
  "doesnt_throw_raw m StartedExc \<Longrightarrow> 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e) e) = 
  bindN m (\<lambda>e. exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) empty e)"
  apply(force simp add: unfold_monad doesnt_throw_raw_def
    split: split_if exceptions.split split_if_asm exceptions.split_asm
    intro!: ext
  )+
  done

lemma doesnt_throw_if: "\<lbrakk> doesnt_throw m1 e; doesnt_throw m2 e \<rbrakk> \<Longrightarrow> 
    doesnt_throw (if c then m1 else m2) e"
  by(simp)

lemma doesnt_throw_raw_if: "\<lbrakk> doesnt_throw_raw m1 e; doesnt_throw_raw m2 e \<rbrakk> \<Longrightarrow> 
    doesnt_throw_raw (if c then m1 else m2) e"
  by(simp)

lemma doesnt_throw_raw_choice:
  "doesnt_throw_raw (m [+] m') e = (doesnt_throw_raw m e \<and> doesnt_throw_raw m' e)"
  by(auto simp add: doesnt_throw_raw_def choice_def)

lemma doesnt_throw_choice:
  "doesnt_throw (m [+] m') e = (doesnt_throw m e \<and> doesnt_throw m' e)"
  by(simp add: doesnt_throw_def project_simps doesnt_throw_raw_choice)

lemma exc_case_pred:
  "\<lbrakk> P m1; P m2; P m3; P m4; \<forall>a. P (m5 a); P m6; P m7 \<rbrakk> \<Longrightarrow>
  P (exceptions_case m1 m2 m3 m4 m5 m6 m7 e)"
  by(auto split: exceptions.split)


lemma doesnt_throw_exceptions_case: "
    \<lbrakk> doesnt_throw m1 e; doesnt_throw m2 e; doesnt_throw m3 e;
        doesnt_throw m4 e; 
        \<forall>c. doesnt_throw (m7 c) e; 
        doesnt_throw m8 e; doesnt_throw m9 e
  \<rbrakk> \<Longrightarrow>
    doesnt_throw (exceptions_case m1 m2 m3 m4 m7 m8 m9 e') e"
  by(cases e', auto)

lemma doesnt_throw_raw_exceptions_case: "
    \<lbrakk> doesnt_throw_raw m1 e; doesnt_throw_raw m2 e; doesnt_throw_raw m3 e;
        doesnt_throw_raw m4 e; 
        \<forall>c. doesnt_throw_raw (m7 c) e;
        doesnt_throw_raw m8 e; doesnt_throw_raw m9 e
  \<rbrakk> \<Longrightarrow>
    doesnt_throw_raw (exceptions_case m1 m2 m3 m4 m7 m8 m9 e') e"
  by(cases e', auto)

lemma doesnt_throw_empty: "doesnt_throw empty e"
  by(clarsimp simp add: empty_def doesnt_throw_def doesnt_throw_raw_def project_defs K_def)

lemma doesnt_throw_raw_empty: "doesnt_throw_raw empty e"
  by(clarsimp simp add: empty_def doesnt_throw_raw_def project_defs K_def)

lemma doesnt_throw_return: "doesnt_throw (return x) e"
  by(auto simp add: doesnt_throw_def doesnt_throw_raw_def project_defs
    throw_def return_def returnR_def returnN_def returnS_def)

lemma returnN_throws_only_one: "e \<noteq> e' \<Longrightarrow> doesnt_throw_raw (returnN e) e'"
  by(auto simp add: doesnt_throw_raw_def returnN_def returnS_def)

lemma throw_throws_only_one: "e \<noteq> e' \<Longrightarrow> doesnt_throw (throw e) e'"
  by(auto simp add: doesnt_throw_def doesnt_throw_raw_def project_defs
    throw_def returnR_def returnN_def returnS_def)

lemma doesnt_throw_raw_bindN': "\<forall>x. doesnt_throw_raw (f x) e \<Longrightarrow> doesnt_throw_raw (m \<guillemotright>=N f) e"
  by(auto simp add: doesnt_throw_raw_def unfold_monad)

lemmas project_aux = 
  projectR_exceptions_cases
  projectN_exceptions_cases
  projectNN_exceptions_cases 
  projectNE_exceptions_cases

lemmas first_stage =
  monad_laws
  monad_simps
  project_simps
  project_aux
  exceptions_case_cancel
  empty_simps
  choice_distrib
  if_cancel
  K_def
  doesnt_throw_def
  choice_assoc
  exceptions_case_choice

lemmas bindN_lcong = arg_cong2[where f="bindN", OF _ refl]

ML_file "no_throw.ML"

declare doesnt_throw_raw_if[no_throw]
declare doesnt_throw_raw_exceptions_case[no_throw]
declare doesnt_throw_empty[no_throw]
declare doesnt_throw_raw_empty[no_throw]
declare returnN_throws_only_one[no_throw]
declare throw_throws_only_one[no_throw]

simproc_setup throw_simproc 
  ("m \<guillemotright>=N (\<lambda>e. (exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e)) e)")
  = {* K (fn ss => fn ctrm =>
        (No_Throw.throw_simproc' @{type_name "exceptions"} ss (term_of ctrm))) *}

simproc_setup throw_simproc' 
  ("m \<guillemotright>=N exceptions_case m1' m2' m3' m4' m7' m8' m9'")
  = {* K (fn ss => fn ctrm =>
        (No_Throw.throw_simproc' @{type_name "exceptions"} ss (term_of ctrm))) *}

simproc_setup throw_bind_simproc
  ("m \<guillemotright>=N (\<lambda>e. (exceptions_case (m1 e) (m2 e) (m3 e) (m4 e) (m7 e) (m8 e) (m9 e)) e \<guillemotright>=N n)")
  = {* K (fn ss => fn ctrm =>
        (No_Throw.throw_simproc' @{type_name "exceptions"} ss (term_of ctrm))) *}

simproc_setup throw_bind_simproc'
 ("m \<guillemotright>=N (\<lambda>e. (exceptions_case m1 m2 m3 m4 m7 m8 m9) \<guillemotright>=N n e)")
  = {* K (fn ss => fn ctrm =>
        (No_Throw.throw_simproc' @{type_name "exceptions"} ss (term_of ctrm))) *}

declare [[simproc del: throw_bind_simproc']]

end
