theory Projections
imports RMonad RMonad_Lemmas RMonad_Decomposition
begin

definition projectN :: "('a, 's)r_monad \<Rightarrow> ('a, 's)nds_monad" ("(_\<Down>N)" [1000] 999) where
  "projectN m \<equiv> \<lambda>s. {(r, s'). (Some r, s') \<in> m s}"

definition projectR :: "('a, 's)r_monad \<Rightarrow> ('s \<Rightarrow> 's set)" ("_\<Down>R" [1000] 999) where
  "projectR m \<equiv> \<lambda>s. {s'. (None, s') \<in> m s}"

definition projectNN :: "('a, 'e, 's)re_monad \<Rightarrow> ('a, 's)nds_monad" ("(_\<Down>NN)" [1000] 999) where
  "projectNN m \<equiv> \<lambda>s. {(r, s'). (Some (Inr r), s') \<in> m s}"

definition projectNE :: "('a, 'e, 's)re_monad \<Rightarrow> ('e, 's)nds_monad" ("(_\<Down>NE)" [1000] 999) where
  "projectNE m \<equiv> \<lambda>s. {(e, s'). (Some (Inl e), s') \<in> m s}"

lemmas project_defs =
  projectN_def
  projectR_def
  projectNN_def
  projectNE_def

subsection  {* projectN/projectR *}

lemma projectR_returnR: "(returnR x)\<Down>R = empty"
  apply(rule ext)
  apply(simp add: projectR_def returnR_def returnN_def returnS_def empty_def K_def)
  done

lemma bindR_empty: "m \<guillemotright>=R (\<lambda>x. empty) = (\<lambda>s. {None} \<times> (m\<Down>R) s)"
  apply(rule ext)
  apply(auto simp add: empty_def K_def bindR_def liftR_def bindN_def bindS_def restart_def 
    returnN_def returnS_def projectR_def split: option.split_asm)
  done

lemma projectN_returnR: "(returnR x)\<Down>N = returnN x"
  apply(rule ext)
  apply(simp add: projectN_def returnR_def returnN_def returnS_def empty_def K_def)
  done

lemma projectR_restart: "restart\<Down>R = returnS"
  apply(rule ext)
  apply(simp add: projectR_def restart_def returnN_def returnS_def)
  done

lemma projectN_restart: "restart\<Down>N = empty"
  apply(rule ext)
  apply(simp add: projectN_def restart_def returnN_def returnS_def empty_def K_def)
  done

lemma projectN_choice:
  "(m1 [+] m2)\<Down>N = m1\<Down>N [+] m2\<Down>N"
  by(auto simp add: choice_def projectN_def)

lemma projectR_choice:
  "(m1 [+] m2)\<Down>R = (m1\<Down>R) [+] (m2\<Down>R)"
  by(auto simp add: choice_def projectR_def)

lemma bindR_projectN: "(m1 \<guillemotright>=R m2)\<Down>N = ((m1\<Down>N) \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>N))"
  apply(rule ext)
  apply(force simp add: projectN_def bindR_some_decompose bindR_def liftR_def restart_def 
    returnN_def returnS_def bindN_def bindS_def
    split: option.split option.split_asm)
  done

lemma bindR_projectR: "(m1 \<guillemotright>=R m2)\<Down>R = (m1\<Down>R) [+] ((m1\<Down>N) \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>R))"
  apply(rule ext)
  apply(clarsimp simp add: projectR_def projectN_def bindR_none_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    returnS_def bindS_def
    split: option.split option.split_asm)
  done

lemma try_catchR_projectN:
  "(try_catchR m1 m2)\<Down>N = m1\<Down>N [+] (\<lambda>s. ((m1\<Down>R) s) \<guillemotright>=S m2\<Down>N)"
  apply(rule ext)
  apply(clarsimp simp add: projectR_def projectN_def try_catchR_some_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    try_catchR_def liftR2_def
    returnS_def bindS_def
    split: option.split option.split_asm)
  done

lemma try_catchR_projectR:
  "(try_catchR m1 m2)\<Down>R = (\<lambda>s. (m1\<Down>R) s \<guillemotright>=S m2\<Down>R)"
  apply(rule ext)
  apply(clarsimp simp add: projectR_def projectN_def try_catchR_none_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    returnS_def bindS_def
    try_catchR_def liftR2_def
    split: option.split option.split_asm)
  done

(*
lemma try_catchR_bindR_projectN:
  "(try_catchR m (h1 \<guillemotright>=R (\<lambda>x. h2 x)))\<Down>N = m\<Down>N [+] ((try_catchR (m\<Down>R) (h1\<Down>N)) \<guillemotright>=R (\<lambda>x. (h2 x)\<Down>N))"
  apply(auto simp add: try_catchR_projectN bindR_projectN)
  apply(rule choice_left_I)
  apply(rule ext)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    try_catchR_def liftR2_def projectN_def projectR_def
    
    split: option.split option.split_asm)
  done
*)

subsection {* projectNN/projectNE *}

(*
lemma projectNN_idem: "(m\<Down>NN)\<Down>NN = m\<Down>NN"
  by(auto simp add: projectNN_def)

lemma projectNE_idem: "(m\<Down>NE)\<Down>NE = m\<Down>NE"
  by(auto simp add: projectNE_def)
*)

(*
lemma project_decomposeN: "m\<Down>N = m\<Down>NN [+] m\<Down>NE"
  apply(rule ext)
  apply(auto simp add: projectN_def projectNE_def projectNN_def choice_def)
  apply(case_tac r)
  apply(auto)
  done
*)

lemma projectNN_choice:
  "(m1 [+] m2)\<Down>NN = m1\<Down>NN [+] m2\<Down>NN"
  by(auto simp add: projectNN_def choice_def)

lemma projectNE_choice:
  "(m1 [+] m2)\<Down>NE = m1\<Down>NE [+] m2\<Down>NE"
  by(auto simp add: projectNE_def choice_def)

(*
lemma bind_empty:
  "bind m (\<lambda>x. empty) = (\<lambda>s. {None} \<times> ((m\<Down>R) s)) [+] (\<lambda>s. (m\<Down>NE) s)"
  apply(rule ext)
  apply(auto simp add: 
    projectNE_def projectR_def
    bind_def liftE_def bindR_def liftR_def bindN_def 
    throw_def returnR_def restart_def returnN_def
    choice_def
    split: sum.split_asm option.split_asm)
  done
*)

lemma projectR_return: "(return x)\<Down>R = empty"
  by(simp add: return_def projectR_returnR empty_def)

(*
lemma projectN_return: "(return x)\<Down>N = returnN x"
  by(simp add: return_def projectN_returnR)
*)

lemma projectNN_return: "(return x)\<Down>NN = returnN x"
  by(auto simp add: return_def returnR_def returnN_def returnS_def projectNN_def)

lemma projectNE_return: "(return x)\<Down>NE = empty"
  by(auto simp add: return_def returnR_def returnN_def returnS_def projectNE_def empty_def K_def)

lemma projectR_throw:
  "(throw x)\<Down>R = empty"
  by(auto simp add: throw_def projectR_returnR)

(*
lemma projectN_throw:
  "(throw x)\<Down>N = throw x"
  by(auto simp add: throw_def projectN_returnR)
*)

lemma projectNE_throw:
  "(throw x)\<Down>NE = returnN x"
  by(auto simp add: throw_def returnR_def returnN_def returnS_def projectNE_def)

lemma projectNN_throw:
  "(throw x)\<Down>NN = empty"
  by(auto simp add: throw_def returnR_def returnN_def returnS_def projectNN_def empty_def K_def)

lemma projectNN_restart:
  "restart\<Down>NN = empty"
  by(auto simp add: projectNN_def restart_def returnN_def returnS_def empty_def K_def)

lemma projectNE_restart:
  "restart\<Down>NE = empty"
  by(auto simp add: projectNE_def restart_def returnN_def returnS_def empty_def K_def)

lemma liftE_projectR: "m1\<Down>N \<guillemotright>=N (\<lambda>x. (liftE m2 x)\<Down>R) = m1\<Down>NN \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>R)"
  apply(rule ext)
  apply(force simp add: liftE_def bindR_def bindN_def liftR_def 
    throw_def restart_def
    returnR_def returnN_def
    returnS_def bindS_def
    projectN_def projectNN_def projectR_def
    split: option.split sum.split sum.split_asm)
  done

lemma bindR_projectNN: "(m1 \<guillemotright>=R m2)\<Down>NN = (
    ((m1\<Down>NN) \<guillemotright>=N (\<lambda>x. (m2 (Inr x))\<Down>NN))
    [+] ((m1\<Down>NE) \<guillemotright>=N (\<lambda>x. (m2 (Inl x))\<Down>NN))
  )"
  apply(rule ext)
  apply(auto simp add: unfold_monad projectNN_def projectNE_def split: option.split)
  apply(case_tac aa)
  apply(force)
  apply(case_tac ab)
  apply(force)+
  done

lemma bindR_projectNE: "(m1 \<guillemotright>=R m2)\<Down>NE = (
    ((m1\<Down>NN) \<guillemotright>=N (\<lambda>x. (m2 (Inr x))\<Down>NE))
    [+] ((m1\<Down>NE) \<guillemotright>=N (\<lambda>x. (m2 (Inl x))\<Down>NE))
  )"
  apply(rule ext)
  apply(auto simp add: unfold_monad projectNN_def projectNE_def split: option.split)
  apply(case_tac aa)
  apply(force)
  apply(case_tac ab)
  apply(force)+
  done

lemma bind_projectR: "(m1 \<guillemotright>= m2)\<Down>R = (m1\<Down>R) [+] ((m1\<Down>NN) \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>R))"
  by(simp add: bind_def bindR_projectR liftE_projectR)

lemma bind_projectN: "(bind m1 m2)\<Down>N = (bindN (m1\<Down>N) (\<lambda>x. (liftE m2 x)\<Down>N))"
  by(simp add: bind_def bindR_projectN)

lemma bind_projectNN: "(m1 \<guillemotright>= m2)\<Down>NN = ((m1\<Down>NN) \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>NN))"
  apply(rule ext)
  apply(force simp add: bind_def projectNN_def 
    bindR_def bindN_def
    returnS_def bindS_def
    liftE_def liftR_def
    returnR_def returnN_def throw_def restart_def
    split: option.split_asm option.split sum.split_asm sum.split
  )
  done

lemma bind_projectNE: "(m1 \<guillemotright>= m2)\<Down>NE = m1\<Down>NE [+] ((m1\<Down>NN) \<guillemotright>=N (\<lambda>x. (m2 x)\<Down>NE))"
  apply(rule ext)
  apply(force simp add: bind_def projectNE_def projectNN_def 
    choice_def
    bindR_def bindN_def
    liftE_def liftR_def
    returnS_def bindS_def
    returnR_def returnN_def throw_def restart_def
    split: option.split_asm option.split sum.split_asm sum.split
  )
  done

lemma try_catch_projectNN:
  "(try_catch m1 m2)\<Down>NN = m1\<Down>NN [+] ((m1\<Down>NE) \<guillemotright>=N (\<lambda>e. (m2 e)\<Down>NN))"
  apply(rule ext)
  apply(simp add: projectNN_def projectNE_def try_catch_norm_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    try_catch_def liftE2_def
    returnS_def bindS_def
    split: option.split option.split_asm sum.split sum.split_asm)
  done

lemma try_catch_projectNE:
  "(try_catch m1 m2)\<Down>NE = ((m1\<Down>NE) \<guillemotright>=N (\<lambda>e. (m2 e)\<Down>NE))"
  apply(rule ext)
  apply(simp add: projectNN_def projectNE_def try_catch_exc_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    try_catch_def liftE2_def
    returnS_def bindS_def
    split: option.split option.split_asm sum.split sum.split_asm)
  done

(*
lemma try_catch_projectN:
  "(try_catch m1 m2)\<Down>N = m1\<Down>NN [+] ((m1\<Down>NE) \<guillemotright>=N (\<lambda>e. (m2 e)\<Down>NN)) [+] 
    ((m1\<Down>NE) \<guillemotright>=N (\<lambda>e. (m2 e)\<Down>NE))"
  by(simp add: project_decomposeN try_catch_projectNN try_catch_projectNE choice_assoc)
*)

lemma try_catch_projectR:
  "(try_catch m1 m2)\<Down>R = m1\<Down>R [+] ((m1\<Down>NE) \<guillemotright>=N (\<lambda>e. (m2 e)\<Down>R))"
  apply(rule ext)
  apply(simp add: projectR_def projectNE_def try_catch_none_decompose)
  apply(force simp add: choice_def bindR_def liftR_def restart_def returnN_def bindN_def
    try_catch_def liftE2_def
    returnS_def bindS_def 
    split: option.split option.split_asm sum.split sum.split_asm)
  done

term "bindN"

lemma try_catchR_projectNN: "(try_catchR m1 m2)\<Down>NN = m1\<Down>NN [+] (\<lambda>s. ((m1\<Down>R) s \<guillemotright>=S (m2\<Down>NN)))"
  apply(rule ext)
  apply(simp add: projectNN_def projectR_def try_catchR_some_decompose bindS_def)
  apply(force simp add: try_catchR_def liftR2_def choice_def split: option.split)
  done

lemma try_catchR_projectNE: "(try_catchR m1 m2)\<Down>NE = m1\<Down>NE [+] (\<lambda>s. (m1\<Down>R) s \<guillemotright>=S (m2\<Down>NE))"
  apply(rule ext)
  apply(simp add: projectNE_def projectR_def try_catchR_some_decompose bindS_def)
  apply(force simp add: try_catchR_def liftR2_def choice_def split: option.split)
  done

lemma projectN_if_then_else:
  "(if b then m1 else m2)\<Down>N = (if b then m1\<Down>N else m2\<Down>N)"
  by(simp)

lemma projectR_if_then_else:
  "(if b then m1 else m2)\<Down>R = (if b then m1\<Down>R else m2\<Down>R)"
  by(simp)

lemma projectNN_if_then_else:
  "(if b then m1 else m2)\<Down>NN = (if b then m1\<Down>NN else m2\<Down>NN)"
  by(simp)

lemma projectNE_if_then_else:
  "(if b then m1 else m2)\<Down>NE = (if b then m1\<Down>NE else m2\<Down>NE)"
  by(simp)


lemma projectN_empty:
  "empty\<Down>N = empty"
  by(auto simp add: empty_def K_def projectN_def)

lemma projectR_empty:
  "empty\<Down>R = empty"
  by(auto simp add: empty_def K_def projectR_def)

lemma projectNN_empty:
  "empty\<Down>NN = empty"
  by(auto simp add: empty_def K_def projectNN_def)

lemma projectNE_empty:
  "empty\<Down>NE = empty"
  by(auto simp add: empty_def K_def projectNE_def)

lemma mapM'_projectR:
  "l \<noteq> [] \<Longrightarrow> (mapM' f l)\<Down>R = doN {
    n \<leftarrow> choose {0..(length l - 1)};
    mapM'_N (\<lambda>x. (f x)\<Down>NN) (take n l);
    (f (l ! n))\<Down>R
  }"
proof(induct l) 
  case Nil
  thus ?case
    by simp
next
  note mapM'.simps[simp del]
  note mapM'_N.simps[simp del]

  case (Cons x xs)
  note ind_hyp=this
  thus ?case 
  proof-
    have "(mapM' f (x # xs))\<Down>R = (f x)\<Down>R [+] ((f x)\<Down>NN \<guillemotright>N (mapM' f xs)\<Down>R)"
      by(simp add: mapM'_Cons bind_projectR)
    moreover have "...=doN {
      n \<leftarrow> choose {0..length (x # xs) - 1};
      mapM'_N (\<lambda>x. (f x)\<Down>NN) (take n (x # xs)) \<guillemotright>N (f ((x # xs) ! n))\<Down>R
    }" (is "?L = ?R")
    proof(cases xs)
      case Nil
      thus ?thesis
        by(clarsimp simp add: mapM'_Nil mapM'_N_Nil choose_def 
          unfold_monad project_defs)
    next
      case (Cons y ys)
      have "xs \<noteq> []" using Cons by simp
      note rew1 = ind_hyp(1)[OF this]
      {
        fix s
        have "((f x)\<Down>NN \<guillemotright>N (mapM' f xs)\<Down>R) s = 
          (doN {
            n \<leftarrow> choose {0..length xs - 1};
            (mapM'_N (\<lambda>x. (f x)\<Down>NN) (take (Suc n) (x # xs)) \<guillemotright>N (f (xs ! n))\<Down>R)
          }) s"
          by(simp add: rew1 choose_comm monad_laws mapM'_N_Cons)
      } note rew2 = this
      show ?thesis
      proof(rule ext, simp add: choice_def, rule equalityI, rule Un_least)
        fix s
        show "((f x)\<Down>R) s
          \<subseteq> (doN {
            n \<leftarrow> choose {0..length xs};
            mapM'_N (\<lambda>x. (f x)\<Down>NN) (take n (x # xs)) \<guillemotright>N (f ((x # xs) ! n))\<Down>R
          }) s"
          apply(auto simp add: projectR_def choose_def unfold_monad mapM'_Nil mapM'_N_Nil)
          apply(rule_tac x=0 in bexI)
          apply(auto simp add: mapM'_N_Nil return_defs)
          done
      next
        fix s
        show "((f x)\<Down>NN \<guillemotright>N (mapM' f xs)\<Down>R) s
              \<subseteq> (doN {
                n \<leftarrow> choose {0..length xs};
                mapM'_N (\<lambda>x. (f x)\<Down>NN) (take n (x # xs)) \<guillemotright>N (f ((x # xs) ! n))\<Down>R
              })
              s"
          apply(simp add: rew2)
          apply(clarsimp simp add: mapM'_Nil projectR_return choose_def unfold_monad)
          apply(rule_tac x="Suc a" in bexI, auto simp add: Cons)[]
          done
      next
        fix s 
        show "(doN {
            n \<leftarrow> choose {0..length xs};
            mapM'_N (\<lambda>x. (f x)\<Down>NN) (take n (x # xs)) \<guillemotright>N (f ((x # xs) ! n))\<Down>R
          })
          s
         \<subseteq> ((f x)\<Down>R) s \<union> ((f x)\<Down>NN \<guillemotright>N (mapM' f xs)\<Down>R) s"
          apply(simp only: rew2)
          apply(clarsimp simp add: mapM'_Nil projectR_return choose_def unfold_monad)
          apply(case_tac a)
          apply(clarsimp simp add: mapM'_N_Nil  mapM'_Nil projectR_return choose_def unfold_monad)
          apply(force)
          done
      qed
    qed
    ultimately show ?thesis 
      by simp
  qed
qed


(*
lemma try_catchR_empty:
  "(try_catchR m empty) = m\<Down>N"
  apply(rule ext)
  apply(auto simp add: try_catchR_def liftR2_def restart_def returnN_def empty_def K_def projectN_def
  split: option.split_asm)
  done
*)

(*
lemma try_catch_empty: "try_catch m (\<lambda>e. empty) = m\<Down>R [+] m\<Down>NN"
  apply(rule ext)
  apply(auto simp add: projectR_def projectNN_def unfold_monad
    split: sum.split_asm option.split_asm)
  done
*)

(*
lemma project_cancel: 
  "(m\<Down>R)\<Down>N = empty"
  "(m\<Down>N)\<Down>R = empty"
  "(m\<Down>NN)\<Down>R = empty"
  "(m\<Down>NE)\<Down>R = empty"
  "(m\<Down>R)\<Down>NN = empty"
  "(m\<Down>R)\<Down>NE = empty"
  "(m\<Down>NN)\<Down>N = (m\<Down>NN)"
  "(m\<Down>NE)\<Down>N = (m\<Down>NE)"
  "(m\<Down>NN)\<Down>NE = empty"
  "(m\<Down>NE)\<Down>NN = empty"
  by(auto simp add: empty_def K_def 
    projectR_def projectN_def projectNN_def projectNE_def)
*)

(*
lemma try_catchR_projectN_bind:
  "try_catchR (m1\<Down>N \<guillemotright>= m2) h = (m1\<Down>N \<guillemotright>= (\<lambda>x. try_catchR (m2 x) h))"
  apply(rule ext)
  apply(auto simp add: unfold_monad project_defs
    split: option.split option.split_asm sum.split sum.split_asm)
  done
*)

(*
lemma try_catchR_projectNN_bind:
  "try_catchR (m1\<Down>NN \<guillemotright>= m2) h = (m1\<Down>NN \<guillemotright>= (\<lambda>x. try_catchR (m2 x) h))"
  apply(rule ext)
  apply(simp only: unfold_monad project_defs)
  apply(force)
  done
*)

(*
lemma try_catchR_projectNE_bind:
  "try_catchR (m1\<Down>NE \<guillemotright>= m2) h = m1\<Down>NE"
  apply(rule ext)
  apply(force simp add: unfold_monad project_defs
    split: option.split option.split_asm sum.split sum.split_asm)
  done
*)

(*
lemma None_notin_projectN: "(None, x) \<notin> projectN m s"
  by(simp add: projectN_def)

lemma None_notin_projectNN: "(None, x) \<notin> projectNN m s"
  by(simp add: projectNN_def)

lemma None_notin_projectNE: "(None, x) \<notin> projectNE m s"
  by(simp add: projectNE_def)

lemma exc_notin_projectNN: "(Some (Inl e), s') \<notin> projectNN m s"
  by(simp add: projectNN_def)

lemma norm_notin_projectNE: "(Some (Inr e), s') \<notin> projectNE m s"
  by(simp add: projectNE_def)

lemma try_catch_projectNN_bind:
  "try_catch (m1\<Down>NN \<guillemotright>= m2) h = (m1\<Down>NN \<guillemotright>= (\<lambda>x. try_catch (m2 x) h))"
  apply(rule ext, rule equalityI, rule_tac[!] subsetI, 
    rename_tac[!] res_pair, case_tac[!] res_pair, rename_tac[!] r s, case_tac[!] r)
  apply(case_tac[2] a)
  apply(case_tac[5] a)
  apply(force simp add: bind_none_decompose bind_norm_decompose bind_exc_decompose 
    try_catch_none_decompose try_catch_norm_decompose try_catch_exc_decompose
    exc_notin_projectNN norm_notin_projectNE None_notin_projectNE None_notin_projectNN)+
  done

lemma try_catch_projectNE_bind:
  "try_catchR (m1\<Down>NE \<guillemotright>= (m2 x)) h = m1\<Down>NE"
  apply(rule ext)
  apply(force simp add: unfold_monad project_defs
    split: option.split option.split_asm sum.split sum.split_asm)
  done
*)

lemma chooseN_project_simps:
  "(chooseN A)\<Down>NN = choose A"
  "(chooseN A)\<Down>R = empty"
  "(chooseN A)\<Down>NE = empty"
  by(auto simp add: unfold_monad project_defs chooseN_def choose_def)

lemmas project_simps =
(*
  projectR_idem
  projectN_idem
*)
  bindR_empty
  projectR_returnR
  projectN_returnR
  projectR_restart
  projectN_restart
  projectN_choice
  projectR_choice
  bindR_projectN
  bindR_projectR
  bindR_projectNN
  bindR_projectNE
  try_catchR_projectN
  try_catchR_projectR
  (* try_catchR_bindR_projectN *)
(*
  projectNN_idem
  projectNE_idem
  bind_empty
*)
  projectNN_choice
  projectNE_choice
  projectR_return
(*
  projectN_return
*)
  projectNN_return
  projectNE_return
  projectR_throw
(*
  projectN_throw
*)
  projectNE_throw
  projectNN_throw
  projectNN_restart
  projectNE_restart
  bind_projectR
  bind_projectN
  bind_projectNN
  bind_projectNE
  try_catch_projectNN
  try_catch_projectNE
(*
  try_catch_projectN
*)
  try_catch_projectR
  try_catchR_projectNN
  try_catchR_projectNE
  projectN_if_then_else
  projectR_if_then_else
  projectNN_if_then_else
  projectNE_if_then_else
  projectN_empty
  projectR_empty
  projectNN_empty
  projectNE_empty
  chooseN_project_simps
(*
  try_catchR_empty
  try_catch_empty
  project_cancel
  try_catchR_projectN_bind
  try_catchR_projectNN_bind
  try_catchR_projectNE_bind
  try_catch_projectNN_bind
*)

lemma projectNN_mono: "(\<forall>x. m x \<subseteq> m' x) \<Longrightarrow> (m\<Down>NN) s \<subseteq> (m'\<Down>NN) s"
  by(auto simp add: project_defs)

lemma projectNE_mono: "(\<forall>x. m x \<subseteq> m' x) \<Longrightarrow> (m\<Down>NE) s \<subseteq> (m'\<Down>NE) s"
  by(auto simp add: project_defs)

lemma projectR_mono: "(\<forall>x. m x \<subseteq> m' x) \<Longrightarrow> (m\<Down>R) s \<subseteq> (m'\<Down>R) s"
  by(auto simp add: project_defs)

lemmas projectNN_mono_all = projectNN_mono[THEN allI]
lemmas projectNE_mono_all = projectNE_mono[THEN allI]
lemmas projectR_mono_all = projectR_mono[THEN allI]


end
