theory RMonad
imports Main "~~/src/HOL/Library/Monad_Syntax"

begin

text {* Sets as non-deterministic monads *}
type_synonym 's nd_monad = "'s set"

definition returnS :: "'s \<Rightarrow> 's nd_monad" where
  "returnS \<equiv> \<lambda>x. {x}"

definition bindS :: "('a nd_monad) \<Rightarrow> ('a \<Rightarrow> 'b nd_monad) \<Rightarrow> 'b nd_monad" 
    (infixr "\<guillemotright>=S" 55) where
  "bindS s g \<equiv> \<Union> g ` s"

text {* Non-deterministic state monad *}
type_synonym ('a, 's)nds_monad = "'s \<Rightarrow> ('a \<times> 's)set"

definition returnN :: "'a \<Rightarrow> ('a, 's)nds_monad" where
  "returnN \<equiv> curry returnS"

(*
definition returnN :: "'a \<Rightarrow> ('a, 's)nds_monad" where
  "returnN \<equiv> \<lambda>a s. {(a, s)}"
*)

(*
definition bindN (* :: "('a, 's)nds_monad \<Rightarrow> ('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> ('b, 's)nds_monad" *)
    (infixr "\<guillemotright>=N" 55) where
  "bindN f g \<equiv> \<lambda>s. \<Union> split g ` f s"
*)

definition bindN (* :: "('a, 's)nds_monad \<Rightarrow> ('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> ('b, 's)nds_monad" *)
    (infixr "\<guillemotright>=N" 55) where
  "bindN f g \<equiv> \<lambda>s. bindS (f s) (split g) "


text {* "Restart monad" *}
type_synonym ('a, 's)r_monad = "('a option, 's)nds_monad"

definition returnR :: "'a \<Rightarrow> ('a, 's)r_monad" where
  "returnR \<equiv> returnN o Some"

definition restart :: "('a, 's)r_monad" where
  "restart \<equiv> returnN None"

definition liftR :: "('a \<Rightarrow> ('b, 's)r_monad) \<Rightarrow> 'a option \<Rightarrow> ('b, 's)r_monad" where
  "liftR f \<equiv> \<lambda>r. (case r of None \<Rightarrow> restart | Some a \<Rightarrow> f a)"

definition liftR2 :: "('a, 's)r_monad \<Rightarrow> 'a option \<Rightarrow> ('a, 's)r_monad" where
  "liftR2 m \<equiv> \<lambda>a s. case a of None \<Rightarrow> m s | Some x \<Rightarrow> {(a, s)}"

definition bindR :: "('a, 's)r_monad \<Rightarrow> ('a \<Rightarrow> ('b, 's)r_monad) \<Rightarrow> ('b, 's)r_monad" (infix "\<guillemotright>=R" 55) where
  "bindR f g \<equiv> bindN f (liftR g)"

text {* Exception monad on top of the restart monad (not really used currently) *}
type_synonym ('a, 'e, 's)re_monad = "(('e + 'a), 's)r_monad"

definition return :: "'a \<Rightarrow> ('a, 'e, 's)re_monad" where
  "return \<equiv> returnR o Inr"

definition throw :: "'e \<Rightarrow> ('a, 'e, 's)re_monad" where
  "throw \<equiv> returnR o Inl"

text {*
  bind and try_catch are symmetric in this case
*}

definition liftE :: "('a \<Rightarrow> ('b, 'e, 's)re_monad) \<Rightarrow> ('e + 'a) \<Rightarrow> ('b, 'e, 's)re_monad" where
  "liftE \<equiv> \<lambda>f r. case r of Inl e \<Rightarrow> throw e | Inr a \<Rightarrow> f a"

definition liftE2 :: "('e \<Rightarrow> ('a, 'e, 's)re_monad) \<Rightarrow> ('e + 'a) \<Rightarrow> ('a, 'e, 's)re_monad" where
  "liftE2 \<equiv> \<lambda>h r. case r of Inl e \<Rightarrow> h e | Inr a \<Rightarrow> return a"

definition bind :: "('a, 'e, 's)re_monad \<Rightarrow> ('a \<Rightarrow> ('b, 'e, 's)re_monad) \<Rightarrow> 
        ('b, 'e,'s)re_monad" (infix "\<guillemotright>=RE" 60) where
  "bind f g \<equiv> bindR f (liftE g)"

(* Can't overload bindR/bindN the same way, not possible to differentiate the types, so
  we have to come up with separate doN/doR notation later... *)

setup {*
  Adhoc_Overloading.add_variant @{const_name Monad_Syntax.bind} @{const_name bind}
*}


text {* This is basically copied straight from Monad_Syntax in the Isabelle/HOL Library *}
syntax
  "_do_blockR" :: "do_binds \<Rightarrow> 'a" ("doR {//(2  _)//}" [12] 62)

translations
  "_do_blockR (_do_cons (_do_then t) (_do_final e))"
    == "CONST bindR t (\<lambda>_. e)"
  "_do_blockR (_do_cons (_do_bind p t) (_do_final e))"
    == "CONST bindR t (\<lambda>p. e)"
  "_do_blockR (_do_cons (_do_let p t) bs)"
    == "let p = t in _do_blockR bs"
  "_do_blockR (_do_cons b (_do_cons c cs))"
    == "_do_blockR (_do_cons b (_do_final (_do_blockR (_do_cons c cs))))"
  "_do_blockR (_do_final e)" => "e"
  "(m >> n)" => "(m >>= (\<lambda>_. n))"

syntax
  "_do_blockN" :: "do_binds \<Rightarrow> 'a" ("doN {//(2  _)//}" [12] 62)

translations
  "_do_blockN (_do_cons (_do_then t) (_do_final e))"
    == "CONST bindN t (\<lambda>_. e)"
  "_do_blockN (_do_cons (_do_bind p t) (_do_final e))"
    == "CONST bindN t (\<lambda>p. e)"
  "_do_blockN (_do_cons (_do_let p t) bs)"
    == "let p = t in _do_blockN bs"
  "_do_blockN (_do_cons b (_do_cons c cs))"
    == "_do_blockN (_do_cons b (_do_final (_do_blockN (_do_cons c cs))))"
  "_do_blockN (_do_final e)" => "e"
  "(m >> n)" => "(m >>= (\<lambda>_. n))"

abbreviation
  "seqN" :: "('a, 's) nds_monad \<Rightarrow> ('s \<Rightarrow> 'b set) \<Rightarrow> 's \<Rightarrow> 'b set"    (infixl "\<guillemotright>N" 55) where
  "m1 \<guillemotright>N m2 \<equiv> doN { m1; m2 }" 

syntax
  "_do_blockS" :: "do_binds \<Rightarrow> 'a" ("doS {//(2  _)//}" [12] 62)

translations
  "_do_blockS (_do_cons (_do_then t) (_do_final e))"
    == "CONST bindS t (\<lambda>_. e)"
  "_do_blockS (_do_cons (_do_bind p t) (_do_final e))"
    == "CONST bindS t (\<lambda>p. e)"
  "_do_blockS (_do_cons (_do_let p t) bs)"
    == "let p = t in _do_blockS bs"
  "_do_blockS (_do_cons b (_do_cons c cs))"
    == "_do_blockS (_do_cons b (_do_final (_do_blockS (_do_cons c cs))))"
  "_do_blockS (_do_final e)" => "e"
  "(m >> n)" => "(m >>= (\<lambda>_. n))"

(* syntax check *)
lemma "doS {
  x \<leftarrow> returnS (1 :: nat);
  y \<leftarrow> {2, 3, 4};
  returnS (x + y) } = {3, 4, 5}"
  by(auto simp add: bindS_def returnS_def)

lemma "doN {
  x \<leftarrow> returnN 1;
  y \<leftarrow> \<lambda>s. {(2, s), (3, s)};
  returnN (x + y) } = (\<lambda>s. {(1 + 2, s), (1 + 3, s)}) "
  by(rule ext, auto simp add: bindN_def returnN_def bindS_def returnS_def)

lemma "doR {
    restart;
    return 1
  } = (\<lambda>s. {(None, s)})"
  by(rule ext, auto simp add: bindR_def liftR_def restart_def bindN_def returnN_def bindS_def returnS_def)

lemmas bind_defs = bind_def bindR_def bindN_def bindS_def
lemmas return_defs = return_def returnR_def returnN_def returnS_def

definition try_catchR :: "('a, 's)r_monad \<Rightarrow> ('a, 's)r_monad \<Rightarrow> ('a, 's)r_monad" 
    ("(<tryR> _ <catchR> _)" [0, 60] 60) where
  "try_catchR m r_handler \<equiv> \<lambda>s. \<Union> split (liftR2 r_handler) ` m s"

lemma "
  <tryR> doR {
    restart;
    returnR 1
  } <catchR>
    (returnR 2)
  = (\<lambda>s. {(Some 2, s)})"
  by(rule ext, auto simp add: bindR_def liftR_def liftR2_def restart_def try_catchR_def returnR_def
    bindN_def returnN_def bindS_def returnS_def)

definition try_catch :: "('a, 'e, 's)re_monad \<Rightarrow> ('e \<Rightarrow> ('a, 'e, 's)re_monad) \<Rightarrow> ('a, 'e, 's)re_monad" 
    ("(<try> _ <catch> _)" [0, 60] 60) where
  "try_catch m handler \<equiv> m \<guillemotright>=R liftE2 handler"

text {* Sanity test: make sure we don't catch restarts with "regular" handlers *}
lemma "
  <try> do {
    restart;
    return 1
  } <catch> (\<lambda>e.
    (return 2))
  = (\<lambda>s. {(None, s)})"
  by(rule ext) (
    auto simp add: bind_defs restart_def returnR_def returnN_def returnS_def liftE_def liftR_def
    try_catch_def liftE2_def)

subsection {* Monad laws *}

lemma bindS_left_unit:
  "returnS a \<guillemotright>=S m = m a"
  by(simp add: bindS_def returnS_def)

lemma bindS_right_unit:
  "m \<guillemotright>=S returnS = m"
  by(simp add: bindS_def returnS_def)

lemma bindS_assoc:
  "(f \<guillemotright>=S g) \<guillemotright>=S h = f \<guillemotright>=S (\<lambda>s. (g s) \<guillemotright>=S h)"
  by(simp add: bindS_def)

lemma bindN_left_unit:
  "returnN a \<guillemotright>=N m = m a"
  by(simp add: bindN_def returnN_def bindS_left_unit)

lemma bindN_right_unit:
  "m \<guillemotright>=N returnN = m"
  by(simp add: bindN_def returnN_def bindS_right_unit)

lemma bindN_assoc:
  "(f \<guillemotright>=N g) \<guillemotright>=N h = f \<guillemotright>=N (\<lambda>a. g a \<guillemotright>=N h)"
  by(simp add: bindN_def bindS_assoc split_def)

lemma bindR_left_unit:
  "returnR a \<guillemotright>=R m = m a"
  by(auto simp add: returnR_def bindR_def liftR_def bindN_left_unit)

lemma liftR_returnR: "liftR returnR = returnN"
  by(rule ext,auto simp add: returnR_def liftR_def restart_def split: option.split)

lemma bindR_right_unit:
  "m \<guillemotright>=R returnR = m"
  by(auto simp add: bindR_def liftR_returnR bindN_right_unit)

lemma liftR_assoc: "(\<lambda>a. liftR g a \<guillemotright>=N liftR h) = liftR (\<lambda>a. g a \<guillemotright>=N liftR h)"
  apply(rule ext)
  apply(case_tac a)
  apply(auto simp add: liftR_def restart_def bindN_left_unit)
  done

lemma bindR_assoc:
  "(f \<guillemotright>=R g) \<guillemotright>=R h = f \<guillemotright>=R (\<lambda>a. g a \<guillemotright>=R h)"
  by(simp only: bindR_def bindN_assoc liftR_assoc)

lemma bind_left_unit:
  "return a \<guillemotright>=RE m = m a"
  by(auto simp add: return_def bind_def liftE_def bindR_left_unit)

lemma liftE_return: "liftE return = returnR"
  by(rule ext, auto simp add: liftE_def return_def throw_def split: sum.split)

lemma bind_right_unit:
  "m \<guillemotright>=RE return = m"
  by(auto simp add: bind_def liftE_return bindR_right_unit)

lemma liftE_assoc: "(\<lambda>a. liftE g a \<guillemotright>=R liftE h) = liftE (\<lambda>a. g a \<guillemotright>=R liftE h)"
  apply(rule ext)
  apply(case_tac a)
  apply(auto simp add: liftE_def throw_def bindR_left_unit)
  done

lemma bind_assoc:
  "(f \<guillemotright>=RE g) \<guillemotright>=RE h = f \<guillemotright>=RE (\<lambda>a. g a \<guillemotright>=RE h)"
  by(simp only: bind_def bindR_assoc liftE_assoc)

subsection {* Interaction of bind/bindR/throw/restart *}

lemma restart_bindR:
  "restart \<guillemotright>=R m = restart"
  apply(rule ext)
  apply(auto simp add: bindR_def restart_def return_defs liftR_def bindN_def bindS_def)
  done

lemma restart_bind:
  "restart \<guillemotright>= m = restart"
  by(auto simp add: bind_def restart_bindR)

lemma try_restart_catchR: "try_catchR restart h = h"
  by(auto simp add: try_catchR_def restart_def returnN_def returnS_def liftR2_def)

lemma throw_bind:
  "throw e \<guillemotright>= m = throw e"
  apply(rule ext)
  apply(auto simp add: bind_def liftE_def bindR_def throw_def return_defs liftR_def bindN_def bindS_def)
  done

lemma try_throw_catch: "try_catch (throw e) h = (h e)"
  by(simp add: try_catch_def liftE2_def throw_def return_defs
    bindR_def liftR_def bindN_def bindS_def)

subsection {* empty monad *}

lemma empty_bindS: "{} \<guillemotright>=S m = {}"
  by(simp add: bindS_def)

definition K :: "'a \<Rightarrow> 'b \<Rightarrow> 'a" where
  "K \<equiv> \<lambda>x y. x"

definition empty :: "'a \<Rightarrow> 'b nd_monad" where
  "empty \<equiv> K {}"

lemma bindS_empty: "m \<guillemotright>=S empty = {}"
  by(simp add: bindS_def empty_def K_def)

lemma empty_bindN: "empty \<guillemotright>=N m = empty"
  by(auto simp add: empty_def bindN_def bindS_def K_def)

lemma empty_bindR: "empty \<guillemotright>=R m = empty"
  by(auto simp add: bindR_def empty_bindN)

lemma empty_bind: "empty \<guillemotright>= m = empty"
  by(auto simp add: bind_def empty_bindR)

lemma bindN_empty: "m \<guillemotright>=N (\<lambda>x. empty) = empty"
  by(auto simp add: empty_def bindN_def bindS_def K_def)

text {* Note that the analogue for bindR/bind doesn't hold, since we still
  might get results out if m restarts/throws. We'll give the corresponding
  lemmas later, after we introduce the projections *}

lemma try_empty_catchR:
  "try_catchR empty h = empty"
  apply(rule ext)
  apply(force simp add: empty_def K_def try_catchR_def)
  done

lemma try_empty_catch:
  "try_catch empty h = empty"
  by(simp add: empty_bindR try_catch_def)

subsection {* choice *}

definition choice 
  (* :: "('a, 's)nds_monad \<Rightarrow> ('a, 's)nds_monad \<Rightarrow> ('a, 's)nds_monad" *) (infixr "[+]" 60) where
  "choice p q \<equiv> \<lambda>s. p s \<union> q s"

definition choiceA :: "('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> ('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> 'a \<Rightarrow> ('b, 's)nds_monad" where
  "choiceA p q \<equiv> \<lambda>x s. p x s \<union> q x s"

lemma choice_left_I: "(b = c) \<Longrightarrow> (a [+] b = a [+] c)"
  by(auto)

lemma choice_absorb: "m [+] m = m"
  by(simp add: choice_def)

lemma choice_comm:
  "m1 [+] m2 = m2 [+] m1"
  by(auto simp add: choice_def)

lemma choice_assoc:
  "(m1 [+] m2) [+] m3 = m1 [+] (m2 [+] m3)"
  by(auto simp add: choice_def)

lemma choice_emptyl:
  "empty [+] m = m"
  by(auto simp add: choice_def empty_def K_def)

lemma choice_emptyr:
  "m [+] empty = m"
  by(auto simp add: choice_def empty_def K_def)

lemma choice_bindS: "(m1 [+] m2) s \<guillemotright>=S m3 = (m1 s \<guillemotright>=S m3) \<union> (m2 s \<guillemotright>=S m3)"
  by(auto simp add: choice_def bindS_def)

lemma choice_bindN:
  "(m1 [+] m2 \<guillemotright>=N m3) = (m1 \<guillemotright>=N m3) [+] (m2 \<guillemotright>=N m3)"
  by(simp only: bindN_def choice_bindS, simp only: choice_def)

lemma choice_bindR:
  "(m1 [+] m2 \<guillemotright>=R m3) = (m1 \<guillemotright>=R m3) [+] (m2 \<guillemotright>=R m3)"
  by(simp add: bindR_def choice_bindN)

lemma choice_bind:
  "(m1 [+] m2 \<guillemotright>= m3) = (m1 \<guillemotright>= m3) [+] (m2 \<guillemotright>= m3)"
  by(simp add: bind_def choice_bindR)

lemma bindS_choice: "s \<guillemotright>=S m1 [+] m2 = (s \<guillemotright>=S m1) \<union> (s \<guillemotright>=S m2)"
  by(auto simp add: choice_def bindS_def)

lemma bindN_choice:
  "(m1 \<guillemotright>=N (\<lambda>x. m2 x [+] m3 x)) = (m1 \<guillemotright>=N m2) [+] (m1 \<guillemotright>=N m3)"
  by(auto simp add: choice_def bindN_def bindS_def)

lemma liftR_choice:
  "liftR (\<lambda>x. m1 x [+] m2 x) = (\<lambda>x. (liftR m1 x) [+] (liftR m2 x))"
  by(auto simp add: liftR_def choice_def split: option.split)

lemma bindR_choice:
  "(m1 \<guillemotright>=R (\<lambda>x. m2 x [+] m3 x)) = (m1 \<guillemotright>=R m2) [+] (m1 \<guillemotright>=R m3)"
  by(auto simp add: bindR_def bindN_choice liftR_choice)

lemma liftE_choice:
  "liftE (\<lambda>x. m1 x [+] m2 x) = (\<lambda>x. (liftE m1 x) [+] (liftE m2 x))"
  by(auto simp add: liftE_def choice_def split: sum.split)

lemma bind_choice:
  "(m1 \<guillemotright>= (\<lambda>x. m2 x [+] m3 x)) = (m1 \<guillemotright>= m2) [+] (m1 \<guillemotright>= m3)"
  by(simp add: bind_def bindR_choice liftE_choice)

lemma try_choice_catchR:  "try_catchR (m1 [+] m2) h = try_catchR m1 h [+] try_catchR m2 h"
  by(simp add: choice_def try_catchR_def)

lemma try_catchR_choice: "try_catchR m (h1 [+] h2) = try_catchR m h1 [+] try_catchR m h2"
  apply(rule ext)
  apply(force simp add: choice_def try_catchR_def liftR2_def 
    split: option.split option.split_asm sum.split sum.split_asm)
  done

lemma try_choice_catch: "try_catch (m1 [+] m2) h = try_catch m1 h [+] try_catch m2 h"
  by(simp add: try_catch_def choice_bindR)

lemma liftE2_choice:
  "liftE2 (\<lambda>x. m1 x [+] m2 x) = (\<lambda>x. (liftE2 m1 x) [+] (liftE2 m2 x))"
  by(auto simp add: liftE2_def choice_def split: sum.split)

lemma try_catch_choice: "try_catch m (\<lambda>e. h1 e [+] h2 e) = try_catch m h1 [+] try_catch m h2"
  by(simp add: try_catch_def bindR_choice liftE2_choice)

fun guard :: "bool \<Rightarrow> (unit, 'e, 's)re_monad" where
  "guard True = return ()"
  | "guard False = empty"

lemma if_then_guard: "(if b then m1 else m2) = ((guard b \<guillemotright> m1) [+] (guard (\<not>b) \<guillemotright> m2))"
  apply(case_tac b)
  apply(rule ext)
  apply(auto simp add: empty_def choice_def return_defs bind_defs
    liftE_def liftR_def K_def)
  done

definition nofail :: "('a, 's)nds_monad \<Rightarrow> bool" where
  "nofail m \<equiv> \<forall>s. m s \<noteq> {}"

lemma[iff]: "x \<in> empty s = False"
  by(auto simp add: empty_def K_def)

definition choose :: "'a set \<Rightarrow> ('a, 's)nds_monad" where
  "choose A \<equiv> \<lambda>s. A \<times> {s}"

definition chooseN :: "'a set \<Rightarrow> ('a, 'e, 's)re_monad" where
  "chooseN A \<equiv> choose ((Some o Inr) ` A)"

definition gets :: "('s \<Rightarrow> 'a) \<Rightarrow> ('a, 's)nds_monad" where
  "gets f \<equiv> \<lambda>s. (returnN (f s) s)"

definition put' :: "'a \<Rightarrow> 's \<Rightarrow> ('a, 's)nds_monad" where
  "put' a s' \<equiv> K (returnN a s')"

definition put :: "'s \<Rightarrow> (unit, 's)nds_monad" where
  "put \<equiv> put' ()"

definition modify :: "('s \<Rightarrow> 's) \<Rightarrow> (unit, 's)nds_monad" where
  "modify f \<equiv> doN { s \<leftarrow> gets f; put s }"

definition getsR :: "('s \<Rightarrow> 'a) \<Rightarrow> ('a, 's)r_monad" where
  "getsR f \<equiv> gets (Some o f)"

definition putR' :: "'a \<Rightarrow> 's \<Rightarrow> ('a, 's)r_monad" where
  "putR' a \<equiv> put' (Some a)"

definition putR :: "'s \<Rightarrow> (unit, 's)r_monad" where
  "putR \<equiv> putR' ()"

definition modifyR :: "('s \<Rightarrow> 's) \<Rightarrow> (unit, 's)r_monad" where
  "modifyR f \<equiv> doR { s \<leftarrow> getsR f; putR s }"

definition getsRE :: "('s \<Rightarrow> 'a) \<Rightarrow> ('a, 'e, 's)re_monad" where
  "getsRE f \<equiv> getsR (Inr o f)"

definition putRE' :: "'a \<Rightarrow> 's \<Rightarrow> ('a, 'e, 's)re_monad" where
  "putRE' a \<equiv> putR' (Inr a)"

definition putRE :: "'s \<Rightarrow> (unit, 'e, 's)re_monad" where
  "putRE \<equiv> putRE' ()"

definition modifyRE :: "('s \<Rightarrow> 's) \<Rightarrow> (unit, 'e, 's)re_monad" where
  "modifyRE f \<equiv> do { s \<leftarrow> getsRE f; putRE s }"

text {* Function @{term mapM} is a monadic version of @{term map}
function. In contrast to the pure version it is important to note that
list elements are evaluated from left to right. *}

fun 
  mapM_rec_N :: "('a \<Rightarrow> ('b, 's) nds_monad) => 'b list => 'a list => ('b list, 's) nds_monad" where
  "mapM_rec_N f l [] = returnN l"
  | "mapM_rec_N f l (a#as) = (f a) \<guillemotright>=N (\<lambda>b. mapM_rec_N f (l@[b]) as)"

definition
  mapM_N :: "('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> 'a list \<Rightarrow> ('b list, 's)nds_monad" where
  "mapM_N f l \<equiv> mapM_rec_N f [] l"

fun mapM'_N :: "('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> 'a list \<Rightarrow> (unit, 's)nds_monad" where
  mapM'_N_Nil: "mapM'_N f [] = returnN ()"
  | mapM'_N_Cons: "mapM'_N f (a#as) = f a \<guillemotright>=N (\<lambda>_. (mapM'_N f as))"

fun mapM' :: "('a \<Rightarrow> ('b, 'e, 's)re_monad) \<Rightarrow> 'a list \<Rightarrow> (unit, 'e, 's)re_monad" where
  mapM'_Nil: "mapM' f [] = return ()"
  | mapM'_Cons: "mapM' f (a#as) = f a \<guillemotright>= (\<lambda>_. (mapM' f as))"

lemma bindN_cong[fundef_cong]:
  "\<lbrakk> m1 = m1'; \<And>s x s'. (x, s') \<in> m1 s \<Longrightarrow> m2 x s' = m2' x s' \<rbrakk>
    \<Longrightarrow> bindN m1 m2 = bindN m1' m2'"
  by(force simp add: bindN_def bindS_def intro!: ext)

lemma bindR_cong[fundef_cong]:
  "\<lbrakk> m1 = m1'; \<And>s s' x. (Some x, s') \<in> m1 s \<Longrightarrow> m2 x s' = m2' x s' \<rbrakk> 
    \<Longrightarrow> bindR m1 m2 = bindR m1' m2'"
  apply(simp add: bindR_def)
  apply(rule bindN_cong)
  apply(auto simp add: liftR_def intro!: ext split: option.split_asm)
  done

lemma bind_cong[fundef_cong]:
  "\<lbrakk> m1 = m1'; \<And> s s' x. (Some (Inr x), s') \<in> m1 s \<Longrightarrow> m2 x s' = m2' x s' \<rbrakk> 
    \<Longrightarrow> bind m1 m2 = bind m1' m2'"
  apply(simp add: bind_def)
  apply(rule bindR_cong)
  apply(auto simp add: liftE_def intro!: ext split: sum.split_asm)
  done

(* TODO: find a more suitable place for this definition *)
definition order_set :: "'a set \<Rightarrow> 'a list set" where
  "order_set A \<equiv> {l. fold_graph Cons [] A l}"

lemmas monad_laws =
  bindS_left_unit
  bindN_left_unit
  bindR_left_unit
  bind_left_unit
  bindS_right_unit
  bindN_right_unit
  bindR_right_unit
  bind_right_unit
  bindS_assoc
  bindN_assoc
  bindR_assoc
  bind_assoc

lemmas unfold_monad =
  bind_def
  return_def
  throw_def
  liftE_def
  bindR_def
  returnR_def
  liftR_def
  restart_def
  bindN_def
  returnN_def
  bindS_def
  returnS_def
  try_catchR_def
  liftR2_def
  try_catch_def
  liftE2_def
  choice_def
  empty_def
  K_def

lemmas monad_simps =
  restart_bindR
  restart_bind
  throw_bind
  try_restart_catchR
  try_throw_catch
  choice_absorb

lemmas empty_simps =
  empty_bindN
  empty_bindR
  empty_bind
  bindN_empty
  choice_emptyl
  choice_emptyr
  try_empty_catchR
  try_empty_catch

lemmas choice_distrib = 
  choice_bindS
  choice_bindN
  choice_bindR
  choice_bind
  bindS_choice
  bindN_choice
  bindR_choice
  bind_choice
  try_choice_catchR
  try_catchR_choice
  try_choice_catch
  try_catch_choice

end
