theory RMonad_Lemmas
  imports RMonad RMonad_Decomposition

begin

lemma bindS_mono1:
  "S' \<subseteq> S \<Longrightarrow> S' \<guillemotright>=S m \<subseteq> S \<guillemotright>=S m"
  by(auto simp add: unfold_monad)

lemma bindS_mono2: "\<forall>s. m' s \<subseteq> m s \<Longrightarrow> S \<guillemotright>=S m' \<subseteq> S \<guillemotright>=S m"
  by(auto simp add: unfold_monad)

lemmas bindS_mono2_all = bindS_mono2[OF allI]

lemma bindS_mono2': "\<forall>s \<in> S. m' s \<subseteq> m s \<Longrightarrow> S \<guillemotright>=S m' \<subseteq> S \<guillemotright>=S m"
  by(auto simp add: unfold_monad)

lemma bindN_mono1:
  "ml' s \<subseteq> ml s \<Longrightarrow> (ml' \<guillemotright>=N m) s \<subseteq> (ml \<guillemotright>=N m) s"
  by(auto simp add: unfold_monad)

lemma bindS_choice_abs:
  "(\<lambda>s. f s \<guillemotright>=S (a [+] b)) = (\<lambda>s. f s \<guillemotright>=S a) [+] (\<lambda>s. f s \<guillemotright>=S b)"
  by(auto simp add: unfold_monad)

lemma bindN_mono2:
  "\<forall>a s. mr' a s \<subseteq> mr a s \<Longrightarrow> (m \<guillemotright>=N mr') s \<subseteq> (m \<guillemotright>=N mr) s"
  by(auto simp add: unfold_monad)

lemmas bindN_mono2_all = bindN_mono2[OF allI]

lemma bindN'_mono2:
  "\<forall>s. mr' s \<subseteq> mr s \<Longrightarrow> (m \<guillemotright>N mr') s \<subseteq> (m \<guillemotright>N mr) s"
  by(auto simp add: unfold_monad)

lemmas bindN'_mono2_all = bindN'_mono2[OF allI]

lemma bindN_bindS: "(m1 \<guillemotright>N m2) s = (snd ` m1 s \<guillemotright>=S m2)"
  by(auto simp add: unfold_monad)

lemma bindS_bindN: "(m1 s \<guillemotright>=S (m2 \<guillemotright>N m3)) = ((\<lambda>s. (m1 s \<guillemotright>=S m2)) \<guillemotright>N m3) s"
  by(auto simp add: unfold_monad)

lemma bind_mono1:
  "m s \<subseteq> m' s \<Longrightarrow>
    (m \<guillemotright>= f) s \<subseteq> (m' \<guillemotright>= f) s"
  by(auto simp add: unfold_monad)

lemma try_catchR_mono2:
  "\<forall>s. h s \<subseteq> h' s \<Longrightarrow> (<tryR> m <catchR> h) s \<subseteq> (<tryR> m <catchR> h') s"
  apply(auto)
  apply(case_tac a)
  apply(auto simp add: try_catchR_none_decompose try_catchR_some_decompose)
  done

lemma choice_subsetI1:
  "x \<in> m1 s \<Longrightarrow> x \<in> (m1 [+] m2) s"
  by(simp add: choice_def)

lemma choice_subsetI2:
  "x \<in> m2 s \<Longrightarrow> x \<in> (m1 [+] m2) s"
  by(simp add: choice_def)

lemma choose_comm:
  "(m1 \<guillemotright>N (choose A \<guillemotright>=N (\<lambda>x. m2 x))) = (choose A \<guillemotright>=N (\<lambda>x. m1 \<guillemotright>N m2 x))"
  by(auto simp add: unfold_monad choose_def)

lemma choice_left_commute:
  "(x [+] y [+] z) = (y [+] x [+] z)"
  by (metis choice_assoc choice_comm)

lemma choice_left_absorb:
  "(m [+] m [+] m') = m [+] m'"
  "(m [+] m' [+] m) = m [+] m'"
  apply (metis choice_absorb choice_assoc)
  by (metis choice_absorb choice_assoc choice_comm)

lemma if_choice: 
  "(if c then (m1 [+] m2) else m3) = (if c then m1 else m3) [+] (if c then m2 else m3)"
  "(if c then m1 else (m2 [+] m3)) = (if c then m1 else m2) [+] (if c then m1 else m3)"
  by(simp_all add: choice_absorb)

lemma empty_bindS: "(empty x \<guillemotright>=S m) = {}"
  by(auto simp add: unfold_monad)

lemma bindS_empty: "(S \<guillemotright>=S empty) = {}"
  by(auto simp add: unfold_monad)

lemma union_bindS: "((S1 \<union> S2) \<guillemotright>=S m) = ((S1 \<guillemotright>=S m) \<union> (S2 \<guillemotright>=S m))"
  by(auto simp add: unfold_monad)

lemma K_beta:
  "K x y = x"
  by(simp add: K_def)

lemma order_set_set:
  "l \<in> order_set A \<Longrightarrow> finite A \<Longrightarrow> (set l = A)"
proof(rule equalityI, unfold order_set_def, simp_all)
  assume asm: "finite A" and l_def: "fold_graph Cons [] A l"
  show "A \<subseteq> set l" using l_def asm
    by(induct set: fold_graph, auto)
next
  assume asm: "finite A" and l_def: "fold_graph Cons [] A l"
  show "set l \<subseteq> A" using l_def
    by(induct set: fold_graph, auto)
qed

lemma distinctI: "\<lbrakk> distinct xs; x \<notin> set xs \<rbrakk> \<Longrightarrow> distinct (x#xs)"
  by(induct xs, auto)

lemma order_set_distinct:
  "l \<in> order_set A \<Longrightarrow> finite A \<Longrightarrow> distinct l"
proof(simp only: order_set_def mem_Collect_eq, induct rule: fold_graph.induct)
  show "distinct []" by simp
next
  fix A x xs
  assume assms: "x \<notin> A" "fold_graph op # [] A xs" "finite A \<Longrightarrow> distinct xs" "finite (insert x A)"
  have "xs \<in> order_set A" using `fold_graph op # [] A xs`
    by(simp add: order_set_def)
  hence "set xs = A" using order_set_set `finite (insert x A)`
    by auto
  thus "distinct (x # xs)" using assms
    by(auto intro: distinctI)
qed

end
