theory Refinement
imports RMonad RMonad_Decomposition Projections EvaluationLogicR HoareLogic
  Repeated_Restarts Restart_Transformations Misc_Lemmas
begin

(*******************************************************************************)
section {* Refinement of monadic programs *}
(*******************************************************************************)

(*******************************************************************************)
subsection {* Non-deterministic monad *}
(*******************************************************************************)

definition simS_wpc ::
  "('a \<Rightarrow> 'a1 set) \<Rightarrow> ('b \<Rightarrow> 'b1 set) \<Rightarrow> ('a1 \<times> 'b1)set \<Rightarrow> ('a \<times> 'b)set" 
  where
  "simS_wpc ma mc S \<equiv> {(sa, sc). \<forall>sc'.
    sc' \<in> mc sc \<longrightarrow>
    (\<exists>sa' \<in> ma sa. (sa', sc') \<in> S)
  }"
 
lemma simS_wpc_monotone_comp_conc:
  "\<forall>s. mc s \<subseteq> mc' s \<Longrightarrow> simS_wpc ma mc' S \<subseteq> simS_wpc ma mc S"
  by(force simp add: simS_wpc_def) 

lemma simS_wpc_monotone_comp_abs:
  "\<forall>s. ma s \<subseteq> ma' s \<Longrightarrow> simS_wpc ma mc S \<subseteq> simS_wpc ma' mc S"
  by(force simp add: simS_wpc_def) 

lemma simS_wpc_choice_conc:
  "simS_wpc ma (mc1 [+] mc2) S = simS_wpc ma mc1 S \<inter> simS_wpc ma mc2 S"
  by(auto simp add: simS_wpc_def choice_def)

lemma simS_wpc_choice_abs:
  "simS_wpc ma1 mc S \<union> simS_wpc ma2 mc S \<subseteq> simS_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: UN_least simS_wpc_monotone_comp_abs simp add: choice_def)

lemma simS_wpc_choice_abs_I1:
  "x \<in> simS_wpc ma1 mc S \<Longrightarrow> x \<in> simS_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: simS_wpc_choice_abs[THEN subsetD])

lemma simS_wpc_choice_abs_I2:
  "x \<in> simS_wpc ma2 mc S \<Longrightarrow> x \<in> simS_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: simS_wpc_choice_abs[THEN subsetD])

lemma simS_wpc_union_lockstep:
  "\<lbrakk> R \<subseteq> simS_wpc ma1 mc1 S;
    R \<subseteq> simS_wpc ma2 mc2 S 
  \<rbrakk> \<Longrightarrow> 
  R \<subseteq> simS_wpc (\<lambda>s. ma1 s \<union> ma2 s) (\<lambda>s. mc1 s \<union> mc2 s) S
  "
  apply(clarsimp simp add: simS_wpc_def)
  by (smt Un_def mem_Collect_eq set_mp splitD)

lemma simS_wpc_choice_lockstep:
  "\<lbrakk> R \<subseteq> simS_wpc ma1 mc1 S;
    R \<subseteq> simS_wpc ma2 mc2 S 
  \<rbrakk> \<Longrightarrow> 
  R \<subseteq> simS_wpc (ma1 [+] ma2) (mc1 [+] mc2) S
  "
  by(clarsimp simp add: choice_def intro!: simS_wpc_union_lockstep del: subsetI)

lemma simS_wpc_empty_conc: "simS_wpc ma empty S = UNIV"
  by(auto simp add: simS_wpc_def empty_def K_def)

(*******************************************************************************)
subsection {* Non-deterministic state monad *}
(*******************************************************************************)

definition simN_wpc :: 
  "('a, 'sa)nds_monad \<Rightarrow> ('b, 'sc)nds_monad \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> ('sa \<times> 'sc)set" 
  where
  "simN_wpc ma mc S \<equiv> {(sa, sc). \<forall>vc sc'. 
    (vc, sc') \<in> mc sc 
    \<longrightarrow> 
    (\<exists>va sa'. (va, sa') \<in> ma sa \<and> (sa', sc') \<in> S va vc)
  }"

lemma simN_wpc_returnN: "simN_wpc (returnN a) (returnN b) S = S a b"
  by(auto simp add: simN_wpc_def returnN_def returnS_def)

(*******************************************************************************)
subsubsection {* Basic Properties *}
(*******************************************************************************)

type_synonym
  's assn = "'s set"

abbreviation (input)
  "true" :: "'a set" where
  "true \<equiv> UNIV"

abbreviation (input)
  "false" :: "'a set" where
  "false \<equiv> {}"

lemma simN_wpc_false [simp]:
  "nofail mc \<Longrightarrow> simN_wpc ma mc (\<lambda>a b. false) = false"
  by(force simp add: simN_wpc_def nofail_def)  

lemma simN_wpc_true [simp]:
  "nofail ma \<Longrightarrow> simN_wpc ma mc (\<lambda>a b. true)
   = true"
  by(force simp add: simN_wpc_def nofail_def)

lemma simN_wpc_Int:
  "simN_wpc ma mc (\<lambda>a b. Sa a b \<inter> Sb a b) \<subseteq>
   simN_wpc ma mc Sa \<inter> simN_wpc ma mc Sb"
  by(force simp add: simN_wpc_def)

lemma simN_wpc_Un:
  "simN_wpc ma mc Sa \<union> simN_wpc ma mc Sb \<subseteq>
  simN_wpc ma mc (\<lambda>a b. Sa a b \<union> Sb a b)"
  by(force simp add: simN_wpc_def)

lemma simN_wpc_UN:
  "(\<Union> i. simN_wpc ma mc (Se i))
    \<subseteq> simN_wpc ma mc (\<lambda>a b. \<Union>i. Se i a b)"
  by(force simp add: simN_wpc_def)

lemma simN_wpc_monotone_rel:
  "\<forall>a b. S' a b \<subseteq> S a b
  \<Longrightarrow> simN_wpc ma mc S' \<subseteq> simN_wpc ma mc S"
  by (force simp add: simN_wpc_def) 

lemma simN_wpc_monotone_comp_conc:
  "\<forall>s. mc s \<subseteq> mc' s \<Longrightarrow> simN_wpc ma mc' S \<subseteq> simN_wpc ma mc S"
  by(force simp add: simN_wpc_def) 

lemma simN_wpc_monotone_comp_abs:
  "\<forall>s. ma s \<subseteq> ma' s \<Longrightarrow> simN_wpc ma mc S \<subseteq> simN_wpc ma' mc S"
  by(force simp add: simN_wpc_def) 

lemma simS_wpc_bindS:
  "\<lbrakk> (sa, sc) \<in> simS_wpc ma1 mc1 S'; S' \<subseteq> simS_wpc ma2 mc2 S \<rbrakk> \<Longrightarrow>
    (sa, sc) \<in> simS_wpc (\<lambda>s. ma1 s \<guillemotright>=S ma2) (\<lambda>s. mc1 s \<guillemotright>=S mc2) S"
  by(force simp add: simS_wpc_def unfold_monad)

lemma simN_wpc_bindS:
  "\<lbrakk> (sa, sc) \<in> simS_wpc ma1 mc1 S'; S' \<subseteq> simN_wpc ma2 mc2 S \<rbrakk> \<Longrightarrow>
    (sa, sc) \<in> simN_wpc (\<lambda>s. ma1 s \<guillemotright>=S ma2) (\<lambda>s. mc1 s \<guillemotright>=S mc2) S"
  by(force simp add: simS_wpc_def simN_wpc_def unfold_monad)

lemma simN_wpc_bindN:
  "\<lbrakk> x \<in> simN_wpc ma1 mc1 S'; \<forall>va vc. S' va vc \<subseteq> simN_wpc (ma2 va) (mc2 vc) S \<rbrakk> \<Longrightarrow>
    x \<in> simN_wpc (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2) S
  "
  (* Takes a while, possibly fragile? *)
  apply(force simp add: simN_wpc_def unfold_monad)
  done

lemma simN_wpc_bindN_set:
  "\<lbrakk> R \<subseteq> simN_wpc ma1 mc1 S'; \<forall>va vc. S' va vc \<subseteq> simN_wpc (ma2 va) (mc2 vc) S \<rbrakk> \<Longrightarrow>
    R \<subseteq> simN_wpc (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2) S"
  by(auto intro: simN_wpc_bindN)

lemma simS_wpc_bindN:
  "\<lbrakk> x \<in> simN_wpc ma1 mc1 S'; \<forall>va vc. S' va vc \<subseteq> simS_wpc (ma2 va) (mc2 vc) S \<rbrakk> \<Longrightarrow>
    x \<in> simS_wpc (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2) S
  "
  by(force simp add: simN_wpc_def simS_wpc_def unfold_monad)

lemma simS_wpc_bindN_set:
  "\<lbrakk> R \<subseteq> simN_wpc ma1 mc1 S'; \<forall>va vc. S' va vc \<subseteq> simS_wpc (ma2 va) (mc2 vc) S \<rbrakk> \<Longrightarrow>
    R \<subseteq> simS_wpc (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2) S
  "
  by(auto intro: simS_wpc_bindN)

lemma simN_wpc_choice_conc:
  "simN_wpc ma (mc1 [+] mc2) S = simN_wpc ma mc1 S \<inter> simN_wpc ma mc2 S"
  by(auto simp add: simN_wpc_def choice_def)

lemma simN_wpc_choice_abs:
  "simN_wpc ma1 mc S \<union> simN_wpc ma2 mc S \<subseteq> simN_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: UN_least simN_wpc_monotone_comp_abs simp add: choice_def)

lemma simN_wpc_choice_abs_I1:
  "x \<in> simN_wpc ma1 mc S \<Longrightarrow> x \<in> simN_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: simN_wpc_choice_abs[THEN subsetD])

lemma simN_wpc_choice_abs_I2:
  "x \<in> simN_wpc ma2 mc S \<Longrightarrow> x \<in> simN_wpc (ma1 [+] ma2) mc S"
  by(auto intro!: simN_wpc_choice_abs[THEN subsetD])

lemma simN_wpc_union_lockstep:
  "\<lbrakk> R \<subseteq> simN_wpc ma1 mc1 S;
    R \<subseteq> simN_wpc ma2 mc2 S 
  \<rbrakk> \<Longrightarrow> 
  R \<subseteq> simN_wpc (\<lambda>s. ma1 s \<union> ma2 s) (\<lambda>s. mc1 s \<union> mc2 s) S
  "
  apply(clarsimp simp add: simN_wpc_def)
  apply (smt CollectD in_mono splitD)
  done

lemma simN_wpc_choice_lockstep:
  "\<lbrakk> R \<subseteq> simN_wpc ma1 mc1 S;
    R \<subseteq> simN_wpc ma2 mc2 S 
  \<rbrakk> \<Longrightarrow> 
  R \<subseteq> simN_wpc (ma1 [+] ma2) (mc1 [+] mc2) S
  "
  apply(clarsimp simp add: simN_wpc_def choice_def)
  apply (smt CollectD in_mono splitD)
  done


lemma simN_wpc_empty_conc: "simN_wpc ma empty S = UNIV"
  by(auto simp add: simN_wpc_def empty_def K_def)

lemma simN_wpc_if:
  "(R \<subseteq> simN_wpc (if x then ma1 else ma2) (if x then mc1 else mc2) S)
   = 
    ((x \<longrightarrow> R \<subseteq> simN_wpc ma1 mc1 S) \<and>
    (\<not>x \<longrightarrow> R \<subseteq> simN_wpc ma2 mc2 S))
  "
  by(simp)

lemma simN_bindS_if:
  "(R \<subseteq> simN_wpc (\<lambda>s. (if x then ma1 else ma2) s \<guillemotright>=S ha)
    (\<lambda>s. (if x then mc1 else mc2) s \<guillemotright>=S hc) S) = 
    ((x \<longrightarrow> R \<subseteq> simN_wpc (\<lambda>s. ma1 s \<guillemotright>=S ha) (\<lambda>s. mc1 s \<guillemotright>=S hc) S) \<and>
    (\<not>x \<longrightarrow> R \<subseteq> simN_wpc (\<lambda>s. ma2 s \<guillemotright>=S ha) (\<lambda>s. mc2 s \<guillemotright>=S hc) S))
  "
  by(simp)

(*******************************************************************************)
subsection {* Restart monad *}
(*******************************************************************************)

definition simR_wpcN ::
  "('a, 'sa)r_monad \<Rightarrow> ('b, 'sc)r_monad \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> ('sa \<times> 'sc)set"
  where
  "simR_wpcN ma mc S \<equiv> {(sa, sc). \<forall>vc sc'.
    (Some vc, sc') \<in> mc sc
    \<longrightarrow>
    (\<exists>va sa'. (Some va, sa') \<in> ma sa \<and> (sa', sc') \<in> S va vc)
  }"

lemma "simR_wpcN ma mc S = simN_wpc (ma\<Down>N) (mc\<Down>N) S"
  by(auto simp add: simR_wpcN_def simN_wpc_def project_defs)

definition simR_wpcR :: 
  "('a, 'sa)r_monad \<Rightarrow> ('b, 'sc)r_monad \<Rightarrow> ('sa \<times> 'sc)set \<Rightarrow> ('sa \<times> 'sc)set "
  where
  "simR_wpcR ma mc S \<equiv> {(sa, sc). \<forall>sc'.
    (None, sc') \<in> mc sc
    \<longrightarrow>
    (\<exists>sa'. (None, sa') \<in> ma sa \<and> (sa', sc') \<in> S)
  }"

   
lemma simR_wpcR_projR: "simR_wpcR ma mc S = simS_wpc (ma\<Down>R) (mc\<Down>R) S"
  by(auto simp add: simR_wpcR_def simS_wpc_def project_defs)

(*******************************************************************************)
subsection {* Forward simulation *}
(*******************************************************************************)

definition simS :: "('sa \<times> 'sc)set \<Rightarrow> ('sa \<Rightarrow> 'sa nd_monad) \<Rightarrow> ('sc \<Rightarrow> 'sc nd_monad) \<Rightarrow> bool" where
  "simS R ma mc \<equiv> R \<subseteq> simS_wpc ma mc R"

definition force_same_and :: "'b set \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'b set" where
  "force_same_and R \<equiv> (\<lambda>va vc. (if va = vc then UNIV else {}) \<inter> R)"

lemma force_same_and_Int:
  "force_same_and (R \<inter> R') = (\<lambda>x y. force_same_and R x y \<inter> R')"
  by(auto simp add: force_same_and_def)

lemma force_same_and_same:
  "force_same_and R x x = R"
  by(auto simp add: force_same_and_def)

definition simN :: "('sa \<times> 'sc)set \<Rightarrow> ('a, 'sa)nds_monad \<Rightarrow> ('a, 'sc)nds_monad \<Rightarrow> bool" where
  "simN R ma mc \<equiv> R \<subseteq> simN_wpc ma mc (force_same_and R)"


(*******************************************************************************)
subsection {* Termination mode specific refinement *}
(*******************************************************************************)

text {*
  Simulation relation which only considers non-restarting and
  non-exceptional termination. 
*}

definition sim_normN_ext :: "('sa \<times> 'sc)set \<Rightarrow> ('aa, 'ea, 'sa)re_monad \<Rightarrow> ('ac, 'ec, 'sc)re_monad \<Rightarrow>
    ('aa \<Rightarrow> 'ac \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> bool" where
  "sim_normN_ext R ma mc Sn \<equiv>
    R \<subseteq> simN_wpc (ma\<Down>NN) (mc\<Down>NN) Sn"

definition sim_normN :: "('sa \<times> 'sc)set \<Rightarrow> ('a, 'ea, 'sa)re_monad \<Rightarrow> ('a, 'ec, 'sc)re_monad \<Rightarrow>
    bool" where
  "sim_normN R ma mc \<equiv>
    sim_normN_ext R ma mc (force_same_and R)"

text {*
  Simulation relation which only considers exceptional (but non-restarting) 
  termination. 
*}

definition sim_normE_ext :: "('sa \<times> 'sc)set \<Rightarrow> ('aa, 'ea, 'sa)re_monad \<Rightarrow> ('ac, 'ec, 'sc)re_monad \<Rightarrow>
    ('ea \<Rightarrow> 'ec \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> bool" where
  "sim_normE_ext R ma mc Sn \<equiv>
    R \<subseteq> simN_wpc (ma\<Down>NE) (mc\<Down>NE) Sn"


definition sim_normE :: "('sa \<times> 'sc)set \<Rightarrow> ('aa, 'e, 'sa)re_monad \<Rightarrow> ('ac, 'e, 'sc)re_monad \<Rightarrow>
    bool" where
  "sim_normE R ma mc \<equiv>
    sim_normE_ext R ma mc (force_same_and R)"

definition sim_norm_ext :: "('sa \<times> 'sc)set \<Rightarrow> ('aa, 'ea, 'sa)re_monad \<Rightarrow> ('ac, 'ec, 'sc)re_monad \<Rightarrow>
    ('aa \<Rightarrow> 'ac \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> ('ea \<Rightarrow> 'ec \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> bool" where
  "sim_norm_ext R ma mc Sn Se \<equiv>
    R \<subseteq> simN_wpc (ma\<Down>NN) (mc\<Down>NN) Sn \<inter> simN_wpc (ma\<Down>NE) (mc\<Down>NE) Se
  "

text {*

  Simulation relation which only considers normal termination. This
  is the relation we will care about when abstracting a computation
  of the form @{term "try_catchR m (rec_tryR h)"}

*}

definition sim_norm :: "('sa \<times> 'sc)set \<Rightarrow> ('a, 'e, 'sa)re_monad \<Rightarrow> ('a, 'e, 'sc)re_monad \<Rightarrow>
    bool" where
  "sim_norm R ma mc \<equiv>
    sim_norm_ext R ma mc (force_same_and R) (force_same_and R)"

definition sim_ext :: "('sa \<times> 'sc)set \<Rightarrow> ('aa, 'ea, 'sa)re_monad \<Rightarrow> ('ac, 'ec, 'sc)re_monad \<Rightarrow>
    ('aa \<Rightarrow> 'ac \<Rightarrow> ('sa \<times> 'sc)set) \<Rightarrow> ('ea \<Rightarrow> 'ec \<Rightarrow> ('sa \<times> 'sc)set)
    \<Rightarrow> ('sa \<times> 'sc)set \<Rightarrow> bool" where
  "sim_ext R ma mc Sn Se Sr \<equiv>
    R \<subseteq> simN_wpc (ma\<Down>NN) (mc\<Down>NN) Sn \<inter> simN_wpc (ma\<Down>NE) (mc\<Down>NE) Se \<inter> simS_wpc (ma\<Down>R) (mc\<Down>R) Sr"

definition sim :: "('sa \<times> 'sc)set \<Rightarrow> ('a, 'e, 'sa)re_monad \<Rightarrow> ('a, 'e, 'sc)re_monad \<Rightarrow>
    bool" where
  "sim R ma mc \<equiv>
    sim_ext R ma mc (force_same_and R) (force_same_and R) R"

lemmas sims_defs =
  sim_def
  sim_norm_def
  sim_normN_def
  sim_normE_def

lemmas sim_exts_defs = 
  sim_ext_def
  sim_norm_ext_def
  sim_normN_ext_def
  sim_normE_ext_def

(*******************************************************************************)
subsubsection {* Links between the simulation definitions *}
(*******************************************************************************)

lemma sim_imp_sim_norm:
  "sim R ma mc \<Longrightarrow> sim_norm R ma mc"
  by(auto simp add: sim_def sim_norm_def sim_ext_def sim_norm_ext_def)

lemma sim_norm_imp_sim_normN:
  "sim_norm R ma mc \<Longrightarrow> sim_normN R ma mc"
  by(auto simp add: sim_normN_def sim_normN_ext_def sim_norm_def sim_norm_ext_def)

lemma sim_norm_imp_sim_normE:
  "sim_norm R ma mc \<Longrightarrow> sim_normE R ma mc"
  by(auto simp add: sim_normE_def sim_normE_ext_def sim_norm_def sim_norm_ext_def)

lemma sim_norm_ext_decompose:
  "sim_norm_ext R ma mc Sn Se = (sim_normN_ext R ma mc Sn \<and> sim_normE_ext R ma mc Se)"
  by(auto simp add: sim_norm_ext_def sim_normN_ext_def sim_normE_ext_def)

lemma sim_norm_decompose:
  "sim_norm R ma mc = (sim_normN R ma mc \<and> sim_normE R ma mc)"
  by(auto simp add: sim_norm_def sim_norm_ext_def sim_normN_def sim_normN_ext_def
    sim_normE_def sim_normE_ext_def)

lemma sim_normN_through_simN:
  "sim_normN R ma mc = simN R (ma\<Down>NN) (mc\<Down>NN)"
  by(auto simp add: sim_normN_ext_def sim_normN_def simN_def)

lemma sim_normE_through_simN:
  "sim_normE R ma mc = simN R (ma\<Down>NE) (mc\<Down>NE)"
  by(auto simp add: sim_normE_ext_def sim_normE_def simN_def)

lemma sim_norm_through_simN:
  "sim_norm R ma mc = (simN R (ma\<Down>NN) (mc\<Down>NN) \<and> simN R (ma\<Down>NE) (mc\<Down>NE))"
  by(simp only: sim_norm_decompose sim_normN_through_simN sim_normE_through_simN)

lemma sim_through_simN_simS:
  "sim R ma mc = (
    simN R (ma\<Down>NN) (mc\<Down>NN)
    \<and> simN R (ma\<Down>NE) (mc\<Down>NE)
    \<and> simS R (ma\<Down>R) (mc\<Down>R)
  )"
  by(auto simp add: sim_def sim_ext_def simN_def simS_def)

lemmas sims_through_simN_simS =
  sim_normN_through_simN
  sim_normE_through_simN 
  sim_norm_through_simN 
  sim_through_simN_simS

(*******************************************************************************)
subsubsection {* Simulation and basic monad operations *}
(*******************************************************************************)

lemma simS_returnS:
  "simS R returnS returnS"
  by(auto simp add: simS_def simS_wpc_def unfold_monad)

lemma simS_bindS:
  "\<lbrakk> simS R ma1 mc1; simS R ma2 mc2 \<rbrakk> \<Longrightarrow>
    simS R (\<lambda>s. (ma1 s \<guillemotright>=S ma2)) (\<lambda>s. (mc1 s \<guillemotright>=S mc2))"
  by(auto simp add: simS_def intro!: simS_wpc_bindS)

lemma simS_choice_lockstep:
  "\<lbrakk> simS R ma1 mc1; simS R ma2 mc2 \<rbrakk> \<Longrightarrow> simS R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(auto simp add: simS_def intro!: simS_wpc_choice_lockstep del: subsetI)

lemma simS_empty:
  "simS R ma empty"
  by(simp add: simS_def simS_wpc_empty_conc)

lemma simS_bindN:
  "\<lbrakk> simN R ma1 mc1; \<forall>x. simS R (ma2 x) (mc2 x) \<rbrakk> \<Longrightarrow> simS R (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2)"
  by(auto simp add: simN_def simS_def force_same_and_def
    intro!: simS_wpc_bindN dest!: set_mp split: split_if_asm)

lemma simN_return:
  "simN R (returnN a) (returnN a)"
  by(auto simp add: simN_def simN_wpc_def unfold_monad force_same_and_def)

lemma simN_bindN:
  "\<lbrakk> simN R ma1 mc1; \<forall>v. simN R (ma2 v) (mc2 v) \<rbrakk> \<Longrightarrow>
    simN R (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2)"
  by(auto simp add: simN_def force_same_and_def intro!: simN_wpc_bindN
    dest!: set_mp split: split_if_asm)

lemma simN_choose:
  "simN R (choose A) (choose A)"
  by(auto simp add: simN_def force_same_and_def chooseN_def choose_def simN_wpc_def)


lemma simN_choice_lockstep:
  "\<lbrakk> simN R ma1 mc1; simN R ma2 mc2 \<rbrakk> \<Longrightarrow> simN R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(auto simp add: simN_def intro!: simN_wpc_choice_lockstep del: subsetI)

lemma simN_union_lockstep:
  "\<lbrakk> simN R ma1 mc1; simN R ma2 mc2 \<rbrakk> \<Longrightarrow> simN R (\<lambda>s. ma1 s \<union> ma2 s) (\<lambda>s. mc1 s \<union> mc2 s)"
  by(blast intro: simN_choice_lockstep[simplified choice_def]) 

lemma simN_empty:
  "simN R ma empty"
  by(auto simp add: simN_def simN_wpc_def unfold_monad)

lemma simN_wpc_bindN_bindS:
  "\<lbrakk> R \<subseteq> simN_wpc ma1 mc1 S';
    \<forall>x y. S' x y \<subseteq> simN_wpc (\<lambda>s. ma2 x s \<guillemotright>=S ha) (\<lambda>s. mc2 y s \<guillemotright>=S hc) S
  \<rbrakk> \<Longrightarrow>
    R \<subseteq> simN_wpc (\<lambda>s. (ma1 \<guillemotright>=N ma2) s \<guillemotright>=S ha) (\<lambda>s. (mc1 \<guillemotright>=N mc2) s \<guillemotright>=S hc) S"
  by(auto simp add: simN_wpc_def unfold_monad, smt mem_Collect_eq set_mp split_conv)  

lemma simN_bindN_bindS:
  "\<lbrakk> simN R ma1 mc1; \<forall>x. simN R (\<lambda>s. ma2 x s \<guillemotright>=S ha) (\<lambda>s. mc2 x s \<guillemotright>=S hc) \<rbrakk> \<Longrightarrow>
    simN R (\<lambda>s. (ma1 \<guillemotright>=N ma2) s \<guillemotright>=S ha) (\<lambda>s. (mc1 \<guillemotright>=N mc2) s \<guillemotright>=S hc)
  "
  by(auto simp add: simN_def force_same_and_def intro!: simN_wpc_bindN_bindS del: subsetI)

lemma simN_bindS:
  "\<lbrakk> simS R ma mc; simN R ha hc \<rbrakk> \<Longrightarrow> simN R (\<lambda>s. ma s \<guillemotright>=S ha) (\<lambda>s. mc s \<guillemotright>=S hc)"
  by(auto simp add: simN_def simS_def intro!: simN_wpc_bindS)

lemma simS_mapM:
  "\<lbrakk> \<forall>x \<in> set l. simN R ((ma x)\<Down>NN) ((mc x)\<Down>NN);
    \<forall>x \<in> set l. simS R ((ma x)\<Down>R) ((mc x)\<Down>R)
  \<rbrakk> \<Longrightarrow>
    simS R ((mapM' ma l)\<Down>R) ((mapM' mc l)\<Down>R)
  "
  apply(induct l)
  apply(simp add: project_simps simS_empty)
  apply(auto simp add: project_simps intro!: simS_choice_lockstep simS_bindN del: subsetI)
  done

lemma simN_mapM:
  "\<lbrakk> \<forall>x \<in> set l. simN R ((ma x)\<Down>NN) ((mc x)\<Down>NN);
    \<forall>x \<in> set l. simN R ((ma x)\<Down>NE) ((mc x)\<Down>NE)
  \<rbrakk> \<Longrightarrow>
    simN R ((mapM' ma l)\<Down>NE) ((mapM' mc l)\<Down>NE)
  "
  "\<lbrakk> \<forall>x \<in> set l. simN R ((ma' x)\<Down>NN) ((mc' x)\<Down>NN) \<rbrakk> \<Longrightarrow>
    simN R ((mapM' ma' l)\<Down>NN) ((mapM' mc' l)\<Down>NN)
  "
  apply(induct l)
  apply(simp_all add: project_simps simN_empty)
  apply(auto simp add: project_simps intro!: simN_choice_lockstep simN_return simN_bindN del: subsetI)
  done

lemma sim_exts_empty_rel:
  "sim_ext {} ma mc Sn Se R"
  "sim_norm_ext {} ma mc Sn Se"
  "sim_normN_ext {} ma mc Sn"
  "sim_normE_ext {} ma mc Se"
  by(simp_all add: sim_ext_def sim_norm_ext_def sim_normN_ext_def sim_normE_ext_def)

text {* Following two lemmas are not derivable from corresponding
    sim/sim_norm/... lemmas, since here we only constrain certain
    result relations *}

lemma sim_ext_return:
  "sim_ext R (return a) (return a) (force_same_and R) Se Sr"
  by(auto simp add: sim_ext_def simN_wpc_empty_conc simS_wpc_empty_conc
    simN_wpc_returnN project_simps force_same_and_def)

lemma sim_ext_throw:
  "sim_ext R (throw a) (throw a) Sn (force_same_and R) Sr"
  by(auto simp add: sim_ext_def simN_wpc_empty_conc simS_wpc_empty_conc simN_wpc_returnN project_simps
    force_same_and_def)

lemma sim_ext_bindI:
  "\<lbrakk> sim_ext R ma1 mc1 Tn Se Sr; \<forall>va vc. sim_ext (Tn va vc) (ma2 va) (mc2 vc) Sn Se Sr \<rbrakk> \<Longrightarrow>
    sim_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se Sr"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
  
lemma sim_ext_choice_abs_I1:
  "sim_ext R ma1 mc Sn Se Sr \<Longrightarrow> sim_ext R (ma1 [+] ma2) mc Sn Se Sr"
  by(auto simp add: sim_ext_def project_simps
    intro: simN_wpc_choice_abs_I1 simS_wpc_choice_abs_I1)

lemma sim_ext_choice_abs_I2:
  "sim_ext R ma2 mc Sn Se Sr \<Longrightarrow> sim_ext R (ma1 [+] ma2) mc Sn Se Sr"
  by(auto simp add: sim_ext_def project_simps
    intro: simN_wpc_choice_abs_I2 simS_wpc_choice_abs_I2)

lemma sim_norm_ext_bindI:
  "\<lbrakk> sim_norm_ext R ma1 mc1 Tn Se; \<forall>va vc. sim_norm_ext (Tn va vc) (ma2 va) (mc2 vc) Sn Se \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn Se"
 by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
  
lemma sim_normN_ext_bindI:
  "\<lbrakk> sim_normN_ext R ma1 mc1 Tn; \<forall>va vc. sim_normN_ext (Tn va vc) (ma2 va) (mc2 vc) Sn \<rbrakk> \<Longrightarrow>
    sim_normN_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Sn"
 by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
  
lemma sim_normE_ext_bindI:
  "\<lbrakk> sim_normN_ext R ma1 mc1 Tn; sim_normE_ext R ma1 mc1 Se;
        \<forall>va vc. sim_normE_ext (Tn va vc) (ma2 va) (mc2 vc) Se \<rbrakk> \<Longrightarrow>
    sim_normE_ext R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2) Se"
 by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)

lemmas sim_exts_bindI =
  sim_ext_bindI sim_norm_ext_bindI sim_normN_ext_bindI sim_normE_ext_bindI
  

lemma sim_ext_try_catchI:
  "\<lbrakk> sim_ext R ma mc Sn Te Sr; \<forall>ea ec. sim_ext (Te ea ec) (ha ea) (hc ec) Sn Se Sr \<rbrakk> \<Longrightarrow>
    sim_ext R (try_catch ma ha) (try_catch mc hc) Sn Se Sr"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
 
lemma sim_norm_ext_try_catchI:
  "\<lbrakk> sim_norm_ext R ma mc Sn Te; \<forall>ea ec. sim_norm_ext (Te ea ec) (ha ea) (hc ec) Sn Se \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catch ma ha) (try_catch mc hc) Sn Se"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
 
lemma sim_normN_ext_try_catchI:
  "\<lbrakk> sim_normN_ext R ma mc Sn; sim_normE_ext R ma mc Te;
    \<forall>ea ec. sim_normN_ext (Te ea ec) (ha ea) (hc ec) Sn \<rbrakk> \<Longrightarrow>
    sim_normN_ext R (try_catch ma ha) (try_catch mc hc) Sn"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
 
lemma sim_normE_ext_try_catchI:
  "\<lbrakk> sim_normE_ext R ma mc Te; \<forall>ea ec. sim_normE_ext (Te ea ec) (ha ea) (hc ec) Se \<rbrakk> \<Longrightarrow>
    sim_normE_ext R (try_catch ma ha) (try_catch mc hc) Se"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set
    del: subsetI)
 
lemmas sim_exts_try_catchI =
  sim_norm_ext_try_catchI sim_norm_ext_try_catchI sim_normN_ext_try_catchI
  sim_normE_ext_try_catchI

lemma sim_ext_mapM':
  "\<forall>x \<in> set l. sim_ext R (ma x) (mc x) (force_same_and R) Se Sr \<Longrightarrow>
    sim_ext R (mapM' ma l) (mapM' mc l) (force_same_and R) Se Sr"
  apply(induct l)
  apply(auto intro: sim_ext_return)
  apply(auto simp add: force_same_and_def intro!: sim_ext_bindI sim_exts_empty_rel sim_ext_return)
  done

lemma simN_choice_abs_I1:
  "simN R ma1 mc \<Longrightarrow> simN R (ma1 [+] ma2) mc"
  apply(simp only: simN_def projectNN_choice)
  apply(auto intro!: Int_greatest  subset_trans[OF _ simN_wpc_choice_abs])
  done

lemma simN_choice_abs_I2:
  "simN R ma2 mc \<Longrightarrow> simN R (ma1 [+] ma2) mc"
  apply(simp only: simN_def projectNN_choice)
  apply(auto intro!: Int_greatest  subset_trans[OF _ simN_wpc_choice_abs])
  done

lemma simS_choice_abs_I1:
  "simS R ma1 mc \<Longrightarrow> simS R (ma1 [+] ma2) mc"
  apply(simp only: simS_def projectNN_choice)
  apply(auto intro!: Int_greatest  subset_trans[OF _ simS_wpc_choice_abs])
  done


lemma simS_choice_abs_I2:
  "simS R ma2 mc \<Longrightarrow> simS R (ma1 [+] ma2) mc"
  apply(simp only: simS_def projectNN_choice)
  apply(auto intro!: Int_greatest  subset_trans[OF _ simS_wpc_choice_abs])
  done

lemma sim_choice_abs_I1:
  "sim R ma1 mc \<Longrightarrow> sim R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I1 simS_choice_abs_I1)

lemma sim_norm_choice_abs_I1:
  "sim_norm R ma1 mc \<Longrightarrow> sim_norm R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I1 simS_choice_abs_I1)

lemma sim_normN_choice_abs_I1:
  "sim_normN R ma1 mc \<Longrightarrow> sim_normN R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I1 simS_choice_abs_I1)

lemma sim_normE_choice_abs_I1:
  "sim_normE R ma1 mc \<Longrightarrow> sim_normE R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I1 simS_choice_abs_I1)

lemmas sims_choice_abs_I1 =
  sim_choice_abs_I1
  sim_norm_choice_abs_I1
  sim_normN_choice_abs_I1
  sim_normE_choice_abs_I1

lemma sim_choice_abs_I2:
  "sim R ma2 mc \<Longrightarrow> sim R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I2 simS_choice_abs_I2)

lemma sim_norm_choice_abs_I2:
  "sim_norm R ma2 mc \<Longrightarrow> sim_norm R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I2 simS_choice_abs_I2)

lemma sim_normN_choice_abs_I2:
  "sim_normN R ma2 mc \<Longrightarrow> sim_normN R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I2 simS_choice_abs_I2)

lemma sim_normE_choice_abs_I2:
  "sim_normE R ma2 mc \<Longrightarrow> sim_normE R (ma1 [+] ma2) mc"
  by(auto simp only: sims_through_simN_simS project_simps
    intro: simN_choice_abs_I2 simS_choice_abs_I2)

lemmas sims_choice_abs_I2 = 
  sim_choice_abs_I2
  sim_norm_choice_abs_I2
  sim_normN_choice_abs_I2
  sim_normE_choice_abs_I2

lemma simS_choice_conc: "simS R ma (mc1 [+] mc2) = (simS R ma mc1 \<and> simS R ma mc2)"
  by(auto simp add: simS_def simS_wpc_def force_same_and_def choice_def)

lemma simN_choice_conc: "simN R ma (mc1 [+] mc2) = (simN R ma mc1 \<and> simN R ma mc2)"
  by(auto simp add: simN_def simN_wpc_def force_same_and_def choice_def)

lemma sim_choice_conc:
  "sim R ma (mc1 [+] mc2) = (sim R ma mc1 \<and> sim R ma mc2)"
  by(auto simp only: sims_through_simN_simS simS_choice_conc simN_choice_conc project_simps)

lemma sim_norm_choice_conc:
  "sim_norm R ma (mc1 [+] mc2) = (sim_norm R ma mc1 \<and> sim_norm R ma mc2)"
  by(auto simp only: sims_through_simN_simS simS_choice_conc simN_choice_conc project_simps)

lemma sim_normN_choice_conc:
  "sim_normN R ma (mc1 [+] mc2) = (sim_normN R ma mc1 \<and> sim_normN R ma mc2)"
  by(auto simp only: sims_through_simN_simS simS_choice_conc simN_choice_conc project_simps)

lemma sim_normE_choice_conc:
  "sim_normE R ma (mc1 [+] mc2) = (sim_normE R ma mc1 \<and> sim_normE R ma mc2)"
  by(auto simp only: sims_through_simN_simS simS_choice_conc simN_choice_conc project_simps)

lemmas sims_choice_conc = 
  sim_choice_conc
  sim_norm_choice_conc
  sim_normN_choice_conc
  sim_normE_choice_conc

lemmas sims_choice_concI = sims_choice_conc[THEN iffD2, OF conjI]

lemma sim_chooseN:
  "sim R (chooseN A) (chooseN A)"
  by(simp_all add: sims_through_simN_simS project_simps simN_empty simS_empty simN_choose)

lemma sim_norm_chooseN:
  "sim_norm R (chooseN A) (chooseN A)"
  by(simp_all add: sims_through_simN_simS project_simps simN_empty simS_empty simN_choose)

lemma sim_normN_chooseN:
  "sim_normN R (chooseN A) (chooseN A)"
  by(simp_all add: sims_through_simN_simS project_simps simN_empty simS_empty simN_choose)

lemma sim_normE_chooseN:
  "sim_normE R (chooseN A) (chooseN A)"
  by(simp_all add: sims_through_simN_simS project_simps simN_empty simS_empty simN_choose)

lemmas sims_chooseN = 
  sim_chooseN sim_norm_chooseN
  sim_normN_chooseN sim_normE_chooseN

lemma sim_bindI:
  "\<lbrakk> sim R ma1 mc1; \<forall>x. sim R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> sim R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_norm_bindI:
  "\<lbrakk> sim_norm R ma1 mc1; \<forall>x. sim_norm R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> sim_norm R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_normN_bindI:
  "\<lbrakk> sim_normN R ma1 mc1; \<forall>x. sim_normN R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> sim_normN R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_normE_bindI:
  "\<lbrakk> sim_normN R ma1 mc1; sim_normE R ma1 mc1; \<forall>x. sim_normE R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> sim_normE R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemmas sims_bindI = sim_bindI sim_norm_bindI sim_normN_bindI sim_normE_bindI

lemma simS_bindN_inv_conc:
  "\<lbrakk> simN R ma1 mc1; 
    Range R \<subseteq> [N x \<leftarrow> mc1] (P x);
    \<forall>x. P x \<noteq> {} \<longrightarrow> simS R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> simS R (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2)"
  apply(auto simp add: simN_def simS_def)
  apply(rule_tac S'="(\<lambda>va vc. (if va = vc then UNIV else {}) \<inter> R \<inter> (UNIV \<times> P vc))" in simS_wpc_bindN)
  apply(auto simp add: simN_wpc_def force_same_and_def Box_def dest!: set_mp
    split: split_if_asm)
  done

lemma simN_bindN_inv_conc:
  "\<lbrakk> simN R ma1 mc1; 
    Range R \<subseteq> [N x \<leftarrow> mc1] (P x);
    \<forall>x. P x \<noteq> {} \<longrightarrow> simN R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> simN R (ma1 \<guillemotright>=N ma2) (mc1 \<guillemotright>=N mc2)"
  apply(auto simp add: simN_def)
  apply(rule_tac S'="(\<lambda>va vc. (if va = vc then UNIV else {}) \<inter> R \<inter> (UNIV \<times> P vc))" in simN_wpc_bindN)
  apply(auto simp add: simN_wpc_def force_same_and_def Box_def dest!: set_mp
    split: split_if_asm)
  done

lemma BoxNN_through_Box_project:
  "BoxNN m P = Box (m\<Down>NN) P"
  by(auto simp add: Box_def BoxN_def BoxNN_def liftPredVal_def project_defs split: sum.split)

lemma sim_bindI_inv_conc:
  "\<lbrakk> sim R ma1 mc1; Range R \<subseteq> [ x \<leftarrow> mc1] (P x); \<forall>x. P x \<noteq> {} \<longrightarrow> sim R (ma2 x) (mc2 x) \<rbrakk>
     \<Longrightarrow> sim R (ma1 \<guillemotright>= ma2) (mc1 \<guillemotright>= mc2)"
  by(auto simp add: sim_through_simN_simS project_simps BoxNN_through_Box_project
    intro!: simN_choice_lockstep simS_choice_lockstep simS_bindN_inv_conc simN_bindN_inv_conc)

lemma sim_try_catchI:
  "\<lbrakk> sim R ma mc; \<forall>e. sim R (ha e) (hc e) \<rbrakk> \<Longrightarrow> sim R (try_catch ma ha) (try_catch mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_norm_try_catchI:
  "\<lbrakk> sim_norm R ma mc; \<forall>e. sim_norm R (ha e) (hc e) \<rbrakk>
    \<Longrightarrow> sim_norm R (try_catch ma ha) (try_catch mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_normN_try_catchI:
  "\<lbrakk> sim_normN R ma mc; sim_normE R ma mc; \<forall>e. sim_normE R (ha e) (hc e) \<rbrakk>
    \<Longrightarrow> sim_normE R (try_catch ma ha) (try_catch mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemma sim_normE_try_catchI:
  "\<lbrakk> sim_normE R ma mc; \<forall>e. sim_normE R (ha e) (hc e) \<rbrakk>
    \<Longrightarrow> sim_normE R (try_catch ma ha) (try_catch mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindN simS_bindN)

lemmas sims_try_catchI =
  sim_try_catchI sim_norm_try_catchI
  sim_normN_try_catchI sim_normE_try_catchI

lemma sim_try_catchRI:
  "\<lbrakk> sim R ma mc; sim R ha hc \<rbrakk> \<Longrightarrow> sim R (try_catchR ma ha) (try_catchR mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindS simS_bindS)

lemma sim_norm_try_catchRI:
  "\<lbrakk> sim R ma mc; sim_norm R ha hc \<rbrakk> \<Longrightarrow> sim_norm R (try_catchR ma ha) (try_catchR mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindS simS_bindS)

lemma sim_normN_try_catchRI:

  "\<lbrakk> sim R ma mc; sim_normN R ha hc \<rbrakk> \<Longrightarrow>
  sim_normN R (try_catchR ma ha) (try_catchR mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindS simS_bindS)

lemma sim_normE_try_catchRI:
  "\<lbrakk> sim R ma mc; sim_normE R ha hc \<rbrakk> \<Longrightarrow>
  sim_normE R (try_catchR ma ha) (try_catchR mc hc)"
  by(auto simp add: sims_through_simN_simS project_simps
    intro!: simN_choice_lockstep simS_choice_lockstep simN_bindS simS_bindS)

lemmas sims_try_catchRI =
  sim_try_catchRI sim_norm_try_catchRI
  sim_normN_try_catchRI sim_normE_try_catchRI

lemma sims_return:
  shows sim_return:
  "sim R (return a) (return a)"
  and sim_norm_return:
  "sim_norm R (return a) (return a)"
  and sim_normN_return:
  "sim_normN R (return a) (return a)"
  and sim_normE_return:
  "sim_normE R (return a) (return a)"
  by(simp_all add: sims_through_simN_simS simN_return project_simps simN_empty simS_empty)

lemma sims_throw:
  "sim R (throw a) (throw a)"
  "sim_norm R (throw a) (throw a)"
  "sim_normN R (throw a) (throw a)"
  "sim_normE R (throw a) (throw a)"
  by(simp_all add: sims_through_simN_simS simN_return project_simps simN_empty simS_empty)

lemma sim_choice_lockstep:
  "\<lbrakk> sim R ma1 mc1; sim R ma2 mc2 \<rbrakk> \<Longrightarrow> sim R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(simp_all add: sims_through_simN_simS project_simps simN_choice_lockstep simS_choice_lockstep)

lemma sim_norm_choice_lockstep:
  "\<lbrakk> sim_norm R ma1 mc1; sim_norm R ma2 mc2 \<rbrakk> \<Longrightarrow> sim_norm R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(simp_all add: sims_through_simN_simS project_simps simN_choice_lockstep simS_choice_lockstep)

lemma sim_normN_choice_lockstep:
  "\<lbrakk> sim_normN R ma1 mc1; sim_normN R ma2 mc2 \<rbrakk> \<Longrightarrow>
  sim_normN R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(simp_all add: sims_through_simN_simS project_simps simN_choice_lockstep simS_choice_lockstep)

lemma sim_normE_choice_lockstep:
  "\<lbrakk> sim_normE R ma1 mc1; sim_normE R ma2 mc2 \<rbrakk> \<Longrightarrow>
  sim_normE R (ma1 [+] ma2) (mc1 [+] mc2)"
  by(simp_all add: sims_through_simN_simS project_simps simN_choice_lockstep simS_choice_lockstep)

lemmas sims_choice_lockstep =
  sim_choice_lockstep sim_norm_choice_lockstep
  sim_normN_choice_lockstep sim_normE_choice_lockstep
 
lemma sim_if_I:
  "\<lbrakk> cond \<longrightarrow> sim R maT mcT; \<not>cond \<longrightarrow> sim R maF mcF \<rbrakk> \<Longrightarrow> 
    sim R (if cond then maT else maF) (if cond then mcT else mcF)
  "
  by(auto)

lemma sim_mapM': 
  "\<forall>x \<in> set l. sim R (ma x) (mc x) \<Longrightarrow> sim R (mapM' ma l) (mapM' mc l)"
  by(auto simp add: sims_through_simN_simS intro!: simN_mapM simS_mapM)

lemma sim_norm_mapM': 
  "\<forall>x \<in> set l. sim_norm R (ma x) (mc x) \<Longrightarrow> sim_norm R (mapM' ma l) (mapM' mc l)"
  by(auto simp add: sims_through_simN_simS intro!: simN_mapM simS_mapM)

lemma sim_normN_mapM': 
  "\<forall>x \<in> set l. sim_normN R (ma x) (mc x) \<Longrightarrow> sim_normN R (mapM' ma l) (mapM' mc l)"
  by(auto simp add: sims_through_simN_simS intro!: simN_mapM simS_mapM)

lemma sim_normE_mapM': 
  "\<forall>x \<in> set l. sim_norm R (ma x) (mc x) \<Longrightarrow> sim_normE R (mapM' ma l) (mapM' mc l)"
  by(auto simp add: sims_through_simN_simS intro!: simN_mapM simS_mapM)

lemmas sim_mapM =
  sim_mapM' sim_norm_mapM' 
  sim_normN_mapM' sim_normE_mapM'

(*******************************************************************************)
subsection {* Refinement and Hoare triples *}
(*******************************************************************************)

lemma simN_wpc_add_inv:
  "\<lbrakk>\<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>;
    R \<inter> I \<times> J \<subseteq> simN_wpc ma mc S'
  \<rbrakk> \<Longrightarrow>
    R \<inter> I \<times> J \<subseteq> simN_wpc ma mc (\<lambda>va vc. S' va vc \<inter> (I \<times> J))
  "
  using [[simp_depth_limit=0]]
  apply(auto simp add: simN_wpc_def inv_explicit)
  apply(blast)
  done

lemma simS_wpc_add_inv: "\<lbrakk> \<lbrace>I\<rbrace> ma \<lbrace>S I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>S J\<rbrace>;
    R \<inter> I \<times> J \<subseteq> simS_wpc ma mc R
  \<rbrakk> \<Longrightarrow>
  R \<inter> I \<times> J \<subseteq> simS_wpc ma mc (R \<inter> I \<times> J)"
  by(auto simp add: simS_wpc_def HoareS_def BoxS_def project_defs
    dest!: set_mp split: split_if_asm)

lemma simN_add_inv:
  "\<lbrakk> simN R ma mc;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> simN (R \<inter> I \<times> J) ma mc"
  apply(auto simp only: simN_def force_same_and_def Int_assoc[THEN sym]
    intro!: simN_wpc_add_inv del: subsetI)
  apply(auto intro: subset_trans)
  done

lemma simS_add_inv:
  "\<lbrakk> simS R ma mc;
    \<lbrace>I\<rbrace> ma \<lbrace>S I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>S J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> simS (R \<inter> I \<times> J) ma mc"
  apply(auto simp only: simS_def force_same_and_def Int_assoc[THEN sym]
    intro!: simS_wpc_add_inv del: subsetI)
  apply(auto intro: subset_trans)
  done

lemma HoareR_HoareS:
  "\<lbrace>I\<rbrace> m \<lbrace>@R I\<rbrace> = \<lbrace>I\<rbrace> (m\<Down>R) \<lbrace>S I\<rbrace>"
  by(auto simp add: HoareS_def HoareR_def Box_defs BoxS_def project_defs)

lemma simN_wpc_monotone_comp_subset:
  "\<lbrakk> R \<subseteq> simN_wpc ma mc' S; \<forall>s\<in> snd ` R. mc s \<subseteq> mc' s \<rbrakk> \<Longrightarrow> R \<subseteq> simN_wpc ma mc S"
  by(force simp add: simN_wpc_def)

lemma sim_normN_ext_add_inv: "
  \<lbrakk>
    sim_normN_ext (R \<inter> I \<times> J) ma mc S;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> sim_normN_ext (R \<inter> I \<times> J) ma mc (\<lambda>va vc. S va vc \<inter> (I \<times> J))
  "
  by(auto simp add: sim_normN_ext_def intro!: simN_wpc_add_inv K_Hoare_to_NN del: subsetI)

lemma sim_normE_ext_add_inv: "
  \<lbrakk>
    sim_normE_ext (R \<inter> I \<times> J) ma mc S;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> sim_normE_ext (R \<inter> I \<times> J) ma mc (\<lambda>va vc. S va vc \<inter> (I \<times> J))
  "
  by(auto simp add: sim_normE_ext_def intro!: simN_wpc_add_inv K_Hoare_to_NE del: subsetI)

lemma sim_norm_ext_add_inv: "
  \<lbrakk>
    sim_norm_ext (R \<inter> I \<times> J) ma mc Sn Se;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> sim_norm_ext (R \<inter> I \<times> J) ma mc (\<lambda>va vc. Sn va vc \<inter> (I \<times> J)) (\<lambda>ea ec. Se ea ec \<inter> (I \<times> J))"
  by(auto simp add: sim_norm_ext_decompose intro!: sim_normN_ext_add_inv sim_normE_ext_add_inv)

lemma sim_norm_add_inv:
  "\<lbrakk> sim_norm R ma mc;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> sim_norm (R \<inter> I \<times> J) ma mc"
  by(auto simp only: sims_through_simN_simS intro!: simN_add_inv K_Hoare_to_NE K_Hoare_to_NN)

lemma sim_normN_add_inv:
  "\<lbrakk> \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>;
   \<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>;
    sim_normN R ma mc
   \<rbrakk> \<Longrightarrow>
  sim_normN (R \<inter> (I \<times> J)) ma mc"
  by(auto simp only: sims_through_simN_simS intro!: simN_add_inv K_Hoare_to_NE K_Hoare_to_NN)

lemma sim_normE_add_inv:
  "\<lbrakk> \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>;
   \<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>;
    sim_normE R ma mc
   \<rbrakk> \<Longrightarrow>
  sim_normE (R \<inter> I \<times> J) ma mc"
  by(auto simp only: sims_through_simN_simS intro!: simN_add_inv K_Hoare_to_NE K_Hoare_to_NN)

lemma sim_add_inv:
  "\<lbrakk> sim R ma mc;
    \<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>; \<lbrace>J\<rbrace> mc \<lbrace>>K J\<rbrace>
  \<rbrakk>
  \<Longrightarrow> sim (R \<inter> I \<times> J) ma mc"
  by(auto simp only: sims_through_simN_simS intro!: simN_add_inv simS_add_inv
    K_Hoare_to_NE K_Hoare_to_NN K_Hoare_to_R)

(*******************************************************************************)
subsection {* Monotonicity properties of the various simulation relations *}
(*******************************************************************************)

lemma sim_normN_ext_mono:
  "\<lbrakk> sim_normN_ext R ma mc Sn; R' \<subseteq> R \<rbrakk> \<Longrightarrow> sim_normN_ext R' ma mc Sn"
  by(auto simp add: sim_normN_ext_def)

lemma sim_normE_ext_mono:
  "\<lbrakk> sim_normE_ext R ma mc Se; R' \<subseteq> R \<rbrakk> \<Longrightarrow> sim_normE_ext R' ma mc Se"
  by(auto simp add: sim_normE_ext_def)

lemma sim_norm_ext_mono:
  "\<lbrakk> sim_norm_ext R ma mc Sn Se; R' \<subseteq> R \<rbrakk> \<Longrightarrow> sim_norm_ext R' ma mc Sn Se"
  apply(auto simp add: sim_norm_ext_decompose)
  apply(blast intro!: sim_normN_ext_mono sim_normE_ext_mono)+
  done

lemma sim_ext_mono:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; R' \<subseteq> R \<rbrakk> \<Longrightarrow> sim_ext R' ma mc Sn Se Sr"
  by(auto simp add: sim_ext_def)
  
lemma sim_ext_mono_norm:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; \<forall>x y. Sn x y \<subseteq> Sn' x y  \<rbrakk>
    \<Longrightarrow> sim_ext R ma mc Sn' Se Sr"
  apply(auto simp add: sim_ext_def del: subsetI)
  apply(rule subset_trans[where B="simN_wpc (ma\<Down>NN) (mc\<Down>NN) Sn"])
  apply(auto intro!: simN_wpc_monotone_rel del: subsetI)
  done

lemma sim_ext_mono_exc:
  "\<lbrakk>sim_ext R ma mc Sn Se Sr; \<forall>x y. Se x y \<subseteq> Se' x y\<rbrakk>
  \<Longrightarrow> sim_ext R ma mc Sn Se' Sr"
  apply(auto simp add: sim_ext_def del: subsetI)
  apply(rule subset_trans[where B="simN_wpc (ma\<Down>NE) (mc\<Down>NE) Se"])
  apply(auto intro!: simN_wpc_monotone_rel del: subsetI)
  done

lemma sim_ext_mono_abs_comp:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; \<forall>s. ma s \<subseteq> ma' s \<rbrakk> \<Longrightarrow> sim_ext R ma' mc Sn Se Sr"
  apply(auto simp add: sim_ext_def del: subsetI)
  apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs])
  apply(force, force simp add: project_defs)
  apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs])
  apply(force, force simp add: project_defs)
  apply(rule subset_trans[OF _ simS_wpc_monotone_comp_abs])
  apply(force, force simp add: project_defs)
  done

lemma projectN_mono:
  "m s \<subseteq> m' s \<Longrightarrow> (m\<Down>N) s \<subseteq> (m'\<Down>N) s"
  by(auto simp add: project_defs)

lemma try_catchR_bound_repeat:
  "((v, s') \<in> (<tryR> m <catchR> rec_tryR h) s) = 
  (\<exists>n. ((v, s') \<in> (<tryR> m <catchR> catchR_N h n) s))"
  apply(case_tac v, auto simp add: try_catchR_none_decompose try_catchR_some_decompose)
  apply(auto dest!: repeat_bounded)
  apply(auto dest!: bounded_into_repeat)
  done

lemma try_rec_tryR_inclusion:
  "(<tryR> m <catchR> rec_tryR m) s \<subseteq> rec_tryR m s"
proof(rule subsetI, case_tac x, simp_all)
  fix v s'
  assume inL:
    "(v, s') \<in> (<tryR> m <catchR> rec_tryR m) s"
  thus "(v, s') \<in> rec_tryR m s"
    apply(auto simp add: try_catchR_bound_repeat)
    apply(rule_tac n="Suc n" in bounded_into_repeat)
    apply(subst (asm) catchR_N.simps(2)[THEN sym])
    apply(assumption)
    done
qed

declare disjCI[rule del]
lemma rec_tryR_inclusion:
  "rec_tryR m s \<subseteq> m s \<union> (<tryR> m <catchR> rec_tryR m) s"
proof(rule subsetI, case_tac x, simp)
  fix v s'
  assume inL:
    "(v, s') \<in> rec_tryR m s"
  from repeat_bounded[OF this] obtain n where "(v, s') \<in> catchR_N m n s" ..
  thus 
    "(v, s') \<in> m s \<or> (v, s') \<in> (<tryR> m <catchR> rec_tryR m) s"
    apply(case_tac n)
    apply(simp)
    apply(auto simp add: try_catchR_bound_repeat)
    apply(rule disjI2)
    apply(force)
    done
qed

lemma try_rec_tryR_same:
  "((<tryR> m <catchR> rec_tryR m)\<Down>N) = ((rec_tryR m)\<Down>N)"
  apply(auto intro!: ext projectN_mono[OF try_rec_tryR_inclusion] del: subsetI)
  apply(rule rec_tryR_inclusion[THEN projectN_mono, THEN order_trans])
  apply(auto simp add: project_defs try_catchR_some_decompose)
  done
  
lemma projectN_incl_imp_projectNN_incl:
  "(m\<Down>N) s \<subseteq> (m'\<Down>N) s \<Longrightarrow> (m\<Down>NN) s \<subseteq> (m'\<Down>NN) s"
  by(auto simp add: project_defs)

lemma projectN_incl_imp_projectNE_incl:
  "(m\<Down>N) s \<subseteq> (m'\<Down>N) s \<Longrightarrow> (m\<Down>NE) s \<subseteq> (m'\<Down>NE) s"
  by(auto simp add: project_defs)

lemma sim_norm_mono_abs:
  "\<lbrakk> sim_norm R ma' mc; \<forall>s. (ma'\<Down>N) s \<subseteq> (ma\<Down>N) s \<rbrakk> \<Longrightarrow> sim_norm R ma mc"
  by(auto simp add: sim_norm_def sim_norm_ext_def del: subsetI
    elim!: order_trans[OF _ simS_wpc_monotone_comp_abs] order_trans[OF _ simN_wpc_monotone_comp_abs]
    dest: projectN_incl_imp_projectNN_incl projectN_incl_imp_projectNE_incl
  )
 
lemma sim_norm_mono_conc:
  "\<lbrakk> sim_norm R ma mc'; \<forall>s. (mc\<Down>N) s \<subseteq> (mc'\<Down>N) s \<rbrakk> \<Longrightarrow> sim_norm R ma mc"
  by(auto simp add: sim_norm_def sim_norm_ext_def del: subsetI
    elim!: simN_wpc_monotone_comp_subset order_trans[OF _ simS_wpc_monotone_comp_conc]
    dest: projectN_incl_imp_projectNN_incl projectN_incl_imp_projectNE_incl
  )
  
lemma sim_norm_ext_mono_abs:
  "\<lbrakk> sim_norm_ext R ma' mc Sn Se; \<forall>s. (ma'\<Down>N) s \<subseteq> (ma\<Down>N) s \<rbrakk> \<Longrightarrow> sim_norm_ext R ma mc Sn Se"
  by(auto simp add: sim_norm_ext_def del: subsetI
    elim!: order_trans[OF _ simS_wpc_monotone_comp_abs] order_trans[OF _ simN_wpc_monotone_comp_abs]
    dest: projectN_incl_imp_projectNN_incl projectN_incl_imp_projectNE_incl
  )
 
lemma sim_norm_ext_mono_conc:
  "\<lbrakk> sim_norm_ext R ma mc' Sn Se; \<forall>s. (mc\<Down>N) s \<subseteq> (mc'\<Down>N) s \<rbrakk> \<Longrightarrow> sim_norm_ext R ma mc Sn Se"
  by(auto simp add: sim_norm_ext_def del: subsetI
    elim!: simN_wpc_monotone_comp_subset order_trans[OF _ simS_wpc_monotone_comp_conc]
    dest: projectN_incl_imp_projectNN_incl projectN_incl_imp_projectNE_incl
  )

lemma bindN_mono:
  "\<lbrakk> \<forall>s. ml s \<subseteq> ml' s; \<forall>a s. mr a s \<subseteq> mr' a s \<rbrakk> \<Longrightarrow>
    (ml \<guillemotright>=N mr) s \<subseteq> (ml' \<guillemotright>=N mr') s"
  apply(auto simp add: bindN_def bindS_def)
  by (metis set_mp split_conv)

(*******************************************************************************)
subsubsection {* try_catchR transformation and simulation *}
(*******************************************************************************)

lemma sim_catchR_N_repeatI:
  "sim R ha hc \<Longrightarrow> sim R (catchR_N ha n) (catchR_N hc n)"
  by(induct n, auto intro: sims_try_catchRI)

lemma repeat_rewrite: "(r, s') \<in> rec_tryR h s = (\<exists>n. (r, s') \<in> catchR_N h n s)"
  by(auto intro: repeat_bounded bounded_into_repeat)

lemma sim_catchR_N1: "sim R (catchR_N ha n) hc \<Longrightarrow> sim R (rec_tryR ha) hc"
  apply(auto simp add: sim_def sim_ext_def project_defs simN_wpc_def simS_wpc_def)
  apply(force simp only: repeat_rewrite dest!: subsetD split: split_if_asm)+
  done

lemma sim_catchR_N2: "\<forall>n. sim R ha (catchR_N hc n) \<Longrightarrow> sim R ha (rec_tryR hc)"
  apply(auto simp add: sim_def sim_ext_def project_defs simN_wpc_def simS_wpc_def)
  apply(force simp only: repeat_rewrite dest!: subsetD split: split_if_asm)+
  done

lemma sim_rec_tryRI:
  "sim R ha hc \<Longrightarrow> sim R (rec_tryR ha) (rec_tryR hc)"
  apply(intro sim_catchR_N2 allI)
  apply(auto intro: sim_catchR_N1 sim_catchR_N_repeatI)
  done

lemma simN_wpc_repeatN_I: "R \<subseteq> simN_wpc ma mc (\<lambda>va vc. R) \<Longrightarrow> 
    R \<subseteq> simN_wpc (repeatN_N ma n) (repeatN_N mc n) (\<lambda>va vc. R)"
  apply(induct n)
  apply(auto simp add: simN_wpc_returnN intro!: simN_wpc_bindN)
  done

lemma simN_wpc_repeatN_N1: "R \<subseteq> simN_wpc (repeatN_N ma n) mc (\<lambda>va vc. R) \<Longrightarrow> 
    R \<subseteq> simN_wpc (repeatN ma) mc (\<lambda>va vc. R)"
  by(force simp add: simN_wpc_def repeatN_rewrite)

lemma simN_wpc_repeatN_N2: "\<forall>n. R \<subseteq> simN_wpc ma (repeatN_N mc n) (\<lambda>va vc. R) \<Longrightarrow> 
    R \<subseteq> simN_wpc ma (repeatN mc) (\<lambda>va vc. R)"
  by(force simp add: simN_wpc_def repeatN_rewrite)

lemma simN_wpc_repeatN:
  "R \<subseteq> simN_wpc ma mc (\<lambda>va vc. R) \<Longrightarrow>
    R \<subseteq> simN_wpc (repeatN ma) (repeatN mc) (\<lambda>va vc. R)"
  apply(intro simN_wpc_repeatN_N2 allI)
  apply(auto intro: simN_wpc_repeatN_N1[THEN subsetD] simN_wpc_repeatN_I[THEN subsetD] )
  done

lemma sim_ext_try_catchRI:
  "\<lbrakk> sim_ext R ma mc Sn Se R'; sim_ext R' ha hc Sn Se R' \<rbrakk> \<Longrightarrow>
  sim_ext R (try_catchR ma ha) (try_catchR mc hc) Sn Se R'"
  apply(auto simp add: sim_ext_def project_simps
  intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindS simS_wpc_bindS
    del: subsetI)
  apply(auto intro!: simN_wpc_bindS simS_wpc_bindS)
  done
 
lemma sim_ext_catchR_N_repeatI:
  "sim_ext R ha hc Sn Se R \<Longrightarrow> sim_ext R (catchR_N ha n) (catchR_N hc n) Sn Se R"
  by(induct n, auto intro!: sim_ext_try_catchRI)

lemma sim_ext_catchR_N1:
  "sim_ext R (catchR_N ha n) hc Sn Se Sr \<Longrightarrow> sim_ext R (rec_tryR ha) hc Sn Se Sr"
  apply(auto simp add: sim_ext_def project_defs simN_wpc_def simS_wpc_def)
  apply(force simp only: repeat_rewrite dest!: subsetD split: split_if_asm)+
  done

lemma sim_ext_catchR_N2:
  "\<forall>n. sim_ext R ha (catchR_N hc n) Sn Se Sr \<Longrightarrow> sim_ext R ha (rec_tryR hc) Sn Se Sr"
  apply(auto simp add: sim_ext_def project_defs simN_wpc_def simS_wpc_def)
  apply(force simp only: repeat_rewrite dest!: subsetD split: split_if_asm)+
  done

lemma sim_ext_rec_tryRI:
  "sim_ext R ha hc Sn Se R \<Longrightarrow> sim_ext R (rec_tryR ha) (rec_tryR hc) Sn Se R"
  apply(intro sim_ext_catchR_N2 allI)
  apply(auto intro: sim_ext_catchR_N1 sim_ext_catchR_N_repeatI)
  done

lemma simN_wpc_bindS_same:
  "\<lbrakk> R \<subseteq> simS_wpc ma mc R; R \<subseteq> simN_wpc ha hc (force_same_and R) \<rbrakk> \<Longrightarrow>
  R \<subseteq> simN_wpc (\<lambda>s. ma s \<guillemotright>=S ha) (\<lambda>s. mc s \<guillemotright>=S hc) (force_same_and R)
  "
  apply(auto simp add: simS_wpc_def simN_wpc_def unfold_monad split: split_if_asm)
  apply(smt CollectE in_mono splitD)
  done

lemma bindS_projectNN: "(\<lambda>s. m s \<guillemotright>=S h)\<Down>NN = (\<lambda>s. m s \<guillemotright>=S h\<Down>NN)"
  by(auto simp add: unfold_monad project_defs)

lemma bindS_projectNE: "(\<lambda>s. m s \<guillemotright>=S h)\<Down>NE = (\<lambda>s. m s \<guillemotright>=S h\<Down>NE)"
  by(auto simp add: unfold_monad project_defs)

lemma bindS_projectN: "(\<lambda>s. m s \<guillemotright>=S h)\<Down>N = (\<lambda>s. m s \<guillemotright>=S h\<Down>N)"
  by(auto simp add: unfold_monad project_defs)

lemmas bindS_projectNs =
  bindS_projectN
  bindS_projectNN
  bindS_projectNE

lemma trans_ref:
  assumes assms1:
    "sim_norm (R :: (('sa \<times> 'sc)set)) ma mc"
    "sim_norm R ha2 hc2"(* note to self: we only ask this for h2, not h1 *)
  and assms2:
    "sim_norm R
    (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1))
    (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1))
    "
    "sim_norm R
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1))
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1))
    "
  shows
  "sim_norm R 
    (try_catchR ma (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2))) 
    (try_catchR mc (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2)))"
proof-
  have path1:
    "
    R \<subseteq> simN_wpc
    (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N ha2\<Down>NE))
    (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N hc2\<Down>NE))
    (force_same_and R)
    "  using assms1 assms2(1)
    apply(simp only: bindS_bindN)
    apply(simp only: sims_through_simN_simS simN_def bindS_projectNs)
    apply(elim conjE, intro simN_wpc_bindN_set allI)
    apply(simp_all add: force_same_and_def)
    done

  moreover have path2:
    "
      R \<subseteq> simN_wpc
      (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN)) \<guillemotright>N
                     (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NE))))
      (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN)) \<guillemotright>N
                     (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE))))
    (force_same_and R) 
      " using assms2
    apply(simp only: bindS_bindN sims_through_simN_simS simN_def bindS_projectNs)
    apply(elim conjE, intro simN_wpc_bindN_set allI)
    apply(simp_all add: sim_normE_def sim_normE_ext_def force_same_and_def)
    apply(rule simN_wpc_repeatN)
    apply(simp_all add: sim_normE_def sim_normE_ext_def)
    done

  moreover have path3:
    "
      R \<subseteq> simN_wpc
      (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN)) \<guillemotright>N
                     (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S
                          ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N ha2\<Down>NE))))
      (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN)) \<guillemotright>N
                     (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S
                          ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N hc2\<Down>NE))))
      (force_same_and R)
      " using assms    
    apply(simp only: bindS_bindN sims_through_simN_simS simN_def bindS_projectNs)
    apply(elim conjE)
    apply(intro simN_wpc_bindN_set[where S'="\<lambda>va vc. R"] allI simN_wpc_repeatN)
    apply(simp_all add: force_same_and_def)
    done

  moreover have path4:
    "
      R \<subseteq> simN_wpc
      (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN)) \<guillemotright>N
                     ha2\<Down>NN))
       
      (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S
                     ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
                     repeatN
                      (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN)) \<guillemotright>N
                     hc2\<Down>NN))
      (force_same_and R)
      " 
    using assms    
    apply(simp only: bindS_bindN sims_through_simN_simS simN_def bindS_projectNs)
    apply(elim conjE)
    apply(intro simN_wpc_bindN_set[where S'="\<lambda>va vc. R"] allI simN_wpc_repeatN)
    apply(simp_all add: force_same_and_def)
    done

  have path5: "R \<subseteq> simN_wpc (ma\<Down>NN) (mc\<Down>NN) (force_same_and R)" using assms1
    by(auto simp add: sims_through_simN_simS simN_def)

  have path6: "R \<subseteq> simN_wpc (ma\<Down>NE) (mc\<Down>NE) (force_same_and R)" using assms1
    by(auto simp add: sims_through_simN_simS simN_def)

  show ?thesis using assms2(1)
    apply(simp add: sim_norm_def sim_norm_ext_def
      sim_normN_def sim_normN_ext_def sim_normE_def sim_normE_ext_def
      try_catchR_projectNN try_catchR_projectNE rec_tryR_transformation_NN 
      rec_tryR_transformation_NE simN_wpc_choice_conc choice_assoc bindS_choice_abs
      bindS_projectNs
    )
    apply(auto dest: path1[THEN subsetD] path2[THEN subsetD] path3[THEN subsetD]
      path4[THEN subsetD] path5[THEN subsetD] path6[THEN subsetD]
      intro!: simN_wpc_choice_abs[THEN subsetD])
    done
qed

lemma trans_ref_fin_inv:
  assumes
    inv:
    "\<lbrace>I\<rbrace> ma \<lbrace>> K I\<rbrace>" "\<lbrace>I\<rbrace> ha1 \<lbrace>> K I\<rbrace>" "\<lbrace>I\<rbrace> ha2 \<lbrace>> K I\<rbrace>"
    "\<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>" "\<lbrace>J\<rbrace> hc1 \<lbrace>> K J\<rbrace>" "\<lbrace>J\<rbrace> hc1' \<lbrace>> K J\<rbrace>" "\<lbrace>J\<rbrace> hc2 \<lbrace>> K J\<rbrace>"
    and assms1:
    "sim_norm (R \<inter> I \<times> J) ma mc"
    "sim_norm (R \<inter> I \<times> J) ha2 hc2"
    and assms2: 
    "sim_norm_ext (R \<inter> I \<times> J)
        (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc1')
        (force_same_and R)
        (force_same_and R)
    "
    "sim_norm_ext (R \<inter> I \<times> J)
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
        (force_same_and R)
        (force_same_and R)
    "
    and fin:
    "sim_norm_ext (Id \<inter> J \<times> J) hc1' (try_catchR hc1 hc1') (force_same_and Id) (force_same_and Id)"
  shows
  "sim_norm (R \<inter> I \<times> J)
    (try_catchR ma (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2))) 
    (try_catchR mc (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2)))"
proof-
   
  {
    fix m1 m2  I
    assume "\<lbrace>I\<rbrace> m1 \<lbrace>S I\<rbrace>" "\<lbrace>I\<rbrace> m2 \<lbrace>> K I\<rbrace>"
    hence "\<lbrace>I\<rbrace> \<lambda>s. (m1 s \<guillemotright>=S m2) \<lbrace>> K I\<rbrace>"
      using [[simp_depth_limit=0]]
      apply(auto simp only: inv_explicit HoareS_def BoxS_def)
      apply(force simp add: project_defs unfold_monad)
      done
  } note thread_inv=this

  have include_inv:
    "sim_norm (R \<inter> I \<times> J)
        (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc1')
    "
    "sim_norm (R \<inter> I \<times> J)
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
    "
    using assms2
    apply(auto simp only: sim_norm_def force_same_and_def Int_assoc[THEN sym]
      intro!: sim_norm_ext_add_inv thread_inv K_Hoare_to_R inv)
    done

  have fin:
    "\<forall>s \<in> J. ((try_catchR hc1 hc1')\<Down>N) s \<subseteq> (hc1'\<Down>N) s" using fin
    apply(clarsimp simp add: sim_norm_ext_def force_same_and_def project_defs simN_wpc_def
      split: split_if_asm
    )
    apply(case_tac a)
    apply(auto dest!: set_mp)
    done
    
  have fin_repeatS:
    "\<forall>s \<in> J. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>N) \<subseteq> (hc1'\<Down>N) s" using fin
    apply(auto simp add: project_defs)
    apply(rule repeat_simp_inv[where h=hc1 and h'=hc1'])
    apply(auto simp add: rec_tryRS_eq project_defs inv)
    done
 
  have finNN:
    "\<forall>s \<in> J. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<subseteq> (hc1'\<Down>NN) s" using fin_repeatS
    by(auto simp add: project_defs unfold_monad)

  have finNE:
    "\<forall>s \<in> J. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE) \<subseteq> (hc1'\<Down>NE) s" using fin_repeatS
    by(auto simp add: project_defs unfold_monad)
  show ?thesis
  proof(rule trans_ref)
    {
      fix s sa sa'
      assume "s \<in> J" "sa \<in> (mc\<Down>R) s" and sa'_def: "((), sa') \<in> repeatS (hc1\<Down>R) sa \<guillemotright>=S hc1\<Down>NN"
      hence "sa \<in> J" using `\<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>`[THEN K_Hoare_to_R]
        by(auto simp only: HoareS_def BoxS_def project_defs)
      
      hence "((), sa') \<in> (hc1'\<Down>NN) sa" using sa'_def finNN
        by(auto)
    } note finNN_inv=this

    {
      fix s sa sa' e
      assume "s \<in> J" "sa \<in> (mc\<Down>R) s" and sa'_def: "(e, sa') \<in> repeatS (hc1\<Down>R) sa \<guillemotright>=S (hc1\<Down>NE)"
      hence "sa \<in> J" using `\<lbrace>J\<rbrace> mc \<lbrace>> K J\<rbrace>`[THEN K_Hoare_to_R]
        by(auto simp only: HoareS_def BoxS_def project_defs)

      hence "(e, sa') \<in> (hc1'\<Down>NE) sa" using sa'_def finNE
        by(auto)
    } note finNE_inv=this

    have N: "
      sim_normN (R \<inter> I \<times> J)
          (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1))
          (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1))
      " 
      apply(auto simp only: sims_through_simN_simS simN_def bindS_projectNs del: subsetI)
      apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs[where ma="(\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha1\<Down>NN)"]])
      apply(rule simN_wpc_monotone_comp_subset[where mc'="(\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (hc1'\<Down>NN))"])
      apply(rule
        include_inv[simplified sims_through_simN_simS bindS_projectNs simN_def, THEN conjunct1])
      apply(auto intro!: bindS_mono2' finNN_inv repeatS_skip_left)
      done

    have E: "
      sim_normE (R \<inter> I \<times> J)
          (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1))
          (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1))
      " 
      apply(auto simp only: sims_through_simN_simS simN_def bindS_projectNs del: subsetI)
      apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs[where ma="(\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha1\<Down>NE)"]])
      apply(rule simN_wpc_monotone_comp_subset[where mc'="(\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (hc1'\<Down>NE))"])
      apply(rule
        include_inv[simplified sims_through_simN_simS bindS_projectNs simN_def, THEN conjunct2])
      apply(auto intro!: bindS_mono2' finNE_inv repeatS_skip_left)
      done

    show "
      sim_norm (R \<inter> I \<times> J)
          (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1))
          (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1))
      " 
      by(auto simp add: sim_norm_decompose intro!: N E)
  next
    {
      fix s sa sa'
      assume "s \<in> J" "sa \<in> (hc2\<Down>R) s" and sa'_def: "((), sa') \<in> repeatS (hc1\<Down>R) sa \<guillemotright>=S hc1\<Down>NN"
      hence "sa \<in> J" using `\<lbrace>J\<rbrace> hc2 \<lbrace>> K J\<rbrace>`[THEN K_Hoare_to_R]
        by(auto simp only: HoareS_def BoxS_def project_defs)
      
      hence "((), sa') \<in> (hc1'\<Down>NN) sa" using sa'_def finNN
        by(auto)
    } note finNN_inv=this

    {
      fix s sa sa' e
      assume "s \<in> J" "sa \<in> (hc2\<Down>R) s" and sa'_def: "(e, sa') \<in> repeatS (hc1\<Down>R) sa \<guillemotright>=S (hc1\<Down>NE)"
      hence "sa \<in> J" using `\<lbrace>J\<rbrace> hc2 \<lbrace>> K J\<rbrace>`[THEN K_Hoare_to_R]
        by(auto simp only: HoareS_def BoxS_def project_defs)

      hence "(e, sa') \<in> (hc1'\<Down>NE) sa" using sa'_def finNE
        by(auto)
    } note finNE_inv=this

    have N: "sim_normN (R \<inter> I \<times> J)
        (\<lambda>s. doS {
                       s \<leftarrow> (ha2\<Down>R) s;
                       repeatS (ha1\<Down>R) s \<guillemotright>=S ha1
                     })
        (\<lambda>s. doS {
               s \<leftarrow> (hc2\<Down>R) s;
               repeatS (hc1\<Down>R) s \<guillemotright>=S hc1
             })
      "
      apply(auto simp only: sims_through_simN_simS simN_def bindS_projectNs del: subsetI)
      apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs[where ma="(\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1\<Down>NN)"]])
      apply(rule simN_wpc_monotone_comp_subset[where mc'="(\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (hc1'\<Down>NN))"])
      apply(rule
        include_inv[simplified sims_through_simN_simS bindS_projectNs simN_def, THEN conjunct1])
      apply(auto intro!: bindS_mono2' finNN_inv repeatS_skip_left)
      done
 
    have E: "sim_normE (R \<inter> I \<times> J)
        (\<lambda>s. doS {
                       s \<leftarrow> (ha2\<Down>R) s;
                       repeatS (ha1\<Down>R) s \<guillemotright>=S ha1
                     })
        (\<lambda>s. doS {
               s \<leftarrow> (hc2\<Down>R) s;
               repeatS (hc1\<Down>R) s \<guillemotright>=S hc1
             })
      "
      apply(auto simp only: sims_through_simN_simS simN_def bindS_projectNs del: subsetI)
      apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs[where ma="(\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1\<Down>NE)"]])
      apply(rule simN_wpc_monotone_comp_subset[where mc'="(\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (hc1'\<Down>NE))"])
      apply(rule
        include_inv[simplified sims_through_simN_simS bindS_projectNs simN_def, THEN conjunct2])
      apply(auto intro!: bindS_mono2' finNE_inv repeatS_skip_left)
      done

    show "sim_norm (R \<inter> I \<times> J)
        (\<lambda>s. doS {
                       s \<leftarrow> (ha2\<Down>R) s;
                       repeatS (ha1\<Down>R) s \<guillemotright>=S ha1
                     })
        (\<lambda>s. doS {
               s \<leftarrow> (hc2\<Down>R) s;
               repeatS (hc1\<Down>R) s \<guillemotright>=S hc1
             })
      "
      by(auto simp add: sim_norm_decompose intro!: N E)
  qed (auto intro!: assms1)
qed

lemma trans_ref_fin:
  assumes assms1:
    "sim_norm R ma mc" "sim_norm R ha2 hc2"
  and assms2: 
    "sim_norm R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha1) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc1')"
    "sim_norm R (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1) (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')"
  and fin:
    "sim_norm Id hc1' (try_catchR hc1 hc1')"
  shows
  "sim_norm R 
    (try_catchR ma (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2))) 
    (try_catchR mc (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2)))"
  apply(subgoal_tac "R = (R \<inter> UNIV \<times> UNIV)")
  apply(erule ssubst)
  apply(rule trans_ref_fin_inv)
  apply(simp_all only: UNIV_Times_UNIV Int_UNIV_right true_post
    sim_norm_def[simplified atomize_eq, THEN sym])
  apply(auto intro!: assms1 assms2 fin)
  done

lemma sim_norm_ext_imp_sim_normN_ext:
  "sim_norm_ext R ma mc Sn Se \<Longrightarrow> sim_normN_ext R ma mc Sn"
  by(auto simp add: sim_norm_ext_def sim_normN_ext_def)

lemma sim_norm_ext_imp_sim_normE_ext:
  "sim_norm_ext R ma mc Sn Se \<Longrightarrow> sim_normE_ext R ma mc Se"
  by(auto simp add: sim_norm_ext_def sim_normE_ext_def)

text {*
  Transformed refinement for the case where the original computation
  and the handler are the same
*}

lemma trans_fin_catchR_ext:
  assumes
    assms1:
    "sim_norm_ext R ha1 hc1' Sn' Se"
    "\<forall>va vc. sim_norm_ext (Sn' va vc) ha2 hc2 Sn Se"
    and assms2: 
    "\<forall>va vc. sim_norm_ext (Sn' va vc)
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
        Sn' Se 
    "
    and fin:
    "sim_norm Id hc1' (try_catchR hc1 hc1')"
  shows
    "sim_norm_ext R 
    (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2)) 
    (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2))
    Sn Se"
proof-
  have fin_subs:
    "\<forall>s. ((try_catchR hc1 hc1')\<Down>N) s \<subseteq> (hc1'\<Down>N) s" using fin
    apply(clarsimp simp add: sim_norm_def sim_norm_ext_def force_same_and_def project_defs simN_wpc_def
      split: split_if_asm
    )
    apply(case_tac a)
    apply(auto dest!: set_mp)
    done
    
  have fin_repeatS:
    "\<forall>s. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>N) \<subseteq> (hc1'\<Down>N) s" using fin_subs
    apply(auto simp add: project_defs)
    apply(rule repeat_simp[where h=hc1 and h'=hc1'])
    apply(auto simp add: rec_tryRS_eq project_defs)
    done
 
  have finNN:
    "\<forall>s. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<subseteq> (hc1'\<Down>NN) s" using fin_repeatS
    by(auto simp add: project_defs unfold_monad)

  have finNE:
    "\<forall>s. (repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE) \<subseteq> (hc1'\<Down>NE) s" using fin_repeatS
    by(auto simp add: project_defs unfold_monad)
  
  hence sim_hc1:
    "sim_norm_ext R ha1 hc1 Sn' Se"
  proof-
    have "\<forall>s. (hc1\<Down>N) s \<subseteq> (hc1'\<Down>N) s" using fin_subs
      by(auto simp add: project_defs try_catchR_some_decompose Collect_subset)
    thus ?thesis 
      by(auto intro!: sim_norm_ext_mono_conc[OF assms1(1)])
  qed
  
  have repeatS_ha1_ref_NN:
    "R \<subseteq> simN_wpc 
        (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN)
        (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) 
        Sn'"
    using assms1(1)[THEN sim_norm_ext_imp_sim_normN_ext]
    apply(auto simp add: sim_normN_ext_def del: subsetI)
    apply(rule simN_wpc_monotone_comp_subset)
    apply(erule order_trans[OF _ simN_wpc_monotone_comp_abs])
    apply(auto simp add: force_same_and_same intro: repeatS_skip_left del: subsetI)
    apply(rule finNN[THEN allE])
    apply(assumption)
    done
 
  have ha2_repeatS_h1_ref_NN:
    "\<forall>va vc. Sn' va vc \<subseteq> simN_wpc
         (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      })
         (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
                      })
         Sn'"
  proof-
    have sub_conc:
      "\<forall>s. (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
      }) s
      \<subseteq> (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1'\<Down>NN) s"
      by(auto intro!: bindS_mono2 repeatS_skip_left bindN_mono finNN
        del: subsetI)
    
    have sub_abs:
      "\<forall>s. (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1\<Down>NN) s
      \<subseteq> (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      }) s"
      by(auto intro!: bindS_mono2 repeatS_skip_left bindN_mono del: subsetI)
    
    show ?thesis using assms2
      apply(simp add: force_same_and_def sim_norm_ext_def bindS_projectNs)
      apply(elim conjE)
      apply(rule simN_wpc_monotone_comp_subset[OF _ sub_conc[THEN All_imp_ball]])
      apply(erule subset_trans[OF _ simN_wpc_monotone_comp_abs[OF sub_abs]])
      done
  qed
  
  def R' \<equiv> "Sn' () ()"
  have Sn'_def': "Sn' = (\<lambda>va vc. R')"
    by(simp add: R'_def)
   
  have repeatN_ha2_repeatS_h1_ref_NN:
    "\<forall>va vc. Sn' va vc \<subseteq> simN_wpc
         (repeatN (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      }))
         (repeatN (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
                      }))
         Sn'" using ha2_repeatS_h1_ref_NN
    apply(auto simp add: force_same_and_def del: subsetI)
    apply(fold R'_def)
    apply(simp only: Sn'_def')
    apply(auto intro!: simN_wpc_repeatN del: subsetI)
    done

  have path1:
    "R \<subseteq> simN_wpc
         ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      }) \<guillemotright>N
          ha2\<Down>NN)
         ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
                      }) \<guillemotright>N
          hc2\<Down>NN)
         Sn" using assms1(2) repeatN_ha2_repeatS_h1_ref_NN
    apply(simp only: bindS_bindN sim_norm_ext_def bindS_projectNs)
    apply(auto intro!: simN_wpc_bindN_set repeatS_ha1_ref_NN
      simp add: force_same_and_same
      del: subsetI)
    done
 
  have path2:
    "R \<subseteq> simN_wpc
          (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NE)
          (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE) Se" 
    using assms1(1)[THEN sim_norm_ext_imp_sim_normE_ext]
    apply(auto simp add: sim_normE_def sim_normE_ext_def
      intro!: del: subsetI)
    apply(rule simN_wpc_monotone_comp_subset)
    apply(erule order_trans[OF _ simN_wpc_monotone_comp_abs])
    apply(auto intro: repeatS_skip_left del: subsetI)
    apply(rule finNE[THEN allE])
    apply(assumption)
    done
  
  have path3:
    "R \<subseteq> simN_wpc
          ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N ha2\<Down>NE)
         ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N hc2\<Down>NE) Se"
    using assms1(2)
    apply(simp only: sim_norm_ext_def bindS_projectNs)
    apply(auto intro!: simN_wpc_bindN_set repeatS_ha1_ref_NN repeatN_ha2_repeatS_h1_ref_NN 
      simp add: force_same_and_same
      del: subsetI)
    done
 
  have path4:
    "R \<subseteq> simN_wpc
         ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      }) \<guillemotright>N
          (\<lambda>s. doS {
                s \<leftarrow> (ha2\<Down>R) s;
                repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NE
              }))
         ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
                      }) \<guillemotright>N
          (\<lambda>s. doS {
                s \<leftarrow> (hc2\<Down>R) s;
                repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE
              }))
         Se"
  proof-
    have
      "\<forall>va vc. Sn' va vc \<subseteq> simN_wpc (\<lambda>s. doS {
                       s \<leftarrow> (ha2\<Down>R) s;
                       repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NE
                     })
         (\<lambda>s. doS {
               s \<leftarrow> (hc2\<Down>R) s;
               repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE
             })
         Se
      "
    proof-
      have sub_conc:
        "\<forall>s. (\<lambda>s. doS {
            s \<leftarrow> (hc2\<Down>R) s;
            repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NE
        }) s
        \<subseteq> (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1'\<Down>NE) s"
        by(auto intro!: bindS_mono2 repeatS_skip_left bindN_mono finNE
          del: subsetI)
      
      have sub_abs:
        "\<forall>s. (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1\<Down>NE) s
        \<subseteq> (\<lambda>s. doS {
            s \<leftarrow> (ha2\<Down>R) s;
            repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NE
        }) s"
        by(auto intro!: bindS_mono2 repeatS_skip_left bindN_mono del: subsetI)
      
      show ?thesis using assms2
        apply(simp only: sim_norm_ext_def bindS_projectNs Int_subset_iff all_conj)
        apply(erule conjE)
        apply(intro allI)
        apply(rule simN_wpc_monotone_comp_subset[OF _ sub_conc[THEN All_imp_ball]])
        apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs[OF sub_abs]])
        apply(auto)
        done
    qed
  
    thus ?thesis using repeatN_ha2_repeatS_h1_ref_NN
      by(auto intro!: simN_wpc_bindN_set[where S'=Sn'] 
        repeatS_ha1_ref_NN 
        simp add: force_same_and_same
        del: subsetI)
  qed
  
  have path5:
    "R \<subseteq> simN_wpc
         ((\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (ha2\<Down>R) s;
                        repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN
                      }) \<guillemotright>N
          (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (ha1\<Down>R) s \<guillemotright>=S ha1\<Down>NN) \<guillemotright>N ha2\<Down>NE))
         ((\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N
          repeatN (\<lambda>s. doS {
                        s \<leftarrow> (hc2\<Down>R) s;
                        repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN
                      }) \<guillemotright>N
          (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (hc1\<Down>R) s \<guillemotright>=S hc1\<Down>NN) \<guillemotright>N hc2\<Down>NE))
         Se"
    using assms1(2) repeatN_ha2_repeatS_h1_ref_NN ha2_repeatS_h1_ref_NN
    apply(simp only: sim_norm_ext_def bindS_bindN bindS_projectNs)
    apply(rule simN_wpc_bindN_set)
    apply(rule simN_wpc_bindN_set)
    apply(rule repeatS_ha1_ref_NN)
    apply(simp add: force_same_and_same)
    apply(intro allI)
    apply(rule simN_wpc_bindN_set)
    apply(auto simp add: force_same_and_same)
    done
 
  show ?thesis
    apply(simp add: sim_norm_def sim_norm_ext_def
      sim_normN_def sim_normN_ext_def sim_normE_def sim_normE_ext_def
      try_catchR_projectNN try_catchR_projectNE 
      simN_wpc_choice_lockstep
      rec_tryR_transformation_NN rec_tryR_transformation_NE
      bindS_projectNs
    )
    apply(auto intro!: path1 path2 path3 path4 path5 simN_wpc_choice_lockstep del: subsetI)
    done

qed

lemma trans_ref_fin_catchR:
  assumes
    assms1:
    "sim_norm R ha1 hc1'"
    "sim_norm R ha2 hc2"
    and assms2: 
    "sim_norm R
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
    "
    and fin:
    "sim_norm Id hc1' (try_catchR hc1 hc1')"
  shows
  "sim_norm R 
    (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2)) 
    (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2))
  "
proof(unfold sim_norm_def, rule trans_fin_catchR_ext)
  show "sim_norm_ext R ha1 hc1' (force_same_and R) (force_same_and R)" 
    using assms1[unfolded sim_norm_def]
    by - assumption
next
  show
    "\<forall>va vc. sim_norm_ext (force_same_and R va vc) ha2 hc2 
        (force_same_and R) (force_same_and R)"
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel assms1[unfolded sim_norm_def])
    done
next
  show
    "\<forall>va vc. sim_norm_ext (force_same_and R va vc) 
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1) (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1') 
        (force_same_and R) (force_same_and R)"
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel assms2[unfolded sim_norm_def])
    done
qed (simp add: fin)
 
lemma trans_ref_fin_same_ext:
  assumes
    assms1:
    "sim_norm_ext R ha1 hc1' Sn' Se"
    "\<forall>va vc. sim_norm_ext (Sn' va vc) ha2 hc2 Sn Se"
    and assms2: 
    "\<forall>va vc. sim_norm_ext (Sn' va vc)
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
        Sn' Se 
    "
    and fin:
    "sim_norm Id hc1' (try_catchR hc1 hc1')"
  shows
  "sim_norm_ext R 
    (try_catchR (ha1 \<guillemotright> ha2) (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2))) 
    (try_catchR (hc1 \<guillemotright> hc2) (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2)))
    Sn Se
  "
proof-
  have repeat_ref:
    "sim_norm_ext R 
    (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2)) 
    (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2))
    Sn Se
    "
    by(rule trans_fin_catchR_ext[OF assms1 assms2 fin])

  show ?thesis
    apply(rule sim_norm_ext_mono_conc)
    apply(rule sim_norm_ext_mono_abs)
    apply(rule repeat_ref)
    apply(auto simp only: try_rec_tryR_same)
    done
qed

lemma trans_ref_fin_same:
  assumes
    assms1:
    "sim_norm R ha1 hc1'"
    "sim_norm R ha2 hc2"
    and assms2: 
    "sim_norm R
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1)
        (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1')
    "
    and fin:
    "sim_norm Id hc1' (try_catchR hc1 hc1')"
  shows
  "sim_norm R 
    (try_catchR (ha1 \<guillemotright> ha2) (rec_tryR ((ha1 :: (unit, 'e, 'sa)re_monad) \<guillemotright> ha2))) 
    (try_catchR (hc1 \<guillemotright> hc2) (rec_tryR ((hc1 :: (unit, 'e, 'sc)re_monad) \<guillemotright> hc2)))
  "
proof(unfold sim_norm_def, rule trans_ref_fin_same_ext)
  show "sim_norm_ext R ha1 hc1' (force_same_and R) (force_same_and R)" 
    using assms1[unfolded sim_norm_def]
    by - assumption
next
  show
    "\<forall>va vc. sim_norm_ext (force_same_and R va vc) ha2 hc2 
        (force_same_and R) (force_same_and R)"
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel assms1[unfolded sim_norm_def])
    done
next
  show
    "\<forall>va vc. sim_norm_ext (force_same_and R va vc) 
        (\<lambda>s. (ha2\<Down>R) s \<guillemotright>=S ha1) (\<lambda>s. (hc2\<Down>R) s \<guillemotright>=S hc1') 
        (force_same_and R) (force_same_and R)"
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel assms2[unfolded sim_norm_def])
    done
qed (simp add: fin)
 
(*******************************************************************************)
subsection {* Miscellaneous lemams *}
(*******************************************************************************)

lemma sim_imp_simS:
  "sim R ma mc \<Longrightarrow> R \<subseteq> simS_wpc (ma\<Down>R) (mc\<Down>R) R"
  by(auto simp add: sim_def sim_ext_def del: subsetI)

lemma sims_empty_conc:
  "simN R ma empty"
  "simS R ma' empty"
  "sim_normN R ma empty"
  "sim_normE R ma empty"
  "sim_norm R ma empty"
  "sim R ma empty"
  by(auto simp add: sims_through_simN_simS simN_def simS_def project_simps
    simN_wpc_empty_conc simS_wpc_empty_conc)

lemma sim_norm_ext_choice_lockstep:
  "
  \<lbrakk>sim_norm_ext R ma1 mc1 Sn Se; sim_norm_ext R ma2 mc2 Sn Se\<rbrakk> \<Longrightarrow>
    sim_norm_ext R (ma1 [+] ma2) (mc1 [+] mc2) Sn Se
  "
  by(auto simp add: sim_norm_ext_def project_simps intro!: simN_wpc_choice_lockstep
    del: subsetI)

lemmas sims_all_intros =
  sims_return
  sims_throw
  sims_bindI
  sims_try_catchI
  sims_try_catchRI
  sims_choice_lockstep
  sims_chooseN
  sim_mapM'
  sims_empty_conc
  simS_choice_lockstep
  simN_choice_lockstep

lemma sim_ext_decompose:
  "sim_ext R ma mc Sn Se Sr = (
    sim_normN_ext R ma mc Sn
    \<and> sim_normE_ext R ma mc Se
    \<and> R \<subseteq> simS_wpc (ma\<Down>R) (mc\<Down>R) Sr
  )"
  by(clarsimp simp add: sim_ext_def sim_normN_ext_def sim_normE_ext_def)

lemma simN_wpc_compose:
  "\<lbrakk>R \<subseteq> simN_wpc ma mb (force_same_and R); R' \<subseteq> simN_wpc mb mc (force_same_and R')\<rbrakk>
    \<Longrightarrow> R O R' \<subseteq> simN_wpc ma mc (force_same_and (R O R'))"
  apply(auto simp add: simN_wpc_def force_same_and_def)
  by (smt CollectD equals0D relcomp.intros set_rev_mp splitD)

lemma simS_wpc_compose:
  "\<lbrakk>R \<subseteq> simS_wpc ma mb R; R' \<subseteq> simS_wpc mb mc R'\<rbrakk>
    \<Longrightarrow> R O R' \<subseteq> simS_wpc ma mc (R O R')"
  apply(auto simp add: simS_wpc_def)
  by (smt in_mono mem_Collect_eq prod.simps(2) relcomp.intros)

lemma simN_compose:
  "\<lbrakk> simN R ma mb; simN R' mb mc \<rbrakk> \<Longrightarrow> simN (R O R') ma mc"
  by(unfold simN_def, rule simN_wpc_compose, assumption+)

lemma simS_compose:
  "\<lbrakk> simS R ma mb; simS R' mb mc \<rbrakk> \<Longrightarrow> simS (R O R') ma mc"
  by(unfold simS_def, rule simS_wpc_compose, assumption+)

lemma sim_norm_compose:
  "\<lbrakk> sim_norm R ma mb; sim_norm R' mb mc \<rbrakk> \<Longrightarrow> sim_norm (R O R') ma mc"
  by(auto simp only: sim_norm_through_simN intro: simN_compose)

lemma sim_compose:
  "\<lbrakk> sim R ma mb; sim R' mb mc \<rbrakk> \<Longrightarrow> sim (R O R') ma mc"
  by(auto simp only: sim_through_simN_simS intro: simN_compose simS_compose)

end
