theory Repeated_Restarts
imports EvaluationLogicR Projections HoareLogic
begin

lemma try_catchR_assoc:
  "try_catchR (try_catchR m h) h' = try_catchR m (try_catchR h h')"
  by(force simp add: try_catchR_def liftR2_def split: option.split_asm intro!: ext)

inductive_set rec_tryR ::
  "('a, 's)r_monad \<Rightarrow> ('a, 's)r_monad"
  for h :: "('a, 's)r_monad"
  and s :: 's
  where
  exec_some: "r \<in> h s \<Longrightarrow> r \<in> rec_tryR h s"
| exec_none: "\<lbrakk> (None, s') \<in> rec_tryR h s; r \<in> h s' \<rbrakk> \<Longrightarrow> 
    r \<in> rec_tryR h s"

fun catchR_N :: "('a, 's)r_monad \<Rightarrow> nat \<Rightarrow> ('a, 's)r_monad" where
  "catchR_N h 0 = h"
  | "catchR_N h (Suc n) = try_catchR h (catchR_N h n)"

lemma catchR_N_alter: "catchR_N h (Suc n) = try_catchR (catchR_N h n) h"
  by(induct n, auto simp add: try_catchR_assoc[THEN sym])

lemma repeat_bounded: "(r, s') \<in> rec_tryR h s \<Longrightarrow> \<exists>n. (r, s') \<in> catchR_N h n s"
proof(induct rule: rec_tryR.induct)
  case (exec_some r)
  thus ?case
    by(rule_tac x=0 in exI, simp)
next
  case (exec_none s' r_full)
  from this obtain r s'' where rs'': "r_full = (r, s'')" by(case_tac r_full)
  from exec_none(2) obtain n where "(None, s') \<in> catchR_N h n s" by auto
  thus ?case using exec_none(3) rs''
    apply(rule_tac x="Suc n" in exI)
    apply(case_tac r)
    apply(auto simp only: catchR_N_alter 
      intro!: try_catchR_none_decompose[THEN iffD2] try_catchR_some_decompose[THEN iffD2])
    done
qed

lemma bounded_into_repeat:
  "(r, s') \<in> catchR_N h n s \<Longrightarrow> (r, s') \<in> rec_tryR h s"
  apply(induct n arbitrary: s' r)
  apply(rule rec_tryR.exec_some, simp)
  apply(simp only: catchR_N_alter)
  apply(case_tac r, simp_all add: try_catchR_none_decompose try_catchR_some_decompose)
  apply(auto intro: intro: rec_tryR.exec_none)
  done

(*
definition inv_for_comp :: "'s set \<Rightarrow> ('a, 's)r_monad \<Rightarrow> bool" where
  "inv_for_comp I m \<equiv> \<forall>r s'. \<forall>s \<in> I. (r, s') \<in> m s \<longrightarrow> s' \<in> I" 
    *)

(*
lemma inv_for_comp_UNIV: "inv_for_comp UNIV m"
  by(auto simp add: inv_for_comp_def)
    *)

lemma inv_explicit:
  "(\<lbrace>I\<rbrace> m \<lbrace>> K I\<rbrace>) = (\<forall>r s'. \<forall>s \<in> I. (r, s') \<in> m s \<longrightarrow> s' \<in> I)"
  by(auto simp add: Hoare_def Box_def K_def)

lemma inv_catchR_N:
  "\<lbrace>I\<rbrace> h \<lbrace>> K I\<rbrace> \<Longrightarrow> \<lbrace>I\<rbrace> (catchR_N h n) \<lbrace>> K I\<rbrace>"
proof(induct n)
  case 0
  thus ?case
    by(simp)
next
  case (Suc n)
  thus ?case (* this proof should be much shorter, but I'm unable to avoid looping *)
    using [[simp_depth_limit=0]]
  proof(simp only: inv_explicit)
    assume IH: "\<forall>r s'. \<forall>s\<in>I. (r, s') \<in> h s \<longrightarrow> s' \<in> I \<Longrightarrow>
                \<forall>r s'. \<forall>s\<in>I. (r, s') \<in> catchR_N h n s \<longrightarrow> s' \<in> I"
      and Inv: "\<forall>r s'. \<forall>s\<in>I. (r, s') \<in> h s \<longrightarrow> s' \<in> I"
    show "\<forall>r s'. \<forall>s\<in>I. (r, s') \<in> catchR_N h (Suc n) s \<longrightarrow> s' \<in> I" 
    proof(intro allI ballI impI)
      fix s r s'
      assume asm: "s \<in> I" "(r, s') \<in> catchR_N h (Suc n) s"
      show "s' \<in> I" 
      proof(cases r)
        case None
        hence "(None, s') \<in> try_catchR h (catchR_N h n) s" using asm
          by simp
        note decomp=this[THEN try_catchR_none_decompose[THEN iffD1]]
        thus ?thesis using asm IH[OF Inv] Inv
          by(blast)
      next
        case (Some a)
        hence "(Some a, s') \<in> try_catchR h (catchR_N h n) s" using asm
          by simp
        note decomp=this[THEN try_catchR_some_decompose[THEN iffD1]]
        thus ?thesis using asm IH[OF Inv] Inv
          by(blast)
      qed
    qed
  qed
qed

lemma inv_catchR_N':
  "\<lbrace>I\<rbrace> h \<lbrace>> K I\<rbrace> \<Longrightarrow> \<forall>n. \<lbrace>I\<rbrace> catchR_N h n \<lbrace>> K I\<rbrace>"
  by(blast dest!: inv_catchR_N)

lemma bounded_simp_inv: "
  \<lbrakk> \<lbrace>I\<rbrace> h \<lbrace>> K I\<rbrace>; 
    \<forall>s \<in> I. \<forall>r s'. (Some r, s') \<in> try_catchR h h' s \<longrightarrow> (Some r, s') \<in> h' s;
    s \<in> I; 
    (Some r, s') \<in> catchR_N h n s
  \<rbrakk> 
  \<Longrightarrow> 
  (Some r, s') \<in> h' s" 
proof(induct n arbitrary: r s s')
  case 0
  also have reduction_base: "\<forall>s \<in> I. \<forall>r s'. (Some r, s') \<in> h s \<longrightarrow> (Some r, s') \<in> h' s" using 0(2)
    by(simp add: try_catchR_some_decompose[THEN iffD2])
  ultimately show ?case
    by(auto)
next
  case (Suc n)
  have reduction_base: "\<forall>s \<in> I. \<forall>r s'. (Some r, s') \<in> h s \<longrightarrow> (Some r, s') \<in> h' s" using Suc(3)
    by(simp add: try_catchR_some_decompose[THEN iffD2])
  note IH=Suc(1)
  note Inv=Suc(2)
  note reduction=Suc(3)
  note s_in_I=Suc(4)
  note s'_in_try_catch=Suc(5)
  hence "(Some r, s') \<in> try_catchR h (catchR_N h n) s"
    by simp
  note decompose=this[THEN try_catchR_some_decompose[THEN iffD1]]
  hence decompose2: "(Some r, s') \<in> h s \<or> 
    (\<exists>s''. (None, s'') \<in> h s \<and> (Some r, s') \<in> catchR_N h n s'' \<and> s'' \<in> I)" using Inv s_in_I
    by(auto simp add: inv_explicit)
  show ?case
  proof(cases "(Some r, s') \<in> h s")
    case True
    thus ?thesis using s_in_I reduction_base
      by auto
  next
    case False
    from this obtain s'' where "(None, s'') \<in> h s \<and> (Some r, s') \<in> catchR_N h n s'' \<and> s'' \<in> I"
      using decompose2
      by auto
    hence "(Some r, s') \<in> try_catchR h h' s" using Suc
      by(auto intro!: try_catchR_some_decompose[THEN iffD2])
    thus ?thesis using reduction s_in_I
      by(auto)
  qed
qed

lemma repeat_simp_inv: "
  \<lbrakk> \<lbrace>I\<rbrace> h \<lbrace>> K I\<rbrace>; 
    \<forall>s \<in> I. \<forall>r s'. (Some r, s') \<in> try_catchR h h' s \<longrightarrow> (Some r, s') \<in> h' s;
    s \<in> I; 
    (Some r, s') \<in> rec_tryR h s
  \<rbrakk> 
  \<Longrightarrow> 
  (Some r, s') \<in> h' s" 
  apply(drule repeat_bounded)
  apply(clarsimp)
  apply(drule bounded_simp_inv)
  apply(auto)
  done

lemma bounded_simp:
 "\<lbrakk> 
    \<forall>s r s'. (Some r, s') \<in> try_catchR h h' s \<longrightarrow> (Some r, s') \<in> h' s;
    (Some r, s') \<in> catchR_N h n s \<rbrakk>
  \<Longrightarrow> 
  (Some r, s') \<in> h' s"
  apply(rule bounded_simp_inv[where I=UNIV])
  apply(auto simp add: true_post)
  done

lemma repeat_simp: "\<lbrakk> \<forall>s r s'. (Some r, s') \<in> try_catchR h h' s \<longrightarrow> (Some r, s') \<in> h' s;
  (Some r, s') \<in> rec_tryR h s \<rbrakk> \<Longrightarrow> (Some r, s') \<in> h' s"
  by(force dest: repeat_bounded bounded_simp)

end

