theory Requirements
  imports 
  "abs4/Abs4_Specification" 
  "abs4/Abs4_Invariant_Lemmas" 
  "abs4/Abs4_Redundancy"
  "abs5/Abs5Defs"
  "abs6/Abs6_Specification" 
  "abs6/Abs6_Lemmas"
  Paper_Notation
  Full_Ref
begin

text {* Here we formalize and prove the requirements on our system *}

(********************************************************************)
subsubsection {* R1: atomic updates *}
(********************************************************************)

thm 
  obs_mem_in_def

text {* We split the atomic update in two parts, one where the
  original view is @{term "\<bottom>"} and one where it is @{term "Some v"} *}

theorem abs6_transaction_None:
  "\<lbrace>(obs abs6_spec)\<^sup>-\<^sup>1[obs_mem_in {\<bottom>}]\<rbrace> 
    <tryR> Abs6_UpdateDefs.update upd_f <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> \<lambda>v. (obs abs6_spec)\<^sup>-\<^sup>1[obs_mem_in {\<bottom>}]\<rbrace>\<^sub>N"
  by(auto simp add: abs6_spec_def obs_ts intro: Abs6_Lemmas.transaction_None[simplified K_def])

text {* An example of how to transfer a property to the concrete
  model. For brevity, we omit this for the other properties, but the
  process is identical. Notice that @{term "obs abs6_spec = abs6_obs"}
  *}

theorem
  "\<lbrace>(obs pmm_spec)\<^sup>-\<^sup>1[full_pi_composition\<^sup>-\<^sup>1[obs_mem_in {\<bottom>}]] \<inter> reach pmm_spec\<rbrace> 
    <tryR> UpdateDefs.update upd_f <catchR> rec_tryR StartupDefs.startup 
  \<lbrace>> \<lambda>v. ((obs pmm_spec)\<^sup>-\<^sup>1[full_pi_composition\<^sup>-\<^sup>1[obs_mem_in {\<bottom>}]])\<rbrace>\<^sub>N"
  by(auto intro!: ref_HoareN_preservation abs6_transaction_None concrete_refines_abs6
    update_full_ref_N concrete_implements_abs6)

text {* We further subdivide the requirement into parts for normal and
  exceptional termination. We define two subtypes of the
  non-restarting Hoare triple, @{term "\<lbrace>P\<rbrace> m \<lbrace>> Q\<rbrace>\<^sub>N"}: the triple for
  normal termination @{term "\<lbrace>P\<rbrace> m \<lbrace>> Q \<rbrace>\<^sub>N\<^isub>N"} and the one for
  exceptional termination, @{term "\<lbrace>P\<rbrace> m \<lbrace>> Q \<rbrace>\<^sub>N\<^isub>E"} *}

text {* The normal termination triple takes advantage of the fact that
  @{term startup} never terminates normally. It either throws a @{term
  StartedExc} if it does not detect any permanent failures, or one of
  the error exceptions (@{term "Degraded EU"}, @{term "FailMode"})
  otherwise *}

theorem transaction_Some_normal:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some mem}]\<rbrace> 
    <tryR> Abs6_UpdateDefs.update upd_f <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> K abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some (mem \<lhd> upd_f)}]\<rbrace>\<^sub>N\<^isub>N"
  by(clarsimp simp add: fun_map_override_def intro!: Abs6_Lemmas.transaction_Some_normal)

theorem transaction_Some_exceptional:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some mem}]\<rbrace> 
    <tryR> Abs6_UpdateDefs.update upd_f <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> K abs6_obs\<^sup>-\<^sup>1[obs_mem_in {\<bottom>, Some mem, Some (mem \<lhd> upd_f)}]\<rbrace>\<^sub>N\<^isub>E"
  by(clarsimp simp add: fun_map_override_def intro!: Abs6_Lemmas.transaction_Some_exceptional)


(********************************************************************)
subsubsection {* R2: correctness of fetch *}
(********************************************************************)

definition abs6_obs_mode_is where
  "abs6_obs_mode_is mode \<equiv> {(sv, sp). sv = mode}"

text {* @{term fetch} never returns a regular value if started
  in @{term ACFail} mode *}

theorem fetch_no_norm_in_Fail:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[abs6_obs_mode_is ACFail]\<rbrace> 
    <tryR> Abs6_FetchDefs.fetch ind <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> K {}\<rbrace>\<^sub>N\<^isub>N"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def abs6_obs_mode_is_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

text {* Moreover, the only exception it throws in that case is @{term FailMode} *}

theorem fetch_no_exc_val_in_Fail:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[abs6_obs_mode_is ACFail]\<rbrace> 
    fetch ind
  \<lbrace>> \<lambda>e. if e = FailMode then UNIV else {}\<rbrace>\<^sub>N\<^isub>E"
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def abs6_obs_mode_is_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

text {* It yields the value from the current view if it terminates normally.
  Again, we take advantage of the fact that @{term startup} never terminates
  normally *}

theorem fetch_obs_correctness_NN:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some mem}]\<rbrace> 
    <tryR> Abs6_FetchDefs.fetch ind <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> \<lambda>cnt. if cnt = mem ind then UNIV else {}\<rbrace>\<^sub>N\<^isub>N"
  apply(rule conseqNN_rule[OF _ fetch_correctness_NN[where mem=mem]])
  apply(auto simp add: obs_mem_in_def memory_eq_def split: split_if_asm)
  done

text {* We cannot separate the exceptional termination of @{term startup}
  and @{term fetch} so cleanly, so we first look at @{term fetch} in 
  isolation. If it terminates with a @{term Degraded} exception, it
  carries the correct value from the current view.
*}

theorem fetch_obs_correctness_no_restart_NE:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some mem}]\<rbrace> 
    Abs6_FetchDefs.fetch ind 
  \<lbrace>> \<lambda>e. if e = FailMode \<or> (\<exists>cnt. e = Degraded (EC cnt) \<and> cnt = mem ind)
    then UNIV else {}\<rbrace>\<^sub>N\<^isub>E"
  apply(rule conseqNE_rule[OF _ fetch_correctness_no_restart_NE[where mem=mem]])
  apply(auto simp add: obs_mem_in_def memory_eq_def split: split_if_asm)
  done

text {* @{term startup} adds a possibility of terminating with 
  a @{term "Degraded EU"} exception, which carries no value, but
  simply reports the @{term ACDegraded} mode *}

theorem fetch_obs_correctness_NE:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {Some mem}]\<rbrace> 
    <tryR> Abs6_FetchDefs.fetch ind <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> \<lambda>e. if e = FailMode \<or> e = StartedExc \<or> e = Degraded EU \<or> 
                (\<exists>cnt. e = Degraded (EC cnt) \<and> cnt = mem ind)
        then UNIV 
        else {}
  \<rbrace>\<^sub>N\<^isub>E"
  apply(rule conseqNE_rule[OF _ fetch_correctness_NE[where mem=mem]])
  apply(auto simp add: obs_mem_in_def memory_eq_def split: split_if_asm)
  done

(********************************************************************)
subsubsection {* R3: no change to the view *}
(********************************************************************)

text {* This is an "internal" Hoare triple (i.e. not based on
observations), which we use to prove R3 *}

lemma fetch_no_mem_upd: 
  "\<lbrace>memory_eq mem\<rbrace> 
    <tryR> Abs6_FetchDefs.fetch ind <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> K (memory_eq mem)\<rbrace>\<^sub>N"
  by(auto simp add: decompose_HoareN intro: fetch_no_mem_upd_NN
    fetch_no_mem_upd_NE)

text {* And this is the resulting observation-based one. The triple
  holds for both normally and exceptionally terminating results *}

theorem fetch_no_obs_mem_upd:
  "\<lbrace>abs6_obs\<^sup>-\<^sup>1[obs_mem_in {mem}]\<rbrace> 
    <tryR> Abs6_FetchDefs.fetch ind <catchR> Abs6_StartupDefs.startup 
  \<lbrace>> K abs6_obs\<^sup>-\<^sup>1[obs_mem_in {mem, \<bottom>}]\<rbrace>\<^sub>N"
  apply(case_tac mem, simp_all)
  apply(rule conseqN_rule[OF _ fetch_call_no_revival[THEN Hoare_imp_HoareN]])
  apply(auto simp add: obs_mem_in_def no_failed_def K_def split: split_if_asm)[2]
  apply(clarsimp)
  apply(rule_tac mem1=a in conseqN_rule[OF _ fetch_no_mem_upd])
  apply(auto simp add: obs_mem_in_def K_def memory_eq_def split: split_if_asm)[2]
  done

text {* More internal Hoare triples *}

lemma restart_mangle_sv_no_revival:
  "\<lbrace>-no_failed\<rbrace> 
    restart_mangle_sv
  \<lbrace>> K (-no_failed)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' v sv sp)
  apply(clarsimp_all simp add: min_def brute_force no_failed_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  done

lemma idle_restart_no_revival:
  "\<lbrace>-no_failed\<rbrace> 
    <tryR> restart_mangle_sv <catchR> startup
  \<lbrace>> K (-no_failed)\<rbrace>"
  by(intro Hoare_try_catchR_inv startup_no_revival restart_mangle_sv_no_revival)

lemma idle_restart_no_mem_upd:
  "\<lbrace>memory_eq mem\<rbrace> 
    (<tryR> restart_mangle_sv <catchR> startup)
  \<lbrace>>R K (memory_eq mem)\<rbrace>"
  apply(auto simp add: decompose_HoareN)
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: abs6_skip_defs
    startup_defs min_def brute_force memory_eq_def 
    split: split_if_asm exceptions.split_asm)
  done

text {* And then R3 for idle restarts *}

theorem idle_restart_no_obs_mem_upd:
  "\<lbrace>abs6_obs -` (obs_mem_in {mem})\<rbrace> 
    (<tryR> restart_mangle_sv <catchR> startup)
  \<lbrace>> K (abs6_obs -` (obs_mem_in {mem, None}))\<rbrace>\<^sub>N"
  apply(case_tac mem, simp_all)
  apply(rule conseqN_rule[OF _ idle_restart_no_revival[THEN Hoare_imp_HoareN]])
  apply(auto simp add: obs_mem_in_def no_failed_def K_def split: split_if_asm)[2]
  apply(clarsimp)
  apply(rule_tac mem1=a in conseqN_rule[OF _ idle_restart_no_mem_upd])
  apply(auto simp add: obs_mem_in_def K_def memory_eq_def split: split_if_asm)
  done

(********************************************************************)
subsubsection {* R4: matching modes of operation and termination *}
(********************************************************************)

text {* On abs6, the mode is the volatile component of the state,
 for both internal and observable states *}

definition abs6_obs_mode_isnt where
  "abs6_obs_mode_isnt mode \<equiv> {(sv, sp). sv \<noteq> mode}"

text {* We employ a similar, already existing definition, @{term
  abs6_obs_mode_is}. The statements are straightforward *}

theorem update_norm_Normal:
  "\<lbrace>P\<rbrace> <tryR> update upd_f <catchR> startup \<lbrace>> K (abs6_obs_mode_is ACNormal) \<rbrace>\<^sub>N\<^isub>N"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def abs6_obs_mode_is_def
    startup_defs
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  apply(case_tac a, simp_all)
  done

theorem update_exc_not_Normal_or_restart:
  "\<lbrace>P\<rbrace> <tryR> update upd_f <catchR> startup 
  \<lbrace>> \<lambda>e. 
    if e = StartedExc 
    then (abs6_obs_mode_is ACNormal) 
    else if e = InvalidInput 
        then P
        else (abs6_obs_mode_isnt ACNormal) 
  \<rbrace>\<^sub>N\<^isub>E"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage split del: split_if)
  apply(rename_tac sv' sp' e sv sp)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def
    startup_defs brute_force abs6_obs_mode_is_def abs6_obs_mode_isnt_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp simp add: min_def split: split_if_asm)+
  done

theorem fetch_norm_Normal:
  "\<lbrace>P\<rbrace> <tryR> fetch ind <catchR> startup \<lbrace>> K (abs6_obs_mode_is ACNormal) \<rbrace>\<^sub>N\<^isub>N"
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force abs6_obs_mode_is_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(case_tac[!] a, simp_all)
  done

theorem fetch_exc_not_Normal_or_restart:
  "\<lbrace>P\<rbrace> <tryR> fetch ind <catchR> startup \<lbrace>> \<lambda>e. 
    if e = StartedExc 
    then (abs6_obs_mode_is ACNormal) 
    else (abs6_obs_mode_isnt ACNormal) 
  \<rbrace>\<^sub>N\<^isub>E"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage split del: split_if)
  apply(rename_tac sv' sp' e sv sp)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force abs6_obs_mode_is_def abs6_obs_mode_isnt_def
    split: split_if_asm exceptions.split_asm)
  done
 
(********************************************************************)
subsubsection {* R5: correctness of the card mode *}
(********************************************************************)

text {* The following invariant expresses R5. On this model, we have
  an explicit state field @{term reliable} which tracks its integrity.
  The set @{term "in_view sp"} is then defined as the set of those
  instances who are both reliable and up-to-date, that is, the arrow
  is bi-directional or points away from them. The refinement relation
  between abs4 and abs5 forces the contents of such instances to match
  the abstract view, making them "active"
*}

definition abs4_correct_mode where
  "abs4_correct_mode \<equiv> {(sv, sp).
    (abs_card_mode sv = ACNormal \<longrightarrow> in_view sp = UNIV \<and> abs4_state.admin_state sp = LOK)
    \<and> (abs_card_mode sv = ACDegraded \<longrightarrow> 
        ((\<exists>ind inst. (ind, inst) \<notin> writable sp) \<or> abs4_state.admin_state sp = LDegraded)
        \<and> (\<forall>ind. \<exists>inst. (ind, inst) \<in> in_view sp)
    )
    \<and> (abs_card_mode sv = ACFail \<longrightarrow> 
        (\<exists>ind. \<forall>inst. (ind, inst) \<notin> in_view sp) 
        \<or> abs4_state.admin_state sp = LDead
    )
  }"

text {* Intermediate lemmas used in the proof *}

lemma abs4_oreach_inclusion: "abs4_abs5_pi ` oreach abs4_spec \<subseteq> oreach abs6_spec"
proof(rule subset_trans[where B="oreach abs5_spec"])
  show "abs4_abs5_pi ` oreach abs4_spec \<subseteq> oreach abs5_spec" 
    by (intro implements_oreach abs4_implements_abs5)
next
  have "id ` oreach abs5_spec \<subseteq> oreach abs6_spec" 
    by(intro implements_oreach abs5_implements_abs6)
  thus "oreach abs5_spec \<subseteq> oreach abs6_spec"
    by simp
qed

definition Normal_imp_all_in_view where 
  "Normal_imp_all_in_view \<equiv> {(sv, sp). abs_card_mode sv = ACNormal \<longrightarrow> in_view sp = UNIV}"

lemma Normal_imp_all_in_view_alt:
  "Normal_imp_all_in_view = all_SEq \<inter> all_reliable_inv"
  apply(auto simp add: Normal_imp_all_in_view_def all_SEq_def all_reliable_inv_def in_view_def)
  apply(subgoal_tac "Primary \<in> arrow_srcs (abs4_state.arrow b ind) \<and> 
                    Secondary \<in> arrow_srcs (abs4_state.arrow b ind)")
  apply(case_tac "abs4_state.arrow b ind", auto)
  done

lemma abs4_6_ref:
  "PO_refines (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)) abs5_spec abs4_spec"
  using abs4_refines_abs5
  by(auto simp add: refines_def)

lemma constrain_concrete:
  "PO_refines R ta tc \<Longrightarrow> reach tc \<subseteq> Range R"
  apply(auto simp add: PO_refines_def PO_rhoare_def)
  apply(erule reach.induct, auto)
  done

lemma abs4_obs_is_id:
  "obs abs4_spec = id"
  by(auto simp add: abs4_spec_def TS.extend_def spec.fields_def)

lemma abs4_oreach_is_reach:
  "oreach abs4_spec = reach abs4_spec"
  by(auto simp add: oreach_def abs4_obs_is_id)

lemma abs4_Normal_all_in_view:
  "oreach abs4_spec \<subseteq> Normal_imp_all_in_view"
proof-
  from abs4_6_ref[THEN constrain_concrete]
  have "reach abs4_spec \<subseteq> Normal_imp_all_in_view"
    by(auto simp add: Normal_imp_all_in_view_alt)
  thus ?thesis using abs4_oreach_is_reach
    by(auto)
qed

lemma comb_rel_writ_rev_cases:
  "(comb_rel_writ (r1, w1) (r2, w2) = LDegraded) = ((r1 \<or> r2) \<and> (\<not>w1 \<or> \<not>w2))"
  "(LDegraded = comb_rel_writ (r1, w1) (r2, w2)) = ((r1 \<or> r2) \<and> (\<not>w1 \<or> \<not>w2))"
  "(comb_rel_writ (r1, w1) (r2, w2) = LDead) = (\<not>r1 \<and> \<not>r2)"
  "(LDead = comb_rel_writ (r1, w1) (r2, w2)) = (\<not>r1 \<and> \<not>r2)"
  by(case_tac[!] r1, case_tac[!] r2, case_tac[!] w1, case_tac[!] w2, simp_all)

text {* Finally, we can prove the desired requirement *}

theorem correct_mode: "oreach abs4_spec \<subseteq> abs4_correct_mode"
proof(clarsimp)
  fix sv sp
  assume reach: "(sv, sp) \<in> oreach abs4_spec"
  from order_trans[OF abs4_oreach_inclusion Abs6_Lemmas.abs6_obs_correct_mode]
    and reach
  have abs_inv: "abs4_abs5_pi (sv, sp) \<in> abs6_obs_correct_mode"
    by(auto simp add: image_eqI simp del: abs4_abs5_pi.simps)
  show "(sv, sp) \<in> abs4_correct_mode"
  proof(clarsimp simp add: abs4_correct_mode_def, intro conjI, rule_tac[!] impI)
    assume "abs_card_mode sv = ACDegraded"
    thus "((\<exists>ind inst. (ind, inst) \<notin> writable sp) \<or> abs4_state.admin_state sp = LDegraded) 
      \<and> (\<forall>ind. \<exists>inst. (ind, inst) \<in> in_view sp)" using abs_inv
      by(auto simp add: abs6_obs_correct_mode_def 
        abs_reg_state_def image_def comb_rel_writ_rev_cases
        split: split_if_asm)
  next
    assume "abs_card_mode sv = ACFail"
    thus "(\<exists>ind. \<forall>inst. (ind, inst) \<notin> in_view sp) \<or> abs4_state.admin_state sp = LDead " 
      using abs_inv
      apply(auto simp add: abs6_obs_correct_mode_def 
        abs_reg_state_def image_def comb_rel_writ_rev_cases
        split: split_if_asm)
      apply (metis (full_types) point_arrow_to.cases)
      done
  next
    assume "abs_card_mode sv = ACNormal"
    thus "in_view sp = UNIV \<and> abs4_state.admin_state sp = LOK"
      using abs_inv abs4_Normal_all_in_view reach
      by(auto simp add: Normal_imp_all_in_view_def abs6_obs_correct_mode_def)
  qed
qed

(********************************************************************)
subsubsection {* R6: maximum redundancy *}
(********************************************************************)

thm writable_in_view_def

lemma startup_redundancy_N:
  "\<lbrace>UNIV\<rbrace> Abs4_StartupDefs.startup \<lbrace>> K writable_in_view \<rbrace>\<^sub>N"
  by(auto simp add: decompose_HoareN intro!: Abs4_Invariant_Lemmas.startup_post_NN
    startup_redundancy_NE)
  
theorem maximum_redundancy:
  "oreach abs4_spec \<subseteq> writable_in_view"
proof(clarsimp simp add: oreach_def abs4_spec_def reach_spec obs_ts)
  fix s
  assume reach: "s \<in> reach abs4_ts"
  thus "s \<in> writable_in_view"
  proof(induct)
    fix s
    assume "s \<in> init_trans abs4_ts `` init abs4_ts"
    thus "s \<in> writable_in_view" using startup_redundancy_N
      apply(auto simp add: abs4_ts_def abs4_trans_def choice_def projectSN_def projectS_def
        set_fun_into_rel_def)
      apply(force simp add: unfold_Hoare UNIV_subset Collect_UNIV)
      done
  next
    fix s t
    assume assms: "s \<in> reach abs4_ts" "s \<in> writable_in_view" "(s, t) \<in> TS.trans abs4_ts" 
    hence rel: "s \<in> all_reliable_inv" using abs4_Normal_all_in_view
      by(auto simp add: abs4_spec_def oreach_def reach_spec obs_ts Normal_imp_all_in_view_alt)
    have trans: "TS.trans abs4_ts = abs4_trans"
      by(simp add: abs4_ts_def)
    show "t \<in> writable_in_view" using assms
      apply(auto simp add: trans abs4_trans_def)
      apply(auto simp add: choice_def projectSN_def projectS_def set_fun_into_rel_def try_rec_tryR_def)
      apply(insert rel update_call_redundancy)[1]
      apply(drule_tac x=upd_f in meta_spec)
      apply(auto simp add: unfold_Hoare)[1]
      apply(insert fetch_call_redundancy)[1]
      apply(drule_tac x=ind in meta_spec, auto simp add: unfold_Hoare)[1]
      apply(insert idle_restart_redundancy[where P=UNIV], auto simp add: unfold_Hoare)[1]
      done
  qed
qed

end