theory Restart_Transformations
  imports Repeated_Restarts 
begin

inductive_set repeatS :: 
  "('s \<Rightarrow> 's set) \<Rightarrow> 's \<Rightarrow> 's set"
  for m :: "('s \<Rightarrow> 's set)"
  and s :: 's
  where 
  skip: "s \<in> repeatS m s"
  | repeat: "\<lbrakk> s' \<in> repeatS m s; s'' \<in> m s' \<rbrakk> \<Longrightarrow> s'' \<in> repeatS m s"

lemma rec_tryRS': 
  "\<lbrakk> (r, s') \<in> rec_tryR m s \<rbrakk> \<Longrightarrow> snd (r, s') \<in> repeatS (m\<Down>R) s \<or> fst (r, s') \<noteq> None "
  apply(induct rule: rec_tryR.induct)
  apply(auto simp add: unfold_monad projectR_def intro: repeatS.skip repeatS.repeat)
  done

lemma rec_tryRS_None:
  "(None, s') \<in> rec_tryR m s \<Longrightarrow> s' \<in> repeatS (m\<Down>R) s"
  by(drule rec_tryRS', simp)

fun the_first where
  "the_first (Some r, s') = (r, s')"

lemma rec_tryRS_Some': 
  "r \<in> rec_tryR m s \<Longrightarrow> (the_first r) \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>N)) \<or> fst r = None"
  apply(induct rule: rec_tryR.induct)
  apply(clarsimp)
  apply(auto simp add: unfold_monad projectN_def 
    dest: rec_tryRS_None intro: repeatS.skip repeatS.repeat)
  done

lemma rec_tryRS_Some: 
  "(Some r, s') \<in> rec_tryR m s \<Longrightarrow> (r, s') \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>N))"
  by(drule rec_tryRS_Some', simp)

fun add_Some where
  "add_Some (a, b) = (Some a, b)"

lemma repeatS_catchR_None':
  "s' \<in> repeatS (m\<Down>R) s \<Longrightarrow> (None, s') \<in> rec_tryR m s \<or> s' = s"
  apply(induct rule: repeatS.induct)
  apply(simp)
  apply(auto simp add: projectN_def projectR_def 
    intro: rec_tryR.exec_none rec_tryR.exec_some)
  done

lemma repeatS_catchR_None:
  "\<lbrakk> s' \<in> repeatS (m\<Down>R) s; s' \<noteq> s  \<rbrakk> \<Longrightarrow> (None, s') \<in> rec_tryR m s"
  by(drule repeatS_catchR_None', simp)

lemma repeatS_try_catchR_Some_unfolded:
  "\<lbrakk>x \<in> repeatS (m\<Down>R) s; (r, s') \<in> (m\<Down>N) x\<rbrakk>
        \<Longrightarrow> (Some r, s') \<in> rec_tryR m s"
  apply(induct rule: repeatS.induct)
  apply(clarsimp simp add: projectN_def intro!: rec_tryR.exec_some)
  apply(drule repeatS_catchR_None')
  apply(safe)
  apply(force simp add: projectR_def projectN_def dest: rec_tryR.exec_none)
  apply(subgoal_tac "(None, s'') \<in> rec_tryR m s")
  apply(clarsimp simp add: projectN_def elim!: rec_tryR.exec_none)
  apply(clarsimp simp add: projectR_def elim!: rec_tryR.exec_some)
  done

lemma repeatS_try_catchR_Some:
  "(r, s') \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>N)) \<Longrightarrow> (Some r, s') \<in> rec_tryR m s"
  by(auto simp add: unfold_monad repeatS_try_catchR_Some_unfolded)

lemma rec_tryRS_eq:
  "((Some r, s') \<in> rec_tryR m s) = ((r, s') \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>N)))"
  by(auto simp add: rec_tryRS_Some repeatS_try_catchR_Some)

text {*

  We want to return an @{typ "('a, 's)nds_monad"} rather than @{typ "'s \<Rightarrow> 's set"}, so
  that we can bind the result on the right-hand side of bindN (see the formulation
  of repeatS_N_rewrite1. This unfortunately neccessitates the use of unit type
  as the return type for the monad, since "repeatN" also includes "repeat 0 times"

*}

inductive_set repeatN :: 
  "('a, 's)nds_monad \<Rightarrow> (unit, 's)nds_monad"
  for
  m :: "('a, 's)nds_monad"
  and s :: 's
  where
  skip: "((), s) \<in> repeatN m s"
  | repeat: "\<lbrakk> ((), s') \<in> repeatN m s; (x, s'') \<in> m s' \<rbrakk> \<Longrightarrow> ((), s'') \<in> repeatN m s"

lemma repeatN_absorb': " \<lbrakk>((), s'') \<in> repeatN m s'; (a, s') \<in> m s \<rbrakk> \<Longrightarrow> ((), s'') \<in> repeatN m s"
  by(induct rule: repeatN.induct, auto intro: repeatN.skip repeatN.repeat)

lemma repeatN_absorb: "(m \<guillemotright>N repeatN m) s \<subseteq> (repeatN m) s"
  by(auto simp add: unfold_monad dest: repeatN_absorb')

lemma repeatS_absorb_L_exp: "\<lbrakk> s' \<in> m s; s'' \<in> repeatS m s' \<rbrakk> \<Longrightarrow> s'' \<in> repeatS m s"
  apply(erule repeatS.induct)
  apply(auto simp add: unfold_monad project_defs intro: repeatS.skip repeatS.repeat)
  done

lemma repeatS_of_repeatS:
  "\<lbrakk> s'' \<in> repeatS m s'; s' \<in> repeatS m s \<rbrakk> \<Longrightarrow> s'' \<in> repeatS m s"
  apply(induct rule: repeatS.induct)
  apply(auto intro: repeatS.skip repeatS.repeat)
  done

lemma repeatS_absorb_L:
  "m s \<guillemotright>=S repeatS m \<subseteq> repeatS m s"
  by(auto simp add: unfold_monad repeatS_absorb_L_exp)

lemma repeatS_absorb_R:
  "repeatS m s \<guillemotright>=S m \<subseteq> repeatS m s"
  by(auto simp add: unfold_monad intro: repeatS.repeat)

fun repeatS_N :: "('s \<Rightarrow> 's set) \<Rightarrow> nat \<Rightarrow> 's \<Rightarrow> 's set" where
  "repeatS_N m 0 = (\<lambda>s. {s})"
  | "repeatS_N m (Suc n) = (\<lambda>s. m s \<guillemotright>=S repeatS_N m n)"

lemma repeatS_N_alter:
  "repeatS_N m (Suc n) s = repeatS_N m n s \<guillemotright>=S m"
  apply(induct n arbitrary: s)
  apply(simp add: unfold_monad)
  apply(atomize)
  apply(simp)
  apply(subgoal_tac "m s \<guillemotright>=S repeatS_N m n = repeatS_N m n s \<guillemotright>=S m")
  apply(erule subst)
  apply(rule subst[OF bindS_assoc])
  apply(simp_all)
  done

lemma repeatS_bound: "s' \<in> repeatS m s \<Longrightarrow> \<exists>n. s' \<in> repeatS_N m n s"
  apply(induct rule: repeatS.induct)
  apply(rule_tac x=0 in exI, simp)
  apply(clarify)
  apply(rule_tac x="Suc n" in exI)
  apply(simp only: repeatS_N_alter)
  apply(auto simp add: unfold_monad)
  done

fun repeatN_N :: "('a, 's)nds_monad \<Rightarrow> nat \<Rightarrow> (unit, 's)nds_monad" where
  "repeatN_N m 0 = returnN ()"
  | "repeatN_N m (Suc n) = (m \<guillemotright>N repeatN_N m n)"

lemma repeatN_N_alter:
  "repeatN_N m (Suc n) = repeatN_N m n \<guillemotright>N (m \<guillemotright>N returnN ())"
  apply(induct n)
  apply(simp add: monad_laws)
  apply(simp)
  apply(simp only: bindN_assoc[THEN sym])
  done

lemma repeatN_bound: "((), s') \<in> repeatN m s \<Longrightarrow> \<exists>n. ((), s') \<in> repeatN_N m n s"
  apply(induct rule: repeatN.induct)
  apply(rule_tac x=0 in exI, simp add: unfold_monad)
  apply(clarify)
  apply(rule_tac x="Suc n" in exI)
  apply(simp only: repeatN_N_alter)
  apply(force simp add: unfold_monad)
  done

lemma repeatS_mono:
  " \<forall>s. m s \<subseteq> m' s \<Longrightarrow> repeatS m s \<subseteq> repeatS m' s"
proof
  fix x   
  assume "x \<in> repeatS m s" "\<forall>s. m s \<subseteq> m' s "
  thus 
    "x \<in> repeatS m' s"
    by(induct rule: repeatS.induct, auto intro: repeatS.skip repeatS.repeat)
qed

lemma repeatS_repeatS:
  "repeatS m s \<guillemotright>=S repeatS m \<subseteq> repeatS m s"
  apply(clarsimp simp add: bindS_def)
  apply(erule_tac x=x in repeatS.induct)
  apply(auto intro: repeatS.repeat)
  done

lemma choice_bindS:
  "(((m1 [+] m2) s) \<guillemotright>=S m3) = ((m1 s \<guillemotright>=S m3) \<union> (m2 s \<guillemotright>=S m3))"
  by(auto simp add: unfold_monad)

lemma bindS_choice_subset:
  "\<lbrakk> m1 s \<guillemotright>=S m2 \<subseteq> m2 s; m1 s \<guillemotright>=S m3 \<subseteq> m3 s \<rbrakk> \<Longrightarrow> m1 s \<guillemotright>=S (m2 [+] m3) \<subseteq> (m2 [+] m3) s"
  by(auto simp add: unfold_monad)

lemma bindN_mono1:
  "\<forall>s. ml' s \<subseteq> ml s \<Longrightarrow> (ml' \<guillemotright>=N m) s \<subseteq> (ml \<guillemotright>=N m) s"
  by(auto simp add: unfold_monad)

lemma bindN_mono2:
  "\<forall>a s. mr' a s \<subseteq> mr a s \<Longrightarrow> (m \<guillemotright>=N mr') s \<subseteq> (m \<guillemotright>=N mr) s"
  by(auto simp add: unfold_monad)

lemma bindN'_mono2:
  "\<forall>s. mr' s \<subseteq> mr s \<Longrightarrow> (m \<guillemotright>N mr') s \<subseteq> (m \<guillemotright>N mr) s"
  by(auto simp add: unfold_monad)

lemma repeatS_skip_left:
  "b s \<subseteq> repeatS a s \<guillemotright>=S b"
  apply(auto simp add: unfold_monad) 
  apply(rule_tac x=s in bexI)
  apply(auto intro: repeatS.skip)
  done

lemma repeatS_choice_rewrite:
  "repeatS (a [+] (b \<guillemotright>N c))
    = repeatS a [+] 
    (
        (\<lambda>s. (repeatS a s) \<guillemotright>=S b)
        \<guillemotright>N 
        repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))
        \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S repeatS a)
    ) 
  "
proof(rule ext, rule equalityI)
  def pt1 \<equiv> "(\<lambda>s. (repeatS a s) \<guillemotright>=S b)"
  def pt2 \<equiv> "repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))"
  def pt3 \<equiv> "(\<lambda>s. c s \<guillemotright>=S repeatS a)"

  fix s
  {
    fix n
    

    have "repeatS_N (a [+] (b \<guillemotright>N c)) n s
      \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s
      "
    proof(induct n arbitrary: s)
      case 0
      show ?case
        by(auto simp add: unfold_monad intro: repeatS.skip)
    next
      case (Suc n')
      note ind_hyp = this
      {
        fix x A
        have
        "x \<in> A \<Longrightarrow> {x} \<subseteq> A"
          by(auto)
      } note singleton_subsetI=this
      have "repeatS_N (a [+] (b \<guillemotright>N c)) (Suc n') s
        \<subseteq> (a [+] (b \<guillemotright>N c)) s \<guillemotright>=S (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3))
       "
        by(auto intro!: bindS_mono2 ind_hyp)
      moreover have "... =
        (a s \<guillemotright>=S (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)))
        \<union> ((b \<guillemotright>N c) s \<guillemotright>=S (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)))
        "
        by(simp add: choice_bindS)
      moreover have "... \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s"
      proof(rule Un_least)
        
        show "a s \<guillemotright>=S (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s"
        proof(rule bindS_choice_subset)
          show "a s \<guillemotright>=S repeatS a \<subseteq> repeatS a s"
            by (rule repeatS_absorb_L)
        next
          have "
            ((\<lambda>s. (a s \<guillemotright>=S repeatS a) \<guillemotright>=S b) \<guillemotright>N pt2 \<guillemotright>N pt3) s
            \<subseteq> ((\<lambda>s. repeatS a s \<guillemotright>=S b) \<guillemotright>N pt2 \<guillemotright>N pt3) s"
            apply(rule subset_trans)
            apply(simp only: bindN_assoc)
            apply(rule bindN_mono1)
            apply(rule allI)
            apply(rule bindS_mono1)
            apply(rule repeatS_absorb_L)
            apply(simp add: monad_laws)
            done

          thus "a s \<guillemotright>=S pt1 \<guillemotright>N pt2 \<guillemotright>N pt3 \<subseteq> (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3) s"
            by(simp add: bindS_bindN pt1_def monad_laws)
        qed
      next
        show "
          (b \<guillemotright>N c) s \<guillemotright>=S repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)
          \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s"
        proof(simp only: bindS_choice, rule Un_least)
          have "\<forall>s. returnN () s \<subseteq> pt2 s"
            by(clarsimp simp add: pt2_def unfold_monad intro!: repeatN.skip)
          hence "\<forall>s. (returnN () \<guillemotright>N pt3) s \<subseteq> (pt2 \<guillemotright>N pt3) s"
            by(intro allI bindN_mono1, simp)
          hence ret_pt2: "\<forall>s. pt3 s \<subseteq> (pt2 \<guillemotright>N pt3) s"
            by(simp add: monad_laws)
          have skip_middle:
            "(pt1 \<guillemotright>N pt3) s \<subseteq> (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3) s"
            by(auto simp add: monad_laws intro!: bindN_mono2 ret_pt2)
          have skip_repeat: "(b \<guillemotright>N pt3) s \<subseteq> ((\<lambda>s. repeatS a s \<guillemotright>=S b) \<guillemotright>N pt3) s "
            by(intro bindN_mono1 allI repeatS_skip_left)
          have
            "(b \<guillemotright>N c) s \<guillemotright>=S repeatS a 
            \<subseteq>
            (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3) s"
            apply(rule subset_trans[OF _ skip_middle])
            apply(simp add: pt1_def)
            apply(rule subset_trans[OF _ skip_repeat])
            apply(simp add: pt3_def bindN_bindS monad_laws)
            done
          thus "(b \<guillemotright>N c) s \<guillemotright>=S repeatS a \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s"
            by(auto simp add: choice_def)
        next
          have "(b \<guillemotright>N c) s \<guillemotright>=S pt1 \<guillemotright>N pt2 \<guillemotright>N pt3 \<subseteq> (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3) s" (is "?L \<subseteq> ?R")
          proof-
            def pt2_int \<equiv> "\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b))" 
            have "?L = (b \<guillemotright>N (pt2_int \<guillemotright>N pt2) \<guillemotright>N pt3) s"
              by(force simp add: pt1_def pt2_int_def unfold_monad)
            moreover have 
              "\<forall>s. (pt2_int \<guillemotright>N pt2) s \<subseteq> pt2 s"
              by(clarsimp simp only: pt2_def pt2_int_def repeatN_absorb)
            ultimately have "?L \<subseteq> (b \<guillemotright>N pt2 \<guillemotright>N pt3) s"
              by(clarsimp intro!: bindN_mono1 bindN_mono2)
            thus "?L \<subseteq> (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3) s"
              apply(elim subset_trans)
              apply(simp add: monad_laws)
              apply(rule bindN_mono1)
              apply(simp add: pt1_def repeatS_skip_left)
              done
          qed
          thus "(b \<guillemotright>N c) s \<guillemotright>=S pt1 \<guillemotright>N pt2 \<guillemotright>N pt3 \<subseteq> (repeatS a [+] (pt1 \<guillemotright>N pt2 \<guillemotright>N pt3)) s"
            by(auto simp add: choice_def)
        qed
      qed
      ultimately show ?case
        by(auto)
    qed
  }
  from this[THEN subsetD] 
  show "repeatS (a [+] (b \<guillemotright>N c)) s
    \<subseteq> (repeatS a [+] 
    (
        (\<lambda>s. (repeatS a s) \<guillemotright>=S b)
        \<guillemotright>N 
        repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))
        \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S repeatS a)
    )) s
  "
    by(clarsimp dest!: repeatS_bound simp add: pt1_def pt2_def pt3_def)
next

  fix s

  def pt1 \<equiv> "(\<lambda>s. (repeatS a s) \<guillemotright>=S b)"
  def pt2 \<equiv> "repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))"
  def pt3 \<equiv> "(\<lambda>s. c s \<guillemotright>=S repeatS a)"

  show "
    (repeatS a [+] 
    (
      (\<lambda>s. (repeatS a s) \<guillemotright>=S b)
      \<guillemotright>N 
      repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))
      \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S repeatS a)
    )) s
    \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s
    " 
  proof(subst choice_def, rule Un_least)
    show "repeatS a s \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s"
      by(clarsimp intro!: repeatS_mono simp add: choice_def)
  next
    {
      fix n
      def pt2_int \<equiv> "(\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))"
      {
        fix m1 m2 m3
        have "(m1 \<guillemotright>N (\<lambda>s. m2 s \<guillemotright>=S m3)) = (\<lambda>s. (m1 \<guillemotright>N m2) s \<guillemotright>=S m3)"
          by(force simp add: unfold_monad intro!: ext)
      } note bindN_bindS2 = this
      have "(pt1 \<guillemotright>N repeatN_N pt2_int n \<guillemotright>N pt3) s \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s"
      proof(induct n arbitrary: s)
        case 0
        fix s'
        have "(pt1 \<guillemotright>N pt3) s' \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s'"
          apply(simp add: pt1_def pt3_def)
          apply(simp only: bindS_bindN[THEN sym])
          apply(simp only: bindN_bindS2)
          apply(rule subset_trans)
          apply(rule_tac S="repeatS (a [+] (b \<guillemotright>N c)) s'" in bindS_mono1)
          apply(clarsimp intro!: repeatS_mono simp add: choice_def)
          apply(rule subset_trans)
          apply(rule_tac m="repeatS (a [+] (b \<guillemotright>N c))" in bindS_mono2)
          apply(rule allI)
          apply(rule subset_trans)
          apply(rule_tac S="repeatS (a [+] (b \<guillemotright>N c)) s" in bindS_mono1)
          apply(force intro: repeatS.repeat repeatS.skip simp add: choice_def)
          apply(rule subset_trans)
          apply(rule_tac m="repeatS (a [+] (b \<guillemotright>N c))" in bindS_mono2)
          apply(auto intro!: repeatS_mono repeatS_repeatS simp add: choice_def)
          done

        thus "(pt1 \<guillemotright>N repeatN_N pt2_int 0 \<guillemotright>N pt3) s' \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s'"
          by(simp add: monad_laws)
      next
        case (Suc n')
        fix s'
        {
          fix m
          have "pt2_int \<guillemotright>N m = (\<lambda>s. c s \<guillemotright>=S (pt1 \<guillemotright>N m))"
            by(simp only: pt2_int_def pt1_def bindS_bindN)
        } note rew1=this
        have rew2: "(pt1 \<guillemotright>N c) s' \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s' "
          apply(simp add: pt1_def)
          apply(simp only: bindS_bindN[THEN sym])
          apply(rule subset_trans)
          apply(rule_tac S="repeatS (a [+] (b \<guillemotright>N c)) s'" in bindS_mono1)
          apply(clarsimp intro!: repeatS_mono simp add: choice_def)
          apply(rule subset_trans)
          apply(rule_tac m="repeatS (a [+] (b \<guillemotright>N c))" in bindS_mono2)
          apply(force intro: repeatS.repeat repeatS.skip simp add: choice_def)
          apply(rule repeatS_repeatS)
          done
       
        have "
            (pt1 \<guillemotright>N repeatN_N pt2_int (Suc n') \<guillemotright>N pt3) s'
           = (pt1 \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S (pt1 \<guillemotright>N repeatN_N pt2_int n' \<guillemotright>N pt3))) s'"
          apply(simp add: monad_laws)
          apply(simp only: rew1)
          done
        moreover have "... \<subseteq> (pt1 \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S repeatS (a [+] (b \<guillemotright>N c)))) s'"
          apply(intro bindN_mono2 allI bindS_mono2)
          apply(simp only: Suc)
          done
        moreover have "... = ((pt1 \<guillemotright>N c) s' \<guillemotright>=S repeatS (a [+] (b \<guillemotright>N c)))"
          by(simp add: bindN_bindS monad_laws)
        moreover have "... \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s'"
          apply(rule subset_trans)
          apply(rule bindS_mono1)
          apply(rule rew2, rule repeatS_repeatS)
          done
        ultimately show 
          "(pt1 \<guillemotright>N repeatN_N pt2_int (Suc n') \<guillemotright>N pt3) s' \<subseteq> repeatS (a [+] (b \<guillemotright>N c)) s'"
          by(auto)
      qed
    }
    note bounded=this
    
    {
      fix m x
      have "x \<in> (pt1 \<guillemotright>N repeatN m \<guillemotright>N pt3) s
        \<Longrightarrow> \<exists>n. x \<in> (pt1 \<guillemotright>N repeatN_N m n \<guillemotright>N pt3) s"
        by(auto simp add: unfold_monad dest!: repeatN_bound)
    } note bounding=this
    
    show 
      "
      (
          (\<lambda>s. (repeatS a s) \<guillemotright>=S b)
          \<guillemotright>N 
          repeatN (\<lambda>s. (c s \<guillemotright>=S (\<lambda>s. repeatS a s \<guillemotright>=S b)))
          \<guillemotright>N (\<lambda>s. c s \<guillemotright>=S repeatS a)
      ) s
      \<subseteq> 
      repeatS (a [+] (b \<guillemotright>N c)) s
    " 
      apply(fold pt1_def pt3_def)
      apply(clarsimp dest!: bounding)
      apply(simp add: pt1_def pt3_def bounded[simplified pt1_def pt3_def, THEN subsetD])
      done
  qed
qed

lemma repeatS_rewrite_right: 
  "
  repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2 
  \<subseteq>
  repeatS m1 s \<guillemotright>=S (m2 \<guillemotright>N repeatN (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2))"
proof-

  {
    fix m1 m2 m3 n
    have
      "
      repeatS_N (m1 [+] (m2 \<guillemotright>N m3)) n s \<guillemotright>=S m2
      \<subseteq>
      repeatS m1 s \<guillemotright>=S (m2 \<guillemotright>N repeatN (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2))"
    proof(induct n arbitrary: s)
      case 0
      show ?case
        by(auto simp add: unfold_monad intro: repeatS.skip repeatN.skip)
    next
      case (Suc n)
      note ind_hyp = this
      {
        fix x A
        have
        "x \<in> A \<Longrightarrow> {x} \<subseteq> A"
          by(auto)
      } note singleton_subsetI=this
      
      have absorption:
        "
        snd ` m2 s \<guillemotright>=S
        (\<lambda>s. m3 s \<guillemotright>=S (\<lambda>a. repeatS m1 a \<guillemotright>=S m2)) \<guillemotright>N
        repeatN (\<lambda>s. m3 s \<guillemotright>=S (\<lambda>s. repeatS m1 s \<guillemotright>=S m2))
        \<subseteq>
        snd ` m2 s \<guillemotright>=S
        repeatN (\<lambda>s. m3 s \<guillemotright>=S (\<lambda>s. repeatS m1 s \<guillemotright>=S m2))
      
        "
        apply(rule bindS_mono2)
        apply(simp add: repeatN_absorb)
        done
      
      show ?case
        apply(simp add: monad_laws)
        apply(rule subset_trans[OF bindS_mono2[OF ind_hyp[THEN allI]]])
        apply(simp only: choice_distrib)
        apply(rule Un_least)
        apply(simp only: monad_laws)
        apply(subst bindS_assoc[THEN sym])
        apply(simp only: repeatS_absorb_L[THEN bindS_mono1])
        apply(rule subset_trans[OF _ bindS_mono1[OF repeatS.skip[THEN singleton_subsetI]]])
        apply(subst bindN_bindS)
        apply(simp only: monad_laws)
        apply(subst bindS_bindN)
        apply(subst bindS_bindN)
        apply(rule subset_trans[OF absorption])
        apply(auto simp add: unfold_monad)
        done
    qed
  } note repeatS_N_rewrite1=this

  {
    fix x m1 m2
    have
      "x \<in> repeatS m1 s \<guillemotright>=S m2 
      \<Longrightarrow> \<exists>n. x \<in> repeatS_N m1 n s \<guillemotright>=S m2"
      by(auto simp add: unfold_monad dest!: repeatS_bound)
  } note repeatS_rewrite1=this

  show ?thesis
    apply(rule subsetI, drule repeatS_rewrite1)
    apply(clarsimp)
    apply(drule repeatS_N_rewrite1[THEN subsetD])
    apply(auto)
    done
qed

lemma repeatS_N_rewrite_left:
  "
  repeatS m1 s \<guillemotright>=S (m2 \<guillemotright>N repeatN_N (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) n)
  \<subseteq>
  repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2
  "
proof(induct n arbitrary: s)
  case 0
  show ?case
    by(auto intro!: bindS_mono1 repeatS_mono simp add: monad_laws choice_def)
next
  case (Suc n)
  note ind_hyp = this

  have rew1: "
    repeatS m1 s \<guillemotright>=S m2 \<guillemotright>N
    (repeatN_N (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) n \<guillemotright>N
     ((\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) \<guillemotright>N returnN ()))
    \<subseteq>
    repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2 \<guillemotright>N
     ((\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) \<guillemotright>N returnN ())
    "
    apply(simp only: bindS_bindN)
    apply(rule subset_trans)
    apply(simp only: bindN_assoc[THEN sym])
    apply(rule bindN_mono1)
    apply(rule allI)
    apply(rule bindN_mono1)
    apply(rule allI)
    apply(rule ind_hyp[simplified bindS_bindN bindN_assoc[THEN sym]])
    apply(simp add: monad_laws)
    done

  {
    fix m1 m2 m3
    have 
    "m1 = m2 \<Longrightarrow> (m1 \<guillemotright>N m3) = (m2 \<guillemotright>N m3)"
      by(clarsimp simp add: unfold_monad)
  } note bindN_eq_leftI=this

  {

    fix m1 m2
    have "m1 \<guillemotright>N m2 = (\<lambda>s. snd ` m1 s \<guillemotright>=S m2)"
      by(rule ext, auto simp add: unfold_monad)
  } note bindN_through_bindS=this

  have rew2:
    "
    (m2 \<guillemotright>N ((\<lambda>s. (((m3 s) \<guillemotright>=S (repeatS m1)) \<guillemotright>=S m2)) \<guillemotright>N (returnN ())))
    = 
    (\<lambda>s. ((m2 \<guillemotright>N m3) s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) \<guillemotright>N (returnN ())
    "
    apply(simp only: bindN_assoc[THEN sym])
    apply(rule bindN_eq_leftI)
    apply(simp only: bindN_through_bindS)
    apply(simp only: monad_laws)
    done

  {
    fix s
    fix m1 m2 m3
    have "repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S (m2 \<guillemotright>N m3)
        \<subseteq> repeatS (m1 [+] (m2 \<guillemotright>N m3)) s"
      apply(rule subset_trans)
      apply(rule_tac m="m1 [+] (m2 \<guillemotright>N m3)" in bindS_mono2)
      apply(auto simp add: choice_def intro!: repeatS_absorb_R)
      done
  } note rew4 = this

  {
    fix m1 m2 m3 s
    have "repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S repeatS m1 \<subseteq> repeatS (m1 [+] (m2 \<guillemotright>N m3)) s"
      apply(rule subset_trans)
      apply(rule bindS_mono2)
      apply(rule allI)
      apply(rule_tac m'="m1 [+] (m2 \<guillemotright>N m3)" in repeatS_mono)
      apply(auto simp add: choice_def intro!: repeatS_repeatS)
      done
  } note rew5 = this

  show "repeatS m1 s \<guillemotright>=S
          (m2 \<guillemotright>N repeatN_N (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2) (Suc n))
          \<subseteq> repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2"
    apply(simp only: repeatN_N_alter)
    apply(rule subset_trans)
    apply(rule rew1)
    apply(simp only: rew2)
    apply(simp only: bindS_bindN)
    apply(simp only: monad_laws)
    apply(rule subset_trans)
    apply(rule bindN_mono1)
    apply(subst bindS_assoc[THEN sym])
    apply(rule allI)
    apply(rule bindS_mono1)
    apply(rule rew4)
    apply(rule subset_trans)
    apply(rule bindN_mono1)
    apply(rule allI)
    apply(subst bindS_assoc[THEN sym])
    apply(rule bindS_mono1)
    apply(rule rew5)
    apply(simp add: monad_laws)
    done
qed

lemma repeatS_rewrite_left:
  "repeatS m1 s \<guillemotright>=S (m2 \<guillemotright>N repeatN (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2))
  \<subseteq>
  repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2
  "
proof-
  {
    fix b m1 m2 m3 s
    have
      "((), b) \<in> m1 s \<guillemotright>=S m2 \<guillemotright>N repeatN m3 \<Longrightarrow> \<exists>n. ((), b) \<in> m1 s \<guillemotright>=S m2 \<guillemotright>N repeatN_N m3 n"
      by(auto simp add: unfold_monad dest!: repeatN_bound)
  } note repeatN_rewrite1=this

  show ?thesis
    by(auto dest!: repeatN_rewrite1 repeatS_N_rewrite_left[THEN subsetD])
qed

lemma repeatS_rewrite:
  "  
  repeatS (m1 [+] (m2 \<guillemotright>N m3)) s \<guillemotright>=S m2
  =
  repeatS m1 s \<guillemotright>=S (m2 \<guillemotright>N repeatN (\<lambda>s. (m3 s \<guillemotright>=S repeatS m1) \<guillemotright>=S m2))
  "
  by(auto intro!: repeatS_rewrite_left repeatS_rewrite_right)

lemma rec_tryR_dec1: "(Some r, s') \<in> rec_tryR (h1 \<guillemotright>R h2) s = 
  ((r, s') \<in> (repeatS (h1\<Down>R [+] (h1\<Down>N \<guillemotright>N h2\<Down>R)) s) \<guillemotright>=S (h1\<Down>N \<guillemotright>N h2\<Down>N))"
  by(simp add: rec_tryRS_eq bindR_projectN bindR_projectR)

lemma rewrite_aux:
  "
  (repeatS (h1 [+] ((h2 :: (unit, 's)nds_monad) \<guillemotright>N h3)) s) \<guillemotright>=S (h2 \<guillemotright>N h4)
  =
  repeatS h1 s \<guillemotright>=S (h2 \<guillemotright>N repeatN (\<lambda>s. (h3 s \<guillemotright>=S repeatS h1) \<guillemotright>=S h2)) \<guillemotright>N h4
  "
  apply(simp only: bindS_bindN)
  apply(simp only: repeatS_rewrite)
  apply(simp only: bindS_bindN)
  done

lemma rec_tryR_dec2: "((Some r, s') \<in> rec_tryR ((h1 :: (unit, 's)r_monad) \<guillemotright>R h2) s) 
  =
  ((r, s') \<in>
  repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>N \<guillemotright>N repeatN (\<lambda>s. ((h2\<Down>R) s \<guillemotright>=S repeatS (h1\<Down>R)) \<guillemotright>=S (h1\<Down>N)) \<guillemotright>N (h2\<Down>N))"
  by(simp only: rec_tryR_dec1 rewrite_aux)

lemma rec_tryR_decomposition: "((Some r, s') \<in> rec_tryR ((h1 :: (unit, 's)r_monad) \<guillemotright>R h2) s) 
  =
  ((r, s') \<in>
  ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>N) \<guillemotright>N 
    repeatN (\<lambda>s. ((h2\<Down>R) s \<guillemotright>=S repeatS (h1\<Down>R)) \<guillemotright>=S (h1\<Down>N)) \<guillemotright>N (h2\<Down>N)) s)
  "
  by(auto simp only: bindS_bindN rec_tryR_dec2)

subsection {* Restart-exception monad *}

subsubsection {* Normal termination *}

lemma rec_tryRS_eq_NN:
  "((Some (Inr r), s') \<in> rec_tryR m s) = ((r, s') \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>NN)))"
  by(auto simp add: rec_tryRS_eq project_defs unfold_monad)

lemma rec_tryR_decomposition_NN_1: "(Some (Inr r), s') \<in> rec_tryR (h1 \<guillemotright> h2) s = 
  ((r, s') \<in> (repeatS (h1\<Down>R [+] (h1\<Down>NN \<guillemotright>N h2\<Down>R)) s) \<guillemotright>=S (h1\<Down>NN \<guillemotright>N h2\<Down>NN))"
  by(simp add: rec_tryRS_eq_NN bind_projectNN bind_projectR)

lemma rec_tryR_decomposition_NN_2: 
  "((Some (Inr r), s') \<in> rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2) s )
  =
  ((r, s') \<in>
  repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN \<guillemotright>N repeatN (\<lambda>s. ((h2\<Down>R) s \<guillemotright>=S repeatS (h1\<Down>R)) \<guillemotright>=S (h1\<Down>NN)) \<guillemotright>N (h2\<Down>NN))"
  by(simp only: rec_tryR_decomposition_NN_1 rewrite_aux)

lemma rec_tryR_decomposition_NN_3: 
  "((Some (Inr r), s') \<in> rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2) s )
  =
  ((r, s') \<in>
  ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N 
    repeatN (\<lambda>s. ((h2\<Down>R) s \<guillemotright>=S repeatS (h1\<Down>R)) \<guillemotright>=S (h1\<Down>NN)) \<guillemotright>N (h2\<Down>NN)) s)
  "
  by(auto simp only: bindS_bindN rec_tryR_decomposition_NN_2)

lemma rec_tryR_transformation_NN_unf: 
  "((Some (Inr r), s') \<in> rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2) s )
  =
  ((r, s') \<in>
  ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N 
    repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S (h1\<Down>NN))) \<guillemotright>N (h2\<Down>NN)) s)
  "
  by(simp only: rec_tryR_decomposition_NN_3 bindS_assoc)

lemma rec_tryR_transformation_NN: 
  "(rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2))\<Down>NN
  =
  ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N 
    repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S (h1\<Down>NN))) \<guillemotright>N (h2\<Down>NN))
  "
  by(simp add: rec_tryR_transformation_NN_unf project_defs)

subsubsection {* Exceptional termination *}

lemma rec_tryRS_eq_NE:
  "((Some (Inl e), s') \<in> rec_tryR m s) = ((e, s') \<in> (repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>NE)))"
  by(auto simp add: rec_tryRS_eq project_defs unfold_monad)

lemma rec_tryR_decomposition_NE_1: "(Some (Inl e), s') \<in> rec_tryR (h1 \<guillemotright> h2) s = 
  ((e, s') \<in> (repeatS (h1\<Down>R [+] (h1\<Down>NN \<guillemotright>N h2\<Down>R)) s) \<guillemotright>=S (h1\<Down>NE [+] (h1\<Down>NN \<guillemotright>N h2\<Down>NE)))"
  by(simp add: rec_tryRS_eq_NE bind_projectNE bind_projectR)

lemma rec_tryR_decomposition_NE_2: "(Some (Inl e), s') \<in> rec_tryR (h1 \<guillemotright> h2) s =
  ((e, s') \<in> (
  (
    (repeatS (h1\<Down>R)
    [+] (
        (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) 
        \<guillemotright>N repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN))
        \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S repeatS (h1\<Down>R))
        )
    )
  s) \<guillemotright>=S (h1\<Down>NE [+] (h1\<Down>NN \<guillemotright>N h2\<Down>NE))))"
  by(simp only: rec_tryR_decomposition_NE_1 repeatS_choice_rewrite)

lemma choice_bindS_bindN:
  "(a [+] b) s \<guillemotright>=S c = ((\<lambda>s. a s \<guillemotright>=S c) [+] (\<lambda>s. b s \<guillemotright>=S c)) s"
  by(auto simp add: unfold_monad)

ML_val {*
  @{term "(repeatS m s) \<guillemotright>= m'"}
*}

lemma rec_tryR_transformation_NE_unf:
"(Some (Inl e), s') \<in> rec_tryR (h1 \<guillemotright> h2) s =
  ((e, s') \<in> (
    (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NE)
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N h2\<Down>NE)
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) 
        \<guillemotright>N repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN))
        \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NE))
        )
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) 
        \<guillemotright>N repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN))
        \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N h2\<Down>NE)
        )
    ) s
  )"
  apply(simp only: rec_tryR_decomposition_NE_2)
  apply(rule eqset_imp_iff)
  apply(simp only: choice_bindS_bindN)
  apply(simp add: choice_distrib)
  apply(simp add: choice_def Un_assoc Un_commute)
  apply(rule arg_cong2[where f="(op Un)"])
  apply(simp)
  apply(rule arg_cong2[where f="(op Un)"])
  apply(simp add: bindS_bindN)
  apply(rule arg_cong2[where f="(op Un)"])
  apply(force simp add: unfold_monad)+
  done

lemma rec_tryR_transformation_NE:
  "(rec_tryR (h1 \<guillemotright> h2))\<Down>NE = 
    ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NE)
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N h2\<Down>NE )
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) 
        \<guillemotright>N repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN))
        \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NE))
        )
    [+] ((\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) 
        \<guillemotright>N repeatN (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN))
        \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (\<lambda>s. repeatS (h1\<Down>R) s \<guillemotright>=S h1\<Down>NN) \<guillemotright>N h2\<Down>NE)
        )
    )"
  by(auto simp add: rec_tryR_transformation_NE_unf projectNE_def)

definition projectN2 :: "('a, 's)r_monad \<Rightarrow> ('a, 's)r_monad" ("(_\<Down>N2)" [1000] 999) where
  "projectN2 m \<equiv> \<lambda>s. {(Some r, s')|r s'. (Some r, s') \<in> m s}"

definition projectR2 :: "('a, 'e, 's)re_monad \<Rightarrow> (unit, 'e, 's)re_monad" ("(_\<Down>R2)" [1000] 999) where
  "projectR2 m \<equiv> \<lambda>s. {(Some (Inr ()), s')|s'. (None, s') \<in> m s}"

lemma try_catchR_projectN2:
  "(try_catchR m1 m2)\<Down>N2 = m1\<Down>N2 [+] ((m1\<Down>R2) \<guillemotright> m2\<Down>N2)"
  apply(rule ext)
  apply(clarsimp simp add: projectR2_def projectN2_def try_catchR_some_decompose)
  apply(auto simp add: unfold_monad split: sum.split sum.split_asm option.split option.split_asm)
  done

lemma projectN2_N:
  "(m\<Down>N2) = ((m\<Down>N) \<guillemotright>=N (returnN o Some))"
  by(auto simp add: projectN2_def projectN_def unfold_monad)

lemma projectR2_R:
  "(m\<Down>R2) = (\<lambda>s. (m\<Down>R) s \<guillemotright>=S (return ()))"
  by(auto simp add: projectR2_def projectR_def unfold_monad)

lemma projectN_NN_NE:
  "(m\<Down>N) = (((m\<Down>NN) \<guillemotright>=N (returnN o Inr)) [+] ((m\<Down>NE) \<guillemotright>=N (returnN o Inl)))"
  apply(auto simp add: project_defs unfold_monad intro!: ext)
  by (metis sumE)

lemma returnN_Inl_bind:
  "(m \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x))))) \<guillemotright>= m' = (m \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x)))))"
  by(auto simp add: unfold_monad intro!: ext)

lemma returnN_Inr_seq:
  "(m \<guillemotright>=N (\<lambda>x. returnN (Some (Inr (f x))))) \<guillemotright> m' = (m \<guillemotright>N m')"
  by(auto simp add: unfold_monad intro!: ext)

lemma rec_tryR_projectN:
  "((rec_tryR m)\<Down>N) = (\<lambda>s. repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>N))"
  apply(subst projectN_def)
  by(auto intro!: ext simp add: rec_tryRS_eq)

lemma rec_tryR_projectNN:
  "((rec_tryR m)\<Down>NN) = (\<lambda>s. repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>NN))"
  apply(subst projectNN_def)
  by(auto intro!: ext simp add: rec_tryRS_eq_NN)

lemma rec_tryR_projectNE:
  "((rec_tryR m)\<Down>NE) = (\<lambda>s. repeatS (m\<Down>R) s \<guillemotright>=S (m\<Down>NE))"
  apply(subst projectNE_def)
  by(auto intro!: ext simp add: rec_tryRS_eq_NE)

lemma push_Inr_to_front: 
  "(m' \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x))))) [+] (m \<guillemotright>=N (\<lambda>x. returnN (Some (Inr (f x)))))
      = (m \<guillemotright>=N (\<lambda>x. returnN (Some (Inr (f x))))) [+] (m' \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x)))))"
  by(simp add: choice_comm)

lemma push_Inr_to_front_left: 
  "(m' \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x))))) [+] (m \<guillemotright>=N (\<lambda>x. returnN (Some (Inr (g x)))))
      [+] m''
      = (m \<guillemotright>=N (\<lambda>x. returnN (Some (Inr (g x))))) [+] (m' \<guillemotright>=N (\<lambda>x. returnN (Some (Inl (f x)))))
      [+] m''"
  by(simp add: choice_left_commute)

lemma bindN_bind_assoc:
  "m \<guillemotright>N (f \<guillemotright>= g) = (m \<guillemotright>N f) \<guillemotright>= g"
  by(auto simp add: unfold_monad)

lemma bindN_bindS_return_seq:
  "(m \<guillemotright>N (\<lambda>s. m' s \<guillemotright>=S return ())) \<guillemotright> m''
    = (m \<guillemotright>N (\<lambda>s. m' s \<guillemotright>=S m''))"
  apply(auto simp add: unfold_monad split: option.split sum.split option.split_asm sum.split_asm 
    intro!: ext)
  apply (metis (no_types) UN_iff split_conv)
  apply(rule_tac x="(aa, ba)" in bexI)
  apply(auto)
  done

lemma bindS_bindN: "(m1 s \<guillemotright>=S (m2 \<guillemotright>=N m3)) = ((\<lambda>s. (m1 s \<guillemotright>=S m2)) \<guillemotright>=N m3) s"
  by(auto simp add: unfold_monad)

lemma repeatN_N_into_unbounded:
  "x \<in> repeatN_N mc n s \<Longrightarrow> x \<in> repeatN mc s"
  apply(induct n arbitrary: x)
  apply(simp_all only: repeatN_N_alter)
  apply(auto simp add: unfold_monad intro: repeatN.skip)
  apply(subgoal_tac "((), ba) \<in> repeatN mc s")
  apply(auto intro: repeatN.repeat)
  done

lemma repeatN_rewrite:
  "(x \<in> repeatN mc s) = (\<exists>n. x \<in> repeatN_N mc n s)"
  apply(case_tac x)
  apply(auto intro: repeatN_bound repeatN_N_into_unbounded)
  done

declare disjCI[rule del]

lemma repeatN_alt:
  "repeatN m = returnN () [+] (m \<guillemotright>N repeatN m)"
  apply(auto intro!: ext)
  apply(erule repeatN.induct)
  apply(auto simp add: unfold_monad intro: repeatN.skip repeatN_absorb')
  apply (metis mem_splitI repeatN.skip)
  apply (metis mem_splitI repeatN.repeat)
  done

lemma rec_tryR_NN_alt:
  "
  (rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2))\<Down>NN =
    ((rec_tryR h1)\<Down>NN \<guillemotright>N h2\<Down>NN) [+]
    ((rec_tryR h1)\<Down>NN \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (rec_tryR (h1 \<guillemotright> h2))\<Down>NN))
  "

  apply(auto simp add: rec_tryR_transformation_NN rec_tryR_projectNN[THEN sym]
    choice_comm choice_left_commute del: subsetI)
  apply(simp only: bindN_choice[THEN sym])
  apply(simp add: bindN_assoc)
  apply(rule arg_cong2[OF refl, where f=bindN])
  apply(intro ext)
  apply(simp add: bindS_bindN)
  apply(subst repeatN_alt)
  apply(simp add: choice_distrib bindN_assoc)
  done

lemma rec_tryR_NE_alt:
  "
    (rec_tryR ((h1 :: (unit, 'e, 's)re_monad) \<guillemotright> h2))\<Down>NE =
    ((rec_tryR h1)\<Down>NE [+] ((rec_tryR h1)\<Down>NN \<guillemotright>N h2\<Down>NE)) [+]
    ((rec_tryR h1)\<Down>NN \<guillemotright>N (\<lambda>s. (h2\<Down>R) s \<guillemotright>=S (rec_tryR (h1 \<guillemotright> h2))\<Down>NE))
  "

  apply(auto simp add: rec_tryR_transformation_NE rec_tryR_projectNN[THEN sym]
    rec_tryR_projectNE[THEN sym]
    choice_comm choice_left_commute del: subsetI)
  apply(simp only: choice_assoc)
  apply(rule arg_cong2[OF refl, where f=choice])
  apply(simp only: bindN_assoc choice_assoc)
  apply(simp only: bindN_choice[THEN sym])
  apply(rule arg_cong2[OF refl, where f=bindN])
  apply(rule ext)
  apply(rule arg_cong2[OF refl, where f=choice])
  apply(subst repeatN_alt)
  apply(simp add: choice_bindN)
  apply(auto simp add: choice_distrib bindN_assoc bindN_bindS
    intro!: ext del: UnCI)
  apply(auto simp add: choice_def)
  apply (metis RMonad_Lemmas.bindS_bindN)+
  done

lemma tryR_catchR_rec_tryR_transformation:
  "((<tryR> m <catchR> (rec_tryR ((h1 :: (unit, 'e, 's)re_monad)\<guillemotright> h2)))\<Down>N2) =
    (m\<Down>N2)
    [+] (
        ((m\<Down>R2) \<guillemotright> ((rec_tryR h1)\<Down>N2))
        \<guillemotright>
        ((<tryR> h2 <catchR> (rec_tryR (h1 \<guillemotright> h2)))\<Down>N2)
    )
  "
proof-

  have expand:
    "((rec_tryR (h1 \<guillemotright>= (\<lambda>_. h2)))\<Down>N \<guillemotright>=N returnN \<circ> Some) =
     ((rec_tryR h1)\<Down>N \<guillemotright>=N returnN \<circ> Some \<guillemotright>=
         (\<lambda>_. (h2\<Down>N \<guillemotright>=N returnN \<circ> Some) [+]
              ((\<lambda>s. (h2\<Down>R) s \<guillemotright>=S return ()) \<guillemotright>=
               (\<lambda>_. (rec_tryR (h1 \<guillemotright>= (\<lambda>_. h2)))\<Down>N \<guillemotright>=N returnN \<circ> Some))))"
    apply(auto simp add: projectN_NN_NE choice_distrib choice_assoc choice_comm choice_left_commute
      bindN_assoc bindN_left_unit returnN_Inl_bind returnN_Inr_seq
      choice_absorb choice_left_absorb ext del: equalityI)
    apply(simp only: bindN_assoc[THEN sym] bindN_bind_assoc bindN_bindS_return_seq bindS_bindN)
    apply(simp only: push_Inr_to_front push_Inr_to_front_left)
    apply(simp only: choice_assoc[THEN sym] choice_bindN[THEN sym])
    apply(auto intro!: arg_cong2[where f=choice] arg_cong2[where f=bindN] rec_tryR_NN_alt rec_tryR_NE_alt
      simp add: bindN_bindS_return_seq)
    done

  show ?thesis
    apply(auto simp add: try_catchR_projectN2 bind_assoc intro!: choice_left_I arg_cong2[where f="bind"])
    apply(clarsimp simp add: projectN2_N projectR2_R intro!: expand intro!: ext)
    done
qed


end

