theory Abs1_Invariant_Lemmas
imports  Abs1_Specification Abs1_StartupDefs "../HoareLogic" "../Pmm_Tacs"
begin

lemmas brute_force = 
  unfold_monad
  project_defs
  getters_setters
  modify_def
  put_def
  restart_mangle_sv_def
  choose_def

lemma pub_ll_invariants:
  "\<lbrace>abs1_init\<rbrace> ll_read ind inst \<lbrace>> K abs1_init \<rbrace>"
  "\<lbrace>abs1_init\<rbrace> write_f ind inst cnt \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: Hoare_def Box_def)
  apply(auto simp add: ll_read_def write_f_def project_simps choice_distrib)
  apply(simp_all add: choice_def, safe)
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: brute_force ll_read_defs write_f_defs 
    abs1_init_def
    split: split_if_asm)
  apply(case_tac[!] inst, clarsimp_all)
  apply (smt INST.exhaust abs1_state.select_convs(1) abs1_state.select_convs(2) 
    abs1_state.select_convs(3) abs1_state.select_convs(4) abs1_state.surjective 
    abs1_state.update_convs(9) hash_by_inst.simps(1) hash_by_inst.simps(2) 
    log_state_by_inst.simps(1) log_state_by_inst.simps(2))+
  done

lemma write_ptr_ll_fail_instance_inv:
  "\<lbrace>abs1_init\<rbrace> write_ptr_ll_fail_instance inst cnt \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: Hoare_def Box_def)
  apply(case_tac inst)
  apply(clarsimp_all simp add: brute_force write_ptr_ll_defs
    min_def abs1_init_def
    split: split_if_asm)
  apply(case_tac[!] inst, simp_all)
  apply (metis hash_by_inst.simps(1) log_state_by_inst.simps(1),
    metis hash_by_inst.simps(2) log_state_by_inst.simps(2))+
  done

lemma ptr_ll_invariant1:
  "\<lbrace>abs1_init\<rbrace> write_ptr_ll inst cnt \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: write_ptr_ll_def choice_inv K_beta intro!: 
    write_ptr_ll_fail_instance_inv try_catch_inv_same bind_inv_same)
  apply(tactic {* distinct_subgoals_tac *})
  apply(case_tac[!] inst)
  apply(auto simp add: Hoare_def Box_def K_def)
  apply(clarsimp_all simp add: brute_force write_ptr_ll_defs ll_read_ptr_defs 
    abs1_init_def
    split: split_if_asm)
  apply (smt INST.exhaust abs1_state.ext_inject abs1_state.surjective abs1_state.update_convs(5) 
    hash_by_inst.simps(1) hash_by_inst.simps(2) log_state_by_inst.simps(1) log_state_by_inst.simps(2))
  apply (smt abs1_state.ext_inject abs1_state.surjective abs1_state.update_convs(7) 
    hash_by_inst.cases hash_by_inst.simps(1) hash_by_inst.simps(2) log_state_by_inst.simps(1) 
    log_state_by_inst.simps(2))
  apply(case_tac[!] inst, simp_all)
  apply (metis hash_by_inst.simps(1) log_state_by_inst.simps(1),
    metis hash_by_inst.simps(2) log_state_by_inst.simps(2))+
  done

lemma ptr_ll_invariant2:
  "\<lbrace>abs1_init\<rbrace> ll_read_ptr inst \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: ll_read_ptr_def choice_inv K_beta intro!: 
    write_ptr_ll_fail_instance_inv try_catch_inv_same bind_inv_same)
  apply(tactic {* distinct_subgoals_tac *})
  apply(case_tac[!] inst)
  apply(auto simp add: Hoare_def Box_def K_def)
  apply(clarsimp_all simp add: brute_force ll_read_ptr_defs 
    abs1_init_def
    split: split_if_asm)
  apply(case_tac[!] inst, simp_all)
  apply (metis hash_by_inst.simps(1) log_state_by_inst.simps(1),
    metis hash_by_inst.simps(2) log_state_by_inst.simps(2))+
  done

lemmas ptr_ll_invariants =
  ptr_ll_invariant1
  ptr_ll_invariant2

lemma hinst_ll_invariant1:
  "\<lbrace>abs1_init\<rbrace> ll_read_hinst inst \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: ll_read_hinst_def choice_inv K_beta intro!: try_catch_inv_same bind_inv_same)
  apply(case_tac[!] inst)
  apply(auto simp add: Hoare_def Box_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: brute_force ll_read_hinst_defs 
    abs1_init_def min_def
    split: split_if_asm)
  apply(case_tac[!] inst, clarsimp_all)
  apply (metis hash_by_inst.simps(2) log_state_by_inst.simps(2) | 
    metis hash_by_inst.simps(1) log_state_by_inst.simps(1))+
  done

definition ht_preserves_init where
  "ht_preserves_init ht \<equiv> {(sv, sp).
    (abs1_state.ptr1_state sp = IUnwr \<longrightarrow> ptr_hashes2 ht Primary \<noteq> abs1_state.ptr1 sp)
    \<and> (abs1_state.ptr2_state sp = IUnwr \<longrightarrow> ptr_hashes2 ht Secondary \<noteq> abs1_state.ptr2 sp)
  }"

lemma hinst_ll_invariant2:
  "\<lbrace>abs1_init \<inter> ht_preserves_init ht\<rbrace> write_hinst inst ht \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: write_hinst_def choice_inv K_beta intro!: try_catch_inv_same bind_inv_same)
  apply(case_tac[!] inst)
  apply(auto simp add: Hoare_def Box_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: brute_force write_hinst_defs 
    ht_preserves_init_def abs1_init_def min_def
    split: split_if_asm)
  apply(case_tac[!] inst, clarsimp_all)
  apply (metis hash_by_inst.simps(2) log_state_by_inst.simps(2) | 
    metis hash_by_inst.simps(1) log_state_by_inst.simps(1))+
  done

lemmas hinst_ll_invariants =
  hinst_ll_invariant1
  hinst_ll_invariant2

lemma get_hash_table_ht_preserves:
  "\<lbrace>abs1_init\<rbrace> get_hash_table inst \<lbrace>>N ht_preserves_init \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(case_tac inst)
  apply(auto simp add: get_hash_table_def is_good_hash_def ll_read_hinst_defs brute_force
    abs1_init_def ht_preserves_init_def min_def
    split: split_if_asm)
  apply (metis hash_by_inst.simps(2) log_state_by_inst.simps(2) | 
    metis hash_by_inst.simps(1) log_state_by_inst.simps(1))+
  done

lemma get_hash_table_init:
  "\<lbrace>abs1_init\<rbrace> get_hash_table inst \<lbrace>> K abs1_init \<rbrace>"
   apply(auto simp add: get_hash_table_def intro!: bind_inv_same hinst_ll_invariant1 
     throw_rule' return_rule')
  apply(auto simp add: Hoare_def Box_def is_good_hash_def brute_force)
  done

(* TODO: move these to HoareLogic *)
lemma Hoare_imp_HoareNN:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>>N K Q\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def K_beta project_defs)

lemma Hoare_imp_HoareNE:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@N K Q\<rbrace>"
  by(auto simp add: HoareNE_alt Hoare_def Box_def K_beta project_defs)

lemma Hoare_imp_HoareR:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@R Q\<rbrace>"
  by(auto simp add: HoareR_def BoxR_def Hoare_def Box_def K_beta project_defs)

lemma get_hash_table_or_fail_full_NN:
  "\<lbrace>abs1_init\<rbrace> get_hash_table_or_fail inst \<lbrace>>N \<lambda>ht. abs1_init \<inter> ht_preserves_init ht \<rbrace>"
  apply(rule HoareNN_join_post)
  apply(auto simp add: get_hash_table_or_fail_def K_beta
    intro!: try_catch_ruleNN_same get_hash_table_ht_preserves
    bind_ruleNN_same
    HoareNN_throw
    get_hash_table_init[THEN Hoare_imp_HoareNN, simplified K_def]
    get_hash_table_init[THEN Hoare_imp_HoareNE]
  )
  apply(auto simp add: HoareNN_alt Hoare_def Box_def abs1_init_def brute_force set_failed_def)
  done

lemma vol_ops_init:
  "\<lbrace>abs1_init\<rbrace> set_failed \<lbrace>> K abs1_init\<rbrace>"
  "\<lbrace>abs1_init\<rbrace> set_degraded_h \<lbrace>> K abs1_init\<rbrace>"
  "\<lbrace>abs1_init\<rbrace> set_degraded_p \<lbrace>> K abs1_init\<rbrace>"
  "\<lbrace>abs1_init\<rbrace> set_degraded \<lbrace>> K abs1_init\<rbrace>"
  "\<lbrace>abs1_init\<rbrace> get_mode \<lbrace>> K abs1_init\<rbrace>"
  "\<lbrace>abs1_init\<rbrace> init_vars \<lbrace>> K abs1_init\<rbrace>"
  by(clarsimp_all simp add: Hoare_def Box_def brute_force abs1_init_def vol_ops_defs)

lemma vol_ops_HoareNE:
  "\<lbrace>P\<rbrace> set_failed \<lbrace>@N Q\<rbrace>"
  "\<lbrace>P\<rbrace> set_degraded \<lbrace>@N Q\<rbrace>"
  "\<lbrace>P\<rbrace> set_degraded_h \<lbrace>@N Q\<rbrace>"
  "\<lbrace>P\<rbrace> get_mode \<lbrace>@N Q\<rbrace>"
  by(clarsimp_all simp add: HoareNE_alt Hoare_def Box_def brute_force abs1_init_def vol_ops_defs)
  
lemma vol_ops_HoareR:
  "\<lbrace>P\<rbrace> set_failed \<lbrace>@R Q\<rbrace>"
  "\<lbrace>P\<rbrace> set_degraded \<lbrace>@R Q\<rbrace>"
  "\<lbrace>P\<rbrace> set_degraded_h \<lbrace>@R Q\<rbrace>"
  "\<lbrace>P\<rbrace> get_mode \<lbrace>@R Q\<rbrace>"
  by(clarsimp_all simp add: HoareR_def BoxR_def brute_force abs1_init_def vol_ops_defs)

lemma bind_ruleR:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>@R Q\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>>N R\<rbrace>; \<forall>x. \<lbrace>R x\<rbrace> f x \<lbrace>@R Q\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<guillemotright>= f \<lbrace>@R Q\<rbrace>"
  apply(auto simp add: HoareR_def BoxR_def HoareNN_alt Hoare_def Box_def brute_force)
  apply(force split: option.split_asm sum.split_asm)
  done

lemma try_catch_ruleR:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>@R Q\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>@N R\<rbrace>; \<forall>x. \<lbrace>R x\<rbrace> h x \<lbrace>@R Q\<rbrace> \<rbrakk> \<Longrightarrow> 
    \<lbrace>P\<rbrace> try_catch m h \<lbrace>@R Q\<rbrace>"
  apply(auto simp add: HoareR_def BoxR_def HoareNE_alt Hoare_def Box_def brute_force)
  apply(force split: option.split_asm sum.split_asm)
  done

lemma HoareR_throw:
  "\<lbrace>P\<rbrace> throw e \<lbrace>@R Q\<rbrace>"
  by(auto simp add: HoareR_def BoxR_def throw_def)

lemma HoareR_return:
  "\<lbrace>P\<rbrace> return x \<lbrace>@R Q\<rbrace>"
  by(auto simp add: HoareR_def BoxR_def return_def)

lemma get_hashes_init:
  "\<lbrace>abs1_init\<rbrace> get_hashes \<lbrace>> K abs1_init \<rbrace>"
  apply(auto simp add: get_hashes_def K_beta
    intro!: 
    try_catch_inv_same get_hash_table_init
    throw_rule' return_rule'
    split: exceptions.split)
  apply(auto simp add: K_beta get_hash_table_or_fail_def
    intro!: bind_inv_same 
    try_catch_inv_same get_hash_table_init
    throw_rule' return_rule'
    vol_ops_init
    split: exceptions.split)[1]
  apply(auto simp add: Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(intro bind_ruleNN[OF get_hash_table_or_fail_full_NN] allI)
  apply(rule bind_ruleNN)
  apply(rule try_catch_ruleNN)
  apply(rule hinst_ll_invariant2[THEN Hoare_imp_HoareNN])
  apply(rule hinst_ll_invariant2[THEN Hoare_imp_HoareNE])
  apply(auto simp add: K_beta HoareNN_return' intro!: Hoare_imp_HoareNN bind_inv_same vol_ops_init 
    throw_rule')[2]
  apply(rule bind_ruleNE)
  apply(auto simp add: K_beta get_hash_table_or_fail_def HoareNE_throw'
    intro!: try_catch_ruleNE bind_ruleNE_same
    get_hash_table_init[THEN Hoare_imp_HoareNE]
    throw_rule' return_rule'
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareNE, simplified K_def]
    split: exceptions.split)[1]
  apply(rule get_hash_table_or_fail_full_NN)
  apply(auto)
  apply(rule bind_ruleNE)
  apply(rule try_catch_ruleNE)
  apply(rule hinst_ll_invariant2[THEN Hoare_imp_HoareNE])
  apply(auto simp add: K_beta HoareNE_throw'
    intro!: bind_ruleNE_same
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareNE, simplified K_def]
    split: exceptions.split)[1]
  apply(rule try_catch_ruleNN)
  apply(rule hinst_ll_invariant2[THEN Hoare_imp_HoareNN])
  apply(rule hinst_ll_invariant2[THEN Hoare_imp_HoareNE])
  apply(auto simp add: K_beta HoareNE_throw'
    intro!: bind_ruleNN_same HoareNN_throw HoareNE_return
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareNE, simplified K_def]
    split: exceptions.split)[2]
  apply(auto simp add: K_beta HoareR_return HoareR_throw
    intro!: 
    get_hash_table_or_fail_full_NN 
    bind_ruleR try_catch_ruleR try_catch_ruleNN bind_ruleNN_same HoareNN_throw
    throw_rule' return_rule'
    hinst_ll_invariant2[THEN Hoare_imp_HoareNN]
    hinst_ll_invariant2[THEN Hoare_imp_HoareNE]
    hinst_ll_invariant2[THEN Hoare_imp_HoareR]
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareR]
    vol_ops_init[THEN Hoare_imp_HoareNE, simplified K_def]
    split: exceptions.split)[1]
  apply(auto simp add: K_beta HoareR_return HoareR_throw get_hash_table_or_fail_def
    intro!: 
    bind_ruleR try_catch_ruleR bind_ruleNN
    get_hash_table_init[THEN Hoare_imp_HoareR]
    get_hash_table_init[THEN Hoare_imp_HoareNE]
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareR]
    split: exceptions.split)[1]
  done
 
lemma hinst_ll_invariant3:
  "\<lbrace>abs1_init \<inter> ht_preserves_init ht\<rbrace> write_hinst inst ht \<lbrace>> K (ht_preserves_init ht) \<rbrace>"
  apply(auto simp add: write_hinst_def choice_inv K_beta intro!: try_catch_inv_same bind_inv_same)
  apply(case_tac[!] inst)
  apply(auto simp add: Hoare_def Box_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: brute_force write_hinst_defs 
    ht_preserves_init_def abs1_init_def min_def
    split: split_if_asm)
  done

(* TODO: move to HoareLogic *)
lemma Hoare_join_post:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>> Q\<rbrace>; \<lbrace>P\<rbrace> m \<lbrace>> Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>> \<lambda>x. Q x \<inter> Q' x\<rbrace>"
  by(auto simp add: Hoare_def Box_def)

lemma Hoare_join_post':
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>> Q\<rbrace>; \<lbrace>P'\<rbrace> m \<lbrace>> Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P \<inter> P'\<rbrace> m \<lbrace>> \<lambda>x. Q x \<inter> Q' x\<rbrace>"
  by(auto simp add: Hoare_def Box_def)

lemma hinst_ll_invariant4:
  "\<lbrace>abs1_init \<inter> ht_preserves_init ht\<rbrace> write_hinst inst ht \<lbrace>> K (abs1_init \<inter> ht_preserves_init ht) \<rbrace>"
  by(auto simp add: K_def intro!: Hoare_join_post
    hinst_ll_invariant2[simplified K_def] hinst_ll_invariant3[simplified K_def])

lemma get_hashes_ht_preserves_NN:
  "\<lbrace>abs1_init\<rbrace> get_hashes \<lbrace>>N ht_preserves_init \<rbrace>"
  apply(unfold get_hashes_def)
  apply(rule try_catch_ruleNN[OF get_hash_table_ht_preserves get_hash_table_init[THEN Hoare_imp_HoareNE]])
  apply(auto simp add: K_beta split: exceptions.split intro!: HoareNN_throw)
  apply(auto intro!: bind_ruleNN HoareNN_throw intro: true_postNN)[1]
  apply(auto simp add: K_beta HoareNN_return'
    intro!: 
    HoareNN_throw
    bind_ruleNN[OF get_hash_table_or_fail_full_NN] bind_ruleNN try_catch_ruleNN
    hinst_ll_invariant3[THEN Hoare_imp_HoareNN]
    hinst_ll_invariant2[THEN Hoare_imp_HoareNE]
    intro: true_postNN
  )
  done

(* TODO: replace the HoareLogic version with this *)
lemma Hoare_all_Hoare:
  "\<lbrace>I\<rbrace> m \<lbrace>& J\<rbrace> = \<lbrace>I\<rbrace> m \<lbrace>> K J\<rbrace>"
  apply(auto simp add: K_def Hoare_all_def Hoare_defs Box_def Box_defs
    split: option.split option.split_asm sum.split sum.split_asm)
  apply(case_tac a)
  apply(auto)
  apply(case_tac aa)
  apply(auto)
  done

lemma get_hashes_full_NN:
  "\<lbrace>abs1_init\<rbrace> get_hashes \<lbrace>>N \<lambda>ht. abs1_init \<inter> ht_preserves_init ht \<rbrace>"
  apply(rule HoareNN_join_post[OF _ get_hashes_ht_preserves_NN])
  apply(intro Hoare_imp_HoareNN[OF get_hashes_init, simplified K_def])
  done

thm bind_inv_same

lemma bind_inv_same':
  "\<lbrakk>\<lbrace>P\<rbrace> m1 \<lbrace>> K P\<rbrace>; \<forall>x. \<lbrace>P\<rbrace> m2 x \<lbrace>> K Q\<rbrace>; P \<subseteq> Q\<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m1 \<guillemotright>= m2 \<lbrace>> K Q\<rbrace>"
  apply(auto simp add: Hoare_def Box_def unfold_monad split: option.split_asm sum.split_asm)
  apply(force dest: set_rev_mp)
  apply(force dest: set_rev_mp)
  apply(force dest!: set_mp)
  done

lemma get_hash_table_preserves_ht_preserves:
  "\<lbrace>abs1_init \<inter> ht_preserves_init ht\<rbrace> get_hash_table inst \<lbrace>> K (abs1_init \<inter> ht_preserves_init ht) \<rbrace>"
  apply(simp only: K_def, rule Hoare_join_post)
  apply(rule conseq_rule[OF _ get_hash_table_init[simplified K_def] _])
  apply(auto)
  apply(auto simp add: Hoare_def Box_def)
  apply(clarsimp simp add: Hoare_def Box_def get_hash_table_def brute_force ll_read_hinst_defs 
    is_good_hash_def
    ht_preserves_init_def
    min_def
    split: split_if_asm
  )
  apply(iterative_simp)+
  done

lemma write_hashes_init:
  "\<lbrace>abs1_init \<inter> ht_preserves_init ht\<rbrace> write_hashes ht \<lbrace>> K abs1_init\<rbrace>"
  apply(unfold write_hashes_def)
  apply(rule bind_inv_same'[OF _ _ Int_lower1])
  apply(auto 
    intro!: try_catch_inv_same bind_inv_same
    hinst_ll_invariant4 allI
    throw_rule'
    split: exceptions.split
  )[1]
  apply(simp add: get_hash_table_or_fail_def)
  apply(rule try_catch_inv_same[OF get_hash_table_preserves_ht_preserves])
  apply(clarsimp simp add: Hoare_def Box_def brute_force set_failed_def 
    ht_preserves_init_def abs1_init_def)
  apply(clarsimp simp add: Hoare_def Box_def brute_force set_degraded_h_def
    ht_preserves_init_def abs1_init_def)
  apply(auto
    intro!: try_catch_inv_same[OF hinst_ll_invariant2]
    throw_rule'
    split: exceptions.split 
  )
  apply(clarsimp simp add: Hoare_def Box_def brute_force set_degraded_h_def
    ht_preserves_init_def abs1_init_def)
  done

lemma update_htable2_alt:
  "pub_update_htable2 ht pind inst cnt = 
        HTable2 ((pub_hashes2 ht)((pind, inst) := Hash cnt)) (ptr_hashes2 ht)"
  "ptr_update_htable2 ht inst ptr_set = 
        HTable2 (pub_hashes2 ht) ((ptr_hashes2 ht)(inst := ptr_set))"
  by(case_tac[!] ht, auto)

lemma update_pub_instance_hash_init:
  "\<lbrace>abs1_init\<rbrace> update_pub_instance_hash ind inst cnt \<lbrace>> K abs1_init \<rbrace>"
  apply(unfold update_pub_instance_hash_def)
  apply(auto simp add: Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(intro bind_ruleNN[OF get_hashes_full_NN] allI)
  apply(rule conseqNN_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareNN]])
  apply(clarsimp simp add: ht_preserves_init_def update_htable2_alt)
  apply(simp_all add: K_beta)
  apply(rule bind_ruleNE)
  apply(rule get_hashes_init[THEN Hoare_imp_HoareNE, simplified K_def])
  apply(rule get_hashes_full_NN)
  apply(intro allI)
  apply(rule conseqNE_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareNE]])
  apply(clarsimp simp add: ht_preserves_init_def update_htable2_alt)
  apply(simp add: K_beta)
  apply(rule bind_ruleR)
  apply(rule get_hashes_init[THEN Hoare_imp_HoareR, simplified K_def])
  apply(rule get_hashes_full_NN)
  apply(intro allI)
  apply(rule conseqR_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareR] order_refl])
  apply(clarsimp simp add: ht_preserves_init_def update_htable2_alt)
  done

definition ptr_not_IUnwr where
  "ptr_not_IUnwr inst \<equiv> {(sv, sp). ptr_state_by_inst inst sp \<noteq> IUnwr }"

lemma get_hash_table_preserves_ptr_state:
  "\<lbrace>ptr_not_IUnwr pinst\<rbrace> get_hash_table inst \<lbrace>> K (ptr_not_IUnwr pinst)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_def)
  apply(simp add: get_hash_table_def ll_read_hinst_def first_stage)
  apply(auto simp add: choice_def)
  apply(case_tac[!] pinst)
  apply(auto simp add: brute_force ll_read_hinst_defs is_good_hash_def min_def ptr_not_IUnwr_def
    split: split_if_asm exceptions.splits)
  done

lemma vol_ops_preserve_ptr_state:
  "\<lbrace>ptr_not_IUnwr inst\<rbrace> set_failed \<lbrace>> K (ptr_not_IUnwr inst)\<rbrace>"
  "\<lbrace>ptr_not_IUnwr inst\<rbrace> set_degraded_h \<lbrace>> K (ptr_not_IUnwr inst)\<rbrace>"
  "\<lbrace>ptr_not_IUnwr inst\<rbrace> set_failed \<lbrace>> K (ptr_not_IUnwr inst)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_def)
  apply(auto simp add: brute_force ll_read_hinst_defs is_good_hash_def min_def ptr_not_IUnwr_def
    vol_ops_defs
    split: split_if_asm exceptions.splits)
  done

lemma write_hinst_preserves_ptr_state:
  "\<lbrace>ptr_not_IUnwr pinst\<rbrace> write_hinst inst cnt \<lbrace>> K (ptr_not_IUnwr pinst)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_def)
  apply(case_tac pinst)
  apply(auto simp add: brute_force is_good_hash_def min_def ptr_not_IUnwr_def
    write_hinst_defs
    vol_ops_defs
    split: split_if_asm exceptions.splits)
  done

lemma get_hashes_preserves_ptr_state:
  "\<lbrace>ptr_not_IUnwr inst\<rbrace> get_hashes \<lbrace>> K (ptr_not_IUnwr inst) \<rbrace>"
  by(auto simp add: K_beta get_hashes_def get_hash_table_or_fail_def
    intro!: try_catch_inv_same bind_inv_same throw_rule' return_rule'
    get_hash_table_preserves_ptr_state vol_ops_preserve_ptr_state write_hinst_preserves_ptr_state
    split: exceptions.splits)

lemma HoareNN_join_post':
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>>N Q\<rbrace>; \<lbrace>P'\<rbrace> m \<lbrace>>N Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P \<inter> P'\<rbrace> m \<lbrace>>N \<lambda>x. Q x \<inter> Q' x\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def)

lemma update_ptr_instance_hash_init:
  "\<lbrace>abs1_init \<inter> ptr_not_IUnwr inst\<rbrace> update_ptr_instance_hash inst cnt \<lbrace>> K abs1_init\<rbrace>"
  apply(unfold update_ptr_instance_hash_def)
  apply(auto simp add: Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(rule bind_ruleNN)
  apply(rule HoareNN_join_post'[OF get_hashes_full_NN 
    Hoare_imp_HoareNN[OF get_hashes_preserves_ptr_state]])
  apply(clarsimp simp add: K_beta)
  apply(rule conseqNN_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareNN]])
  apply(force simp add: ht_preserves_init_def ptr_not_IUnwr_def update_htable2_alt)
  apply(simp add: K_beta)
  apply(rule bind_ruleNE)
  apply(rule conseqNE_rule[OF _ get_hashes_init[THEN Hoare_imp_HoareNE] _])
  apply(auto simp add: K_beta)
  apply(rule HoareNN_join_post'[OF get_hashes_full_NN 
    Hoare_imp_HoareNN[OF get_hashes_preserves_ptr_state]])
  apply(clarsimp simp add: K_beta)
  apply(rule conseqNE_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareNE]])
  apply(force simp add: ht_preserves_init_def ptr_not_IUnwr_def update_htable2_alt)
  apply(auto simp add: K_beta)
  apply(rule bind_ruleR)
  apply(rule conseqR_rule[OF _ get_hashes_init[THEN Hoare_imp_HoareR]])
  apply(auto)
  apply(rule HoareNN_join_post'[OF get_hashes_full_NN 
    Hoare_imp_HoareNN[OF get_hashes_preserves_ptr_state]])
  apply(clarsimp simp add: K_beta)
  apply(rule conseqR_rule[OF _ write_hashes_init[THEN Hoare_imp_HoareR]])
  apply(force simp add: ht_preserves_init_def ptr_not_IUnwr_def update_htable2_alt)
  apply(simp)
  done

lemma write_ptr_ll_post_writable_NN:
  "\<lbrace>abs1_init\<rbrace> write_ptr_ll inst cnt \<lbrace>>N K (abs1_init \<inter> ptr_not_IUnwr inst) \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp add: write_ptr_ll_def first_stage)
  apply(auto simp add: choice_def)
  apply(auto simp add: write_ptr_ll_defs brute_force abs1_init_def ptr_not_IUnwr_def
    split: split_if_asm exceptions.splits)
  apply (smt abs1_state.ext_inject abs1_state.surjective abs1_state.update_convs(5) 
    hash_by_inst.cases hash_by_inst.simps(1) hash_by_inst.simps(2) log_state_by_inst.simps(1) 
    log_state_by_inst.simps(2))
  by (smt INST.exhaust abs1_state.ext_inject abs1_state.surjective abs1_state.update_convs(7) 
    hash_by_inst.simps(1) hash_by_inst.simps(2) log_state_by_inst.simps(1) log_state_by_inst.simps(2))

lemma fold_K: "(\<lambda>_. P) = K P"
  by(auto simp add: K_def)

lemma read_ptr_DegradedH_init:
  "\<lbrace>abs1_init\<rbrace> read_ptr_DegradedH inst \<lbrace>> K abs1_init\<rbrace>"
  by(auto simp add: read_ptr_DegradedH_def get_hashes_DegradedH_def K_beta
    get_hash_table_or_fail_def
    intro!: bind_inv_same try_catch_inv_same ptr_ll_invariant2 vol_ops_init throw_rule' return_rule'
    get_hash_table_init)

lemma write_ptr_instance_init:
  "\<lbrace>abs1_init\<rbrace> write_ptr_instance inst cnt \<lbrace>> K abs1_init\<rbrace>"
  apply(auto simp add: write_ptr_instance_def Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(simp_all only: fold_K[where P=abs1_init])
  apply(auto simp add: write_ptr_instance_def HoareNN_return' HoareNE_throw' K_beta 
    intro!: 
    bind_ruleR bind_ruleNN bind_ruleNE 
    try_catch_ruleNN try_catch_ruleNE try_catch_ruleR
    HoareNN_throw HoareNE_return HoareR_throw HoareR_return
    write_ptr_ll_post_writable_NN
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_init[THEN Hoare_imp_HoareNE]
    vol_ops_init[THEN Hoare_imp_HoareR]
    update_ptr_instance_hash_init[THEN Hoare_imp_HoareNN]
    update_ptr_instance_hash_init[THEN Hoare_imp_HoareNE]
    update_ptr_instance_hash_init[THEN Hoare_imp_HoareR]
    read_ptr_DegradedH_init[THEN Hoare_imp_HoareNN]
    read_ptr_DegradedH_init[THEN Hoare_imp_HoareNE]
    read_ptr_DegradedH_init[THEN Hoare_imp_HoareR]
    ptr_ll_invariant1[THEN Hoare_imp_HoareNE]
    ptr_ll_invariant1[THEN Hoare_imp_HoareR]
    split: exceptions.splits
  )
  done

lemma sync_hashes_init:
  "\<lbrace>abs1_init\<rbrace> sync_hashes Secondary Primary \<lbrace>> K abs1_init\<rbrace>"
  apply(auto simp add: check_hash_region_def Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(simp_all only: fold_K[where P=abs1_init])
  apply(auto simp add: sync_hashes_def K_beta HoareNE_throw'
    intro!: 
    bind_ruleNN bind_ruleNE bind_ruleR
    try_catch_ruleNN try_catch_ruleNE try_catch_ruleR
    HoareNN_throw HoareR_throw
    HoareNN_join_post[OF get_hash_table_init[THEN Hoare_imp_HoareNN] get_hash_table_ht_preserves]
    get_hash_table_init[THEN Hoare_imp_HoareNE]
    get_hash_table_init[THEN Hoare_imp_HoareR]
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_HoareNE
    vol_ops_HoareR
    hinst_ll_invariant2[THEN Hoare_imp_HoareNN]
    hinst_ll_invariant2[THEN Hoare_imp_HoareNE]
    hinst_ll_invariant2[THEN Hoare_imp_HoareR]
  )
  done

lemma HoareR_choice:
  "\<lbrace>P\<rbrace> m1 [+] m2 \<lbrace>@R Q\<rbrace> = (\<lbrace>P\<rbrace> m1 \<lbrace>@R Q\<rbrace> \<and> \<lbrace>P\<rbrace> m2 \<lbrace>@R Q\<rbrace>)"
  by(auto simp add: HoareR_def BoxR_choice)

lemma check_hash_region_init:
  "\<lbrace>abs1_init\<rbrace> check_hash_region \<lbrace>> K abs1_init\<rbrace>"
  apply(auto simp add: check_hash_region_def Hoare_all_Hoare[THEN sym] Hoare_all_def)
  apply(simp_all only: fold_K[where P=abs1_init])
  apply(force simp add: K_beta HoareNN_choice HoareNE_choice HoareR_choice
    HoareNN_return' HoareNE_throw'
    fold_K[where P=abs1_init]
    intro!: 
    bind_ruleNN bind_ruleNE bind_ruleR
    try_catch_ruleNN try_catch_ruleNE try_catch_ruleR
    HoareNN_throw HoareNE_return HoareR_throw HoareR_return
    hinst_ll_invariant2[THEN Hoare_imp_HoareNN]
    hinst_ll_invariant2[THEN Hoare_imp_HoareNE]
    hinst_ll_invariant2[THEN Hoare_imp_HoareR]
    get_hash_table_preserves_ht_preserves[THEN Hoare_imp_HoareNN]
    get_hash_table_preserves_ht_preserves[THEN Hoare_imp_HoareNE]
    get_hash_table_init[THEN Hoare_imp_HoareNE]
    get_hash_table_init[THEN Hoare_imp_HoareR]
    get_hash_table_init[THEN Hoare_imp_HoareR, THEN conseqR_rule[OF Int_lower1], where S="abs1_init"]
    HoareNN_join_post[OF get_hash_table_init[THEN Hoare_imp_HoareNN] get_hash_table_ht_preserves]
    vol_ops_init[THEN Hoare_imp_HoareNN]
    vol_ops_HoareNE
    vol_ops_HoareR
    vol_ops_init[THEN Hoare_imp_HoareNN, THEN conseqNN_rule[OF Int_lower1], where S="K abs1_init", 
      simplified]
    sync_hashes_init[THEN Hoare_imp_HoareNN]
    sync_hashes_init[THEN Hoare_imp_HoareNE]
    sync_hashes_init[THEN Hoare_imp_HoareR]
    split: exceptions.splits
    intro: get_hash_table_init[THEN Hoare_imp_HoareNN]
  )+
  done

lemma startup_init:
  "\<lbrace>abs1_init\<rbrace> startup \<lbrace>> K abs1_init\<rbrace>"
  apply(auto simp add: K_beta Let_def
    startup_def startup1_def startup2_def check_ptr_def read_ptr_def get_hashes_DegradedH_def
    get_hash_table_or_fail_def just_sync_ptr_def sync_ptr_def read_ptr_or_fail_def
    read_ptr_or_fail_silence_deg_def check_all_public_def check_pub_region_def throw_mode_error_def
    get_sync_direction_def get_hashes_silence_deg_def get_ptr_def
    sync_and_remove_from_ptr_silence_deg_def sync_def read_or_fail_def read_def
    write_instance_def read_DegradedH_def remove_from_ptr_silence_deg_def check_pub_same_def
    remove_from_ptr_def just_sync_silence_deg_def just_sync_def write_ptr_def
    read_or_fail_silence_deg_def
    intro!: 
    bind_inv_same
    try_catch_inv_same
    return_rule'
    throw_rule'
    mapM'_rule
    vol_ops_init
    check_hash_region_init
    ptr_ll_invariant2
    ptr_ll_invariant1
    pub_ll_invariants
    get_hashes_init
    get_hash_table_init
    write_ptr_instance_init
    chooseN_inv
    update_pub_instance_hash_init
    split: exceptions.splits
  )
  apply(erule_tac[!] contrapos_np)
  apply(auto simp add: K_beta
    intro!: 
    vol_ops_init
    ptr_ll_invariant1
    bind_inv_same try_catch_inv_same return_rule' throw_rule')
  done

end