theory Abs1_LLInterface
  imports Abs1Defs
begin

definition write_success :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_success ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        size cnt \<le> capacity (Public ind)
        \<and> hw_state sp ind inst = IOK
        \<and> sp' = sp\<lparr> memory := fun_upd_second (memory sp) ind inst cnt \<rparr>
    }
  "

definition doesnt_match_hashes where
  "doesnt_match_hashes sp ind inst cnt \<equiv>
    (hash1_state sp \<noteq> EcOK \<or> Hash cnt \<noteq> pub_hashes2 (hash1 sp) (ind, inst))
    \<and> (hash2_state sp \<noteq> EcOK \<or> Hash cnt \<noteq> pub_hashes2 (hash2 sp) (ind, inst))"

definition write_fail_instance ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "write_fail_instance ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailWrite), sv, sp')|sp' cnt'. 
        size cnt \<le> capacity (Public ind)
        \<and> cnt' \<noteq> cnt
        \<and> cnt' \<noteq> memory sp ind inst
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> doesnt_match_hashes sp ind inst cnt'
        \<and> sp' = sp\<lparr> 
            memory := fun_upd_second (memory sp) ind inst cnt', 
            hw_state := fun_upd_second (hw_state sp) ind inst
                            (min (hw_state sp ind inst) IUnwr)
            \<rparr>
    }"

definition write_fail_readonly ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "write_fail_readonly ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(Some (Inl FailWrite), sv, sp')|sp'.
        size cnt \<le> capacity (Public ind)
        \<and> cnt \<noteq> memory sp ind inst
        \<and> sp' = sp\<lparr> hw_state := fun_upd_second (hw_state sp) ind inst
                            (min (hw_state sp ind inst) IUnwr)
        \<rparr>
  }"

definition write_corrupt_restart ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv \<times> 'c abs1_state)r_monad" where
  "write_corrupt_restart ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(None, sv', sp')|sv' sp' cnt'. 
        size cnt \<le> capacity (Public ind)
        \<and> cnt' \<noteq> cnt
        \<and> cnt' \<noteq> memory sp ind inst
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> doesnt_match_hashes sp ind inst cnt'
        \<and> sp' = sp\<lparr> 
            memory := fun_upd_second (memory sp) ind inst cnt'
        \<rparr>
    }"

definition write_too_large
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "write_too_large ind inst cnt \<equiv> \<lambda>(sv, sp). 
    {(Some (Inl InvalidInput), sv, sp')|sp'. size cnt > capacity (Public ind) \<and> sp' = sp }"

definition write_f :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_f ind inst cnt \<equiv> write_success ind inst cnt
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_fail_instance ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_fail_readonly ind inst cnt <catch> K restart_mangle_sv)
  "

lemmas write_f_defs = 
    write_f_def
    write_success_def
    write_fail_instance_def
    write_corrupt_restart_def
    write_fail_readonly_def
    write_too_large_def
    doesnt_match_hashes_def

definition write_hinst_success :: "INST \<Rightarrow> 'c htable2 \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_hinst_success inst ht \<equiv> \<lambda>(sv, sp). 
    {(Some (Inr ()), sv, sp')|sp'.
        (inst = Primary \<longrightarrow> EcBad \<le> hash1_state sp)
        \<and> (inst = Secondary \<longrightarrow> EcBad \<le> hash2_state sp)
        \<and> (if inst = Primary 
        then
            sp' = sp \<lparr>
                hash1_state := EcOK,
                hash1 := ht
            \<rparr>
        else
            sp' = sp \<lparr>
                hash2_state := EcOK,
                hash2 := ht
            \<rparr>
        )
    }
  "

definition write_hinst_fail_instance :: "INST \<Rightarrow> 'c htable2 \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_hinst_fail_instance inst ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailWrite), sv, sp')|sp' ht'. 
        sp' = (if inst = Primary
            then
                sp\<lparr> 
                    hash1_state := min (hash1_state sp) EcUnwr,
                    hash1 := ht'
                \<rparr>
            else
                sp\<lparr> 
                    hash2_state := min (hash2_state sp) EcUnwr,
                    hash2 := ht'
                \<rparr>
        )

    }"

definition write_hinst_corrupt_restart :: "INST \<Rightarrow> 'c htable2 \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_hinst_corrupt_restart inst ht \<equiv> \<lambda>(sv, sp). 
    {(None, sv', sp')|sv' sp' ht'. 
        sp' = (if inst = Primary
            then sp\<lparr> 
                hash1_state := min (hash1_state sp) EcBad,
                hash1 := ht'
            \<rparr>
            else sp\<lparr> 
                hash2_state := min (hash2_state sp) EcBad,
                hash2 := ht'
            \<rparr>
        )
    }"

definition write_hinst ::
  "INST \<Rightarrow> 'c htable2 \<Rightarrow> (unit, 'sv, 'c)abs1_monad" where
  "write_hinst inst ht \<equiv> write_hinst_success inst ht
    [+] write_hinst_fail_instance inst ht
    [+] write_hinst_corrupt_restart inst ht
    [+] restart_mangle_sv
    [+] do { write_hinst_success inst ht; restart_mangle_sv }
    [+] (<try> write_hinst_fail_instance inst ht <catch> K restart_mangle_sv)
  "

lemmas write_hinst_defs = 
    write_hinst_def
    write_hinst_success_def
    write_hinst_fail_instance_def
    write_hinst_corrupt_restart_def

definition write_ptr_ll_success :: "INST \<Rightarrow> pub_indx set \<Rightarrow> 
  (unit, 'sv, 'c)abs1_monad" where
  "write_ptr_ll_success inst ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        ptr_state_by_inst inst sp = IOK
        \<and> sp' = (if inst = Primary
            then sp \<lparr>
                    ptr1 := ptr_set
                \<rparr>
            else sp \<lparr>
                    ptr2 := ptr_set
                \<rparr>
        )
  }"

definition write_ptr_ll_fail_instance ::
  "INST \<Rightarrow> pub_indx set \<Rightarrow> ('a, 'sv, 'c :: serialization_cl)abs1_monad" where
  "write_ptr_ll_fail_instance inst ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailWrite), sv, sp')|sp' ptr_set'. 
        ptr_set' \<noteq> ptr_set
        \<and> ptr_set' \<noteq> ptr_by_inst inst sp
        \<and> (hash1_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash1 sp) inst)
        \<and> (hash2_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash2 sp) inst)
        \<and> sp' = (if inst = Primary
            then sp\<lparr> 
                ptr1 := ptr_set',
                ptr1_state := min (ptr1_state sp) IUnwr
                \<rparr>
            else sp\<lparr> 
                ptr2 := ptr_set',
                ptr2_state := min (ptr2_state sp) IUnwr
                \<rparr>
        )
    }"

definition write_ptr_ll_corrupt_restart ::
  "INST \<Rightarrow> pub_indx set \<Rightarrow> ('a, 'sv \<times> ('c :: serialization_cl) abs1_state)r_monad" where
  "write_ptr_ll_corrupt_restart inst ptr_set \<equiv> \<lambda>(sv, sp). 
    {(None, sv', sp')|sv' sp' ptr_set'. 
        ptr_set' \<noteq> ptr_set
        \<and> ptr_set' \<noteq> ptr_by_inst inst sp
        \<and> (hash1_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash1 sp) inst)
        \<and> (hash2_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash2 sp) inst)
        \<and> sp' = (if inst = Primary
            then sp\<lparr> 
                ptr1 := ptr_set'
            \<rparr>
            else sp\<lparr> 
                ptr2 := ptr_set'
            \<rparr>
        )
    }"

definition write_ptr_ll ::
  "INST \<Rightarrow> pub_indx set \<Rightarrow> (unit, 'sv, 'c :: serialization_cl)abs1_monad" where
  "write_ptr_ll inst cnt \<equiv> write_ptr_ll_success inst cnt
    [+] write_ptr_ll_fail_instance inst cnt
    [+] write_ptr_ll_corrupt_restart inst cnt
    [+] restart_mangle_sv
    [+] do { write_ptr_ll_success inst cnt; restart_mangle_sv }
    [+] (<try> write_ptr_ll_fail_instance inst cnt <catch> K restart_mangle_sv)
  "

lemmas write_ptr_ll_defs = 
    write_ptr_ll_def
    write_ptr_ll_success_def
    write_ptr_ll_fail_instance_def
    write_ptr_ll_corrupt_restart_def

definition ll_read_failure ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "ll_read_failure ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailLLRead), 
        (sv, sp\<lparr> 
            hw_state := fun_upd_second (hw_state sp) ind inst IDead
        \<rparr>)
    )}
  "

definition ll_read_any ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, 'sv, 'c)abs1_monad" where
  "ll_read_any ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr cnt), sv, sp')|cnt sp'.
        hw_state sp ind inst \<noteq> IDead
        \<and> (cnt = memory sp ind inst \<or> 
            (doesnt_match_hashes sp ind inst cnt \<and> size cnt \<le> capacity (Public ind)))
        \<and> sp' = sp\<lparr> memory := fun_upd_second (memory sp) ind inst cnt \<rparr>
    }
  "

definition ll_read ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, 'sv, 'c)abs1_monad" where
  "ll_read ind inst \<equiv>
    ll_read_failure ind inst 
    [+] ll_read_any ind inst
    [+] restart_mangle_sv
    [+] (<try> ll_read_failure ind inst <catch> K restart_mangle_sv)
    [+] do { ll_read_any ind inst; restart_mangle_sv }
  "

lemmas ll_read_defs = 
  ll_read_def 
  ll_read_failure_def 
  ll_read_any_def
  doesnt_match_hashes_def

definition ll_read_success_hinst :: 
  "INST \<Rightarrow> ('c htable2, 'sv, 'c)abs1_monad" where
  "ll_read_success_hinst inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr (hash_by_inst inst sp)), sv, sp')|sp'. EcDead < log_state_by_inst inst sp \<and> sp' = sp}
  "

definition ll_read_failure_hinst ::
  "INST \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "ll_read_failure_hinst inst \<equiv> \<lambda>(sv, sp).
    if inst = Primary 
    then 
        {(Some (Inl FailLLRead), 
                (sv, sp\<lparr> 
                    hash1_state := EcDead
                \<rparr>)
            )}
    else
        {(Some (Inl FailLLRead), 
                (sv, sp\<lparr> 
                    hash2_state := EcDead
                \<rparr>)
            )}

  "

definition ll_read_corrupt_hinst ::
  "INST \<Rightarrow> ('c htable2, 'sv, 'c)abs1_monad" where
  "ll_read_corrupt_hinst inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ht'), sv, sp')|ht' sp'.
        EcDead < log_state_by_inst inst sp
        \<and> sp' = (if inst = Primary
            then sp\<lparr> 
                hash1 := ht',
                hash1_state := min EcBad (hash1_state sp)
            \<rparr>
            else sp\<lparr> 
                hash2 := ht',
                hash2_state := min EcBad (hash2_state sp)
            \<rparr>
        )
    }
  "

definition ll_read_hinst ::
  "INST \<Rightarrow> ('c htable2, 'sv, 'c)abs1_monad" where
  "ll_read_hinst inst \<equiv>
    ll_read_success_hinst inst 
    [+] ll_read_failure_hinst inst 
    [+] ll_read_corrupt_hinst inst
    [+] restart_mangle_sv
    [+] do { ll_read_success_hinst inst; restart_mangle_sv }
    [+] (<try> ll_read_failure_hinst inst <catch> K restart_mangle_sv)
    [+] do { ll_read_corrupt_hinst inst; restart_mangle_sv }
  "

lemmas ll_read_hinst_defs = 
  ll_read_hinst_def 
  ll_read_success_hinst_def 
  ll_read_failure_hinst_def 
  ll_read_corrupt_hinst_def


definition ll_read_ptr_failure ::
  "INST \<Rightarrow> ('a, 'sv, 'c)abs1_monad" where
  "ll_read_ptr_failure inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailLLRead), sv,
            (if inst = Primary
                then sp\<lparr> ptr1_state := IDead \<rparr>
                else sp\<lparr> ptr2_state := IDead \<rparr>
            )
    )}
  "

definition ll_read_ptr_any ::
  "INST \<Rightarrow> (pub_indx set, 'sv, 'c :: serialization_cl)abs1_monad" where
  "ll_read_ptr_any inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ptr_set'), sv, sp')|ptr_set' sp'.
        ptr_state_by_inst inst sp \<noteq> IDead
        \<and> (ptr_set' = ptr_by_inst inst sp
            \<or> (
                (hash1_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash1 sp) inst)
                \<and> (hash2_state sp \<noteq> EcOK \<or> ptr_set' \<noteq> ptr_hashes2 (hash2 sp) inst)
            )
        )
        \<and> sp' = (if inst = Primary
                    then sp\<lparr> ptr1 := ptr_set' \<rparr>
                    else sp\<lparr> ptr2 := ptr_set' \<rparr>
            )
    }
  "

definition ll_read_ptr ::
  "INST \<Rightarrow> (pub_indx set, 'sv, 'c :: serialization_cl)abs1_monad" where
  "ll_read_ptr inst \<equiv>
    ll_read_ptr_failure inst 
    [+] ll_read_ptr_any inst
    [+] restart_mangle_sv
    [+] (<try> ll_read_ptr_failure inst <catch> K restart_mangle_sv)
    [+] do { ll_read_ptr_any inst; restart_mangle_sv }
  "

lemmas ll_read_ptr_defs = 
  ll_read_ptr_def 
  ll_read_ptr_failure_def 
  ll_read_ptr_any_def


end
