theory Abs1_Ref
  imports "../concrete/ConcreteDefs" Abs1Defs "../Refinement" 
  "../Pmm_Refinement"
  Abs1_LLInterface Abs1_LLInterface_Lemmas
  "../concrete/LLInterface" "../concrete/LLInterface_Lemmas"
  "../concrete/StartupDefs" Abs1_StartupDefs
  "../concrete/UpdateDefs" Abs1_UpdateDefs
  "../concrete/FetchDefs" Abs1_FetchDefs
  "../Pmm_Tacs"
  "../Pmm_Lemmas"
begin

fun ec_state_by_flags :: "bool \<Rightarrow> hw_state \<Rightarrow> ec_state" where
  "ec_state_by_flags _ IDead = EcDead"
  | "ec_state_by_flags _ IUnwr = EcUnwr"
  | "ec_state_by_flags False _ = EcBad"
  | "ec_state_by_flags True _ = EcOK"

lemma ec_state_by_flags_rev_cases:
  "(ec_state_by_flags f s = EcOK) = (f \<and> s = IOK)"
  "(EcOK = ec_state_by_flags f s) = (f \<and> s = IOK)"
  "(ec_state_by_flags f s = EcBad) = (\<not>f \<and> s = IOK)"
  "(EcBad = ec_state_by_flags f s) = (\<not>f \<and> s = IOK)"
  "(ec_state_by_flags f s = EcDead) = (s = IDead)"
  "(EcDead = ec_state_by_flags f s) = (s = IDead)"
  "(ec_state_by_flags f s = EcUnwr) = (s = IUnwr)"
  "(EcUnwr = ec_state_by_flags f s) = (s = IUnwr)"
  by(case_tac[!] f, case_tac[!] s, simp_all)

fun htable2_to_htable :: "'c htable2 \<Rightarrow> 'c :: serialization_cl htable" where
  "htable2_to_htable (HTable2 pub_h ptr_h) =
    HTable pub_h (\<lambda>inst. Hash (serialize_ptr (ptr_h inst)))"

definition abs1_conc_strel :: "(('sv \<times> 'c :: contents_cl abs1_state) \<times> ('sv \<times> 'c state))set" 
  where
  "abs1_conc_strel \<equiv> {((sva, spa), (svc, spc)). 
    sva = svc
    \<and> (\<forall>pind. Abs1Defs.hw_state spa pind = ConcreteDefs.hw_state spc (Public pind))
    \<and> (\<forall>pind. Abs1Defs.memory spa pind = ConcreteDefs.memory spc (Public pind))

    \<and> ConcreteDefs.memory spc (Admin PTR_REG) Primary = serialize_ptr (Abs1Defs.ptr1 spa)
    \<and> ConcreteDefs.memory spc (Admin PTR_REG) Secondary = serialize_ptr (Abs1Defs.ptr2 spa)
    \<and> Abs1Defs.ptr1_state spa = hw_state spc (Admin PTR_REG) Primary
    \<and> Abs1Defs.ptr2_state spa = hw_state spc (Admin PTR_REG) Secondary
    \<and> Abs1Defs.hash1_state spa =
        ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Primary)) 
                (hw_state spc (Admin HASH_REG) Primary)
    \<and> Abs1Defs.hash2_state spa =
        ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Secondary)) 
                (hw_state spc (Admin HASH_REG) Secondary)
    \<and> (ok_hash_reg_read spc Primary \<longrightarrow> 
        state.memory spc (Admin HASH_REG) Primary =
        htable_to_hinst_cnt (htable2_to_htable (Abs1Defs.hash1 spa)))
    \<and> (ok_hash_reg_read spc Secondary \<longrightarrow> 
        state.memory spc (Admin HASH_REG) Secondary =
        htable_to_hinst_cnt (htable2_to_htable (Abs1Defs.hash2 spa)))
    \<and> (\<forall>inst.
        (hw_state spc (Admin HASH_REG) inst = IUnwr \<longrightarrow> 
            \<not>ok_hash_reg_read spc inst
        )
    )
    \<and> (\<forall>ind inst. size (memory spc ind inst) \<le> capacity ind)
  }"

lemmas ref_rel_defs =
  abs1_conc_strel_def
  written_hash_def

lemmas brute_force =
  unfold_monad
  project_defs
  force_same_and_def
  LLInterface.ll_read_def 
  LLInterface.ll_read_success_def 
  LLInterface.ll_read_failure_def 
  LLInterface.ll_read_corrupt_def
  LLInterface.write_f_defs
  Abs1_LLInterface.ll_read_def 
  Abs1_LLInterface.ll_read_failure_def 
  Abs1_LLInterface.ll_read_any_def
  Abs1_LLInterface.ll_read_hinst_defs
  Abs1_LLInterface.ll_read_ptr_defs
  Abs1_LLInterface.write_f_defs
  Abs1_LLInterface.write_hinst_defs
  Abs1_LLInterface.write_ptr_ll_defs
  restart_mangle_sv_raw
  Let_def

lemma restart_mangle_sv_refines_itself[ref_thm]:
  "sim abs1_conc_strel 
    restart_mangle_sv
    restart_mangle_sv
  "
  apply(auto simp only: sim_def sim_ext_def)
  apply(auto simp add: abs1_conc_strel_def brute_force simN_wpc_def simS_wpc_def intro!: ext)
  done

lemma restart_mangle_sv_sim_ext:
  "sim_ext abs1_conc_strel
    restart_mangle_sv
    restart_mangle_sv
    Sn Se abs1_conc_strel"
  apply(auto simp only: sim_ext_def)
  apply(auto simp add: abs1_conc_strel_def brute_force simN_wpc_def simS_wpc_def intro!: ext)
  done

lemma restart_mangle_sv_simS:
  "abs1_conc_strel \<subseteq> simS_wpc
    (restart_mangle_sv\<Down>R)
    (restart_mangle_sv\<Down>R)
    abs1_conc_strel
  "
  apply(auto simp add: abs1_conc_strel_def brute_force simN_wpc_def simS_wpc_def intro!: ext)
  done

lemma no_concrete_hinst_readonly_fails:
  "(LLInterface.write_fail_readonly (Admin HASH_REG) inst cnt) = empty"
  by(auto simp add: empty_def K_def write_fail_readonly_def intro!: ext)

lemma choice_left_eqI:
  "ml = mr \<Longrightarrow> (m [+] ml) = (m [+] mr)"
  by(auto simp add: choice_def)

lemma corrupt_restart_hinst_bind:
  "(write_hinst_corrupt_restart inst ht \<guillemotright>= f) = write_hinst_corrupt_restart inst ht"
  apply(clarsimp simp add: write_hinst_defs unfold_monad min_def intro!: ext)
  apply(safe)
  apply(case_tac[!] b)
  apply(auto split: option.split_asm sum.split_asm option.split sum.split)
  done

lemma write_hinst_success_seq_is_seqR:
  "(write_hinst_success inst ht \<guillemotright> m) = (write_hinst_success inst ht \<guillemotright>R m)"
  apply(rule ext)
  apply(clarsimp simp add: unfold_monad write_hinst_defs split: option.split sum.split)
  apply(safe)
  apply(case_tac[!] b)
  apply(auto split: option.split_asm sum.split_asm option.split sum.split)
  done

lemma restart_mangle_sv_bindR:
  "(restart_mangle_sv \<guillemotright>=R f) = restart_mangle_sv"
  by(auto simp add: brute_force intro!: ext)

lemma abs1_write_hinst_alt:
  "Abs1_LLInterface.write_hinst inst ht =
    Abs1_LLInterface.write_hinst_success inst ht
    [+] Abs1_LLInterface.write_hinst_fail_instance inst ht
    [+] Abs1_LLInterface.write_hinst_corrupt_restart inst ht
    [+] empty
    [+] empty
    [+] restart_mangle_sv
    [+] do { Abs1_LLInterface.write_hinst_success inst ht; restart_mangle_sv }
    [+] (<try> Abs1_LLInterface.write_hinst_fail_instance inst ht <catch> K restart_mangle_sv)
    [+] empty
  "
  apply(auto simp only: Abs1_LLInterface.write_hinst_def first_stage choice_comm 
    corrupt_restart_hinst_bind
    choice_left_commute choice_left_absorb)
  done

lemma forall_INST:
  "(\<forall>(x :: INST). P x) = (P Primary \<and> P Secondary)"
  apply(auto)
  apply(case_tac x, auto)
  done

lemma written_hash_inv:
  "state.memory spc' (Admin HASH_REG) Primary = state.memory spc (Admin HASH_REG) Primary \<Longrightarrow>
    written_hash spc' Primary = written_hash spc Primary"
  "state.memory spc' (Admin HASH_REG) Secondary = state.memory spc (Admin HASH_REG) Secondary \<Longrightarrow>
    written_hash spc' Secondary = written_hash spc Secondary"
  by(auto simp add: written_hash_def)

lemma written_hash_inv_hw_state:
  "written_hash (spc\<lparr> hw_state := f \<rparr>) = written_hash spc"
  by(rule ext, simp add: written_hash_def)

lemma written_hash_hash_mem_upd:
  "written_hash (s\<lparr>
        state.memory := fun_upd_second (state.memory s) (Admin HASH_REG) Primary x \<rparr>
        ) Secondary 
    = written_hash s Secondary"
  "written_hash (s\<lparr>
        state.memory := fun_upd_second (state.memory s) (Admin HASH_REG) Secondary x \<rparr>
        ) Secondary 
    = (parse_htable o fst o parse_hinst) x"
  "written_hash (s\<lparr>
        state.memory := fun_upd_second (state.memory s) (Admin HASH_REG) Secondary x \<rparr>
        ) Primary 
    = written_hash s Primary"
  "written_hash (s\<lparr>
        state.memory := fun_upd_second (state.memory s) (Admin HASH_REG) Primary x \<rparr>
        ) Primary 
    = (parse_htable o fst o parse_hinst) x"
  by(simp_all add: written_hash_def)

lemma written_hash_of_htable_to_hinst_cnt:
  "state.memory spc (Admin HASH_REG) inst = htable_to_hinst_cnt ht
    \<Longrightarrow> written_hash spc inst = ht"
  by(auto simp add: written_hash_def parse_written_ht)

lemma inj_htable2_to_htable:
  "inj htable2_to_htable"
  apply(rule injI)
  apply(case_tac x, case_tac y, simp)
  apply(auto intro!: ext)
  apply (metis hash.inject parse_serialize_ptr)+
  done

lemma write_f_hash_ref:
  "sim abs1_conc_strel 
        (Abs1_LLInterface.write_hinst inst ht)
        (LLInterface.write_f (Admin HASH_REG) inst (htable_to_hinst_cnt (htable2_to_htable ht)))"
  apply(auto simp only: LLInterface_Lemmas.write_f_alt abs1_write_hinst_alt
    first_stage restart_mangle_sv_bindR
    intro!: sims_all_intros restart_mangle_sv_refines_itself
    simp del: choice_emptyl choice_emptyr
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: simN_wpc_def simS_wpc_def abs1_conc_strel_def
    brute_force ec_state_simps written_hash_hash_mem_upd ok_hash_reg_read_def
    size_serialize_hashes htable_size
    htable_size[THEN linorder_class.leD]
    ec_state_by_flags_rev_cases
    written_hash_of_htable_to_hinst_cnt
    written_hash_inv_hw_state
    inj_eq[OF inj_htable2_to_htable]
    simp del: written_hash_def
  )
  apply(case_tac[!] inst)
  apply(safe)
  apply(clarsimp_all simp add: ec_state_simps ec_state_by_flags_rev_cases 
    parse_hinst_htable_to_hinst_cnt written_hash_def
    parse_serialize_htable
    hash_of_parse_written_ht written_hash_of_htable_to_hinst_cnt)
  apply(case_tac[!] "hw_state ba (Admin HASH_REG) Secondary")
  apply(clarsimp_all simp add: min_def ec_state_by_flags_rev_cases ec_state_simps)
  apply(case_tac[!] "hw_state ba (Admin HASH_REG) Primary")
  apply(clarsimp_all simp add: min_def ec_state_by_flags_rev_cases ec_state_simps)
  apply (metis hash_of_parse_written_ht parse_hinst_htable_to_hinst_cnt)+
  done

lemma write_f_hash_ref'[ref_thm]:
  "(htable_to_hinst_cnt (htable2_to_htable ht) = cnt) \<Longrightarrow>
    sim abs1_conc_strel 
        (Abs1_LLInterface.write_hinst inst ht)
        (LLInterface.write_f (Admin HASH_REG) inst cnt)"
  by(auto intro: write_f_hash_ref)

lemma corrupt_restart_ptr_ll_bind:
  "(write_ptr_ll_corrupt_restart inst ptr_set \<guillemotright>= f) = write_ptr_ll_corrupt_restart inst ptr_set"
  apply(clarsimp simp add: write_ptr_ll_defs unfold_monad min_def intro!: ext)
  apply(rule equalityI)
  apply(case_tac[!] inst)
  apply(clarsimp_all)
  apply(case_tac[!] b)
  apply(clarsimp_all)
  done

lemma write_ptr_ll_success_seq_is_seqR:
  "(write_ptr_ll_success inst ht \<guillemotright> m) = (write_ptr_ll_success inst ht \<guillemotright>R m)"
  apply(rule ext)
  apply(clarsimp simp add: unfold_monad write_ptr_ll_defs split: option.split sum.split)
  apply(safe)
  apply(case_tac[!] b)
  apply(auto split: option.split_asm sum.split_asm option.split sum.split)
  done

lemma abs1_write_ptr_ll_alt:
  "Abs1_LLInterface.write_ptr_ll inst ht =
    do { Abs1_LLInterface.write_ptr_ll_success inst ht; restart_mangle_sv }
    [+] (<try> Abs1_LLInterface.write_ptr_ll_fail_instance inst ht <catch> K restart_mangle_sv)
    [+] do { Abs1_LLInterface.write_ptr_ll_corrupt_restart inst ht; restart_mangle_sv }
    [+] empty
    [+] empty
    [+] restart_mangle_sv
    [+] Abs1_LLInterface.write_ptr_ll_success inst ht
    [+] Abs1_LLInterface.write_ptr_ll_fail_instance inst ht
    [+] Abs1_LLInterface.write_ptr_ll_corrupt_restart inst ht
    [+] empty
    [+] restart_mangle_sv
  "
  apply(auto simp only: Abs1_LLInterface.write_ptr_ll_def first_stage choice_comm 
    corrupt_restart_ptr_ll_bind
    choice_left_commute choice_left_absorb)
  done

lemma inj_neq:
  "inj f \<Longrightarrow> (f x \<noteq> f y) = (x \<noteq> y)"
  by(auto dest: injD)

lemma no_concrete_ptr_ll_readonly_fails:
  "(LLInterface.write_fail_readonly (Admin PTR_REG) inst cnt) = empty"
  by(auto simp add: empty_def K_def write_fail_readonly_def intro!: ext)

lemma ht2_to_ht_ptr_eq:
  "htable2_to_htable f = g \<Longrightarrow> (x = ptr_hashes2 f inst) = 
  (Hash (serialize_ptr x) = ptr_hashes g inst)"
  by (metis (lifting, no_types) hash.inject htable2.exhaust htable2_to_htable.simps 
    parse_serialize_ptr ptr_hashes.simps ptr_hashes2.simps)
  
lemma write_f_ptr_ref[ref_thm]:
  "\<lbrakk> aind = (Admin PTR_REG); cnt = serialize_ptr ptr_set \<rbrakk> \<Longrightarrow>
  sim abs1_conc_strel 
        (Abs1_LLInterface.write_ptr_ll inst ptr_set)
        (LLInterface.write_f aind inst cnt)"
  apply(subst abs1_write_ptr_ll_alt)
  apply(auto simp only: LLInterface.write_f_def no_concrete_ptr_ll_readonly_fails 
    first_stage restart_mangle_sv_bindR
    intro!: sims_all_intros restart_mangle_sv_refines_itself
    simp del: choice_emptyl choice_emptyr
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: simN_wpc_def simS_wpc_def abs1_conc_strel_def
    brute_force ec_state_simps ok_hash_reg_read_def
    size_serialize_ptr ptr_size
    ptr_size[THEN linorder_class.leD]
    ec_state_by_flags_rev_cases
    inj_eq[OF inj_serialize_ptr]
    inj_neq[OF inj_serialize_ptr]
    (* written_hash_of_htable_to_hinst_cnt 
    simp del: written_hash_def *)
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: inj_neq[OF inj_serialize_ptr] ht2_to_ht_ptr_eq)
  apply (metis hw_state.distinct(3) parse_written_ht)+
  done

lemma ht2_to_ht_pub_eq:
  "htable2_to_htable f = g \<Longrightarrow> pub_hashes2 f  = pub_hashes g"
  by(case_tac f, case_tac g, simp)
 
lemma fun_upd_eq:
  "(f(x := y) = f(x := z)) = (y = z)"
  by(auto dest: fun_cong[where x=x])

lemma write_f_rest_ref[ref_thm]:
  "ind' = Public ind \<Longrightarrow>
  sim abs1_conc_strel 
        (Abs1_LLInterface.write_f ind inst cnt)
        (LLInterface.write_f ind' inst cnt)"
  apply(auto simp only: write_f_alt Abs1_LLInterface.write_f_def
    first_stage
    intro!: sims_all_intros restart_mangle_sv_refines_itself)
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp add: sim_def sim_ext_def)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs1_conc_strel_def simN_wpc_def simS_wpc_def
    brute_force 
    intro!: ext
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: min_def hw_state_simps ok_hash_reg_read_def ok_hash_cnt_def Let_def
    ht2_to_ht_pub_eq fun_upd_eq ec_state_by_flags_rev_cases)
  apply (metis hw_state.distinct(3) parse_written_ht)+
  done

lemma ll_read_pub_ref[ref_thm]:
  "ind' = Public ind \<Longrightarrow>
  sim abs1_conc_strel
    (Abs1_LLInterface.ll_read ind inst)
    (LLInterface.ll_read ind' inst)
  "
  apply(auto simp only: LLInterface_Lemmas.ll_read_alt Abs1_LLInterface_Lemmas.ll_read_alt first_stage 
    intro!: sims_all_intros restart_mangle_sv_refines_itself)
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def)
  using [[simp_depth_limit=2]]
  apply(auto simp add: abs1_conc_strel_def simN_wpc_def simS_wpc_def
    brute_force
    ok_hash_reg_read_def
    hw_state_simps
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: min_def hw_state_simps ok_hash_reg_read_def ok_hash_cnt_def Let_def
    ht2_to_ht_pub_eq fun_upd_eq ec_state_by_flags_rev_cases)
  apply (metis parse_written_ht)+
  done

lemmas for_auto_solve =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  LLInterface.ll_read_defs
  Abs1_LLInterface.ll_read_defs

lemma sim_ext_choice_lockstep:
  "\<lbrakk> sim_ext R ma1 mc1 Sn Se Sr; sim_ext R ma2 mc2 Sn Se Sr \<rbrakk> \<Longrightarrow>
    sim_ext R (ma1 [+] ma2) (mc1 [+] mc2) Sn Se Sr"
  by(auto simp only: sim_ext_def project_simps Int_subset_iff
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep del: subsetI)

lemma sim_ext_mono_exc:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; \<forall>x y. Se x y \<subseteq> Se' x y  \<rbrakk>
    \<Longrightarrow> sim_ext R ma mc Sn Se' Sr"
  apply(auto simp add: sim_ext_def del: subsetI)
  apply(rule subset_trans[where B="simN_wpc (ma\<Down>NE) (mc\<Down>NE) Se"])
  apply(auto intro!: simN_wpc_monotone_rel del: subsetI)
  done

lemma EcDead_bot: "EcDead = bot"
  by(simp)

lemma less_eq_EcDead: "(x \<le> EcDead) = (x = EcDead)"
  by(simp only: EcDead_bot bot_unique)

definition htable_from_cnt where
  "htable_from_cnt \<equiv> (parse_htable o fst o parse_hinst)"

lemma ll_read_hinst_sim_ext:
  "sim_ext abs1_conc_strel (ll_read_hinst inst) (ll_read (Admin HASH_REG) inst) 
    (\<lambda>va vc. (if vc = htable_to_hinst_cnt (htable2_to_htable va) \<or> \<not> ok_hash_cnt vc
        then UNIV else {}) \<inter> abs1_conc_strel)
    (force_same_and abs1_conc_strel) 
    abs1_conc_strel"
proof-
  {
    fix ma mc R
    assume 
    "sim_ext R ma mc
      (\<lambda>va vc. (if vc = htable_to_hinst_cnt (htable2_to_htable va) \<or> \<not> ok_hash_cnt vc
        then UNIV else {}) \<inter> R)
      (force_same_and R) 
      R"
    hence 
    "sim_ext R ma mc (\<lambda>va vc. R) (force_same_and R) R"
      by(rule sim_ext_mono_norm, auto)
  } note post_strenghten1=this

  {
    fix ma mc R
    assume 
    "sim_ext R ma mc
      (\<lambda>va vc. (if vc = htable_to_hinst_cnt (htable2_to_htable va) \<or> \<not> ok_hash_cnt vc
        then UNIV else {}) \<inter> R)
      (force_same_and R) 
      R"
    hence 
    "sim_ext R ma mc
      (\<lambda>va vc. (if vc = htable_to_hinst_cnt (htable2_to_htable va) \<or> \<not> ok_hash_cnt vc
        then UNIV else {}) \<inter> R)
      (\<lambda>ea ec. R) R"
      by(rule sim_ext_mono_exc, auto simp add: force_same_and_def)
  } note post_strenghten2=this

  have success_ref:
    "sim_ext abs1_conc_strel (ll_read_success_hinst inst)
     (ll_read_success (Admin HASH_REG) inst)
     (\<lambda>va vc. (if vc = htable_to_hinst_cnt (htable2_to_htable va) \<or> \<not>ok_hash_cnt vc
        then UNIV else {}) 
        \<inter> abs1_conc_strel)
     (force_same_and abs1_conc_strel) abs1_conc_strel
    "
    apply(auto simp add: sim_ext_def)
    using [[simp_depth_limit=2]]
    apply(clarsimp_all simp add: simN_wpc_def simS_wpc_def brute_force htable_from_cnt_def
      forall_INST less_eq_EcDead
      ec_state_by_flags_rev_cases
    )
    apply(safe)
    apply(clarsimp_all simp add: min_def ec_state_by_flags_rev_cases ec_state_simps
      hash_of_parse_written_ht ref_rel_defs)
    apply (metis (full_types) ec_state_by_flags_rev_cases(6) hash_by_inst.cases 
      log_state_by_inst.simps(1) log_state_by_inst.simps(2))
    apply (metis (full_types) ec_state_by_flags_rev_cases(6) hash_by_inst.cases 
      log_state_by_inst.simps(1) log_state_by_inst.simps(2))
    apply(case_tac inst)
    apply(clarsimp_all simp add: ok_hash_reg_read_def ok_hash_cnt_def Let_def)
    done

  show ?thesis
    apply(auto simp only: Abs1_LLInterface.ll_read_hinst_def LLInterface_Lemmas.ll_read_alt K_def
      intro!: sim_ext_choice_lockstep sim_ext_bindI sim_ext_try_catchI restart_mangle_sv_sim_ext
        post_strenghten1 post_strenghten2)
    apply(tactic {* distinct_subgoals_tac *})
    apply(rule success_ref)
    apply(case_tac[!] inst)
    apply(auto simp add: sim_ext_def)
    using [[simp_depth_limit=2]]
    apply(clarsimp_all simp add: simN_wpc_def simS_wpc_def brute_force htable_from_cnt_def
      ref_rel_defs forall_INST less_eq_EcDead
      ec_state_by_flags_rev_cases
    )
    apply(clarsimp_all simp add: ok_hash_reg_read_def ok_hash_cnt_def Let_def)
    apply(case_tac[!] "hw_state ba (Admin HASH_REG) Secondary")
    apply(clarsimp_all simp add: min_def ec_state_by_flags_rev_cases ec_state_simps)
    apply(case_tac[!] "hw_state ba (Admin HASH_REG) Primary")
    apply(clarsimp_all simp add: min_def ec_state_by_flags_rev_cases ec_state_simps)
    done

qed

lemma ll_read_ptr_ref[ref_thm]:
  "ind = Admin PTR_REG \<Longrightarrow>
    sim_ext abs1_conc_strel
    (Abs1_LLInterface.ll_read_ptr inst)
    (LLInterface.ll_read ind inst)
    (\<lambda>va vc. if vc = serialize_ptr va then abs1_conc_strel else {})
    (force_same_and abs1_conc_strel)
    abs1_conc_strel
  "
  apply(auto simp only: Abs1_LLInterface_Lemmas.ll_read_ptr_alt 
    LLInterface_Lemmas.ll_read_alt first_stage
    intro!: sim_ext_choice_lockstep restart_mangle_sv_refines_itself 
    sim_ext_bindI[where Tn="(\<lambda>va vc. if vc = serialize_ptr va then abs1_conc_strel else {})"])
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def LLInterface_Lemmas.ll_read_simps 
    restart_mangle_sv_empty
    simN_wpc_empty_conc simS_wpc_empty_conc)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: abs1_conc_strel_def simN_wpc_def simS_wpc_def
        brute_force
        hw_state_simps
        ok_hash_reg_read_def
        inj_eq[OF inj_serialize_ptr]
        size_serialize_ptr ptr_size
    split: split_if_asm
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: inj_eq[OF inj_serialize_ptr] inj_neq[OF inj_serialize_ptr]
    ec_state_by_flags_rev_cases ht2_to_ht_ptr_eq)
  apply (metis hw_state.distinct(3) parse_written_ht)+
  done

definition arrowN :: "('a \<Rightarrow> ('b, 's)nds_monad) \<Rightarrow> ('b \<Rightarrow> ('c, 's)nds_monad) \<Rightarrow> 'a \<Rightarrow> ('c, 's)nds_monad"
  where
  "arrowN f g \<equiv> \<lambda>a. f a \<guillemotright>=N g"

lemma bindN_through_arrowN:
  "(m \<guillemotright>=N f) = (arrowN (\<lambda>_. m) f ())"
  by(auto simp add: arrowN_def)

definition arrow :: "('a \<Rightarrow> ('b, 'e, 'sv, 'sp)pre_monad) \<Rightarrow> ('b \<Rightarrow> ('c, 'e, 'sv, 'sp)pre_monad) \<Rightarrow>
    'a \<Rightarrow> ('c, 'e, 'sv, 'sp)pre_monad"
  where
  "arrow f g \<equiv> \<lambda>a. f a \<guillemotright>= g"

lemma arrow_assoc: "arrow (arrow f g) h = arrow f (arrow g h)"
  by(auto simp add: arrow_def bind_assoc)

lemma bind_through_arrow:
  "(m \<guillemotright>= f) = (arrow (\<lambda>_. m) f ())"
  by(auto simp add: arrow_def)

lemma is_good_hash_project_simp:
  "(is_good_hash inst)\<Down>NE = empty"
  "(is_good_hash inst)\<Down>R = empty"
  by(auto simp add: brute_force is_good_hash_def)

lemma bind_associate_left:
  "f \<guillemotright>= (\<lambda>v. g v \<guillemotright>= h v) = (f \<guillemotright>= (\<lambda>v. g v \<guillemotright>= (\<lambda>v'. return (v, v')))) \<guillemotright>= (\<lambda>(v, v'). h v v')"
  by(simp add: bind_assoc bind_left_unit)

definition conc_mem_eq where
  "conc_mem_eq ind inst cnt \<equiv> {(sv, sp). state.memory sp ind inst = cnt }"

definition conc_hw_state_neq where
  "conc_hw_state_neq ind inst hws \<equiv> {(sv, sp). state.hw_state sp ind inst \<noteq> hws }"

lemma ll_read_post_NN:
  "\<lbrace>UNIV\<rbrace> 
    LLInterface.ll_read ind inst 
  \<lbrace>>N \<lambda>cnt. conc_mem_eq ind inst cnt \<inter> conc_hw_state_neq ind inst IDead \<rbrace>"
  apply(auto simp add: Hoare_alt Hoare_def Box_def)
  apply(auto simp add: brute_force conc_mem_eq_def conc_hw_state_neq_def)
  done

(*
lemma
  "htable_to_hinst_cnt (htable_from_cnt cnt) = (cnt :: 'c :: contents_cl)"
  apply(auto simp add: htable_from_cnt_def htable_to_hinst_cnt_def)
  apply(simp add: htable_to_hinst_cnt_def serialize_parse_htable)
    *)

lemma ll_read_bind_is_good_hash_ref:
  "sim_ext abs1_conc_strel
    (ll_read_hinst inst \<guillemotright>= (\<lambda>v. is_good_hash inst \<guillemotright>= (\<lambda>v'. return (v, v'))))
    (LLInterface.ll_read (Admin HASH_REG) inst) 
    (\<lambda>(cnt_a, ok_a) cnt_c. 
        bool_into_set ((cnt_c = htable_to_hinst_cnt (htable2_to_htable cnt_a) \<or> \<not>ok_hash_cnt cnt_c)
            \<and> ok_a = ok_hash_cnt cnt_c) 
        \<inter> abs1_conc_strel)
    (force_same_and abs1_conc_strel) 
    abs1_conc_strel
  "
  apply(subst bind_right_unit[where m="LLInterface.ll_read (Admin HASH_REG) inst", THEN sym])
  apply(rule sim_ext_bindI)
  apply(rule sim_ext_add_NN_inv[where I=UNIV and J=UNIV and I'="K UNIV", simplified, 
    OF true_postNN ll_read_post_NN ll_read_hinst_sim_ext])
  apply(auto simp add: K_beta intro!: sim_exts_empty_rel)
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force is_good_hash_def)
  apply(auto simp add: bool_into_set_def conc_mem_eq_def conc_hw_state_neq_def abs1_conc_strel_def 
    ok_hash_reg_read_def ok_hash_cnt_def Let_def)
  apply(case_tac[!] inst)
  apply(auto simp add: ec_state_by_flags_rev_cases)
  apply(case_tac "state.hw_state ba (Admin HASH_REG) Primary", simp_all)
  apply(case_tac "state.hw_state ba (Admin HASH_REG) Secondary", simp_all)
  apply (metis (full_types) exhaust_IUnwr)+
  done
  
lemma get_hash_table_ref[ref_thm]:
  "sim_ext abs1_conc_strel
    (Abs1_CommonDefs.get_hash_table inst)
    (CommonDefs.get_hash_table inst)
    (\<lambda>va vc. (if vc = htable2_to_htable va then UNIV else {}) \<inter> abs1_conc_strel)
    (force_same_and abs1_conc_strel)
    abs1_conc_strel
  "
  apply(unfold Abs1_CommonDefs.get_hash_table_def CommonDefs.get_hash_table_def sim_def)
  apply(subst bind_associate_left)
  apply(rule sim_ext_bindI)
  apply(rule ll_read_bind_is_good_hash_ref)
  apply(auto simp add: sim_def[THEN symmetric] bool_into_set_def htable_from_cnt_def
    force_same_and_same parse_written_ht
    intro: sim_exts_empty_rel sims_all_intros intro!: sim_ext_return_diffI sim_ext_throw_diffI)
  done

lemma vol_ops_ref[ref_thm]:
  "sim abs1_conc_strel (PmmDefs.init_vars) (PmmDefs.init_vars)"
  "sim abs1_conc_strel (PmmDefs.get_mode) (PmmDefs.get_mode)"
  "sim abs1_conc_strel (PmmDefs.set_degraded) (PmmDefs.set_degraded)"
  "sim abs1_conc_strel (PmmDefs.set_degraded_p) (PmmDefs.set_degraded_p)"
  "sim abs1_conc_strel (PmmDefs.set_degraded_h) (PmmDefs.set_degraded_h)"
  "sim abs1_conc_strel (PmmDefs.set_failed) (PmmDefs.set_failed)"
  "sim abs1_conc_strel (PmmDefs.get_updated) (PmmDefs.get_updated)"
  "sim abs1_conc_strel (PmmDefs.add_to_updated pind) (PmmDefs.add_to_updated pind)"
  "sim abs1_conc_strel (PmmDefs.set_updated upd) (PmmDefs.set_updated upd)"
  apply(auto simp add: sim_def sim_ext_def)
  using [[simp_depth_limit=2]]
  apply(auto simp add:
    PmmDefs.vol_ops_defs
    PmmDefs.getters_setters 
    brute_force
    simN_wpc_def simS_wpc_def
    abs1_conc_strel_def
  )
  done

lemma write_hash_table_ref'[ref_thm]:
  "ht' = htable2_to_htable ht \<Longrightarrow>
  sim abs1_conc_strel
    (Abs1_LLInterface.write_hinst inst ht)
    (CommonDefs.write_hash_table inst ht')
  "
  by(auto simp add: write_hash_table_def write_f_hash_ref')

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_ext_return[ref_thm]
  sim_ext_throw[ref_thm]
  sim_chooseN[ref_thm]

declare
  sim_ext_bind_rt[ref_bind]
  sim_ext_try_catch_rt[ref_try_catch]
  sim_ext_exc_case_rt[ref_exc_case]
  sim_ext_if_full_rt[ref_if_full]


(*
lemma write_hashes_ref':
  "sim abs1_conc_strel
    (Abs1_CommonDefs.write_hashes ht)
    (CommonDefs.write_hashes (htable2_to_htable ht))
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  done

(* TODO: this throws a "prove a different theorem" *)
  *)

lemma write_hashes_ref':
  "sim abs1_conc_strel
    (Abs1_CommonDefs.write_hashes ht)
    (CommonDefs.write_hashes (htable2_to_htable ht))
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  done

lemma write_hashes_ref[ref_thm]:
  "ht' = htable2_to_htable ht \<Longrightarrow>
    sim abs1_conc_strel
    (Abs1_CommonDefs.write_hashes ht)
    (CommonDefs.write_hashes ht')
  "
  by(simp add: write_hashes_ref')

declare sims_throw(1)[ref_thm del]
declare sim_ext_throw_diffI[ref_thm]
declare sim_ext_return_diffI[ref_thm]

lemma sim_sim_ext_try_catchI:
  "\<lbrakk> sim R ma1 mc1; \<forall>x. sim_ext R (ma2 x) (mc2 x) (force_same_and R) Se R \<rbrakk> \<Longrightarrow>
    sim_ext R (try_catch ma1 ma2) (try_catch mc1 mc2) (force_same_and R) Se R"
  by(auto simp add: force_same_and_def sim_def
    intro!: sim_ext_try_catchI[where Te="force_same_and R"] sim_exts_empty_rel)

declare sim_sim_ext_try_catchI[ref_try_catch]
lemma read_ptr_ref:
  "sim_ext abs1_conc_strel
    (Abs1_CommonDefs.read_ptr inst)
    (CommonDefs.read (Admin PTR_REG) inst)
    (\<lambda>va vc. if vc = serialize_ptr va then abs1_conc_strel else {})
    (force_same_and abs1_conc_strel)
    abs1_conc_strel
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: force_same_and_def ht2_to_ht_ptr_eq matches_htable_def ptr_matches_htable2_def)
  done

lemma read_ptr_DegradedH_ref:
  "sim_ext abs1_conc_strel
    (Abs1_CommonDefs.read_ptr_DegradedH inst)
    (CommonDefs.read_DegradedH (Admin PTR_REG) inst)
    (\<lambda>va vc. if vc = serialize_ptr va then abs1_conc_strel else {})
    (force_same_and abs1_conc_strel)
    abs1_conc_strel
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: force_same_and_def ht2_to_ht_ptr_eq matches_htable_def ptr_matches_htable2_def)
  done

lemma read_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.read pind inst)
    (CommonDefs.read (Public pind) inst)
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: matches_htable_def pub_matches_htable2_def ht2_to_ht_pub_eq)
  done

lemma read_DegradedH_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.read_DegradedH pind inst)
    (CommonDefs.read_DegradedH (Public pind) inst)
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: matches_htable_def pub_matches_htable2_def ht2_to_ht_pub_eq)
  done

lemma update_pub_instance_hash_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.update_pub_instance_hash ind inst cnt)
    (CommonDefs.update_instance_hash (Public ind) inst cnt)
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: matches_htable_def pub_matches_htable2_def ht2_to_ht_pub_eq)
  apply(case_tac va, simp_all)
  done

lemma update_ptr_instance_hash_ref':
  "
    sim abs1_conc_strel
    (Abs1_CommonDefs.update_ptr_instance_hash inst ptr_set)
    (CommonDefs.update_instance_hash (Admin PTR_REG) inst (serialize_ptr ptr_set))
  "
  apply(tactic {* (refinement_tac false default_simp @{context}) 1 *})
  apply(auto simp add: matches_htable_def pub_matches_htable2_def ht2_to_ht_pub_eq)
  apply(case_tac va, auto)
  done

lemma update_ptr_instance_hash_ref[ref_thm]:
  "cnt = serialize_ptr ptr_set \<Longrightarrow>
    sim abs1_conc_strel
    (Abs1_CommonDefs.update_ptr_instance_hash inst ptr_set)
    (CommonDefs.update_instance_hash (Admin PTR_REG) inst cnt)
  "
  by(auto intro: update_ptr_instance_hash_ref')

declare sim_ext_if_rt[ref_if]

lemma abs1_conc_strel_not_empty:
  "\<exists>x. x \<in> (abs1_conc_strel :: (('sv \<times> ('c :: contents_cl)abs1_state) \<times> ('sv \<times> 'c state))set)"
proof-
  fix svc :: 'sv and spc' :: "'c state"
  
  def spc \<equiv> "spc'\<lparr> ConcreteDefs.hw_state := K (K IOK),
    memory := (\<lambda>ind inst. SOME cnt. size cnt = capacity ind)
        (Admin (PTR_REG) := \<lambda>inst. SOME x. x \<in> range serialize_ptr,
        Admin (HASH_REG) := \<lambda>inst. SOME x. x \<in> range (htable_to_hinst_cnt o htable2_to_htable)) 
    \<rparr>"
  
  have spc_hash: "(\<forall>inst. ConcreteDefs.hw_state spc (Admin HASH_REG) inst = IOK)"
    by(auto simp add: spc_def K_def)
  
  have spc_sizes: "(\<forall>ind inst. size (state.memory spc ind inst) \<le> capacity ind)"
  proof(intro allI)
    fix ind inst
    have "\<exists>cnt. size (cnt :: 'c) = capacity ind" (is "\<exists>cnt. ?P cnt")
      by(auto simp add: all_sizes)
    then obtain cnt :: 'c where cnt_def: "?P cnt" ..
    have "\<exists>x :: 'c. x \<in> range serialize_ptr" (is "\<exists>x. ?P' x") by auto
    then obtain x :: 'c where x_def: "?P' x" ..
    have "\<exists>ht :: 'c. ht \<in> range (htable_to_hinst_cnt \<circ> htable2_to_htable)" (is "\<exists>ht. ?P'' ht") by auto
    then obtain ht where ht_def: "?P'' ht" ..
    show "size (state.memory spc ind inst) \<le> capacity ind"
      apply(simp add: spc_def)
      apply(intro conjI impI)
      apply(auto simp add: size_serialize_ptr intro!: ptr_size x_def intro: someI2[where a=x])[1]
      apply(rule someI2[where a=ht])
      apply(auto simp add: ht_def size_serialize_hashes htable_size)[2]
      apply(rule someI2[where a=cnt])
      apply(auto intro!: cnt_def)
      done
  qed

  def spa \<equiv> "
    \<lparr>
        hash1_state = ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Primary)) 
                (hw_state spc (Admin HASH_REG) Primary)
        , hash2_state = ec_state_by_flags
          (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Secondary)) 
              (hw_state spc (Admin HASH_REG) Secondary)
        , hash1 = (SOME x. htable_to_hinst_cnt (htable2_to_htable x) 
            = state.memory spc (Admin HASH_REG) Primary)
        , hash2 = (SOME x. htable_to_hinst_cnt (htable2_to_htable x) 
            = state.memory spc (Admin HASH_REG) Secondary)
        , ptr1 = parse_ptr (ConcreteDefs.memory spc (Admin PTR_REG) Primary),
        ptr1_state = ConcreteDefs.hw_state spc (Admin PTR_REG) Primary,
        ptr2 = parse_ptr (ConcreteDefs.memory spc (Admin PTR_REG) Secondary),
        ptr2_state = ConcreteDefs.hw_state spc (Admin PTR_REG) Secondary,
        memory = (\<lambda>pind. ConcreteDefs.memory spc (Public pind)),
        hw_state = (\<lambda>pind. ConcreteDefs.hw_state spc (Public pind))
    \<rparr>"
  have "((svc, spa), (svc, spc)) \<in> abs1_conc_strel" using spc_sizes spc_hash
    apply(auto simp add: abs1_conc_strel_def spa_def)
    apply(auto simp add: spc_def K_beta parse_ptr_spec intro!: f_inv_into_f[THEN sym] 
      intro: range_eqI someI)[2]
    apply(simp_all add: spc_def)
    apply (smt UNIV_I comp_apply image_iff someI)+
    done
  thus ?thesis ..
qed

lemma elem_if: "(x \<in> (if cond then UNIV else {})) = cond"
  by(simp)

lemma ex_elem_set: "(\<exists>x y. x \<in> set y) = True" "(\<exists>x y. y \<in> set x) = True"
  by (metis ListMem_iff elem)+

lemma subset_rewr: "(\<forall>x. x \<in> A \<longrightarrow> x \<in> B) = (A \<subseteq> B)"
  by auto

lemmas abs1_ref_simp_thms = 
    (* Int_commute Int_left_commute *)
    UNIV_def[THEN sym] Set.empty_def[THEN sym]
    Int_UNIV_right Int_UNIV_left Int_empty_left Int_empty_right empty_subsetI
    Int_Collects Collect_subset Set.Int_Collect Pmm_Refinement.Int_Collect  mem_Collect_eq
    subset_Collect subset_bCollect empty_iff Int_iff imp_conj_same imp_conj_simps
    conj_commute conj_assoc conj_left_absorb iso_tuple_UNIV_I
    HOL.simp_thms HOL.ex_simps
    HOL.True_implies_equals
    Product_Type.split
    triv_precond bool_into_set_def
    admin_indx.simps
    indx.simps
    comm_same_alter
    abs1_conc_strel_not_empty
    force_same_and_diff force_same_and_subset force_same_and_same
    if_True
    elem_if ex_elem_set subset_rewr
    subset_refl_eq
    imp_ex[THEN sym]
    set_equalityI_HOL
    Int_lower1_eq Int_lower2_eq

thm imp_cong

thm imp_conj_simps

lemmas abs1_ref_congs = imp_cong if_weak_cong Collect_cong

ML {*
fun is_Trueprop (Const ("HOL.Trueprop", _) $ _) = true
  | is_Trueprop _ = false

*}

ML {*

fun instantiate_eq_Vars thm =
    let
        fun does_match spat obj =
            Pattern.matches @{theory}
                                 (spat |> Syntax.read_term_global @{theory} |> Logic.varify_global, obj)

        val cprems = cprems_of thm
        val eq_cprems = cprems
                               |> filter (is_Trueprop o term_of)
                               |> map Thm.dest_arg
                               |> filter (does_match "(x :: 'a set) = y" o term_of)

        val ct_pairs =  eq_cprems
                            |> map Thm.dest_binop
                            |> map (fn (x, y) =>
                                   if is_Var (term_of x)
                                   then SOME (x, y)
                                   else
                                       if is_Var (term_of y)
                                       then SOME (y, x)
                                       else NONE
                               )
                            |> take_somes
    in
        cterm_instantiate ct_pairs thm
    end

*}
 
ML {*
  val abs1_ref_simpset = (HOL_basic_ss addsimps @{thms abs1_ref_simp_thms}
                                        addsimprocs [@{simproc HOL.defined_All}])
    |> fold Raw_Simplifier.add_cong @{thms abs1_ref_congs}

  fun abs1_simp _ = map (Ref_Tac.instantiate_redundant_subsets
                            #> asm_full_simplify abs1_ref_simpset 
                            #> Ref_Tac.instantiate_redundant_subsets
                            #> instantiate_eq_Vars
                            #> full_simplify abs1_ref_simpset)

*}

declare sims_throw(1)[ref_thm]

lemma ht2_to_ht_eq:
  "(htable2_to_htable ht' = htable2_to_htable ht) = (ht' = ht)"
proof(cases ht', cases ht, auto, rule ext)
  fix ptr_f ptr_f' inst
  assume 
    eq: "(\<lambda>inst. Hash (serialize_ptr (ptr_f inst))) = (\<lambda>inst. Hash (serialize_ptr (ptr_f' inst)))"
  from eq have "Hash (serialize_ptr (ptr_f inst)) = Hash (serialize_ptr (ptr_f' inst))"
    by (metis hash.inject parse_serialize_ptr)
  thus "ptr_f inst = ptr_f' inst"
    by(simp add: hash.inject inj_eq[OF inj_serialize_ptr])
qed

lemma check_hash_region_ref[ref_thm]:
  "sim abs1_conc_strel
   (Abs1_StartupDefs.check_hash_region) 
   (StartupDefs.check_hash_region) 
  "
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})
  apply(auto simp add: ht2_to_ht_eq)
  done

lemma read_ptr_instance_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.read_ptr inst)
    (CommonDefs.read_ptr_instance inst)
  "
  apply(auto simp add: read_ptr_instance_def sim_def)
  apply(subst bind_right_unit[THEN sym])
  apply(auto intro!: sim_ext_bindI[OF read_ptr_ref]sim_exts_empty_rel sim_ext_return_diffI
    simp add: parse_serialize_ptr force_same_and_same
    split: split_if_asm)
  done


lemma read_ptr_instance_DegradedH_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.read_ptr_DegradedH inst)
    (CommonDefs.read_ptr_instance_DegradedH inst)
  "
  apply(auto simp add: read_ptr_instance_DegradedH_def sim_def)
  apply(subst bind_right_unit[THEN sym])
  apply(auto intro!: sim_ext_bindI[OF read_ptr_DegradedH_ref]sim_exts_empty_rel sim_ext_return_diffI
    simp add: parse_serialize_ptr force_same_and_same
    split: split_if_asm)
  done

declare
  sim_return[ref_thm del]
  sims_throw(1)[ref_thm del]
  sim_ext_return[ref_thm del]
  sim_ext_throw[ref_thm del]
  sim_chooseN[ref_thm del]

declare
  sim_ext_bind_rt[ref_bind del]
  sim_ext_try_catch_rt[ref_try_catch del]
  sim_ext_exc_case_rt[ref_exc_case del]
  sim_ext_if_full_rt[ref_if_full del]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_chooseN[ref_thm]

declare
  plain_sim_bind_rt[ref_bind]
  plain_sim_try_catch_rt[ref_try_catch]
  plain_sim_exc_case_rt[ref_exc_case]
  plain_sim_if_rt[ref_if]
  plain_sim_if_full_rt[ref_if_full]
  plain_mapM_rt[ref_mapM]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_chooseN[ref_thm]

declare
  plain_sim_bind_rt[ref_bind]
  plain_sim_try_catch_rt[ref_try_catch]
  plain_sim_exc_case_rt[ref_exc_case]
  plain_sim_if_rt[ref_if]
  plain_sim_if_full_rt[ref_if_full]
  plain_mapM_rt[ref_mapM]

lemma get_ptr_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.get_ptr silence_deg)
    (CommonDefs.get_ptr silence_deg)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

declare
  sim_return[ref_thm del]
  sims_throw(1)[ref_thm del]
  sim_chooseN[ref_thm del]

declare
  plain_sim_bind_rt[ref_bind del]
  plain_sim_try_catch_rt[ref_try_catch del]
  plain_sim_exc_case_rt[ref_exc_case del]
  plain_sim_if_rt[ref_if del]
  plain_sim_if_full_rt[ref_if_full del]
  plain_mapM_rt[ref_mapM del]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_ext_return[ref_thm]
  sim_ext_throw[ref_thm]
  sim_chooseN[ref_thm]

declare
  sim_ext_bind_rt[ref_bind]
  sim_ext_try_catch_rt[ref_try_catch]
  sim_ext_exc_case_rt[ref_exc_case]
  sim_ext_if_full_rt[ref_if_full]


lemma get_sync_direction_ref[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.get_sync_direction pind)
    (CommonDefs.get_sync_direction pind)
  "
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})
  apply(auto simp add: ht2_to_ht_pub_eq)
  done
  
declare
  sim_return[ref_thm del]
  sims_throw(1)[ref_thm del]
  sim_ext_return[ref_thm del]
  sim_ext_throw[ref_thm del]
  sim_chooseN[ref_thm del]

declare
  sim_ext_bind_rt[ref_bind del]
  sim_ext_try_catch_rt[ref_try_catch del]
  sim_ext_exc_case_rt[ref_exc_case del]
  sim_ext_if_full_rt[ref_if_full del]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_chooseN[ref_thm]

declare
  plain_sim_bind_rt[ref_bind]
  plain_sim_try_catch_rt[ref_try_catch]
  plain_sim_exc_case_rt[ref_exc_case]
  plain_sim_if_rt[ref_if]
  plain_sim_if_full_rt[ref_if_full]
  plain_mapM_rt[ref_mapM]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_chooseN[ref_thm]

declare
  plain_sim_bind_rt[ref_bind]
  plain_sim_try_catch_rt[ref_try_catch]
  plain_sim_exc_case_rt[ref_exc_case]
  plain_sim_if_rt[ref_if]
  plain_sim_if_full_rt[ref_if_full]
  plain_mapM_rt[ref_mapM]

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.write_ptr_instance inst cnt)
    (CommonDefs.write_ptr_instance inst cnt)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.write_instance ind inst cnt)
    (CommonDefs.write_instance (Public ind) inst cnt)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.sync ind f to)
    (CommonDefs.sync ind f to)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_CommonDefs.just_sync ind f to)
    (CommonDefs.just_sync ind f to)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

lemma[ref_thm]:
  "ind = ind' \<Longrightarrow> sim abs1_conc_strel
    (Abs1_CommonDefs.just_sync_silence_deg ind f inst)
    (CommonDefs.just_sync_silence_deg ind' f inst)
  "
  apply(auto)
  apply(tactic {* (refinement_ins_tac false abs1_simp @{context}) 1 *})
  apply(auto)
  done

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_StartupDefs.check_pub_same pind)
    (StartupDefs.check_pub_same pind)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

lemma[ref_thm]:
  "sim abs1_conc_strel
    (Abs1_StartupDefs.check_ptr)
    (StartupDefs.check_ptr)
  "
  by(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})

ML {*
  build_sim_thms false abs1_simp @{context}
    @{term "(Abs1_StartupDefs.check_all_public)"}
    @{term "(StartupDefs.check_all_public)"}
    ||> (Context.>> o Ref_Thms.transfer)
*}

(* TODO: the above is not fully simplified;
  the simplifier should add a premise @{term "P \<equiv> True"}
  after seeing a term @{term "P \<Longrightarrow> Q"}, but this only
  works within a lemma, see below
*)

ML {*
fun my_mk_eq_True (_: simpset) r = let
  val _ = tracing "my_mk_eq_True called";
in
  SOME (r RS @{thm meta_eq_to_obj_eq} RS @{thm Eq_TrueI}) handle Thm.THM _ => NONE
end
*}

ML {*
  Ref_Thms.get @{context}
  |> hd
  (*
  |> asm_full_simp_tac (simpset_of @{context} addsimps @{thms some_simp Eq_TrueI})
  *)
  |> asm_full_simp_tac (HOL_basic_ss
                            |> Raw_Simplifier.add_cong @{thm imp_cong}
                            |> Simplifier.set_mkeqTrue my_mk_eq_True
                    ) 1
  |> Seq.pull
*}

schematic_lemma
  "
  \<forall>e. e \<in> abs1_conc_strel \<longrightarrow>
        ?R'34 \<subseteq> abs1_conc_strel \<and>
        abs1_conc_strel \<subseteq> ?R'34 \<and> e \<in> abs1_conc_strel \<Longrightarrow>
    sim abs1_conc_strel Abs1_StartupDefs.check_all_public
     StartupDefs.check_all_public
  "
  (* this also doesn't seem to have any effect? *)
  ML_prf {* 
    Context.>> (Context.map_proof (map_simpset (Simplifier.set_mkeqTrue my_mk_eq_True)))
 *}
  apply(simp)
  oops


lemma choice_sim_if[ref_thm]:
  "\<lbrakk> sim R ma1 mc1; sim R ma2 mc2 \<rbrakk> \<Longrightarrow>
  sim R (ma1 [+] ma2) (if cond then mc1 else mc2)"
  by(auto intro: sim_choice_abs_I1 sim_choice_abs_I2)

lemma startup_ref:
  "sim abs1_conc_strel
    (Abs1_StartupDefs.startup)
    (StartupDefs.startup)
  "
  apply(simp only: Abs1_StartupDefs.startup_def StartupDefs.startup_def 
    Abs1_StartupDefs.startup1_def Abs1_StartupDefs.startup2_def bind_assoc)
  apply(auto intro!: sims_all_intros)
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})+
  done

lemma fetch_ref:
  "sim abs1_conc_strel
    (Abs1_FetchDefs.fetch pind)
    (FetchDefs.fetch pind)
  "
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})
  done

lemma[ref_thm]:
  "sim R
    (PmmDefs.check_sizes upd_f)
    (PmmDefs.check_sizes upd_f)
  "
  apply(auto simp add: PmmDefs.check_sizes_def)
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})
  done

lemma update_ref:
  "sim abs1_conc_strel
    (Abs1_UpdateDefs.update regs)
    (UpdateDefs.update regs)
  "
  apply(tactic {* (refinement_tac false abs1_simp @{context}) 1 *})
  done

lemma update_full_ref:
  "sim abs1_conc_strel
    (<tryR>
        Abs1_UpdateDefs.update regs   
    <catchR>
        rec_tryR (Abs1_StartupDefs.startup)
    )
    (<tryR>
        UpdateDefs.update regs   
    <catchR>
        rec_tryR (StartupDefs.startup)
    )
  "
  by(auto intro!: sim_try_catchRI sim_rec_tryRI update_ref startup_ref)

lemma fetch_full_ref:
  "sim abs1_conc_strel
    (<tryR>
        Abs1_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs1_StartupDefs.startup)
    )
    (<tryR>
        FetchDefs.fetch ind
    <catchR>
        rec_tryR (StartupDefs.startup)
    )
  "
  by(auto intro!: sim_try_catchRI sim_rec_tryRI fetch_ref startup_ref)

lemma abs1_startup_full_ref_N:
  "sim abs1_conc_strel
    (<tryR>
        Abs1_StartupDefs.startup
    <catchR>
        rec_tryR (Abs1_StartupDefs.startup)
    )
    (<tryR>
        StartupDefs.startup
    <catchR>
        rec_tryR (StartupDefs.startup)
    )
  "
  by(auto intro!: sim_try_catchRI sim_rec_tryRI startup_ref)


end
