theory Abs1_SpecRef
  imports Abs1_Specification "../concrete/Concrete_Specification" Abs1_Ref "../SpecRef_Lemmas"
begin

lemma[simp]: "Concrete_Specification.htable2_to_htable = Abs1_Ref.htable2_to_htable"
  by(rule ext, case_tac x, auto)

fun dest_hash :: "'a hash \<Rightarrow> 'a" where
  "dest_hash (Hash c) = c"

fun htable_to_htable2 :: "'c htable \<Rightarrow> 'c :: serialization_cl htable2" where
  "htable_to_htable2 (HTable pub_h ptr_h) =
    HTable2 pub_h (\<lambda>inst. parse_ptr (dest_hash (ptr_h inst)))"

fun pmm_abs1_pi :: "('sv \<times> 'c state) \<Rightarrow> ('sv \<times> 'c :: contents_cl abs1_obs_state)" 
  where
  "pmm_abs1_pi (svc, spc) = (svc, 
    \<lparr>
        abs1_obs_state.hash1_state = ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Primary)) 
                (hw_state spc (Admin HASH_REG) Primary)
        , abs1_obs_state.hash2_state = ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Secondary)) 
                (hw_state spc (Admin HASH_REG) Secondary)
        , abs1_obs_state.hash1 = (
            if ok_hash_reg_read spc Primary
            then (Some o htable_to_htable2 o parse_htable o fst o parse_hinst) 
                (state.memory spc (Admin HASH_REG) Primary)
            else None)
        , abs1_obs_state.hash2 = (
            if ok_hash_reg_read spc Secondary
            then (Some o htable_to_htable2 o parse_htable o fst o parse_hinst) 
                (state.memory spc (Admin HASH_REG) Secondary)
            else None)
        , abs1_obs_state.ptr1 = parse_ptr (state.memory spc (Admin PTR_REG) Primary)
        , abs1_obs_state.ptr1_state = state.hw_state spc (Admin PTR_REG) Primary
        , abs1_obs_state.ptr2 = parse_ptr (state.memory spc (Admin PTR_REG) Secondary)
        , abs1_obs_state.ptr2_state = state.hw_state spc (Admin PTR_REG) Secondary
        , abs1_obs_state.memory = (\<lambda>pind. state.memory spc (Public pind))
        , abs1_obs_state.hw_state = (\<lambda>pind. state.hw_state spc (Public pind))
    \<rparr>
  )"
 
lemma htable_to_htable2_is_inv:
  "htable_to_htable2 (htable2_to_htable h) = h"
  by(case_tac h, clarsimp_all simp add: parse_serialize_ptr)

fun pmm_abs1_init_map :: "('sv \<times> 'c state) \<Rightarrow> ('sv \<times> 'c :: contents_cl abs1_state)" 
  where
  "pmm_abs1_init_map (svc, spc) = (svc, 
    \<lparr>
        abs1_state.hash1_state = ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Primary)) 
                (hw_state spc (Admin HASH_REG) Primary)
        , abs1_state.hash2_state = ec_state_by_flags
            (ok_hash_cnt (ConcreteDefs.memory spc (Admin HASH_REG) Secondary)) 
                (hw_state spc (Admin HASH_REG) Secondary)
        , abs1_state.hash1 =
            (htable_to_htable2 o parse_htable o fst o parse_hinst) 
                (state.memory spc (Admin HASH_REG) Primary)
        , abs1_state.hash2 =
            (htable_to_htable2 o parse_htable o fst o parse_hinst) 
                (state.memory spc (Admin HASH_REG) Secondary)
        , abs1_state.ptr1 = parse_ptr (state.memory spc (Admin PTR_REG) Primary)
        , abs1_state.ptr1_state = state.hw_state spc (Admin PTR_REG) Primary
        , abs1_state.ptr2 = parse_ptr (state.memory spc (Admin PTR_REG) Secondary)
        , abs1_state.ptr2_state = state.hw_state spc (Admin PTR_REG) Secondary
        , abs1_state.memory = (\<lambda>pind. state.memory spc (Public pind))
        , abs1_state.hw_state = (\<lambda>pind. state.hw_state spc (Public pind))
    \<rparr>
  )"
 
theorem pmm_refines_abs1:
  "refines abs1_conc_strel pmm_abs1_pi abs1_spec pmm_spec"
proof(intro refinesI)
  show "obs_consistent abs1_conc_strel pmm_abs1_pi abs1_spec pmm_spec"
    apply(clarsimp simp add: obs_consistent_def pmm_spec_def abs1_spec_def abs1_conc_strel_def obs_ts)
    apply(rename_tac spa spc)
    apply(clarsimp_all simp add: parse_serialize_ptr parse_written_ht ec_state_by_flags_rev_cases)
    apply(safe)
    apply(case_tac[!] spa, clarsimp_all)
    apply(clarsimp_all simp add: ok_hash_reg_read_def htable_to_htable2_is_inv intro!: ext)
    apply (metis exhaust_IOK)+
    done
next
  show "PO_refines abs1_conc_strel abs1_spec pmm_spec"
  proof(rule refine_basic)
    show "init_trans pmm_spec `` init pmm_spec
            \<subseteq> abs1_conc_strel `` init_trans abs1_spec `` init abs1_spec"
      apply(simp add: pmm_spec_def pmm_ts_def abs1_spec_def abs1_ts_def TS.defs spec.defs)
      apply(rule sim_norm_init_inclusion)
      apply(rule Abs1_Ref.startup_ref[THEN sim_rec_tryRI, THEN sim_imp_sim_norm, 
        simplified sim_norm_def]) 
      apply(auto simp add: abs1_conc_strel_def pmm_init_def abs1_init_def)
      apply(simp_all add: forall_INST)
      apply(clarsimp)
      apply(rule_tac x="snd (pmm_abs1_init_map (undefined, b))" in exI)
      apply(clarsimp_all simp add: hw_state_simps ec_state_by_flags_rev_cases parse_serialize_ptr)
      apply(clarsimp_all simp add: ok_hash_reg_read_def)
      apply(intro impI conjI)
      apply(clarsimp_all simp add: parse_written_ht htable_to_htable2_is_inv
        ht2_to_ht_ptr_eq matches_htable_def 
        pmm_matches_hash_def written_hash_of_htable_to_hinst_cnt abs1_ref_simp_thms)
      apply (metis ht2_to_ht_ptr_eq)+
      done
  next
    show "{abs1_conc_strel} TS.trans abs1_spec, TS.trans pmm_spec {> abs1_conc_strel}"
      apply(simp add: abs1_spec_def abs1_ts_def pmm_spec_def pmm_ts_def spec.defs TS.defs 
        abs1_trans_def pmm_trans_def)
      apply(intro relhoare_UN_lockstep allI relhoare_set_fun_choice_lockstep)
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare simp only: sim_norm_def[THEN symmetric])
      apply(auto intro!: 
        Abs1_Ref.fetch_full_ref[THEN sim_imp_sim_norm]
        Abs1_Ref.update_full_ref[THEN sim_imp_sim_norm]
        sim_try_catchRI[THEN sim_imp_sim_norm, OF Abs1_Ref.restart_mangle_sv_refines_itself]
        Abs1_Ref.startup_ref[THEN sim_rec_tryRI]
        simp add: try_rec_tryR_def)
      done
  qed
qed

theorem pmm_implements_abs1:
  "implements pmm_abs1_pi abs1_spec pmm_spec"
  by(auto intro: refinement_soundness pmm_refines_abs1)

end
