theory Abs1_StartupDefs
imports Abs1_CommonDefs
begin

definition sync_hashes :: "INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  sync_hashes_def[monad_def]:
  "sync_hashes to from \<equiv> do {
    ht \<leftarrow> (<try> get_hash_table from <catch> K (do { set_failed; throw FailMode }));
    <try> write_hinst to ht <catch> K set_degraded_h
  }"

definition check_hash_region :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  check_hash_region_def[monad_def]:
  "check_hash_region \<equiv> <try> do {
    ht2 \<leftarrow> get_hash_table Secondary;
    <try> do {
        ht1 \<leftarrow> get_hash_table Primary;
        (if ht1 \<noteq> ht2
            then 
                <try>
                    write_hinst Primary ht2 
                <catch> K set_degraded_h
            else return ())
        }
    <catch>
        (\<lambda>e. case e of 
            FailRead \<Rightarrow>
                <try>
                    write_hinst Primary ht2 
                <catch> K set_degraded_h
            | FailLLRead \<Rightarrow> set_degraded_h
            | e' \<Rightarrow> throw e'
        )
  }
  <catch>
    (\<lambda>e. case e of 
        FailRead \<Rightarrow> sync_hashes Secondary Primary
        | FailLLRead \<Rightarrow> do {
            (<try> get_hash_table Primary <catch> K (do { set_failed; throw FailMode }));
            set_degraded_h
        }
        | _ \<Rightarrow> throw e
    )
  "

definition check_ptr :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  check_ptr_def[monad_def]:
  "check_ptr \<equiv> <try> do {
    ptr_set2 \<leftarrow> read_ptr Secondary;
    <try> do {
        ptr_set1 \<leftarrow> read_ptr Primary;
        (if (ptr_set2 \<noteq> ptr_set1)
            then
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded v \<Rightarrow>
                        if v = EB False
                        then
                            do {
                                set_degraded_p;
                                <try> 
                                    write_ptr_ll Primary ptr_set2
                                <catch> K (return ())
                            }
                        else return ()
                    | _ \<Rightarrow> throw e
                )
            else return ())
    }
    <catch> 
        (\<lambda>e. case e of 
            FailRead \<Rightarrow>
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded _ \<Rightarrow> set_degraded_p
                    | _ \<Rightarrow> throw e
                )
            | Degraded v \<Rightarrow>
                if v = EU
                then set_degraded_p
                else
                    if (parse_ptr (get_exc_val_cnt v) \<noteq> ptr_set2)
                    then <try> write_ptr_ll Primary ptr_set2 <catch> K (return ())
                    else return ()
            | _ \<Rightarrow> throw e
        )
  }
  <catch> (\<lambda>e. case e of 
    Degraded v \<Rightarrow>
        if v = EU
        then do { read_ptr_or_fail_silence_deg Primary; set_degraded_p }
        else
            <try>
                write_ptr_ll  Primary (parse_ptr (get_exc_val_cnt v))
            <catch> K (return ())
    | FailRead \<Rightarrow>
        <try>
            just_sync_ptr Secondary Primary
        <catch> (\<lambda>e. case e of Degraded _ \<Rightarrow> return () | _ \<Rightarrow> throw e)
    | _ \<Rightarrow> throw e)
  "

definition check_pub_same
    :: "pub_indx  \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  check_pub_same_def[monad_def]:
  "check_pub_same pind \<equiv>
    <try> do {
        read pind Primary;
        (<try> do {
            read pind Secondary;
            return ()
        } <catch>
            (\<lambda>e. case e of
                FailRead \<Rightarrow> just_sync_silence_deg Secondary Primary pind
                | Degraded _ \<Rightarrow> return ()
                | _ \<Rightarrow> throw e
            )
        )
   } <catch> (\<lambda>e. case e of
        FailRead \<Rightarrow> just_sync_silence_deg Primary Secondary pind 
        | Degraded v \<Rightarrow>
            if v = EU
            then read_or_fail_silence_deg pind Secondary \<guillemotright> return ()
            else return ()
        | _ \<Rightarrow> throw e
    )
  "

definition check_pub_region
  :: "pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  check_pub_region_def[monad_def]:
  "check_pub_region pind \<equiv> do {
    dir \<leftarrow> get_sync_direction pind;
    if dir = SEq
    then do { check_pub_same pind; remove_from_ptr_silence_deg pind }
    else
        if dir = SR
        then sync_and_remove_from_ptr_silence_deg Secondary Primary pind 
        else sync_and_remove_from_ptr_silence_deg Primary Secondary pind 
  }"
 
definition check_all_public :: "
  (unit, pmm_vol_mem, 'c :: contents_cl) abs1_monad" where
  check_all_public_def[monad_def]:
  "check_all_public \<equiv> do {
    l \<leftarrow> chooseN (order_set UNIV);
    mapM' check_pub_region l
  }"

definition startup1 where
  startup1_def[monad_def]:
  "startup1 \<equiv> do {
    init_vars; 
    check_hash_region
  }
  "

definition startup2 where
  startup2_def[monad_def]:
  "startup2 \<equiv> do {
     check_ptr;
     check_all_public;
     throw_mode_error EU;
     throw StartedExc
   }"
  
definition startup :: "('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  startup_def[monad_def]:
  "startup \<equiv> do {
    startup1;
    startup2
  }"

end
