theory Abs1_Startup_Lemmas
imports Abs1_StartupDefs Abs1_LLInterface_Lemmas (* Abs1_Invariant_Lemmas *)
  "../Refinement"
  "../Projections" "../EvaluationLogicR" "../Pmm_Tacs" "../Misc_Lemmas" "../Pmm_Lemmas"
  "../Pmm_Refinement"
begin

ML_file "../no_throw.ML"

lemmas brute_force =
  unfold_monad
  project_defs
  Abs1_LLInterface.ll_read_defs
  Abs1_LLInterface.ll_read_hinst_defs
  Abs1_LLInterface.ll_read_ptr_defs
  Abs1_LLInterface.write_f_defs
  Abs1_LLInterface.write_hinst_defs
  Abs1_LLInterface.write_ptr_ll_defs
  restart_mangle_sv_raw
  getters_setters
  vol_ops_defs

lemma write_f_NN: "(write_f ind inst cnt)\<Down>NN = (write_success ind inst cnt)\<Down>NN"
  apply(rule ext)
  apply(auto simp add: brute_force split: split_if_asm)
  done

lemma ll_read_NE: "((ll_read ind inst)\<Down>NE) = (ll_read_failure ind inst)\<Down>NE"
  apply(rule ext)
  apply(auto simp add: brute_force split: split_if_asm)
  done

lemma ll_read_NN: "(ll_read ind inst)\<Down>NN = 
  (ll_read_any ind inst)\<Down>NN"
  apply(rule ext)
  apply(auto simp add: ll_read_def projectNN_choice projectNN_restart choice_emptyr)
  apply(auto simp add: brute_force split: option.split split_if_asm)
  done

lemma write_f_NE: "(write_f ind inst cnt)\<Down>NE = 
    (write_fail_instance ind inst cnt)\<Down>NE
    [+] (write_too_large ind inst cnt)\<Down>NE
    [+] (write_fail_readonly ind inst cnt)\<Down>NE
  "
  apply(rule ext)
  apply(auto simp add: write_f_def projectNE_choice projectNE_restart choice_emptyr)
  apply(auto simp add: brute_force split: option.split split_if_asm)
  done

definition get_hash_table_NN :: "INST \<Rightarrow> ('c :: contents_cl htable2, ('sv \<times> 'c abs1_state))nds_monad" where
  "get_hash_table_NN inst \<equiv> \<lambda>(sv, sp). {
    (r', sv, sp)|r'.
    r' = (hash_by_inst inst sp )
    \<and> log_state_by_inst inst sp = EcOK
  }"

lemma hol_mp: "((x \<longrightarrow> y) \<and> x) = (x \<and> y)"
  by(blast)

lemma get_hash_table_NN_simp: "(get_hash_table inst)\<Down>NN = get_hash_table_NN inst"
  apply(simp only: get_hash_table_def)
  apply(simp only: first_stage K_def)
  apply(simp only: ll_read_hinst_def choice_distrib)
  apply(rule ext)
  apply(auto simp only: choice_def)
  apply(auto simp add: brute_force get_hash_table_NN_def is_good_hash_def hol_mp
    ec_state_simps 
    split: split_if exceptions.split split_if_asm exceptions.split_asm option.split sum.split
  )
  done

definition get_hash_table_NE 
  :: "INST \<Rightarrow> ('e exceptions, ('sv \<times> ('c :: contents_cl)abs1_state))nds_monad" where
  "get_hash_table_NE inst \<equiv> \<lambda>(sv, sp :: 'c abs1_state ). 
    {(FailLLRead, sv, sp')|sp'.
        inst = Primary
        \<and> sp' = sp\<lparr> 
            hash1_state := EcDead
        \<rparr>
    } \<union>
    {(FailLLRead, sv, sp')|sp'.
        inst = Secondary
        \<and> sp' = sp\<lparr> 
            hash2_state := EcDead
        \<rparr>
    } \<union>
    {(FailRead, sv, sp')|sp' cnt.
        inst = Primary
        \<and> hash1_state sp > EcDead
        \<and> sp' = sp\<lparr> 
            hash1 := cnt,
            hash1_state := min EcBad (hash1_state sp)
        \<rparr>
    }  \<union>
    {(FailRead, sv, sp')|sp' cnt.
        inst = Secondary
        \<and> hash2_state sp > EcDead
        \<and> sp' = sp\<lparr> 
            hash2 := cnt, 
            hash2_state := min EcBad (hash2_state sp)
        \<rparr>
    } 

  "

lemma is_good_hash_project_simp:
  "(is_good_hash inst)\<Down>NE = empty"
  "(is_good_hash inst)\<Down>R = empty"
  by(auto simp add: brute_force is_good_hash_def)

lemma set_degraded_h_project_simps:
  "set_degraded_h\<Down>NE = empty"
  "set_degraded_h\<Down>R = empty"
  by(auto simp add: brute_force set_degraded_h_def)

lemma set_failed_project_simps:
  "set_failed\<Down>NE = empty"
  "set_failed\<Down>R = empty"
  by(auto simp add: brute_force set_failed_def)

lemma get_hash_table_NE_simp: "(get_hash_table inst)\<Down>NE = get_hash_table_NE inst"
  apply(simp only: get_hash_table_def first_stage K_def)
  apply(simp only: ll_read_hinst_def choice_distrib first_stage ll_read_simps restart_mangle_sv_empty
    is_good_hash_project_simp)
  apply(rule ext)
  apply(case_tac inst)
  apply(auto simp only: choice_def)
  apply(auto simp add: brute_force get_hash_table_NE_def is_good_hash_def hol_mp
    ec_state_simps
    split: split_if exceptions.split split_if_asm exceptions.split_asm
  )
  apply(rule_tac x = "hash1 b" in exI, case_tac "hash1_state b", simp_all add: min_def)[]
  apply(rule_tac x = "hash2 b" in exI, case_tac "hash2_state b", simp_all add: min_def)
  done

lemmas get_hash_table_raw = 
  get_hash_table_NN_simp
  get_hash_table_NE_simp

lemma sync_hashes_doesnt_throw_InvalidInput: "doesnt_throw_raw ((sync_hashes to from)\<Down>NE) InvalidInput"
  apply(simp only: doesnt_throw_raw_def sync_hashes_def first_stage get_hash_table_raw
    set_degraded_h_project_simps set_failed_project_simps)
  apply(auto simp add: doesnt_throw_def unfold_monad)
  done

lemma sync_hashes_doesnt_throw_FailRead: "doesnt_throw_raw ((sync_hashes to from)\<Down>NE) FailRead"
  apply(simp only: doesnt_throw_raw_def sync_hashes_def first_stage get_hash_table_raw
    set_degraded_h_project_simps set_failed_project_simps)
  apply(auto simp add: doesnt_throw_def unfold_monad)
  done

lemma sync_hashes_doesnt_throw_FailLLRead: "doesnt_throw_raw ((sync_hashes to from)\<Down>NE) FailLLRead"
  apply(simp only: doesnt_throw_raw_def sync_hashes_def first_stage get_hash_table_raw
    set_degraded_h_project_simps set_failed_project_simps)
  apply(auto simp add: doesnt_throw_def unfold_monad)
  done

lemma sync_hashes_doesnt_throw_FailWrite: "doesnt_throw_raw ((sync_hashes to from)\<Down>NE) FailWrite"
  apply(simp only: doesnt_throw_raw_def sync_hashes_def first_stage get_hash_table_raw
    set_degraded_h_project_simps set_failed_project_simps)
  apply(auto simp add: doesnt_throw_def unfold_monad)
  done

lemma sync_hashes_doesnt_throw_Degraded: "\<forall>v. doesnt_throw_raw ((sync_hashes to from)\<Down>NE) (Degraded v)"
  apply(simp only: doesnt_throw_raw_def sync_hashes_def first_stage get_hash_table_raw
    set_degraded_h_project_simps set_failed_project_simps)
  apply(auto simp add: doesnt_throw_def unfold_monad)
  done

lemmas sync_hashes_no_throw =
  sync_hashes_doesnt_throw_InvalidInput
  sync_hashes_doesnt_throw_FailLLRead
  sync_hashes_doesnt_throw_FailRead
  sync_hashes_doesnt_throw_FailWrite
  sync_hashes_doesnt_throw_Degraded

declare sync_hashes_no_throw[no_throw]

lemma get_raw_hashes_no_throw:
  "doesnt_throw_raw ((get_hash_table inst)\<Down>NE) InvalidInput"
  "doesnt_throw_raw ((get_hash_table inst)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((get_hash_table inst)\<Down>NE) FailMode"
  "\<forall>v. doesnt_throw_raw ((get_hash_table inst)\<Down>NE) (Degraded v)"
  thm get_hash_table_NE_simp
  apply(simp add: get_hash_table_NE_simp)
  apply(auto simp only: get_hash_table_NE_simp)
  apply(auto simp add: doesnt_throw_raw_def get_hash_table_NE_def project_defs)
  done

declare get_raw_hashes_no_throw[no_throw]

lemma bindN_returnN_no_throw:
"e' \<noteq> e \<Longrightarrow> doesnt_throw_raw (m \<guillemotright>N returnN e) e'"
  by(auto simp add: unfold_monad doesnt_throw_raw_def)

declare bindN_returnN_no_throw[no_throw]

(*
lemma exists_wrong_cnt: "\<exists>(cnt :: 'a :: contents_cl). \<not>ok_hash cnt"
proof-
  fix cnt :: 'a
  obtain cnt' where "cnt' \<noteq> cnt" using infinite_has_two
    by(auto)
  hence hash_differs: "Hash cnt \<noteq> Hash cnt'"
    by simp
  def cnt'' \<equiv> "serialize_hash_reg (cnt, Hash cnt')"
  have "\<not>ok_hash cnt''" using cnt''_def hash_differs
    by(simp add: ok_hash_def s_hash_reg_spec surj_f_inv_f[OF p_hash_reg_bij[THEN bij_is_surj]])
  thus ?thesis 
    ..
qed
  *)
 
definition check_hash_region_NN :: "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs1_state)nds_monad" where
  "check_hash_region_NN \<equiv>
  (\<lambda>(sv, sp). {((), sv, sp')|sp' cnt.
    (hash1_state sp = EcOK \<and> cnt = hash1 sp \<or> hash2_state sp = EcOK \<and> cnt = hash2 sp)
    \<and> (hash1_state sp \<ge> EcBad \<and> hash2_state sp \<ge> EcBad)
    \<and> sp' = sp\<lparr> hash1 := cnt, hash2 := cnt, hash1_state := EcOK, hash2_state := EcOK \<rparr>
  })
  [+] (\<lambda>(sv, sp). {((), sv', sp')|sv' sp' cnt.
    (hash1_state sp = EcOK \<and> cnt = hash1 sp \<or> hash2_state sp = EcOK \<and> cnt = hash2 sp)
    \<and> (hash1_state sp' = EcOK \<or> hash2_state sp' = EcOK)
    \<and> (hash1_state sp' = EcOK \<longrightarrow> hash1 sp' = cnt)
    \<and> (hash2_state sp' = EcOK \<longrightarrow> hash2 sp' = cnt)
    \<and> (hash1_state sp' \<le> EcUnwr \<or> hash2_state sp' \<le> EcUnwr)
    \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
    \<and> (hash1_state sp = EcDead \<longrightarrow> hash1_state sp' = EcDead)
    \<and> (hash2_state sp = EcDead \<longrightarrow> hash2_state sp' = EcDead)
    \<and> (hash1_state sp = EcUnwr \<longrightarrow> hash1_state sp' \<le> EcUnwr)
    \<and> (hash2_state sp = EcUnwr \<longrightarrow> hash2_state sp' \<le> EcUnwr)
    \<and> abs1_same_rest sp' sp
  })
  "

lemma not_all_diff:
  "(\<forall>x. x \<noteq> y) = False"
  "(\<forall>x. y \<noteq> x) = False"
  by(auto)

lemma log_state_upd_eq_iff:
  "hash2_state sp = x \<Longrightarrow> sp\<lparr>hash2_state := x\<rparr> = sp"
  "hash1_state sp = x \<Longrightarrow> sp\<lparr>hash1_state := x\<rparr> = sp"
  by(auto)

lemma check_hash_region_NN_right: "(r, sv', sp') \<in> (check_hash_region\<Down>NN) (sv, sp) \<Longrightarrow> 
        (r, sv', sp') \<in> check_hash_region_NN (sv, sp)"
  apply(simp only: check_hash_region_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp)
  apply(simp_all only: sync_hashes_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp
    set_degraded_h_project_simps set_failed_project_simps)
  apply(clarsimp_all simp add: brute_force set_degraded_h_def
    get_hash_table_NN_def get_hash_table_NE_def
    split: split_if exceptions.split split_if_asm exceptions.split_asm
  )
  apply(iterative_simp)+
  apply(clarsimp_all simp add: check_hash_region_NN_def choice_def abs1_same_rest_def all_conj
    min_def log_state_upd_eq_iff
    not_all_diff)
  apply(simp_all (no_asm_use) add: HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(iterative_simp simp add: ec_state_simps min_def split: split_if_asm)+
  done

declare disjCI[rule del]
declare UnCI[rule del]

lemma init_varsR: 
  "(init_vars\<Down>R) (sv, sp) = {}"
  by(auto simp add: init_vars_def brute_force)

lemma init_vars_simp: "((init_vars\<Down>NN) \<guillemotright>N m) (sv, sp) = 
  m (sv \<lparr> card_mode := CNormal, updated := {} \<rparr>, sp)"
  by(auto simp add: brute_force init_vars_def)

lemma init_vars_simp': "(init_vars\<Down>NN) = (\<lambda>(sv, sp). 
  {((), sv \<lparr> card_mode := CNormal, updated := {} \<rparr>, sp)})"
  by(auto simp add: brute_force init_vars_def)

lemma init_varsR':
  "(init_vars)\<Down>R = empty"
  by(auto simp add: init_vars_def brute_force)

definition get_hash_table_NE'
  :: "INST \<Rightarrow> ('e exceptions, ('sv \<times> ('c :: contents_cl)abs1_state))nds_monad" where
  "get_hash_table_NE' inst \<equiv> (\<lambda>(sv, sp :: 'c abs1_state ). 
    {(FailLLRead, sv, sp')|sp'.
        inst = Primary
        \<and> sp' = sp\<lparr> 
            hash1_state := EcDead
        \<rparr>
    }) [+]
    (\<lambda>(sv, sp). {(FailLLRead, sv, sp')|sp'.
        inst = Secondary
        \<and> sp' = sp\<lparr> 
            hash2_state := EcDead
        \<rparr>
    }) [+]
    (\<lambda>(sv, sp). {(FailRead, sv, sp')|sp' cnt.
        inst = Primary
        \<and> hash1_state sp > EcDead
        \<and> sp' = sp\<lparr> 
            hash1 := cnt,
            hash1_state := min EcBad (hash1_state sp)
        \<rparr>
    }) [+]
    (\<lambda>(sv, sp). {(FailRead, sv, sp')|sp' cnt.
        inst = Secondary
        \<and> hash2_state sp > EcDead
        \<and> sp' = sp\<lparr> 
            hash2 := cnt, 
            hash2_state := min EcBad (hash2_state sp)
        \<rparr>
    })
  "

lemma not_Inl: "(\<forall>x. y \<noteq> Inl x) = (\<exists>z. y = Inr z)"
  by(case_tac y, auto)

lemma has_two_exists_different:
  "\<exists>x. x \<noteq> (y :: 'a :: has_two)"
  apply(case_tac "y = has_two_a")
  apply(auto simp add: diff_vals)
  apply(rule_tac x="has_two_b" in exI)
  apply(auto simp add: diff_vals[THEN not_sym])
  done

text {* A couple of technical lemmas *}
lemma aux1:
  "(\<forall>a. ((\<forall>x. a \<noteq> Inl (x :: 'e exceptions)) \<or> ab \<noteq> Some a) \<and> (ab = Some a \<longrightarrow> Q a)) = 
        ((\<not>(\<exists>y. ab = Some y)) \<or> (\<exists>z. ab = Some (Inr z) \<and> Q (Inr z)))"
  apply(case_tac ab)
  apply(auto)
  apply(case_tac a)
  apply(auto)
  apply(case_tac aa)
  apply(auto)
  done

lemma card_mode_full_upd:
  "((a :: pmm_vol_mem)\<lparr> updated := l, card_mode := m \<rparr>) = \<lparr> card_mode = m, updated = l \<rparr> "
  "((a :: pmm_vol_mem)\<lparr> card_mode := m, updated := l \<rparr>) = \<lparr> card_mode = m, updated = l \<rparr> "
  by(auto)

lemma abs1_NN_reduction: 
  "(e, sv', sp') \<in>
    ((init_vars\<Down>NN) \<guillemotright>N (check_hash_region\<Down>R)) (sv, sp) \<guillemotright>=S ((init_vars\<Down>NN) \<guillemotright>N check_hash_region_NN)
  \<Longrightarrow>
  (e, sv', sp') \<in>
    ((init_vars\<Down>NN) \<guillemotright>N check_hash_region_NN)
    (sv, sp)"
  apply(subst (asm) check_hash_region_NN_def)
  apply(simp only: first_stage union_bindS)
  apply(safe)
  apply(simp_all only: check_hash_region_def first_stage union_bindS)
  apply(safe)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp 
    set_degraded_h_project_simps set_failed_project_simps)
  apply(simp_all only: sync_hashes_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hash_table_def first_stage empty_bindS)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only:
    get_hash_table_NN_def get_hash_table_NE_def first_stage union_bindS
    is_good_hash_project_simp init_varsR' empty_bindS)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: init_vars_def brute_force aux1 is_good_hash_def
    abs1_same_rest_def
    split: split_if_asm option.split_asm sum.split_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: log_state_upd_eq_iff min_def split: split_if_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: check_hash_region_NN_def choice_def
   abs1_same_rest_def split: split_if_asm)
  apply(simp_all (no_asm_use) add: HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: log_state_upd_eq_iff min_def card_mode_full_upd split: split_if_asm)
  apply(iterative_simp simp add: ec_state_simps min_def split: split_if_asm)+
  done

definition check_hash_region_NE ::
  "('c exceptions, pmm_vol_mem \<times> 'c :: contents_cl abs1_state)nds_monad" where
  "check_hash_region_NE \<equiv>
    \<lambda>(sv, sp).
    {
        (FailMode, sv', sp')|sv' sp'.
            hash1_state sp' \<le> EcBad
            \<and> hash2_state sp' \<le> EcBad
            \<and> (hash1_state sp = EcDead \<longrightarrow> hash1_state sp' = EcDead)
            \<and> (hash2_state sp = EcDead \<longrightarrow> hash2_state sp' = EcDead)
            \<and> (hash1_state sp = EcUnwr \<longrightarrow> hash1_state sp' \<le> EcUnwr)
            \<and> (hash2_state sp = EcUnwr \<longrightarrow> hash2_state sp' \<le> EcUnwr)
            \<and> sv' = sv\<lparr> card_mode := CFail \<rparr>
            \<and> abs1_same_rest sp sp'
    }
  "

lemma check_hash_region_NE_right:
  "(check_hash_region\<Down>NE) (sv, sp) \<subseteq> (check_hash_region_NE) (sv, sp)"
  apply(simp only: check_hash_region_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp
    set_degraded_h_project_simps set_failed_project_simps)
  apply(simp only: sync_hashes_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp 
    set_degraded_h_project_simps set_failed_project_simps
  )
  apply(clarsimp_all simp add: brute_force get_hash_table_NN_def get_hash_table_NE_def
    abs1_same_rest_def ec_state_simps
    vol_ops_defs
    split: split_if exceptions.split split_if_asm exceptions.split_asm
  )
  apply(iterative_simp)+
  apply(clarsimp_all simp add: check_hash_region_NE_def abs1_same_rest_def ec_state_simps min_def)
  done

lemma abs1_NE_reduction': 
  "(e, sv', sp') \<in>
    ((init_vars\<Down>NN) \<guillemotright>N (check_hash_region\<Down>R)) (sv, sp) \<guillemotright>=S ((init_vars\<Down>NN) \<guillemotright>N check_hash_region_NE)
  \<Longrightarrow>
  (e, sv', sp') \<in>
    ((init_vars\<Down>NN) \<guillemotright>N check_hash_region_NE)
    (sv, sp)"
  apply(subst (asm) check_hash_region_NE_def)
  apply(simp_all only: check_hash_region_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp 
    set_degraded_h_project_simps set_failed_project_simps)
  apply(simp_all only: sync_hashes_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(safe)
  apply(simp_all only: get_hash_table_NN_simp get_hash_table_NE_simp 
    set_degraded_h_project_simps set_failed_project_simps)
  apply(simp_all only: get_hash_table_def)
  apply(simp_all add: first_stage if_choice vol_ops_project_simps is_good_hash_project_simp
    bindS_empty empty_bindS
    cong: if_cong exceptions.case_cong)
  apply(simp_all only:
    get_hash_table_NN_def get_hash_table_NE_def is_good_hash_project_simp init_varsR')
  apply(simp_all (no_asm_use))
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: init_vars_def brute_force aux1
    abs1_same_rest_def
    split: split_if_asm option.split_asm sum.split_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=4]]
  apply(clarsimp_all simp add: check_hash_region_NE_def choice_def abs1_same_rest_def
    min_def split: split_if_asm)
  apply(iterative_simp simp add: ec_state_simps min_def split: split_if_asm)+
  done

definition abs1_check_hash_region' ::  "(unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  "abs1_check_hash_region' \<equiv> compose
  check_hash_region_NN
  check_hash_region_NE
  (check_hash_region\<Down>R)"

definition abs1_startup1' ::  "(unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad" where
  "abs1_startup1' \<equiv> do { init_vars;  abs1_check_hash_region' }"

lemma init_vars_NE:
  "(init_vars\<Down>NE) = empty"
  by(auto simp add: init_vars_def brute_force)

ML {*

  (*

    Uses the given monad inclusion theorem to weaken the subgoal premises.

    The theorem should be of the form @{term "m s \<subseteq> m' s"}. The tactic
    then weakens the first premise of the shape @{term "x \<in> (f m) s"}
    into a premise of the form @{term "x \<in> (f m') s"}, where @{term f}
    is some combination of basic monadic operations (bind, bindN, bindS
    try_catch, projectNN, projectNE, projectR)

  *)

  fun get_fun_term' (f $ x) = f

  fun first_some [] = NONE
  |   first_some (SOME x :: _) = SOME x
  |   first_some (_ :: xs) = first_some xs


  fun strip_Trueprops t = case t of
     @{term "Trueprop"} $ t' => strip_Trueprops t'
     | _ => t

  val subset_pat = @{term_pat "?A \<subseteq> ?B"}

  fun get_target_term thy thm = thm
     |> tap (fn th => length (prems_of th) = 0
         orelse error "Wrong format for the subset theorem (has premises)"
         )
     |> prop_of
     |> strip_Trueprops
     |> tap (fn t => if Pattern.matches thy (subset_pat, t) 
             then ()
             else error "Wrong format for the subset theorem"
         )
     |> get_arg 1
     |> (fn t => get_fun_term' t
         handle MATCH => error "Wrong format for the subset theorem (not a function application)")
 

  fun find_term thy l t t' = if Pattern.matches thy (t, t')
    then SOME l
    else case t' of
        Const (@{const_name "bindN"}, _) $ m $ f =>
            first_some [
                find_term thy (@{thm "bindN_mono1"} :: l) t m
                , find_term thy (@{thm "bindN'_mono2_all"} :: l) t f
            ]
        | Const (@{const_name "bindS"}, _) $ m $ f =>
            first_some [
                find_term thy (@{thm "bindS_mono1"} :: l) t m
                , find_term thy (@{thm "bindS_mono2_all"} :: l) t f
            ]
        | Const (@{const_name "projectNN"}, _) $ m =>
            find_term thy (@{thm "projectNN_mono_all"} :: l) t m

        | Const (@{const_name "projectNE"}, _) $ m =>
            find_term thy (@{thm "projectNE_mono_all"} :: l) t m
  
        | Const (@{const_name "projectR"}, _) $ m =>
            find_term thy (@{thm "projectR_mono_all"} :: l) t m

        | f $ s => find_term thy l t f

        | Abs (_, _, body) => find_term thy l t body
        | _ => NONE

  fun exploit_mono_tac ctxt thm i st =
  let
    val thy = Proof_Context.theory_of ctxt

    val target_term = get_target_term thy thm

    (* val _ = (tracing "Target term: "; tracing (PolyML.makestring target_term)) *)

    val ({asms = assms, ...}, _) = Subgoal.focus ctxt i st


    fun make_tac [] = no_tac st
    |  make_tac (ct :: cts) = let
        val t = strip_Trueprops (term_of ct)
        val in_pat = @{term_pat "?x \<in> ?A"} 
        val is_match = Pattern.matches thy (in_pat, t)
  
        fun make_thm t t' = 
            find_term thy [] t t'

        fun finalize_result l = (Library.foldl (op RS) (thm, l)) RS @{thm subsetD}
    in
        if is_match
            then (
                case make_thm target_term (get_arg 0 t) of
                    NONE => make_tac cts
                    | SOME l =>
                        dtac (finalize_result l) i st
            )
            else make_tac cts
    end
    
  in
    make_tac assms  
  end

  *}

lemma check_hash_region_NN_right':
 "(check_hash_region\<Down>NN) s \<subseteq> check_hash_region_NN s"
  by(case_tac s, auto intro!: check_hash_region_NN_right)

lemma check_hash_region_NE_right':
 "(check_hash_region\<Down>NE) s \<subseteq> check_hash_region_NE s"
  by(case_tac s, auto intro!: check_hash_region_NE_right)

lemma empty_union:
  "empty s \<union> S = S"
  by(simp add: empty_def K_def)

(*
lemma abs1_startup1'_NN: "(abs1_startup1'\<Down>NN) = (startup1\<Down>NN)"
  by(auto simp add: abs1_startup1'_def startup1_def abs1_check_hash_region'_def
    compose_project project_simps check_hash_region_NN_simp)
    *)

lemma abs1_startup1_fin_NN:
  "sim_normN Id  (abs1_startup1') ((try_catchR startup1 abs1_startup1'))" 
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def
    first_stage abs1_startup1'_def
    choice_def
    abs1_check_hash_region'_def compose_project 
  )
  apply(simp add: startup1_def project_simps init_vars_NE first_stage)
  apply(tactic {* exploit_mono_tac @{context} @{thm check_hash_region_NN_right'} 1 *})
  apply(auto simp add: startup1_def first_stage init_varsR' empty_union
    intro!: abs1_NN_reduction)
  done

lemma abs1_startup1_fin_NE:
    "sim_normE Id abs1_startup1' (try_catchR startup1 abs1_startup1')" 
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def
    first_stage abs1_startup1'_def init_vars_NE
    choice_def
    abs1_check_hash_region'_def compose_project 
  )
  apply(simp add: startup1_def project_simps init_vars_NE first_stage)
  apply(tactic {* exploit_mono_tac @{context} @{thm check_hash_region_NE_right'} 1 *})
  apply(auto simp add: startup1_def first_stage init_varsR' empty_union
    intro!: abs1_NE_reduction')
  done

lemma simN_wpc_projectN:
  "simN_wpc (ma\<Down>NN) (mc\<Down>NN) (\<lambda>va vc. S (Inr va) (Inr vc))
    \<inter> simN_wpc (ma\<Down>NE) (mc\<Down>NE) (\<lambda>ea ec. S (Inl ea) (Inl ec))
   \<subseteq> simN_wpc (ma\<Down>N) (mc\<Down>N) S
  "
  apply(auto simp add: simN_wpc_def project_defs)
  apply(metis not_Inl)
  done

lemma force_same_inj_f:
  "inj f \<Longrightarrow> force_same_and R (f x) (f y) = force_same_and R x y"
  by(simp add: force_same_and_def injD)

lemma abs1_startup1_fin:
    "sim_norm Id abs1_startup1' (try_catchR startup1 abs1_startup1')" 
  by(auto simp add: sim_norm_decompose abs1_startup1_fin_NE abs1_startup1_fin_NN)

lemma eq_by_project: "(m = m') = (m\<Down>NN = m'\<Down>NN \<and> m\<Down>NE = m'\<Down>NE \<and> m\<Down>R = m'\<Down>R)"
  apply(auto simp add: project_defs mem_Collect_eq intro!: ext)
  apply (smt mem_Collect_eq not_Inl not_None_eq split_conv)+
  done

end
