theory Abs2_CommonDefs
  imports Abs2_LLInterface
begin

definition get_hashes_success  :: "('c htable2, 'sv, 'c :: contents_cl)abs2_monad" where
  "get_hashes_success \<equiv> \<lambda>(sv, sp).
    {(Some (Inr h), sv, sp)|h.
        (hstate sp = LOK \<or> hstate sp = LDegraded)
        \<and> h = hash sp
  }"

definition get_hashes_fail :: "('a, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "get_hashes_fail \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' h'.
        sv' = sv\<lparr> card_mode := CFail \<rparr>
        \<and> sp' = sp\<lparr> hash := h', hstate := LDead \<rparr>
  }"

definition get_hashes_degrade :: "('a, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "get_hashes_degrade \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp'.
        hstate sp \<noteq> LDead
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded \<rparr>
  }"

definition get_hashes :: "('c htable2, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "get_hashes \<equiv>
    get_hashes_success
    [+] get_hashes_fail
    [+] get_hashes_degrade
    [+] restart_mangle_sv"

lemmas get_hashes_defs =
  get_hashes_def
  get_hashes_success_def
  get_hashes_fail_def
  get_hashes_degrade_def

definition get_hashes_degrade' :: "('c htable2, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "get_hashes_degrade' \<equiv> \<lambda>(sv, sp).
    {(Some (Inr h), sv, sp')|h sp'.
        hstate sp = LOK
        \<and> h = hash sp
        \<and> sp' = sp\<lparr> hstate := LDegraded \<rparr>
  }"

definition get_hashes_DegradedH :: "('c htable2, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "get_hashes_DegradedH \<equiv>
    get_hashes_success
    [+] get_hashes_degrade'
    [+] get_hashes_fail
    [+] restart_mangle_sv"

lemmas get_hashes_DegradedH_defs =
  get_hashes_DegradedH_def
  get_hashes_def
  get_hashes_success_def
  get_hashes_fail_def
  get_hashes_degrade'_def

lemmas get_hashes_all_defs =
  get_hashes_DegradedH_defs
  get_hashes_degrade_def

definition write_hashes_success :: 
  "'c htable2 \<Rightarrow> (unit, 'sv, 'c)abs2_monad" where
  "write_hashes_success ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        hstate sp = LOK
        \<and> sp' = sp\<lparr> hash := ht \<rparr>
  }"

definition write_hashes_degrade :: 
  "'c htable2 \<Rightarrow> ('a, pmm_vol_mem, 'c)abs2_monad" where
  "write_hashes_degrade ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB False))), sv', sp')|sv' sp'.
        sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := min (hstate sp) LDegraded \<rparr>
  }"

definition write_hashes_degrade_write :: 
  "'c htable2 \<Rightarrow> ('a, pmm_vol_mem, 'c)abs2_monad" where
  "write_hashes_degrade_write ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB True))), sv', sp')|sv' sp'.
        hstate sp \<noteq> LDead
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hash := ht, hstate := LDegraded \<rparr>
  }"

definition write_hashes_fail ::
  "'c htable2 \<Rightarrow> ('a, pmm_vol_mem, 'c)abs2_monad" where
  "write_hashes_fail ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp'.
        sv' = sv\<lparr> card_mode := CFail \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDead \<rparr>
  }"

definition write_hashes_partial_restart ::
  "'c htable2 \<Rightarrow> ('a, 'sv, 'c)abs2_monad" where
  "write_hashes_partial_restart ht \<equiv> \<lambda>(sv, sp).
    {(None, sv, sp')|sp'.
        hstate sp \<ge> LDegraded
        \<and> sp' = sp\<lparr> hash := ht \<rparr>
  }"

definition write_hashes :: 
  "'c htable2 \<Rightarrow> (unit, pmm_vol_mem, 'c)abs2_monad" where
  "write_hashes ht \<equiv>
  write_hashes_success ht
  [+] write_hashes_degrade ht
  [+] write_hashes_degrade_write ht
  [+] write_hashes_fail ht
  [+] write_hashes_partial_restart ht
  [+] restart_mangle_sv"

lemmas write_hashes_defs =
  write_hashes_def
  write_hashes_success_def
  write_hashes_degrade_def
  write_hashes_degrade_write_def
  write_hashes_fail_def
  write_hashes_partial_restart_def

definition read_DegradedH :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)abs2_monad" where
   read_DegradedH_def[monad_def]: 
  "read_DegradedH ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> get_hashes_DegradedH;
    (if \<not> pub_matches_htable2 ht ind inst cnt
        then throw FailRead
        else return cnt)
  }"

definition read_DegradedH_or_fail
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)abs2_monad" where
   read_DegradedH_or_fail_def[monad_def]: 
  "read_DegradedH_or_fail ind inst \<equiv>
  <try> read_DegradedH ind inst <catch> K (do { set_failed; throw FailMode })"

definition read :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)abs2_monad" where
   read_def[monad_def]:
  "read ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> (
        <try>
            get_hashes
        <catch> (\<lambda>e. case e of
            Degraded _ \<Rightarrow> do {
                ht \<leftarrow> get_hashes_DegradedH;
                (if pub_matches_htable2 ht ind inst cnt
                    then throw (Degraded (EC cnt))
                    else throw (Degraded EU)
                )
            }
            | _ \<Rightarrow> throw e
        )
    );
    (if \<not> pub_matches_htable2 ht ind inst cnt
        then throw FailRead
        else return cnt)
  }"

definition read_ptr_DegradedH :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  read_ptr_DegradedH_def[monad_def]:
  "read_ptr_DegradedH inst \<equiv> do {
    ptr_set \<leftarrow> (
        <try>
            ll_read_ptr inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> get_hashes_DegradedH;
    (if \<not> ptr_matches_htable2 ht inst ptr_set
        then throw FailRead
        else return ptr_set)
  }"

definition read_ptr :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  read_ptr_def[monad_def]:
  "read_ptr inst \<equiv> do {
    ptr_set \<leftarrow> (
        <try>
            ll_read_ptr inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> (
        <try>
            get_hashes
        <catch> (\<lambda>e. case e of
            Degraded _ \<Rightarrow> do {
                ht \<leftarrow> get_hashes_DegradedH;
                (if ptr_matches_htable2 ht inst ptr_set
                    then throw (Degraded (EC (serialize_ptr ptr_set)))
                    else throw (Degraded EU)
                )
            }
            | _ \<Rightarrow> throw e
        )
    );
    (if \<not>ptr_matches_htable2 ht inst ptr_set
        then throw FailRead
        else return ptr_set)
  }"

definition read_or_fail :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  read_or_fail_def[monad_def]:
  "read_or_fail pind inst \<equiv> 
    <try>
        read pind inst
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do { set_failed; throw FailMode }
            else throw (Degraded v)
        | _ \<Rightarrow> do { set_failed; throw FailMode }
    )
  "

definition read_ptr_or_fail :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  read_ptr_or_fail_def[monad_def]:
  "read_ptr_or_fail inst \<equiv> 
    <try>
        read_ptr inst
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do { set_failed; throw FailMode }
            else throw (Degraded v)
        | _ \<Rightarrow> do { set_failed; throw FailMode }
    )
  "

definition read_or_fail_silence_deg
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where read_or_fail_silence_deg_def[monad_def]:
  "read_or_fail_silence_deg ind inst \<equiv>
    <try> read_or_fail ind inst <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (get_exc_val_cnt v)
        | _ \<Rightarrow> throw e
    )
  "
 
definition read_ptr_or_fail_silence_deg
  :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where read_ptr_or_fail_silence_deg_def[monad_def]:
  "read_ptr_or_fail_silence_deg inst \<equiv>
    <try> read_ptr_or_fail inst <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (parse_ptr (get_exc_val_cnt v))
        | _ \<Rightarrow> throw e
    )
  "

definition update_pub_instance_hash :: 
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, pmm_vol_mem, 'c)abs2_monad" where
  update_pub_instance_hash_def[monad_def]:
  "update_pub_instance_hash ind inst cnt \<equiv> do {
        ht \<leftarrow> get_hashes;
        write_hashes (pub_update_htable2 ht ind inst cnt)
  }
  "

definition update_ptr_instance_hash :: 
  "INST \<Rightarrow> pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  update_ptr_instance_hash_def[monad_def]:
  "update_ptr_instance_hash inst ptr_set \<equiv> do {
        ht \<leftarrow> get_hashes;
        write_hashes (ptr_update_htable2 ht inst ptr_set)
  }
  "

definition write_instance :: 
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  write_instance_def[monad_def]:
  "write_instance ind inst cnt \<equiv> do {
    mode \<leftarrow> get_mode;
    (if mode = CDegradedH
    then throw (Degraded (EB False))
    else return());
    (<try>
        write_f ind inst cnt
    <catch> (\<lambda>e. case e of
        InvalidInput \<Rightarrow> throw InvalidInput
        | FailWrite \<Rightarrow> do {
            set_degraded;
            (<try> 
                read ind inst
            <catch> (\<lambda>e. case e of
                Degraded v \<Rightarrow>
                    if v = EU
                    then throw (Degraded (EB False))
                    else throw (Degraded (EB True))
                | FailRead \<Rightarrow> throw (Degraded (EB False))
                | _ \<Rightarrow> throw e)
            );
            throw (Degraded (EB True))
        }
        | _ \<Rightarrow> throw e
    ));
    <try>
        update_pub_instance_hash ind inst cnt
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then throw (Degraded v)
            else
                do {
                      cnt \<leftarrow> (
                          <try>
                              read_DegradedH ind inst
                          <catch> (\<lambda>e. case e of
                              FailMode \<Rightarrow> throw FailMode
                              | _ \<Rightarrow> throw (Degraded (EB False))
                          )
                      );
                      throw (Degraded (EB True))
                  }
        | _ \<Rightarrow> throw e
    )
  }"


definition write_ptr_instance :: 
  "INST \<Rightarrow> pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  write_ptr_instance_def[monad_def]:
  "write_ptr_instance inst ptr_set \<equiv> do {
    mode \<leftarrow> get_mode;
    (if mode = CDegradedH
    then throw (Degraded (EB False))
    else return());
    (<try>
        write_ptr_ll inst ptr_set
    <catch> K (do { set_degraded; throw (Degraded (EB False)) })
    );
    <try>
        update_ptr_instance_hash inst ptr_set
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then throw (Degraded v)
            else
                do {
                    (<try>
                        read_ptr_DegradedH inst
                    <catch> (\<lambda>e. case e of
                        FailMode \<Rightarrow> throw FailMode
                        | _ \<Rightarrow> throw (Degraded (EB False))
                    ));
                    throw (Degraded (EB True))
                }
        | _ \<Rightarrow> throw e
    )
  }"

definition sync :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  sync_def[monad_def]:
  "sync ind to from \<equiv> do {
    cnt \<leftarrow> (
        <try>
            read_or_fail ind from
        <catch> (\<lambda>e. case e of
            Degraded v  \<Rightarrow> throw (Degraded (EC (get_exc_val_cnt v)))
            | _ \<Rightarrow> throw e
        )
    );
    (<try>
        write_instance ind to cnt
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> throw (Degraded (EC cnt))
        | _ \<Rightarrow> throw e
    ));
    return cnt
  }"

definition sync_ptr :: "INST \<Rightarrow> INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  sync_ptr_def[monad_def]:
  "sync_ptr to from \<equiv> do {
    ptr_set \<leftarrow> (
        <try>
            read_ptr_or_fail from
        <catch> (\<lambda>e. case e of
            Degraded v  \<Rightarrow> throw (Degraded (EC (get_exc_val_cnt v)))
            | _ \<Rightarrow> throw e
        )
    );
    (<try>
        write_ptr_instance to ptr_set
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> do { set_degraded_p; throw (Degraded (EC (serialize_ptr ptr_set))) }
        | _ \<Rightarrow> throw e
    ));
    return ptr_set
  }"

definition just_sync
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where just_sync_def[monad_def]:
  "just_sync ind to from \<equiv> sync ind to from \<guillemotright> return ()"

definition just_sync_ptr :: "INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  just_sync_ptr_def[monad_def]:
  "just_sync_ptr to from \<equiv> sync_ptr to from \<guillemotright> return ()"

definition just_sync_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where just_sync_silence_deg_def[monad_def]:
  "just_sync_silence_deg to from ind \<equiv>
    <try> 
        just_sync ind to from
    <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "

definition get_ptr :: "bool \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl) abs2_monad" where
  get_ptr_def[monad_def]:
  "get_ptr silence_deg \<equiv>
    <try> 
        read_ptr Secondary
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do {
                ptr_set \<leftarrow> read_ptr_or_fail_silence_deg Primary;
                set_degraded_p;
                if silence_deg
                then return ptr_set
                else throw (Degraded EU)
            }
            else
                if silence_deg
                then return (parse_ptr (get_exc_val_cnt v))
                else throw (Degraded EU)
        | FailRead \<Rightarrow> 
                    <try>
                        sync_ptr Secondary Primary
                    <catch> (\<lambda>e. case e of
                        Degraded v \<Rightarrow>
                            if silence_deg
                            then return (parse_ptr (get_exc_val_cnt v))
                            else throw (Degraded EU)
                        | _ \<Rightarrow> throw e
                    )
        | _ \<Rightarrow> throw e
    )
  "

definition write_ptr :: "pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  write_ptr_def[monad_def]:
  "write_ptr ptr_set \<equiv> 
  do {
    mode \<leftarrow> get_mode;
    (if mode \<in> {CFail, CDegradedH, CDegradedP}
    then throw (Degraded (EB False))
    else return ());
    (<try>
        write_ptr_instance Secondary ptr_set
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            do {
                (if v = EB True
                then <try> write_ptr_ll Primary ptr_set <catch> K (return ())
                else read_ptr_or_fail_silence_deg Primary \<guillemotright> return ()
                );
                set_degraded_p;
                throw (Degraded (if v = EB True then v else EB False))
            }
        | _ \<Rightarrow> throw e
    ));
    (<try>
        write_ptr_instance Primary ptr_set
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> do { set_degraded_p; throw (Degraded (EB True)) }
        | _ \<Rightarrow> throw e
    ))
  }
  "

definition remove_from_ptr :: "pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  remove_from_ptr_def[monad_def]:
  "remove_from_ptr ind \<equiv> do {
    ptr \<leftarrow> get_ptr False;
    let ptr' = ptr - {ind};
    write_ptr ptr'
  }"

definition set_ptr :: "pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  set_ptr_def[monad_def]:
  "set_ptr ptr \<equiv> do {
    get_ptr False;
    write_ptr ptr
  }"

definition get_hashes_silence_deg
  :: "('c htable2, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  get_hashes_silence_deg_def[monad_def]:
  "get_hashes_silence_deg \<equiv>
    <try>
        get_hashes
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> get_hashes_DegradedH
        | _ \<Rightarrow> throw e
    )
  "

definition get_sync_direction
  :: "pub_indx \<Rightarrow> (sync_direction, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where
  get_sync_direction_def[monad_def]:
  "get_sync_direction pind \<equiv> do {
    ht \<leftarrow> get_hashes_silence_deg;
    let pub_h = pub_hashes2 ht;
    if (pub_h (pind, Primary) = pub_h (pind, Secondary))
    then (return SEq)
    else do {
        ptr_set \<leftarrow> get_ptr True;
        (if pind \<in> ptr_set
        then return SL
        else return SR)
    }
  }"

definition remove_from_ptr_silence_deg where
  remove_from_ptr_silence_deg_def[monad_def]:
  "remove_from_ptr_silence_deg ind \<equiv>
        <try>
            remove_from_ptr ind
        <catch> (\<lambda>e. if e = FailMode then throw FailMode else return ())
  "

definition sync_and_remove_from_ptr_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad"
  where sync_and_remove_from_ptr_silence_deg_def[monad_def]:
  "sync_and_remove_from_ptr_silence_deg to from ind \<equiv>
    <try> do {
        sync ind to from;
        remove_from_ptr_silence_deg ind
    } <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> return ()
        | _ \<Rightarrow> throw e
    )
  "

end