theory Abs2_SpecRef
  imports "../abs1/Abs1_Specification" "../abs1/Abs1_Invariant_Lemmas" 
  Abs2_Specification Abs2_Ref Abs2_Startup_Lemmas "../SpecRef_Lemmas"
begin

fun is_some :: "'a option \<Rightarrow> bool"
  where
  "is_some (Some _) = True"
  | "is_some _ = False"

fun abs1_abs2_pi :: "('sv \<times> 'c abs1_obs_state) \<Rightarrow> ('sv \<times> 'c abs2_obs_state)" 
  where
  "abs1_abs2_pi (svc, spc) = (svc, 
    \<lparr> 
        abs2_obs_state.hstate = 
            combine_ec_states (abs1_obs_state.hash1_state spc) (abs1_obs_state.hash2_state spc)
        , abs2_obs_state.hash = 
            (if abs1_obs_state.hash1 spc = None \<and> abs1_obs_state.hash2 spc = None
            then None
            else (if is_some (abs1_obs_state.hash1 spc)
                    then abs1_obs_state.hash1 spc 
                    else abs1_obs_state.hash2 spc)
            )
        , abs2_obs_state.ptr1 = abs1_obs_state.ptr1 spc
        , abs2_obs_state.ptr1_state = abs1_obs_state.ptr1_state spc
        , abs2_obs_state.ptr2 = abs1_obs_state.ptr2 spc
        , abs2_obs_state.ptr2_state = abs1_obs_state.ptr2_state spc
        , abs2_obs_state.memory = abs1_obs_state.memory spc
        , abs2_obs_state.hw_state = abs1_obs_state.hw_state spc
    \<rparr>
  )"

lemma abs1_startup1'_SN_incl:
  "\<forall>sc\<in>abs1_init.
  ((Abs1_StartupDefs.startup1 \<guillemotright>= 
    (\<lambda>_. Abs1_StartupDefs.startup2 :: (unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad))\<Down>SN) sc
  \<subseteq> ((abs1_startup1' \<guillemotright>= 
    (\<lambda>_. Abs1_StartupDefs.startup2 :: (unit, pmm_vol_mem, 'c :: contents_cl)abs1_monad))\<Down>SN) sc"
  apply(rule ballI)
  apply(simp add: projectSN_def)
  apply(intro projectS_mono fin_imp_bind_NN_incl abs1_startup1_fin)
  done

lemma Collect_pair_False:
  "{(a, b). False} = {}"
  by(simp)

lemma abs2_state_equality:
  "((x :: 'c abs2_state) = (y :: 'c abs2_state)) = (
    abs2_state.hstate x = abs2_state.hstate y
    \<and> abs2_state.hash x = abs2_state.hash y
    \<and> abs2_state.ptr1 x = abs2_state.ptr1 y
    \<and> abs2_state.ptr1_state x = abs2_state.ptr1_state y
    \<and> abs2_state.ptr2 x = abs2_state.ptr2 y
    \<and> abs2_state.ptr2_state x = abs2_state.ptr2_state y
    \<and> abs2_state.memory x = abs2_state.memory y
    \<and> abs2_state.hw_state x = abs2_state.hw_state y
  )"
  by(case_tac x, case_tac y, simp)

lemma comb_EcOK_LDegraded':
  "(combine_ec_states x EcOK = LDegraded) = (x \<le> EcUnwr)"
  by(case_tac x, simp_all)

lemma silly:
  "(\<lambda>(sv, sp) y. {(Some (Inr ()), sv \<lparr>card_mode := CNormal, updated := {}\<rparr>, sp)}) x x
    = {(Some (Inr ()), (fst x) \<lparr>card_mode := CNormal, updated := {}\<rparr>, (snd x))}"
  by (metis (lifting, no_types) prod_case_beta)

lemma ex_abs2_state:
  "(\<exists>b. P b) = 
    (\<exists>hstate hash ptr1 ptr1_state ptr2 ptr2_state memory hw_state.
        P \<lparr> abs2_state.hstate = hstate, abs2_state.hash = hash, 
            abs2_state.ptr1 = ptr1, abs2_state.ptr1_state = ptr1_state, 
            abs2_state.ptr2 = ptr2, abs2_state.ptr2_state = ptr2_state, abs2_state.memory = memory, 
            abs2_state.hw_state = hw_state
    \<rparr>)"
  apply(auto)
  apply(case_tac b, force)
  done
 
fun abs1_abs2_int_map :: "('sv \<times> 'c abs1_state) \<Rightarrow> ('sv \<times> 'c abs2_state)" 
  where
  "abs1_abs2_int_map (svc, spc) = (svc, 
    \<lparr> 
        abs2_state.hstate = 
            combine_ec_states (abs1_state.hash1_state spc) (abs1_state.hash2_state spc)
        , abs2_state.hash = (if abs1_state.hash1_state spc \<noteq> EcOK \<and> abs1_state.hash2_state spc \<noteq> EcOK
            then undefined
            else (if abs1_state.hash1_state spc = EcOK 
                then abs1_state.hash1 spc else abs1_state.hash2 spc)
            )
        , abs2_state.ptr1 = abs1_state.ptr1 spc
        , abs2_state.ptr1_state = abs1_state.ptr1_state spc
        , abs2_state.ptr2 = abs1_state.ptr2 spc
        , abs2_state.ptr2_state = abs1_state.ptr2_state spc
        , abs2_state.memory = abs1_state.memory spc
        , abs2_state.hw_state = abs1_state.hw_state spc
    \<rparr>
  )"

lemma forall_INST:
  "(\<forall>(x :: INST). P x) = (P Primary \<and> P Secondary)"
  apply(auto)
  apply(case_tac x, auto)
  done

lemma abs1_init_NN_incl:
    "set_fun_into_rel (abs1_startup1'\<Down>SNN) `` abs1_init
    \<subseteq> abs2_abs1_strel `` set_fun_into_rel (Abs2_StartupDefs.startup1\<Down>SNN) `` abs2_init"
  apply(simp_all add: projectSNN_def projectS_def abs1_startup1'_def 
    Abs2_StartupDefs.startup1_def
    project_simps
    abs1_check_hash_region'_def compose_project)
  apply(auto simp add: set_fun_into_rel_def Image_def projectS_defs image_def)
  apply(clarsimp simp add: brute_force silly)
  apply(rename_tac svc spc svc' spc')
  apply(rule_tac x="svc'" in exI)
  apply(rule_tac x="snd (abs1_abs2_int_map (undefined, spc'))" in exI)
  apply(auto simp add: abs2_abs1_strel_def abs2_abs1_strel_core_def)
  apply(simp_all add: abs1_same_rest_def abs1_init_def abs2_init_def
    Abs2_StartupDefs.check_hash_region_defs
    check_hash_region_NN_def
    min_def brute_force forall_INST
    split: split_if_asm
  )
  apply(simp_all add: HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(clarsimp_all simp add: card_mode_full_upd Collect_pair_False abs2_state_equality
    comb_EcOK_LDegraded'
    comb_EcOK_is_LDegraded)
  apply(clarsimp_all simp add: ex_abs2_state combine_ec_states_simps
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    conj_commute conj_left_commute state_ineq_simps)
  apply(case_tac[!] inst, clarsimp_all)
  done

lemma abs1_init_NE_incl:
  "set_fun_into_rel (abs1_startup1'\<Down>SNE) `` abs1_init
    \<subseteq> abs2_abs1_strel ``
      set_fun_into_rel (Abs2_StartupDefs.startup1\<Down>SNE) `` abs2_init"
  apply(simp_all add: projectSNE_def projectS_def abs1_startup1'_def 
    Abs2_StartupDefs.startup1_def
    project_simps
    abs1_check_hash_region'_def compose_project)
  apply(auto simp add: set_fun_into_rel_def Image_def projectS_defs image_def)
  apply(clarsimp simp add: brute_force silly)
  apply(rename_tac svc spc svc' spc')
  apply(auto simp add: abs2_abs1_strel_def abs2_abs1_strel_core_def)
  apply(simp_all add: abs1_same_rest_def abs1_init_def abs2_init_def
    Abs2_StartupDefs.check_hash_region_defs
    check_hash_region_NE_def
    min_def brute_force
    forall_INST
    split: split_if_asm
  )
  apply(simp_all add: HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  using[[simproc del: Product_Type.split_beta]]
  apply(iterative_simp)+
  apply(clarsimp_all simp add: card_mode_full_upd Collect_pair_False abs2_state_equality
    comb_EcOK_LDegraded'
    comb_EcOK_is_LDegraded)
  apply(case_tac[!] "abs1_state.hash1_state spc'", case_tac[!] "abs1_state.hash2_state spc'", simp_all)
  apply (metis abs2_state.select_convs(1) abs2_state.select_convs(3) 
    abs2_state.select_convs(4) abs2_state.select_convs(5) abs2_state.select_convs(6) 
    abs2_state.select_convs(7) abs2_state.select_convs(8))+
  done

lemma repeatS_inv:
  "\<lbrace>P\<rbrace> m \<lbrace>> K P\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> repeatS (m\<Down>R) \<lbrace>S P\<rbrace>"
  apply(auto simp add: Hoare_def HoareS_def Box_def BoxS_def)
  apply(erule repeatS.induct)
  apply(auto simp add: K_beta project_defs)
  done

lemma abs1_remove_repeat:
  "
  set_fun_into_rel ((rec_tryR 
  (Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad))\<Down>SN) `` abs1_init
  \<subseteq>
  set_fun_into_rel ((Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad)\<Down>SN) `` 
  abs1_init
  "
proof(clarsimp simp add: projectS_defs set_fun_into_rel_def image_def)
  fix sv and sp :: "('c :: contents_cl) abs1_state" and v sv' sp'
  assume 
    in_init: "(sv, sp) \<in> abs1_init" 
    and in_rep: "(Some v, sv', sp') \<in> rec_tryR 
    (Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad) (sv, sp)"
  from in_rep[THEN rec_tryRS_Some, simplified bindS_decompose, simplified] 
  obtain sv'' sp'' where mid:
    "(sv'', sp'') \<in> repeatS 
    ((Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad)\<Down>R) (sv, sp) \<and>
    (v, sv', sp') \<in> 
    ((Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad)\<Down>N) (sv'', sp'')" 
    by(auto)
  have in_init_mid:
    "(sv'', sp'') \<in> abs1_init" using in_init mid Abs1_Invariant_Lemmas.startup_init[THEN repeatS_inv]
    by(auto simp add: HoareS_def BoxS_def)
  hence "(sv'', sp'') \<in> abs1_init \<and> (Some v, sv', sp') \<in> (Abs1_StartupDefs.startup ) (sv'', sp'')"
    using mid
    by(auto simp add: project_defs)
  thus "\<exists>x\<in>abs1_init. \<exists>a. (Some a, sv', sp') \<in> 
    (Abs1_StartupDefs.startup :: ('a, pmm_vol_mem, 'c :: contents_cl)abs1_monad) x"
    by(auto)
qed

theorem abs1_refines_abs2:
  "refines abs2_abs1_strel abs1_abs2_pi abs2_spec abs1_spec"
proof(intro refinesI)
  show "obs_consistent abs2_abs1_strel abs1_abs2_pi abs2_spec abs1_spec"
    apply(auto simp add: obs_consistent_def abs2_spec_def abs1_spec_def 
      abs2_abs1_strel_def abs2_abs1_strel_core_def obs_ts)
    apply(case_tac[!] "abs1_state.hash1_state ba", case_tac[!] "abs1_state.hash2_state ba", auto)
    done
next
  show "PO_refines abs2_abs1_strel abs2_spec abs1_spec"
  proof(rule refine_basic)
    show "init_trans abs1_spec `` init abs1_spec
        \<subseteq> abs2_abs1_strel `` init_trans abs2_spec `` init abs2_spec"
      apply(simp add: abs1_spec_def abs1_ts_def abs2_spec_def abs2_ts_def TS.defs spec.defs)
      apply(rule subset_trans[OF abs1_remove_repeat])
      apply(simp only: Abs2_StartupDefs.startup_def Abs1_StartupDefs.startup_def)
      apply(rule init_conc_mono_cond[where 
        mc'="((abs1_startup1' \<guillemotright>= (\<lambda>_. Abs1_StartupDefs.startup2))\<Down>SN)", OF _ abs1_startup1'_SN_incl])
      apply(intro sim_norm_init_bind_inclusion[where R="abs2_abs1_strel"]
        Abs2_Ref.startup2_ref_N[simplified sim_norm_def] abs1_init_NN_incl abs1_init_NE_incl)
      done
  next
    show "{abs2_abs1_strel} TS.trans abs2_spec, TS.trans abs1_spec {> abs2_abs1_strel}"
      apply(simp add: abs2_spec_def abs2_ts_def abs1_spec_def abs1_ts_def spec.defs TS.defs 
        abs2_trans_def abs1_trans_def)
      apply(intro relhoare_UN_lockstep allI relhoare_set_fun_choice_lockstep)
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare simp only: sim_norm_def[THEN symmetric])
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare 
        Abs2_Ref.fetch_full_ref
        Abs2_Ref.update_full_ref
        simp add: try_rec_tryR_def)
      apply(auto intro!: sim_norm_try_catchRI trans_ref_fin_catchR
        Abs2_Ref.restart_mangle_sv_refines_itself 
        simp add: try_rec_tryR_def
        Abs2_StartupDefs.startup_def Abs1_StartupDefs.startup_def
        intro!:
        Abs2_Ref.startup1_ref Abs2_Ref.startup2_ref_N 
        Abs2_Ref.startup2_ref_R_N Abs1_Startup_Lemmas.abs1_startup1_fin
      )
      done
  qed
qed

theorem abs1_implements_abs2:
  "implements abs1_abs2_pi abs2_spec abs1_spec"
  by(auto intro: refinement_soundness abs1_refines_abs2)

end
