theory Abs2_StartupDefs
imports Abs2_CommonDefs
begin

definition degrade_hash_region :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "degrade_hash_region \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv', sp')|sv' sp'.
        hstate sp \<noteq> LDead
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded \<rparr>
  }"

definition fail_hash_region :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "fail_hash_region \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp'.
        sv' = sv\<lparr> card_mode := CFail \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDead \<rparr>
  }"

definition check_hash_region_success ::  "(unit, 'sv, 'c :: contents_cl)abs2_monad" where
  "check_hash_region_success \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        hstate sp = LOK
        \<and> sp' = sp
  }"

definition check_hash_region :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  "check_hash_region \<equiv>
    check_hash_region_success [+] fail_hash_region [+] degrade_hash_region
  "

lemmas check_hash_region_defs =
  check_hash_region_def
  degrade_hash_region_def
  fail_hash_region_def
  check_hash_region_success_def

definition check_ptr :: "(unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  check_ptr_def[monad_def]:
  "check_ptr \<equiv> <try> do {
    ptr_set2 \<leftarrow> read_ptr Secondary;
    <try> do {
        ptr_set1 \<leftarrow> read_ptr Primary;
        (if (ptr_set2 \<noteq> ptr_set1)
            then
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded v \<Rightarrow>
                        if v = EB False
                        then
                            do {
                                set_degraded_p;
                                <try> 
                                    write_ptr_ll Primary ptr_set2
                                <catch> K (return ())
                            }
                        else return ()
                    | _ \<Rightarrow> throw e
                )
            else return ())
    }
    <catch> 
        (\<lambda>e. case e of 
            FailRead \<Rightarrow>
                <try>
                    write_ptr_instance Primary ptr_set2
                <catch> (\<lambda>e. case e of
                    Degraded _ \<Rightarrow> set_degraded_p
                    | _ \<Rightarrow> throw e
                )
            | Degraded v \<Rightarrow>
                if v = EU
                then set_degraded_p
                else
                    if (parse_ptr (get_exc_val_cnt v) \<noteq> ptr_set2)
                    then <try> write_ptr_ll Primary ptr_set2 <catch> K (return ())
                    else return ()
            | _ \<Rightarrow> throw e
        )
  }
  <catch> (\<lambda>e. case e of 
    Degraded v \<Rightarrow>
        if v = EU
        then do { read_ptr_or_fail_silence_deg Primary; set_degraded_p }
        else
            <try>
                write_ptr_ll  Primary (parse_ptr (get_exc_val_cnt v))
            <catch> K (return ())
    | FailRead \<Rightarrow>
        <try>
            just_sync_ptr Secondary Primary
        <catch> (\<lambda>e. case e of Degraded _ \<Rightarrow> return () | _ \<Rightarrow> throw e)
    | _ \<Rightarrow> throw e)
  "

definition check_pub_same
    :: "pub_indx  \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  check_pub_same_def[monad_def]:
  "check_pub_same pind \<equiv>
    <try> do {
        read pind Primary;
        (<try> do {
            read pind Secondary;
            return ()
        } <catch>
            (\<lambda>e. case e of
                FailRead \<Rightarrow> just_sync_silence_deg Secondary Primary pind
                | Degraded _ \<Rightarrow> return ()
                | _ \<Rightarrow> throw e
            )
        )
   } <catch> (\<lambda>e. case e of
        FailRead \<Rightarrow> just_sync_silence_deg Primary Secondary pind 
        | Degraded v \<Rightarrow>
            if v = EU
            then read_or_fail_silence_deg pind Secondary \<guillemotright> return ()
            else return ()
        | _ \<Rightarrow> throw e
    )
  "

definition check_pub_region
  :: "pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  check_pub_region_def[monad_def]:
  "check_pub_region pind \<equiv> do {
    dir \<leftarrow> get_sync_direction pind;
    if dir = SEq
    then do { check_pub_same pind; remove_from_ptr_silence_deg pind }
    else
        if dir = SR
        then sync_and_remove_from_ptr_silence_deg Secondary Primary pind 
        else sync_and_remove_from_ptr_silence_deg Primary Secondary pind 
  }"
 
definition check_all_public :: "
  (unit, pmm_vol_mem, 'c :: contents_cl) abs2_monad" where
  check_all_public_def[monad_def]:
  "check_all_public \<equiv> do {
    l \<leftarrow> chooseN (order_set UNIV);
    mapM' check_pub_region l
  }"

definition startup1 where
  startup1_def[monad_def]:
  "startup1 \<equiv> do {
    init_vars; 
    check_hash_region
  }
  "

definition startup2 where
  startup2_def[monad_def]:
  "startup2 \<equiv> do {
     check_ptr;
     check_all_public;
     throw_mode_error EU;
     throw StartedExc
   }"
  
definition startup :: "('a, pmm_vol_mem, 'c :: contents_cl)abs2_monad" where
  startup_def[monad_def]:
  "startup \<equiv> do {
    startup1;
    startup2
  }"

end
