theory Abs2_Startup_Lemmas
imports "../Projections" "../Pmm_Tacs" "../HoareLogic" "../Pmm_Hoare"
  "../Pmm_Refinement" "../Pmm_Lemmas"
  Abs2_Lemmas
  Abs2_CommonDefs Abs2_StartupDefs
  uses ("trans_simproc.ML")
begin

lemmas brute_force =
  unfold_monad
  project_defs
  vol_ops_defs
  Abs2_LLInterface.ll_read_defs
  Abs2_LLInterface.ll_read_ptr_defs
  Abs2_LLInterface.write_f_defs
  Abs2_LLInterface.write_ptr_ll_defs
  restart_mangle_sv_raw
  getters_setters

declare disjCI[rule del]
declare UnCI[rule del]

lemma infinite_has_two': "\<forall>cnt' :: 'a. cnt' = cnt \<Longrightarrow> finite (UNIV :: 'a set)"
proof-
  fix cnt :: 'a
  def U \<equiv> "UNIV :: 'a set" (* seems to be the only way to disambiguate UNIV? *)
  assume "\<forall>cnt'. cnt' = cnt"
  hence "U = insert cnt {}" using U_def
    by auto 
  hence "finite U"
    by(force intro!: Finite_Set.finite.insertI)
  thus ?thesis using U_def
    apply(auto)
    done
qed

lemma infinite_has_two: "\<exists>cnt' :: 'a :: infinite. cnt' \<noteq> cnt"
  apply(insert infinite_UNIV)
  apply(erule contrapos_np)
  apply(blast intro!: infinite_has_two')
  done

lemma infinite_has_two_rev: 
  "\<exists>cnt' :: 'a :: infinite. cnt \<noteq> cnt'"
proof-
  from infinite_has_two obtain cnt' where "cnt' \<noteq> cnt" ..
  thus ?thesis by auto
qed

lemma check_hash_region_project_simps:
  "check_hash_region\<Down>R = empty"
  "fail_hash_region\<Down>NN = empty"
  "check_hash_region_success\<Down>NE = empty"
  "degrade_hash_region\<Down>NE = empty"
  by(auto simp add: check_hash_region_defs init_vars_def brute_force intro!: ext)

fun ptr_matches_hash :: "'c abs2_state \<Rightarrow> INST \<Rightarrow> bool" where
  "ptr_matches_hash sp inst =
        (ptr_state_by_inst inst sp  \<noteq> IDead \<and> hstate sp \<noteq> LDead
            \<and> (ptr_hashes2 (hash sp) inst = ptr_by_inst inst sp))
  "

fun abs2_unwr_bad_hash_by_inst :: "'c abs2_state \<Rightarrow> 'c abs2_state \<Rightarrow> INST \<Rightarrow> bool" where
  "abs2_unwr_bad_hash_by_inst sp sp' inst =
    (ptr_state_by_inst inst sp' = IOK \<or> \<not> ptr_matches_hash sp' inst \<or>
    (ptr_state_by_inst inst sp' = IUnwr \<and> ptr_state_by_inst inst sp = IUnwr \<and>
        ptr_matches_hash sp inst \<and>
        ptr_by_inst inst sp' = ptr_by_inst inst sp))
  "

fun abs2_unwr_bad_hash :: "'c abs2_state \<Rightarrow> 'c abs2_state \<Rightarrow> bool" where
  "abs2_unwr_bad_hash sp sp' =
    (abs2_unwr_bad_hash_by_inst sp sp' Primary
    \<and> abs2_unwr_bad_hash_by_inst sp sp' Secondary)"

lemma abs2_unw_bad_hash_trans:
  "\<lbrakk> abs2_unwr_bad_hash sp sp'; abs2_unwr_bad_hash sp' sp'' \<rbrakk> \<Longrightarrow>
    abs2_unwr_bad_hash sp sp''"
  apply(clarsimp)
  by smt
 
definition abs2_ptr_same_rest :: "'c abs2_state \<Rightarrow> 'c abs2_state \<Rightarrow> bool" where
  "abs2_ptr_same_rest sp sp' \<equiv>
            memory sp' = memory sp
            \<and> hw_state sp' = hw_state sp
  "

definition read_ptr_NN where
  "read_ptr_NN inst \<equiv> 
    (\<lambda>(sv, sp). {(r, sv, sp)|r.
        ptr_matches_hash sp inst
        \<and> r = ptr_by_inst inst sp
    })
  "

definition write_ptr_instance_NN 
  :: "INST \<Rightarrow> pub_indx set \<Rightarrow> (unit, pmm_vol_mem \<times> ('c :: contents_cl) abs2_state)nds_monad"
  where
  "write_ptr_instance_NN inst ptr_set \<equiv> 
    (\<lambda>(sv, sp). {((), sv, sp')|sp'.
        hstate sp = LOK
        \<and> card_mode sv \<noteq> CDegradedH
        \<and> ptr_state_by_inst inst sp = IOK
        \<and> sp' = (if inst = Primary
            then sp\<lparr> ptr1 := ptr_set, hash :=
                    HTable2 (pub_hashes2 (hash sp)) ((ptr_hashes2 (hash sp))(Primary := ptr_set))
                \<rparr>
            else sp\<lparr> ptr2 := ptr_set, hash :=
                    HTable2 (pub_hashes2 (hash sp)) ((ptr_hashes2 (hash sp))(Secondary := ptr_set))
                \<rparr>
        )
    })
  "

lemma state_ineq_simps:
  "hs \<noteq> LDead = (hs = LDegraded \<or> hs = LOK)"
  "is \<ge> IUnwr = (is = IUnwr \<or> is = IOK)"
  "is \<noteq> IDead = (is = IUnwr \<or> is = IOK)"
  apply(case_tac hs, auto)
  apply(case_tac "is", auto)+
  done

lemma read_ptr_NN:
  "(read_ptr inst)\<Down>NN = read_ptr_NN inst"
  apply(rule ext)
  apply(auto simp only: read_ptr_def first_stage Let_def)
  apply(clarsimp_all simp add: brute_force read_ptr_NN_def get_hashes_defs
    hw_state_simps 
    split: split_if_asm split_if)
  apply(iterative_simp simp add: hw_state_simps hstate_simps state_ineq_simps)+
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: ptr_matches_htable2_def)
  done

lemma write_ptr_instance_NN:
  "(write_ptr_instance inst cnt)\<Down>NN = write_ptr_instance_NN inst cnt"
  apply(rule ext)
  apply(auto simp only: write_ptr_instance_def update_ptr_instance_hash_def first_stage Let_def)
  apply(clarsimp_all simp add: brute_force write_ptr_instance_NN_def
    get_hashes_defs write_hashes_defs
    hw_state_simps
    split: split_if_asm split_if)
  apply(case_tac[!] "hash b", clarsimp_all)
  done

declare doesnt_throw_raw_if[no_throw]
declare doesnt_throw_raw_exceptions_case[no_throw]
declare doesnt_throw_empty[no_throw]
declare doesnt_throw_raw_empty[no_throw]
declare returnN_throws_only_one[no_throw]
declare throw_throws_only_one[no_throw]

declare doesnt_throw_exceptions_case[no_throw]

lemma push_hstate_to_front:
  "s\<lparr> ptr1 := a, hstate := b \<rparr> = s\<lparr> hstate := b, ptr1 := a \<rparr>"
  "s\<lparr> ptr2 := a, hstate := b \<rparr> = s\<lparr> hstate := b, ptr2 := a \<rparr>"
  by(auto)

lemma choice_mono1:
  "ml' s \<subseteq> ml s \<Longrightarrow> (ml' [+] m) s \<subseteq> (ml [+] m) s"
  by(auto simp add: choice_def)

lemma choice_mono2:
  "(mr' s \<subseteq> mr s) \<Longrightarrow> (m [+] mr') s \<subseteq> (m [+] mr) s"
  by(auto simp add: choice_def)

definition ll_read_ptr_failure_NE ::
  "INST \<Rightarrow> ('c exceptions, 'sv \<times> 'c :: contents_cl abs2_state)nds_monad"
  where
  "ll_read_ptr_failure_NE inst \<equiv> ((ll_read_ptr_failure :: INST \<Rightarrow> ('c, 'sv, 'c)abs2_monad) inst)\<Down>NE"

definition ll_read_ptr_any_NN ::
  "INST \<Rightarrow> (pub_indx set, 'sv \<times> 'c :: contents_cl abs2_state)nds_monad"
  where
  "ll_read_ptr_any_NN inst \<equiv> ((ll_read_ptr_any :: INST \<Rightarrow> (pub_indx set, 'sv, 'c)abs2_monad) inst)\<Down>NN"

lemma ll_read_ptr_proj_simps:
  "(ll_read_ptr inst)\<Down>NE = ll_read_ptr_failure_NE inst"
  "(ll_read_ptr inst)\<Down>NN = ll_read_ptr_any_NN inst"
  apply(case_tac[!] inst)
  apply(auto simp add: brute_force ll_read_ptr_failure_NE_def ll_read_ptr_any_NN_def)
  done

definition check_ptr_NN_success ::
  "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "check_ptr_NN_success \<equiv>
  (\<lambda>(sv, sp). {((), sv, sp')|sp'.
        ptr_matches_hash sp Secondary
        \<and> card_mode sv \<noteq> CDegradedH
        \<and> ptr1_state sp = IOK
        \<and> sp' = sp\<lparr> ptr1 := ptr2 sp, hash :=
            HTable2 (pub_hashes2 (hash sp)) ((ptr_hashes2 (hash sp))(Primary := ptr2 sp)) \<rparr>
  })
  [+] (\<lambda>(sv, sp). {((), sv, sp')|sp'.
       ptr_matches_hash sp Primary
        \<and> card_mode sv \<noteq> CDegradedH
        \<and> ptr2_state sp = IOK
        \<and> sp' = sp\<lparr> ptr2 := ptr1 sp, hash := 
            HTable2 (pub_hashes2 (hash sp)) ((ptr_hashes2 (hash sp))(Secondary := ptr1 sp)) \<rparr>
  })
  [+] (\<lambda>(sv, sp). {((), sv, sp')|sp'.
    ptr_matches_hash sp Primary
    \<and> ptr_matches_hash sp Secondary 
    \<and> ptr1 sp = ptr2 sp
    \<and> sp' = sp
  })
  "

definition check_ptr_NN_degrade ::
  "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "check_ptr_NN_degrade \<equiv>  (\<lambda>(sv, sp). {((), sv', sp')|cnt sv' sp'.
    ((ptr_matches_hash sp Secondary \<and> cnt = ptr2 sp) \<or>
     (ptr_matches_hash sp Primary \<and> cnt = ptr1 sp))
    \<and> card_mode sv \<noteq> CDegradedH
    \<and> sv' = sv\<lparr> card_mode := min CDegradedP (card_mode sv) \<rparr>
    \<and> (ptr_matches_hash sp' Secondary \<longrightarrow> ptr2 sp' = cnt)
    \<and> (ptr_matches_hash sp' Primary \<longrightarrow> ptr1 sp' = cnt)
    \<and> (ptr_matches_hash sp' Primary \<or> ptr_matches_hash sp' Secondary)
    \<and> (\<not> ptr_matches_hash sp' Secondary \<or> \<not> ptr_matches_hash sp' Primary)
    \<and> (\<not> (ptr_matches_hash sp' Secondary) \<longrightarrow> ptr2_state sp' < IOK)
    \<and> (\<not> (ptr_matches_hash sp' Primary) \<longrightarrow> ptr1_state sp' < IOK)
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> pub_hashes2 (hash sp') = pub_hashes2 (hash sp)
    \<and> hstate sp' = hstate sp
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })"

definition check_ptr_NN_deg_h ::
  "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "check_ptr_NN_deg_h \<equiv>
  (\<lambda>(sv, sp). {((), sv', sp')|sv' sp' cnt.
    sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>

    \<and> ((ptr_matches_hash sp Secondary \<and> cnt = ptr2 sp) \<or>
     (ptr_matches_hash sp Primary \<and> cnt = ptr1 sp))

    \<and> (ptr_matches_hash sp' Secondary \<longrightarrow> ptr2 sp' = cnt)
    \<and> (ptr_matches_hash sp' Primary \<longrightarrow> ptr1 sp' = cnt)
    \<and> (ptr_matches_hash sp' Secondary \<or> ptr_matches_hash sp' Primary)

    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp

    \<and> pub_hashes2 (hash sp') = pub_hashes2 (hash sp)
    \<and> hstate sp' = LDegraded
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })
  "

definition check_ptr_NN :: "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "check_ptr_NN \<equiv>
  check_ptr_NN_success
  [+] check_ptr_NN_degrade
  [+] check_ptr_NN_deg_h
  "

lemma exc_exhaust:
  "\<lbrakk> e \<noteq> FailWrite; e \<noteq> FailLLRead; e \<noteq> FailRead; e \<noteq> FailMode;
         \<forall>exc_val. e \<noteq> Degraded exc_val; e \<noteq> InvalidInput;
         e \<noteq> StartedExc \<rbrakk> \<Longrightarrow> False"
  by(case_tac e, auto)

lemma Degraded_exhaust:
  "\<lbrakk> e \<noteq> FailWrite; e \<noteq> FailLLRead; e \<noteq> FailRead; e \<noteq> FailMode; e \<noteq> InvalidInput;
    e \<noteq> StartedExc \<rbrakk> \<Longrightarrow> \<exists>v. e = Degraded v"
  by(case_tac e, auto)

declare doesnt_throw_exceptions_case[no_throw del]
declare doesnt_throw_raw_bindN'[no_throw]
declare [[simproc add: throw_bind_simproc']]

lemma aux:
  "(\<forall>exc_val. exc_val = EB True \<or> v \<noteq> exc_val) = (v = EB True)"
  apply(auto)
  apply(metis)
  done

definition one_gone where
  "one_gone sp \<equiv> hstate sp = LDegraded"

definition mode_inv where
  "mode_inv \<equiv> {(sv, sp). card_mode sv = CDegradedH \<longleftrightarrow> one_gone sp }"

lemma ge_DegP_imp_ge_DegH:
  "CDegradedP \<le> x \<Longrightarrow> (CDegradedH \<le> x) = True"
  by(case_tac x, simp_all)

lemma min2:
  "(x :: 'a :: linorder) \<le> y \<Longrightarrow> min x (min y z) = min x z"
  "x \<le> y \<Longrightarrow> min y (min x z) = min x z"
  "x \<le> y \<Longrightarrow> min y (min z x) = min x z"
  "x \<le> y \<Longrightarrow> min x (min z y) = min x z"
  by(auto simp add: min_def)

lemma update_htable2_alt:
  "pub_update_htable2 ht pind inst cnt = 
        HTable2 ((pub_hashes2 ht)((pind, inst) := Hash cnt)) (ptr_hashes2 ht)"
  "ptr_update_htable2 ht inst ptr_set = 
        HTable2 (pub_hashes2 ht) ((ptr_hashes2 ht)(inst := ptr_set))"
  by(case_tac[!] ht, auto)

lemma check_ptr_NN_right:
  "\<lbrakk> (r, sv', sp') \<in> (check_ptr\<Down>NN) (sv, sp); (sv, sp) \<in> mode_inv \<rbrakk> \<Longrightarrow> 
        (r, sv', sp') \<in> check_ptr_NN (sv, sp)"
  apply(simp only: check_ptr_def check_ptr_NN_def)
  apply(simp)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: just_sync_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: sync_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_or_fail_silence_deg_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_or_fail_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_NN write_ptr_instance_NN)
  apply(simp_all only: read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: write_ptr_instance_def update_ptr_instance_hash_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_DegradedH_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(simp_all add: get_hashes_all_defs write_hashes_defs brute_force
    read_ptr_NN_def write_ptr_instance_NN_def mode_inv_def one_gone_def
    update_htable2_alt (* serialize_parse_ptr *)
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(case_tac[!] sv)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms Degraded_exhaust} *},
    clarsimp_all simp add: all_conj aux)
  apply(iterative_simp)+
  apply(simp_all (no_asm_use) add: check_ptr_NN_success_def check_ptr_NN_degrade_def
    check_ptr_NN_deg_h_def choice_def update_htable2_alt)
  apply(simp_all (no_asm_use) add:
    HOL.conj_disj_distribR ex_disj_distrib conj_assoc abs2_ptr_same_rest_def)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: min2 ptr_matches_htable2_def parse_serialize_ptr)
  apply(clarsimp_all simp add: min_def split: split_if_asm)
  using [[simp_depth_limit=2]]
  apply(case_tac[!] "ptr2_state sp", clarsimp_all simp add: ge_DegP_imp_ge_DegH
    split: split_if_asm)
  apply(case_tac[!] "ptr1_state sp", clarsimp_all simp add: ge_DegP_imp_ge_DegH
    split: split_if_asm)
  apply(case_tac[!] "card_mode")
  apply(simp_all)
  apply(case_tac[!] sp)
  apply(clarsimp_all)
  done

definition check_ptr_fail_NE where
  "check_ptr_fail_NE \<equiv>
  (\<lambda>(sv, sp). {(FailMode, sv', sp')|sv' sp'.
    (hstate sp' = LDead
        \<or> (\<not> ptr_matches_hash sp' Primary \<and> \<not> ptr_matches_hash sp' Secondary))
    \<and> sv' = sv\<lparr> card_mode := CFail \<rparr>
    \<and> hstate sp' \<le> hstate sp
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> (hstate sp' \<noteq> LDead \<longrightarrow> hash sp' = hash sp)
    \<and> abs2_ptr_same_rest sp sp'
  })
  "

lemma get_hashes_DegradedH_NE_no_throw[no_throw]:
  "doesnt_throw_raw (get_hashes_DegradedH_NE) FailRead"
  "doesnt_throw_raw (get_hashes_DegradedH_NE) FailLLRead"
  "doesnt_throw_raw (get_hashes_DegradedH_NE) FailWrite"
  "doesnt_throw_raw (get_hashes_DegradedH_NE) StartedExc"
  "\<forall>v. doesnt_throw_raw (get_hashes_DegradedH_NE) (Degraded v)"
  by(auto simp add: doesnt_throw_raw_def get_hashes_DegradedH_NE_def
    get_hashes_DegradedH_defs project_defs)

lemma[no_throw]:
  "doesnt_throw_raw (get_hashes_DegradedH\<Down>NE) FailRead"
  "doesnt_throw_raw (get_hashes_DegradedH\<Down>NE) FailLLRead"
  "doesnt_throw_raw (get_hashes_DegradedH\<Down>NE) FailWrite"
  "doesnt_throw_raw (get_hashes_DegradedH\<Down>NE) StartedExc"
  "\<forall>v. doesnt_throw_raw (get_hashes_DegradedH\<Down>NE) (Degraded v)"
  by(auto simp add: get_hashes_proj_simps intro!: get_hashes_DegradedH_NE_no_throw)

declare sefN_bind[simp del]

lemma exc_aux2:
  "(\<forall>exc_val. exc_val = EB True \<or> e \<noteq> Degraded exc_val) =
  (e = Degraded (EB True) \<or> (\<forall>v. e \<noteq> Degraded v))"
  "(\<forall>exc_val. e = Degraded exc_val \<longrightarrow> exc_val = EB True) =
  (e = Degraded (EB True) \<or> (\<forall>v. e \<noteq> Degraded v))"
  "(\<forall>exc_val. e = Degraded exc_val \<longrightarrow> exc_val = EB True \<or> P exc_val)
  = ((e = Degraded (EB True)) \<or> (\<forall>exc_val. e \<noteq> Degraded exc_val) \<or>
  (\<exists>exc_val. e = Degraded exc_val \<and> P exc_val))
  "
  apply(case_tac[!] e, simp_all)
  apply(metis)
  done

lemma check_ptr_NE_right:
  "\<And>sv sp. (check_ptr\<Down>NE) (sv, sp) \<subseteq> (check_ptr_fail_NE) (sv, sp)"
  apply(simp_all only: check_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: just_sync_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: sync_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only:  read_ptr_or_fail_silence_deg_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_or_fail_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_NN write_ptr_instance_NN)
  apply(simp_all only: read_ptr_def Let_def)
  (* TODO: choking hazard, small parts. Immediately trying the usuall simp_all add: ... seems to
    cause the simplifier to loop *)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: write_ptr_instance_def update_ptr_instance_hash_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_DegradedH_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(simp_all add: get_hashes_all_defs write_hashes_defs brute_force
    read_ptr_NN_def write_ptr_instance_NN_def
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all)
  apply(clarsimp_all simp add: min_def split: split_if_asm)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *}, clarsimp_all)+
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: FailMode_exhaust all_conj aux exc_aux2)
  apply(clarsimp_all simp add: check_ptr_fail_NE_def)
  apply(clarsimp_all simp add: abs2_ptr_same_rest_def not_BothGone_le_OneGone ptr_matches_htable2_def)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *},
    clarsimp_all simp add: FailMode_exhaust)+
  done

use "trans_simproc.ML"

simproc_setup trans_simproc 
  ("r \<le> s")
  = {* K Trans_Simproc.trans_simproc *}

definition startup1' where
  "startup1' \<equiv> do {
    startup1;
    check_ptr
  }"

definition startup1'_NN  ::
  "(unit, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "startup1'_NN \<equiv>
  (\<lambda>(sv, sp). {((), sv', sp')|sv' sp' cnt.
    hstate sp = LOK

    \<and> ((ptr_matches_hash sp Secondary \<and> cnt = ptr2 sp) \<or>
     (ptr_matches_hash sp Primary \<and> cnt = ptr1 sp))
    \<and> ptr_matches_hash sp' Secondary
    \<and> ptr_matches_hash sp' Primary
    \<and> ptr1 sp' = cnt
    \<and> ptr2 sp' = cnt
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp

    \<and> sv' = \<lparr> card_mode = CNormal, updated = {} \<rparr>
    \<and> hstate sp' = hstate sp
    \<and> pub_hashes2 (hash sp') = pub_hashes2 (hash sp)
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })
  [+] (\<lambda>(sv, sp). {((), sv', sp')|sv' sp' cnt.
    hstate sp = LOK

    \<and> ((ptr_matches_hash sp Secondary \<and> cnt = ptr2 sp) \<or>
     (ptr_matches_hash sp Primary \<and> cnt = ptr1 sp))
    \<and> (ptr_matches_hash sp' Secondary \<longrightarrow> ptr2 sp' = cnt)
    \<and> (ptr_matches_hash sp' Primary \<longrightarrow> ptr1 sp' = cnt)
    \<and> (ptr_matches_hash sp' Secondary \<or> ptr_matches_hash sp' Primary)
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> (ptr1_state sp' \<le> IUnwr \<or> ptr2_state sp' \<le> IUnwr)

    \<and> sv' = \<lparr> card_mode = CDegradedP, updated = {} \<rparr>

    \<and> hstate sp' = hstate sp
    \<and> pub_hashes2 (hash sp') = pub_hashes2 (hash sp)
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })
  [+] (\<lambda>(sv, sp). {((), sv', sp')|sv' sp' cnt.
    hstate sp \<noteq> LDead

    \<and> ((ptr_matches_hash sp Secondary \<and> cnt = ptr2 sp) \<or>
     (ptr_matches_hash sp Primary \<and> cnt = ptr1 sp))
    \<and> (ptr_matches_hash sp' Secondary \<longrightarrow> ptr2 sp' = cnt)
    \<and> (ptr_matches_hash sp' Primary \<longrightarrow> ptr1 sp' = cnt)
    \<and> (ptr_matches_hash sp' Secondary \<or> ptr_matches_hash sp' Primary)
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
 
    \<and> hstate sp' = LDegraded
    \<and> sv' = \<lparr> card_mode = CDegradedH, updated = {} \<rparr>
    \<and> pub_hashes2 (hash sp') = pub_hashes2 (hash sp)
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })
  "

lemma check_ptr_NN_right':
  "(sv, sp) \<in> mode_inv \<Longrightarrow> (check_ptr\<Down>NN) (sv, sp) \<subseteq> check_ptr_NN (sv, sp)"
  by(auto dest: check_ptr_NN_right[rotated 1])

lemma UNIV_le:
  "(UNIV \<subseteq> A) = (A = UNIV)"
  by(blast)

lemma Collect_eq_UNIV:
  "({x. P x} = UNIV) = (\<forall>x. P x)"
  by(blast)

lemma check_ptr_bindN:
  "\<lbrace>true\<rbrace> m \<lbrace>> K mode_inv\<rbrace> \<Longrightarrow>
    (m \<guillemotright>N (check_ptr\<Down>NN)) (sv, sp) \<subseteq> (m \<guillemotright>N check_ptr_NN) (sv, sp)"
  by(auto simp add: Hoare_def Box_def unfold_monad UNIV_le Collect_eq_UNIV
    dest: check_ptr_NN_right'[THEN subsetD, rotated 1]
  )

lemma startup_mode_inv: "\<lbrace>true\<rbrace> (startup1\<Down>NN) \<lbrace>> K mode_inv \<rbrace>"
  apply(auto simp add: K_def mode_inv_def one_gone_def Hoare_def Box_def)
  apply(auto simp add: startup1_def check_hash_region_defs brute_force min_def)
  done

lemma startup1'_NN_right:
  "(startup1'\<Down>NN) (sv, sp) \<subseteq> ((startup1\<Down>NN) \<guillemotright>N check_ptr_NN) (sv, sp)"
  by(auto simp only: startup1'_def project_simps intro!: check_ptr_bindN startup_mode_inv)

lemma startup1'_NE_right:
  "(startup1'\<Down>NE) (sv, sp) \<subseteq> ((startup1\<Down>NE) [+] (startup1\<Down>NN \<guillemotright>N check_ptr_fail_NE)) (sv, sp)"
  by(auto simp only: startup1'_def first_stage
    intro!: choice_mono2 check_ptr_NE_right intro: bindN_mono2 del: subsetI)

lemma st_aux:
  "(sp\<lparr> updated := v1, card_mode := v2\<rparr> = \<lparr> card_mode = v2', updated = v1' \<rparr>) = (v1 = v1' \<and> v2 = v2')"
  by (smt pmm_vol_mem.iffs pmm_vol_mem.simps(4) pmm_vol_mem.simps(5) pmm_vol_mem.surjective unit.exhaust)
  
lemma startup1'_NN_right':
  "(startup1'\<Down>NN) (sv, sp) \<subseteq> startup1'_NN (sv, sp)"
  apply(rule subset_trans[OF startup1'_NN_right])
  apply(simp add: startup1_def check_hash_region_def check_ptr_NN_def first_stage)
  apply(simp add: choice_def, safe)
  apply(clarsimp_all simp add: vol_ops_defs brute_force check_hash_region_defs
    check_ptr_NN_success_def check_ptr_NN_degrade_def check_ptr_NN_deg_h_def
    abs2_ptr_same_rest_def)
  apply(case_tac[!] sp)
  apply((iterative_simp)+)
  apply(simp_all (no_asm_use) add: startup1'_NN_def choice_def)
  apply(simp_all (no_asm_use) add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    abs2_ptr_same_rest_def)
  apply(case_tac[!] sv)
  apply(simp_all add: min_def)
  apply (metis less_eq_IUnwr state_ineq_simps(3))+
  done

definition startup1'_NE ::
  "('c exceptions, pmm_vol_mem \<times> ('c::contents_cl)abs2_state)nds_monad" where
  "startup1'_NE \<equiv>
  (\<lambda>(sv, sp). {(FailMode, sv', sp')|sv' sp'.
    sv' = \<lparr> card_mode = CFail, updated = {} \<rparr>
    \<and> (hstate sp' = LDead
        \<or>  (\<not>ptr_matches_hash sp' Primary \<and> \<not>ptr_matches_hash sp' Secondary))
    \<and> hstate sp' \<le> hstate sp
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> (hstate sp' \<noteq> LDead \<longrightarrow> pub_hashes2 (hash sp') = pub_hashes2 (hash sp))
    \<and> abs2_ptr_same_rest sp sp'
  })"

lemma startup1'_NE_right':
  "(startup1'\<Down>NE) (sv, sp) \<subseteq> startup1'_NE (sv, sp)"
  apply(rule subset_trans[OF startup1'_NE_right])
  apply(simp only: startup1_def check_hash_region_def first_stage vol_ops_project_simps)
  apply(simp only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: vol_ops_defs brute_force check_hash_region_defs
    check_ptr_fail_NE_def split: split_if_asm)
  apply(iterative_simp)+
  apply(simp_all (no_asm_use) add: startup1'_NE_def)
  apply(simp_all (no_asm_use) add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc
    abs2_ptr_same_rest_def)
  apply(clarsimp_all)
  apply (metis less_eq_log_state.simps(1) not_BothGone_le_OneGone order_antisym)+
  done

lemma check_hash_region_NN:
  "(check_hash_region\<Down>NN) = (check_hash_region_success\<Down>NN [+] degrade_hash_region\<Down>NN)"
  by(auto simp add: check_hash_region_def first_stage check_hash_region_project_simps)

lemma aux1:
  "doN {
    init_vars\<Down>NN;
    check_hash_region_success\<Down>NN;
    mode \<leftarrow> get_mode\<Down>NN;
    if mode = CDegradedH then m1 else m2
  } = doN { init_vars\<Down>NN; check_hash_region_success\<Down>NN; m2 }"
  apply(rule ext)
  apply(auto simp add: check_hash_region_success_def vol_ops_defs brute_force split: split_if_asm)
  done

lemma aux2:
  "doN {
    init_vars\<Down>NN;
    degrade_hash_region\<Down>NN;
    mode \<leftarrow> get_mode\<Down>NN;
    if mode = CDegradedH then m1 else m2
  } = doN { init_vars\<Down>NN; degrade_hash_region\<Down>NN; m1 }"
  apply(rule ext)
  apply(auto simp add: degrade_hash_region_def vol_ops_defs brute_force min_def split: split_if_asm)
  done

definition restart_mangle_sv_R where
  "restart_mangle_sv_R \<equiv> (restart_mangle_sv :: ('c htable, 'sv, 'c :: contents_cl)abs2_monad)\<Down>R"

lemma aux3:
  "get_hashes\<Down>R = restart_mangle_sv_R"
  "get_hashes_DegradedH\<Down>R = restart_mangle_sv_R"
  apply(rule_tac[!] ext)
  apply(clarsimp_all simp add: get_hashes_defs get_hashes_DegradedH_def
    get_hashes_degrade'_def
    restart_mangle_sv_R_def restart_mangle_sv_def
    modify_def put_def brute_force
    choose_def brute_force)
  done

lemma choice_absorb_left: "(m1 [+] m2 = m2) = (\<forall>s. m1 s \<subseteq> m2 s)"
  apply(auto simp add: choice_def intro!: ext)
  apply(metis UnI1)
  done

lemma choice_absorb_right: "(m1 [+] m2 = m1) = (\<forall>s. m2 s \<subseteq> m1 s )"
  apply(auto simp add: choice_def intro!: ext)
  apply(metis UnI2)
  done

lemma aux5:
  "(read_ptr_DegradedH inst)\<Down>R = (ll_read_ptr inst)\<Down>R"
  apply(simp add: read_ptr_DegradedH_def first_stage choice_absorb_right)
  apply(simp only: ll_read_ptr_def first_stage aux3 restart_mangle_sv_R_def)
  apply(simp only: choice_def)
  apply(safe)
  apply(clarsimp_all simp add: brute_force split: split_if_asm)
  done

lemma choice_absorb_right':
  "m2 = m1 [+] m2' \<Longrightarrow> m1 [+] m2 = m1 [+] m2'"
  by(auto simp add: choice_def intro!: ext)

lemma bindN_mono1_all:
  "(\<And>s. ml' s \<subseteq> ml s) \<Longrightarrow> (ml' \<guillemotright>=N m) s \<subseteq> (ml \<guillemotright>=N m) s"
  by(rule RMonad_Lemmas.bindN_mono1, simp)

thm bindN'_mono2_all

thm check_ptr_NE_right[THEN bindN_mono1_all, THEN bindN'_mono2_all]

lemma bindS_decompose: "(r \<in> (m s \<guillemotright>=S hr)) = (\<exists>s'. s' \<in> m s \<and> r \<in> hr s')"
  by(auto simp add: bindS_def)

definition startup1'_NN' where
  "startup1'_NN' \<equiv> startup1'_NN"

definition check_ptr_R where
  "check_ptr_R \<equiv>  (\<lambda>(sv, sp). {(sv', sp')|sv' sp'.
    (\<not>ptr_matches_hash sp' Secondary
        \<or> (ptr_matches_hash sp Secondary \<and> ptr2 sp' = ptr2 sp)
        \<or> (ptr_matches_hash sp Primary \<and> ptr2 sp' = ptr1 sp)
    )
    \<and> (\<not>ptr_matches_hash sp' Primary
        \<or> (ptr_matches_hash sp Secondary \<and> ptr1 sp' = ptr2 sp)
        \<or> (ptr_matches_hash sp Primary \<and> ptr1 sp' = ptr1 sp)
    )
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> (hstate sp' \<noteq> LDead \<longrightarrow> pub_hashes2 (hash sp') = pub_hashes2 (hash sp))
    \<and> hstate sp' \<le> hstate sp
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  }) [+]
  (\<lambda>(sv, sp). {(sv', sp')|sv' sp'.
    \<not>ptr_matches_hash sp' Secondary
    \<and> \<not>ptr_matches_hash sp' Primary
    \<and> ptr1_state sp' \<le> ptr1_state sp
    \<and> ptr2_state sp' \<le> ptr2_state sp
    \<and> hstate sp' \<le> hstate sp
    \<and> (hstate sp' \<noteq> LDead \<longrightarrow> pub_hashes2 (hash sp') = pub_hashes2 (hash sp))
    \<and> abs2_unwr_bad_hash sp sp'
    \<and> abs2_ptr_same_rest sp sp'
  })
  "

lemma check_ptr_R_right:
  "(check_ptr\<Down>R) (sv, sp) \<subseteq> check_ptr_R (sv, sp)"
  apply(clarsimp_all simp only: check_ptr_def check_ptr_def first_stage
    vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_def write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: just_sync_ptr_def sync_ptr_def Let_def first_stage if_choice
    vol_ops_project_simps aux3)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_or_fail_silence_deg_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: update_ptr_instance_hash_def read_ptr_or_fail_def read_ptr_def 
    read_ptr_DegradedH_def)
  apply(simp_all only: first_stage Let_def vol_ops_project_simps if_choice aux3)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: update_ptr_instance_hash_def write_ptr_instance_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_DegradedH_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_proj_simps)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(clarsimp_all simp add: get_hashes_DegradedH_defs
    get_hashes_defs restart_mangle_sv_R_def write_hashes_defs
    brute_force get_hashes_DegradedH_NN_def get_hashes_NN_def
    vol_ops_defs split: split_if_asm exceptions.split_asm)
  apply(simp_all add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc abs2_ptr_same_rest_def
    disj_commute disj_left_commute disj_absorb disj_left_absorb all_conj
  )
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *}, clarsimp_all)+
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: FailMode_exhaust all_conj aux exc_aux2 min_def
    ptr_matches_htable2_def
    split: split_if_asm del: disjCI)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *}, clarsimp_all)+
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  apply(simp_all only: check_ptr_R_def choice_def update_htable2_alt)
  apply(simp_all add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc abs2_ptr_same_rest_def
    disj_commute disj_left_commute disj_absorb disj_left_absorb
  )
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: FailMode_exhaust all_conj aux exc_aux2 min_def split: split_if_asm)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms Degraded_exhaust} *},
    clarsimp_all simp add: all_conj aux parse_serialize_ptr)
  apply (smt linorder_linear not_BothGone_le_OneGone state_ineq_simps(3))+
  done

lemma LOK_less_eq:
  "LOK \<le> x = (x = LOK)"
  by(case_tac x, simp_all)

lemma less_eq_LDead:
  "x \<le> LDead = (x = LDead)"
  by(case_tac x, simp_all)

lemma check_ptr_R_NN_reduction:
  "(v, sv', sp') \<in> (check_ptr_R) (sv, sp) \<guillemotright>=S (startup1'_NN)
    \<Longrightarrow> (v, sv', sp') \<in> (startup1'_NN) (sv, sp)"
  apply(clarsimp simp only: bindS_decompose)
  apply(subst startup1'_NN'_def[THEN symmetric])
  apply(clarsimp_all simp add: startup1'_NN_def init_vars_def check_hash_region_defs brute_force
    abs2_ptr_same_rest_def
    check_ptr_R_def split: split_if_asm
  )
  apply(iterative_simp simp add: LOK_less_eq IOK_less_eq)+
  apply(simp_all (no_asm) only: startup1'_NN'_def startup1'_NN_def choice_def)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc abs2_ptr_same_rest_def
    disj_commute disj_left_commute disj_absorb disj_left_absorb
  )
  apply(clarsimp_all)
  apply(simp_all add: less_eq_IUnwr)
  apply(iterative_simp)
  done

definition startup1'_NE' where
  "startup1'_NE' \<equiv> startup1'_NE"

lemma check_ptr_R_NE_reduction:
  "(e, sv', sp') \<in> (check_ptr_R) (sv, sp) \<guillemotright>=S (startup1'_NE)
    \<Longrightarrow> (e, sv', sp') \<in> (startup1'_NE) (sv, sp)"
  apply(clarsimp simp only: bindS_decompose)
  apply(rename_tac sv'' sp'')
  apply(subst startup1'_NE'_def[THEN symmetric])
  apply(clarsimp_all simp add: startup1'_NE_def init_vars_def check_hash_region_defs brute_force
    abs2_ptr_same_rest_def
    check_ptr_R_def split: split_if_asm
  )
  apply(iterative_simp simp add: LOK_less_eq IOK_less_eq)+
  apply(simp_all (no_asm) add: startup1'_NE'_def startup1'_NE_def choice_def)
  apply(simp_all add:
    HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc abs2_ptr_same_rest_def
    disj_commute disj_left_commute disj_absorb disj_left_absorb
  )
  apply(clarsimp_all simp add: less_eq_IDead less_eq_LDead)
  done

lemma bindN_bindS_decomp:
  "x \<in> (m \<guillemotright>=N f) s \<guillemotright>=S h \<Longrightarrow> \<exists>v s'. (v, s') \<in> m s \<and> x \<in> f v s' \<guillemotright>=S h"
  by(auto simp add: unfold_monad)

lemma startup1'_NN_reduction:
  "(e, sv', sp') \<in> (startup1'\<Down>R) (sv, sp) \<guillemotright>=S (startup1'_NN)
    \<Longrightarrow> (e, sv', sp') \<in> (startup1'_NN) (sv, sp)"
  apply(subst startup1'_NN'_def[THEN symmetric])
  apply(simp only: startup1'_def startup1_def first_stage
    vol_ops_project_simps check_hash_region_project_simps check_hash_region_NN)
  apply(auto dest!: bindN_bindS_decomp check_ptr_R_right[THEN bindS_mono1, THEN subsetD]
    check_ptr_R_NN_reduction
  )
  apply(clarsimp_all simp add: check_hash_region_defs init_vars_def brute_force)
  apply(clarsimp_all simp add: startup1'_NN_def init_vars_def check_hash_region_defs brute_force
    abs2_ptr_same_rest_def
    split: split_if_asm
  )
  apply(iterative_simp simp add: IOK_less_eq EcOK_less_eq)+
  apply(simp_all add: startup1'_NN'_def startup1'_NN_def choice_def abs2_ptr_same_rest_def)
  done

lemma less_eq_LOK:
  "(x \<le> LOK) = True"
  by(case_tac x, simp_all)

lemma startup1'_NE_reduction:
  "(e, sv', sp') \<in> (startup1'\<Down>R) (sv, sp) \<guillemotright>=S (startup1'_NE)
    \<Longrightarrow> (e, sv', sp') \<in> (startup1'_NE) (sv, sp)"
  apply(subst startup1'_NE'_def[THEN symmetric])
  apply(simp only: startup1'_def startup1_def first_stage
    vol_ops_project_simps check_hash_region_project_simps check_hash_region_NN)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto dest!: bindN_bindS_decomp check_ptr_R_right[THEN bindS_mono1, THEN subsetD]
    check_ptr_R_NE_reduction)
  apply(clarsimp_all simp add: check_hash_region_defs init_vars_def brute_force)
  apply(clarsimp_all simp add: startup1'_NE_def init_vars_def check_hash_region_defs brute_force
    abs2_ptr_same_rest_def
    split: split_if_asm
  )
  apply(iterative_simp simp add: less_eq_LOK IOK_less_eq EcOK_less_eq)+
  apply(simp_all add: startup1'_NE'_def startup1'_NE_def choice_def abs2_ptr_same_rest_def)
  apply(clarsimp_all dest!: not_BothGone_le_OneGone simp add: less_eq_LOK)
  done

declare [[simproc del: trans_simproc]]

definition abs2_startup1' where
  "abs2_startup1' \<equiv> compose
    startup1'_NN
    startup1'_NE
    (startup1'\<Down>R)
  "

lemma abs2_startup1'_fin:
  "sim_norm Id abs2_startup1' (try_catchR startup1' abs2_startup1')"
  by(auto simp add: sim_norm_decompose sim_normN_def sim_normN_ext_def
    sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def project_simps choice_def
    abs2_startup1'_def compose_project
    intro: startup1'_NN_right'[THEN subsetD] startup1'_NE_right'[THEN subsetD]
    startup1'_NN_reduction startup1'_NE_reduction

  )
  
end
