theory Abs3_CommonDefs
  imports Abs3_LLInterface
begin

definition get_hashes_success  :: "('c pub_hashes, 'c :: size)abs3_monad" where
  "get_hashes_success \<equiv> \<lambda>(sv, sp).
    {(Some (Inr h), sv, sp)|h.
        (hstate sp = LOK \<or> hstate sp = LDegraded)
        \<and> h = pub_hashes sp
  }"

definition get_hashes_fail :: "('a, 'c :: size)abs3_monad" where
  "get_hashes_fail \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' h'.
        sv' = sv\<lparr> card_mode := CFail \<rparr>
        \<and> sp' = sp\<lparr> pub_hashes := h', hstate := LDead \<rparr>
  }"

definition get_hashes_degrade :: "('a, 'c :: size)abs3_monad" where
  "get_hashes_degrade \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp'.
        hstate sp \<noteq> LDead
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded \<rparr>
  }"

definition get_hashes :: "('c pub_hashes, 'c :: size)abs3_monad" where
  "get_hashes \<equiv>
    get_hashes_success
    [+] get_hashes_fail
    [+] get_hashes_degrade
    [+] restart_mangle_sv"

lemmas get_hashes_defs =
  get_hashes_def
  get_hashes_success_def
  get_hashes_fail_def
  get_hashes_degrade_def

definition get_hashes_degrade' :: "('c pub_hashes, 'c :: size)abs3_monad" where
  "get_hashes_degrade' \<equiv> \<lambda>(sv, sp).
    {(Some (Inr h), sv, sp')|h sp'.
        hstate sp = LOK
        \<and> h = pub_hashes sp
        \<and> sp' = sp\<lparr> hstate := LDegraded \<rparr>
  }"

definition get_hashes_DegradedH :: "('c pub_hashes, 'c :: size)abs3_monad" where
  "get_hashes_DegradedH \<equiv>
    get_hashes_success
    [+] get_hashes_degrade'
    [+] get_hashes_fail
    [+] restart_mangle_sv"

lemmas get_hashes_DegradedH_defs =
  get_hashes_DegradedH_def
  get_hashes_def
  get_hashes_success_def
  get_hashes_fail_def
  get_hashes_degrade'_def

lemmas get_hashes_all_defs =
  get_hashes_def
  get_hashes_degrade_def
  get_hashes_DegradedH_defs

definition write_hashes_success :: 
  "'c pub_hashes \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_hashes_success ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        hstate sp = LOK
        \<and> sp' = sp\<lparr> pub_hashes := ht \<rparr>
  }"

definition write_hashes_degrade :: 
  "'c pub_hashes \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_hashes_degrade ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB False))), sv', sp')|sv' sp'.
        sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := min (hstate sp) LDegraded \<rparr>
  }"

definition write_hashes_degrade_write :: 
  "'c pub_hashes \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_hashes_degrade_write ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB True))), sv', sp')|sv' sp'.
        hstate sp \<noteq> LDead
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> pub_hashes := ht, hstate := LDegraded \<rparr>
  }"

definition write_hashes_fail ::
  "'c pub_hashes \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_hashes_fail ht \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp'.
        sv' = sv\<lparr> card_mode := CFail \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDead \<rparr>
  }"

definition write_hashes_partial_restart ::
  "'c pub_hashes \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_hashes_partial_restart ht \<equiv> \<lambda>(sv, sp).
    {(None, sv, sp')|sp'.
        hstate sp \<ge> LDegraded
        \<and> sp' = sp\<lparr> pub_hashes := ht \<rparr>
  }"

definition write_hashes :: 
  "'c pub_hashes \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_hashes ht \<equiv>
  write_hashes_success ht
  [+] write_hashes_degrade ht
  [+] write_hashes_degrade_write ht
  [+] write_hashes_fail ht
  [+] write_hashes_partial_restart ht
  [+] restart_mangle_sv"

lemmas write_hashes_defs =
  write_hashes_def
  write_hashes_success_def
  write_hashes_degrade_def
  write_hashes_degrade_write_def
  write_hashes_fail_def
  write_hashes_partial_restart_def

definition get_ptr_success ::
  "(pub_indx set, 'c) abs3_monad" where
  "get_ptr_success \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ptr_set), sv, sp)|ptr_set.
        ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> ptr_set = ptr sp
  }"

definition get_ptr_fail ::
  "(pub_indx set, 'c) abs3_monad" where
  "get_ptr_fail \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' hst pst.
        (hst = LDead \<or> pst = LDead)
        \<and> hst \<le> hstate sp
        \<and> pst \<le> ptr_state sp
        \<and> sp' = sp\<lparr> hstate := hst, ptr_state := pst \<rparr>
        \<and> sv' = sv\<lparr> card_mode := CFail \<rparr>
  }"

definition get_ptr_deg ::
  "(pub_indx set, 'c ) abs3_monad" where
  "get_ptr_deg \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp'.
        ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> sv' = sv\<lparr> card_mode := min CDegradedP (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> ptr_state := LDegraded \<rparr>
  }"
        
definition get_ptr_deg_h ::
  "(pub_indx set, 'c ) abs3_monad" where
  "get_ptr_deg_h \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp' pst.
        pst \<le> ptr_state sp
        \<and> LDegraded \<le> pst
        \<and> hstate sp \<ge> LDegraded
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded, ptr_state := pst \<rparr>
  }"

definition get_ptr_silence_deg ::
  "(pub_indx set, 'c ) abs3_monad" where
  "get_ptr_silence_deg \<equiv> \<lambda>(sv, sp).
    {(Some (Inr (ptr sp)), sv', sp')|sv' sp'.
        ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> sv' = sv\<lparr> card_mode := min CDegradedP (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> ptr_state := LDegraded \<rparr>
  }"

definition get_ptr_sv_deg_h ::
  "(pub_indx set, 'c ) abs3_monad" where
  "get_ptr_sv_deg_h \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv, sp')|sp'.
        card_mode sv = CDegradedH
        \<and> ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> sp' = sp
  }"
        
definition get_ptr_silence_deg_h ::
  "(pub_indx set, 'c ) abs3_monad" where
  "get_ptr_silence_deg_h \<equiv> \<lambda>(sv, sp).
    {(Some (Inr (ptr sp)), sv', sp')|sv' sp' pst.
        pst \<le> ptr_state sp
        \<and> LDegraded \<le> pst
        \<and> hstate sp \<ge> LDegraded
        \<and> sv' = sv\<lparr> card_mode := min CDegradedH (card_mode sv) \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded, ptr_state := pst \<rparr>
  }"

definition get_ptr ::
  "bool \<Rightarrow> (pub_indx set, 'c :: size) abs3_monad" where
  "get_ptr silence_deg \<equiv>
    get_ptr_success
    [+] get_ptr_fail
    [+] restart_mangle_sv
    [+] (if silence_deg
        then get_ptr_silence_deg [+] get_ptr_silence_deg_h
        else get_ptr_deg [+] get_ptr_deg_h [+] get_ptr_sv_deg_h
    )
  "

thm get_ptr_def

lemmas get_ptr_defs =
  get_ptr_def
  get_ptr_success_def
  get_ptr_fail_def
  get_ptr_deg_def
  get_ptr_deg_h_def
  get_ptr_sv_deg_h_def
  get_ptr_silence_deg_def
  get_ptr_silence_deg_h_def

definition write_ptr_success ::
  "pub_indx set \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_ptr_success ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        hstate sp = LOK
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> ptr_state sp = LOK
        \<and> sp' = sp\<lparr> ptr := ptr_set \<rparr>
  }"

definition write_ptr_fail ::
  "pub_indx set \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_ptr_fail ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' hst pst.
        (hst = LDead \<or> pst = LDead)
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> hst \<le> hstate sp
        \<and> pst \<le> ptr_state sp
        \<and> sp' = sp\<lparr> hstate := hst, ptr_state := pst \<rparr>
        \<and> sv' = sv\<lparr> card_mode := CFail \<rparr>
  }"

definition write_ptr_deg ::
  "pub_indx set \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_ptr_deg ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB False))), sv', sp')|sv' sp'.
        ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> sv' = sv\<lparr> card_mode := CDegradedP \<rparr>
        \<and> sp' = sp\<lparr> ptr_state := LDegraded \<rparr>
  }"
        
definition write_ptr_part_deg ::
  "pub_indx set \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_ptr_part_deg ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB True))), sv', sp')|sv' sp'.
        ptr_state sp \<ge> LDegraded
        \<and> hstate sp \<ge> LDegraded
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> sv' = sv\<lparr> card_mode := CDegradedP \<rparr>
        \<and> sp' = sp\<lparr> ptr_state := LDegraded, ptr := ptr_set \<rparr>
  }"
 
definition write_ptr_deg_h ::
  "pub_indx set \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_ptr_deg_h ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB False))), sv', sp')|sv' sp' pst.
        pst \<le> ptr_state sp
        \<and> LDegraded \<le> pst
        \<and> hstate sp \<ge> LDegraded
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> sv' = sv\<lparr> card_mode := CDegradedH \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded, ptr_state := pst \<rparr>
  }"

definition write_ptr_part_deg_h ::
  "pub_indx set \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_ptr_part_deg_h ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB True))), sv', sp')|sv' sp' pst.
        ptr_state sp \<ge> LDegraded
        \<and> pst \<le> ptr_state sp
        \<and> LDegraded \<le> pst
        \<and> hstate sp \<ge> LDegraded
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> sv' = sv\<lparr> card_mode := CDegradedH \<rparr>
        \<and> sp' = sp\<lparr> hstate := LDegraded, ptr_state := pst, ptr := ptr_set \<rparr>
  }"

(* TODO: is this necessary? In the previous version, we
  forgot to change the volatile memory, yet the refinement still
  seemingly went through, so something looks wrong *)
definition write_ptr_partial_restart ::
  "pub_indx set \<Rightarrow> ('a, 'c :: size)abs3_monad" where
  "write_ptr_partial_restart ptr_set \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp'.
        hstate sp \<ge> LDegraded
        \<and> ptr_state sp \<ge> LDegraded
        \<and> card_mode sv \<notin> {CFail, CDegradedH, CDegradedP}
        \<and> sp' = sp\<lparr> ptr := ptr_set \<rparr>
  }"

definition write_ptr_throw_immediately
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_ptr_throw_immediately ptr_set \<equiv> \<lambda>(sv, sp).
    {(Some (Inl e), sv, sp)|e.
        e = Degraded (EB False)
        \<and> card_mode sv \<in> {CFail, CDegradedH, CDegradedP}
    }
  "

definition write_ptr :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  "write_ptr ptr_set \<equiv>
    write_ptr_success ptr_set
    [+] write_ptr_fail ptr_set
    [+] write_ptr_deg ptr_set
    [+] write_ptr_part_deg ptr_set
    [+] write_ptr_deg_h ptr_set
    [+] write_ptr_part_deg_h ptr_set
    [+] write_ptr_partial_restart ptr_set
    [+] write_ptr_throw_immediately ptr_set
    [+] restart_mangle_sv
  "

lemmas write_ptr_defs =
  write_ptr_def
  write_ptr_success_def
  write_ptr_fail_def
  write_ptr_deg_def
  write_ptr_part_deg_def
  write_ptr_deg_h_def
  write_ptr_part_deg_h_def
  write_ptr_partial_restart_def
  write_ptr_throw_immediately_def

definition read_DegradedH :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad" where
   read_DegradedH_def[monad_def]: 
  "read_DegradedH ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> get_hashes_DegradedH;
    (if ht(ind, inst) \<noteq> Hash cnt
        then throw FailRead
        else return cnt)
  }"

definition read_DegradedH_or_fail
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad" where
   read_DegradedH_or_fail_def[monad_def]: 
  "read_DegradedH_or_fail ind inst \<equiv>
  <try> read_DegradedH ind inst <catch> K (do { set_failed; throw FailMode })"

definition read :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad" where
   read_def[monad_def]:
  "read ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> (
        <try>
            get_hashes
        <catch> (\<lambda>e. case e of
            Degraded _ \<Rightarrow> do {
                ht \<leftarrow> get_hashes_DegradedH;
                (if ht(ind, inst) = Hash cnt
                    then throw (Degraded (EC cnt))
                    else throw (Degraded EU)
                )
            }
            | _ \<Rightarrow> throw e
        )
    );
    (if ht(ind, inst) \<noteq> Hash cnt
        then throw FailRead
        else return cnt)
  }"

definition read_or_fail :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad" where
  read_or_fail_def[monad_def]:
  "read_or_fail pind inst \<equiv> 
    <try>
        read pind inst
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do { set_failed; throw FailMode }
            else throw (Degraded v)
        | _ \<Rightarrow> do { set_failed; throw FailMode }
    )
  "

definition read_or_fail_silence_deg
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad"
  where read_or_fail_silence_deg_def[monad_def]:
  "read_or_fail_silence_deg ind inst \<equiv>
    <try> read_or_fail ind inst <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (get_exc_val_cnt v)
        | _ \<Rightarrow> throw e
    )
  "
 
definition update_instance_hash :: 
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  update_instance_hash_def[monad_def]:
  "update_instance_hash ind inst cnt \<equiv> do {
        ht \<leftarrow> get_hashes;
        write_hashes (ht((ind, inst) := Hash cnt))
  }
  "

definition write_instance :: 
  "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  write_instance_def[monad_def]:
  "write_instance ind inst cnt \<equiv> do {
    mode \<leftarrow> get_mode;
    (if mode = CDegradedH
    then throw (Degraded (EB False))
    else return());
    (<try>
        write_f ind inst cnt
    <catch> (\<lambda>e. case e of
        InvalidInput \<Rightarrow> throw InvalidInput
        | FailWrite \<Rightarrow> do {
            set_degraded;
            (<try> 
                read ind inst
            <catch> (\<lambda>e. case e of
                Degraded v \<Rightarrow>
                    if v = EU
                    then throw (Degraded (EB False))
                    else throw (Degraded (EB True))
                | FailRead \<Rightarrow> throw (Degraded (EB False))
                | _ \<Rightarrow> throw e)
            );
            throw (Degraded (EB True))
        }
        | _ \<Rightarrow> throw e
    ));
    <try>
        update_instance_hash ind inst cnt
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then throw (Degraded v)
            else
                do {
                      cnt \<leftarrow> (
                          <try>
                              read_DegradedH ind inst
                          <catch> (\<lambda>e. case e of
                              FailMode \<Rightarrow> throw FailMode
                              | _ \<Rightarrow> throw (Degraded (EB False))
                          )
                      );
                      throw (Degraded (EB True))
                  }
        | _ \<Rightarrow> throw e
    )
  }"

definition sync ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs3_monad" where
   sync_def[monad_def]:
  "sync ind to from \<equiv> do {
    cnt \<leftarrow> (
        <try>
            read_or_fail ind from
        <catch> (\<lambda>e. case e of
            Degraded v  \<Rightarrow> throw (Degraded (EC (get_exc_val_cnt v)))
            | _ \<Rightarrow> throw e
        )
    );
    (<try>
        write_instance ind to cnt
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> throw (Degraded (EC cnt))
        | _ \<Rightarrow> throw e
    ));
    return cnt
  }"

definition just_sync ::
  "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> (unit, 'c :: size)abs3_monad" where
   just_sync_def[monad_def]:
  "just_sync ind to from \<equiv> sync ind to from \<guillemotright> return ()"

(* TODO: in the more concrete models,
  change the case to an if in the catch cause below *)
definition just_sync_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c :: size)abs3_monad"
  where just_sync_silence_deg_def[monad_def]:
  "just_sync_silence_deg to from ind \<equiv>
    <try> 
        just_sync ind to from
    <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "

definition remove_from_ptr :: "pub_indx \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  remove_from_ptr_def[monad_def]:
  "remove_from_ptr ind \<equiv> do {
    ptr \<leftarrow> get_ptr False;
    let ptr' = ptr - {ind};
    write_ptr ptr'
  }"

text {* This definition serves just to facilitate the abs2/4
  refinement proof; after get_ptr, we know that at least one PTR is
  good, so we don't "bring the PTR back from the dead"
*}

definition set_ptr :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs3_monad" where
  set_ptr_def[monad_def]:
  "set_ptr ptr_set \<equiv> do {
    get_ptr False;
    write_ptr ptr_set
  }"

definition get_hashes_silence_deg
  :: "('c pub_hashes, 'c :: size)abs3_monad" where
  get_hashes_silence_deg_def[monad_def]:
  "get_hashes_silence_deg \<equiv>
    <try>
        get_hashes
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> get_hashes_DegradedH
        | _ \<Rightarrow> throw e
    )
  "

definition get_sync_direction
  :: "pub_indx \<Rightarrow> (sync_direction, 'c :: size)abs3_monad"
  where
  get_sync_direction_def[monad_def]:
  "get_sync_direction pind \<equiv> do {
    ht \<leftarrow> get_hashes_silence_deg;
    if (ht(pind, Primary) = ht(pind, Secondary))
    then (return SEq)
    else do {
        ptr_set \<leftarrow> get_ptr True;
        (if pind \<in> ptr_set
        then return SL
        else return SR)
    }
  }"

definition remove_from_ptr_silence_deg where
  remove_from_ptr_silence_deg_def[monad_def]:
  "remove_from_ptr_silence_deg ind \<equiv>
        <try>
            remove_from_ptr ind
        <catch> (\<lambda>e. if e = FailMode then throw FailMode else return ())
  "

definition sync_and_remove_from_ptr_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c :: size)abs3_monad"
  where sync_and_remove_from_ptr_silence_deg_def[monad_def]:
  "sync_and_remove_from_ptr_silence_deg to from ind \<equiv>
    <try> do {
        sync ind to from;
        remove_from_ptr_silence_deg ind
    } <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> return ()
        | _ \<Rightarrow> throw e
    )
  "

end