theory Abs3_Invariant_Lemmas
  imports Abs3_StartupDefs Abs3_UpdateDefs Abs3_FetchDefs Abs3_Lemmas 
  "../HoareLogic" "../Pmm_Tacs" "../Pmm_Lemmas" "../RMonad_Lemmas"
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs
  Abs3_LLInterface.write_f_defs
  Abs3_LLInterface.ll_read_defs

definition ptr_not_dead where
  "ptr_not_dead \<equiv> {(svc, spc). ptr_state spc \<noteq> LDead}"

definition card_mode_inv where
  "card_mode_inv \<equiv> {(sv, sp).
    (card_mode sv = CDegradedH \<longrightarrow> (hstate sp = LDegraded \<and> LDegraded \<le> ptr_state sp ))
    \<and> (card_mode sv = CDegradedP \<longrightarrow> (hstate sp = LOK \<and> ptr_state sp = LDegraded ))
    \<and> (card_mode sv = CDegraded \<or> card_mode sv = CNormal \<longrightarrow>
        hstate sp = LOK \<and> ptr_state sp = LOK)
  }"

definition abs3_capacity_inv where
  "abs3_capacity_inv \<equiv> {(sv, sp). \<forall>ind inst. size (memory sp ind inst) \<le> capacity (Public ind)}"

definition not_CFail_inv where
  "not_CFail_inv \<equiv> {(sv, sp). card_mode sv \<noteq> CFail }"
    
definition abs3_inv where
  "abs3_inv = card_mode_inv"

definition abs3_norm_inv where
  "abs3_norm_inv = abs3_inv \<inter> not_CFail_inv"

lemmas abs3_inv_defs =
  abs3_inv_def
  abs3_norm_inv_def
  not_CFail_inv_def
  card_mode_inv_def
  abs3_capacity_inv_def
 
definition empty_ptr where
  "empty_ptr = {(svc, spc). card_mode svc = CNormal \<longrightarrow> ptr spc = {}}"

definition precommit_inv where
  "precommit_inv \<equiv> abs3_norm_inv \<inter> sv_mode_equals CNormal \<inter> empty_ptr"

lemmas precommit_inv_defs =
  abs3_inv_defs
  precommit_inv_def
  empty_ptr_def
  sv_mode_equals_def

lemma read_post_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> 
    Abs3_CommonDefs.read pind inst
  \<lbrace>>N K abs3_norm_inv \<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_def K_def intro!: bind_ruleNN_same)
  apply(clarsimp simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs abs3_inv_defs
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    split: exceptions.split_asm split_if_asm)
  done

lemma read_post_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> 
    Abs3_CommonDefs.read pind inst
  \<lbrace>@N \<lambda>e. if e \<noteq> FailMode then abs3_norm_inv else UNIV \<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_def K_def intro!: bind_ruleNE_same)
  apply(clarsimp simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs abs3_inv_defs
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    split: exceptions.split_asm split_if_asm)
  apply(auto simp add: min_def split: split_if_asm dest: card_mode_exhaust)
  done

lemma HoareNN_return_same:
  "\<lbrace>P\<rbrace> return x \<lbrace>>N K P\<rbrace>"
  by(auto simp add: HoareNN_alt project_simps intro!: returnN_rule')

lemma get_hashes_silence_deg_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> get_hashes_silence_deg \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def project_defs)
  apply(rename_tac svc spc v svc' spc')
  apply(clarsimp simp add: get_hashes_silence_deg_def 
    get_hashes_all_defs abs3_inv_defs
    brute_force split: exceptions.split_asm split_if_asm)
  apply(case_tac "card_mode svc", iterative_simp simp add: min_def split: split_if_asm)
  done

lemma get_sync_direction_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> get_sync_direction pind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: get_sync_direction_def first_stage HoareNN_return'
    intro!: bind_ruleNN_same get_hashes_silence_deg_inv_NN)
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def project_defs)
  apply(rename_tac svc spc v svc' spc')
  apply(clarsimp simp add: get_ptr_defs abs3_inv_defs brute_force
    split: exceptions.split_asm split_if_asm)
  apply(case_tac "card_mode svc", iterative_simp simp add: min_def split: split_if_asm)
  done

lemma get_hashes_silence_deg_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> get_hashes_silence_deg \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def project_defs)
  apply(rename_tac svc spc v svc' spc')
  apply(clarsimp simp add: get_hashes_silence_deg_def 
    get_hashes_all_defs abs3_inv_defs
    brute_force split: exceptions.split_asm split_if_asm)
  apply(case_tac "card_mode svc", iterative_simp simp add: min_def split: split_if_asm)
  done

lemma get_sync_direction_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> get_sync_direction pind \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: get_sync_direction_def first_stage HoareNN_return'
    intro!: bind_ruleNN_same get_hashes_silence_deg_capacity_inv_NN)
  apply(clarsimp_all simp add: HoareNN_alt Hoare_def Box_def project_defs)
  apply(rename_tac[!] svc spc v svc' spc')
  apply(clarsimp simp add: get_ptr_defs abs3_capacity_inv_def brute_force
    split: exceptions.split_asm split_if_asm)
  apply(case_tac[!] "card_mode svc")
  apply(iterative_simp simp add: min_def split: split_if_asm)
  done

lemma get_sync_direction_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> get_sync_direction pind \<lbrace>@N K abs3_inv\<rbrace>"
  apply(auto simp only: get_sync_direction_def
    intro!: bind_ruleNE get_hashes_silence_deg_inv_NN)
  apply(clarsimp simp add: HoareNE_alt Hoare_def Box_def project_defs)
  apply(rename_tac svc spc e svc' spc')
  apply(clarsimp simp add: get_hashes_silence_deg_def 
    get_hashes_all_defs abs3_inv_defs
    brute_force split: exceptions.split_asm split_if_asm)
  apply(case_tac "card_mode svc", iterative_simp simp add: min_def split: split_if_asm)
  apply(auto simp add: K_beta intro!: HoareNE_return)
  apply(clarsimp simp add: HoareNE_alt Hoare_def Box_def project_defs)
  apply(rename_tac svc spc e svc' spc')
  apply(clarsimp simp add: get_ptr_defs abs3_inv_defs brute_force
    split: exceptions.split_asm split_if_asm)
  done

lemma get_mode_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> get_mode \<lbrace>>N K abs3_norm_inv\<rbrace>"
  by(clarsimp simp add: HoareNN_alt Hoare_def Box_def abs3_inv_defs brute_force min_def)

lemma get_mode_inv_NN':
  "\<lbrace>abs3_norm_inv\<rbrace> get_mode \<lbrace>>N \<lambda>m. abs3_norm_inv \<inter> sv_mode_equals m \<rbrace>"
  by(clarsimp simp add: HoareNN_alt Hoare_def Box_def abs3_inv_defs sv_mode_equals_def
    brute_force min_def)

lemma read_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.read pind inst
  \<lbrace>>N K abs3_norm_inv \<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp simp add: abs3_inv_defs)
  apply(clarsimp_all simp add: get_hashes_all_defs brute_force split: split_if_asm)
  done
 
lemma read_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.read pind inst
  \<lbrace>@N \<lambda>e.
    if e = FailMode
    then abs3_inv
    else abs3_norm_inv
  \<rbrace>"
  apply(clarsimp simp add: HoareNE_alt Hoare_def Box_def K_beta split: split_if)
  apply(rename_tac svc spc e)
  apply(safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_hashes_all_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  apply(clarsimp_all simp add: abs3_inv_defs sv_mode_equals_def FailMode_exhaust)
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all)
  done
 
lemma write_instance_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>>N K abs3_norm_inv \<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs
    brute_force abs3_inv_defs split: split_if_asm)
  done

lemma write_instance_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>@N \<lambda>e.
    if e = FailMode
    then abs3_inv
    else abs3_norm_inv
  \<rbrace>"
  apply(clarsimp simp add: HoareNE_alt Hoare_def Box_def K_beta)
  apply(rename_tac svc spc e)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs3_CommonDefs.get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs min_def
    brute_force abs3_inv_defs sv_mode_equals_def split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms card_mode_exhaust} *}, simp_all)
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(iterative_simp)+
  apply(case_tac[!] aa)
  apply(clarsimp_all)
  apply(iterative_simp)+
  apply(case_tac[!] spc)
  using [[simp_depth_limit=100]]
  apply(case_tac[!] "svc")
  apply(clarsimp_all simp add: fun_upd_second_same)
  done

lemma remove_from_ptr_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> remove_from_ptr pind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(rename_tac svc spc svc' spc')
  apply(simp_all only: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_ptr_defs write_ptr_defs
    split: exceptions.split_asm split_if_asm)
  apply(clarsimp_all simp add: abs3_inv_defs)
  done

lemma remove_from_ptr_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> remove_from_ptr pind \<lbrace>@N \<lambda>e. if e = FailMode then abs3_inv else abs3_norm_inv\<rbrace>"
  apply(auto simp only: HoareNE_alt Hoare_def Box_def)
  apply(rename_tac svc spc e svc' spc')
  apply(auto)
  apply(simp_all only: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_ptr_defs write_ptr_defs
    split: exceptions.split_asm split_if_asm)
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all simp add: abs3_inv_defs min_def)
  done

lemma read_or_fail_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_CommonDefs.read_or_fail pind inst \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(simp only: Abs3_CommonDefs.read_or_fail_def)
  apply(auto simp add: K_beta
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    HoareNN_throw
    read_inv_NN
    read_inv_NE
    true_postNN[where m="set_failed"]
    split: exceptions.split
  )
  done

lemma set_failed_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> set_failed \<lbrace>>N K abs3_inv\<rbrace>"
  "\<lbrace>abs3_inv\<rbrace> set_failed \<lbrace>>N K abs3_inv\<rbrace>"
  by(clarsimp_all simp add: HoareNN_alt Hoare_def Box_def brute_force abs3_inv_defs)

lemma read_or_fail_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_CommonDefs.read_or_fail pind inst
  \<lbrace>@N \<lambda>e.
    if e = FailMode
    then abs3_inv
    else abs3_norm_inv 
  \<rbrace>"
  apply(simp only: Abs3_CommonDefs.read_or_fail_def)
  apply(auto simp add: K_beta HoareNE_throw'
    intro!:
    try_catch_ruleNE
    bind_ruleNE
    HoareNN_throw
    HoareNE_vol_ops
    read_inv_NN
    read_inv_NE
    set_failed_inv_NN
    split: exceptions.split
  )
  done

lemma sync_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.sync to from pind
  \<lbrace>>N K abs3_norm_inv \<rbrace>"
  apply(simp_all only: Abs3_CommonDefs.sync_def)
  apply(auto simp add: K_beta
    intro!: bind_ruleNN try_catch_ruleNN
    read_inv_NN 
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    read_inv_NE
    HoareNN_return_same
    HoareNN_throw
    write_instance_inv_NN write_instance_inv_NE
    split: exceptions.split
  )
  done
 
lemma sync_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace>
    Abs3_CommonDefs.sync to from pind
  \<lbrace>@N \<lambda>e. if e = FailMode then abs3_inv else abs3_norm_inv \<rbrace>"
  by(auto simp add: Abs3_CommonDefs.sync_def HoareNE_throw' K_beta
    intro!:
    try_catch_ruleNN
    try_catch_ruleNE
    bind_ruleNE
    bind_ruleNN
    HoareNN_throw
    HoareNE_return
    HoareNE_vol_ops
    read_inv_NN 
    read_inv_NE
    read_or_fail_inv_NN 
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    set_failed_inv_NN
    split: exceptions.split
  )

lemma check_pub_same_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_StartupDefs.check_pub_same pind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(simp only: Abs3_StartupDefs.check_pub_same_def
    Abs3_CommonDefs.just_sync_silence_deg_def Abs3_CommonDefs.just_sync_def
    Abs3_CommonDefs.read_or_fail_silence_deg_def
  )
  apply(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    try_catch_ruleNE
    bind_ruleNE
    HoareNN_throw
    HoareNE_return
    HoareNN_return_same
    read_inv_NN
    read_inv_NE
    sync_inv_NN
    sync_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    split: exceptions.split
    del: subsetI
  )
  done


lemma norm_inv_subset_inv:
  "abs3_norm_inv \<subseteq> abs3_inv"
  by(simp add: abs3_norm_inv_def)

lemma check_pub_same_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_StartupDefs.check_pub_same pind \<lbrace>@N K abs3_inv\<rbrace>"
  apply(simp only: Abs3_StartupDefs.check_pub_same_def
    Abs3_CommonDefs.just_sync_silence_deg_def Abs3_CommonDefs.just_sync_def
    Abs3_CommonDefs.read_or_fail_silence_deg_def
  )
  apply(auto simp add: K_beta HoareNE_throw' HoareNN_return'
    intro!:
    bind_ruleNE
    try_catch_ruleNE
    try_catch_ruleNN
    read_inv_NN
    read_inv_NE
    sync_inv_NN
    sync_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    HoareNE_return
    HoareNN_throw
    norm_inv_subset_inv
    intro: ex_falsumNE ex_falsumNN
    split: exceptions.split
    del: subsetI
  )
  done

lemma check_pub_region_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace>
  Abs3_StartupDefs.check_pub_region pind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(simp only:
    Abs3_StartupDefs.check_pub_region_def
    Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg_def
    Abs3_CommonDefs.remove_from_ptr_silence_deg_def
  )
  apply(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    try_catch_ruleNE
    bind_ruleNE
    split: exceptions.split
    del: subsetI
  )
  apply(auto simp add: K_beta
    intro!:
    HoareNN_throw
    HoareNE_return
    HoareNN_return_same
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NN[simplified K_def]
    remove_from_ptr_inv_NE
    read_inv_NN
    read_inv_NE
    sync_inv_NN
    sync_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    check_pub_same_inv_NN
  )
  done

lemma check_pub_region_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace>
  Abs3_StartupDefs.check_pub_region pind \<lbrace>@N K abs3_inv\<rbrace>"
  apply(simp only:
    Abs3_StartupDefs.check_pub_region_def
    Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg_def
    Abs3_CommonDefs.remove_from_ptr_silence_deg_def
  )
  apply(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    try_catch_ruleNE
    bind_ruleNE
    split: exceptions.split
    del: subsetI
  )
  apply(auto simp add: K_beta
    intro!:
    HoareNN_throw
    HoareNE_return
    HoareNN_return_same
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NN[simplified K_def]
    remove_from_ptr_inv_NE
    read_inv_NN
    read_inv_NE
    sync_inv_NN
    sync_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    check_pub_same_inv_NN
    check_pub_same_inv_NE[simplified K_def]
    norm_inv_subset_inv
    del: subsetI
  )
  done

lemma startup1_inv_NN:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup1 \<lbrace>>N K abs3_norm_inv\<rbrace>" 
  by(auto simp add: HoareNN_alt Hoare_def Box_def abs3_inv_defs
    Abs3_StartupDefs.startup1_def Abs3_StartupDefs.startup1_NN_def compose_project
    choice_def K_def 
  )

lemma startup1_inv_NE:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup1 \<lbrace>@N K abs3_inv\<rbrace>" 
  by(auto simp add: HoareNE_alt Hoare_def Box_def abs3_inv_defs
    Abs3_StartupDefs.startup1_def Abs3_StartupDefs.startup1_NE_def compose_project
    choice_def K_def 
  )

lemma set_degraded_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> set_degraded \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def abs3_inv_defs brute_force min_def)
  by (metis card_mode.exhaust less_eq_card_mode.simps(11) less_eq_card_mode.simps(16))

declare doesnt_throw_raw_bindN'[no_throw]


definition removed_one_from_ptr :: "pub_indx \<Rightarrow> (pmm_vol_mem \<times> 'c abs3_state)set" where
  "removed_one_from_ptr pind \<equiv> {(svc, spc). card_mode svc = CNormal \<longrightarrow> pind \<notin> ptr spc}"

lemma remove_from_ptr_removed_inv_NN:
  "\<lbrace>P\<rbrace> remove_from_ptr_silence_deg pind \<lbrace>>N K (removed_one_from_ptr pind) \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(simp_all only: remove_from_ptr_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: get_ptr_defs write_ptr_defs brute_force
    removed_one_from_ptr_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp simp add: min_def split: split_if_asm)+
  done

lemma remove_from_ptr_fails_post_NE:
  "\<lbrace>P\<rbrace> remove_from_ptr_silence_deg pind
  \<lbrace>@N \<lambda>e. if e = FailMode then sv_mode_equals CFail else {}\<rbrace>" 
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: remove_from_ptr_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_ptr_defs write_ptr_defs brute_force
    sv_mode_equals_def
    split: exceptions.split_asm split_if_asm)
  done

lemma InvalidInput_exhaust:
  "\<lbrakk>e \<noteq> FailLLRead; e \<noteq> FailRead; e \<noteq> FailMode; e \<noteq> FailWrite;
  \<forall>c. e \<noteq> Degraded c; e \<noteq> StartedExc\<rbrakk>
    \<Longrightarrow> (e = InvalidInput) = True"
  "\<lbrakk>e \<noteq> FailLLRead; e \<noteq> FailRead; e \<noteq> FailMode; e \<noteq> FailWrite ;
  \<forall>c. e \<noteq> Degraded c; e \<noteq> StartedExc\<rbrakk>
    \<Longrightarrow> (InvalidInput = e) = True"
  by(case_tac[!] e, simp_all)

lemma pmm_vol_mem_equality_parts:
  "((s :: pmm_vol_mem) = s') = (card_mode s = card_mode s' \<and> updated s = updated s')"
  by(case_tac s, case_tac[!] s', simp_all)

lemma sync_mode_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_CommonDefs.sync pind to from
  \<lbrace>@N \<lambda>e. if e = FailMode then sv_mode_equals CFail
    else if e \<in> {FailRead, FailLLRead, FailWrite, StartedExc, InvalidInput} 
        then {} else sv_mode_neq CNormal\<rbrace>" 
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: Abs3_CommonDefs.sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs3_CommonDefs.write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps write_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  apply(simp_all add: brute_force get_hashes_all_defs write_hashes_defs
    abs3_capacity_inv_def
    split: exceptions.split_asm split_if_asm)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, clarsimp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms InvalidInput_exhaust} *}, clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: sv_mode_neq_def sv_mode_equals_def
    pmm_vol_mem_equality_parts
    FailMode_exhaust min_def split: split_if_asm)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, clarsimp_all)
  apply(blast dest!: leD)+
  apply(case_tac[!] b, case_tac[!] bb, clarsimp_all simp add: fun_upd_second_same)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm conjE})))) *})
  apply(simp_all (asm_lr))
  apply(clarsimp_all)
  done

lemma just_sync_mode_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_CommonDefs.just_sync pind to from
  \<lbrace>@N \<lambda>e. if e = FailMode then sv_mode_equals CFail
    else if e \<in> {FailRead, FailLLRead, FailWrite, StartedExc, InvalidInput} 
        then {} else sv_mode_neq CNormal\<rbrace>" 
  by(auto simp add: Abs3_CommonDefs.just_sync_def
    intro!: bind_ruleNE HoareNE_return sync_mode_post_NE intro: true_postNN)

lemma not_CNormal_imp_removed:
  "sv_mode_neq CNormal \<subseteq> removed_one_from_ptr ind"
  by(auto simp add: sv_mode_neq_def removed_one_from_ptr_def)

lemma sync_and_remove_from_ptr_silence_deg_removed_post_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync_and_remove_from_ptr_silence_deg to from pind  
    \<lbrace>>N K (removed_one_from_ptr pind) \<rbrace>"
  apply(simp only: sync_and_remove_from_ptr_silence_deg_def)
  apply(auto intro!: try_catch_ruleNN bind_ruleNN bind_ruleNE HoareNN_throw
    remove_from_ptr_removed_inv_NN
    sync_mode_post_NE
    conseqNE_rule[OF _ remove_from_ptr_fails_post_NE]
    dest: not_CNormal_imp_removed[THEN subsetD]
    intro: true_postNN true_postNE
    split: exceptions.split
    simp add: HoareNN_return' K_beta
  )
  done

lemma sync_and_remove_from_ptr_silence_deg_failed_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync_and_remove_from_ptr_silence_deg pind to from \<lbrace>@N K (sv_mode_equals CFail) \<rbrace>"
  apply(simp only: sync_and_remove_from_ptr_silence_deg_def)
  apply(auto intro!: try_catch_ruleNE try_catch_ruleNN bind_ruleNN bind_ruleNE
    HoareNN_throw HoareNE_return
    remove_from_ptr_removed_inv_NN
    sync_mode_post_NE
    conseqNE_rule[OF _ remove_from_ptr_fails_post_NE]
    intro: true_postNN true_postNE
    split: exceptions.split
    simp add: HoareNN_return' HoareNE_throw' K_beta
  )
  done

lemma read_capacity_inv:
  "\<lbrace>abs3_capacity_inv\<rbrace>
    Abs3_CommonDefs.read pind inst
  \<lbrace>>K abs3_capacity_inv \<rbrace>"
  apply(clarsimp simp add: Hoare_def Box_def K_beta)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp simp add: abs3_inv_defs)
  apply(clarsimp_all simp add: get_hashes_all_defs brute_force split: split_if_asm)
  apply(iterative_simp)+
  done
 
(* TODO: move these to HoareLogic, together with the copies from Abs1_Invariant_Lemmas *)
lemma Hoare_imp_HoareNN:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>>N K Q\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def K_beta project_defs)

lemma Hoare_imp_HoareNE:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@N K Q\<rbrace>"
  by(auto simp add: HoareNE_alt Hoare_def Box_def K_beta project_defs)

lemma Hoare_imp_HoareR:
  "\<lbrace>P\<rbrace> m \<lbrace>> K Q\<rbrace> \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>@R Q\<rbrace>"
  by(auto simp add: HoareR_def BoxR_def Hoare_def Box_def K_beta project_defs)

lemma write_instance_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace>
    Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>>N K abs3_capacity_inv \<rbrace>"
  apply(clarsimp simp add: Hoare_def Hoare_alt Box_def K_beta)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs
    brute_force abs3_inv_defs split: split_if_asm)
  done

lemma write_instance_capacity_inv_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace>
    Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>@N K abs3_capacity_inv \<rbrace>"
  apply(clarsimp simp add: HoareNE_alt Hoare_def Box_def K_beta)
  apply(rename_tac svc spc e svc' spc')
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs3_CommonDefs.get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs min_def
    brute_force abs3_inv_defs split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp)+
  apply(case_tac[!] "a")
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all)
  apply(iterative_simp)+
  done

lemma vol_ops_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> set_failed \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(auto simp add: brute_force abs3_inv_defs)
  done

lemma sync_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync pind to from \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: sync_def)
  apply(auto intro!: try_catch_ruleNN_same try_catch_ruleNE_same bind_ruleNE_same
    HoareNN_throw bind_ruleNN_same ex_falsumNN 
    read_capacity_inv[THEN Hoare_imp_HoareNN]
    read_capacity_inv[THEN Hoare_imp_HoareNE]
    vol_ops_capacity_inv_NN HoareNE_vol_ops
    write_instance_capacity_inv_NN write_instance_capacity_inv_NE
    simp add: HoareNN_return' HoareNE_throw' K_beta
    read_or_fail_def
    split: exceptions.split)
  done

lemma sync_capacity_inv_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync pind to from \<lbrace>@N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: sync_def)
  apply(auto intro!: try_catch_ruleNN_same try_catch_ruleNE_same bind_ruleNE_same
    HoareNN_throw HoareNE_return bind_ruleNN_same ex_falsumNN 
    read_capacity_inv[THEN Hoare_imp_HoareNN]
    read_capacity_inv[THEN Hoare_imp_HoareNE]
    vol_ops_capacity_inv_NN HoareNE_vol_ops
    write_instance_capacity_inv_NN write_instance_capacity_inv_NE
    simp add: HoareNN_return' HoareNE_throw' K_beta
    read_or_fail_def
    split: exceptions.split)
  done

lemma check_pub_same_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_same pind \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(simp only: Abs3_StartupDefs.check_pub_same_def
    Abs3_CommonDefs.just_sync_silence_deg_def Abs3_CommonDefs.just_sync_def
    Abs3_CommonDefs.read_or_fail_silence_deg_def read_or_fail_def
  )
  apply(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    intro!:
    try_catch_ruleNN_same
    bind_ruleNN_same
    try_catch_ruleNE_same
    bind_ruleNE_same
    HoareNN_throw
    HoareNE_return
    HoareNN_return_same
    read_capacity_inv[THEN Hoare_imp_HoareNN]
    read_capacity_inv[THEN Hoare_imp_HoareNE]
    sync_capacity_inv_NN
    sync_capacity_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    vol_ops_capacity_inv_NN
    HoareNE_vol_ops
    split: exceptions.split
    del: subsetI
  )
  done

lemma check_pub_same_capacity_inv_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_same pind \<lbrace>@N K abs3_capacity_inv\<rbrace>"
  apply(simp only: Abs3_StartupDefs.check_pub_same_def
    Abs3_CommonDefs.just_sync_silence_deg_def Abs3_CommonDefs.just_sync_def
    Abs3_CommonDefs.read_or_fail_silence_deg_def read_or_fail_def
  )
  apply(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    intro!:
    try_catch_ruleNN_same
    bind_ruleNN_same
    try_catch_ruleNE_same
    bind_ruleNE_same
    HoareNN_throw
    HoareNE_return
    HoareNN_return_same
    read_capacity_inv[THEN Hoare_imp_HoareNN]
    read_capacity_inv[THEN Hoare_imp_HoareNE]
    sync_capacity_inv_NN
    sync_capacity_inv_NE
    read_or_fail_inv_NN
    read_or_fail_inv_NE
    write_instance_inv_NN
    write_instance_inv_NE
    vol_ops_capacity_inv_NN
    HoareNE_vol_ops
    split: exceptions.split
    del: subsetI
  )
  done

lemma check_pub_region_removed_post_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_region pind \<lbrace>>N K (removed_one_from_ptr pind) \<rbrace>"
  by(auto simp add: Abs3_StartupDefs.check_pub_region_def K_beta
    intro!:
    get_sync_direction_capacity_inv_NN
    sync_and_remove_from_ptr_silence_deg_removed_post_NN
    bind_ruleNN bind_ruleNE
    remove_from_ptr_removed_inv_NN
    intro: true_postNN
  )

lemma read_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> Abs3_CommonDefs.read pind inst \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_def K_def intro!: bind_ruleNN_same)
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    split: exceptions.split_asm split_if_asm)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  done

lemma read_doesnt_add_post_NE:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> Abs3_CommonDefs.read pind inst \<lbrace>@N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_def K_def intro!:
    bind_ruleNN_same try_catch_ruleNE_same)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps restart_mangle_sv_proj_simps if_choice
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  done

lemma read_DegradedH_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> read_DegradedH pind inst \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_DegradedH_def K_def intro!: bind_ruleNN_same)
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(auto simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    split: exceptions.split_asm split_if_asm)
  done

lemma read_DegradedH_doesnt_add_post_NE:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> Abs3_CommonDefs.read_DegradedH pind inst
  \<lbrace>@N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.read_DegradedH_def K_def intro!:
    bind_ruleNN_same try_catch_ruleNE_same)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps restart_mangle_sv_proj_simps if_choice
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    min_def
    split: exceptions.split_asm split_if_asm)
  done

lemma vol_ops_dont_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> set_failed \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> set_degraded \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(auto simp add: brute_force removed_one_from_ptr_def min_def
    split: exceptions.split_asm split_if_asm)
  done

lemma write_instance_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.write_instance_def K_beta
    intro!: bind_ruleNN_same bind_ruleNE_same try_catch_ruleNN_same
    HoareNN_throw
    read_doesnt_add_post_NN
    read_doesnt_add_post_NE
    read_DegradedH_doesnt_add_post_NN
    read_DegradedH_doesnt_add_post_NE
    vol_ops_dont_add_post_NN
    HoareNN_get_mode
    split: exceptions.split
  )
  apply(auto simp add: Abs3_CommonDefs.update_instance_hash_def K_def intro!: bind_ruleNN_same)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    write_hashes_defs
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp simp add: min_def split: split_if_asm)+
  done

lemma write_instance_doesnt_add_post_NE:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> Abs3_CommonDefs.write_instance pind inst cnt
  \<lbrace>@N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: Abs3_CommonDefs.write_instance_def K_beta HoareNE_throw' HoareNN_return'
    intro!: bind_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same try_catch_ruleNN_same
    HoareNE_return HoareNN_throw
    read_doesnt_add_post_NN
    read_doesnt_add_post_NE
    read_DegradedH_doesnt_add_post_NN
    read_DegradedH_doesnt_add_post_NE
    vol_ops_dont_add_post_NN
    HoareNN_get_mode
    HoareNE_vol_ops
    split: exceptions.split
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp add: Abs3_CommonDefs.update_instance_hash_def K_def intro!: bind_ruleNE_same)
  apply(clarsimp_all simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage
    vol_ops_project_simps)
  apply(auto simp add: brute_force get_hashes_all_defs removed_one_from_ptr_def
    write_hashes_defs
    HoareNN_alt HoareNE_alt Hoare_def Box_def first_stage vol_ops_project_simps
    min_def
    split: exceptions.split_asm split_if_asm)
  done

lemma remove_from_ptr_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> remove_from_ptr pind \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(auto simp add: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(auto simp add: brute_force get_ptr_defs write_ptr_defs removed_one_from_ptr_def
    split: exceptions.split_asm split_if_asm)
  done

lemma remove_from_ptr_doesnt_add_post_NE:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> remove_from_ptr pind \<lbrace>@N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(auto simp add: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: brute_force get_ptr_defs write_ptr_defs removed_one_from_ptr_def
    min_def
    split: exceptions.split_asm split_if_asm)
  done

lemma check_pub_same_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace>
  Abs3_StartupDefs.check_pub_same pind
  \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    Abs3_StartupDefs.check_pub_same_def
    Abs3_CommonDefs.just_sync_silence_deg_def
    Abs3_CommonDefs.read_or_fail_silence_deg_def
    Abs3_CommonDefs.read_or_fail_def
    Abs3_CommonDefs.just_sync_def
    Abs3_CommonDefs.sync_def
    read_DegradedH_or_fail_def
    intro!:
    bind_ruleNN_same try_catch_ruleNN_same
    bind_ruleNE_same try_catch_ruleNE_same
    HoareNN_throw HoareNE_return
    HoareNN_return_same
    HoareNN_get_mode
    HoareNE_vol_ops
    read_doesnt_add_post_NN
    read_doesnt_add_post_NE
    write_instance_doesnt_add_post_NN
    write_instance_doesnt_add_post_NE
    read_DegradedH_doesnt_add_post_NN
    read_DegradedH_doesnt_add_post_NE
    remove_from_ptr_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NE
    vol_ops_dont_add_post_NN
    split: exceptions.split
  )

lemma get_sync_direction_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace> get_sync_direction pind \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(rename_tac sv sp v sv' sp')
  apply(simp_all only: get_sync_direction_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: get_hashes_proj_simps first_stage vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: get_hashes_proj_simps first_stage vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_hashes_all_defs get_ptr_defs brute_force
    removed_one_from_ptr_def min_def
    split: exceptions.split_asm split_if_asm)
  done

lemma get_sync_direction_fails_post_NE:
  "\<lbrace>P\<rbrace> get_sync_direction pind \<lbrace>@N K (sv_mode_equals CFail)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta)
  apply(rename_tac sv sp v sv' sp')
  apply(simp_all only: get_sync_direction_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: get_hashes_proj_simps first_stage vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: get_hashes_proj_simps first_stage vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_hashes_all_defs get_ptr_defs brute_force
    min_def sv_mode_equals_def
    split: exceptions.split_asm split_if_asm)
  done

lemma sync_and_remove_from_ptr_silence_deg_doesnt_add_post_NN:
  "to \<noteq> from \<Longrightarrow>
  \<lbrace>removed_one_from_ptr pind'\<rbrace>
    sync_and_remove_from_ptr_silence_deg to from pind
  \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    sync_and_remove_from_ptr_silence_deg_def
    remove_from_ptr_silence_deg_def
    Abs3_CommonDefs.read_or_fail_def
    Abs3_CommonDefs.sync_def
    intro!:
    bind_ruleNN_same try_catch_ruleNN_same
    bind_ruleNE_same try_catch_ruleNE_same
    HoareNN_throw HoareNE_return
    HoareNN_return_same
    HoareNN_get_mode
    HoareNE_vol_ops
    read_doesnt_add_post_NN
    read_doesnt_add_post_NE
    write_instance_doesnt_add_post_NN
    write_instance_doesnt_add_post_NE
    read_DegradedH_doesnt_add_post_NN
    read_DegradedH_doesnt_add_post_NE
    remove_from_ptr_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NE
    vol_ops_dont_add_post_NN
    split: exceptions.split
  )

lemma check_pub_region_doesnt_add_post_NN:
  "\<lbrace>removed_one_from_ptr pind'\<rbrace>
  Abs3_StartupDefs.check_pub_region pind
  \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    Abs3_StartupDefs.check_pub_region_def
    remove_from_ptr_silence_deg_def
    intro!:
    bind_ruleNN_same try_catch_ruleNN_same
    bind_ruleNE_same try_catch_ruleNE_same
    HoareNN_throw HoareNE_return
    HoareNN_return_same
    HoareNN_get_mode
    HoareNE_vol_ops
    check_pub_same_doesnt_add_post_NN
    get_sync_direction_doesnt_add_post_NN
    sync_and_remove_from_ptr_silence_deg_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NE
    split: exceptions.split
  )

lemma read_mode_post_NE:
  "\<lbrace>P\<rbrace>
  Abs3_CommonDefs.read pind inst
  \<lbrace>@N \<lambda>e.
  if e = FailMode then sv_mode_equals CFail
  else if e \<in> {FailWrite, FailLLRead, StartedExc, InvalidInput} then {}
  else UNIV\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs 
    sv_mode_equals_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: FailMode_exhaust)
  done

lemma read_or_fail_silence_deg_fails_post_NE:
  "\<lbrace>P\<rbrace>
    Abs3_CommonDefs.read_or_fail_silence_deg pind inst
  \<lbrace>@N K (sv_mode_equals CFail)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp only: Abs3_CommonDefs.read_or_fail_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs sv_mode_equals_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: FailMode_exhaust)
  done

lemma check_pub_same_fails_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_same pind \<lbrace>@N K (sv_mode_equals CFail)\<rbrace>"
  apply(auto simp add: Abs3_StartupDefs.check_pub_same_def
    intro!: bind_ruleNN bind_ruleNE try_catch_ruleNN try_catch_ruleNE)
  apply(auto intro!:
    HoareNE_join[OF read_mode_post_NE[where P=UNIV] read_capacity_inv[THEN Hoare_imp_HoareNE],
    simplified, simplified K_beta]
    read_capacity_inv[THEN Hoare_imp_HoareNN]
    HoareNE_return 
    sync_capacity_inv_NN
    HoareNE_join[OF sync_mode_post_NE sync_capacity_inv_NE, simplified]
    read_or_fail_silence_deg_fails_post_NE[simplified K_def]
    try_catch_ruleNE bind_ruleNE
    simp add: HoareNE_throw' HoareNN_return'
    just_sync_silence_deg_def just_sync_def
    K_beta
    intro: true_postNN
    split: exceptions.split
  )
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  done

lemma mode_not_Normal_imp_removed:
  "m \<noteq> CNormal \<Longrightarrow> sv_mode_equals m \<subseteq> removed_one_from_ptr pind'"
  by(auto simp add: sv_mode_equals_def removed_one_from_ptr_def)

lemma check_pub_region_fails_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_region pind \<lbrace>@N K (sv_mode_equals CFail) \<rbrace>"
  apply(auto simp add: Abs3_StartupDefs.check_pub_region_def K_beta
    intro!: bind_ruleNE bind_ruleNN try_catch_ruleNE
    get_sync_direction_capacity_inv_NN
    get_sync_direction_fails_post_NE
    sync_and_remove_from_ptr_silence_deg_failed_post_NE
    check_pub_same_fails_post_NE
  )
  apply(rule check_pub_same_capacity_inv_NN)
  apply(simp add: K_beta)
  apply(rule conseqNE_rule[OF _ remove_from_ptr_fails_post_NE])
  apply(auto simp add: K_beta)
  done

lemma check_pub_region_doesnt_add_inv_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_pub_region pind \<lbrace>@N K (removed_one_from_ptr pind') \<rbrace>"
  by(auto intro!: conseqNE_rule[OF _ check_pub_region_fails_post_NE] mode_not_Normal_imp_removed
    simp add: K_beta del: subsetI)

lemma chooseN_ruleNN:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>>N \<lambda>x. if x \<in> A then P else {}\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def chooseN_def brute_force)

lemma chooseN_ruleNE:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>@N Q\<rbrace>"
  by(auto simp add: HoareNE_alt Hoare_def Box_def chooseN_def choose_def project_defs unfold_monad)

lemma remove_from_ptr_capacity_inv:
  "\<lbrace>abs3_capacity_inv\<rbrace> remove_from_ptr ind \<lbrace>> K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta)
  apply(auto simp add: remove_from_ptr_def get_ptr_defs write_ptr_defs
    Let_def brute_force abs3_capacity_inv_def)
  done

lemma check_pub_region_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> check_pub_region ind \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  by(auto simp add: check_pub_region_def K_beta HoareNN_return' HoareNE_throw'
    sync_and_remove_from_ptr_silence_deg_def remove_from_ptr_silence_deg_def
    intro!: bind_ruleNN_same try_catch_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same
    check_pub_same_capacity_inv_NN
    get_sync_direction_capacity_inv_NN
    sync_capacity_inv_NN
    sync_capacity_inv_NE
    remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNN]
    remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNE]
    HoareNN_throw HoareNE_return
    split: exceptions.split
  )

lemma check_all_public_empty_ptr_post_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_all_public \<lbrace>>N K empty_ptr\<rbrace>"
proof(subst Abs3_StartupDefs.check_all_public_def, rule bind_ruleNN[OF chooseN_ruleNN], intro allI)
  fix l
  show
    "\<lbrace>if l \<in> order_set UNIV then abs3_capacity_inv else {}\<rbrace>
    mapM' Abs3_StartupDefs.check_pub_region l
    \<lbrace>>N K empty_ptr\<rbrace>"
  proof(cases "l \<in> order_set UNIV")
    case True
    thus ?thesis
    proof(simp)
      have characterization: "empty_ptr = (\<Inter>x' \<in> set l. removed_one_from_ptr x')"
        using order_set_set[OF True]
        by(auto simp add: empty_ptr_def removed_one_from_ptr_def)
      show "\<lbrace>abs3_capacity_inv\<rbrace> mapM' Abs3_StartupDefs.check_pub_region l \<lbrace>>N K empty_ptr\<rbrace>"
        by(auto simp add: characterization
          intro!: mapM'_ruleNN
          check_pub_region_removed_post_NN
          check_pub_region_capacity_inv_NN
          check_pub_region_doesnt_add_post_NN)
    qed
  qed(auto simp add: ex_falsumNN)
qed

lemma check_all_public_fails_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_all_public \<lbrace>@N K (sv_mode_equals CFail) \<rbrace>"
  by(auto simp add: Abs3_StartupDefs.check_all_public_def K_beta
    intro!: bind_ruleNE mapM'_ruleNE check_pub_region_fails_post_NE chooseN_ruleNE
    chooseN_inv_NN
    check_pub_region_capacity_inv_NN)

lemma mode_not_Normal_imp_empty_ptr:
  "m \<noteq> CNormal \<Longrightarrow> sv_mode_equals m \<subseteq> empty_ptr"
  by(auto simp add: sv_mode_equals_def empty_ptr_def)

lemma check_all_public_empty_ptr_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.check_all_public \<lbrace>@N K empty_ptr \<rbrace>"
  by(auto intro!: conseqNE_rule[OF _ check_all_public_fails_post_NE] mode_not_Normal_imp_empty_ptr
    del: subsetI simp add: K_beta)

lemma throw_mode_error_mode_post_NE:
  "\<lbrace>empty_ptr\<rbrace> (throw_mode_error EU) :: (unit, 'c)abs3_monad \<lbrace>@N K empty_ptr\<rbrace>"
  by(auto simp add: throw_mode_error_def HoareNE_throw' K_beta
    intro!: bind_ruleNE
    HoareNN_get_mode
    HoareNE_return
    HoareNE_vol_ops
    elim!: mode_not_Normal_imp_empty_ptr[THEN subsetD, rotated 1]
  )

lemma throw_mode_error_mode_post_NN:
  "\<lbrace>empty_ptr\<rbrace> (throw_mode_error EU) :: (unit, 'c)abs3_monad \<lbrace>>N K empty_ptr\<rbrace>"
  by(auto simp add: throw_mode_error_def HoareNN_return' K_beta
    intro!: bind_ruleNN
    HoareNN_get_mode
    HoareNN_throw
  )

lemma startup1_mode_post_NE:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup1 \<lbrace>@N K (sv_mode_equals CFail)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def Abs3_StartupDefs.startup1_def compose_project)
  apply(auto simp add: Abs3_StartupDefs.startup1_NE_def brute_force sv_mode_equals_def)
  done

lemma startup1_empty_ptr_post_NE:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup1 \<lbrace>@N K empty_ptr\<rbrace>"
  by(auto simp add: K_def intro!: conseqNE_rule[OF _ startup1_mode_post_NE]
    elim!: mode_not_Normal_imp_empty_ptr[THEN subsetD, rotated 1]
  )

lemma startup_true_post_NN:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup \<lbrace>>N Q\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def 
    Abs3_StartupDefs.startup_def Abs3_StartupDefs.startup2_def first_stage)

lemma startup_empty_ptr_post_NN:
  "\<lbrace>P\<rbrace> Abs3_StartupDefs.startup \<lbrace>>N K empty_ptr\<rbrace>"
  by(auto intro: conseqNN_rule[OF _ startup_true_post_NN])

lemma startup1_capacity_inv:
  "\<lbrace>abs3_capacity_inv\<rbrace> startup1 \<lbrace>> K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta)
  apply(auto simp add: abs3_capacity_inv_def brute_force startup1_def compose_def
    startup1_NN_def startup1_NE_def)
  done

lemma startup_empty_ptr_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_StartupDefs.startup \<lbrace>@N K empty_ptr\<rbrace>"
  by(auto simp add: Abs3_StartupDefs.startup_def Abs3_StartupDefs.startup2_def K_beta
    HoareNE_throw'
    intro!: bind_ruleNE
    startup1_capacity_inv[THEN Hoare_imp_HoareNN]
    startup1_empty_ptr_post_NE
    check_all_public_empty_ptr_post_NE
    check_all_public_empty_ptr_post_NN
    throw_mode_error_mode_post_NN
    throw_mode_error_mode_post_NE
    HoareNN_throw
  )

lemma write_secondary_precommit_inv_NN:
  "\<lbrace>precommit_inv\<rbrace> Abs3_UpdateDefs.write_secondary upd_f ind \<lbrace>>N K precommit_inv\<rbrace>"
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(rename_tac sv sp sv' sp')
  apply(simp_all only: write_secondary_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs
    precommit_inv_defs
    brute_force abs3_inv_defs split: split_if_asm)
  done

lemma just_sync_silence_deg_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_CommonDefs.just_sync_silence_deg to from ind \<lbrace>>N K abs3_norm_inv \<rbrace>"
  by(auto simp add: K_beta HoareNN_return'
    Abs3_CommonDefs.just_sync_silence_deg_def
    Abs3_CommonDefs.just_sync_def
    intro!:
    try_catch_ruleNN bind_ruleNN bind_ruleNE HoareNN_throw HoareNE_return
    sync_inv_NN sync_inv_NE
  )

lemma just_sync_silence_deg_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_CommonDefs.just_sync_silence_deg to from ind \<lbrace>@N
  \<lambda>e. if e = FailMode then abs3_inv else abs3_norm_inv \<rbrace>"
  by(auto simp add: K_beta HoareNN_return' HoareNE_throw'
    Abs3_CommonDefs.just_sync_silence_deg_def
    Abs3_CommonDefs.just_sync_def
    intro!:
    try_catch_ruleNE bind_ruleNN bind_ruleNE HoareNN_throw HoareNE_return
    sync_inv_NN sync_inv_NE)

lemma write_secondary_rollback_post_NE:
  "\<lbrace>precommit_inv\<rbrace>
    write_secondary upd_f ind
  \<lbrace>@N \<lambda>e. if \<exists>v. e = Degraded v then abs3_norm_inv else UNIV\<rbrace>" 
  apply(simp_all only: write_secondary_def)
  apply(rule bind_ruleNE)
  apply(rule conseqNE_rule[OF _ write_instance_inv_NE])
  apply(auto intro: true_postNN HoareNE_vol_ops simp add: precommit_inv_def)
  done

lemma set_ptr_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> set_ptr pinds \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(simp only: set_ptr_def)
  apply(rule bind_ruleNN_same)
  apply(auto simp add: K_beta HoareNN_alt Hoare_def Box_def)
  apply(simp_all only: get_ptr_def write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: precommit_inv_defs get_ptr_defs write_ptr_defs brute_force)
  done

lemma set_ptr_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> set_ptr pinds
  \<lbrace>@N \<lambda>e. if \<exists>v. e = Degraded v then abs3_norm_inv else abs3_inv\<rbrace>"
  apply(simp only: set_ptr_def)
  apply(rule bind_ruleNE_same)
  apply(auto simp add: K_beta HoareNN_alt HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: get_ptr_def write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs3_inv_defs get_ptr_defs write_ptr_defs brute_force min_def)
  apply(auto)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms card_mode_exhaust} *}, simp_all)
  by (metis card_mode_exhaust less_eq_card_mode.simps(10))

lemma sync_and_remove_from_ptr_silence_deg_inv_NN:
 "\<lbrace>abs3_norm_inv\<rbrace> sync_and_remove_from_ptr_silence_deg Primary Secondary ind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: sync_and_remove_from_ptr_silence_deg_def remove_from_ptr_silence_deg_def
    K_beta HoareNN_return' HoareNE_throw'
    intro!: try_catch_ruleNN try_catch_ruleNE bind_ruleNN bind_ruleNE HoareNN_throw HoareNE_return
    sync_inv_NN
    sync_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NE
    split: exceptions.split
  )
  done
 
lemma mapM'_sync_and_remove_from_ptr_silence_deg_removed_post_NN':
  assumes distinct: "distinct l"
  shows
    "\<lbrace>abs3_capacity_inv\<rbrace>
    mapM' (sync_and_remove_from_ptr_silence_deg Primary Secondary) l
    \<lbrace>>N K (\<Inter>x\<in>set l. removed_one_from_ptr x)\<rbrace>"
proof(cases l)
  case Nil
  thus ?thesis
    by(clarsimp intro!: true_postNN)
next
  case (Cons x xs)
  hence complicated: "abs3_capacity_inv=(\<Inter>x \<in> set l. abs3_capacity_inv)"
    by(simp)
  show ?thesis using distinct
    apply(subst complicated)
    apply(erule mapM'_distinct_ruleNN)
    apply(auto intro!: sync_and_remove_from_ptr_silence_deg_removed_post_NN
      sync_and_remove_from_ptr_silence_deg_doesnt_add_post_NN
      intro: true_postNN)
    apply(auto simp add: sync_and_remove_from_ptr_silence_deg_def remove_from_ptr_silence_deg_def
      K_beta
      HoareNN_return' HoareNE_throw'
      intro!: bind_ruleNN_same try_catch_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same
      sync_capacity_inv_NN
      sync_capacity_inv_NE
      remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNN]
      remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNE]
      HoareNN_throw HoareNE_return
      split: exceptions.split
    )
    done
qed

lemma sync_if_not_SEq_remove_from_ptr_doesnt_add_post_NN:
  "to \<noteq> from \<Longrightarrow>
  \<lbrace>removed_one_from_ptr pind'\<rbrace>
    sync_if_not_SEq_remove_from_ptr to from pind
  \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def
    remove_from_ptr_silence_deg_def
    Abs3_CommonDefs.read_or_fail_def
    Abs3_CommonDefs.sync_def
    intro!:
    bind_ruleNN_same try_catch_ruleNN_same
    bind_ruleNE_same try_catch_ruleNE_same
    HoareNN_throw HoareNE_return
    HoareNN_return_same
    HoareNN_get_mode
    HoareNE_vol_ops
    sync_and_remove_from_ptr_silence_deg_doesnt_add_post_NN
    get_sync_direction_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NN
    remove_from_ptr_doesnt_add_post_NE
    vol_ops_dont_add_post_NN
    split: exceptions.split
  )

thm sync_and_remove_from_ptr_silence_deg_removed_post_NN

lemma sync_if_not_SEq_remove_from_ptr_removed_post_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync_if_not_SEq_remove_from_ptr to from pind  
    \<lbrace>>N K (removed_one_from_ptr pind) \<rbrace>"
  by(auto simp add: check_arrow_SEq_def
    sync_if_not_SEq_remove_from_ptr_def K_beta HoareNN_return'
    intro!:
    bind_ruleNN
    sync_and_remove_from_ptr_silence_deg_removed_post_NN
    remove_from_ptr_removed_inv_NN
    get_sync_direction_capacity_inv_NN
  )

lemma mapM'_sync_if_not_SEq_remove_from_ptr_removed_post_NN':
  assumes distinct: "distinct l"
  shows
    "\<lbrace>abs3_capacity_inv\<rbrace>
    mapM' (sync_if_not_SEq_remove_from_ptr Primary Secondary) l
    \<lbrace>>N K (\<Inter>x\<in>set l. removed_one_from_ptr x)\<rbrace>"
proof(cases l)
  case Nil
  thus ?thesis
    by(clarsimp intro!: true_postNN)
next
  case (Cons x xs)
  hence complicated: "abs3_capacity_inv=(\<Inter>x \<in> set l. abs3_capacity_inv)"
    by(simp)
  show ?thesis using distinct
    apply(subst complicated)
    apply(erule mapM'_distinct_ruleNN)
    apply(auto intro!: 
      sync_if_not_SEq_remove_from_ptr_removed_post_NN
      sync_if_not_SEq_remove_from_ptr_doesnt_add_post_NN
      intro: true_postNN)
    apply(auto simp add: sync_if_not_SEq_remove_from_ptr_def
      check_arrow_SEq_def remove_from_ptr_silence_deg_def
      sync_and_remove_from_ptr_silence_deg_def
      HoareNN_return' HoareNE_throw' K_beta
      intro!: 
      bind_ruleNN_same try_catch_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same
      HoareNN_throw HoareNE_return
      get_sync_direction_capacity_inv_NN
      remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNN]
      remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNE]
      sync_capacity_inv_NN
      sync_capacity_inv_NE
      split: exceptions.split
    )
    done
qed


lemma mapM'_sync_if_not_SEq_remove_from_ptr_removed_post_NN:
  "distinct l \<Longrightarrow>
    \<lbrace>abs3_capacity_inv\<rbrace>
    mapM' (sync_if_not_SEq_remove_from_ptr Primary Secondary) l
  \<lbrace>>N K (\<Inter>x\<in>set l. removed_one_from_ptr x)\<rbrace>"
  by(auto intro!: conseqNN_rule [OF _ mapM'_sync_if_not_SEq_remove_from_ptr_removed_post_NN'])

definition ptr_eq where
  "ptr_eq pinds \<equiv> {(sv, sp). ptr sp = pinds}"

definition ptr_eq_updated where
  "ptr_eq_updated \<equiv> {(sv, sp). updated sv = ptr sp}"

lemma abs3_rollback_post_NN:
  "\<lbrace>P\<rbrace> Abs3_UpdateDefs.rollback \<lbrace>>N Q\<rbrace>"
  by(auto simp add: Abs3_UpdateDefs.rollback_def
    intro!: bind_ruleNN HoareNN_throw
    intro: true_postNN
  )

lemma set_ptr_post_NE:
  "\<lbrace>updated_equals val\<rbrace> set_ptr val \<lbrace>@N \<lambda>e. if e = Degraded (EB True) then ptr_eq_updated else UNIV\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def ptr_eq_updated_def updated_equals_def)
  apply(simp_all only: set_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_ptr_defs write_ptr_defs brute_force split: split_if_asm)
  apply(clarsimp_all)
  done

lemma do_commit_post_NN:
  "\<lbrace>P\<rbrace> Abs3_UpdateDefs.do_commit \<lbrace>>N K ptr_eq_updated\<rbrace>"
  apply(auto simp add: Abs3_UpdateDefs.do_commit_def K_beta HoareNN_return'
    intro!: bind_ruleNN try_catch_ruleNN HoareNN_throw
    abs3_rollback_post_NN
    get_updated_post_NN
    set_ptr_post_NE
    split: exceptions.split
  )
  apply(auto simp add: HoareNN_alt Hoare_def Box_def ptr_eq_updated_def updated_equals_def K_beta)
  apply(simp_all add: set_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_ptr_defs write_ptr_defs brute_force split: split_if_asm)
  done

lemma Collect_UNIV:
  "(Collect P = UNIV) = (All P)"
  by(auto)

definition ptr_subset where
  "ptr_subset pinds \<equiv> {(sv, sp). card_mode sv = CNormal \<longrightarrow> ptr sp \<subseteq> pinds}"

lemma ptr_subset_inv_NN_I:
  "\<lbrakk>
    \<And>pind'. \<lbrace>removed_one_from_ptr pind'\<rbrace> m \<lbrace>>N K (removed_one_from_ptr pind')\<rbrace>;
    \<lbrace>sv_mode_neq CNormal\<rbrace> m \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>
  \<rbrakk> \<Longrightarrow>
    \<lbrace>ptr_subset pinds\<rbrace> m \<lbrace>>N K (ptr_subset pinds)\<rbrace>" 
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(rename_tac sv sp v sv' sp')
  apply(auto simp add: removed_one_from_ptr_def ptr_subset_def brute_force sv_mode_neq_def)
  apply(simp add: Collect_subset)
  apply(metis)
  apply(simp add: Collect_subset)
  apply(subgoal_tac "card_mode sv=CNormal")
  apply(metis (hide_lams, no_types) set_mp)+
  done

lemma mapM'_rule_NN_same:
  "(\<And>x. \<lbrace>P\<rbrace> m x \<lbrace>>N K P\<rbrace>) \<Longrightarrow> \<lbrace>P\<rbrace> mapM' m l \<lbrace>>N K P\<rbrace>"
  by(induct l, auto simp add: HoareNN_return' K_beta intro!: bind_ruleNN)

lemma sync_and_remove_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace>
  sync_and_remove_from_ptr_silence_deg Primary Secondary ind
  \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: sync_and_remove_from_ptr_silence_deg_def HoareNN_return' HoareNE_throw'
    remove_from_ptr_silence_deg_def sync_def K_beta
    remove_from_ptr_def read_or_fail_def
    intro!: bind_ruleNN_same try_catch_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same
    HoareNE_return HoareNN_throw
    HoareNE_vol_ops
    split: exceptions.split
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp add: HoareNN_return' HoareNE_throw' K_beta
    read_def write_instance_def
    update_instance_hash_def read_DegradedH_def
    intro!: bind_ruleNN_same try_catch_ruleNN_same bind_ruleNE_same try_catch_ruleNE_same
    HoareNE_return HoareNN_throw
    HoareNE_vol_ops
    HoareNN_get_mode
    split: exceptions.split
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def K_def sv_mode_neq_def)
  apply(auto simp add: get_hashes_all_defs write_hashes_defs brute_force min_def
    get_ptr_defs write_ptr_defs
    split: split_if_asm)
  done

lemma sync_if_not_SEq_remove_from_ptr_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace>
  sync_if_not_SEq_remove_from_ptr Primary Secondary ind
  \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNN_return' K_beta
    sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def
    intro!:
    bind_ruleNN_same
    sync_and_remove_no_mode_upgrade_NN
  )
  apply(auto simp add: get_sync_direction_def remove_from_ptr_silence_deg_def remove_from_ptr_def)
  apply(auto simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def K_def sv_mode_neq_def)
  apply(simp_all add: first_stage vol_ops_project_simps
    get_hashes_silence_deg_def
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_hashes_all_defs write_hashes_defs brute_force min_def
    get_ptr_defs write_ptr_defs
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(metis exceptions.exhaust)+
  done

lemma mapM'_ptr_subset:
  "\<lbrace>ptr_subset pinds\<rbrace>
    mapM' (sync_if_not_SEq_remove_from_ptr Primary Secondary) l
  \<lbrace>>N K (ptr_subset pinds)\<rbrace>
  "
  by(auto intro!: mapM'_rule_NN_same ptr_subset_inv_NN_I
    sync_if_not_SEq_remove_from_ptr_doesnt_add_post_NN
    sync_if_not_SEq_remove_from_ptr_no_mode_upgrade_NN
  )

lemma sync_if_not_SEq_remove_from_ptr_ptr_post_NN:
  assumes l: "l \<in> order_set pinds"
  shows
  "\<lbrace>ptr_eq pinds \<inter> abs3_capacity_inv\<rbrace>
    mapM' (sync_if_not_SEq_remove_from_ptr Primary Secondary) l
  \<lbrace>>N K empty_ptr\<rbrace>
  "
proof-
  have sub: "ptr_eq pinds \<subseteq> ptr_subset pinds"
    by(auto simp add: ptr_eq_def ptr_subset_def)
  have comb:
    "\<lbrace>(ptr_subset pinds) \<inter> abs3_capacity_inv\<rbrace>
    mapM' (sync_if_not_SEq_remove_from_ptr Primary Secondary) l
    \<lbrace>>N K (ptr_subset pinds \<inter> (\<Inter>x\<in>set l. removed_one_from_ptr x))\<rbrace>"
    apply(auto simp only: intro!: HoareNN_Int mapM'_ptr_subset
      mapM'_sync_if_not_SEq_remove_from_ptr_removed_post_NN'
      order_set_distinct[OF l]
    )
    apply(auto)
    done
  show ?thesis
    apply(rule conseqNN_rule[OF _ comb])
    apply(clarsimp intro!: sub[THEN subsetD])
    apply(auto simp add: K_def ptr_subset_def removed_one_from_ptr_def empty_ptr_def
      order_set_set[OF l])[1]
    done
qed

lemma rollforward_post_NN:
  "\<lbrace>ptr_eq_updated \<inter> abs3_capacity_inv\<rbrace> Abs3_UpdateDefs.rollforward \<lbrace>>N K empty_ptr\<rbrace>" 
  apply(auto simp add: Abs3_UpdateDefs.rollforward_def)
  apply(rule bind_ruleNN[where Q="\<lambda>x. ptr_eq x \<inter> abs3_capacity_inv"])
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def ptr_eq_updated_def ptr_eq_def brute_force)
  apply(clarsimp)
  apply(rule bind_ruleNN[OF chooseN_ruleNN])
  apply(auto intro: ex_falsumNN)
  apply(rule bind_ruleNN[where Q="K empty_ptr"])
  apply(rule sync_if_not_SEq_remove_from_ptr_ptr_post_NN)
  apply(auto simp add: HoareNN_alt Hoare_def Box_def brute_force empty_ptr_def)
  done

lemma get_mode_ruleNN:
  "\<lbrace>P\<rbrace> get_mode \<lbrace>>N \<lambda>v. sv_mode_equals v \<inter> P\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def brute_force sv_mode_equals_def)

lemma set_degraded_not_normal_NN:
  "\<lbrace>P\<rbrace> set_degraded \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def sv_mode_neq_def brute_force min_def
    split: split_if_asm)

lemma read_no_mode_upgrade_post_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> read ind inst \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all add: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs split: split_if_asm)
  done

lemma read_no_mode_upgrade_post_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> read ind inst \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs min_def
    split: split_if_asm)
  apply(auto)
  done

lemma read_DegradedH_no_mode_upgrade_post_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> read_DegradedH ind inst \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all add: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs split: split_if_asm)
  done

lemma read_DegradedH_no_mode_upgrade_post_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> read_DegradedH ind inst \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs min_def
    split: split_if_asm)
  done

lemma update_instance_hash_no_mode_upgrade_post_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> update_instance_hash ind inst cnt \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all add: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs min_def
    split: split_if_asm)
  done

lemma update_instance_hash_no_mode_upgrade_post_NE:
  "\<lbrace>P\<rbrace> update_instance_hash ind inst cnt \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all add: update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs min_def
    split: split_if_asm)
  done

lemma write_instance_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> write_instance ind inst cnt \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: K_beta HoareNE_throw' HoareNN_return'
    write_instance_def
    intro!:
    bind_ruleNN_same
    try_catch_ruleNN_same
    mapM'_ruleNE
    HoareNE_return
    HoareNN_throw
    chooseN_ruleNE
    HoareNE_vol_ops
    set_degraded_not_normal_NN
    sv_mode_eq_neq_subset
    read_DegradedH_no_mode_upgrade_post_NN
    read_DegradedH_no_mode_upgrade_post_NE
    update_instance_hash_no_mode_upgrade_post_NN
    update_instance_hash_no_mode_upgrade_post_NE
    read_no_mode_upgrade_post_NN
    read_no_mode_upgrade_post_NE
    HoareNN_get_mode
    split: exceptions.split
    del: subsetI
  )
  apply(auto simp add: HoareNN_alt HoareNE_alt Hoare_def Box_def sv_mode_neq_def K_def)
  apply(simp_all only: write_f_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force split: split_if_asm)
  done
  
lemma write_instance_mode_post_NE:
  "size cnt \<le> capacity (Public ind)
    \<Longrightarrow> \<lbrace>P\<rbrace> write_instance ind inst cnt \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: K_beta HoareNE_throw' HoareNN_return'
    write_instance_def
    intro!:
    bind_ruleNE 
    mapM'_ruleNE
    HoareNE_return
    HoareNN_throw
    chooseN_ruleNE
    HoareNE_vol_ops
    get_mode_ruleNN
    set_degraded_not_normal_NN
    sv_mode_eq_neq_subset
    split: exceptions.split
    intro: true_postNN
    del: subsetI
  )
  apply(clarsimp simp add: sv_mode_equals_def sv_mode_neq_def)
  apply(rule try_catch_ruleNE[where Q="\<lambda>e. if e \<noteq> FailWrite then {} else UNIV"])
  apply(clarsimp simp add: Hoare_alt Hoare_def Box_def brute_force)
  apply(auto simp add: K_beta HoareNE_throw' split: exceptions.split)
  apply(rule bind_ruleNE, rule HoareNE_vol_ops)
  apply(rule set_degraded_not_normal_NN)
  apply(clarsimp simp add: K_beta)
  apply(rule bind_ruleNE[where Q="K (sv_mode_neq CNormal)" and R="K (sv_mode_neq CNormal)"])
  apply(rule try_catch_ruleNE[OF read_no_mode_upgrade_post_NE])
  apply(auto simp add: HoareNE_throw' K_beta split: exceptions.split)[1]
  apply(rule try_catch_ruleNN[OF read_no_mode_upgrade_post_NN read_no_mode_upgrade_post_NE])
  apply(auto simp add: K_beta intro!: HoareNN_throw split: exceptions.split)[1]
  apply(clarsimp simp add: K_beta HoareNE_throw')
  apply(rule try_catch_ruleNE[OF update_instance_hash_no_mode_upgrade_post_NE])
  apply(auto simp add: HoareNE_throw' K_beta split: exceptions.split)[1]
  apply(rule bind_ruleNE)
  apply(rule try_catch_ruleNE[OF read_DegradedH_no_mode_upgrade_post_NE])
  apply(auto simp add: HoareNE_throw' K_beta split: exceptions.split)[1]
  apply(rule try_catch_ruleNN[OF read_DegradedH_no_mode_upgrade_post_NN])
  apply(rule read_DegradedH_no_mode_upgrade_post_NE)
  apply(auto simp add: HoareNE_throw' K_beta intro: HoareNN_throw split: exceptions.split)
  done

lemma get_updated_post_NN:
  "\<lbrace>P\<rbrace> get_updated \<lbrace>>N K P\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def brute_force)

lemma set_failed_post_NN:
  "\<lbrace>P\<rbrace> set_failed \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def brute_force sv_mode_neq_def)

lemma chooseN_post_NN:
  "\<lbrace>P\<rbrace> chooseN A \<lbrace>>N K P\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def chooseN_def brute_force)

lemma write_instance_no_mode_upgrade_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> write_instance xa Secondary x \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: write_instance_def K_beta HoareNE_throw' HoareNN_return'
    intro!: bind_ruleNE_same bind_ruleNN_same try_catch_ruleNN_same try_catch_ruleNE_same
    HoareNN_get_mode
    set_failed_post_NN HoareNE_vol_ops
    set_degraded_not_normal_NN
    HoareNN_throw HoareNE_return
    read_no_mode_upgrade_post_NN
    read_no_mode_upgrade_post_NE
    read_DegradedH_no_mode_upgrade_post_NN
    read_DegradedH_no_mode_upgrade_post_NE
    update_instance_hash_no_mode_upgrade_post_NE
    split: exceptions.split
  )
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta)
  apply(auto simp add: brute_force sv_mode_neq_def)
  done

lemma sync_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> sync xa Secondary Primary \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: sync_def read_or_fail_def K_beta HoareNE_throw' HoareNN_return'
    intro!: bind_ruleNE_same bind_ruleNN_same try_catch_ruleNN_same try_catch_ruleNE_same
    read_no_mode_upgrade_post_NN
    read_no_mode_upgrade_post_NE
    write_instance_no_mode_upgrade_NN
    write_instance_no_mode_upgrade_NE
    set_failed_post_NN HoareNE_vol_ops
    HoareNN_throw
    split: exceptions.split
  )
  done

lemma sync_no_mode_upgrade_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> sync xa Secondary Primary \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: sync_def read_or_fail_def K_beta HoareNE_throw' HoareNN_return'
    intro!: bind_ruleNE_same bind_ruleNN_same try_catch_ruleNN_same try_catch_ruleNE_same
    read_no_mode_upgrade_post_NN
    read_no_mode_upgrade_post_NE
    write_instance_no_mode_upgrade_NN
    write_instance_no_mode_upgrade_NE
    set_failed_post_NN HoareNE_vol_ops
    HoareNN_throw HoareNE_return
    split: exceptions.split
  )
  done

lemma just_sync_silence_deg_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace>
  just_sync_silence_deg Secondary Primary xa
  \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  by(auto simp add: HoareNN_return' K_beta
    just_sync_silence_deg_def just_sync_def
    intro!:
    sync_no_mode_upgrade_NN
    sync_no_mode_upgrade_NE
    bind_ruleNE
    bind_ruleNN
    try_catch_ruleNN
    HoareNN_throw
    HoareNE_return
  )
 
lemma get_sync_direction_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> get_sync_direction xa \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(simp_all only: get_sync_direction_def get_ptr_def get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_hashes_all_defs get_ptr_defs brute_force sv_mode_neq_def
    min_def split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  done

lemma check_arrow_SEq_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> check_arrow_SEq xa \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  apply(simp only: check_arrow_SEq_def)
  apply(rule bind_ruleNN_same)
  apply(rule get_sync_direction_no_mode_upgrade_NN)
  apply(clarsimp simp add: HoareNN_return' K_beta)
  done

lemma sync_if_not_SEq_no_mode_upgrade_NN:
  "\<lbrace>sv_mode_neq CNormal\<rbrace>
  sync_if_not_SEq Secondary Primary xa
  \<lbrace>>N K (sv_mode_neq CNormal)\<rbrace>"
  by(auto simp add: HoareNN_return' K_beta
    sync_if_not_SEq_def
    intro!:
    bind_ruleNE
    bind_ruleNN
    try_catch_ruleNN
    HoareNN_throw
    HoareNE_return
    just_sync_mode_post_NE
    just_sync_silence_deg_no_mode_upgrade_NN
    check_arrow_SEq_no_mode_upgrade_NN
  )

lemma get_sync_direction_no_mode_upgrade_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> get_sync_direction ind \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta sv_mode_neq_def)
  apply(auto simp add: get_sync_direction_def get_hashes_all_defs get_hashes_silence_deg_def
    get_ptr_defs
    brute_force sv_mode_neq_def min_def split: split_if_asm)
  done

lemma just_sync_silence_deg_no_mode_upgrade_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace>
  just_sync_silence_deg Secondary Primary xa
  \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  by(auto simp add: HoareNE_throw' HoareNN_return' K_beta
    just_sync_silence_deg_def just_sync_def
    intro!:
    sync_no_mode_upgrade_NN
    sync_no_mode_upgrade_NE
    bind_ruleNE
    bind_ruleNN
    try_catch_ruleNN
    try_catch_ruleNE
    HoareNN_throw
    HoareNE_return
  )
 
lemma rollback_no_mode_upgrade_NE:
  "\<lbrace>sv_mode_neq CNormal\<rbrace> rollback \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: rollback_def sync_if_not_SEq_def K_beta HoareNN_return' HoareNE_throw'
    check_arrow_SEq_def
    intro!:
    bind_ruleNN_same
    bind_ruleNE_same
    HoareNE_vol_ops
    get_updated_post_NN
    chooseN_ruleNE
    chooseN_inv_NN
    mapM'_ruleNE
    mapM'_rule_NN_same
    HoareNE_return HoareNN_throw
    get_sync_direction_no_mode_upgrade_NN
    get_sync_direction_no_mode_upgrade_NE
    just_sync_silence_deg_no_mode_upgrade_NN
    just_sync_silence_deg_no_mode_upgrade_NE
  )
  apply(auto simp add: Hoare_alt Hoare_def Box_def sv_mode_neq_def brute_force)
  done

lemma precommit_mode_post_NE:
  "\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind) \<Longrightarrow>
  \<lbrace>UNIV\<rbrace> pre_commit upd_f \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: K_beta
    pre_commit_def
    intro!: bind_ruleNE[where R="K (sv_mode_neq CNormal)"]
    chooseN_ruleNE
    HoareNE_vol_ops
  )
  apply(rule true_postNN)
  apply(auto simp add: K_beta intro!: chooseN_ruleNN ex_falsumNE)
  apply(rule try_catch_ruleNE[where Q="K (sv_mode_neq CNormal)"])
  apply(rule mapM'_ruleNE)
  apply(simp add: write_secondary_def)
  apply(rule ballI, rule bind_ruleNE)
  apply(rule write_instance_mode_post_NE)
  apply(simp add: order_set_set)
  apply(auto intro!: HoareNE_vol_ops true_postNN 
    rollback_no_mode_upgrade_NE
    simp add: HoareNE_throw' K_beta
    split: exceptions.split)
  done

lemma set_ptr_mode_post_NE:
  "\<lbrace>P\<rbrace> set_ptr pinds \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def sv_mode_neq_def K_beta)
  apply(auto simp add: set_ptr_def brute_force get_ptr_defs write_ptr_defs min_def
    split: exceptions.split_asm split_if_asm )
  done

lemma do_commit_mode_post_NE:
  "\<lbrace>P\<rbrace> do_commit \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: K_beta
    do_commit_def
    intro!:
    bind_ruleNE
    HoareNE_vol_ops
    true_postNN
  )
  apply(auto simp add: K_beta HoareNE_throw'
    intro!: 
    try_catch_ruleNE
    set_ptr_mode_post_NE
    HoareNE_return
    rollback_no_mode_upgrade_NE
    split: exceptions.split
  )
  done

lemma sync_if_not_SEq_remove_from_ptr_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> sync_if_not_SEq_remove_from_ptr Primary Secondary ind
  \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def K_beta
    HoareNN_return'
    elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1]
    intro!:
    conseqNE_rule[OF _ get_sync_direction_fails_post_NE]
    bind_ruleNE_same
    bind_ruleNN_same
    HoareNE_return
    get_sync_direction_capacity_inv_NN
  )
  apply(auto simp add: sync_and_remove_from_ptr_silence_deg_def K_beta
    HoareNE_throw'
    intro!: try_catch_ruleNE_same bind_ruleNE_same
    conseqNE_rule[OF _ sync_mode_post_NE]
    sync_capacity_inv_NN
    conseqNE_rule[OF _ remove_from_ptr_fails_post_NE[where P=UNIV]]
    HoareNE_return
    elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1]
    split: split_if_asm exceptions.split
  )
  apply(auto simp add: HoareNN_return' HoareNE_throw' K_beta
    remove_from_ptr_silence_deg_def
    intro!: try_catch_ruleNE_same bind_ruleNE_same try_catch_ruleNN_same bind_ruleNN_same
    HoareNN_throw HoareNE_return
    remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNN]
    remove_from_ptr_capacity_inv[THEN Hoare_imp_HoareNE]
    sync_capacity_inv_NN
    sync_capacity_inv_NE
    split: exceptions.split
  )
  done

lemma rollforward_mode_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> rollforward \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: K_beta
    rollforward_def
    intro!:
    bind_ruleNE_same
    chooseN_ruleNE
    chooseN_inv_NN
    get_updated_post_NN
    HoareNE_vol_ops
    mapM'_ruleNE
    mapM'_invNN
    sync_if_not_SEq_remove_from_ptr_capacity_inv_NN
  )
  apply(auto simp add: sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def K_beta
    HoareNN_return'
    elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1]
    intro!:
    conseqNE_rule[OF _ get_sync_direction_fails_post_NE]
    bind_ruleNE_same
    bind_ruleNN_same
    HoareNE_return
    get_sync_direction_capacity_inv_NN
  )
  apply(rule conseqNE_rule[OF _ remove_from_ptr_fails_post_NE[where P=UNIV]])
  apply(auto simp add: K_beta elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1])
  apply(auto simp add: sync_and_remove_from_ptr_silence_deg_def K_beta
    HoareNE_throw'
    intro!: try_catch_ruleNE_same bind_ruleNE_same
    conseqNE_rule[OF _ sync_mode_post_NE]
    sync_capacity_inv_NN
    conseqNE_rule[OF _ remove_from_ptr_fails_post_NE[where P=UNIV]]
    HoareNE_return
    elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1]
    split: split_if_asm exceptions.split
  )
  done

lemma check_sizes_post_NN:
  "\<lbrace>P\<rbrace> check_sizes upd_f 
    \<lbrace>>N K (if (\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind))
        then P
        else {}
    )\<rbrace>"
proof(auto simp add: check_sizes_def HoareNE_throw' HoareNN_return' K_beta
    intro!: bind_ruleNN
    HoareNE_return
    HoareNN_throw
    mapM'_invNN
    chooseN_ruleNN ex_falsumNN)
  fix cnt ind l
  assume 
    sz: "\<not> size cnt \<le> capacity (Public ind)" 
    and in_dom: "upd_f ind = Some cnt" 
    and l_def: "l \<in> order_set (dom upd_f)"
  thm mapM'_ruleNN
  have rewr:
    "K {} = K (\<Inter>ind' \<in> set l. if size (the (upd_f ind')) \<le> capacity (Public ind') then UNIV else {})"
    apply(auto simp add: K_def intro!: ext)
    apply(erule_tac x=ind in ballE)
    apply(auto simp add: sz in_dom order_set_set[OF l_def] split: split_if_asm)
    done
  show 
    "\<lbrace>P\<rbrace> mapM' 
        (\<lambda>pind. if capacity (Public pind) < size (the (upd_f pind)) 
            then throw InvalidInput else return ()) 
        l
    \<lbrace>>N K {}\<rbrace>"
    apply(subst rewr)
    apply(rule mapM'_ruleNN)
    apply(auto simp add: HoareNN_return' HoareNE_throw' K_beta
      intro!: HoareNN_throw)
    done
qed

lemma check_sizes_post_NE:
  "\<lbrace>P\<rbrace> check_sizes upd_f \<lbrace>@N \<lambda>e. if e = InvalidInput then P else {}\<rbrace>"
  by(auto simp add: check_sizes_def HoareNE_throw' HoareNN_return' K_beta
    intro!: bind_ruleNE chooseN_ruleNE 
    HoareNE_return
    HoareNN_throw
    mapM'_ruleNE
    chooseN_ruleNN ex_falsumNE)


lemma throw_mode_error_post_NE:
  "\<lbrace>P\<rbrace> throw_mode_error v \<lbrace>@N K (sv_mode_neq CNormal) \<rbrace>"
  by(auto simp add: Hoare_alt Hoare_def Box_def throw_mode_error_def brute_force
    sv_mode_neq_def)

lemma throw_mode_error_inv:
  "\<lbrace>P\<rbrace> throw_mode_error v \<lbrace>>N K P \<rbrace>"
  by(auto simp add: Hoare_alt Hoare_def Box_def throw_mode_error_def brute_force
    sv_mode_neq_def)

lemma pre_commit_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> pre_commit upd_f \<lbrace>>N K abs3_capacity_inv \<rbrace>"
  apply(auto simp add: pre_commit_def
    write_secondary_def
    intro!: bind_ruleNN_same bind_ruleNE_same
    try_catch_ruleNN_same
    HoareNN_throw
    chooseN_inv_NN 
    mapM'_invNN
    mapM'_ruleNE
    write_instance_capacity_inv_NN
    write_instance_capacity_inv_NE
    abs3_rollback_post_NN
    split: exceptions.split
  )
  apply(auto simp add: Hoare_alt Hoare_def Box_def brute_force abs3_capacity_inv_def)
  done

lemma do_commit_capacity_inv_NN:
  "\<lbrace>abs3_capacity_inv\<rbrace> do_commit \<lbrace>>N K abs3_capacity_inv\<rbrace>"
  apply(auto simp add: do_commit_def HoareNN_return' K_beta
    intro!: bind_ruleNN_same try_catch_ruleNN_same
    get_updated_post_NN
    HoareNN_throw
    abs3_rollback_post_NN
    split: exceptions.split
  )
  apply(auto simp add: Hoare_alt Hoare_def Box_def set_ptr_def get_ptr_defs write_ptr_defs 
    brute_force abs3_capacity_inv_def)
  done

lemma update_mode_post_NE:
  "P \<subseteq> abs3_capacity_inv 
    \<Longrightarrow> \<lbrace>P\<rbrace> update upd_f 
    \<lbrace>@N \<lambda>e. if e = InvalidInput then P \<union> sv_mode_neq CNormal else (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: update_def)
  apply(rule bind_ruleNE_same)
  apply(auto simp add: Hoare_alt Hoare_def Box_def throw_mode_error_def brute_force
    sv_mode_neq_def)[2]
  apply(auto)
  apply(rule bind_ruleNE)
  apply(rule conseqNE_rule[OF order_refl check_sizes_post_NE])
  apply(clarsimp)
  apply(rule check_sizes_post_NN)
  apply(auto simp add: K_beta intro!: ex_falsumNE)
  thm conseqNE_rule
  apply(erule conseqNE_rule[where R="K (sv_mode_neq CNormal)"])
  apply(auto)
  apply(auto simp add: K_beta
    update_def
    intro!: bind_ruleNE
    conseqNE_rule[OF _ throw_mode_error_post_NE]
    conseqNE_rule[OF _ check_sizes_post_NE]
    check_sizes_post_NN
    throw_mode_error_inv
    abs3_rollback_post_NN
  )
  apply(rule conseqNE_rule[OF _ precommit_mode_post_NE], simp_all)
  apply(rule pre_commit_capacity_inv_NN, simp_all add: K_beta)
  apply(rule do_commit_mode_post_NE)
  apply(rule do_commit_capacity_inv_NN, simp_all add: K_beta)
  apply(rule rollforward_mode_post_NE)
  apply(rule true_postNN)
  done

lemma set_updated_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> set_updated pinds \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(auto simp add: abs3_inv_defs brute_force)
  done

lemma sync_if_not_SEq_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> sync_if_not_SEq Secondary Primary xb \<lbrace>>N K abs3_norm_inv\<rbrace>"
  by(auto simp add: K_beta HoareNN_return'
    sync_if_not_SEq_def
    check_arrow_SEq_def
    intro!:
    bind_ruleNN_same
    HoareNE_return
    chooseN_ruleNE
    chooseN_post_NN
    HoareNE_vol_ops
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    just_sync_silence_deg_inv_NN
    norm_inv_subset_inv
    del: subsetI
  )

lemma sync_if_not_SEq_inv_NE:
  "to \<noteq> from \<Longrightarrow> \<lbrace>abs3_norm_inv\<rbrace> sync_if_not_SEq to from ind \<lbrace>@N K abs3_inv\<rbrace>"
  by(auto simp add: K_beta HoareNN_return'
    sync_if_not_SEq_def
    check_arrow_SEq_def
    intro!:
    bind_ruleNN_same
    bind_ruleNE_same
    HoareNE_return
    chooseN_ruleNE
    chooseN_post_NN
    HoareNE_vol_ops
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    just_sync_silence_deg_inv_NN
    conseqNE_rule[OF _ just_sync_silence_deg_inv_NE]
    norm_inv_subset_inv
    del: subsetI
  )

lemma rollback_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> rollback \<lbrace>@N K abs3_inv\<rbrace>"
  apply(auto simp add: K_beta HoareNE_throw'
    rollback_def
    intro!:
    bind_ruleNN
    bind_ruleNE
    try_catch_ruleNE
    mapM'_ruleNE
    mapM'_ruleNE
    HoareNE_return
    chooseN_ruleNE
    chooseN_post_NN
    HoareNE_vol_ops
    set_updated_inv_NN
    write_instance_inv_NN
    write_instance_inv_NE
    norm_inv_subset_inv
    get_updated_post_NN
    conseqNE_rule[OF _ just_sync_silence_deg_inv_NE]
    sync_if_not_SEq_inv_NE 
    sync_if_not_SEq_inv_NN
    split: exceptions.split
    del: subsetI
  )
  apply(intro mapM'_rule_NN_same[simplified K_def] sync_if_not_SEq_inv_NN[simplified K_def])
  done

lemma add_to_updated_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> add_to_updated xb \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(auto simp add: abs3_inv_defs brute_force)
  done

lemma pre_commit_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> pre_commit upd_f \<lbrace>@N K abs3_inv\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    pre_commit_def
    write_secondary_def
    intro!:
    bind_ruleNN
    bind_ruleNE
    try_catch_ruleNE
    mapM'_ruleNE
    HoareNE_return
    chooseN_ruleNE
    chooseN_post_NN
    HoareNE_vol_ops
    set_updated_inv_NN
    write_instance_inv_NN
    write_instance_inv_NE
    norm_inv_subset_inv
    rollback_inv_NE
    add_to_updated_inv_NN
    split: exceptions.split
    del: subsetI
  )

lemma do_commit_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> do_commit \<lbrace>@N K abs3_inv\<rbrace>"
  by(auto simp add: K_beta HoareNE_throw'
    do_commit_def
    intro!:
    bind_ruleNN
    bind_ruleNE
    HoareNE_return
    try_catch_ruleNE
    HoareNE_vol_ops
    set_ptr_inv_NE
    rollback_inv_NE
    get_updated_post_NN
    split: exceptions.split
)

lemma sync_if_not_SEq_remove_from_ptr_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> sync_if_not_SEq_remove_from_ptr Primary Secondary xb \<lbrace>@N K abs3_inv\<rbrace>"
  apply(auto simp add: K_beta HoareNE_throw' HoareNN_return'
    sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def
    remove_from_ptr_silence_deg_def
    intro!:
    bind_ruleNN
    bind_ruleNE
    mapM'_ruleNE
    HoareNE_return
    try_catch_ruleNE
    HoareNE_vol_ops
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NE
    sync_and_remove_from_ptr_silence_deg_inv_NN
    split: exceptions.split
  )
  apply(rule remove_from_ptr_inv_NE)
  apply(auto)
  apply(auto simp add: HoareNE_throw' K_beta
    sync_and_remove_from_ptr_silence_deg_def
    remove_from_ptr_silence_deg_def
    intro!:
    try_catch_ruleNE
    bind_ruleNN_same
    bind_ruleNE
    HoareNE_return
    sync_inv_NN
    sync_inv_NE
    norm_inv_subset_inv
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    remove_from_ptr_inv_NE
    split: exceptions.split
    del: subsetI
  )
  done

lemma rollforward_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> rollforward \<lbrace>@N K abs3_inv\<rbrace>"
  apply(auto simp add: K_beta HoareNE_throw'
    rollforward_def
    intro!:
    bind_ruleNN
    bind_ruleNE
    mapM'_ruleNE
    HoareNE_return
    try_catch_ruleNE
    HoareNE_vol_ops
    set_ptr_inv_NE
    rollback_inv_NE
    get_updated_post_NN
    chooseN_ruleNE
    chooseN_post_NN
    sync_if_not_SEq_remove_from_ptr_inv_NE
    split: exceptions.split
    intro: true_postNN
  )
  apply(auto simp add: HoareNE_throw' HoareNN_return'
    sync_if_not_SEq_remove_from_ptr_def
    check_arrow_SEq_def
    K_beta
    remove_from_ptr_silence_deg_def
    intro!:
    try_catch_ruleNN
    try_catch_ruleNE
    bind_ruleNN_same
    bind_ruleNE
    HoareNE_return
    HoareNN_throw
    sync_inv_NN
    sync_inv_NE
    norm_inv_subset_inv
    get_sync_direction_inv_NN
    get_sync_direction_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NE
    sync_and_remove_from_ptr_silence_deg_inv_NN
    split: exceptions.split
    del: subsetI
  )
  done

lemma doesnt_add_imp_preserves_empty_ptr_NN:
  "(\<And>ind. \<lbrace>removed_one_from_ptr ind\<rbrace> m \<lbrace>>N K (removed_one_from_ptr ind)\<rbrace>) \<Longrightarrow>
    \<lbrace>empty_ptr\<rbrace> m \<lbrace>>N K empty_ptr\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def brute_force)
  apply(clarsimp simp add: removed_one_from_ptr_def empty_ptr_def Collect_subset)
  by (metis all_not_in_conv)

lemma doesnt_add_imp_preserves_empty_ptr_NE:
  "(\<And>ind. \<lbrace>removed_one_from_ptr ind\<rbrace> m \<lbrace>@N K (removed_one_from_ptr ind)\<rbrace>) \<Longrightarrow>
    \<lbrace>empty_ptr\<rbrace> m \<lbrace>@N K empty_ptr\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_def brute_force)
  apply(clarsimp simp add: removed_one_from_ptr_def empty_ptr_def Collect_subset)
  by (metis all_not_in_conv)

lemma set_failed_empty_ptr_NN:
  "\<lbrace>P\<rbrace> set_failed \<lbrace>>N K empty_ptr \<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def K_def brute_force empty_ptr_def)

lemma fetch_empty_ptr:
  "\<lbrace>empty_ptr\<rbrace> Abs3_FetchDefs.fetch pind \<lbrace>>N K empty_ptr\<rbrace>"
  apply(auto simp add: HoareNN_return' HoareNE_throw' K_beta
    Abs3_FetchDefs.fetch_def
    Abs3_FetchDefs.fetch_same_def
    Abs3_CommonDefs.read_or_fail_def
    Abs3_CommonDefs.sync_def
    throw_mode_error_def
    intro!:
    bind_ruleNN_same
    try_catch_ruleNN
    try_catch_ruleNE
    bind_ruleNE
    HoareNN_throw
    HoareNE_vol_ops
    HoareNN_get_mode
    get_sync_direction_doesnt_add_post_NN[THEN doesnt_add_imp_preserves_empty_ptr_NN]
    read_doesnt_add_post_NN[THEN doesnt_add_imp_preserves_empty_ptr_NN]
    set_failed_empty_ptr_NN
    split: exceptions.split
  )
  apply(auto intro!:
    set_failed_empty_ptr_NN[simplified K_beta]
    read_doesnt_add_post_NN[THEN doesnt_add_imp_preserves_empty_ptr_NN, simplified K_def]
    read_doesnt_add_post_NE[THEN doesnt_add_imp_preserves_empty_ptr_NE, simplified K_def]
    write_instance_doesnt_add_post_NN[THEN doesnt_add_imp_preserves_empty_ptr_NN, simplified K_def]
    write_instance_doesnt_add_post_NE[THEN doesnt_add_imp_preserves_empty_ptr_NE, simplified K_def]
    simp add: K_beta
  )
  apply(auto intro!: set_failed_empty_ptr_NN[simplified K_def])
  done

lemma fetch_same_inv_NN:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_FetchDefs.fetch_same pind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  by(auto simp add:
    Abs3_FetchDefs.fetch_same_def
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    read_inv_NN
    read_or_fail_inv_NN
    sync_inv_NN
    read_inv_NE
    HoareNN_throw
    split: exceptions.split
  )

lemma fetch_same_inv_NE:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_FetchDefs.fetch_same pind \<lbrace>@N K abs3_inv\<rbrace>"
  by(auto simp add: HoareNE_throw' K_beta
    Abs3_FetchDefs.fetch_same_def
    intro!:
    try_catch_ruleNE
    bind_ruleNE
    read_inv_NN
    read_or_fail_inv_NN
    sync_inv_NN
    conseqNE_rule[OF _ sync_inv_NE]
    conseqNE_rule[OF _ read_or_fail_inv_NE]
    read_or_fail_inv_NE
    read_inv_NE
    HoareNN_throw
    norm_inv_subset_inv
    split: exceptions.split
    del: subsetI
  )

lemma read_mode_post_NE':
  "\<lbrace>P\<rbrace>
  Abs3_CommonDefs.read pind inst
  \<lbrace>@N \<lambda>e.
  if e = FailMode then sv_mode_equals CFail
  else if \<exists>v. e = Degraded v then sv_mode_neq CNormal
  else if e \<in> {FailWrite, FailLLRead, StartedExc, InvalidInput} then {}
  else UNIV \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs min_def
    sv_mode_equals_def sv_mode_neq_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: FailMode_exhaust)
  done

lemma read_or_fail_mode_post_NE:
  "\<lbrace>P\<rbrace> Abs3_CommonDefs.read_or_fail pind inst \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs sv_mode_neq_def min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: FailMode_exhaust)
  done

lemma fetch_same_mode_post_NE:
  "\<lbrace>abs3_capacity_inv\<rbrace> Abs3_FetchDefs.fetch_same pind \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_throw' K_beta
    Abs3_FetchDefs.fetch_same_def
    Abs3_CommonDefs.read_or_fail_def)
  apply(rule try_catch_ruleNE)
  apply(rule 
    HoareNE_join[OF read_mode_post_NE'[where P=UNIV] read_capacity_inv[THEN Hoare_imp_HoareNE],
    simplified])
  apply(auto simp add: HoareNE_throw' K_beta
    intro!: 
    bind_ruleNE
    conseqNE_rule[OF _ sync_mode_post_NE]
    read_or_fail_mode_post_NE
    elim: sv_mode_eq_neq_subset[THEN subsetD, rotated 1]
    split: exceptions.split split_if_asm
  )
  apply(auto simp add: HoareNE_throw' K_beta
    Abs3_CommonDefs.read_or_fail_def
    intro!:
    try_catch_ruleNN
    bind_ruleNN
    HoareNN_throw
    HoareNE_vol_ops
    read_mode_post_NE'
    conseqNN_rule[OF _ read_no_mode_upgrade_post_NN]
    set_failed_post_NN
    split: exceptions.split
    del: subsetI
  )
  done

end