theory Abs3_Ref
imports
  "../Composition"
  "../Pmm_Refinement"
  "../Misc_Lemmas"
  "../abs2/Abs2_CommonDefs"
  "../abs2/Abs2_UpdateDefs"
  Abs3_CommonDefs
  "../abs2/Abs2_Startup_Lemmas"
  "Abs3_StartupDefs"
  "Abs3_FetchDefs"
  "Abs3_UpdateDefs"
begin

lemmas brute_force =
  unfold_monad
  project_defs
  Abs2_LLInterface.ll_read_defs
  Abs2_LLInterface.ll_read_ptr_defs
  Abs2_LLInterface.write_f_defs
  Abs2_LLInterface.write_ptr_ll_defs
  Abs3_LLInterface.ll_read_defs
  Abs3_LLInterface.write_f_defs
  restart_mangle_sv_raw
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs
  force_same_and_def
  restart_mangle_sv_raw

fun pinst_state' :: "hw_state \<Rightarrow> bool \<Rightarrow> ec_state" where
  "pinst_state' IDead _ = EcDead"
  | "pinst_state' IUnwr _ = EcUnwr"
  | "pinst_state' IOK False = EcBad"
  | "pinst_state' IOK True = EcOK"

fun pinst_state :: "'c abs2_state \<Rightarrow> INST \<Rightarrow> ec_state" where
  "pinst_state sp Primary = pinst_state' (ptr1_state sp) (ptr_matches_hash sp Primary)"
  | "pinst_state sp Secondary = pinst_state' (ptr2_state sp) (ptr_matches_hash sp Secondary)"

definition abs3_abs2_strel :: "(('sv \<times> 'c abs3_state) \<times> ('sv \<times> 'c :: contents_cl abs2_state))set" 
  where
  "abs3_abs2_strel \<equiv> {((sva, spa), (svc, spc)). 
    sva = svc
    \<and> Abs3Defs.hw_state spa = Abs2Defs.hw_state spc
    \<and> Abs3Defs.memory spa = Abs2Defs.memory spc
    \<and> Abs3Defs.hstate spa = Abs2Defs.hstate spc
    \<and> (Abs3Defs.hstate spa = LDead \<or> (
        Abs3Defs.pub_hashes spa = pub_hashes2 (hash spc)
        \<and> Abs3Defs.ptr_state spa = combine_ec_states (pinst_state spc Primary) (pinst_state spc Secondary)
      ))
    \<and> (pinst_state spc Primary = EcOK \<longrightarrow> ptr1 spc = ptr spa)
    \<and> (pinst_state spc Secondary = EcOK \<longrightarrow> ptr2 spc = ptr spa)
    \<and> (ptr_matches_hash spc Primary \<longrightarrow> ptr1_state spc = IOK)
    \<and> (ptr_matches_hash spc Secondary \<longrightarrow> ptr2_state spc = IOK)
    \<and> (\<forall>ind inst. size (abs2_state.memory spc ind inst) \<le> capacity (Public ind))
  }"

definition min_one_good_ptr :: "('sv \<times> 'c abs2_state) set" where
  "min_one_good_ptr \<equiv> {(sv, spc). pinst_state spc Primary = EcOK \<or> pinst_state spc Secondary = EcOK}"

lemma restart_mangle_sv_refines_itself:
  "sim abs3_abs2_strel
    restart_mangle_sv
    restart_mangle_sv
  "
  apply(auto simp only: sim_def sim_ext_def)
  apply(clarsimp_all simp add: abs3_abs2_strel_def brute_force simN_wpc_def simS_wpc_def
        split: split_if_asm
        intro!: ext
  )
  done

lemma fun_upd_second_same:
  "(fun_upd_second f x y r = fun_upd_second f x y r') = (r = r')"
  by (metis (lifting) fun_upd_same)

lemma sim_ext_choice_lockstep:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; sim_ext R ma' mc' Sn Se Sr \<rbrakk>
    \<Longrightarrow> sim_ext R (ma [+] ma') (mc [+] mc') Sn Se Sr"
  by(auto simp add: sim_ext_def project_simps del: subsetI
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep)

lemma write_f_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_LLInterface.write_f ind inst cnt)
    (Abs2_LLInterface.write_f ind inst cnt)
  "
  apply(auto simp only: Abs2_LLInterface.write_f_def Abs3_LLInterface.write_f_def K_def
    intro!: restart_mangle_sv_refines_itself sims_all_intros)
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def simN_wpc_def simS_wpc_def)
  apply(clarsimp_all simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def abs3_abs2_strel_def
    brute_force force_same_and_def fun_upd_second_same
  )
  done

lemma ll_read_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_LLInterface.ll_read ind inst)
    (Abs2_LLInterface.ll_read ind inst)
  "
  apply(auto simp only: Abs2_LLInterface.ll_read_def Abs3_LLInterface.ll_read_def
    K_def intro!: restart_mangle_sv_refines_itself sims_all_intros
  )
  apply(tactic {* distinct_subgoals_tac *})
  apply(auto simp only: sim_def sim_ext_def simN_wpc_def simS_wpc_def)
  apply(clarsimp_all simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def abs3_abs2_strel_def
    brute_force force_same_and_def 
  )
  apply(metis)
  done

lemma LDegraded_le_comb:
  "(LDegraded \<le> combine_ec_states EcOK x) = True"
  "(LDegraded \<le> combine_ec_states x EcOK) = True"
  by(case_tac[!] x, simp_all add: min_def)

lemma hstate_comb_ptr_hstate:
  "abs2_state.hstate spc = LDead \<Longrightarrow>
    combine_ec_states (pinst_state spc Primary) (pinst_state spc Secondary) = LDead"
  apply(auto)
  apply(case_tac "ptr1_state spc")
  apply(auto)
  apply(case_tac[!] "ptr2_state spc")
  apply(auto)
  done

lemma pinst_state_EcOK_I:
  "\<lbrakk>((sva, spa), svc, spc) \<in> abs3_abs2_strel; ptr_hashes2 (hash spc) Primary = ptr1 spc;
    ptr1_state spc \<noteq> IDead; abs2_state.hstate spc \<noteq> LDead
  \<rbrakk> \<Longrightarrow>
  pinst_state' (ptr1_state spc) True = EcOK"
  by(case_tac "ptr1_state spc", clarsimp_all simp add: abs3_abs2_strel_def)

 
lemma exc_val_aux:
  "(\<forall>exc_val. exc_val = EB True \<and> P exc_val \<or> v \<noteq> exc_val \<and> Q exc_val) =
    (v = EB True \<and> P (EB True) \<and> (\<forall>exc_val. exc_val \<noteq> EB True \<longrightarrow> Q exc_val))"
  by(metis)

lemma combine_ec_states_2_False:
  "combine_ec_states (pinst_state' x False) (pinst_state' y False) = LDead"
  by(case_tac x, case_tac[!] y, simp_all)

lemma combine_ec_states_False_Dead:
  "combine_ec_states (pinst_state' x False) EcDead = LDead"
  by(case_tac x, simp_all)

lemma combine_ec_states_Dead_False:
  "combine_ec_states EcDead (pinst_state' x False) = LDead"
  by(case_tac x, simp_all)

lemma pinst_state_False_notOK:
  "pinst_state' x False \<noteq> EcOK"
  by(case_tac x, simp_all)

lemma comb_EcOK:
  "combine_ec_states EcOK (pinst_state' (min x IUnwr) False) = LDegraded"
  "combine_ec_states (pinst_state' (min x IUnwr) False) EcOK  = LDegraded"
  by(case_tac[!] x, simp_all add: min_def)

lemma ex_leqt: "\<exists>x :: 'a :: preorder. x \<le> y"
  by (metis order_refl)

lemma combined_state_False_EcBad:
  "combine_ec_states (pinst_state' x False) EcBad = LDead"
  by(case_tac x, simp_all)

lemma pinst_state_min_IUnwr:
  "pinst_state' (min x IUnwr) y \<noteq> EcOK"
  by(case_tac x, simp_all add: min_def)

lemma comb_EcDead_Unwr:
  "combine_ec_states EcDead (pinst_state' (min x IUnwr) y) = LDead"
  "combine_ec_states (pinst_state' (min x IUnwr) y) EcDead  = LDead"
  by(case_tac[!] x, simp_all add: min_def)

lemma pinst_state_IOK_EcOK:
  "(pinst_state' IOK t = EcOK) = (t = True)"
  by(case_tac t, simp_all)
  
lemma comb_EcOK_le:
  "(combine_ec_states EcOK t1 \<le> combine_ec_states EcOK t2) = (t1 \<le> EcUnwr \<or> EcBad \<le> t2)"
  by(case_tac t1, case_tac[!] t2, simp_all)

lemma comb_EcOK_IOK:
  "combine_ec_states EcOK (pinst_state' IOK x) = LOK"
  "combine_ec_states (pinst_state' IOK x) EcOK  = LOK"
  by(case_tac[!] x, simp_all)

lemma le_EcOK:
  "(x \<le> EcOK) = True"
  by(case_tac x, simp_all)

lemma comb_hstate_EcOK_not_LOK:
  "(combine_ec_states EcOK (pinst_state' x f) \<noteq> LOK) = (x \<noteq> IOK)"
  "(combine_ec_states (pinst_state' x f) EcOK  \<noteq> LOK) = (x \<noteq> IOK)"
  by(case_tac[!] x, case_tac[!] f, simp_all)

lemma comb_hstate_EcOK_LDegraded:
  "(combine_ec_states EcOK (pinst_state' x f) = LDegraded) = (x \<le> IUnwr)"
  "(combine_ec_states (pinst_state' x f) EcOK  = LDegraded) = (x \<le> IUnwr)"
  "(LDegraded = combine_ec_states EcOK (pinst_state' x f)) = (x \<le> IUnwr)"
  "(LDegraded = combine_ec_states (pinst_state' x f) EcOK ) = (x \<le> IUnwr)"
  by(case_tac[!] x, case_tac[!] f, simp_all)

lemmas abs3_ref_simps =
  LDegraded_le_comb
  hstate_comb_ptr_hstate
  pinst_state_EcOK_I
  exc_val_aux
  combine_ec_states_2_False
  combine_ec_states_False_Dead
  combine_ec_states_Dead_False
  pinst_state_False_notOK
  comb_EcOK
  ex_leqt
  combined_state_False_EcBad
  pinst_state_min_IUnwr
  comb_EcDead_Unwr
  pinst_state_IOK_EcOK
  comb_EcOK_le
  comb_EcOK_IOK
  le_EcOK
  comb_hstate_EcOK_not_LOK
  comb_hstate_EcOK_LDegraded

lemma get_ptr_false_ref_NN: "sim_normN_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr False)
  (Abs2_CommonDefs.get_ptr False)
  (force_same_and (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr))"
  apply(auto simp only: sim_normN_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp only: Abs2_CommonDefs.get_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.sync_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_defs Abs2_CommonDefs.write_hashes_defs
    restart_mangle_sv_def
    min_one_good_ptr_def
    update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm)
  apply(iterative_simp)+
  apply(simp_all (no_asm) add: get_ptr_defs restart_mangle_sv_def brute_force)
  apply(simp (no_asm) add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: abs3_abs2_strel_def ec_state_simps)
  apply(clarsimp_all simp add: abs3_ref_simps)
  apply(metis)
  done
 
definition write_hashes_NE :: "'c htable2 \<Rightarrow>
  ('c exceptions, pmm_vol_mem \<times> 'c :: contents_cl abs2_state)nds_monad"
  where
  "write_hashes_NE ht \<equiv> (
  (Abs2_CommonDefs.write_hashes_fail ht [+]
  Abs2_CommonDefs.write_hashes_degrade ht [+] Abs2_CommonDefs.write_hashes_degrade_write ht)
  :: (unit, pmm_vol_mem, 'c)abs2_monad
  )\<Down>NE"

lemma write_hashes_proj_simps:
  "(Abs2_CommonDefs.write_hashes ht)\<Down>NE = write_hashes_NE ht"
  by(auto simp add: Abs2_CommonDefs.write_hashes_defs brute_force write_hashes_NE_def intro!: ext)

lemma get_ptr_true_ref_NN: "sim_normN_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr True)
  (Abs2_CommonDefs.get_ptr True)
  (force_same_and (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr))"
  apply(auto simp only: sim_normN_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp only: Abs2_CommonDefs.get_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.sync_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_DegradedH_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_all_defs Abs2_CommonDefs.write_hashes_defs
    get_hashes_NN_def get_hashes_DegradedH_NN_def
    restart_mangle_sv_def
    min_one_good_ptr_def update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=2]]
  apply(simp_all add: pinst_state_EcOK_I min2)
  apply(case_tac[!] svc, clarsimp_all del: disjCI)
  apply(simp_all (no_asm) add: get_ptr_defs restart_mangle_sv_def brute_force)
  apply(simp_all add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: abs3_abs2_strel_def ec_state_simps del: disjCI)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps hstate_simps not_BothGone_le_OneGone min_def
    parse_serialize_ptr
    split: split_if_asm
    del: disjCI
  )
  apply(clarsimp_all simp add: all_conj ge_DegP_imp_ge_DegH del: disjCI)
  apply(clarsimp_all simp add: conj_commute exc_val_aux ge_DegP_imp_ge_DegH all_conj)
  apply (smt exhaust_EcOK combine_ec_states.simps(4) less_eq_IUnwr pinst_state'.simps(2)
    combine_ec_states.simps(1) combine_ec_states.simps(2) order_eq_iff pinst_state'.simps(3)
    pinst_state'.simps(4))+
  done
 
lemma get_ptr_ref_NN: "sim_normN_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr b)
  (Abs2_CommonDefs.get_ptr b)
  (force_same_and (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr))"
  by(case_tac b, auto intro: get_ptr_true_ref_NN get_ptr_false_ref_NN)

lemma get_ptr_false_ref_NE: "sim_normE_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr False)
  (Abs2_CommonDefs.get_ptr False)
  (force_same_and abs3_abs2_strel)"
  apply(auto simp only: sim_normE_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp only: Abs2_CommonDefs.get_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.sync_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_DegradedH_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_all_defs Abs2_CommonDefs.write_hashes_defs
    get_hashes_NN_def get_hashes_DegradedH_NN_def
    restart_mangle_sv_def
    min_one_good_ptr_def update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=1]]
  apply(case_tac[!] svc, clarsimp_all simp add: abs3_ref_simps min2 del: disjCI)
  apply(simp_all (no_asm) add: get_ptr_defs restart_mangle_sv_def brute_force parse_serialize_ptr)
  apply(simp_all add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: abs3_abs2_strel_def ec_state_simps del: disjCI)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_abs2_strel_def abs3_ref_simps hstate_simps
    not_BothGone_le_OneGone min_def
    split: split_if_asm
    del: disjCI
  )
  apply(iterative_simp)+
  apply (smt exhaust_EcOK combine_ec_states.simps(4) less_eq_IUnwr pinst_state'.simps(2)
    combine_ec_states.simps(1) combine_ec_states.simps(2) order_eq_iff pinst_state'.simps(3)
    pinst_state'.simps(4))+
  done
 
lemma get_ptr_true_ref_NE: "sim_normE_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr True)
  (Abs2_CommonDefs.get_ptr True)
  (force_same_and abs3_abs2_strel)"
  apply(auto simp only: sim_normE_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp only: Abs2_CommonDefs.get_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.sync_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_DegradedH_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_all_defs Abs2_CommonDefs.write_hashes_defs
    get_hashes_NN_def get_hashes_DegradedH_NN_def
    restart_mangle_sv_def
    min_one_good_ptr_def update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=1]]
  apply(case_tac[!] svc, clarsimp_all simp add: abs3_ref_simps min2 parse_serialize_ptr del: disjCI)
  apply(simp_all (no_asm) add: get_ptr_defs restart_mangle_sv_def brute_force)
  apply(simp_all add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: abs3_abs2_strel_def ec_state_simps del: disjCI)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_abs2_strel_def abs3_ref_simps hstate_simps
    not_BothGone_le_OneGone min_def
    split: split_if_asm
    del: disjCI
  )
  apply(iterative_simp)+
  done

lemma get_ptr_ref_NE: "sim_normE_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_ptr b)
  (Abs2_CommonDefs.get_ptr b)
  (force_same_and abs3_abs2_strel)"
  by(case_tac b, auto intro: get_ptr_false_ref_NE get_ptr_true_ref_NE)

lemma get_ptr_ref_N[ref_thm]:
  "sim_norm_ext abs3_abs2_strel
    (Abs3_CommonDefs.get_ptr b)
    (Abs2_CommonDefs.get_ptr b)
    (force_same_and (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr))
    (force_same_and abs3_abs2_strel)
  "
  by(auto simp only: sim_norm_ext_decompose sim_normN_def sim_normE_def
    intro!: get_ptr_ref_NN get_ptr_ref_NE)

lemma not_IOK_is_le_IUnwr:
  "(x \<noteq> IOK) = (x \<le> IUnwr)"
  by(case_tac x, simp_all)

lemma write_ptr_ref_NN: " sim_normN_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (Abs3_CommonDefs.write_ptr ptr_set)
  (Abs2_CommonDefs.write_ptr ptr_set)
  (force_same_and abs3_abs2_strel)"
  apply(auto simp only: sim_normN_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp only: Abs2_CommonDefs.write_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp only: write_ptr_instance_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp only: update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(clarsimp_all simp add: Abs2_CommonDefs.get_hashes_defs Abs2_CommonDefs.write_hashes_defs
    brute_force split: split_if_asm)
  apply(clarsimp_all simp add: write_ptr_defs brute_force)
  apply(clarsimp_all simp add: abs3_abs2_strel_def abs3_ref_simps update_htable2_alt 
    ptr_matches_htable2_def)
  apply(case_tac "ptr1_state spc")
  apply(simp_all)
  apply(case_tac[!] "ptr2_state spc")
  apply(iterative_simp simp add: abs3_ref_simps min_one_good_ptr_def parse_serialize_ptr)+
  done

lemma restart_mangle_sv_proj_simps:
  "restart_mangle_sv\<Down>NN = empty"
  "restart_mangle_sv\<Down>NE = empty"
  by(auto simp add: brute_force)

lemma write_ptr_ref_NE: "sim_normE_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (Abs3_CommonDefs.write_ptr ht)
  (Abs2_CommonDefs.write_ptr ht)
  (force_same_and abs3_abs2_strel)"
  apply(auto simp only: sim_normE_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp only: Abs2_CommonDefs.write_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_instance_def read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: read_ptr_DegradedH_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: update_ptr_instance_hash_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_proj_simps cong: if_cong exceptions.case_cong)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_defs Abs2_CommonDefs.get_hashes_DegradedH_def
    Abs2_CommonDefs.get_hashes_degrade'_def Abs2_CommonDefs.write_hashes_defs
    get_hashes_DegradedH_NN_def get_hashes_NN_def
    update_htable2_alt ptr_matches_htable2_def parse_serialize_ptr
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp simp add: abs3_ref_simps min2)+
  apply(simp_all (no_asm) add: write_ptr_defs restart_mangle_sv_raw choice_def)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all del: disjCI)
  apply(clarsimp_all simp add: abs3_abs2_strel_def abs3_ref_simps hstate_simps
    not_BothGone_le_OneGone min_def
    split: split_if_asm
    del: disjCI
  )
  using [[simp_depth_limit=1]]
  apply(case_tac[!] "card_mode svc")
  apply(case_tac [!] svc, clarsimp_all del: disjCI)
  apply(simp_all add: min_one_good_ptr_def)
  apply(iterative_simp del: disjCI)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_abs2_strel_def abs3_ref_simps hstate_simps
    parse_serialize_ptr not_BothGone_le_OneGone 
    split: split_if_asm del: disjCI
  )
  apply(case_tac[!] "ptr1_state spc", clarsimp_all simp add: abs3_ref_simps del: disjCI)
  apply(case_tac[!] "ptr2_state spc", clarsimp_all simp add: abs3_ref_simps del: disjCI)
  done

lemma write_ptr_ref_N:
  "sim_norm_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (Abs3_CommonDefs.write_ptr ht)
  (Abs2_CommonDefs.write_ptr ht)
  (force_same_and abs3_abs2_strel)
  (force_same_and abs3_abs2_strel)
  "
  by(auto simp add: sim_norm_ext_decompose intro!: write_ptr_ref_NN write_ptr_ref_NE)

lemma get_hashes_ref: "sim_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_hashes)
  (Abs2_CommonDefs.get_hashes)
  (\<lambda>va vc. (if va = (pub_hashes2 vc)
            then UNIV \<times> {(svc, spc). ptr_hashes2 (hash spc) = ptr_hashes2 vc \<and>
                abs2_state.hstate spc \<noteq> LDead }
            else {}
           ) \<inter> abs3_abs2_strel)
  (force_same_and abs3_abs2_strel)
  (abs3_abs2_strel)
  "
  apply(auto simp add: Abs3_CommonDefs.get_hashes_def Abs2_CommonDefs.get_hashes_def
    intro!: sim_ext_choice_lockstep)
  apply(auto simp add: sim_ext_def)
  apply(clarsimp_all simp add: simN_def simS_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(clarsimp_all simp add: brute_force Abs3_CommonDefs.get_hashes_defs
    Abs2_CommonDefs.get_hashes_defs abs3_abs2_strel_def)
  apply(iterative_simp simp add: pinst_state_False_notOK combine_ec_states_2_False)+
  done

lemma get_hashes_DegradedH_ref: "sim_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_hashes_DegradedH)
  (Abs2_CommonDefs.get_hashes_DegradedH)
  (\<lambda>va vc. (if va = (pub_hashes2 vc)
            then UNIV \<times> {(svc, spc). ptr_hashes2 (hash spc) = ptr_hashes2 vc \<and>
                abs2_state.hstate spc \<noteq> LDead }
            else {}
           ) \<inter> abs3_abs2_strel)
  (force_same_and abs3_abs2_strel)
  (abs3_abs2_strel)
  "
  apply(auto simp add: Abs3_CommonDefs.get_hashes_DegradedH_def Abs2_CommonDefs.get_hashes_DegradedH_def
    intro!: sim_ext_choice_lockstep)
  apply(auto simp add: sim_ext_def)
  apply(clarsimp_all simp add: simN_def simS_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(clarsimp_all simp add: brute_force Abs3_CommonDefs.get_hashes_DegradedH_defs
    Abs2_CommonDefs.get_hashes_DegradedH_defs abs3_abs2_strel_def)
  apply(iterative_simp simp add: pinst_state_False_notOK combine_ec_states_2_False)+
  done

lemma update_instance_hash_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_CommonDefs.update_instance_hash ind inst cnt)
    (Abs2_CommonDefs.update_pub_instance_hash ind inst cnt)
  "
  apply(simp add: Abs3_CommonDefs.update_instance_hash_def Abs2_CommonDefs.update_pub_instance_hash_def)
  apply(simp only: sim_def)
  apply(rule sim_ext_bindI[OF get_hashes_ref])
  apply(auto intro: sim_exts_empty_rel)
  apply(auto simp add: Abs3_CommonDefs.write_hashes_def Abs2_CommonDefs.write_hashes_def
    intro!: sim_ext_choice_lockstep restart_mangle_sv_refines_itself
  )
  apply(auto simp add: sim_ext_def simN_wpc_empty_conc simS_wpc_empty_conc write_hashes_simps
    del: subsetI)
  apply(clarsimp_all simp add: simN_def simS_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(clarsimp_all simp add: brute_force Abs3_CommonDefs.write_hashes_defs
    Abs2_CommonDefs.write_hashes_defs abs3_abs2_strel_def update_htable2_alt)
  apply(iterative_simp simp add: pinst_state_False_notOK min_def split: split_if_asm)+
  done

lemma get_ptrs_R:
  "
  ((Abs3_CommonDefs.get_ptr b
    :: (pub_indx set, 'c :: size) abs3_monad
  )\<Down>R) =
  ((restart_mangle_sv
    :: (pub_indx set, 'c) abs3_monad
  )\<Down>R)"
  apply(auto simp add: project_defs intro!: ext)
  apply(auto simp add: get_ptr_defs choice_def split: split_if_asm)
  done

lemma write_hashes_R:
  "(((Abs2_CommonDefs.write_hashes ht) :: (unit, pmm_vol_mem, 'c)abs2_monad)\<Down>R) =
    ((Abs2_CommonDefs.write_hashes_partial_restart ht)
        :: (unit, pmm_vol_mem, 'c)abs2_monad
    )\<Down>R [+]
    (restart_mangle_sv
        :: (unit, pmm_vol_mem, 'c)abs2_monad
    )\<Down>R"
  apply(auto simp add: project_defs intro!: ext)
  apply(auto simp add: Abs2_CommonDefs.write_hashes_defs choice_def)
  done

declare disjCI[rule del]

definition ll_read_ptr_R where
  "ll_read_ptr_R inst \<equiv>
    (<try> ll_read_ptr_failure inst <catch> K restart_mangle_sv)
    [+] do { ll_read_ptr_any inst; restart_mangle_sv }
  "

lemma pinst_state_OK_cond:
  "(pinst_state' x f = EcOK) = (x = IOK \<and> f = True)"
  by(case_tac x, case_tac[!] f, simp_all)

lemma ll_read_ptr_R:
  "(ll_read_ptr inst :: (pub_indx set, 'sv, 'c :: serialization_cl)abs2_monad)\<Down>R =
  (ll_read_ptr_R inst)\<Down>R
  "
  apply(rule ext)
  apply(auto simp add: ll_read_ptr_R_def first_stage ll_read_ptr_def)
  apply(clarsimp_all simp add: choice_def)
  apply(clarsimp_all simp add: brute_force split: split_if split_if_asm)
  apply (smt abs2_state.iffs abs2_state.simps(12) abs2_state.simps(13) abs2_state.surjective)
  by (smt abs2_state.iffs abs2_state.simps(14) abs2_state.simps(15) abs2_state.surjective)

lemma abs2_get_ptr_R:
  "((Abs2_CommonDefs.get_ptr b)\<Down>R) (sv, sp) \<subseteq> check_ptr_R (sv, sp)"
  apply(auto)
  apply(rename_tac sv' sp')
  apply(simp only: Abs2_CommonDefs.get_ptr_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.sync_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: Abs2_CommonDefs.write_ptr_instance_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.update_ptr_instance_hash_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs2_CommonDefs.read_ptr_DegradedH_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_all_defs Abs2_CommonDefs.write_hashes_defs
    get_hashes_NN_def get_hashes_DegradedH_NN_def
    restart_mangle_sv_def
    min_one_good_ptr_def update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  using [[simp_depth_limit=1]]
  apply(case_tac[!] sv, clarsimp_all simp add: abs3_ref_simps min2 del: disjCI)
  apply(simp_all (no_asm) add: check_ptr_R_def brute_force)
  apply(simp_all add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs2_ptr_same_rest_def abs3_ref_simps hstate_simps
    not_BothGone_le_OneGone min_def
    split: split_if_asm
    del: disjCI
  )
  apply(iterative_simp)+
  apply(case_tac[!] "ptr2_state sp", clarsimp_all)
  apply(case_tac[!] "ptr1_state sp", clarsimp_all)
  done

lemma restart_mangle_sv_startup1:
  "(restart_mangle_sv\<Down>R) (sva, spa) \<guillemotright>=S startup1_NN = startup1_NN (sva, spa)"
  apply(auto simp add: brute_force)
  apply(auto simp add: startup1_NN_def choice_def)
  done

lemma combine_ec_states_is_LDead:
  "(combine_ec_states x y = LDead) = (x \<noteq> EcOK \<and> y \<noteq> EcOK)"
  "(LDead = combine_ec_states x y) = (x \<noteq> EcOK \<and> y \<noteq> EcOK)"
  by(case_tac[!] x, case_tac[!] y, simp_all)

lemma startup1_ref_NN':
  "\<lbrakk>((sva, spa), svc, spc) \<in> abs3_abs2_strel;
    ((), svc', spc') \<in> startup1'_NN (svc, spc)\<rbrakk>
       \<Longrightarrow> \<exists>a b. ((), a, b) \<in> startup1_NN (sva, spa) \<and>
                ((a, b), svc', spc') \<in> abs3_abs2_strel
  "
  apply(clarsimp simp add: startup1'_NN_def brute_force)
  apply(iterative_simp simp add: IOK_less_eq EcOK_less_eq)+
  apply(clarsimp_all simp add: abs3_abs2_strel_def startup1_NN_def choice_def abs2_ptr_same_rest_def
    comb_EcOK_IOK)
  apply(simp_all (no_asm)
    add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps min_one_good_ptr_def
    hstate_simps
    not_BothGone_le_OneGone 
    parse_serialize_ptr 
    pinst_state_OK_cond IOK_less_eq
    comb_EcOK_IOK less_eq_IUnwr
    not_IOK_is_le_IUnwr
    combine_ec_states_is_LDead
  )
  apply (smt IOK_less_eq abs3_ref_simps(20) exhaust_IUnwr hw_state.simps(2)
    hw_state.simps(3) min_max.inf.idem min_max.le_iff_inf)
  apply (metis comb_EcOK_IOK(1) state_ineq_simps(3))
  apply (metis comb_EcOK_IOK(2) state_ineq_simps(3))
  apply (metis (no_types) abs3_ref_simps(22) less_eq_IUnwr less_eq_log_state.simps(7)
    not_IOK_is_le_IUnwr order_trans)
  done

lemma startup1_ref_NN:
  "sim_normN abs3_abs2_strel
    Abs3_StartupDefs.startup1
    abs2_startup1'
  "
  by(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def
    Abs3_StartupDefs.startup1_def abs2_startup1'_def compose_project
    intro!: startup1_ref_NN')

lemma less_eq_LDegraded:
  "(x \<le> LDegraded) = (x = LDegraded \<or> x = LDead)"
  by(case_tac x, simp_all)

lemma LDegraded_less_eq:
  "(LDegraded \<le> x) = (x = LDegraded \<or> x = LOK)"
  by(case_tac x, simp_all)

lemma startup1_ref_NE':
  "\<lbrakk>((sva, spa), svc, spc) \<in> abs3_abs2_strel;
    (e, svc', spc') \<in> startup1'_NE (svc, spc)\<rbrakk>
       \<Longrightarrow> \<exists>a b. (e, a, b) \<in> startup1_NE (sva, spa) \<and>
                ((a, b), svc', spc') \<in> abs3_abs2_strel
  "
  apply(clarsimp simp add: startup1'_NE_def brute_force)
  apply(iterative_simp simp add: IOK_less_eq EcOK_less_eq)+
  apply(clarsimp_all simp add: abs3_abs2_strel_def startup1_NE_def choice_def abs2_ptr_same_rest_def
    comb_EcOK_IOK)
  apply(simp_all (no_asm)
    add: HOL.conj_disj_distribL HOL.conj_disj_distribR ex_disj_distrib conj_assoc)
  apply(clarsimp_all simp add: abs3_ref_simps min_one_good_ptr_def
    hstate_simps
    not_BothGone_le_OneGone 
    parse_serialize_ptr 
    pinst_state_OK_cond IOK_less_eq
    comb_EcOK_IOK less_eq_IUnwr
    not_IOK_is_le_IUnwr
    combine_ec_states_is_LDead
  )
  apply(iterative_simp simp add: ex_leqt less_eq_LDead)+
  apply(metis)+
  done

lemma startup1_ref_NE:
  "sim_normE abs3_abs2_strel
    Abs3_StartupDefs.startup1
    abs2_startup1'
  "
  by(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def
    Abs3_StartupDefs.startup1_def abs2_startup1'_def compose_project
    intro!: startup1_ref_NE')

lemma get_ptr_ref_R_NN:
  "sim_normN abs3_abs2_strel
  (\<lambda>s. (((Abs3_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  "
  apply(auto simp only: sim_normN_def sim_normN_ext_def simN_wpc_def bindS_projectNN)
  apply(simp_all add: get_ptrs_R Abs3_StartupDefs.startup1_def compose_project force_same_and_def
    abs2_startup1'_def 
  )
  apply(auto simp add: restart_mangle_sv_startup1
    dest!:
    bindS_mono1[OF abs2_get_ptr_R, THEN subsetD]
    check_ptr_R_NN_reduction
    intro!: startup1_ref_NN'
  )
  done
 
lemma restart_mangle_sv_startup1_NE:
  "(restart_mangle_sv\<Down>R) (sva, spa) \<guillemotright>=S startup1_NE = startup1_NE (sva, spa)"
  apply(auto simp add: brute_force)
  apply(auto simp add: startup1_NE_def choice_def)
  done

lemma get_ptr_ref_R_NE:
  "sim_normE abs3_abs2_strel
  (\<lambda>s. (((Abs3_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  "
  apply(auto simp only: sim_normE_def sim_normE_ext_def simN_wpc_def bindS_projectNE)
  apply(simp_all add: get_ptrs_R Abs3_StartupDefs.startup1_def compose_project force_same_and_def
    abs2_startup1'_def 
  )
  apply(auto simp add: restart_mangle_sv_startup1_NE
    dest!:
    bindS_mono1[OF abs2_get_ptr_R, THEN subsetD]
    check_ptr_R_NE_reduction
    intro!: startup1_ref_NE'
  )
  done

lemma get_ptrs_refs_R_N[ref_thm]:
  "sim_norm abs3_abs2_strel
  (\<lambda>s. (((Abs3_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.get_ptr b)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  "
  by(auto simp add: sim_norm_decompose intro!: get_ptr_ref_R_NN get_ptr_ref_R_NE)

lemma write_ptr_ref_R_NN:
  "sim_normN_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (\<lambda>s. (((Abs3_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  (force_same_and abs3_abs2_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def force_same_and_def
    bindS_projectNN startup1_def abs2_startup1'_def compose_project bindS_decompose)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all (no_asm) add: write_ptr_defs startup1_NN_def choice_def brute_force)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  apply(simp only: Abs2_CommonDefs.write_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_instance_def read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_DegradedH_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: update_ptr_instance_hash_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice restart_mangle_sv_proj_simps
    cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_defs Abs2_CommonDefs.get_hashes_DegradedH_def
    Abs2_CommonDefs.get_hashes_degrade'_def Abs2_CommonDefs.write_hashes_defs
    get_hashes_DegradedH_NN_def get_hashes_NN_def
    update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp simp add: abs3_ref_simps min2 del: disjCI)+
  apply(clarsimp_all simp add: startup1'_NN_def choice_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(clarsimp_all simp add: IOK_less_eq LOK_less_eq)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: min_one_good_ptr_def abs3_abs2_strel_def abs2_ptr_same_rest_def)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: abs3_ref_simps hstate_simps not_BothGone_le_OneGone parse_serialize_ptr 
    ex_leqt
  )
  apply(clarsimp_all simp add: pinst_state_OK_cond)
  apply ((smt comb_hstate_EcOK_LDegraded comb_hstate_EcOK_not_LOK
    not_IOK_is_le_IUnwr pinst_state'.simps IOK_less_eq
      parse_serialize_ptr less_eq_log_state.simps
    )+)
  done
 
lemma write_ptr_ref_R_NE:
  "sim_normE_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (\<lambda>s. (((Abs3_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  (force_same_and abs3_abs2_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def
    bindS_projectNE startup1_def abs2_startup1'_def compose_project bindS_decompose)
  apply(rename_tac sva spa svc spc ec svc'' spc'' svc' spc')
  apply(simp_all (no_asm) add: write_ptr_defs startup1_NE_def choice_def brute_force)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  apply(simp only: Abs2_CommonDefs.write_ptr_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_instance_def read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: read_ptr_or_fail_silence_deg_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_or_fail_def Let_def)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(simp_all only: read_ptr_def Let_def)
  apply(simp_all only: vol_ops_project_simps first_stage if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_ptr_DegradedH_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice cong: if_cong exceptions.case_cong)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: update_ptr_instance_hash_def Let_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: vol_ops_project_simps first_stage if_choice restart_mangle_sv_proj_simps
    cong: if_cong exceptions.case_cong)
  apply(simp_all only: get_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: get_hashes_NE_def get_hashes_DegradedH_NE_def)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_hashes_proj_simps cong: exceptions.case_cong)
  apply(simp_all only: write_hashes_NE_def cong: exceptions.case_cong)
  apply(simp_all only: first_stage vol_ops_project_simps if_choice)
  apply(simp_all only: choice_def Un_iff, safe)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(clarsimp_all simp add: brute_force
    Abs2_CommonDefs.get_hashes_defs Abs2_CommonDefs.get_hashes_DegradedH_def
    Abs2_CommonDefs.get_hashes_degrade'_def Abs2_CommonDefs.write_hashes_defs
    get_hashes_DegradedH_NN_def get_hashes_NN_def
    update_htable2_alt ptr_matches_htable2_def
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp simp add: abs3_ref_simps min2 del: disjCI)+
  apply(clarsimp_all simp add: startup1'_NE_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(clarsimp_all simp add: IOK_less_eq LOK_less_eq)
  apply(clarsimp_all simp add: min_one_good_ptr_def abs3_abs2_strel_def abs2_ptr_same_rest_def)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: abs3_ref_simps hstate_simps not_BothGone_le_OneGone parse_serialize_ptr 
    ex_leqt
  )
  apply(clarsimp_all simp add: pinst_state_OK_cond)
  apply (smt combine_ec_states_2_False less_eq_LDegraded not_BothGone_le_OneGone less_eq_LOK
    pub_hashes.simps
  )+
  done

lemma write_ptr_ref_R_N[ref_thm]:
  "sim_norm_ext (abs3_abs2_strel \<inter> UNIV \<times> min_one_good_ptr)
  (\<lambda>s. (((Abs3_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1))
  (\<lambda>s. (((Abs2_CommonDefs.write_ptr ptr_set)\<Down>R) s \<guillemotright>=S abs2_startup1'))
  (force_same_and abs3_abs2_strel)
  (force_same_and abs3_abs2_strel)
  "
  by(auto simp add: sim_norm_ext_decompose intro!: write_ptr_ref_R_NN write_ptr_ref_R_NE)

lemma remove_from_ptr_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_CommonDefs.remove_from_ptr pind)
    (Abs2_CommonDefs.remove_from_ptr pind)
  "
  apply(simp only: Abs3_CommonDefs.remove_from_ptr_def Abs2_CommonDefs.remove_from_ptr_def
    sim_norm_def)
  apply(rule sim_norm_ext_bind_sameI[OF get_ptr_ref_N])
  apply(force simp add: Let_def intro!: write_ptr_ref_N)
  apply(simp_all add: force_same_and_same force_same_and_diff)
  done

lemma remove_from_ptr_ref_R_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_CommonDefs.remove_from_ptr pind)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((Abs2_CommonDefs.remove_from_ptr pind)\<Down>R) s \<guillemotright>=S abs2_startup1')
  "
  apply(simp only: Abs3_CommonDefs.remove_from_ptr_def Abs2_CommonDefs.remove_from_ptr_def
    sim_norm_def)
  apply(rule sim_norm_ext_bind_bindS[OF get_ptr_ref_N])
  apply(clarsimp simp only: sim_norm_def[THEN symmetric] intro!: get_ptrs_refs_R_N)
  apply(clarsimp simp add: Let_def)
  apply(rule write_ptr_ref_R_N)
  apply(simp_all add: force_same_and_same force_same_and_diff)
  done

lemma set_ptr_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_CommonDefs.set_ptr ptr_set)
    (Abs2_CommonDefs.set_ptr ptr_set)
  "
  apply(simp only: Abs3_CommonDefs.set_ptr_def Abs2_CommonDefs.set_ptr_def  
    sim_norm_def)
  apply(rule sim_norm_ext_bind_sameI[OF get_ptr_ref_N])
  apply(force simp add: Let_def intro!: write_ptr_ref_N)
  apply(simp_all add: force_same_and_same force_same_and_diff)
  done

lemma set_ptr_ref_R_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_CommonDefs.set_ptr ptr_set)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((Abs2_CommonDefs.set_ptr ptr_set)\<Down>R) s \<guillemotright>=S abs2_startup1')
  "
  apply(simp only: Abs3_CommonDefs.set_ptr_def Abs2_CommonDefs.set_ptr_def
    sim_norm_def)
  apply(rule sim_norm_ext_bind_bindS[OF get_ptr_ref_N])
  apply(clarsimp simp only: sim_norm_def[THEN symmetric] intro!: get_ptrs_refs_R_N)
  apply(clarsimp simp add: Let_def)
  apply(rule write_ptr_ref_R_N)
  apply(simp_all add: force_same_and_same force_same_and_diff)
  done

declare get_hashes_ref[ref_thm]
declare get_hashes_DegradedH_ref[ref_thm]
thm ll_read_ref

lemma vol_ops_ref[ref_thm]:
  "sim abs3_abs2_strel (PmmDefs.init_vars) (PmmDefs.init_vars)"
  "sim abs3_abs2_strel (PmmDefs.get_mode) (PmmDefs.get_mode)"
  "sim abs3_abs2_strel (PmmDefs.set_degraded) (PmmDefs.set_degraded)"
  "sim abs3_abs2_strel (PmmDefs.set_degraded_p) (PmmDefs.set_degraded_p)"
  "sim abs3_abs2_strel (PmmDefs.set_degraded_h) (PmmDefs.set_degraded_h)"
  "sim abs3_abs2_strel (PmmDefs.set_failed) (PmmDefs.set_failed)"
  "sim abs3_abs2_strel (PmmDefs.get_updated) (PmmDefs.get_updated)"
  "sim abs3_abs2_strel (PmmDefs.add_to_updated pind) (PmmDefs.add_to_updated pind)"
  "sim abs3_abs2_strel (PmmDefs.set_updated upd) (PmmDefs.set_updated upd)"
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(clarsimp_all simp add: abs3_abs2_strel_def brute_force split: split_if_asm)
  done

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]

declare
  sim_ext_bind_rt[ref_bind]
  sim_ext_try_catch_rt[ref_try_catch]
  sim_ext_exc_case_rt[ref_exc_case]
  sim_ext_if_rt[ref_if]
  sim_ext_if_full_rt[ref_if_full]
  sim_ext_mapM_rt[ref_mapM]

declare
  sim_ext_throw_diffI[ref_thm]
  sim_ext_return_diffI[ref_thm]

lemma read_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_CommonDefs.read ind inst) 
    (Abs2_CommonDefs.read ind inst) 
  "
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(auto simp add: pub_matches_htable2_def)
  done

lemma read_DegradedH_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_CommonDefs.read_DegradedH ind inst) 
    (Abs2_CommonDefs.read_DegradedH ind inst) 
  "
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(auto simp add: pub_matches_htable2_def)
  done

declare
  sim_return[ref_thm del]
  sims_throw(1)[ref_thm del]

declare
  sim_ext_bind_rt[ref_bind del]
  sim_ext_try_catch_rt[ref_try_catch del]
  sim_ext_exc_case_rt[ref_exc_case del]
  sim_ext_if_rt[ref_if del]
  sim_ext_if_full_rt[ref_if_full del]
  sim_ext_mapM_rt[ref_mapM del]

declare
  sim_ext_throw_diffI[ref_thm del]
  sim_ext_return_diffI[ref_thm del]

declare sim_return[ref_thm]
declare sim_norm_if_rt[ref_if]
declare
  sim_norm_return[ref_thm]
  sims_throw(1)[ref_thm]
  sims_throw(2)[ref_thm]
  sim_norm_chooseN[ref_thm]

declare
  sim_norm_bind_rt[ref_bind]
  sim_norm_try_catch_rt[ref_try_catch]
  sim_norm_exc_case_rt[ref_exc_case]
  sim_norm_if_rt[ref_if]
  sim_norm_mapM_rt[ref_mapM]

declare sim_sim_norm_bindI[ref_bind]
declare sim_norm_sim_try_catchI[ref_try_catch]
declare get_ptr_ref_N[ref_thm del]

lemma get_hashes_silence_deg_ref':
  "sim_ext abs3_abs2_strel
  (Abs3_CommonDefs.get_hashes_silence_deg)
  (Abs2_CommonDefs.get_hashes_silence_deg)
  (\<lambda>va vc. (if va = (pub_hashes2 vc)
            then UNIV \<times> {(svc, spc). ptr_hashes2 (hash spc) = ptr_hashes2 vc \<and>
                abs2_state.hstate spc \<noteq> LDead }
            else {}
           ) \<inter> abs3_abs2_strel)
  (force_same_and abs3_abs2_strel)
  (abs3_abs2_strel)
  "
  by(auto simp add:
    Abs3_CommonDefs.get_hashes_silence_deg_def Abs2_CommonDefs.get_hashes_silence_deg_def
    intro!: sim_ext_try_catch_sameI[OF get_hashes_ref] sim_ext_exceptions_caseI
    sim_ext_throw sim_ext_mono_norm[OF get_hashes_DegradedH_ref])

lemma sim_normN_ext_mono_concl:
  "\<lbrakk>sim_normN_ext R ma mc Sn; \<forall>x y. Sn x y \<subseteq> Sn' x y\<rbrakk> \<Longrightarrow> sim_normN_ext R ma mc Sn'"
  by(auto simp add: sim_normN_ext_def intro: simN_wpc_mono_concl del: subsetI)

declare get_ptr_ref_N[ref_thm del]
lemma get_ptr_ref_N'[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_CommonDefs.get_ptr b)
    (Abs2_CommonDefs.get_ptr b)
  "
  apply(auto simp add: sim_norm_def sim_norm_ext_decompose
    intro!: sim_normN_ext_mono_concl[OF get_ptr_ref_NN] get_ptr_ref_NE)
  apply(simp add: force_same_and_def)
  done

lemma get_sync_direction_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_CommonDefs.get_sync_direction pind) 
    (Abs2_CommonDefs.get_sync_direction pind) 
  "
  apply(simp only: 
    Abs3_CommonDefs.get_sync_direction_def 
    Abs2_CommonDefs.get_sync_direction_def 
  )
  apply(simp only: sim_norm_def)
  apply(rule sim_norm_ext_bindI[OF get_hashes_silence_deg_ref'[THEN sim_ext_imp_sim_norm_ext]])
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule_tac[!] sim_norm_ext_mono[where R=abs3_abs2_strel])
  apply(auto simp add: sim_norm_def[THEN symmetric] intro!: sims_all_intros sim_norm_return
    get_ptr_ref_N')
  done

lemma write_instance_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_CommonDefs.write_instance pind inst cnt)
    (Abs2_CommonDefs.write_instance pind inst cnt)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma sync_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_CommonDefs.sync pind inst cnt)
    (Abs2_CommonDefs.sync pind inst cnt)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma sim_norm_sim_if[ref_if]:
  "\<lbrakk>sim_norm R maT mcT; sim R maF mcF\<rbrakk>
    \<Longrightarrow> sim_norm R (if cond then maT else maF) (if cond then mcT else mcF)
  "
  by(auto intro: sim_imp_sim_norm)

lemma sim_sim_norm_if[ref_if]:
  "\<lbrakk>sim R maT mcT; sim_norm R maF mcF\<rbrakk>
    \<Longrightarrow> sim_norm R (if cond then maT else maF) (if cond then mcT else mcF)
  "
  by(auto intro: sim_imp_sim_norm)

declare [[ref_tac_store_non_gen]]

lemma sync_and_remove_from_ptr_silence_deg_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg to from ind)
    (Abs2_CommonDefs.sync_and_remove_from_ptr_silence_deg to from ind)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma check_pub_same_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_StartupDefs.check_pub_same ind)
    (Abs2_StartupDefs.check_pub_same ind)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma check_all_public_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    Abs3_StartupDefs.check_all_public 
    Abs2_StartupDefs.check_all_public 
  "
  apply(auto simp add:
    Abs3_StartupDefs.check_all_public_def Abs2_StartupDefs.check_all_public_def
    intro!: sims_all_intros sim_norm_mapM'
  )
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  done

definition abs2_startup2' where
  abs2_startup2'_def[monad_def]:
  "abs2_startup2' \<equiv> do {
    Abs2_StartupDefs.check_all_public;
    throw_mode_error EU;
    throw StartedExc
  }"

lemma startup2_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_StartupDefs.startup2)
    (abs2_startup2')
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma fetch_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_FetchDefs.fetch regs)
    (Abs2_FetchDefs.fetch regs)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

declare sim_chooseN[ref_thm]

declare Abs3_CommonDefs.sync_def[monad_def del]
declare Abs2_CommonDefs.sync_def[monad_def del]

lemma sync_if_not_SEq_ref[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.sync_if_not_SEq to from ind)
    (Abs2_UpdateDefs.sync_if_not_SEq to from ind)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma rollback_ref[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.rollback)
    (Abs2_UpdateDefs.rollback)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma write_secondary_ref[ref_thm]:
  "sim abs3_abs2_strel
    (Abs3_UpdateDefs.write_secondary regs ind)
    (Abs2_UpdateDefs.write_secondary regs ind)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

declare sim_sim_norm_try_catchI[ref_try_catch]

lemma pre_commit_ref[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.pre_commit regs)
    (Abs2_UpdateDefs.pre_commit regs)
  "
  apply(auto simp add: Abs3_UpdateDefs.pre_commit_def Abs2_UpdateDefs.pre_commit_def)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  done

lemma sync_if_not_SEq_remove_from_ptr_ref[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.sync_if_not_SEq_remove_from_ptr to from ind)
    (Abs2_UpdateDefs.sync_if_not_SEq_remove_from_ptr to from ind)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma rollforward_ref[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.rollforward)
    (Abs2_UpdateDefs.rollforward)
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma[ref_thm]:
  "sim R
    (PmmDefs.check_sizes upd_f)
    (PmmDefs.check_sizes upd_f)
  "
  apply(auto simp add: PmmDefs.check_sizes_def
    intro!: sims_all_intros)
  done

lemma update_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (Abs3_UpdateDefs.update regs)
    (Abs2_UpdateDefs.update regs)
  "
  by(tactic {* (refinement_tac false default_simp @{context}) 1 *})

lemma startup1_ref_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    Abs3_StartupDefs.startup1
    abs2_startup1'
  "
  by(auto simp add: sim_norm_decompose intro: startup1_ref_NN startup1_ref_NE)

declare plain_bindS_rt[ref_bindS]
declare try_catch_bindS_rt[ref_try_catch_bindS]
declare mapM_bindS_rt[ref_mapM_bindS]
declare sim_chooseN[ref_thm]
declare bind_bindS_rt[ref_bind_bindS]

declare vol_ops_ref[THEN sim_imp_sim_norm, ref_thm]
declare read_DegradedH_ref[THEN sim_imp_sim_norm, ref_thm]
declare read_ref[THEN sim_imp_sim_norm, ref_thm]
declare write_instance_ref[THEN sim_imp_sim_norm, ref_thm]
declare ll_read_ref[THEN sim_imp_sim_norm, ref_thm]
declare write_f_ref[THEN sim_imp_sim_norm, ref_thm]
declare write_secondary_ref[THEN sim_imp_sim_norm, ref_thm]

lemma sim_norm_ext_bind_bindS:
  "\<lbrakk>sim_norm_ext R ma1 mc1 Sn' Se';
    sim_norm_ext R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>x y. sim_norm_ext (R' x y)  (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S rha) (\<lambda>s. ((mc2 y)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se;
    \<forall>x y. Sn' x y \<subseteq> R' x y
  \<rbrakk> \<Longrightarrow>
    sim_norm_ext R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S rha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S rhc)
        Sn Se
  "
  apply(auto simp add: bind_projectR choice_bindS_bindN
    intro!: sim_norm_ext_choice_lockstep)
  apply(auto simp add: sim_norm_def sim_norm_ext_def bindS_projectNN bindS_projectNE
    sims_through_simN_simS simS_def force_same_and_def
    intro!: simN_wpc_bindN_bindS
    del: subsetI)
  apply(smt order_trans)+
  done

lemma sim_ext_sim_norm_bindS:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; sim_norm R ha hc; Sr \<subseteq> R
    \<rbrakk> \<Longrightarrow>
    sim_norm R (\<lambda>s. (ma\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mc\<Down>R) s \<guillemotright>=S hc)"
  apply(auto simp add: sim_norm_def sim_norm_ext_def sim_ext_def sim_def
    bindS_projectNN bindS_projectNE
    del: subsetI)
  apply(auto simp add: force_same_and_def intro!: simN_wpc_bindS del: subsetI)
  done

lemma sim_norm_if_bindS_diff_cond:
  "\<lbrakk> cond = cond';
     cond \<Longrightarrow> sim_norm R (\<lambda>s. (mat\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mct\<Down>R) s \<guillemotright>=S hc);
     \<not>cond \<Longrightarrow> sim_norm R (\<lambda>s. (maf\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mcf\<Down>R) s \<guillemotright>=S hc)
  \<rbrakk> \<Longrightarrow>
  sim_norm R
    (\<lambda>s. ((if cond then mat else maf)\<Down>R) s \<guillemotright>=S ha)
    (\<lambda>s. ((if cond' then mct else mcf)\<Down>R) s \<guillemotright>=S hc)
  "
  by(simp)

lemma get_sync_direction_ref_R_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_CommonDefs.get_sync_direction pind)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1) 
    (\<lambda>s. ((Abs2_CommonDefs.get_sync_direction pind)\<Down>R) s \<guillemotright>=S abs2_startup1') 
  "
  apply(simp only: 
    Abs3_CommonDefs.get_sync_direction_def 
    Abs2_CommonDefs.get_sync_direction_def 
    Let_def
  )
  apply(simp add: sim_norm_def)
  apply(rule sim_norm_ext_bind_bindS[where R'="\<lambda>x y.
    (if x = pub_hashes2 y then UNIV else {}) \<inter> abs3_abs2_strel"])
  apply(rule get_hashes_silence_deg_ref'[THEN sim_ext_imp_sim_norm_ext])
  apply(auto simp add: sim_norm_def[THEN symmetric] intro!: sim_exts_empty_rel)
  apply(rule sim_ext_sim_norm_bindS)
  apply(rule get_hashes_silence_deg_ref')
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})+
  done

lemma fetch_ref_R_N[ref_thm]:
  "
  sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_FetchDefs.fetch regs)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((Abs2_FetchDefs.fetch regs)\<Down>R) s \<guillemotright>=S abs2_startup1')
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma sim_sim_norm_bind_bindS:
  "\<lbrakk>
    sim R ma1 mc1;
    sim_norm R (\<lambda>s. (ma1\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. (mc1\<Down>R) s \<guillemotright>=S hc);
    \<forall>x. sim_norm R (\<lambda>s. ((ma2 x)\<Down>R) s \<guillemotright>=S ha) (\<lambda>s. ((mc2 x)\<Down>R) s \<guillemotright>=S hc)
  \<rbrakk> \<Longrightarrow>
    sim_norm R
        (\<lambda>s. ((ma1 \<guillemotright>= ma2)\<Down>R) s \<guillemotright>=S ha)
        (\<lambda>s. ((mc1 \<guillemotright>= mc2)\<Down>R) s \<guillemotright>=S hc)
  "
  by(drule sim_imp_sim_norm, rule sim_norm_bind_bindS, assumption+)

declare sim_sim_norm_bind_bindS[ref_bind_bindS]

lemma sync_and_remove_from_ptr_silence_deg_ref_R_N[ref_thm]:
  "sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg to from ind)\<Down>R) s
        \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((Abs2_CommonDefs.sync_and_remove_from_ptr_silence_deg to from ind)\<Down>R) s
        \<guillemotright>=S abs2_startup1')
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma startup2_ref_R_N:
  "sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_StartupDefs.startup2 :: ('a, 'c)abs3_monad)\<Down>R) s
        \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((abs2_startup2' :: ('a, pmm_vol_mem, 'c :: contents_cl)abs2_monad)\<Down>R) s \<guillemotright>=S abs2_startup1')
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma update_ref_R_N:
  "
  sim_norm abs3_abs2_strel
    (\<lambda>s. ((Abs3_UpdateDefs.update regs)\<Down>R) s \<guillemotright>=S Abs3_StartupDefs.startup1)
    (\<lambda>s. ((Abs2_UpdateDefs.update regs)\<Down>R) s \<guillemotright>=S abs2_startup1')
  "
  apply(auto simp add: Abs3_UpdateDefs.update_def Abs2_UpdateDefs.update_def Let_def)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  done

lemma abs2_startup':
  "Abs2_StartupDefs.startup = do {
    startup1';
    abs2_startup2'
  }"
  by(simp add: Abs2_StartupDefs.startup_def
    Abs2_StartupDefs.startup1_def Abs2_StartupDefs.startup2_def
    startup1'_def abs2_startup2'_def bind_assoc
  )

lemma update_full_ref:
  "sim_norm abs3_abs2_strel
    (<tryR>
        Abs3_UpdateDefs.update regs   
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
    (<tryR>
        Abs2_UpdateDefs.update regs   
    <catchR>
        rec_tryR (Abs2_StartupDefs.startup)
    )
  "
  by(auto simp add: Abs3_StartupDefs.startup_def abs2_startup'
    intro!: trans_ref_fin update_ref_N startup2_ref_N update_ref_R_N startup2_ref_R_N
    abs2_startup1'_fin)

lemma fetch_full_ref:
  "sim_norm abs3_abs2_strel
    (<tryR>
        Abs3_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
    (<tryR>
        Abs2_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs2_StartupDefs.startup)
    )
  "
  by(auto simp add: Abs3_StartupDefs.startup_def abs2_startup'
    intro!: trans_ref_fin fetch_ref_N startup2_ref_N fetch_ref_R_N startup2_ref_R_N
    abs2_startup1'_fin)

lemma abs3_startup_full_ref_N:
  "sim_norm abs3_abs2_strel
    (<tryR>
        Abs3_StartupDefs.startup
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
    (<tryR>
        Abs2_StartupDefs.startup
    <catchR>
        rec_tryR (Abs2_StartupDefs.startup)
    )
  "
  apply(auto simp add: Abs3_StartupDefs.startup_def abs2_startup')
  apply(rule trans_ref_fin_same)
  apply(auto intro!: trans_ref_fin_same Abs3_Ref.startup1_ref_N 
    Abs3_Ref.startup2_ref_N Abs3_Ref.startup2_ref_R_N Abs2_Startup_Lemmas.abs2_startup1'_fin)
  done

end

