theory Abs3_SpecRef 
  imports "../abs2/Abs2_Specification" Abs3_Specification Abs3_Ref "../SpecRef_Lemmas"
begin

fun ptr_by_inst_obs :: "INST \<Rightarrow> 'c abs2_obs_state \<Rightarrow> pub_indx set" where
  "ptr_by_inst_obs Primary = ptr1"
  | "ptr_by_inst_obs Secondary = ptr2"

fun ptr_state_by_inst_obs :: "INST \<Rightarrow> 'c abs2_obs_state \<Rightarrow> hw_state" where
  "ptr_state_by_inst_obs Primary sp = ptr1_state sp"
  | "ptr_state_by_inst_obs Secondary sp = ptr2_state sp"

fun ptr_matches_hash_obs :: "'c abs2_obs_state \<Rightarrow> INST \<Rightarrow> bool" where
  "ptr_matches_hash_obs sp inst =
        (ptr_state_by_inst_obs inst sp  \<noteq> IDead
            \<and> abs2_obs_state.hash sp \<noteq> None
            \<and> (ptr_hashes2 (the (hash sp)) inst = (ptr_by_inst_obs inst sp)))
  "

fun pinst_obs_state :: "'c abs2_obs_state \<Rightarrow> INST \<Rightarrow> ec_state" where
  "pinst_obs_state sp Primary = pinst_state' (ptr1_state sp) (ptr_matches_hash_obs sp Primary)"
  | "pinst_obs_state sp Secondary = pinst_state' (ptr2_state sp) (ptr_matches_hash_obs sp Secondary)"

declare disjCI[intro]
fun abs2_abs3_pi :: "('sv \<times> 'c abs2_obs_state) \<Rightarrow> ('sv \<times> 'c :: contents_cl abs3_obs_state)" 
  where
  "abs2_abs3_pi (svc, spc) = (svc, 
        \<lparr> 
            abs3_obs_state.hstate = abs2_obs_state.hstate spc
            , abs3_obs_state.ptr = if 
                    pinst_obs_state spc Primary \<noteq> EcOK 
                    \<and> pinst_obs_state spc Secondary \<noteq> EcOK
                then None
                else Some 
                    (if pinst_obs_state spc Primary = EcOK 
                        then abs2_obs_state.ptr1 spc
                        else abs2_obs_state.ptr2 spc
                    )
            , abs3_obs_state.ptr_state = 
                combine_ec_states (pinst_obs_state spc Primary) (pinst_obs_state spc Secondary)
            , abs3_obs_state.memory = abs2_obs_state.memory spc
            , abs3_obs_state.hw_state = abs2_obs_state.hw_state spc
            , abs3_obs_state.pub_hashes = Option.map pub_hashes2 (hash spc)
        \<rparr>
  )"

lemma abs2_startup1'_SN_incl:
  "\<forall>sc\<in>abs2_init.
  ((Abs2_Startup_Lemmas.startup1' \<guillemotright>= 
    (\<lambda>_. abs2_startup2' :: (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad))\<Down>SN) sc
  \<subseteq> ((abs2_startup1' \<guillemotright>= 
    (\<lambda>_. abs2_startup2' :: (unit, pmm_vol_mem, 'c :: contents_cl)abs2_monad))\<Down>SN) sc"
  apply(rule ballI)
  apply(simp add: projectSN_def)
  apply(intro projectS_mono fin_imp_bind_NN_incl abs2_startup1'_fin)
  done


declare disjCI[rule del]

lemma Collect_Un:
  "({y. P y} \<union> {y. Q y}) = {y. P y \<or> Q y}"
  by(auto)

   
lemma abs3_state_equality:
  "((x :: 'c abs3_state) = (y :: 'c abs3_state)) = (
    abs3_state.hstate x = abs3_state.hstate y
    \<and> abs3_state.ptr x = abs3_state.ptr y
    \<and> abs3_state.ptr_state x = abs3_state.ptr_state y
    \<and> abs3_state.memory x = abs3_state.memory y
    \<and> abs3_state.hw_state x = abs3_state.hw_state y
    \<and> abs3_state.pub_hashes x = abs3_state.pub_hashes y
  )"
  by(case_tac x, case_tac y, simp)

lemma ex_abs3_state:
  "(\<exists>b. P b) = 
    (\<exists>hstate ptr ptr_state memory hw_state pub_hashes.
        P \<lparr> abs3_state.hstate = hstate, 
            abs3_state.ptr = ptr, abs3_state.ptr_state = ptr_state, 
            abs3_state.memory = memory, 
            abs3_state.hw_state = hw_state,
            abs3_state.pub_hashes = pub_hashes
    \<rparr>)"
  apply(auto)
  apply(case_tac b, force)
  done

fun abs2_abs3_int_map :: "('sv \<times> 'c abs2_state) \<Rightarrow> ('sv \<times> 'c :: contents_cl abs3_state)" 
  where
  "abs2_abs3_int_map (svc, spc) = (svc, 
        \<lparr> 
            abs3_state.hstate = abs2_state.hstate spc
            , abs3_state.ptr = if 
                    pinst_state spc Primary \<noteq> EcOK 
                    \<and> pinst_state spc Secondary \<noteq> EcOK
                then undefined
                else 
                    (if pinst_state spc Primary = EcOK 
                        then abs2_state.ptr1 spc
                        else abs2_state.ptr2 spc
                    )
            , abs3_state.ptr_state = 
                combine_ec_states (pinst_state spc Primary) (pinst_state spc Secondary)
            , abs3_state.memory = abs2_state.memory spc
            , abs3_state.hw_state = abs2_state.hw_state spc
            , abs3_state.pub_hashes = pub_hashes2 (abs2_state.hash spc)
        \<rparr>
  )"

lemma abs2_init_NN_incl:
  "set_fun_into_rel (abs2_startup1'\<Down>SNN) `` abs2_init
    \<subseteq> abs3_abs2_strel `` set_fun_into_rel (Abs3_StartupDefs.startup1\<Down>SNN) `` abs3_init"
  apply(simp_all add: projectSNN_def projectSNE_def projectS_def abs2_startup1'_def 
    Abs3_StartupDefs.startup1_def
    project_simps
    compose_project)
  apply(clarsimp simp add: set_fun_into_rel_def Image_def projectS_defs image_def)
  apply(rename_tac svc spc svc' spc')
  apply(simp add: startup1'_NN_def choice_def)
  apply(iterative_simp simp add: abs2_ptr_same_rest_def abs2_init_def IOK_less_eq)+
  apply(clarsimp_all simp add: abs3_abs2_strel_def Abs3_StartupDefs.startup1_NN_def Collect_Un
    brute_force abs3_init_def)
  apply(rule_tac[!] x="snd (abs2_abs3_int_map (undefined, spc'))" in exI)
  apply(clarsimp_all)
  apply(simp_all add: HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc)
  apply(case_tac[!] spc')
  apply(clarsimp_all simp add: ex_abs3_state
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    conj_commute conj_left_commute state_ineq_simps
    comb_hstate_EcOK_not_LOK combine_ec_states.simps not_IOK_is_le_IUnwr
    pinst_state_False_notOK
    comb_hstate_EcOK_LDegraded
    pinst_state_False_notOK
  )
  apply(smt comb_hstate_EcOK_LDegraded(2) hw_state.distinct(1) not_IOK_is_le_IUnwr pinst_state_False_notOK
    comb_hstate_EcOK_LDegraded(3) combine_ec_states.simps(1) combine_ec_states.simps(2) 
    combine_ec_states.simps(3)
    hw_state.distinct(1) ec_state.distinct(11) 
    less_eq_hw_state.simps(4) less_eq_log_state.simps(4) less_eq_log_state.simps(7)
    log_state.distinct(5) Abs3_Ref.abs3_ref_simps(10) min_absorb1 pinst_state'.simps(3) 
    pinst_state'.simps(4)
  )+
  done

lemma pinst_state'_EcOK:
  "(pinst_state' x y = EcOK) = (x = IOK \<and> y)"
  by(case_tac x, case_tac[!] y, simp_all)

lemma abs2_init_NE_incl:
  "set_fun_into_rel (abs2_startup1'\<Down>SNE) `` abs2_init
    \<subseteq> abs3_abs2_strel `` set_fun_into_rel (Abs3_StartupDefs.startup1\<Down>SNE) `` abs3_init"
  apply(simp_all add: projectSNE_def projectS_def abs2_startup1'_def 
    Abs3_StartupDefs.startup1_def
    project_simps
    compose_project)
  apply(clarsimp simp add: set_fun_into_rel_def Image_def projectS_defs image_def)
  apply(rename_tac svc spc svc' spc')
  apply(simp add: startup1'_NE_def)
  apply(iterative_simp simp add: abs2_ptr_same_rest_def abs2_init_def IOK_less_eq)+
  apply(clarsimp_all simp add: abs3_abs2_strel_def Abs3_StartupDefs.startup1_NE_def Collect_Un
    pinst_state'_EcOK 
    combine_ec_states_Dead_False combine_ec_states_False_Dead combine_ec_states_2_False
    brute_force abs3_init_def)
  apply (metis abs3_state.select_convs(1) abs3_state.select_convs(4) abs3_state.select_convs(5) 
    abs3_state.select_convs(6) less_eq_log_state.simps(1))+
  done

theorem abs2_refines_abs3:
  "refines abs3_abs2_strel abs2_abs3_pi abs3_spec abs2_spec"
proof(intro refinesI)
  show "obs_consistent abs3_abs2_strel abs2_abs3_pi abs3_spec abs2_spec"
    apply(clarsimp simp add: obs_consistent_def abs3_spec_def abs2_spec_def 
      abs3_abs2_strel_def obs_ts)
    apply(intro conjI impI allI)
    apply(clarsimp_all simp add: combine_ec_states_2_False pinst_state_False_notOK 
      combine_ec_states_is_LDead)
    done
next
  show "PO_refines abs3_abs2_strel abs3_spec abs2_spec"
  proof(rule refine_basic)
    show "init_trans abs2_spec `` init abs2_spec
        \<subseteq> abs3_abs2_strel `` init_trans abs3_spec `` init abs3_spec"
      apply(simp add: abs2_spec_def abs2_ts_def abs3_spec_def abs3_ts_def TS.defs spec.defs)
      apply(simp only: Abs3_StartupDefs.startup_def abs2_startup')
      apply(rule init_conc_mono_cond[where 
        mc'="((abs2_startup1' \<guillemotright>= (\<lambda>_. abs2_startup2'))\<Down>SN)", OF _ abs2_startup1'_SN_incl])
      apply(intro sim_norm_init_bind_inclusion[where R="abs3_abs2_strel"]
        abs2_init_NN_incl abs2_init_NE_incl Abs3_Ref.startup2_ref_N[simplified sim_norm_def])
      done
  next
    show "{abs3_abs2_strel} TS.trans abs3_spec, TS.trans abs2_spec {> abs3_abs2_strel}"
      apply(simp add: abs3_spec_def abs3_ts_def abs2_spec_def abs2_ts_def spec.defs TS.defs 
        abs3_trans_def abs2_trans_def)
      apply(intro relhoare_UN_lockstep allI relhoare_set_fun_choice_lockstep)
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare simp only: sim_norm_def[THEN symmetric]
        try_rec_tryR_def
        intro!: Abs3_Ref.fetch_full_ref
        Abs3_Ref.update_full_ref

      )
      apply(auto intro!: sim_norm_try_catchRI trans_ref_fin_catchR
        Abs3_Ref.restart_mangle_sv_refines_itself 
        simp add: try_rec_tryR_def
        Abs3_StartupDefs.startup_def abs2_startup'
        intro!:
        abs2_startup1'_fin
        Abs3_Ref.startup1_ref_N Abs3_Ref.startup2_ref_N startup2_ref_R_N
      )
      done
  qed
qed

theorem abs2_implements_abs3:
  "implements abs2_abs3_pi abs3_spec abs2_spec"
  by(auto intro: refinement_soundness abs2_refines_abs3)

end
