theory Abs4_CommonDefs
  imports Abs4Defs
begin

definition sync_success :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('c, 'c)abs4_monad" where
  "sync_success ind to from \<equiv> \<lambda>(sv, sp).
    {(Some (Inr cnt), sv, sp')|sp' cnt.
        (ind, from) \<in> reliable sp
        \<and> (ind, to) \<in> writable sp
        \<and> cnt = memory sp ind from
        \<and> sp' = sp\<lparr>
            arrow := (arrow sp)(ind := SEq),
            memory := fun_upd_second (memory sp) ind to cnt,
            reliable := insert (ind, to) (reliable sp)
        \<rparr>
    }"

definition sync_degrade
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "sync_degrade ind to from \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded o EC) cnt, sv', sp')|sv' sp' cnt cnt' unr unwr a_st.
        (ind, from) \<in> reliable sp
        \<and> (ind, to) \<in> unwr
        \<and> cnt = memory sp ind from
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> (unr = {(ind, to)} \<or> (unr = {} \<and> cnt' = memory sp ind to))
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            writable := writable sp - unwr,
            memory := fun_upd_second (memory sp) ind to cnt',
            reliable := reliable sp - unr,
            admin_state := a_st
        \<rparr>
    }
  "

definition sync_degrade_write
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('a, 'c)abs4_monad" where
  "sync_degrade_write ind to from \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded o EC) cnt, sv', sp')|sv' sp' cnt unwr a_st.
        (ind, from) \<in> reliable sp
        \<and> (ind, to) \<in> writable sp
        \<and> (ind, to) \<in> unwr
        \<and> cnt = memory sp ind from
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := (arrow sp)(ind := SEq),
            writable := writable sp - unwr,
            memory := fun_upd_second (memory sp) ind to cnt,
            reliable := insert (ind, to) (reliable sp),
            admin_state := a_st
        \<rparr>
    }
  "

definition sync_fail
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "sync_fail ind to from \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' unr unwr arr_del m' a_st.
        (ind, from) \<in> unr
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> a_st \<le> admin_state sp
        \<and> (\<forall>inst. size (m' inst) \<le> capacity (Public ind))
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            memory := (memory sp)(ind := m'),
            admin_state := a_st
        \<rparr>
    }
  "

definition sync_restart
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "sync_restart ind to from \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp' unr unwr cnt' cnt'' arr_del a_st.
        ((ind, from) \<in> unr \<or> cnt' = memory sp ind from)
        \<and> ((ind, to) \<in> unr \<or> cnt'' = memory sp ind to)
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> size cnt'' \<le> capacity (Public ind)
        \<and> (unwr = {} \<or> (ind, to) \<in> unwr \<or> (ind, from) \<in> unwr)
        \<and> a_st \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            memory := (memory sp)(ind := (\<lambda>inst. if inst = from then cnt' else cnt'')),
            admin_state := a_st
        \<rparr>
    }
  "

definition sync_restart_write
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('a, 'c)abs4_monad" where
  "sync_restart_write ind to from \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp' unwr a_st.
        (ind, from) \<in> reliable sp
        \<and> (ind, to) \<in> writable sp
        \<and> (unwr = {} \<or> (ind, to) \<in> unwr \<or> (ind, from) \<in> unwr)
        \<and> a_st \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            arrow := (arrow sp)(ind := SEq),
            reliable := insert (ind, to) (reliable sp),
            writable := writable sp - unwr,
            memory := fun_upd_second (memory sp) ind to (memory sp ind from),
            admin_state := a_st
        \<rparr>
    }
  "

definition sync
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs4_monad"
  where
  "sync ind to from \<equiv>
    sync_success ind to from
    [+] sync_degrade ind to from
    [+] sync_degrade_write ind to from
    [+] sync_fail ind to from
    [+] sync_restart ind to from
    [+] sync_restart_write ind to from
  "

lemmas sync_defs =
  sync_def
  sync_success_def
  sync_degrade_def
  sync_degrade_write_def
  sync_fail_def
  sync_restart_def
  sync_restart_write_def

definition just_sync
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> (unit, 'c :: size)abs4_monad" 
   where just_sync_def[monad_def]:
  "just_sync ind to from \<equiv> sync ind to from \<guillemotright> return ()"
        
definition read_success
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c)abs4_monad" where
  "read_success ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inr cnt), sv, sp)|cnt.
        (ind, inst) \<in> reliable sp
        \<and> cnt = memory sp ind inst
    }
  "

definition read_corrupt
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "read_corrupt ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailRead), sv, sp')|sp' cnt'.
        size cnt' \<le> capacity (Public ind)
        \<and> sp' = sp\<lparr>
            reliable := reliable sp - {(ind, inst)},
            memory := fun_upd_second (memory sp) ind inst cnt'
        \<rparr>
    }
  "

definition read_degrade
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs4_monad" where
  "read_degrade ind inst \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded) EU, sv', sp')|sv' sp' unwr cnt' a_st.
        (ind, inst) \<in> unwr
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := reliable sp - {(ind, inst)},
            writable := writable sp - unwr,
            admin_state := a_st
        \<rparr>
    }
  "

definition read_degrade_reliable
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c)abs4_monad" where
  "read_degrade_reliable ind inst \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded) (EC cnt), sv', sp')|sv' sp' unwr cnt a_st.
        (ind, inst) \<in> reliable sp
        \<and> (ind, inst) \<in> unwr
        \<and> (ind, other_i inst) \<in> unwr
        \<and> cnt = memory sp ind inst
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            writable := writable sp - unwr,
            admin_state := a_st
        \<rparr>
    }
  "

definition focused_adm_fail
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "focused_adm_fail ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' cnt' unwr unr arr_del.
        (ind, inst) \<in> unr
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            admin_state := LDead 
        \<rparr>
  }"

definition read :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs4_monad" where
  "read ind inst \<equiv>
    read_success ind inst
    [+] read_corrupt ind inst
    [+] read_degrade ind inst
    [+] read_degrade_reliable ind inst
    [+] focused_adm_fail ind inst
    [+] restart_mangle_sv
    [+] do { read_success ind inst; restart_mangle_sv }
    [+] (<try> read_corrupt ind inst <catch> K restart_mangle_sv)
    [+] (<try> read_degrade ind inst <catch> K restart_mangle_sv)
    [+] (<try> read_degrade_reliable ind inst <catch> K restart_mangle_sv)
    [+] (<try> focused_adm_fail ind inst <catch> K restart_mangle_sv)
  "

lemmas read_defs =
  read_def
  read_success_def
  read_corrupt_def
  read_degrade_def
  read_degrade_reliable_def
  focused_adm_fail_def

definition focused_fail'
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "focused_fail' ind inst \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' cnt' unwr unr arr_del a_st.
        (ind, inst) \<in> unr
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> a_st \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            admin_state := a_st
        \<rparr>
  }"

definition read_or_fail
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('c, 'c :: size)abs4_monad" where
  "read_or_fail ind inst \<equiv>
    read_success ind inst
    [+] read_degrade_reliable ind inst
    [+] focused_fail' ind inst
    [+] restart_mangle_sv
    [+] (<try> read_degrade_reliable ind inst <catch> K restart_mangle_sv)
    [+] (<try> focused_fail' ind inst <catch> K restart_mangle_sv)
  "

lemmas read_or_fail_defs =
  read_or_fail_def
  read_success_def
  read_degrade_reliable_def
  focused_fail'_def

fun point_arrow_to :: "INST \<Rightarrow> sync_direction" where
  "point_arrow_to Primary = SL"
  | "point_arrow_to Secondary = SR"

definition write_new_success
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "write_new_success ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv', sp')|sv' sp' arr.
        (ind, inst) \<in> writable sp
        \<and> size cnt \<le> capacity (Public ind)
        \<and> arr = (if cnt = memory sp ind (other_i inst) then SEq else point_arrow_to inst)
        \<and> sv' = sv\<lparr> abs_updated := insert ind (abs_updated sv) \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := (arrow sp)(ind := arr),
            memory := fun_upd_second (memory sp) ind inst cnt,
            reliable := insert (ind, inst) (reliable sp)
        \<rparr>
    }
  "

definition write_new_degrade
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c  \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "write_new_degrade ind inst cnt \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded) (EB False), sv', sp')|sv' sp' cnt' unr unwr a_st.
        (ind, inst) \<in> unwr
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> unr = {(ind, inst)}
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            admin_state := a_st
        \<rparr>
    }
  "

definition write_new_degrade_correct
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c  \<Rightarrow> ('a, 'c :: size)abs4_monad" where
  "write_new_degrade_correct ind inst cnt \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded) (EB True), sv', sp')|sv' sp' unwr arr a_st cnt'.
        (ind, inst) \<in> unwr
        \<and> size cnt' \<le> capacity (Public ind)
        \<and> arr = (if cnt' = memory sp ind (other_i inst) then SEq else point_arrow_to inst)
        \<and> a_st \<le> admin_state sp
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> ((ind, inst) \<in> reliable sp \<and> cnt' = memory sp ind inst \<or> cnt' = cnt)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := (arrow sp)(ind := arr),
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := insert (ind, inst) (reliable sp),
            writable := writable sp - unwr,
            admin_state := a_st
        \<rparr>
    }
  "

definition write_new_corrupt_restart
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "write_new_corrupt_restart ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp' cnt'.
        size cnt' \<le> capacity (Public ind)
        \<and> sp' = sp\<lparr>
            memory := fun_upd_second (memory sp) ind inst cnt',
            reliable := reliable sp - {(ind, inst)}
        \<rparr>
    }
  "

definition write_new_too_large
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "write_new_too_large ind inst cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (InvalidInput)), sv, sp')|sp'.
        capacity (Public ind) < size cnt
        \<and> sp' = sp
    }"

definition write_new
  :: "pub_indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "write_new ind inst cnt \<equiv>
    write_new_success ind inst cnt
    [+] write_new_degrade ind inst cnt
    [+] write_new_degrade_correct ind inst cnt
    [+] focused_adm_fail ind inst
    [+] write_new_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_new_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_new_degrade ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_new_degrade_correct ind inst cnt <catch> K restart_mangle_sv)
    [+] write_new_corrupt_restart ind inst cnt
  "

lemmas write_new_defs =
  write_new_def
  write_new_success_def
  write_new_degrade_def
  write_new_degrade_correct_def
  write_new_corrupt_restart_def
  write_new_too_large_def
  focused_adm_fail_def

definition get_arrow_success :: "pub_indx \<Rightarrow> (sync_direction, 'c)abs4_monad" where
  "get_arrow_success ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inr dir), sv, sp)|dir.
        dir = arrow sp ind
        \<and> dir \<noteq> SBot
    }"

definition get_arrow_degrade :: "pub_indx \<Rightarrow> (sync_direction, 'c)abs4_monad" where
  "get_arrow_degrade ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inr dir), sv', sp')|sv' sp' dir unwr.
        dir = arrow sp ind
        \<and> LDegraded \<le> admin_state sp 
        \<and> dir \<noteq> SBot
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            admin_state := LDegraded,
            writable := writable sp - unwr
        \<rparr>
  }"

definition get_arrow_fail :: "pub_indx \<Rightarrow> ('a, 'c)abs4_monad" where
  "get_arrow_fail ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' arr_del unr unwr.
        ind \<in> arr_del
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            writable := writable sp - unwr,
            reliable := reliable sp - unr,
            admin_state := LDead
        \<rparr>
  }"

definition get_arrow_restart :: "pub_indx \<Rightarrow> ('a, 'c)abs4_monad" where
  "get_arrow_restart ind \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp' arr_del unr unwr a_st.
        a_st \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            writable := writable sp - unwr,
            reliable := reliable sp - unr,
            admin_state := a_st
        \<rparr>
  }"


definition get_arrow :: "pub_indx \<Rightarrow> (sync_direction, 'c :: size)abs4_monad" where
  "get_arrow ind \<equiv>
    get_arrow_success ind
    [+] get_arrow_degrade ind
    [+] get_arrow_fail ind
    [+] restart_mangle_sv
    [+] get_arrow_restart ind
  "

lemmas get_arrow_defs =
  get_arrow_def
  get_arrow_success_def
  get_arrow_degrade_def
  get_arrow_fail_def
  get_arrow_restart_def

definition maybe_mess_up_admin
  :: "pub_indx \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "maybe_mess_up_admin pind \<equiv>
    return ()
    [+] (get_arrow_degrade pind \<guillemotright> return ())
    [+] get_arrow_fail pind
    [+] get_arrow_restart pind
  "

definition just_sync_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c :: size)abs4_monad"
  where just_sync_silence_deg_def[monad_def]:
  "just_sync_silence_deg to from ind \<equiv>
    <try>
        just_sync ind to from
    <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "
 
definition just_sync_silence_deg'
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c :: size)abs4_monad"
  where just_sync_silence_deg_def'[monad_def]:
  "just_sync_silence_deg' to from ind \<equiv>
    <try> do {
        sync ind to from;
        maybe_mess_up_admin ind
    } <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "
  
end