theory Abs4_Ref
imports
  "../Composition"
  "../Pmm_Refinement"
  "../Misc_Lemmas"
  "../Pmm_Lemmas"
  "../abs3/Abs3_CommonDefs"
  "../abs3/Abs3_Lemmas"
  "../abs3/Abs3_UpdateDefs"
  "../abs3/Abs3_StartupDefs"
  "../abs3/Abs3_Invariant_Lemmas"
  Abs4_CommonDefs
  Abs4_UpdateDefs
  Abs4_FetchDefs
  Abs4_StartupDefs
  Abs4_Lemmas
  Abs4_Invariant_Lemmas
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs
  force_same_and_def
  Abs3_LLInterface.write_f_defs
  Abs3_LLInterface.ll_read_defs

definition abs3_reliable :: "'c abs3_state \<Rightarrow> (pub_indx \<times> INST)set" where
  "abs3_reliable spc \<equiv> {(ind, inst).
    hw_state spc ind inst \<noteq> IDead
    \<and> hstate spc \<noteq> LDead
    \<and> pub_hashes spc (ind, inst) = Hash (abs3_state.memory spc ind inst)
  }
  "

definition abs3_writable :: "'c abs3_state \<Rightarrow> (pub_indx \<times> INST)set" where
  "abs3_writable spc \<equiv> {(ind, inst).
    hw_state spc ind inst = IOK
    \<and> hstate spc = LOK
  }
  "

definition abs3_arrow :: "'c abs3_state \<Rightarrow> pub_indx \<Rightarrow> sync_direction" where
  "abs3_arrow spc \<equiv> (\<lambda>ind.
    if hstate spc = LDead \<or> ptr_state spc = LDead
    then SBot
    else
        if pub_hashes spc (ind, Primary) = pub_hashes spc (ind, Secondary)
        then SEq
        else
            if ind \<in> ptr spc
            then SL
            else SR
  )"


fun abs4_card_mode :: "card_mode \<Rightarrow> abs_card_mode" where
  "abs4_card_mode CDegradedH = ACDegraded"
  | "abs4_card_mode CDegradedP = ACDegraded"
  | "abs4_card_mode CDegraded = ACDegraded"
  | "abs4_card_mode CNormal = ACNormal"
  | "abs4_card_mode CFail = ACFail"

definition abs4_abs3_strel ::"((abs_vol_mem \<times> 'c abs4_state) \<times> (pmm_vol_mem \<times> 'c :: size abs3_state))set" 
  where
  "abs4_abs3_strel \<equiv> {((sva, spa), (svc, spc)). 
    sva = \<lparr> abs_card_mode = abs4_card_mode (PmmDefs.card_mode svc), abs_updated = updated svc \<rparr>
    \<and> spa = \<lparr>
        arrow = abs3_arrow spc,
        reliable = abs3_reliable spc,
        writable = abs3_writable spc,
        memory = abs3_state.memory spc,
        admin_state = min (hstate spc) (ptr_state spc)
    \<rparr>
    \<and> (\<forall>ind inst. size (abs3_state.memory spc ind inst) \<le> capacity (Public ind))
  }"

lemmas abs4_abs3_strel_defs =
  abs4_abs3_strel_def
  abs3_reliable_def
  abs3_writable_def
  abs3_arrow_def

definition hashes_ok where
  "hashes_ok \<equiv> {(svc, spc). hstate spc = LOK }"

lemma abs3_reliableI:
  "\<lbrakk> hstate spc \<noteq> LDead; hw_state spc ind inst \<noteq> IDead;
    pub_hashes spc (ind, inst) = Hash (abs3_state.memory spc ind inst) \<rbrakk> \<Longrightarrow>
  (ind, inst) \<in> abs3_reliable spc
  "
  by(auto simp add: abs3_reliable_def)

lemma abs3_writableI:
  "\<lbrakk> hstate spc = LOK; hw_state spc ind inst = IOK \<rbrakk> \<Longrightarrow>
  (ind, inst) \<in> abs3_writable spc
  "
  by(auto simp add: abs3_writable_def)

lemma set_ref_simps:
  "(\<exists>A. x \<in> A \<and> A \<inter> B = {x}) = (x \<in> B)"
  "(\<exists>A. x \<in> A \<and> B - A = B) = (x \<notin> B)"
  "(\<exists>A. x \<in> A \<and> B \<subseteq> A) = True"
  "(\<exists>A. B \<subseteq> A \<and> x \<in> A) = True"
  "(\<exists>A. x \<in> A \<and> B - A = B - {x}) = True"
  "(\<exists>A. B \<subseteq> A) = True"
  "(\<exists>A. A \<subseteq> B) = True"
  "(\<exists>A. x \<in> A \<and> A \<subseteq> B) = (x \<in> B)"
  "(\<exists>A. x \<in> A \<and> B - {x} = B - A) = True"
  "(\<forall>A. \<not> B \<subseteq> A) = False"
  "(Ex (op \<subseteq> A)) = True"
  "(Ex (op \<in> x)) = True"
  "(A - B = A) = (B \<subseteq> -A)"
  "(A = A - B) = (B \<subseteq> -A)"
  "(x \<notin> A - B) = (x \<notin> A \<or> x \<in> B)"
  "(\<exists>A. A \<inter> B = C) = (C \<subseteq> B)"
  "(\<exists>A. B \<inter> A = C) = (C \<subseteq> B)"
  "({} = A - B) = (A \<subseteq> B)"
  "(\<exists>A. B - A = B - {x})"
  "(insert x A = A) = (x \<in> A)"
  "(A = insert x A) = (x \<in> A)"
  "(A \<inter> B = {}) = (A \<subseteq> -B)"
  "(\<exists>A\<subseteq> -B. x \<in> A) = (x \<notin> B)"
  "(\<exists>A\<subseteq> -B. x \<in> A \<and> y \<in> A) = (x \<notin> B \<and> y \<notin> B)"
  "(\<exists>A. B - A = B - {x} \<and> x \<in> A) = True"
  "(\<exists>A. B - {x} = B - A \<and> x \<in> A) = True"
  "(\<exists>A. B - A = B - C) = True"
  "(\<exists>A. B - C = B - A) = True"
  "(\<exists>A. insert x B = B - A) = (x \<in> B)"
  "({x} \<inter> A = {x}) = (x \<in> A)"
  "(\<exists>A. x \<in> A \<and> A \<subseteq> - B \<and> x \<notin> B) = (x \<notin> B)"
  "(\<exists>A. B \<subseteq> A \<and> x \<in> A \<and> y \<in> A) = True"
  "(\<exists>A. x \<in> A \<and> y \<in> A) = True"
  by(auto)

lemma abs3_state_ref_simps:
  "((spc\<lparr> hw_state := fun_upd_second (hw_state spc) ind inst hws\<rparr>) = spc) =
    (hw_state spc ind inst = hws)"
  "(spc = (spc\<lparr> hw_state := fun_upd_second (hw_state spc) ind inst hws\<rparr>)) =
    (hw_state spc ind inst = hws)"

  "(spc\<lparr>hstate := hs, ptr_state := ps\<rparr> = spc\<lparr>hstate := hs', ptr_state := ps'\<rparr>)
    = (hs = hs' \<and> ps = ps')"
  "hstate spc = hs \<Longrightarrow> spc \<lparr> hstate := hs \<rparr> = spc"
  "ptr_state spc = ps \<Longrightarrow> spc \<lparr> ptr_state := ps \<rparr> = spc"
  "(sp \<lparr> pub_hashes := ht, hstate := hs \<rparr> = sp \<lparr> pub_hashes := ht', hstate := hs' \<rparr>)
    = (ht = ht' \<and> hs = hs')"
  "(sp \<lparr> hstate := hs \<rparr> = sp) = (hstate sp = hs)"
  "(sp = sp \<lparr> hstate := hs \<rparr>) = (hstate sp = hs)"
  "(sp = sp\<lparr>ptr_state := ps, hstate := hs\<rparr>) = (ptr_state sp = ps \<and> hstate sp = hs)"
  "(sp\<lparr>ptr_state := ps, hstate := hs\<rparr> = sp) = (ptr_state sp = ps \<and> hstate sp = hs)"
  "(ptr_state sp \<noteq> ptr_state sp') \<Longrightarrow> ((sp = sp') = False)"
  "(spc\<lparr>hstate := hs, hw_state := fun_upd_second (hw_state spc) ind inst hws \<rparr> = spc\<lparr>hstate := hs\<rparr>)
   = (hw_state spc ind inst = hws)"
  "(spc\<lparr>abs3_state.memory :=
      fun_upd_second (abs3_state.memory spc) ind inst a,
      hstate := LDegraded,
      hw_state := fun_upd_second (hw_state spc) ind inst hws\<rparr> =
  spc\<lparr>abs3_state.memory :=
  fun_upd_second (abs3_state.memory spc) ind inst a,
      hstate := LDegraded\<rparr>)
  = (hw_state spc ind inst = hws)"
  "
  (spc\<lparr>hstate := LDegraded,
    abs3_state.memory := fun_upd_second (abs3_state.memory spc) ind inst cnt \<rparr> =
    spc\<lparr>hstate := LDegraded\<rparr>)
  = (cnt = abs3_state.memory spc ind inst)
  "
  "hstate spc \<noteq> hstate spc' \<Longrightarrow> (spc = spc') = False"
  "hs = hstate spc \<Longrightarrow> spc\<lparr> hstate := hs \<rparr> = spc"
  apply(auto dest:
    arg_cong[where f=hstate] arg_cong[where f=ptr_state] arg_cong[where f=pub_hashes]
    arg_cong[where f=hw_state]
    arg_cong[where f="\<lambda>sp. hw_state sp ind inst"]
    arg_cong[where f="\<lambda>sp. abs3_state.memory sp ind inst"]
  )
  done

lemma inst_state_ref_simps:
  "\<not>x \<le> IUnwr \<Longrightarrow> x \<noteq> IDead"
  by(case_tac x, simp_all)

lemma log_state_ref_simps:
  "(LOK \<le> x) = (x = LOK)"
  "(x \<le> LOK) = True"
  "LDegraded \<le> x \<Longrightarrow> x \<noteq> LDead"
  "(x \<le> LDegraded) \<Longrightarrow> (x \<noteq> LOK)"
  "(x \<le> LDead) = (x = LDead)"
  "x \<noteq> LDead \<Longrightarrow> LDegraded \<le> x"
  "(min LOK x = min LDegraded x) = (x \<le> LDegraded)"
  "(min LDegraded x = min LOK x) = (x \<le> LDegraded)"
  by(case_tac[!] x, clarsimp_all simp add: min_def)

lemma override_on_ref_simps:
  "(\<exists>A. override_on f g A = g)"
  "(\<exists>A. g = override_on f g A)"
  "(\<exists>A. override_on f g A = f)"
  "(\<exists>A. f = override_on f g A)"
  "\<exists>A. x \<in> A \<and> override_on f g A = g" 
  by(auto simp add: override_on_def intro: exI[where x=UNIV] exI[where x="{}"])
 
lemma misc_ref_simps:
  "(\<forall>x. x \<noteq> y) = False"
  "(\<forall>x. y \<noteq> x) = False"
  "(\<exists>v. f(x := v) = f)"
  "(\<exists>v. f(x := v) = f) = True"
  "(\<exists>x. (x = a \<longrightarrow> P x) \<and> (x \<noteq> a \<longrightarrow> Q x)) = (P a \<or> (\<exists>y. y \<noteq> a \<and> Q y))"
  by auto

lemma bool_ref_simps:
  "(Ex Not) = True"
  "(All Not) = False"
  "(\<exists>x. x) = True"
  "\<exists>x. (y \<and> x) = y"
  "\<exists>x. (x \<and> y) = y"
  by auto

lemma pmm_vol_mem_ref_simps:
  "(sv\<lparr> card_mode := x \<rparr> = sv) = (card_mode sv = x)"
  "(sv = sv\<lparr> card_mode := x \<rparr>) = (card_mode sv = x)"
  "(sv \<lparr> card_mode := x \<rparr> = sv \<lparr> card_mode := x' \<rparr>) = (x = x')"
  "((sv :: pmm_vol_mem) \<lparr> card_mode := x \<rparr> = (sv' :: pmm_vol_mem) \<lparr> card_mode := x' \<rparr>)
  = (x = x' \<and> updated sv = updated sv')"
  apply(auto dest: arg_cong[where f=card_mode])
  apply (metis (lifting) pmm_vol_mem.iffs pmm_vol_mem.simps(4) pmm_vol_mem.surjective)+
  done

lemma abs3_reliable_ref_simps:
  "hw_s = hw_state spc \<Longrightarrow>
    abs3_reliable (spc\<lparr>hw_state := fun_upd_second hw_s ind inst IDead\<rparr>)
    = abs3_reliable spc - {(ind, inst)}"
  "\<lbrakk> hstate spc \<noteq> LDead; hw_state spc ind inst \<noteq> IDead;
    pub_hashes spc (ind, inst) = Hash (abs3_state.memory spc ind inst) \<rbrakk> \<Longrightarrow>
  ((ind, inst) \<in> abs3_reliable spc) = True
  "
  "hstate spc \<noteq> LDead \<Longrightarrow> abs3_reliable (spc\<lparr> hstate:= LDegraded \<rparr>) = abs3_reliable spc"
  "hstate spc \<noteq> LDead \<Longrightarrow>
    abs3_reliable (spc\<lparr> hstate:= LDegraded,
        hw_state := fun_upd_second (hw_state spc) ind inst IDead \<rparr>)
  = abs3_reliable spc - {(ind, inst)}"
  "hstate spc \<noteq> LDead \<Longrightarrow>
  abs3_reliable
  (spc\<lparr>abs3_state.memory := fun_upd_second (abs3_state.memory spc) ind inst cnt,
        hstate := LDegraded, hw_state := fun_upd_second (hw_state spc) ind inst IDead\<rparr>) 
  = abs3_reliable spc - {(ind, inst)}
  "
  "\<lbrakk> hstate spc \<noteq> LDead; ph_f = pub_hashes spc;
    ph_f (ind, inst) = Hash cnt; hws_f = hw_state spc; hws_f ind inst \<noteq> IDead;
    mem_f = abs3_state.memory spc
  \<rbrakk> \<Longrightarrow>
  abs3_reliable (spc\<lparr>hstate := LDegraded,
    abs3_state.memory := fun_upd_second mem_f ind inst cnt\<rparr>) 
  = insert (ind, inst) (abs3_reliable spc)
  "
  "\<lbrakk>f = pub_hashes spc; f (ind, inst) \<noteq> Hash cnt; mem_f = abs3_state.memory spc\<rbrakk> \<Longrightarrow>
    abs3_reliable (spc\<lparr>abs3_state.memory :=
        fun_upd_second mem_f ind inst cnt\<rparr>)
    = abs3_reliable spc - {(ind, inst)}
  "
  "hw_s' = hw_state spc \<Longrightarrow> abs3_reliable (spc\<lparr>hw_state :=
    fun_upd_second hw_s' ind inst (min (hw_s' ind inst) IUnwr)\<rparr>) 
   = abs3_reliable spc
  "
  "\<lbrakk> hw_s' ind inst \<noteq> IDead ;hw_s' = hw_state spc \<rbrakk> \<Longrightarrow> abs3_reliable (spc\<lparr>hw_state :=
    fun_upd_second hw_s' ind inst IUnwr\<rparr>)
   = abs3_reliable spc
  "
  "hstate spc = LDead \<Longrightarrow> abs3_reliable spc = {}"
  "\<lbrakk> hstate spc \<noteq> LDead; hw_state spc ind to \<noteq> IDead;
  pub_hashes spc (ind, from) = Hash (abs3_state.memory spc ind from) \<rbrakk> \<Longrightarrow>
  abs3_reliable (spc \<lparr>
    abs3_state.memory :=
        fun_upd_second (abs3_state.memory spc) ind to cnt,
        pub_hashes :=
            (abs3_state.pub_hashes spc) ((ind, to) := Hash cnt)\<rparr>)
  = insert (ind, to) (abs3_reliable spc)
  "  
  "\<lbrakk> hs = hstate spc; hs \<noteq> LDead; hw_s' = hw_state spc; hw_s' ind inst \<noteq> IDead;
    mem_f = abs3_state.memory spc
  \<rbrakk>
    \<Longrightarrow> abs3_reliable
    (spc\<lparr>abs3_state.memory :=
        fun_upd_second mem_f ind inst cnt\<rparr>) = 
  (if pub_hashes spc (ind, inst) = Hash cnt
  then insert (ind, inst) (abs3_reliable spc)
  else abs3_reliable spc - {(ind, inst)})
  "
  "pub_hashes spc (ind, inst) \<noteq> Hash (abs3_state.memory spc ind inst)
    \<Longrightarrow> (ind, inst) \<in> abs3_reliable spc = False"
  "abs3_reliable (spc \<lparr> ptr := ptr' \<rparr>) = abs3_reliable spc"
  "abs3_reliable (spc \<lparr> ptr_state := pst' \<rparr>) = abs3_reliable spc"
  "hw_state spc ind inst = IDead \<Longrightarrow> ((ind, inst) \<in> abs3_reliable spc) = False"
  "hw_s = hw_state spc \<Longrightarrow> abs3_reliable
  (spc\<lparr>hw_state := hw_s(pind := (hw_s pind) (Primary := IDead, Secondary := IDead))\<rparr>) =
    abs3_reliable spc - {(pind, Primary), (pind, Secondary)}"
  "\<lbrakk> ph_f = pub_hashes spc; mem_f = abs3_state.memory spc;
    ph_f (pind, Primary) \<noteq> Hash cnt_p; ph_f (pind, Secondary) \<noteq> Hash cnt_s \<rbrakk>
    \<Longrightarrow>
    abs3_reliable (spc\<lparr>abs3_state.memory := mem_f
                         (pind := (mem_f pind)
                            (Primary := cnt_p, Secondary := cnt_s))\<rparr>)
    = abs3_reliable spc - {(pind, Primary), (pind, Secondary)}"
  by(auto simp add: min_def abs3_reliable_def split: split_if_asm)

lemma abs3_reliable_more_ref_simps:
  "hst \<le> hstate spc \<Longrightarrow> \<exists>unr. abs3_reliable spc - unr = abs3_reliable (spc\<lparr>hstate := hst\<rparr>)"
  apply(case_tac "hst")
  apply(clarsimp_all simp add: abs3_reliable_ref_simps set_ref_simps
    log_state_ref_simps abs3_state_ref_simps)
  done

lemma abs3_writable_ref_simps:
  "hw_s' = hw_state spc \<Longrightarrow>
    abs3_writable (spc\<lparr>hw_state := fun_upd_second hw_s' ind inst IDead\<rparr>)
    = abs3_writable spc - {(ind, inst)}"
  "hw_s' = hw_state spc \<Longrightarrow>
  abs3_writable (spc\<lparr>hw_state :=
    fun_upd_second hw_s' ind inst (min (hw_s' ind inst) IUnwr)\<rparr>) =
  abs3_writable spc - {(ind, inst)}
  "
  "hw_s' = hw_state spc \<Longrightarrow>
  abs3_writable (spc\<lparr>hw_state :=
    fun_upd_second hw_s' ind inst IUnwr\<rparr>) =
  abs3_writable spc - {(ind, inst)}
  "
  "\<lbrakk> hstate spc = LOK; hw_state spc ind inst = IOK \<rbrakk> \<Longrightarrow>
  ((ind, inst) \<in> abs3_writable spc) = True
  "
  "abs3_writable (spc \<lparr> abs3_state.pub_hashes := h' \<rparr>) = abs3_writable spc"
  "abs3_writable (spc \<lparr> abs3_state.memory := m' \<rparr>) = abs3_writable spc"
  "hstate spc = LDegraded \<Longrightarrow> abs3_writable spc = {}"
  "hstate spc = LDead \<Longrightarrow> abs3_writable spc = {}"
  "hw_state spc ind inst \<le> IUnwr \<Longrightarrow> (ind, inst) \<in> abs3_writable spc = False"
  "abs3_writable (spc \<lparr> ptr_state := pst' \<rparr>) = abs3_writable spc"
  "abs3_writable (spc \<lparr> ptr := ptr' \<rparr>) = abs3_writable spc"
  by(auto simp add: min_def abs3_writable_def split: split_if_asm)
 
lemma abs3_writable_more_ref_simps:
  "hst \<le> hstate spc \<Longrightarrow> \<exists>unwr. abs3_writable spc - unwr = abs3_writable (spc\<lparr>hstate := hst\<rparr>)"
  apply(case_tac "hst")
  apply(clarsimp_all simp add: abs3_writable_ref_simps set_ref_simps
    log_state_ref_simps abs3_state_ref_simps)
  done

lemma abs3_arrow_ref_simps:
  "hstate spc = LDead \<Longrightarrow> abs3_arrow spc = K SBot"
  "ptr_state spc = LDead \<Longrightarrow> abs3_arrow spc = K SBot"
  "abs3_arrow (spc\<lparr> hw_state := hw_f \<rparr>) = abs3_arrow spc"
  "abs3_arrow (spc \<lparr> abs3_state.memory := mem_f \<rparr>) = abs3_arrow spc"
  "hstate spc \<noteq> LDead \<Longrightarrow> abs3_arrow (spc\<lparr> hstate := LDegraded \<rparr>) = abs3_arrow spc"
  "abs3_arrow (spc\<lparr>abs3_state.memory := m', hw_state := hw' \<rparr>) = abs3_arrow spc"
  "hstate spc = LDead \<Longrightarrow> abs3_arrow spc = K SBot"
  "(abs3_arrow spc ind = SEq) = (hstate spc \<noteq> LDead \<and> ptr_state spc \<noteq> LDead \<and>
    pub_hashes spc (ind, Primary) = pub_hashes spc (ind, Secondary))"
  "(SEq = abs3_arrow spc ind) = (hstate spc \<noteq> LDead \<and> ptr_state spc \<noteq> LDead \<and>
    pub_hashes spc (ind, Primary) = pub_hashes spc (ind, Secondary))"
  "\<lbrakk> pub_hashes spc = f; f (ind, from) = ph; hstate spc \<noteq> LDead; ptr_state spc \<noteq> LDead;
    to \<noteq> from \<rbrakk> \<Longrightarrow>
    abs3_arrow (spc\<lparr> pub_hashes := f((ind, to) := ph) \<rparr>) = 
    (abs3_arrow spc)(ind := SEq)
  "
  "ptr_state spc \<noteq> LDead \<Longrightarrow> abs3_arrow (spc\<lparr>ptr_state := LDegraded\<rparr>) = abs3_arrow spc"
  "\<lbrakk> pub_h = pub_hashes spc; pub_h (pind, Primary) = pub_h (pind, Secondary);
    hst = hstate spc; ptr' = ptr spc; hst \<noteq> LDead\<rbrakk>
     \<Longrightarrow> abs3_arrow (spc\<lparr>ptr := ptr' - {pind}\<rparr>) = abs3_arrow spc"
  "\<lbrakk> ptr_state spc \<noteq> LDead; pst \<noteq> LDead \<rbrakk> \<Longrightarrow> abs3_arrow (spc\<lparr>ptr_state := pst\<rparr>) = abs3_arrow spc"
  by(auto simp add: abs3_arrow_def K_def split: split_if intro!: ext)
  
lemma abs4_card_mode_min_comm:
  "abs4_card_mode (min CDegraded (card_mode svc)) =
    min (abs4_card_mode (card_mode svc)) ACDegraded"
  "abs4_card_mode (min CDegradedP (card_mode svc)) =
    min (abs4_card_mode (card_mode svc)) ACDegraded"
  "abs4_card_mode (min CDegradedH (card_mode svc)) =
    min (abs4_card_mode (card_mode svc)) ACDegraded"
  by(case_tac[!] "card_mode svc", auto simp add: min_def)
 
lemmas fun_upd_ref_simps =
  fun_upd_second_is_triv_iff
  fun_upd_second_is_upd_iff
  fun_upd_second_same
  fun_upd_is_triv_iff
  fun_upd_eq_iff

lemmas abs3_ref_simps =
  set_ref_simps
  bool_ref_simps
  misc_ref_simps
  abs3_reliable_ref_simps
  abs3_reliable_more_ref_simps
  abs3_writable_ref_simps
  abs3_writable_more_ref_simps
  abs3_arrow_ref_simps
  abs3_state_ref_simps
  pmm_vol_mem_ref_simps
  override_on_ref_simps
  abs4_card_mode_min_comm
  inst_state_ref_simps
  log_state_ref_simps
  fun_upd_ref_simps
  ex_conj_distribs

lemma sync_ref_NN:
  "from \<noteq> to \<Longrightarrow>
  sim_normN_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.sync ind to from)
    (Abs3_CommonDefs.sync ind to from)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.sync_def)
  apply(simp_all only: first_stage)
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: brute_force write_hashes_defs get_hashes_defs 
    abs3_inv_defs
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: Abs4_CommonDefs.sync_defs choice_def abs4_abs3_strel_defs)
  apply(intro conjI ext, clarsimp_all split: split_if intro!: set_eqI)+
  apply(safe)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms card_mode_exhaust} *}, simp_all)
  done

lemma sync_ref_NE:
  "\<lbrakk> from \<noteq> to \<rbrakk> \<Longrightarrow>
  sim_normE_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.sync ind to from)
    (Abs3_CommonDefs.sync ind to from)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: Abs3_CommonDefs.sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_hashes_def)
  apply(simp_all only: first_stage vol_ops_project_simps restart_mangle_sv_proj_simps
    write_hashes_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc'
    Product_Type.split_beta]]
  apply(case_tac[!] ec)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs
    abs3_inv_defs
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all)
  apply(iterative_simp simp add: FailMode_exhaust FailWrite_exhaust)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailMode_exhaust} *}, simp_all)
  apply(clarsimp_all)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailMode_exhaust} *}, simp_all)
  apply(clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(clarsimp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms InvalidInput_exhaust} *}, simp_all)
  apply(clarsimp_all)
  apply(clarsimp_all simp add: Abs4_CommonDefs.sync_defs choice_def abs4_abs3_strel_def
    brute_force fun_upd_second_idem_iff abs3_ref_simps
    split: split_if split_if_asm del: disjCI)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc ex_conj_distribs
  )
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps fun_upd_second_is_upd_iff del: disjCI)
  apply(iterative_simp del: disjCI)+
  apply(clarsimp_all simp add: min_def split: split_if_asm del: disjCI)
  apply(clarsimp_all simp add: abs3_ref_simps K_def)
  apply(case_tac[!] to, clarsimp_all simp add: fun_upd_triv 
    fun_upd_is_triv_iff[THEN trans[OF eq_commute]])
  apply((metis leD)+)[6]
  apply(case_tac[!] aa, simp_all)
  apply(clarsimp_all)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms exc_exhaust} *}, simp_all)
  apply((metis leD)+)
  done

lemma ex_subset_conj_distrib:
  "(\<exists>A \<subseteq> B. P A \<and> Q) = ((\<exists>A \<subseteq> B. P A) \<and> Q)"
  by(auto)

lemma INST_fun_eq_iff:
  "(f = f') = (f Primary = f' Primary \<and> f Secondary = f' Secondary)"
  apply(rule iffI, simp)
  apply(rule ext, case_tac x, simp_all)
  done

lemma sync_ref:
  "\<lbrakk> from \<noteq> to \<rbrakk> \<Longrightarrow>
  sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.sync ind to from)
    (Abs3_CommonDefs.sync ind to from)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    (abs4_abs3_strel)
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro!: sim_normN_ext_mono[OF sync_ref_NN]
    sync_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_hashes_def)
  apply(simp_all only: first_stage vol_ops_project_simps restart_mangle_sv_proj_simps
    write_hashes_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc'
    Product_Type.split_beta]]
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs
    abs3_inv_defs
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp simp add: FailMode_exhaust FailWrite_exhaust)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms FailWrite_exhaust} *}, simp_all)
  apply(clarsimp_all)
  apply(tactic {* (REPEAT1 (CHANGED (TRYALL (etac @{thm disjE})))) *})
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms InvalidInput_exhaust} *}, simp_all)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all)
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: Abs4_CommonDefs.sync_defs abs4_abs3_strel_def
    brute_force abs3_ref_simps INST_fun_eq_iff
    split: split_if split_if_asm del: disjCI)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    ex_subset_conj_distrib
    abs3_ref_simps
  )
  apply(case_tac[!] to)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add:
    abs3_ref_simps INST_fun_eq_iff
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    ex_subset_conj_distrib
    min_def
    del: disjCI
  )
  apply(clarsimp_all simp add: fun_upd_is_triv_iff
    fun_upd_is_triv_iff[THEN trans[OF eq_commute]])
  apply(iterative_simp)+
  apply(auto)
  done

lemma read_ref_NN:
  "sim_normN abs4_abs3_strel
    (Abs4_CommonDefs.read ind inst)
    (Abs3_CommonDefs.read ind inst)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage)
  apply(clarsimp_all simp add: brute_force write_hashes_defs get_hashes_defs
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs choice_def abs4_abs3_strel_defs)
  apply(auto intro!: ext)
  done

lemma LDead_le_trans: "\<lbrakk> x \<noteq> LDead; \<not> LDegraded \<le> y \<rbrakk> \<Longrightarrow> y \<le> x"
  by(case_tac x, case_tac[!] y, simp_all)

lemma read_ref_NE:
  "sim_normE_ext abs4_abs3_strel
    (Abs4_CommonDefs.read ind inst)
    (Abs3_CommonDefs.read ind inst)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  (* TODO: simplifier bug here in the above procedure? *)
  apply(simp_all add: brute_force get_hashes_all_defs write_hashes_defs hashes_ok_def
    split: split_if_asm exceptions.split_asm)
  using [[simproc add: Product_Type.split_beta]]
  apply(clarsimp_all)
  apply(iterative_simp)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs abs4_abs3_strel_def
    brute_force abs3_ref_simps
    del: disjCI
    split: split_if split_if_asm)
  apply(clarsimp_all simp add: abs3_ref_simps LDead_le_trans min_def split: split_if)
  apply(auto)
  done

lemma read_ref:
  "sim_ext abs4_abs3_strel
    (Abs4_CommonDefs.read ind inst)
    (Abs3_CommonDefs.read ind inst)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro: sim_normN_ext_mono[OF read_ref_NN[simplified sim_normN_def]]
    read_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(simp_all add: brute_force get_hashes_all_defs write_hashes_defs hashes_ok_def
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all)
  apply(iterative_simp)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs choice_def abs4_abs3_strel_def
    brute_force abs3_ref_simps
    split: split_if split_if_asm del: disjCI)
  apply(clarsimp_all simp add: fun_upd_is_triv_iff
    fun_upd_is_triv_iff[THEN trans[OF eq_commute]])
  apply (metis log_state_ref_simps(5) log_state_ref_simps(6)
    less_eq_log_state.simps(2) min_le_iff_disj)+
  done

lemma arrow_SEqI:
  "\<lbrakk> abs3_state.pub_hashes spc (ind, Primary) = abs3_state.pub_hashes spc (ind, Secondary);
        hstate spc \<noteq> LDead; ptr_state spc \<noteq> LDead \<rbrakk> \<Longrightarrow>
    abs3_arrow spc ind = SEq"
  by(auto simp add: abs3_arrow_def)

lemma arrow_SLI:
  "\<lbrakk> abs3_state.pub_hashes spc (ind, Primary) \<noteq> abs3_state.pub_hashes spc (ind, Secondary);
      hstate spc \<noteq> LDead; ptr_state spc \<noteq> LDead; ind \<in> ptr spc \<rbrakk> \<Longrightarrow>
    abs3_arrow spc ind = SL"
  by(auto simp add: abs3_arrow_def)

lemma arrow_SRI:
  "\<lbrakk> abs3_state.pub_hashes spc (ind, Primary) \<noteq> abs3_state.pub_hashes spc (ind, Secondary);
      hstate spc \<noteq> LDead; ptr_state spc \<noteq> LDead; ind \<notin> ptr spc \<rbrakk> \<Longrightarrow>
    abs3_arrow spc ind = SR"
  by(auto simp add: abs3_arrow_def)
 
lemma inf_commute_eq:
  "(min (x :: 'a :: linorder) y = min y x) = True"
  by(auto simp add: min_def)

lemma LDegraded_leq:
  "(LDegraded \<le> x) = (x = LOK \<or> x = LDegraded)"
  by(case_tac x, simp_all)

lemma le_LDegraded: "(x \<le> LDegraded) = (x = LDead \<or> x = LDegraded)"
  by(case_tac x, simp_all)

lemma get_arrow_ref_NN:
  "sim_normN_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.get_arrow ind)
    (Abs3_CommonDefs.get_sync_direction ind)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: get_sync_direction_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs get_ptr_defs
    split: split_if_asm exceptions.split_asm)
  apply(tactic {* distinct_subgoals_tac *})
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs3_inv_defs)
  apply(clarsimp_all simp add: get_arrow_defs abs4_abs3_strel_def brute_force min_def del: disjCI)
  using [[simp_depth_limit=5]]
  apply(case_tac[!] "card_mode svc")
  apply(clarsimp_all simp add: abs3_ref_simps arrow_SEqI arrow_SLI arrow_SRI)
  done

lemma get_arrow_ref_NE:
  "sim_normE_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.get_arrow ind)
    (Abs3_CommonDefs.get_sync_direction ind)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: get_sync_direction_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs get_ptr_defs
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs3_inv_defs)
  apply(clarsimp_all simp add: get_arrow_defs abs4_abs3_strel_def brute_force min_def del: disjCI)
  apply(iterative_simp simp add: le_LDegraded)+
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps K_def)
  done

lemma get_arrow_ref:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.get_arrow ind)
    (Abs3_CommonDefs.get_sync_direction ind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro: sim_normN_ext_mono[OF get_arrow_ref_NN]
    get_arrow_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: get_sync_direction_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    get_hashes_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs get_ptr_defs
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs3_inv_defs)
  apply(clarsimp_all simp add: get_arrow_defs abs4_abs3_strel_def brute_force min_def del: disjCI)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps K_def)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms card_mode_exhaust} *}, simp_all)
  done

lemma all_reliable_no_IDead:
  "abs3_reliable spc = UNIV \<Longrightarrow> hw_state spc ind inst \<noteq> IDead"
  by(auto simp add: abs3_reliable_def)

lemma write_secondary_arrow_upd:
  "\<lbrakk> (ind, Primary) \<in> abs3_reliable spc;
  hstate spc \<noteq> LDead; ptr_state spc \<noteq> LDead; ptr spc = {} \<rbrakk> \<Longrightarrow>
  abs3_arrow
    (spc\<lparr>abs3_state.memory :=
                     fun_upd_second (abs3_state.memory spc) ind Secondary
                      cnt,
                     pub_hashes := (abs3_state.pub_hashes spc)
                       ((ind, Secondary) :=
                          Hash cnt)\<rparr>)
  = (abs3_arrow spc)(ind := (if cnt = abs3_state.memory spc ind Primary then SEq else SR))
  "
  by(auto simp add: abs3_reliable_def abs3_arrow_def intro!: ext)

lemma write_secondary_reliable_upd:
  "((ind', inst) \<in> abs3_reliable
    (spc\<lparr>abs3_state.memory :=
        fun_upd_second (abs3_state.memory spc) ind Secondary
         cnt,
        pub_hashes := (abs3_state.pub_hashes spc)
          ((ind, Secondary) :=
             Hash cnt)\<rparr>)
  ) = (hstate spc \<noteq> LDead \<and> hw_state spc ind' inst \<noteq> IDead
    \<and> ((ind', inst) \<in> abs3_reliable spc \<or> ind = ind' \<and> inst = Secondary))"
  by(auto simp add: abs3_reliable_def)

lemma write_secondary_ref_NN:
  "cnt = the (upd_f ind) \<Longrightarrow>
  sim_normN_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (Abs4_CommonDefs.write_new ind Secondary cnt) 
    (Abs3_UpdateDefs.write_secondary upd_f ind)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: write_secondary_def)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: update_instance_hash_def)
  apply(simp_all add: first_stage)
  apply(simp_all only: add_to_updated_def)
  apply(simp_all add: first_stage)
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs
    split: split_if_asm exceptions.split_asm)
  apply(auto simp add: write_new_defs brute_force abs4_abs3_strel_def del: disjCI)
  apply(clarsimp_all simp add: write_secondary_arrow_upd write_secondary_reliable_upd abs3_ref_simps
    precommit_inv_defs all_reliable_inv_def all_reliable_no_IDead)
  done

lemma all_reliable_imp_hash_matches:
  "abs3_reliable spc = UNIV \<Longrightarrow> pub_hashes spc (ind, inst) = Hash (abs3_state.memory spc ind inst)"
  by(auto simp add: abs3_reliable_def)

lemma write_secondary_ref_NE:
  "cnt = the (upd_f ind) \<Longrightarrow>
  sim_normE_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (Abs4_CommonDefs.write_new ind Secondary cnt) 
    (Abs3_UpdateDefs.write_secondary upd_f ind)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: write_secondary_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice get_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice get_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_f_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs empty_ptr_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: write_new_defs brute_force abs4_abs3_strel_def
    del: disjCI)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
  )
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: precommit_inv_defs hashes_ok_def min_def
    write_secondary_arrow_upd write_secondary_reliable_upd abs3_ref_simps K_def
    all_reliable_inv_def all_reliable_no_IDead all_reliable_imp_hash_matches
    del: disjCI)
  apply(safe del: disjCI)
  apply(simp_all (no_asm_simp) add: 
        HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
        disj_commute disj_left_commute disj_absorb disj_left_absorb
        conj_commute conj_left_commute conj_left_absorb 
        fun_upd_is_triv_iff[THEN trans[OF eq_commute]]
    split: split_if_asm
  )
  apply(clarsimp_all simp add: abs3_ref_simps abs3_arrow_def all_reliable_imp_hash_matches)
  done

lemma write_secondary_ref:
  "cnt = the (upd_f ind) \<Longrightarrow>
  sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (Abs4_CommonDefs.write_new ind Secondary cnt) 
    (Abs3_UpdateDefs.write_secondary upd_f ind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)

  abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro: write_secondary_ref_NN
    write_secondary_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: write_secondary_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_instance_def)
  apply(simp_all add: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_instance_hash_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice get_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice get_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: write_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_f_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice write_hashes_proj_simps
    restart_mangle_sv_proj_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  using [[simproc del: Product_Type.split_beta]]
  apply(clarsimp_all simp add: brute_force get_hashes_all_defs write_hashes_defs empty_ptr_def
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: write_new_defs brute_force abs4_abs3_strel_def
    del: disjCI)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_left_absorb 
  )
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: precommit_inv_defs hashes_ok_def min_def
    write_secondary_arrow_upd write_secondary_reliable_upd abs3_ref_simps K_def
    all_reliable_no_IDead all_reliable_imp_hash_matches all_reliable_inv_def
    del: disjCI)
  apply(safe del: disjCI)
  apply(simp_all (no_asm_simp) add: 
        HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
        disj_commute disj_left_commute disj_absorb disj_left_absorb
        conj_commute conj_left_commute conj_left_absorb 
        fun_upd_is_triv_iff[THEN trans[OF eq_commute]]
    split: split_if_asm
  )
  apply(clarsimp_all simp add: abs3_ref_simps abs3_arrow_def)
  apply (smt UNIV_I all_reliable_imp_hash_matches hash.inject subset_UNIV)+
  done

lemma set_ptr_ref_NN:
  "sim_normN_ext (abs4_abs3_strel \<inter> UNIV \<times> precommit_inv)
    (Abs4_UpdateDefs.switch_arrows pinds)
    (Abs3_CommonDefs.set_ptr pinds)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: set_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: get_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_ptr_defs write_ptr_defs brute_force
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: switch_arrows_defs brute_force abs4_abs3_strel_defs
    precommit_inv_defs
    override_on_def
    intro!: ext
    del: disjCI)
  done

lemma more_stuff:
  "abs3_reliable (spc\<lparr> ptr := pinds \<rparr>) = abs3_reliable spc"
  "abs3_writable (spc\<lparr> ptr := pinds \<rparr>) = abs3_writable spc"
  by(clarsimp_all simp add: abs3_reliable_def abs3_writable_def)

lemma arrow_switch:
  "\<lbrakk> ptr spc = ptr'; ptr' = {}; arr = abs3_arrow spc;
    hstate spc = hst; hst \<noteq> LDead; ptr_state spc = pst; pst \<noteq> LDead \<rbrakk> \<Longrightarrow>
    override_on arr (switch_arrow arr) pinds
    = abs3_arrow (spc\<lparr>ptr := pinds\<rparr>)"
  by(auto simp add: override_on_def abs3_arrow_def intro!: ext)

lemma more_arrow:
  "ptr_state spc \<noteq> LDead \<Longrightarrow>
    abs3_arrow (spc\<lparr>ptr_state := LDegraded, ptr := pinds\<rparr>) = abs3_arrow (spc\<lparr>ptr := pinds\<rparr>)"
  "hstate spc \<noteq> LDead \<Longrightarrow>
    abs3_arrow (spc\<lparr>hstate := LDegraded, ptr := pinds\<rparr>) = abs3_arrow (spc\<lparr>ptr := pinds\<rparr>)"
  by(auto simp add: abs3_arrow_def intro!: ext)

lemma set_ptr_ref_NE:
  "sim_normE_ext (abs4_abs3_strel \<inter> UNIV \<times> precommit_inv)
    (Abs4_UpdateDefs.switch_arrows pinds)
    (Abs3_CommonDefs.set_ptr pinds)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: set_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: get_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_ptr_defs write_ptr_defs brute_force
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: switch_arrows_defs brute_force abs4_abs3_strel_def
    intro!: ext
    del: disjCI)
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: precommit_inv_defs hashes_ok_def min_def
    abs3_ref_simps K_def
    del: disjCI)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    LDegraded_leq
  )
  apply(iterative_simp del: disjCI)+
  apply(clarsimp_all simp add: precommit_inv_defs hashes_ok_def min_def
    abs3_ref_simps K_def more_stuff arrow_switch more_arrow
    del: disjCI)
  done

lemma set_ptr_ref:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> precommit_inv)
    (Abs4_UpdateDefs.switch_arrows pinds)
    (Abs3_CommonDefs.set_ptr pinds)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro: set_ptr_ref_NN
    set_ptr_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: set_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: get_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: write_ptr_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_ptr_defs write_ptr_defs brute_force
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: switch_arrows_defs brute_force abs4_abs3_strel_def
    intro!: ext
    del: disjCI)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    LDegraded_leq
  )
  apply(clarsimp_all simp add: precommit_inv_defs hashes_ok_def min_def
    abs3_ref_simps K_def more_stuff arrow_switch more_arrow
    del: disjCI)
  done

lemma abs3_arrow_SBot_iff:
  "(abs3_arrow spc ind = SBot) = (hstate spc = LDead \<or> ptr_state spc = LDead)"
  by(auto simp add: abs3_arrow_def)

lemma full_override_nonempty:
  "(\<exists>A. A \<noteq> {} \<and> override_on f g A = g) = True"
  by(simp, rule_tac x=UNIV in exI, simp add: override_on_def)

lemma startup1_ref:
  "sim abs4_abs3_strel
    (Abs4_StartupDefs.startup1)
    (Abs3_StartupDefs.startup1)
  "
  apply(auto simp add: sim_def sim_ext_def force_same_and_def simN_wpc_def simS_wpc_def)
  apply(simp_all add: Abs3_StartupDefs.startup1_def compose_project)
  apply(simp_all add: Abs4_StartupDefs.startup1_def compose_project)
  apply(clarsimp_all simp add:
    Abs3_StartupDefs.startup1_NN_def
    Abs3_StartupDefs.startup1_NE_def
    brute_force)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs4_abs3_strel_def
    Abs4_StartupDefs.startup1_NN_def
    Abs4_StartupDefs.startup1_NE_def
    brute_force del: disjCI)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    ex_conj_distribs
    LDegraded_leq
  )
  apply(iterative_simp simp add: abs3_arrow_SBot_iff del: disjCI)+
  apply(clarsimp_all simp add: abs3_ref_simps K_def full_override_nonempty del: disjCI)
  apply(clarsimp_all simp add: min_def abs3_ref_simps)
  apply (metis log_state.exhaust)
  done

declare sim_ext_return[ref_thm]
declare sim_ext_return_diffI[ref_thm]
declare sim_ext_throw[ref_thm]

declare
  sim_return[ref_thm]
  sims_throw(1)[ref_thm]
  sim_chooseN[ref_thm]

declare
  sim_ext_bind_rt[ref_bind]
  sim_ext_try_catch_rt[ref_try_catch]
  sim_ext_exc_case_rt[ref_exc_case]
  sim_ext_if_rt[ref_if]
  sim_ext_if_full_rt[ref_if_full]
  sim_ext_mapM_rt[ref_mapM]

declare read_ref[ref_thm]
declare sync_ref[ref_thm]

lemma remove_from_ptr_silence_deg_ref:
  "
    sim_ext (abs4_abs3_strel \<inter> (arrow_is SEq pind \<times> UNIV))
    (maybe_mess_up_admin pind)
    (remove_from_ptr_silence_deg pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
  abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def
    force_same_and_def force_same_and_def)
  apply(simp_all only: remove_from_ptr_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: remove_from_ptr_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force arrow_is_def get_ptr_defs write_ptr_defs
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs4_abs3_strel_def brute_force maybe_mess_up_admin_def
    get_arrow_degrade_def get_arrow_restart_def get_arrow_fail_def del: disjCI)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc
    ex_conj_distribs
    LDegraded_leq le_LDegraded
  )
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: abs3_ref_simps del: disjCI)
  apply(iterative_simp del: disjCI)+
  using [[simp_depth_limit=5]]
  apply(clarsimp_all simp add: min_def K_def
    abs3_ref_simps le_LDegraded LDegraded_leq 
    del: disjCI)
  apply(case_tac[!] "card_mode aa", simp_all)
  done

lemma sync_ref':
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.sync pind to from)
    (Abs3_CommonDefs.sync pind to from)
    (force_same_and (abs4_abs3_strel \<inter> (arrow_is SEq pind \<times> UNIV)))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: force_same_and_Int 
    intro!: sim_ext_add_NN_inv[OF sync_post_NN true_postNN, simplified K_beta])
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *}, simp_all)
  done

lemma just_sync_ref':
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.just_sync pind to from)
    (Abs3_CommonDefs.just_sync pind to from)
    (force_same_and (abs4_abs3_strel \<inter> (arrow_is SEq pind \<times> UNIV)))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: force_same_and_Int 
    intro!: sim_ext_add_NN_inv[OF just_sync_post_NN true_postNN, simplified K_beta])
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *}, simp_all)
  done

declare sync_ref'[ref_thm]
declare just_sync_ref'[ref_thm]
declare remove_from_ptr_silence_deg_ref[ref_thm]

lemma sync_and_remove_from_ptr_ref[ref_thm]:
  "from \<noteq> to \<Longrightarrow> sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.just_sync_silence_deg' to from pind)
    (Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg to from pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(simp_all only:
    Abs4_CommonDefs.just_sync_silence_deg_def'
    Abs3_CommonDefs.sync_and_remove_from_ptr_silence_deg_def
  )
  apply(rule sim_ext_try_catchI[where Te="force_same_and abs4_abs3_strel"])
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(auto simp add: force_same_and_def split: exceptions.split intro!: sim_exts_empty_rel)
  apply(auto simp add: sim_ext_def simN_wpc_def brute_force simS_wpc_def)
  done

lemma force_same_and_subset:
  "(force_same_and R e e') \<subseteq> R"
  by(case_tac "e=e'", simp_all add: force_same_and_def)

declare sim_ext_throw_diffI[ref_thm]

definition degraded_hashes :: "('sv \<times> 'c abs3_state)set" where
  "degraded_hashes \<equiv> {(svc, spc). hstate spc = LDegraded }"

lemmas read_ref' =
  read_ref[THEN sim_ext_mono, THEN sim_ext_add_NE_inv[OF true_postNE read_post_NE],
  where R=abs4_abs3_strel, simplified,
  THEN sim_ext_add_NN_inv[OF true_postNN read_post_NN], simplified K_def]

declare read_ref[ref_thm del]
declare read_ref'[ref_thm]

lemma read_or_fail_ref_NN:
  "sim_normN abs4_abs3_strel
    (Abs4_CommonDefs.read_or_fail ind inst)
    (Abs3_CommonDefs.read_or_fail ind inst)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all only: first_stage)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage)
  apply(clarsimp_all simp add: brute_force write_hashes_defs get_hashes_defs
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_or_fail_defs choice_def abs4_abs3_strel_defs)
  apply(auto intro!: ext)
  done

lemma read_or_fail_ref_NE:
  "sim_normE_ext abs4_abs3_strel
    (Abs4_CommonDefs.read_or_fail ind inst)
    (Abs3_CommonDefs.read_or_fail ind inst)
    (force_same_and abs4_abs3_strel)
  "
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all add: first_stage cong: exceptions.case_cong if_cong)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_DegradedH_def)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(simp_all add: brute_force get_hashes_all_defs write_hashes_defs hashes_ok_def
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all)
  apply(iterative_simp)
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_or_fail_defs abs4_abs3_strel_def
    brute_force abs3_ref_simps ex_conj_distribs
    del: disjCI
    split: split_if split_if_asm)
  apply(clarsimp_all simp add: min_def abs3_ref_simps)
  apply (metis LDead_le_trans UNIV_I subset_UNIV)+
  done

lemma read_or_fail_ref:
  "sim_ext abs4_abs3_strel
    (Abs4_CommonDefs.read_or_fail ind inst)
    (Abs3_CommonDefs.read_or_fail ind inst)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_decompose simS_wpc_def
    intro: sim_normN_ext_mono[OF read_or_fail_ref_NN[simplified sim_normN_def]]
    read_or_fail_ref_NE
  )
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: Abs3_CommonDefs.read_or_fail_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: Abs3_CommonDefs.read_def)
  apply(simp_all only: first_stage vol_ops_project_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: get_hashes_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  apply(simp_all only: get_hashes_DegradedH_def)
  apply(simp_all add: first_stage vol_ops_project_simps get_hashes_proj_simps
    restart_mangle_sv_proj_simps)
  using [[simp_depth_limit=0]]
  using [[simproc del: throw_simproc throw_simproc' throw_bind_simproc throw_bind_simproc']]
  apply(simp_all add: brute_force get_hashes_all_defs write_hashes_defs hashes_ok_def
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all)
  apply(iterative_simp)+
  using [[simp_depth_limit=2]]
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_or_fail_defs choice_def abs4_abs3_strel_def
    brute_force abs3_ref_simps
    split: split_if split_if_asm del: disjCI)
  apply(clarsimp_all simp add: min_def abs3_ref_simps
    fun_upd_is_triv_iff fun_upd_is_triv_iff[THEN trans[OF eq_commute]]
  )
  apply (metis log_state_ref_simps(6) bot_least not_less_bot order_neq_le_trans)+
  done

declare read_or_fail_ref[ref_thm]

lemma read_or_fail_silence_deg_ref[ref_thm]:
  "sim_ext abs4_abs3_strel
    (Abs4_StartupDefs.read_or_fail_silence_deg ind inst)
    (Abs3_CommonDefs.read_or_fail_silence_deg ind inst)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  done

declare just_sync_ref'[ref_thm del]

lemma just_sync_silence_deg_ref[ref_thm]:
  "from \<noteq> to \<Longrightarrow> sim_ext (abs4_abs3_strel \<inter> (UNIV \<times> abs3_norm_inv))
    (Abs4_CommonDefs.just_sync_silence_deg to from pind)
    (Abs3_CommonDefs.just_sync_silence_deg to from pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *}, auto)

lemma strel_contains_capacity_inv:
  "abs4_abs3_strel \<inter> I \<times> J = abs4_abs3_strel \<inter> I \<times> (J \<inter> abs3_capacity_inv)"
  by(auto simp add: abs4_abs3_strel_def abs3_capacity_inv_def)

lemma check_pub_same_ref:
  "sim_ext (abs4_abs3_strel \<inter> (UNIV  \<times> abs3_norm_inv))
    (Abs4_StartupDefs.check_pub_same pind)
    (Abs3_StartupDefs.check_pub_same pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(simp_all only: 
    Abs4_StartupDefs.check_pub_same_def
    Abs3_StartupDefs.check_pub_same_def
  )
  apply(rule sim_ext_try_catchI)
  apply(rule read_ref'[THEN sim_ext_mono, THEN sim_ext_bindI])
  apply(rule order_refl)
  apply(clarsimp_all)
  apply(rule sim_ext_try_catchI)
  apply(rule read_ref'[THEN sim_ext_mono, THEN sim_ext_bindI])
  apply(rule Int_mono)
  apply(simp add: force_same_and_subset)
  apply(rule order_refl)
  apply(clarsimp)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(clarsimp simp add: force_same_and_def)
  apply(safe)
  apply(auto split: exceptions.split
    simp add: force_same_and_diff' 
    intro!: sim_exts_empty_rel
    del: subsetI
  )
  apply(tactic {* TRYALL (refinement_tac false Ref_Tac.default_simp @{context}) *})
  apply(clarsimp_all simp add:
    force_same_and_subset force_same_and_same
    force_same_and_diff
    intro!: sim_exts_empty_rel
    del: subsetI)
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(clarsimp)
  apply(drule sim_ext_add_NE_inv[rotated 2, OF _ true_postNE])
  apply(rule sim_ext_mono_exc just_sync_silence_deg_inv_NE)
  apply(erule sim_ext_mono_exc)
  apply(auto simp add: K_def intro!: sim_exts_empty_rel
    dest: force_same_and_subset[THEN subsetD])
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(rule sim_ext_mono)
  apply(rule sim_ext_mono_exc)
  apply(force)
  apply(clarsimp simp add: force_same_and_def)
  apply(clarsimp simp add: force_same_and_def)
  done

declare remove_from_ptr_silence_deg_ref[ref_thm]

lemma check_pub_same_ref'[ref_thm]:
  "sim_ext (abs4_abs3_strel \<inter> (arrow_is SEq pind  \<times> abs3_norm_inv))
    (Abs4_StartupDefs.check_pub_same pind)
    (Abs3_StartupDefs.check_pub_same pind)
    (force_same_and (abs4_abs3_strel \<inter> arrow_is SEq pind \<times> UNIV))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(rule sim_ext_add_NN_inv[OF check_pub_same_arrow_inv_NN true_postNN, 
    OF sim_ext_mono, OF check_pub_same_ref, simplified K_beta force_same_and_Int[THEN sym]])
  apply(auto)
  done

lemma bindN_rule_same:
  "\<lbrakk>\<lbrace>P\<rbrace> m \<lbrace>> K P\<rbrace>; \<forall>x. \<lbrace>P\<rbrace> f x \<lbrace>> R\<rbrace>\<rbrakk> \<Longrightarrow> \<lbrace>P\<rbrace> m \<guillemotright>=N f \<lbrace>> R\<rbrace>"
  by(auto simp add: K_beta intro!: bindN_ruleN)

lemmas get_arrow_ref' =
  get_arrow_ref[THEN sim_ext_add_NN_inv[OF true_postNN get_sync_direction_inv_NN], simplified K_def]

lemma get_arrow_post_NN:
  "\<lbrace>P\<rbrace> get_arrow pind \<lbrace>>N \<lambda>dir. arrow_is dir pind \<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def get_arrow_defs brute_force arrow_is_def)

lemma join_invs:
  "(UNIV \<times> J) \<inter> (I \<times> UNIV) = I \<times> J"
  "(I \<times> UNIV) \<inter> (UNIV \<times> J) = I \<times> J"
  by(auto)

lemmas get_arrow_ref'' = get_arrow_ref'[THEN sim_ext_add_NN_inv[OF get_arrow_post_NN true_postNN],
  simplified K_beta force_same_and_Int[THEN sym] Int_assoc join_invs]

lemma check_pub_region_ref:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_StartupDefs.check_pub_region pind)
    (Abs3_StartupDefs.check_pub_region pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(simp_all only: Abs4_StartupDefs.check_pub_region_def Abs3_StartupDefs.check_pub_region_def)
  apply(intro sim_ext_bindI[OF get_arrow_ref''] allI)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(tactic {* TRYALL (refinement_ins_tac false Ref_Tac.default_simp @{context}) *})
  apply(auto intro: sim_ext_mono)
  done

lemma check_pub_region_ref':
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_StartupDefs.check_pub_region pind)
    (Abs3_StartupDefs.check_pub_region pind)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
  apply(insert
    sim_ext_add_NN_inv[OF true_postNN check_pub_region_inv_NN, OF check_pub_region_ref,
    THEN sim_ext_add_NE_inv[OF true_postNE check_pub_region_inv_NE]])
  apply(simp only: force_same_and_Int[THEN sym] force_same_and_Int[THEN sym] K_beta)
  apply(assumption)
  done

declare sim_ext_mapM'[ref_mapM]

lemma force_same_and_Int:
  "force_same_and (A \<inter> B) = (\<lambda>va vc. force_same_and A va vc \<inter> B)"
  by(auto simp add: force_same_and_def)

lemma startup1_ref':
  "sim_ext abs4_abs3_strel
    (Abs4_StartupDefs.startup1)
    (Abs3_StartupDefs.startup1)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
  apply(insert startup1_ref)
  apply(simp add: sim_def)
  apply(drule
    sim_ext_add_NN_inv[OF true_postNN[where P=UNIV] startup1_inv_NN[where P=UNIV], OF sim_ext_mono])
  apply(simp, rule order_refl)
  apply(drule
    sim_ext_add_NE_inv[OF true_postNE[where P=UNIV] startup1_inv_NE[where P=UNIV], OF sim_ext_mono])
  apply(simp, rule order_refl)
  apply(simp only: K_beta force_same_and_Int[THEN sym])
  apply(auto simp add: K_beta force_same_and_Int[THEN sym] elim: sim_ext_mono intro: Int_lower1)
  done

declare startup1_ref'[ref_thm]
declare check_pub_region_ref'[ref_thm]

lemma sim_ext_chooseN:
  "sim_ext R (chooseN A) (chooseN A) (force_same_and R) Se Sr"
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def chooseN_def brute_force)

declare sim_ext_chooseN[ref_thm]

lemma check_all_public_ref[ref_thm]:
 "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
  Abs4_StartupDefs.check_all_public Abs3_StartupDefs.check_all_public
  (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
  (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv)) abs4_abs3_strel
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
 
lemma throw_mode_error_ref[ref_thm]:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (throw_abs_mode_error val)
    (PmmDefs.throw_mode_error val)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force
    abs4_abs3_strel_def
    force_same_and_def force_same_and_def PmmDefs.throw_mode_error_def 
    throw_abs_mode_error_def)
  apply(case_tac[!] "card_mode aa", clarsimp_all dest!: subsetD[OF norm_inv_subset_inv])
  done

lemma startup_ref'':
  "sim_ext abs4_abs3_strel
    Abs4_StartupDefs.startup
    Abs3_StartupDefs.startup
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(auto simp add: force_same_and_same norm_inv_subset_inv[THEN subsetD])
  done

lemma startup_ref':
  "sim_ext abs4_abs3_strel
    Abs4_StartupDefs.startup
    Abs3_StartupDefs.startup
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> abs3_inv))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> abs3_inv))
    abs4_abs3_strel
  "
  by(auto simp add: K_beta force_same_and_def
    intro!:
    startup_ref''[THEN sim_ext_add_NN_inv[OF startup_post_NN true_postNN sim_ext_mono],
    THEN sim_ext_add_NE_inv[OF startup_post_NE true_postNE],
    THEN sim_ext_mono, THEN sim_ext_mono_exc, THEN sim_ext_mono_norm])

lemma startup_ref:
  "sim_ext abs4_abs3_strel
    Abs4_StartupDefs.startup
    Abs3_StartupDefs.startup
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    abs4_abs3_strel
  "
  apply(auto simp add: K_beta force_same_and_def
    intro!:
    startup_ref'[THEN sim_ext_add_NN_inv[OF true_postNN startup_empty_ptr_post_NN sim_ext_mono],
    THEN sim_ext_add_NE_inv[OF true_postNE startup_empty_ptr_post_NE],
    THEN sim_ext_mono, THEN sim_ext_mono_exc, THEN sim_ext_mono_norm])
  apply(auto simp add: abs4_abs3_strel_def abs3_capacity_inv_def)
  done

lemma aux1million: "(P \<and> (P \<longrightarrow> Q)) = (P \<and> Q)"
  by blast

lemma get_arrow_return_ref:
  "sim_ext
     (abs4_abs3_strel \<inter> (arrow_is SEq pind \<union> arrow_is SR pind) \<times> abs3_norm_inv)
     (get_arrow pind) (return a)
     (\<lambda>va _. if va = SL then {} else abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
     (force_same_and abs4_abs3_strel) abs4_abs3_strel"
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def)
  apply(clarsimp_all simp add: arrow_is_def get_arrow_defs brute_force
    abs4_abs3_strel_def abs3_inv_defs)
  apply(clarsimp_all simp add: aux1million)
  apply(simp_all (no_asm) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc ex_conj_distribs
  )
  done
 
lemma get_arrow_arrow_inv_NN:
  "\<lbrace>arrow_is x ind' \<union> arrow_is y ind'\<rbrace>
  get_arrow ind
  \<lbrace>>N K (arrow_is x ind' \<union> arrow_is y ind')\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(auto simp add: arrow_is_def get_arrow_defs brute_force)
  done
 
lemma just_sync_silence_deg_SR_ref:
  "sim_ext (abs4_abs3_strel \<inter> arrow_is SR pind \<times> abs3_norm_inv)
     (just_sync_silence_deg' Secondary Primary pind)
     (Abs3_CommonDefs.just_sync_silence_deg Secondary Primary pind)
     (force_same_and abs4_abs3_strel) (force_same_and abs4_abs3_strel)
     abs4_abs3_strel
  "
  apply(simp only: just_sync_silence_deg_def')
  apply(simp only: Abs4_CommonDefs.just_sync_silence_deg_def'
    Abs3_CommonDefs.just_sync_silence_deg_def)
  apply(rule sim_ext_try_catchI[where Te="force_same_and abs4_abs3_strel"])
  apply(simp only: Abs3_CommonDefs.just_sync_def)
  apply(rule sim_ext_bindI)
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(force intro: sim_ext_mono)
  apply(clarsimp)
  apply(rule sim_ext_mono[where R=abs4_abs3_strel])
  apply(simp only: maybe_mess_up_admin_def)
  apply(rule sim_ext_choice_abs_I1)
  apply(rule sim_ext_return)
  apply(clarsimp simp add: force_same_and_def)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel sim_ext_return sim_ext_throw)
  done

lemma sync_if_not_SEq_ref[ref_thm]:
  "sim_ext (abs4_abs3_strel \<inter> (arrow_is SEq pind \<union> arrow_is SR pind) \<times> abs3_norm_inv)
    (Abs4_UpdateDefs.sync_by_arrow_silence_deg pind)
    (Abs3_UpdateDefs.sync_if_not_SEq Secondary Primary pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
    "
  apply(simp only: sync_by_arrow_silence_deg_def sync_if_not_SEq_def check_arrow_SEq_def)
  apply(simp only: first_stage)
  apply(rule sim_ext_bindI)
  apply(rule get_arrow_ref''[THEN sim_ext_mono,
      THEN sim_ext_add_NN_inv[OF get_arrow_arrow_inv_NN true_postNN],
      THEN sim_ext_mono, THEN sim_ext_mono_norm,
    where R2=abs4_abs3_strel and x4=SEq and y4=SR and ind'4=pind and J2=abs3_norm_inv]
  )
  apply(clarsimp)
  apply(clarsimp)
  apply(intro allI, rule order_refl)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel simp add: K_beta)
  apply(auto intro: just_sync_silence_deg_SR_ref[THEN sim_ext_mono])
  apply(rule sim_ext_mono[OF sim_exts_empty_rel(1)])
  apply(force simp add: arrow_is_def)
  apply(simp only: maybe_mess_up_admin_def)
  apply(rule sim_ext_choice_abs_I1)
  apply(rule sim_ext_return[THEN sim_ext_mono])
  apply(clarsimp)
  apply(rule sim_ext_mono[OF sim_exts_empty_rel(1)])
  apply(force simp add: arrow_is_def)
  done

lemma sim_ext_get_updated_rollback_ref:
  "sim_ext (abs4_abs3_strel \<inter> abs_updated_not_SL \<times> abs3_norm_inv)
    get_abs_updated get_updated
    (\<lambda>va vc. force_same_and abs4_abs3_strel va vc \<inter>
        (\<Inter>ind \<in> va. (arrow_is SEq ind \<union> arrow_is SR ind)) \<times> abs3_norm_inv)
    Se
    Sr
  "
  apply(auto simp add: sim_ext_def)
  apply(auto simp add: simN_wpc_def simS_wpc_def brute_force abs_updated_not_SL_def
    arrow_is_def abs3_inv_defs
    abs4_abs3_strel_def)
  done

lemma set_updated_ref:
  "sim_ext abs4_abs3_strel
  (set_abs_updated l) (set_updated l)
  (force_same_and abs4_abs3_strel)
  (force_same_and abs4_abs3_strel)
  abs4_abs3_strel
  "
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def
    brute_force abs4_abs3_strel_def)

lemma rollback_ref:
  "sim_ext (abs4_abs3_strel \<inter> abs_updated_not_SL \<times> abs3_norm_inv)
    Abs4_UpdateDefs.rollback 
    Abs3_UpdateDefs.rollback 
    Sn
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(simp add: Abs4_UpdateDefs.rollback_def Abs4_UpdateDefs.sync_updated_def
    Abs3_UpdateDefs.rollback_def first_stage)
  apply(rule sim_ext_bindI[OF sim_ext_get_updated_rollback_ref])
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_bindI)
  apply(rule sim_ext_chooseN[THEN sim_ext_add_NN_inv[OF chooseN_ruleNN true_postNN],
    THEN sim_ext_mono, THEN sim_ext_mono_norm])
  apply(rule order_refl)
  apply(intro allI)
  apply(rule order_refl)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel simp add: K_beta)
  apply(subst Int_assoc)
  apply(subst Sigma_Int_distrib2[THEN sym])
  apply(rule sim_ext_bindI)
  apply(rule sim_ext_mapM')
  apply(simp add: order_set_set)
  apply(clarsimp)
  apply(rule 
    sim_ext_add_NN_inv[OF sync_by_arrow_silence_deg_arrow_inv_NN
    sync_if_not_SEq_inv_NN
    sync_if_not_SEq_ref[THEN sim_ext_mono], THEN sim_ext_mono_norm]
  )
  apply(force)
  apply(clarsimp simp add: K_def force_same_and_def)
  apply(clarsimp simp add: force_same_and_same)
  apply(auto simp add: force_same_and_same
    intro!: sim_ext_bindI sim_ext_throw set_updated_ref intro: sim_ext_mono)
  done

lemma mapM'_write_secondary_ref''':
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (mapM' (\<lambda>ind. write_new ind Secondary ((the o upd_f) ind)) ord)
    (mapM' (Abs3_UpdateDefs.write_secondary upd_f) ord)
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
  apply(clarsimp intro!: sim_ext_mapM')
  apply(rule sim_ext_mono_norm[OF sim_ext_add_NN_inv])
  apply(auto intro!: write_new_Secondary_all_reliable_post_NN
    write_secondary_precommit_inv_NN
    write_secondary_ref 
    intro: sim_ext_mono_norm[OF sim_ext_add_NN_inv]
  )
  apply(auto simp add: K_def force_same_and_def)
  done

lemma mapM'_write_secondary_ref'':
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
    (mapM' (\<lambda>ind. write_new ind Secondary ((the o upd_f) ind)) ord)
    (mapM' (Abs3_UpdateDefs.write_secondary upd_f) ord)
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv))
    (\<lambda>ea ec. force_same_and abs4_abs3_strel ea ec \<inter>
        (if \<exists>v. ea = Degraded v then abs_updated_not_SL else UNIV) \<times>
        (if \<exists>v. ea = Degraded v then abs3_norm_inv else UNIV))
    abs4_abs3_strel
  "
  apply(rule sim_ext_mono_exc[OF sim_ext_add_NE_inv])
  apply(rule mapM'_ruleNE, clarsimp_all)
  apply(rule conseqNE_rule[OF Int_lower2 write_new_Secondary_abs_updated_post_NE,
    where S="\<lambda>e. if \<exists>v. e = Degraded v then abs_updated_not_SL else UNIV"])
  apply(simp)
  apply(rule write_new_Secondary_post_NN)
  apply(rule mapM'_ruleNE, clarsimp_all)
  apply(rule write_secondary_rollback_post_NE)
  apply(rule write_secondary_precommit_inv_NN)
  apply(rule mapM'_write_secondary_ref'''[THEN sim_ext_mono, simplified o_def])
  apply(auto simp add: force_same_and_def)
  done

lemma mapM'_write_secondary_ref':
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
    (mapM' (\<lambda>ind. write_new ind Secondary ((the o upd_f) ind)) ord)
    (mapM' (Abs3_UpdateDefs.write_secondary upd_f) ord)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
    (\<lambda>ea ec. force_same_and abs4_abs3_strel ea ec \<inter>
        (if \<exists>v. ea = Degraded v then abs_updated_not_SL else UNIV) \<times>
        (if \<exists>v. ea = Degraded v then abs3_norm_inv else UNIV))
    abs4_abs3_strel
  "
proof(cases ord)
  case Nil
  thus ?thesis
    by(clarsimp simp add: force_same_and_same intro!: sim_ext_return_diffI)
next
  case (Cons x xs)
  have sim':
  "sim_ext ((abs4_abs3_strel \<inter> (all_reliable_inv \<times> precommit_inv)) \<inter> abs_updated_not_SL \<times> UNIV)
    (mapM' (\<lambda>ind. write_new ind Secondary ((the o upd_f) ind)) ord)
    (mapM' (Abs3_UpdateDefs.write_secondary upd_f) ord)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
    (\<lambda>ea ec. force_same_and abs4_abs3_strel ea ec \<inter>
        (if \<exists>v. ea = Degraded v then abs_updated_not_SL else UNIV) \<times>
        (if \<exists>v. ea = Degraded v then abs3_norm_inv else UNIV))
    abs4_abs3_strel
  "
    apply(rule sim_ext_mono_norm[OF sim_ext_add_NN_inv])
    apply(rule mapM'_ruleNN, clarsimp_all)
    apply(rule write_new_Secondary_abs_updated_post_NN)
    apply(rule write_new_Secondary_abs_updated_post_NN)
    apply(rule write_new_Secondary_abs_updated_post_NN)
    apply(rule mapM'_ruleNN, clarsimp_all)
    apply(rule true_postNN)
    apply(rule true_postNN)
    apply(rule true_postNN)
    apply(rule mapM'_write_secondary_ref''[THEN sim_ext_mono, simplified o_def])
    apply(auto simp add: Cons K_beta force_same_and_same)
    done
  show ?thesis
    by(rule sim'[THEN sim_ext_mono, THEN sim_ext_mono_norm], auto)
qed

lemma write_new_subset: "cnt = (the o upd_f) ind \<Longrightarrow>
  write_new ind Secondary cnt (sv, sp) \<subseteq> write_new' upd_f ind (sv, sp)"
  by(auto simp add: write_new'_def brute_force chooseN_def)

lemma mapM'_subsetI: "(\<And>x s. m x s \<subseteq> m' x s) \<Longrightarrow> mapM' m l s \<subseteq> mapM' m' l s"
  apply(induct l arbitrary: s)
  apply(simp)
  apply(rule proj_subset)
  apply(auto simp add: project_defs bind_exc_decompose bind_norm_decompose bind_none_decompose)
  apply(blast)+
  done

lemma mapM'_write_new_subset:
  "(mapM' (\<lambda>ind. write_new ind Secondary ((the \<circ> upd_f) ind)) l) (sv, sp)
    \<subseteq>
  (mapM' (write_new' upd_f) l) (sv, sp)"
  by(auto intro!: mapM'_subsetI write_new_subset del: subsetI)
  
lemma mapM'_write_secondary_ref:
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
    (mapM' (Abs4_UpdateDefs.write_new' upd_f) ord)
    (mapM' (Abs3_UpdateDefs.write_secondary upd_f) ord)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
    (\<lambda>ea ec. force_same_and abs4_abs3_strel ea ec \<inter>
        (if \<exists>v. ea = Degraded v then abs_updated_not_SL else UNIV) \<times>
        (if \<exists>v. ea = Degraded v then abs3_norm_inv else UNIV))
    abs4_abs3_strel
  "
  apply(rule sim_ext_mono_abs_comp)
  apply(auto intro!: mapM'_write_secondary_ref' mapM'_write_new_subset del: subsetI)
  done

lemma precommit_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (Abs4_UpdateDefs.pre_commit upd_f) 
    (Abs3_UpdateDefs.pre_commit upd_f)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
proof-
  have set_updated_ref:
    "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
        (set_abs_updated {})
        (set_updated {})
     (force_same_and
       (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
     (force_same_and abs4_abs3_strel)
     abs4_abs3_strel
    " 
    apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
    apply(clarsimp_all simp add: abs4_abs3_strel_def all_reliable_inv_def precommit_inv_defs
      abs_updated_not_SL_def brute_force)
    done

  have ref':
     "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)
    (Abs4_UpdateDefs.pre_commit upd_f) 
    (Abs3_UpdateDefs.pre_commit upd_f)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
  "
    apply(simp only: Abs4_UpdateDefs.pre_commit_def Abs3_UpdateDefs.pre_commit_def)
    apply(rule sim_ext_bindI[OF set_updated_ref])
    apply(clarsimp simp add: force_same_and_same)
    apply(rule sim_ext_bindI[OF sim_ext_chooseN])
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel)
    apply(rule sim_ext_try_catchI[OF mapM'_write_secondary_ref])
    apply(auto)
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel)
    apply(rule rollback_ref)
    apply(auto simp add: force_same_and_def
      intro: sim_exts_empty_rel sim_ext_throw_diffI split: exceptions.split)
    done

  show ?thesis
    apply(rule
      ref'[THEN sim_ext_mono,
        THEN sim_ext_add_NE_inv[OF true_postNE pre_commit_inv_NE],
        THEN sim_ext_mono, THEN sim_ext_mono_exc,
        where I2=all_reliable_inv and R2="abs4_abs3_strel \<inter> UNIV \<times> precommit_inv"
      ]
    )
    apply(auto simp add: K_beta force_same_and_def precommit_inv_def)
    done
qed

lemma get_updated_commit_ref:
  "sim_ext
     (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
        get_abs_updated
        get_updated
     (\<lambda>va vc.
        force_same_and
        (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL \<inter> abs_updated_equals va) \<times> precommit_inv)
        va vc
     )
     (force_same_and abs4_abs3_strel)
    abs4_abs3_strel"
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force force_same_and_def
    abs4_abs3_strel_def all_reliable_inv_def abs_updated_not_SL_def abs_updated_equals_def 
    precommit_inv_defs)

lemma set_ptr_commit_ref:
  "sim_ext (abs4_abs3_strel \<inter> 
        (all_reliable_inv \<inter> abs_updated_not_SL \<inter> abs_updated_equals pinds) \<times> precommit_inv)
    (switch_arrows pinds)
    (set_ptr pinds)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv))
    (\<lambda>ea ec. force_same_and abs4_abs3_strel ea ec \<inter>
     (if ea = Degraded (EB True)
        then (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv
        else if \<exists>v. ea = Degraded v
            then (all_reliable_inv \<inter> abs_updated_not_SL) \<times> abs3_norm_inv
            else UNIV )
  )
    abs4_abs3_strel
  "
proof-
  note hint = set_ptr_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NN_inv[OF switch_arrows_post_NN set_ptr_inv_NN],
    THEN sim_ext_add_NE_inv[OF switch_arrows_post_NE set_ptr_inv_NE],
    THEN sim_ext_mono, THEN sim_ext_mono_norm, THEN sim_ext_mono_exc]
  show ?thesis
    apply(rule hint[where R3="abs4_abs3_strel \<inter> UNIV \<times> precommit_inv"])
    apply(auto simp add: K_def force_same_and_def precommit_inv_def)
    done
qed
 
lemma do_commit_ref:
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
    (Abs4_UpdateDefs.do_commit) 
    (Abs3_UpdateDefs.do_commit)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times>
        (ptr_eq_updated \<inter> abs3_norm_inv)))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
proof-
  have ref':
    "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)
    (Abs4_UpdateDefs.do_commit) 
    (Abs3_UpdateDefs.do_commit)
    (force_same_and (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
    "
    apply(simp only: Abs4_UpdateDefs.do_commit_def Abs3_UpdateDefs.do_commit_def)
    apply(rule sim_ext_bindI[OF get_updated_commit_ref])
    apply(clarsimp)
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel)
    apply(rule sim_ext_try_catchI[OF set_ptr_commit_ref])
    apply(auto split: exceptions.split intro!: sim_ext_return sim_ext_throw sim_exts_empty_rel
      simp add: force_same_and_same force_same_and_diff')
    apply(subst force_same_and_def)
    apply(auto intro: sim_exts_empty_rel)
    apply(rule rollback_ref[THEN sim_ext_mono], auto)
    done

  show ?thesis
    apply(rule ref'[THEN sim_ext_mono,
      THEN sim_ext_add_NN_inv[OF true_postNN do_commit_post_NN],
      THEN sim_ext_add_NE_inv[OF true_postNE do_commit_inv_NE],
      THEN sim_ext_mono_norm,
      THEN sim_ext_mono_exc,
      THEN sim_ext_mono,
      where R3="(abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SL) \<times> precommit_inv)" and I3=UNIV
      ]
    )
    apply(auto simp add: K_def force_same_and_def precommit_inv_def)
    done
qed

lemma get_updated_rollforward_ref:
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv)
   get_abs_updated get_updated
   (\<lambda>va. force_same_and
          (abs4_abs3_strel \<inter>
           (all_reliable_inv \<inter> abs_updated_not_SR \<inter> abs_updated_equals va) \<times>
           abs3_norm_inv)
          va)
   (force_same_and abs4_abs3_strel) abs4_abs3_strel"
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(clarsimp_all simp add: brute_force
    abs4_abs3_strel_def all_reliable_inv_def abs_updated_not_SR_def abs_updated_equals_def abs3_inv_defs)
  done

lemma get_arrow_rollforward_ref:
  "
  \<lbrakk>ord \<in> order_set pinds; ind \<in> set ord\<rbrakk> \<Longrightarrow>
  sim_ext
  (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds)) \<times> abs3_norm_inv)
    (get_arrow ind) (get_sync_direction ind)
    (\<lambda>va vc. if va = SR \<or> va = SBot \<or> va \<noteq> vc then {} else
        abs4_abs3_strel \<inter>
            (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds) \<inter> arrow_is va ind)
            \<times> abs3_norm_inv)
    (force_same_and abs4_abs3_strel) abs4_abs3_strel
  "
  apply(rule
    sim_ext_add_NN_inv[OF _ true_postNN get_arrow_ref''[THEN sim_ext_mono], THEN sim_ext_mono_norm
    , where I'1="K (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds))"
    ]
  )
  apply(auto simp add: K_beta abs_updated_equals_def abs_updated_not_SR_def arrow_is_def order_set_set
    force_same_and_def HoareNN_alt Hoare_def Box_def)
  apply(auto simp add: all_reliable_inv_def get_arrow_defs brute_force min_def split: split_if_asm)
  done

lemma maybe_mess_up_admin_rollforward_post_NN:
  "\<lbrace>(all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds))\<rbrace>
    Abs4_CommonDefs.maybe_mess_up_admin ind
  \<lbrace>>N K (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds))\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(simp_all only: maybe_mess_up_admin_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: sync_defs brute_force all_reliable_inv_def abs_updated_not_SR_def
    get_arrow_defs min_def
    abs_updated_equals_def insert_UNIV split: split_if_asm)
  apply(auto)
  done

lemma remove_from_ptr_silence_deg_inv_NN:
 "\<lbrace>abs3_norm_inv\<rbrace> remove_from_ptr_silence_deg ind \<lbrace>>N K abs3_norm_inv\<rbrace>"
  apply(auto simp add: remove_from_ptr_silence_deg_def
    K_beta HoareNN_return' HoareNE_throw'
    intro!: try_catch_ruleNN try_catch_ruleNE bind_ruleNN bind_ruleNE HoareNN_throw HoareNE_return
    sync_inv_NN
    sync_inv_NE
    remove_from_ptr_inv_NN
    remove_from_ptr_inv_NE
    split: exceptions.split
  )
  done
 
lemma sync_if_not_SEq_and_remove_from_ptr_rollforward_ref:
  "
  \<lbrakk>ord \<in> order_set pinds; ind \<in> set ord\<rbrakk> \<Longrightarrow>
    sim_ext
        (abs4_abs3_strel \<inter>
            (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds)) \<times>
            abs3_norm_inv)
        (sync_by_arrow_silence_deg ind)
        (sync_if_not_SEq_remove_from_ptr Primary Secondary ind)
        (force_same_and
             (abs4_abs3_strel \<inter>
              (all_reliable_inv \<inter> (abs_updated_not_SR \<inter> abs_updated_equals pinds)) \<times>
              abs3_norm_inv))

        (force_same_and abs4_abs3_strel)
        abs4_abs3_strel
  "
  apply(simp only: sync_by_arrow_silence_deg_def sync_if_not_SEq_remove_from_ptr_def check_arrow_SEq_def)
  apply(simp only: first_stage)
  apply(rule sim_ext_bindI)
  apply(rule get_arrow_rollforward_ref)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sync_and_remove_from_ptr_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NN_inv[OF
        just_sync_silence_deg'_rollforward_post_NN
        sync_and_remove_from_ptr_silence_deg_inv_NN],
    THEN sim_ext_mono,
    THEN sim_ext_mono_norm])
  apply(auto simp add: K_beta force_same_and_same)
  apply(rule remove_from_ptr_silence_deg_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NN_inv[OF
        maybe_mess_up_admin_rollforward_post_NN 
        remove_from_ptr_silence_deg_inv_NN],
    THEN sim_ext_mono,
    THEN sim_ext_mono_norm
    , where ind3=ind and R2="abs4_abs3_strel \<inter> arrow_is SEq ind \<times> UNIV"]
  )
  apply(auto simp add: K_beta force_same_and_same)
  apply(metis sync_direction.exhaust)
  done

lemma abs4_rollback_post_NN:
  "\<lbrace>P\<rbrace> Abs4_UpdateDefs.rollback \<lbrace>>N Q\<rbrace>"
  by(auto simp add: Abs4_UpdateDefs.rollback_def
    intro!: bind_ruleNN HoareNN_throw
    intro: true_postNN
  )

lemma rollforward_ref:
  "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> (ptr_eq_updated \<inter> abs3_norm_inv))
    Abs4_UpdateDefs.rollforward 
    Abs3_UpdateDefs.rollforward 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (empty_ptr \<inter> abs3_norm_inv)))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
  "
proof-
  have ref':
    "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv)
    Abs4_UpdateDefs.rollforward 
    Abs3_UpdateDefs.rollforward 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> abs3_norm_inv))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
    "
    apply(simp only: Abs4_UpdateDefs.rollforward_def sync_updated_def 
      Abs3_UpdateDefs.rollforward_def first_stage)
    apply(rule sim_ext_bindI[OF get_updated_rollforward_ref])
    apply(clarsimp)
    apply(subst force_same_and_def)
    apply(auto simp add: K_beta intro: sim_exts_empty_rel)
    apply(rule sim_ext_bindI[OF sim_ext_chooseN[THEN sim_ext_add_NN_inv[OF chooseN_ruleNN true_postNN]]])
    apply(clarsimp)
    apply(subst force_same_and_def)
    apply(auto simp add: K_beta intro: sim_exts_empty_rel)
    apply(rule sim_ext_bindI)
    apply(rule sim_ext_mapM')
    apply(clarsimp)
    apply(simp add: Int_assoc Sigma_Int_distrib2[THEN sym])
    apply(rule sync_if_not_SEq_and_remove_from_ptr_rollforward_ref)
    apply(auto)
    apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
    apply(clarsimp_all simp add: all_reliable_inv_def abs3_inv_defs brute_force abs4_abs3_strel_def)
    done

  have ref'':
    "sim_ext (abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> (ptr_eq_updated \<inter> abs3_norm_inv))
    Abs4_UpdateDefs.rollforward 
    Abs3_UpdateDefs.rollforward 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (empty_ptr \<inter> abs3_norm_inv)))
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
    "
    apply(rule ref'[THEN sim_ext_mono,
      THEN sim_ext_add_NN_inv[OF true_postNN rollforward_post_NN],
      THEN sim_ext_mono,
      THEN sim_ext_mono_norm,
      where R2="(abs4_abs3_strel \<inter> (all_reliable_inv \<inter> abs_updated_not_SR) \<times> abs3_norm_inv)"
      and I2=UNIV
      ]
    )
    apply(auto simp add: K_def force_same_and_def precommit_inv_def)
    apply(auto simp add: abs4_abs3_strel_def abs3_capacity_inv_def)
    done

  show ?thesis
    apply(rule ref''[THEN sim_ext_mono,
      THEN sim_ext_add_NE_inv[OF true_postNE rollforward_inv_NE],
      THEN sim_ext_mono,
      THEN sim_ext_mono_exc,
      where R2="(abs4_abs3_strel \<inter>
      (all_reliable_inv \<inter> abs_updated_not_SR) \<times> (ptr_eq_updated \<inter> abs3_norm_inv))"
      and I2=UNIV
      ]
    )
    apply(auto simp add: K_def force_same_and_def precommit_inv_def)
    done
qed

lemma throw_mode_error_pre_commit_ref:
  "sim abs4_abs3_strel (throw_abs_mode_error val) (PmmDefs.throw_mode_error val)"
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all only: PmmDefs.throw_mode_error_def throw_abs_mode_error_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: brute_force
    abs4_abs3_strel_def abs3_inv_defs all_reliable_inv_def
    split: split_if_asm
  )
  apply(auto)
  apply (metis abs4_card_mode.cases abs4_card_mode.simps(4) abs_card_mode.distinct(5))
  apply (metis abs4_card_mode.cases abs4_card_mode.simps(4) abs_card_mode.distinct(3))
  done
    
lemma check_size_ref:
  "sim R (PmmDefs.check_sizes upd_f) (PmmDefs.check_sizes upd_f)"
  by(auto simp add: PmmDefs.check_sizes_def
    intro!: sims_all_intros)

lemma check_size_ext_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> precommit_inv))
    (PmmDefs.check_sizes upd_f) (PmmDefs.check_sizes upd_f) 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> precommit_inv)))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv)) 
    abs4_abs3_strel"
  apply(auto simp add: PmmDefs.check_sizes_def PmmDefs.check_sizes_def)
  apply(rule sim_ext_bindI[where Tn="\<lambda>va vc. 
    if va = vc \<and> vc \<in> order_set (dom (upd_f)) 
        then (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> precommit_inv)) 
        else {}"])
  apply(force simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force chooseN_def)
  apply(auto intro!: sim_exts_empty_rel sim_ext_mapM')
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force chooseN_def order_set_set)
  done

definition abs3_all_reliable where
  "abs3_all_reliable \<equiv> {(sv, sp). card_mode sv = CNormal \<longrightarrow> 
    (\<forall>ind inst. (ind, inst) \<in> abs3_reliable sp)}"

lemma move_rel_to_abs3:
  "abs4_abs3_strel \<inter> all_reliable_inv \<times> J
  = abs4_abs3_strel \<inter> UNIV \<times> (abs3_all_reliable \<inter> J)"
  apply(auto simp add: abs4_abs3_strel_def all_reliable_inv_def abs3_all_reliable_def)
  apply(case_tac "card_mode aa", simp_all)
  done

lemma update_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (Abs4_UpdateDefs.update upd_f) 
    (Abs3_UpdateDefs.update upd_f) 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    abs4_abs3_strel
  "
proof-
    
  have ref':
    "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (Abs4_UpdateDefs.update upd_f) 
    (Abs3_UpdateDefs.update upd_f) 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (empty_ptr \<inter> abs3_norm_inv)))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
    "
    apply(simp only: Abs4_UpdateDefs.update_def Abs3_UpdateDefs.update_def)
    apply(rule sim_ext_bindI[where
      Tn="force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> precommit_inv))"])
    defer
    apply(clarsimp simp add: force_same_and_same)
    apply(rule sim_ext_bindI)
    thm sim_ext_mono_norm
    apply(rule check_size_ext_ref[THEN sim_ext_mono_norm[where
        Sn'="force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> precommit_inv)"]])
    apply(clarsimp simp add: force_same_and_same)
    apply(auto simp add: force_same_and_same intro: sim_exts_empty_rel)[1]
    apply(rule sim_ext_bindI[OF precommit_ref])
    apply(clarsimp simp add: force_same_and_same)
    apply(rule sim_ext_bindI[OF do_commit_ref])
    apply(clarsimp simp add: force_same_and_same)
    apply(rule sim_ext_bindI[OF rollforward_ref])
    apply(auto simp add: force_same_and_same)
    apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
    apply(simp_all only: throw_abs_mode_error_def PmmDefs.throw_mode_error_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
    apply(clarsimp_all simp add: brute_force
      abs4_abs3_strel_def abs3_inv_defs all_reliable_inv_def
      precommit_inv_def empty_ptr_def PmmDefs.sv_mode_equals_def
      split: split_if_asm
    )
    apply(iterative_simp)+
    apply(case_tac[!] "card_mode aa")
    apply(clarsimp_all)
    done

  hence ref'':
    "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> (abs3_all_reliable \<inter> (abs3_inv \<inter> empty_ptr)))
    (Abs4_UpdateDefs.update upd_f) 
    (Abs3_UpdateDefs.update upd_f) 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (empty_ptr \<inter> abs3_norm_inv)))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_inv))
    abs4_abs3_strel
    "
    apply(subst (asm) move_rel_to_abs3)
    apply(assumption)
    done

  show ?thesis
    thm ref''[THEN sim_ext_mono
      , THEN sim_ext_add_NE_inv[OF true_postNE update_mode_post_NE, rotated 1]
      , THEN sim_ext_mono, THEN sim_ext_mono_exc, THEN sim_ext_mono_norm
      , where R3="abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> empty_ptr)" and I3=UNIV
        and J3="(abs3_inv \<inter> empty_ptr \<inter> abs3_capacity_inv)"
      ]
    apply(rule ref'[THEN sim_ext_mono
      , THEN sim_ext_add_NE_inv[OF true_postNE update_mode_post_NE, rotated 1]
      , THEN sim_ext_mono, THEN sim_ext_mono_exc, THEN sim_ext_mono_norm
      , where R3="abs4_abs3_strel \<inter> UNIV \<times> (abs3_all_reliable \<inter> (abs3_inv \<inter> empty_ptr))" and I3=UNIV
        and J3="(abs3_all_reliable \<inter> abs3_inv \<inter> empty_ptr \<inter> abs3_capacity_inv)"
      ]
    )
    apply(auto simp add: move_rel_to_abs3)[3]
    apply(clarsimp simp add: abs3_capacity_inv_def abs4_abs3_strel_def)
    apply(auto simp add: K_def force_same_and_def PmmDefs.sv_mode_neq_def all_reliable_inv_def 
      empty_ptr_def
      abs4_abs3_strel_def
      abs3_norm_inv_def)
    apply(case_tac[!] "card_mode aa", simp_all add: abs3_all_reliable_def)
    done
qed

lemma update_full_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (<tryR>
        Abs4_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (<tryR>
        Abs3_UpdateDefs.update upd_f   
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (abs4_abs3_strel)
  "
  by(intro sim_ext_try_catchRI[OF update_ref] sim_ext_rec_tryRI startup_ref)

lemma get_mode_ref:
  "R \<subseteq> abs4_abs3_strel \<Longrightarrow>
    sim_ext R get_abs_mode PmmDefs.get_mode
    (\<lambda>va vc. (R \<inter> sv_abs_mode_equals va \<times> sv_mode_equals vc)
        \<inter> (if va = abs4_card_mode vc then UNIV else {}))
    Se
    Sr
  "
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force
    sv_abs_mode_equals_def PmmDefs.sv_mode_equals_def abs4_abs3_strel_def)

lemma HoareNE_Int:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>@N K Q\<rbrace>; \<lbrace>P'\<rbrace> m \<lbrace>@N K Q'\<rbrace> \<rbrakk> \<Longrightarrow> \<lbrace>P \<inter> P'\<rbrace> m \<lbrace>@N K (Q \<inter> Q')\<rbrace>"
  by(force simp add: HoareNE_alt Hoare_def Box_def K_def)

lemma fetch_same_ref:
  "sim_ext (abs4_abs3_strel \<inter> (UNIV \<times> abs3_norm_inv))
    (Abs4_FetchDefs.fetch_same pind)
    (Abs3_FetchDefs.fetch_same pind)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> PmmDefs.sv_mode_neq CNormal)))
    abs4_abs3_strel
  "
proof-
  have ref':
    "sim_ext (abs4_abs3_strel \<inter> (UNIV \<times> abs3_norm_inv))
    (Abs4_FetchDefs.fetch_same pind)
    (Abs3_FetchDefs.fetch_same pind)
    (force_same_and abs4_abs3_strel)
    (force_same_and abs4_abs3_strel)
    abs4_abs3_strel
    "
    apply(simp_all only: 
      Abs4_FetchDefs.fetch_same_def
      Abs3_FetchDefs.fetch_same_def
    )
    apply(rule sim_ext_try_catchI)
    apply(rule read_ref'[THEN sim_ext_mono_norm])
    apply(simp add: force_same_and_subset)
    apply(auto simp add: force_same_and_diff' intro!: sim_exts_empty_rel split: exceptions.split)
    apply(tactic {* TRYALL (refinement_tac false Ref_Tac.default_simp @{context}) *})
    apply(simp_all)
    apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
    apply(rule sim_ext_mono, auto)
    done

  note inv = HoareNE_Int[OF fetch_same_inv_NE Abs3_Invariant_Lemmas.fetch_same_mode_post_NE, simplified]
  thm inv
  show ?thesis
    thm sim_ext_add_NN_inv
    apply(rule sim_ext_mono_exc)
    thm ref'[THEN sim_ext_mono
        , THEN sim_ext_add_NN_inv[OF true_postNN fetch_same_inv_NN]
        , THEN sim_ext_mono
        , THEN sim_ext_add_NE_inv[OF true_postNE inv]
        , THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
    ]
    thm sim_ext_mono_exc
    apply(subst strel_contains_capacity_inv) 
    apply(rule ref'[THEN sim_ext_mono
        , THEN sim_ext_add_NN_inv[OF true_postNN fetch_same_inv_NN]
        , THEN sim_ext_mono
        , THEN sim_ext_add_NE_inv[OF true_postNE inv]
        , THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
      ]
    )
    apply(rule order_refl)
    apply(simp add: strel_contains_capacity_inv[THEN sym])
    apply(simp add: force_same_and_def K_def)
    apply(simp add: K_beta)
    apply(intro allI)  
    apply(rule order_refl)
    apply(auto simp add: force_same_and_def)
    done
qed

lemma pmm_sv_mode_eq_neq_subset: "m \<noteq> m' \<Longrightarrow> PmmDefs.sv_mode_equals m \<subseteq> PmmDefs.sv_mode_neq m'"
  by(auto simp add: PmmDefs.sv_mode_equals_def PmmDefs.sv_mode_neq_def)

lemma get_arrow_ref''':
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv) (get_arrow pind)
  (get_sync_direction pind)
  (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
  (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> PmmDefs.sv_mode_neq CNormal)))
  abs4_abs3_strel
  "
proof-

  note invNE =
    HoareNE_Int[OF get_sync_direction_inv_NE get_sync_direction_fails_post_NE[where P=UNIV], simplified]
  note invNN = get_sync_direction_inv_NN
  show ?thesis
    apply(rule get_arrow_ref'[THEN sim_ext_mono,
      THEN sim_ext_add_NE_inv[OF true_postNE invNE],
      THEN sim_ext_add_NN_inv[OF true_postNN invNN],
      THEN sim_ext_mono, THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
      ]
    )
    apply(auto simp add: K_beta force_same_and_def)
    apply(force intro!: pmm_sv_mode_eq_neq_subset[THEN subsetD])
    done
qed
  
declare get_arrow_ref'''[ref_thm]

thm read_or_fail_ref
declare read_or_fail_ref[ref_thm del]
lemma read_or_fail_ref'[ref_thm]:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (Abs4_CommonDefs.read_or_fail ind inst)
    (Abs3_CommonDefs.read_or_fail ind inst)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> PmmDefs.sv_mode_neq CNormal)))
    abs4_abs3_strel
  "
proof-
  have inv1:
  "\<lbrace>abs3_norm_inv\<rbrace> Abs3_CommonDefs.read_or_fail ind inst \<lbrace>@N K abs3_inv \<rbrace>"
    apply(intro conseqNE_rule[OF order_refl read_or_fail_inv_NE])
    apply(auto simp add: K_beta intro!: norm_inv_subset_inv del: subsetI)
    done

  note invNE = HoareNE_Int[OF inv1 
    Abs3_Invariant_Lemmas.read_or_fail_mode_post_NE[where P=UNIV], simplified]
  show ?thesis
  apply(rule read_or_fail_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NE_inv[OF true_postNE invNE],
    THEN sim_ext_add_NN_inv[OF true_postNN read_or_fail_inv_NN],
    THEN sim_ext_mono, THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
    ]
  )
  apply(auto simp add: K_beta force_same_and_def)
  done
qed

declare get_mode_ref[ref_thm]
declare fetch_same_ref[ref_thm]

(*
lemma throw_mode_error_mode_post_NE:
  "\<lbrace>P\<rbrace> throw_mode_error val \<lbrace>@N K (sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_def)
  apply(simp_all only: throw_mode_error_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: sv_mode_neq_def brute_force split: split_if_asm)
  done
  *)

lemma throw_mode_error_mode_post_NE:
  "\<lbrace>P\<rbrace> PmmDefs.throw_mode_error val \<lbrace>@N K (PmmDefs.sv_mode_neq CNormal)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_def)
  apply(simp_all only: PmmDefs.throw_mode_error_def)
  apply(simp_all add: first_stage vol_ops_project_simps cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: PmmDefs.sv_mode_neq_def brute_force split: split_if_asm)
  done

declare throw_mode_error_ref[ref_thm del]

lemma throw_mode_error_ref'[ref_thm]:
  "sim_ext (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)
    (throw_abs_mode_error val)
    (PmmDefs.throw_mode_error val)
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> PmmDefs.sv_mode_neq CNormal)))
    abs4_abs3_strel
  "
  apply(rule throw_mode_error_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NE_inv[OF true_postNE throw_mode_error_mode_post_NE[where P=abs3_norm_inv]],
    THEN sim_ext_mono, THEN sim_ext_mono_exc])
  apply(auto simp add: force_same_and_def K_def)
  done

lemma fetch_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (Abs4_FetchDefs.fetch ind) 
    (Abs3_FetchDefs.fetch ind) 
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv))
    (force_same_and (abs4_abs3_strel \<inter> UNIV \<times> (abs3_inv \<inter> PmmDefs.sv_mode_neq CNormal)))
    abs4_abs3_strel
  "
  apply(auto simp add: Abs4_FetchDefs.fetch_def Abs3_FetchDefs.fetch_def)
  apply(rule sim_ext_bindI[OF get_mode_ref])
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF sim_ext_throw_diffI, where Tn="\<lambda>_ _. {}"])
  apply(auto intro!: sim_exts_empty_rel simp add: force_same_and_same)
  apply(force intro!: pmm_sv_mode_eq_neq_subset[THEN subsetD])
  apply(case_tac vc, auto)
  apply(rule sim_ext_bindI[OF sim_ext_return_diffI,
    where Tn="force_same_and (abs4_abs3_strel \<inter> UNIV \<times> abs3_norm_inv)"])
  apply(force intro!: simp add: force_same_and_same abs3_inv_defs PmmDefs.sv_mode_equals_def)
  apply(auto simp add: force_same_and_same)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(simp_all)
  done

lemma fetch_ref_with_invs:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (Abs4_FetchDefs.fetch ind) 
    (Abs3_FetchDefs.fetch ind) 
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    abs4_abs3_strel
  "
  apply(rule fetch_ref[THEN sim_ext_mono,
    THEN sim_ext_add_NN_inv[OF fetch_reliable_inv fetch_empty_ptr],
    THEN sim_ext_mono, THEN sim_ext_mono_norm, THEN sim_ext_mono_exc,
      where R3="abs4_abs3_strel \<inter> UNIV \<times> abs3_inv"
    ]
  )
  apply(auto simp add: K_def force_same_and_def abs3_norm_inv_def
    PmmDefs.sv_mode_neq_def empty_ptr_def all_reliable_inv_def abs4_abs3_strel_def)
  apply(case_tac "card_mode aa", simp_all)
  done

lemma fetch_full_ref:
  "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (<tryR>
        Abs4_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (<tryR>
        Abs3_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
    (abs4_abs3_strel)
  "
  by(intro sim_ext_try_catchRI[OF fetch_ref_with_invs] sim_ext_rec_tryRI startup_ref)

lemma abs4_startup_full_ref_N:
  "sim_norm (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (<tryR>
        Abs4_StartupDefs.startup
    <catchR>
        rec_tryR Abs4_StartupDefs.startup
    )
    (<tryR>
        Abs3_StartupDefs.startup
    <catchR>
        rec_tryR Abs3_StartupDefs.startup
    )
  "
  apply(unfold sim_norm_def)
  apply(rule sim_ext_imp_sim_norm_ext)
  apply(rule sim_ext_try_catchRI)
  apply(rule sim_ext_mono)
  apply(rule Abs4_Ref.startup_ref)
  apply(simp)
  apply(rule sim_ext_rec_tryRI[OF Abs4_Ref.startup_ref])
  done

lemma abs4_fetch_full_ref_N:
  "sim_norm (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (<tryR>
        Abs4_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (<tryR>
        Abs3_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
  "
  by(auto simp add: sim_norm_def intro!: sim_ext_imp_sim_norm_ext Abs4_Ref.fetch_full_ref)

lemma abs4_update_full_ref_N:
  "sim_norm (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
    (<tryR>
        Abs4_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (<tryR>
        Abs3_UpdateDefs.update upd_f   
    <catchR>
        rec_tryR (Abs3_StartupDefs.startup)
    )
  "
  by(auto simp add: sim_norm_def intro!: sim_ext_imp_sim_norm_ext Abs4_Ref.update_full_ref)

end
