theory Abs4_SpecRef
  imports "../abs3/Abs3_Specification" Abs4_Specification Abs4_Ref "../SpecRef_Lemmas"
begin

definition abs3_obs_reliable :: "'c abs3_obs_state \<Rightarrow> (pub_indx \<times> INST)set" where
  "abs3_obs_reliable spc \<equiv> {(ind, inst).
    hw_state spc ind inst \<noteq> IDead
    (* \<and> hstate spc \<noteq> LDead *)
    \<and> pub_hashes spc \<noteq> None
    \<and> the (pub_hashes spc) (ind, inst) = Hash (abs3_obs_state.memory spc ind inst)
  }
  "

definition abs3_obs_writable :: "'c abs3_obs_state \<Rightarrow> (pub_indx \<times> INST)set" where
  "abs3_obs_writable spc \<equiv> {(ind, inst).
    hw_state spc ind inst = IOK
    \<and> hstate spc = LOK
  }
  "

definition abs3_obs_arrow :: "'c abs3_obs_state \<Rightarrow> pub_indx \<Rightarrow> sync_direction" where
  "abs3_obs_arrow spc \<equiv> (\<lambda>ind.
    (* if hstate spc = LDead \<or> ptr_state spc = LDead
        \<or> pub_hashes spc = None \<or> ptr spc = None *)
    if pub_hashes spc = None \<or> ptr spc = None
    then SBot
    else
        if the (pub_hashes spc) (ind, Primary) = the (pub_hashes spc) (ind, Secondary)
        then SEq
        else
            if ind \<in> the (ptr spc)
            then SL
            else SR
  )"

fun abs3_abs4_pi
  :: "(pmm_vol_mem \<times> 'c abs3_obs_state) \<Rightarrow> (abs_vol_mem \<times> 'c :: contents_cl abs4_state)" where
  "abs3_abs4_pi (svc, spc) = (
    \<lparr> abs_card_mode = abs4_card_mode (PmmDefs.card_mode svc), abs_updated = updated svc \<rparr>,
        \<lparr>
            arrow = abs3_obs_arrow spc,
            reliable = abs3_obs_reliable spc,
            writable = abs3_obs_writable spc,
            memory = abs3_obs_state.memory spc,
            admin_state = min (hstate spc) (ptr_state spc)
        \<rparr>
    )"
 
(* TODO: move this to Refinement.thy *)
lemma sim_norm_ext_try_catchRI:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; sim_norm_ext Sr ha hc Sn Se \<rbrakk> \<Longrightarrow>
    sim_norm_ext R (try_catchR ma ha) (try_catchR mc hc) Sn Se"
  by(auto simp add: sim_exts_defs all_conj project_simps
    intro!: simN_wpc_choice_lockstep simS_wpc_choice_lockstep
    simN_wpc_bindN_set simS_wpc_bindN_set simN_wpc_bindS
    del: subsetI)
 
theorem abs3_refines_abs4:
  "refines 
    (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)) 
    abs3_abs4_pi abs4_spec abs3_spec"
proof(intro refinesI)
  show "obs_consistent (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
        abs3_abs4_pi abs4_spec abs3_spec"
    apply(auto simp add: obs_consistent_def abs4_spec_def abs3_spec_def 
      abs4_abs3_strel_def obs_ts min_def 
      abs3_arrow_def abs3_obs_arrow_def abs3_reliable_def abs3_obs_reliable_def
      abs3_writable_def abs3_obs_writable_def
    )
    done
next
  show "PO_refines (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)) abs4_spec abs3_spec"
  proof(rule refine_basic)
    show "init_trans abs3_spec `` init abs3_spec
        \<subseteq> (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)) `` 
            init_trans abs4_spec `` init abs4_spec"
      apply(simp add: abs3_spec_def abs3_ts_def abs4_spec_def abs4_ts_def TS.defs spec.defs)
      apply(rule sim_norm_init_inclusion)
      apply(rule Abs4_Ref.startup_ref[THEN sim_ext_imp_sim_norm_ext])
      apply(auto simp add: abs4_abs3_strel_def abs4_init_def abs3_reliable_def abs3_arrow_ref_simps
        in_view_def abs3_init_def)
      apply(auto simp add: min_def split: split_if_asm)
      apply (metis abs3_arrow_SBot_iff arrow_srcs.simps(4) ex_in_conv)+
      done
  next
    have restart_ref:
      "sim_ext (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr))
      restart_mangle_sv restart_mangle_sv
      (force_same_and
      (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
      (force_same_and (abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)))
      abs4_abs3_strel
      "
      by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force
        abs4_abs3_strel_def)

    show "{abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)} 
            TS.trans abs4_spec, TS.trans abs3_spec 
        {> abs4_abs3_strel \<inter> all_reliable_inv \<times> (abs3_inv \<inter> empty_ptr)}"
      apply(simp add: abs4_spec_def abs4_ts_def abs3_spec_def abs3_ts_def spec.defs TS.defs 
        abs4_trans_def abs3_trans_def)
      apply(intro relhoare_UN_lockstep allI relhoare_set_fun_choice_lockstep)
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare 
        Abs4_Ref.fetch_full_ref[THEN sim_ext_imp_sim_norm_ext]
        Abs4_Ref.update_full_ref[THEN sim_ext_imp_sim_norm_ext]
        abs4_startup_full_ref_N[ simplified sim_norm_def]
        sim_norm_ext_try_catchRI[OF restart_ref]
        Abs4_Ref.startup_ref[THEN sim_ext_rec_tryRI, THEN sim_ext_imp_sim_norm_ext]
        simp add: try_rec_tryR_def)
      done
  qed
qed

theorem abs3_implements_abs4:
  "implements abs3_abs4_pi abs4_spec abs3_spec"
  by(auto intro: refinement_soundness abs3_refines_abs4)

end