theory Abs4_UpdateDefs
imports Abs4_CommonDefs
begin

definition write_new'
  :: "(pub_indx \<rightharpoonup> 'c) \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "write_new' upd_f pind \<equiv> do {
    inst \<leftarrow> chooseN UNIV;
    write_new pind inst ((the o upd_f) pind)
  }"

definition sync_by_arrow_silence_deg
  :: "pub_indx \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "sync_by_arrow_silence_deg ind \<equiv> do {
    arr \<leftarrow> get_arrow ind;
    if arr = SEq
    then maybe_mess_up_admin ind
    else if arr = SL
    then just_sync_silence_deg' Primary Secondary ind
    else if arr = SR
    then just_sync_silence_deg' Secondary Primary ind
    else return ()
  }"

definition sync_updated
  :: "(unit, 'c :: size)abs4_monad" where
  sync_updated_def[monad_def]:
  "sync_updated \<equiv> do {
    abs_updated_s \<leftarrow> get_abs_updated;
    abs_updated \<leftarrow> chooseN (order_set abs_updated_s);
    mapM' sync_by_arrow_silence_deg abs_updated;
    set_abs_updated {}
  }"

definition rollback 
  :: "(unit, 'c :: size)abs4_monad" where
  rollback_def[monad_def]:
  "rollback \<equiv> do {
    sync_updated;
    throw (Degraded (EB False))
  }"

definition pre_commit
  :: "(pub_indx \<rightharpoonup> 'c :: size) \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  pre_commit_def[monad_def]:
  "pre_commit upd_f \<equiv> do {
    set_abs_updated {};
    ord \<leftarrow> chooseN (order_set (dom upd_f));
    <try>
        mapM' (write_new' upd_f) ord
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> rollback
        | _ \<Rightarrow> throw e)
  }"

fun switch_arrow :: "(pub_indx \<Rightarrow> sync_direction) \<Rightarrow> pub_indx \<Rightarrow> sync_direction" where
  "switch_arrow arr_f pind = (case arr_f pind of
    SL \<Rightarrow> SR
    | SR \<Rightarrow> SL
    | _ \<Rightarrow> SEq
  )"

definition switch_arrows_success
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "switch_arrows_success pinds \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        admin_state sp = LOK
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (switch_arrow (arrow sp)) pinds
        \<rparr>
    }"

definition switch_arrows_degrade
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "switch_arrows_degrade pinds \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded) exc, sv', sp')|sv' sp' unwr exc.
        sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> exc \<in> {EU, EB False}
        \<and> LDegraded \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            writable := writable sp - unwr,
            admin_state := LDegraded
        \<rparr>
    }"

definition switch_arrows_degrade_write
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "switch_arrows_degrade_write pinds \<equiv> \<lambda>(sv, sp).
    {((Some o Inl o Degraded o EB) True, sv', sp')|sv' sp' unwr.
        sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> LDegraded \<le> admin_state sp
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (switch_arrow (arrow sp)) pinds,
            writable := writable sp - unwr,
            admin_state := LDegraded
        \<rparr>
    }"

definition switch_arrows_fail
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "switch_arrows_fail pinds \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' arr_del unr unwr.
        sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            arrow := override_on (arrow sp) (K SBot) arr_del,
            reliable := reliable sp - unr,
            writable := writable sp - unwr,
            admin_state := LDead
        \<rparr>
    }"

definition switch_arrows
  :: "pub_indx set \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  "switch_arrows pinds \<equiv>
    switch_arrows_success pinds
    [+] switch_arrows_degrade pinds
    [+] switch_arrows_degrade_write pinds
    [+] switch_arrows_fail pinds
    [+] restart_mangle_sv
    [+] do { switch_arrows_success pinds; restart_mangle_sv }
  "

lemmas switch_arrows_defs =
  switch_arrows_def
  switch_arrows_success_def
  switch_arrows_degrade_def
  switch_arrows_degrade_write_def
  switch_arrows_fail_def

(* "commit" seems to be a keyword *)
definition do_commit :: "(unit, 'c :: size)abs4_monad" where
  do_commit_def[monad_def]:
  "do_commit \<equiv> do {
    pinds \<leftarrow> get_abs_updated;
    <try> do {
        switch_arrows pinds
    } <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then return ()
            else rollback
        | _ \<Rightarrow> throw e
    )
  }
  "

definition rollforward where
  rollforward_def[monad_def]:
  "rollforward \<equiv> sync_updated"

definition update :: "(pub_indx \<rightharpoonup> 'c :: size) \<Rightarrow> (unit, 'c :: size)abs4_monad" where
  update_def[monad_def]:
  "update upd_f \<equiv> do {
    throw_abs_mode_error (EB False);
    check_sizes upd_f;
    pre_commit upd_f;
    do_commit;
    rollforward;
    throw_abs_mode_error (EB True)
  }"

end
