theory Abs5_CommonDefs
  imports Abs5Defs
begin

definition read_success :: "pub_indx \<Rightarrow> ('c, 'c)abs5_monad" where
  "read_success ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inr cnt), sv, sp)|cnt.
        reg_state sp ind \<ge> LDegraded
        \<and> cnt = memory sp ind
  }"
        
definition read_degrade :: "pub_indx \<Rightarrow> ('c, 'c)abs5_monad" where
  "read_degrade ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EC cnt))), sv', sp')|cnt deg sv' sp' a_st.
        reg_state sp ind \<ge> LDegraded
        \<and> cnt = memory sp ind
        \<and> ind \<in> deg
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> a_st \<le> admin_state sp
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
            admin_state := a_st
        \<rparr>
  }"

definition read_fail :: "pub_indx \<Rightarrow> ('c, 'c)abs5_monad" where
  "read_fail ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' reg_state' adm_state'.
        (reg_state' ind = LDead \<or> adm_state' = LDead)
        \<and> (\<forall>ind'. reg_state' ind' \<le> reg_state sp ind')
        \<and> adm_state' \<le> admin_state sp
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := reg_state',
            admin_state := adm_state'
        \<rparr>
  }"

definition restart_mangle_sp :: "('a, 'c)abs5_monad" where
  "restart_mangle_sp \<equiv> \<lambda>(sv, sp).
    {(None, sv', sp')|sv' sp' reg_state' adm_state'.
        (\<forall>ind. reg_state' ind \<le> reg_state sp ind)
        \<and> adm_state' \<le> admin_state sp
        \<and> sp' = sp\<lparr> reg_state := reg_state', admin_state := adm_state' \<rparr>
    }"

definition read :: "pub_indx \<Rightarrow> ('c, 'c)abs5_monad" where
  "read ind \<equiv>
    read_success ind
    [+] read_degrade ind
    [+] read_fail ind
    [+] restart_mangle_sv
    [+] restart_mangle_sp
    [+] try_catch (read_degrade ind) (K restart_mangle_sv)
    [+] try_catch (read_fail ind) (K restart_mangle_sv)
  "

lemmas read_defs =
  read_def
  read_success_def
  read_degrade_def
  read_fail_def
  restart_mangle_sp_def

definition access :: "pub_indx \<Rightarrow> (unit, 'c)abs5_monad" where
  "access ind \<equiv> read ind \<guillemotright> return ()"

definition access_silence_deg :: "pub_indx \<Rightarrow> (unit, 'c)abs5_monad" where
  "access_silence_deg ind \<equiv> 
    <try>
        access ind 
    <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "
  
definition write_upd_mem_success :: "pub_indx \<Rightarrow> 'c \<Rightarrow> (unit, 'c)abs5_monad" where
  "write_upd_mem_success ind cnt \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv', sp)|sv'.
        LDegraded \<le> reg_state sp ind 
        \<and> sv' = sv\<lparr> abs_updated := insert ind (abs_updated sv), upd_mem := (upd_mem sv)(ind \<mapsto> cnt) \<rparr>
  }"

definition write_upd_mem :: "pub_indx \<Rightarrow> 'c \<Rightarrow> (unit, 'c)abs5_monad" where
  "write_upd_mem ind cnt \<equiv>
    write_upd_mem_success ind cnt
    [+] try_catch (read_degrade ind \<guillemotright> return ())
        (K (throw (Degraded (EB False)) [+] throw (Degraded (EB True))))
    [+] (read_fail ind \<guillemotright> return ())
    [+] restart_mangle_sv
    [+] restart_mangle_sp
  "

lemmas write_upd_mem_defs =
  write_upd_mem_def
  write_upd_mem_success_def
  read_defs

definition degrade_admin
  :: "(unit, 'c)abs5_monad" where
  "degrade_admin \<equiv> (\<lambda>(sv, sp).
  {(Some (Inr ()), sv', sp')|sv' sp' deg.
    LDegraded \<le> admin_state sp
    \<and> sp' = sp\<lparr>
        reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
        admin_state := LDegraded
    \<rparr>
    \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
  })"

definition fail_admin
  :: "(unit, 'c)abs5_monad" where
  "fail_admin \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp' reg_state'.
        (\<forall>ind. reg_state' ind \<le> reg_state sp ind)
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := reg_state',
            admin_state := LDead
        \<rparr>
    }"

definition maybe_mess_up_admin
  :: "(unit, 'c)abs5_monad" where
  "maybe_mess_up_admin \<equiv>
    return ()
    [+] (degrade_admin \<guillemotright> throw (Degraded (EB False)))
    [+] fail_admin
    [+] restart_mangle_sv
    [+] restart_mangle_sp
  "

definition maybe_mess_up_admin_silence_deg
  :: "(unit, 'c)abs5_monad" where
  "maybe_mess_up_admin_silence_deg \<equiv>
    return ()
    [+] degrade_admin
    [+] fail_admin
    [+] restart_mangle_sv
    [+] restart_mangle_sp
  "

lemmas maybe_mess_up_admin_defs =
  maybe_mess_up_admin_def
  degrade_admin_def
  fail_admin_def

end