theory Abs5_Lemmas
imports Abs5_StartupDefs Abs5_UpdateDefs Abs5_FetchDefs "../Pmm_Lemmas"
uses ("../abs2/trans_simproc.ML")
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs

lemma mapM'_inclusion:
  "\<lbrakk> \<forall>x s. m x s \<subseteq> m' s; \<forall>s. (m' \<guillemotright> m') s \<subseteq> m' s; \<forall>s. (Some (Inr ()), s) \<in> m' s \<rbrakk> \<Longrightarrow>
    mapM' m l s \<subseteq> m' s"
  apply(induct l arbitrary: s, auto)
  apply(clarsimp simp add: unfold_monad)
  apply(case_tac aa, simp add: bind_none_decompose, elim disjE, clarsimp_all)
  apply(force simp add: all_conj dest: set_mp)
  apply(subgoal_tac "(None, b) \<in> ((m' \<guillemotright> m') s)")
  apply(force dest: set_mp)
  apply (metis (no_types) bind_none_decompose subsetD)
  apply(case_tac ab, simp_all add: bind_norm_decompose bind_exc_decompose, elim disjE, clarsimp_all)
  apply(force dest: set_mp)
  apply(subgoal_tac "(Some (Inl aa), b) \<in> ((m' \<guillemotright> m') s)")
  apply(force dest: set_mp)
  apply (metis (no_types) bind_exc_decompose subsetD)
  apply(subgoal_tac "(Some (Inr ()), b) \<in> ((m' \<guillemotright> m') s)")
  apply(force dest: set_mp)
  apply (metis (no_types) bind_norm_decompose subsetD)
  done

definition abs5_degrade where
  "abs5_degrade \<equiv> (\<lambda>(sv, sp).
  {(Some (Inr ()), sv', sp')|sv' sp' a_st' deg.
    a_st' \<le> admin_state sp
    \<and> (a_st' = LDead \<longrightarrow> admin_state sp = LDead)
    \<and> (a_st' = LDegraded \<or> deg \<noteq> {})
    \<and> sp' = sp\<lparr>
        reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
        admin_state := a_st'
    \<rparr>
    \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
  })"

definition abs5_fail where
  "abs5_fail \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sv' sp'  reg_state' a_st'.
        (\<forall>ind. reg_state' ind \<le> reg_state sp ind)
        \<and> a_st' \<le> admin_state sp
        \<and> (a_st' = LDead \<or> LDead \<in> range reg_state')
        \<and> sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := reg_state',
            admin_state := a_st'
        \<rparr>
    }
  "

definition abs5_skip where
  "abs5_skip \<equiv>
    return ()
    [+] abs5_degrade
    [+] abs5_fail
    [+] restart_mangle_sp
  "

definition none_LDead where
  "none_LDead \<equiv> {(sv, sp). LDead \<notin> range (reg_state sp) \<and> admin_state sp \<noteq> LDead}"

definition one_not_LDead where
  "one_not_LDead ind \<equiv> {(sv, sp). reg_state sp ind \<noteq> LDead}"

lemma restart_mangle_sp_proj_simps:
  "(restart_mangle_sp)\<Down>NN = empty"
  "(restart_mangle_sp)\<Down>NE = empty"
  by(auto simp add: restart_mangle_sp_def brute_force intro!: ext)

lemma maybe_mess_up_admin_proj_simps:
  "(fail_admin)\<Down>NN = empty"
  "(fail_admin)\<Down>R = empty"
  "(degrade_admin)\<Down>NE = empty"
  "(degrade_admin)\<Down>R = empty"
  by(auto simp add: brute_force maybe_mess_up_admin_defs)
  
lemma read_proj_simps:
  "(read_success ind)\<Down>NE = empty"
  "(read_success ind)\<Down>R = empty"
  "(read_degrade ind)\<Down>NN = empty"
  "(read_degrade ind)\<Down>R = empty"
  "(read_fail ind)\<Down>NN = empty"
  "(read_fail ind)\<Down>R = empty"
  by(auto simp add: read_defs brute_force)

lemma check_pub_region_state_post_NN:
  "\<lbrace>UNIV\<rbrace> check_pub_region ind \<lbrace>>N K (one_not_LDead ind)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp_all only: check_pub_region_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: access_silence_deg_def access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps read_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
    restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
    maybe_mess_up_admin_defs read_defs brute_force min_def
    one_not_LDead_def
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: override_on_def min_def o_def split: split_if_asm)
  done

definition admin_not_LDead where
  "admin_not_LDead \<equiv> {(sv, sp). admin_state sp \<noteq> LDead}"

lemma check_pub_region_admin_not_LDead_inv_NN:
  "\<lbrace>admin_not_LDead\<rbrace> check_pub_region ind \<lbrace>>N K admin_not_LDead\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp_all only: check_pub_region_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: access_silence_deg_def access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps read_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
    restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
    maybe_mess_up_admin_defs read_defs brute_force min_def
    admin_not_LDead_def
    split: split_if_asm exceptions.split_asm)
  done

lemma check_pub_region_other_not_LDead_inv_NN:
  "\<lbrace>one_not_LDead ind'\<rbrace> check_pub_region ind \<lbrace>>N K (one_not_LDead ind')\<rbrace>"
  apply(case_tac "ind'=ind")
  apply(auto intro: conseqNN_rule[OF subset_UNIV check_pub_region_state_post_NN])
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta)
  apply(simp_all only: check_pub_region_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: access_silence_deg_def access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    restart_mangle_sp_proj_simps read_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
    restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
    maybe_mess_up_admin_defs read_defs brute_force min_def
    one_not_LDead_def
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: override_on_def min_def o_def split: split_if_asm)
  done

lemma mapM'_check_pub_region_post:
  assumes "l \<in> order_set UNIV"
  shows "\<lbrace>admin_not_LDead\<rbrace> mapM' check_pub_region l \<lbrace>>N K none_LDead \<rbrace>"
proof-
  have post:
    "\<lbrace>\<Inter>ind \<in> set l. admin_not_LDead\<rbrace>
        mapM' check_pub_region l
    \<lbrace>>N K (\<Inter>ind \<in> set l. one_not_LDead ind) \<rbrace>"
    apply(rule mapM'_distinct_ruleNN)
    apply(auto intro:
      order_set_distinct[OF `l \<in> order_set UNIV`]
      conseqNN_rule[OF subset_UNIV check_pub_region_state_post_NN]
      check_pub_region_admin_not_LDead_inv_NN
      check_pub_region_other_not_LDead_inv_NN)
    done

  have inv:
    "\<lbrace>admin_not_LDead\<rbrace> mapM' check_pub_region l \<lbrace>>N K admin_not_LDead \<rbrace>"
    by(auto intro!: mapM'_invNN check_pub_region_admin_not_LDead_inv_NN)

  show ?thesis
    apply(rule conseqNN_rule[OF _ HoareNN_join[OF post inv]])
    apply(auto simp add: K_beta one_not_LDead_def admin_not_LDead_def none_LDead_def
      order_set_set[OF `l \<in> order_set UNIV`])
    apply(metis)
    done
qed
        
lemma check_all_public_mode_post_NN:
  "\<lbrace>admin_not_LDead\<rbrace> check_all_public \<lbrace>>N K none_LDead \<rbrace>"
  by(auto simp add: check_all_public_def intro!: bind_ruleNN
    intro: ex_falsumNN mapM'_check_pub_region_post chooseN_ruleNN)

use "../abs2/trans_simproc.ML"

simproc_setup trans_simproc 
  ("r \<le> s")
  = {* K Trans_Simproc.trans_simproc *}

declare disjCI[rule del]

lemma override_on_with_min_le:
  "override_on f (min LDegraded \<circ> f) deg ind \<le> f ind"
  by(auto simp add: override_on_def)

lemma ex_override:
  "deg \<noteq> {} \<Longrightarrow>
      (\<exists>degb. degb \<noteq> {} \<and>
        override_on
            (override_on f (min LDegraded \<circ> f) deg)
            (min LDegraded \<circ> override_on f (min LDegraded \<circ> f) deg)
            dega
        = override_on f (min LDegraded \<circ> f) degb) = True
  "
  apply(rule eqTrueI)
  apply(rule_tac x="deg \<union> dega" in exI)
  apply(auto simp add: override_on_def min_def o_def split: split_if_asm)
  done

lemma ex_arg:
  "(\<exists>y. f x = f y) = True"
  "(\<exists>y. f y = f x) = True"
  by(auto)

lemma abs5_skip_idempotent:
  "(abs5_skip \<guillemotright> abs5_skip) (sv, sp) \<subseteq> abs5_skip (sv, sp)"
proof(rule proj_subset, clarsimp_all del: subsetI)
  fix sv sp
  show
    "((abs5_skip \<guillemotright>= (\<lambda>_. abs5_skip))\<Down>R) (sv, sp) \<subseteq> (abs5_skip\<Down>R) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all add: abs5_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs5_degrade_def abs5_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    apply(iterative_simp)+
    apply(case_tac[!] sp, clarsimp_all simp add: override_on_with_min_le)
    apply (metis order_trans override_on_with_min_le)+
    done
next
  fix sv sp
  show
    "((abs5_skip \<guillemotright>= (\<lambda>_. abs5_skip))\<Down>NN) (sv, sp) \<subseteq> (abs5_skip\<Down>NN) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all add: abs5_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs5_degrade_def abs5_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    apply(iterative_simp)+
    apply(case_tac[!] sp, case_tac[!] sv, clarsimp_all simp add: min_def override_on_with_min_le
      ex_override
      split: split_if_asm)
    apply (metis log_state.simps(2))+
    apply ((metis ex_override log_state.simps(2) override_on_emptyset)+)[2]
    apply(case_tac[!] "deg = {}", clarsimp_all simp add: ex_arg ex_override)
    apply (smt ex_override log_state.simps(2) order_trans)+
    done
next
  fix sv sp
  show "((abs5_skip \<guillemotright>= (\<lambda>_. abs5_skip))\<Down>NE) (sv, sp) \<subseteq> (abs5_skip\<Down>NE) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all add: abs5_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs5_degrade_def abs5_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    apply(iterative_simp)+
    apply(case_tac[!] sp, clarsimp_all simp add: override_on_with_min_le)
    apply(smt order_trans override_on_with_min_le)+
    done
qed
 
lemma mapM'_check_pub_region_inclusion:
  "mapM' (check_pub_region) l s \<subseteq> abs5_skip s"
proof(rule mapM'_inclusion)
  {
    fix x sv sp
    have "((check_pub_region x)\<Down>R) (sv, sp) \<subseteq> (abs5_skip\<Down>R) (sv, sp)"
      apply(simp_all only: check_pub_region_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: access_silence_deg_def access_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(simp_all only: read_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        restart_mangle_sp_proj_simps read_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
        restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
        maybe_mess_up_admin_defs read_defs brute_force min_def
        one_not_LDead_def
        split: split_if_asm exceptions.split_asm)
      apply(iterative_simp)+
      apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
        restart_mangle_sp_def)
      apply(simp_all add:
        HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
        disj_commute disj_left_commute disj_absorb disj_left_absorb
        conj_commute conj_left_commute conj_absorb conj_left_absorb
        ex_conj_distribs all_conj
      )
      apply(case_tac[!] sp, clarsimp_all simp add: override_on_with_min_le)
      apply((metis (lifting) order_trans override_on_with_min_le)+)[8]
      done
  } note incR = this

  {
    fix x sv sp
    have "((check_pub_region x)\<Down>NN) (sv, sp) \<subseteq> (abs5_skip\<Down>NN) (sv, sp)"
      apply(simp_all only: check_pub_region_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: access_silence_deg_def access_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(simp_all only: read_def)
      apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
        restart_mangle_sp_proj_simps read_proj_simps
        cong: exceptions.case_cong if_cong)
      apply(simp_all only: choice_def Un_iff, safe)
      apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
        restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
        maybe_mess_up_admin_defs read_defs brute_force min_def
        one_not_LDead_def
        split: split_if_asm exceptions.split_asm)
      apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
        restart_mangle_sp_def)
      apply(simp_all add:
        HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
        disj_commute disj_left_commute disj_absorb disj_left_absorb
        conj_commute conj_left_commute conj_absorb conj_left_absorb
        ex_conj_distribs all_conj
      )
      apply(case_tac[!] sp, case_tac[!] sv, clarsimp_all simp add:
        override_on_with_min_le min_def ex_arg)
      apply(case_tac[!] "deg = {}", clarsimp_all simp add: ex_arg ex_override)
      apply (metis equals0D ex_override)+
      done
  } note incNN = this
    
  show "\<forall>x s. check_pub_region x s \<subseteq> abs5_skip s"
    apply(auto del: subsetI intro!: proj_subset incR incNN)
    apply(simp_all only: check_pub_region_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: access_silence_deg_def access_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: read_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
      restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
      maybe_mess_up_admin_defs read_defs brute_force min_def
      one_not_LDead_def
      split: split_if_asm exceptions.split_asm)
    apply(iterative_simp)+
    apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
      restart_mangle_sp_def)
    apply(simp_all add:
      HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
      disj_commute disj_left_commute disj_absorb disj_left_absorb
      conj_commute conj_left_commute conj_absorb conj_left_absorb
      ex_conj_distribs all_conj
    )
    apply(case_tac[!] b, clarsimp_all simp add: override_on_with_min_le range_eqI[OF sym])
    apply (metis order_trans override_on_with_min_le)+
    done
qed(auto intro!: abs5_skip_idempotent, auto simp add: abs5_skip_def brute_force)


definition abs5_check_all_normal where
  "abs5_check_all_normal \<equiv> (\<lambda>(sv, sp).
  {(r, sv, sp)|r.
    r = Some (Inr ())
    \<and> admin_state sp \<noteq> LDead
    \<and> LDead \<notin> range (reg_state sp)
  })"

definition abs5_check_all_degrade where
  "abs5_check_all_degrade \<equiv> (\<lambda>(sv, sp).
  {(Some (Inr ()), sv', sp')|sv' sp' a_st' deg.
    a_st' \<le> admin_state sp
    \<and> a_st' \<noteq> LDead
    \<and> LDead \<notin> range (reg_state sp)
    \<and> (a_st' = LDegraded \<or> deg \<noteq> {})
    \<and> sp' = sp\<lparr>
        reg_state := override_on (reg_state sp) (K LDegraded) deg,
        admin_state := a_st'
    \<rparr>
    \<and> sv' = sv\<lparr> abs_card_mode := ACDegraded \<rparr>
  })"

definition abs5_check_all where
  "abs5_check_all \<equiv>
    abs5_check_all_normal
    [+] abs5_check_all_degrade
    [+] abs5_fail
    [+] restart_mangle_sp
  "

lemmas abs5_check_all_defs =
  abs5_check_all_def
  abs5_check_all_normal_def
  abs5_check_all_degrade_def
  abs5_fail_def
  restart_mangle_sp_def

lemma override_range:
  "(LDead \<notin> range (override_on f (min LDegraded \<circ> f) deg)) = (LDead \<notin> range f)"
  by(auto simp add: override_on_def min_def o_def)

lemma override_remove_min:
  "LDead \<notin> range f \<Longrightarrow>
  override_on f (min LDegraded \<circ> f) deg
  = override_on f (K LDegraded) deg 
  "
  apply(auto simp add: override_on_def min_def o_def K_beta intro!: ext)
  apply(case_tac "f a", auto dest!: range_eqI[OF sym, where f=f])
  done

definition post_check_admin where
   "post_check_admin \<equiv> admin_not_LDead \<inter> sv_abs_mode_neq ACFail"

lemma check_all_inclusion:
  assumes adm_state: "(sv, sp) \<in> post_check_admin"
  shows "check_all_public (sv, sp) \<subseteq> abs5_check_all (sv, sp)"
proof(rule proj_subset')
  show "(check_all_public\<Down>NN) (sv, sp) \<subseteq> (abs5_check_all\<Down>NN) (sv, sp)"
  proof(rule subsetI, case_tac x, simp_all)
    fix sv' sp'
    assume origin: "((), sv', sp') \<in> (check_all_public\<Down>NN) (sv, sp)"
    hence none_LDead: "(sv', sp') \<in> none_LDead" using check_all_public_mode_post_NN adm_state
      by(auto simp add: HoareNN_alt Hoare_def Box_def K_beta post_check_admin_def)

    have
      "\<exists>l. l \<in> order_set UNIV \<and> ((), sv', sp') \<in> ((mapM' check_pub_region l)\<Down>NN) (sv, sp)"
        (is "\<exists>x. ?P x")
      using origin
      apply(auto simp add: projectNN_def check_all_public_def bind_norm_decompose)
      apply(auto simp add: chooseN_def brute_force)
      done
    from this obtain l where inc1: "?P l" ..
    hence in_skip: "((), sv', sp') \<in> (abs5_skip\<Down>NN) (sv, sp)"
      apply -
      apply(erule conjE, erule rev_subsetD, intro projectNN_mono allI mapM'_check_pub_region_inclusion)
      done
    show "((), sv', sp') \<in> (abs5_check_all\<Down>NN) (sv, sp)" using none_LDead in_skip adm_state
      apply(auto simp add: abs5_skip_def none_LDead_def brute_force)
      apply(clarsimp_all simp add: abs5_check_all_def brute_force)
      apply(clarsimp_all simp add: abs5_degrade_def)
      apply(clarsimp_all simp add: abs5_check_all_defs brute_force admin_not_LDead_def
        override_remove_min override_range)
      apply(iterative_simp)+
      apply(clarsimp_all simp add: post_check_admin_def sv_mode_neq_def)
      apply(case_tac[!] sv, case_tac[!] sp, clarsimp_all simp add: min_def)
      apply(safe, clarsimp_all simp add: post_check_admin_def sv_abs_mode_neq_def override_range)
      apply(simp_all add:
        HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
        disj_commute disj_left_commute disj_absorb disj_left_absorb
        conj_commute conj_left_commute conj_absorb conj_left_absorb
        ex_conj_distribs all_conj
      )
      apply(case_tac[!] abs_card_mode)
      apply(clarsimp_all)
      apply(metis equals0D)+
      done
  qed
next
  show "(check_all_public\<Down>NE) (sv, sp) \<subseteq> (abs5_check_all\<Down>NE) (sv, sp)"
    apply(rule subset_trans[where B="(abs5_skip\<Down>NE) (sv, sp)"])
    apply(clarsimp simp add: check_all_public_def chooseN_def brute_force
      dest!: mapM'_check_pub_region_inclusion[THEN subsetD]
    )
    apply(auto simp add: abs5_skip_def abs5_check_all_def brute_force)
    apply(clarsimp simp add: abs5_degrade_def)
    done
next
  show "(check_all_public\<Down>R) (sv, sp) \<subseteq> (abs5_check_all\<Down>R) (sv, sp)"
    apply(rule subset_trans[where B="(abs5_skip\<Down>R) (sv, sp)"])
    apply(clarsimp simp add: check_all_public_def chooseN_def brute_force
      dest!: mapM'_check_pub_region_inclusion[THEN subsetD]
    )
    apply(auto simp add: abs5_skip_def abs5_check_all_def brute_force)
    apply(auto simp add: abs5_degrade_def restart_mangle_sp_def)
    done
qed

text {* The following could probably somehow be unified with 
  the check_pub_region theorem, but will leave it as a TODO at
  the moment *}

lemma touch_one_inclusion:
  "touch_one ind (sv, sp) \<subseteq> abs5_skip (sv, sp)"
proof(rule proj_subset')
  show "((touch_one ind)\<Down>NN) (sv, sp) \<subseteq> (abs5_skip\<Down>NN) (sv, sp)"
    apply(simp_all only: touch_one_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: read_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
      restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
      maybe_mess_up_admin_defs read_defs brute_force min_def
      one_not_LDead_def
      split: split_if_asm exceptions.split_asm)
    apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
      restart_mangle_sp_def)
    apply(simp_all add:
      HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
      disj_commute disj_left_commute disj_absorb disj_left_absorb
      conj_commute conj_left_commute conj_absorb conj_left_absorb
      ex_conj_distribs all_conj
    )
    apply(case_tac[!] sp, case_tac[!] sv, clarsimp_all simp add:
      override_on_with_min_le min_def ex_arg)
    apply(case_tac[!] "deg = {}", clarsimp_all simp add: ex_arg ex_override)
    apply (metis equals0D ex_override)+
    done
next
  show "((touch_one ind)\<Down>NE) (sv, sp) \<subseteq> (abs5_skip\<Down>NE) (sv, sp)"
    apply(simp_all only: touch_one_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: read_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
      restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
      maybe_mess_up_admin_defs read_defs brute_force min_def
      one_not_LDead_def
      split: split_if_asm exceptions.split_asm)
    apply(iterative_simp)+
    apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
      restart_mangle_sp_def)
    apply(simp_all add:
      HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
      disj_commute disj_left_commute disj_absorb disj_left_absorb
      conj_commute conj_left_commute conj_absorb conj_left_absorb
      ex_conj_distribs all_conj
    )
    apply(case_tac[!] sp, clarsimp_all simp add: override_on_with_min_le range_eqI[OF sym])
    apply (metis order_trans override_on_with_min_le)+
    done
next
  show "((touch_one ind)\<Down>R) (sv, sp) \<subseteq> (abs5_skip\<Down>R) (sv, sp)"
    apply(simp_all only: touch_one_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all only: read_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def
      restart_mangle_sv_proj_simps restart_mangle_sp_proj_simps
      maybe_mess_up_admin_defs read_defs brute_force min_def
      one_not_LDead_def
      split: split_if_asm exceptions.split_asm)
    apply(iterative_simp)+
    apply(simp_all add: abs5_skip_def brute_force abs5_degrade_def abs5_fail_def
      restart_mangle_sp_def)
    apply(simp_all add:
      HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
      disj_commute disj_left_commute disj_absorb disj_left_absorb
      conj_commute conj_left_commute conj_absorb conj_left_absorb
      ex_conj_distribs all_conj
    )
    apply(case_tac[!] sp, clarsimp_all simp add: override_on_with_min_le)
    apply(metis (lifting) order_trans override_on_with_min_le)+
    done
qed

lemma mapM'_touch_one_inclusion:
  "mapM' touch_one l (sv, sp) \<subseteq> abs5_skip (sv, sp)"
  apply(rule mapM'_inclusion)
  apply(auto del: subsetI intro!: touch_one_inclusion abs5_skip_idempotent)
  apply(auto simp add: abs5_skip_def brute_force)
  done

lemma touch_updated_inclusion:
  "\<forall>s. touch_updated s \<subseteq> (abs5_skip \<guillemotright> set_abs_updated {}) s"
  apply(auto)
  apply(auto simp add: touch_updated_def get_abs_updated_def choose_def unfold_monad
    getters_setters chooseN_def split: option.split_asm sum.split_asm option.split sum.split
    dest!: mapM'_touch_one_inclusion[THEN subsetD]
  )
  done

definition upd_mem_eq :: "(pub_indx \<rightharpoonup> 'c) \<Rightarrow> ('c abs5_vol_state \<times> 'c abs5_state)set" where
  "upd_mem_eq upd_f \<equiv> {(sv, sp). upd_mem sv = upd_f}"

lemma clear_upd_mem_post_NN:
  "\<lbrace>P\<rbrace> clear_upd_mem \<lbrace>>N K (upd_mem_eq Map.empty)\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def K_def
    clear_upd_mem_def modifyRE_def brute_force upd_mem_eq_def)

definition upd_dom_subset where
  "upd_dom_subset upd_f \<equiv> {(sv, sp). dom (upd_mem sv) \<subseteq> dom upd_f}"

definition upd_mem_eq_one where
  "upd_mem_eq_one ind upd_f \<equiv> {(sv, sp). upd_mem sv ind = upd_f ind}"

lemma mapM'_write_new'_post_NN:
  assumes l_set: "l \<in> order_set (dom upd_f)"
  shows
    "\<lbrace>upd_mem_eq Map.empty\<rbrace> mapM' (write_new' upd_f) l \<lbrace>>N K (upd_mem_eq upd_f)\<rbrace>"
proof-
  have subs_inv:
    "\<lbrace>upd_dom_subset upd_f\<rbrace> mapM' (write_new' upd_f) l \<lbrace>>N K (upd_dom_subset upd_f)\<rbrace>"
    apply(clarsimp intro!: mapM'_invNN)
    apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
    apply(auto simp add: write_new'_def write_upd_mem_defs brute_force upd_dom_subset_def
      order_set_set[OF l_set])
    done
  have notin_dom:
    "\<lbrace>upd_mem_eq Map.empty\<rbrace> mapM' (write_new' upd_f) l \<lbrace>>N K (upd_dom_subset upd_f)\<rbrace>"
    apply(rule conseqNN_rule[OF _ subs_inv])
    apply(auto simp add: upd_mem_eq_def upd_dom_subset_def)
    done
  have in_dom:
    "\<lbrace>\<Inter>ind \<in> set l. UNIV\<rbrace>
        mapM' (write_new' upd_f) l
    \<lbrace>>N K (\<Inter>ind \<in> set l. upd_mem_eq_one ind upd_f)\<rbrace>"
    apply(intro mapM'_distinct_ruleNN)
    apply(auto intro: order_set_distinct[OF l_set] true_postNN)
    apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
    apply(auto simp add: write_new'_def write_upd_mem_defs brute_force upd_mem_eq_one_def
      order_set_set[OF l_set])
    done
  show ?thesis
    apply(rule conseqNN_rule[OF _ HoareNN_join[OF notin_dom in_dom]])
    apply(auto simp add: K_def upd_dom_subset_def upd_mem_eq_def upd_mem_eq_one_def
      order_set_set[OF l_set]
      intro!: ext)
    apply(case_tac "upd_f x", auto simp add: dom_def)
    done
qed
    
lemma pre_commit_post_NN:
  "\<lbrace>UNIV\<rbrace> pre_commit upd_f \<lbrace>>N K (upd_mem_eq upd_f) \<rbrace>"
  apply(simp only: pre_commit_def)
  apply(rule bind_ruleNN[OF true_postNN])
  apply(clarsimp simp only: K_beta)
  apply(rule bind_ruleNN[where Q="K (upd_mem_eq Map.empty)"])
  apply(rule clear_upd_mem_post_NN)
  apply(clarsimp simp only: K_beta)
  apply(rule bind_ruleNN[OF chooseN_ruleNN])
  apply(auto simp add: K_beta intro: ex_falsumNN)
  apply(rule try_catch_ruleNN)
  apply(rule mapM'_write_new'_post_NN)
  apply(auto  simp add: K_beta rollback_def
    intro!: HoareNN_throw
    bind_ruleNN
    intro: true_postNE true_postNN
    split: exceptions.split)
  done

definition dom_upd_mem_is_updated where
  "dom_upd_mem_is_updated \<equiv> {(sv, sp). abs_updated sv = dom (upd_mem sv)}"

lemma set_abs_updated_post_NN:
  "\<lbrace>P\<rbrace> set_abs_updated inds \<lbrace>>N K (abs_updated_equals inds)\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def brute_force abs_updated_equals_def)

lemma pre_commit_updated_post_NN:
  "\<lbrace>UNIV\<rbrace> Abs5_UpdateDefs.pre_commit upd_f \<lbrace>>N (K dom_upd_mem_is_updated)\<rbrace>"
  apply(simp only: Abs5_UpdateDefs.pre_commit_def)
  apply(rule bind_ruleNN[OF set_abs_updated_post_NN])
  apply(clarsimp simp add: K_beta)
  apply(rule bind_ruleNN[where Q="K (dom_upd_mem_is_updated)"])
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def brute_force abs_updated_equals_def
    dom_upd_mem_is_updated_def clear_upd_mem_def modifyRE_def)
  apply(clarsimp simp add: K_beta)
  apply(rule bind_ruleNN[OF chooseN_ruleNN])
  apply(auto intro: ex_falsumNN)
  apply(rule try_catch_ruleNN)
  apply(rule mapM'_invNN)
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def brute_force 
    dom_upd_mem_is_updated_def write_new'_def write_upd_mem_defs)
  apply(rule true_postNE)
  apply(auto simp add: K_beta rollback_def
    split: exceptions.split intro!: bind_ruleNN intro: HoareNN_throw true_postNN)
  done

end