theory Abs5_Ref
imports
  "../Composition"
  "../Pmm_Refinement"
  "../Misc_Lemmas"
  "../Pmm_Lemmas"
  "../abs4/Abs4_CommonDefs"
  "../abs4/Abs4_UpdateDefs"
  "../abs4/Abs4_FetchDefs"
  "../abs4/Abs4_StartupDefs"
  "../abs4/Abs4_Invariant_Lemmas"
  Abs5_CommonDefs
  Abs5_UpdateDefs
  Abs5_FetchDefs
  Abs5_StartupDefs
  uses ("../abs2/trans_simproc.ML")
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs
  force_same_and_def

fun is_arrow_tgt :: "sync_direction \<Rightarrow> INST \<Rightarrow> bool" where
  "is_arrow_tgt SR Secondary = True"
  | "is_arrow_tgt SL Primary = True"
  | "is_arrow_tgt _ _ = False"

fun comb_rel_writ :: "(bool \<times> bool) \<Rightarrow> (bool \<times> bool) \<Rightarrow> log_state" where
  "comb_rel_writ (True, True) (_, True) = LOK"
  | "comb_rel_writ (_, True) (True, True) = LOK"
  | "comb_rel_writ (True, _) (_, _) = LDegraded"
  | "comb_rel_writ (_, _) (True, _) = LDegraded"
  | "comb_rel_writ _ _ = LDead"

definition abs_reg_state :: "'c abs4_state \<Rightarrow> pub_indx \<Rightarrow> log_state" where
  "abs_reg_state spc \<equiv> \<lambda>ind.
        comb_rel_writ
            ((ind, Primary) \<in> in_view spc, (ind, Primary) \<in> writable spc)
            ((ind, Secondary) \<in> in_view spc, (ind, Secondary) \<in> writable spc)"

definition abs5_abs4_strel ::"(('c abs5_vol_state \<times> 'c abs5_state) \<times> 
    (abs_vol_mem \<times> ('c :: size) abs4_state))set" 
  where
  "abs5_abs4_strel \<equiv> {((sva, spa), (svc, spc)). 
    svc = abs_vol_mem.truncate sva
    \<and> (\<forall>ind.
        ((ind, Primary) \<in> in_view spc \<longrightarrow> memory spa ind = abs4_state.memory spc ind Primary)
        \<and> ((ind, Secondary) \<in> in_view spc \<longrightarrow> memory spa ind = abs4_state.memory spc ind Secondary)
    )
    \<and> admin_state spa = abs4_state.admin_state spc
    \<and> reg_state spa = abs_reg_state spc
  }"

definition is_sync_tgt :: "pub_indx \<Rightarrow> INST \<Rightarrow> ('sv \<times> 'c abs4_state)set" where
  "is_sync_tgt ind inst = {(sv, sp).
    is_arrow_tgt (arrow sp ind) inst
    \<or> (arrow sp ind = SEq \<and> (ind, inst) \<notin> reliable sp)}"

lemma ex_not_Dead:
  "(\<exists>x. x \<noteq> LDead \<and> P x) = (P LDegraded \<or> P LOK)"
  apply(auto)
  by (metis log_state.exhaust)

lemma min_aux: "(x :: 'a :: linorder) \<le> y \<Longrightarrow> (min y z = x) = (z = x \<or> (y < z \<and> y = x))"
  by(auto simp add: min_def)

lemma le_LDead:
  "(x \<le> LDead) = (x = LDead)"
  by(case_tac[!] x, auto)

lemma LOK_le:
  "(x \<le> LOK) = True"
  by(case_tac[!] x, auto)

lemma le_LOK:
  "(LOK \<le> x) = (x = LOK)"
  by(case_tac[!] x, auto)

lemma min_LDead:
  "min x LDead = LDead"
  "min LDead x = LDead"
  by(case_tac[!] x, auto simp add: min_def)

lemma ex2_log_states:
  "(\<exists>x :: log_state. x \<noteq> y) = True"
  by(case_tac y, auto)

lemmas log_state_simps =
  le_LDead
  le_LOK
  LOK_le
  min_LDead
  ex2_log_states

lemma LDegraded_le_comb_x_True:
  "LDegraded \<le> comb_rel_writ (x, y) (True, z) = True"
  by(case_tac x, case_tac[!] y, case_tac[!] z, simp_all)

lemma LDegraded_le_comb_True_x:
  "LDegraded \<le> comb_rel_writ (True, z) (x, y) = True"
  by(case_tac x, case_tac[!] y, case_tac[!] z, simp_all)

lemma comb_is_LDead:
  "(comb_rel_writ (x, y) (u, v) = LDead) = (\<not>x \<and> \<not>u)"
  "(LDead = comb_rel_writ (x, y) (u, v)) = (\<not>x \<and> \<not>u)"
  by(case_tac[!] x, case_tac[!] y, case_tac[!] u, case_tac[!] v, simp_all)

lemma LDegraded_le_comb:
  "(LDegraded \<le> comb_rel_writ (x, y) (u, v)) = (x \<or> u)"
  by(case_tac[!] x, case_tac[!] y, case_tac[!] u, case_tac[!] v, simp_all)

lemma comb_is_LDegraded:
  "(comb_rel_writ (x, y) (u, v) = LDegraded) = ((x \<or> u) \<and> (\<not>y \<or> \<not>v))"
  "(LDegraded = comb_rel_writ (x, y) (u, v)) = ((x \<or> u) \<and> (\<not>y \<or> \<not>v))"
  by(case_tac[!] x, case_tac[!] y, case_tac[!] u, case_tac[!] v, simp_all)

lemmas comb_simps =
  LDegraded_le_comb_x_True
  LDegraded_le_comb_True_x
  LDegraded_le_comb
  comb_is_LDead
  comb_is_LDegraded

declare disjCI[rule del]

lemma aux:
  "(\<exists>f. P f \<and> Q f \<and> (\<forall>x. f x = g x)) = (P g \<and> Q g)"
  apply(auto intro!: ext)
  apply(metis ext)+
  done

lemma in_view_ref_simps:
  "in_view (spc\<lparr> abs4_state.memory := mem' \<rparr>) = in_view spc"
  "in_view (spc\<lparr> abs4_state.admin_state := a_st' \<rparr>) = in_view spc"
  "in_view (spc\<lparr> writable := w' \<rparr>) = in_view spc"
  "rel = reliable spc \<Longrightarrow>
    in_view (spc\<lparr> abs4_state.reliable := rel - A \<rparr>) = in_view spc - A"
  "\<lbrakk> rel = reliable spc; inst \<in> arrow_srcs (arrow spc ind) \<rbrakk> \<Longrightarrow>
    in_view (spc\<lparr> abs4_state.reliable := insert (ind, inst) rel \<rparr>) = insert (ind, inst) (in_view spc)"
  "arrow spc ind = SEq \<Longrightarrow> ((ind, inst) \<in> in_view spc) = ((ind, inst) \<in> reliable spc)"
  "inst \<in> arrow_srcs (arrow spc ind) \<Longrightarrow> ((ind, inst) \<in> in_view spc) = ((ind, inst) \<in> reliable spc)"
  "arr = arrow spc \<Longrightarrow>
    in_view (spc\<lparr>arrow := override_on arr (\<lambda>y. SBot) arr_del\<rparr>) = in_view spc - (arr_del \<times> UNIV)" 
  by(case_tac[!] inst, auto simp add: in_view_def override_on_def split: split_if_asm)

lemma in_view_eqI:
  "\<lbrakk> (ind, inst) \<in> reliable spc; arrow spc ind = SEq \<rbrakk> \<Longrightarrow> ((ind, inst) \<in> in_view spc) = True"
  "\<lbrakk> (ind, inst) \<in> reliable spc; inst \<in> arrow_srcs (arrow spc ind) \<rbrakk> \<Longrightarrow>
    ((ind, inst) \<in> in_view spc) = True"
  by(case_tac inst, auto simp add: in_view_def)

lemma abs_reg_state_ref_simps_1:
  "abs_reg_state (spc\<lparr> abs4_state.admin_state := a_st' \<rparr>) = abs_reg_state spc"
  "writ = writable spc \<Longrightarrow>
    abs_reg_state (spc \<lparr> writable := writ - unwr \<rparr>) =
    override_on (abs_reg_state spc) ((min LDegraded) o (abs_reg_state spc)) (fst ` unwr)
  "
  apply(auto simp add: abs_reg_state_def in_view_ref_simps override_on_def K_def
    in_view_ref_simps comb_simps in_view_eqI min_def
    intro!: ext)
  apply(auto simp add: in_view_def)
  apply(metis INST.exhaust)
  apply (smt Pair_eq fst_conv image_iff)
  apply(metis INST.exhaust)
  apply (smt Pair_eq fst_conv image_iff)
  done

lemma abs_reg_state_ref_simps_2:
  "abs_reg_state (spc\<lparr> abs4_state.memory := mem' \<rparr>) = abs_reg_state spc"
  "rel = reliable spc \<Longrightarrow>
    abs_reg_state (spc\<lparr> reliable := rel - {(ind, inst)} \<rparr>) =
        (abs_reg_state spc)(ind :=
            if (ind, other_i inst) \<in> in_view spc
            then abs_reg_state spc ind
            else LDead)
  "
  "\<lbrakk> arrow spc ind = SEq; (ind, inst) \<in> reliable spc \<rbrakk> \<Longrightarrow> (LDegraded \<le> abs_reg_state spc ind) = True"
  "\<lbrakk> inst \<in> arrow_srcs (arrow spc ind); (ind, inst) \<in> reliable spc \<rbrakk> \<Longrightarrow>
  (LDegraded \<le> abs_reg_state spc ind) = True"
  "arr = arrow spc \<Longrightarrow>
    abs_reg_state (spc\<lparr>arrow := override_on arr (\<lambda>y. SBot) arr_del\<rparr>)
    = override_on (abs_reg_state spc) (K LDead) arr_del"
  apply(auto simp add: abs_reg_state_def in_view_ref_simps intro!: ext)
  apply(case_tac[!] inst, auto simp add: comb_simps in_view_def override_on_def in_view_def K_def)
  apply (smt Pair_eq comb_rel_writ.simps(1) comb_rel_writ.simps(2) comb_simps(6))+
  done

lemma abs_reg_state_ref_simps_3:
  "rel = reliable spc \<Longrightarrow>
    abs_reg_state (spc\<lparr>reliable := rel - unr\<rparr>)
    = override_on (abs_reg_state spc) (K LDead)
        {ind|ind inst.
        (ind, inst) \<in> unr \<and> ((ind, other_i inst) \<in> unr \<or> (ind, other_i inst) \<notin> reliable spc
            \<or> other_i inst \<notin> arrow_srcs (arrow spc ind))}"
  apply(auto simp add: abs_reg_state_def in_view_def override_on_def K_def intro!: ext)
  apply(case_tac inst, auto)
  apply(case_tac inst, auto)
  apply(case_tac inst, auto)
  apply(case_tac "(ind, Primary) \<in> unr", auto)
  apply(case_tac[!] "(ind, Secondary) \<in> unr", auto)
  apply (smt comb_rel_writ.simps(1) comb_rel_writ.simps(2) comb_simps(7) other_i.simps(1))
  apply (smt comb_rel_writ.simps(1) comb_rel_writ.simps(2) comb_simps(7) other_i.simps(2))
  done
  
lemmas abs_reg_state_ref_simps =
  abs_reg_state_ref_simps_1
  abs_reg_state_ref_simps_2
  abs_reg_state_ref_simps_3

lemma override_on_simp:
  "(f(ind := LDead) = override_on (g(ind := LDead)) (K LDegraded) A)
  = (ind \<notin> A \<and> (\<forall>ind'. ind' \<noteq> ind \<longrightarrow> ((ind' \<in> A \<longrightarrow> f ind' = LDegraded) \<and> (ind' \<notin> A \<longrightarrow> f ind' = g ind'))))"
  apply(rule iffI)
  apply(auto simp add: override_on_def K_def intro!: ext)
  apply (metis fun_upd_apply log_state.simps(1))
  apply (metis fun_upd_apply log_state.simps(1))
  apply (metis fun_upd_apply log_state.simps(1))
  done

lemma override_updated_simp:
  "(override_on f (min LDegraded o f) A)(ind := LDead)
  = (override_on (f(ind := LDead)) (min LDegraded \<circ> f(ind := LDead))) A
  "
  by(auto simp add: override_on_def min_def intro!: ext)

lemma ex_override_simp:
  "(ind, inst) \<in> unwr \<Longrightarrow>
  (\<exists>deg. override_on f g deg = override_on f g (fst ` unwr)
        \<and> ind \<in> deg)
  "
  "(ind, inst) \<in> unwr \<Longrightarrow>
  (\<exists>deg. ind \<in> deg \<and> override_on f g deg = override_on f g (fst ` unwr))
  "
  by(rule_tac x="fst ` unwr" in exI, force simp add: image_def)+

lemma double_override: "override_on (override_on f g A) g B = override_on f g (A \<union> B)"
  by(auto simp add: override_on_def)
 
lemma override_le_simp:
  "((override_on (override_on f (K LDead) A) (min LDegraded \<circ> override_on f (K LDead) B) C) ind 
  \<le> f ind) = True"
  "(override_on (override_on f (K LDead) A)
    (min LDegraded \<circ> override_on f (K LDead) B) C
        ind
    = LDead)
  = (f ind = LDead \<or> (ind \<in> A \<and> ind \<notin> C) \<or> (ind \<in> B \<and> ind \<in> C))"
  by(auto simp add: override_on_def min_def K_def)

lemmas override_on_ref_simps =
  override_on_simp
  override_updated_simp
  double_override
  override_le_simp

lemma is_arrow_tgt_2nd_simp:
  "is_arrow_tgt arr Primary = (arr = SL)"
  "is_arrow_tgt arr Secondary = (arr = SR)"
  by(case_tac[!] arr, simp_all)

lemma comb_mono:
  "\<lbrakk> x \<le> x'; y \<le> y'; u \<le> u'; v \<le> v' \<rbrakk> \<Longrightarrow>
  (comb_rel_writ (x, y) (u, v) \<le> comb_rel_writ (x', y') (u', v')) = True"
  by(case_tac[!] x, case_tac[!] y, case_tac[!] u, case_tac[!] v, simp_all,
    case_tac[!] x', case_tac[!] y', case_tac[!] u', case_tac[!] v',
    simp_all)

lemma forall_INST_decompose:
  "(\<forall>inst :: INST. P inst) = (P Primary \<and> P Secondary)" 
  apply(auto)
  apply(rule_tac y=inst in INST.exhaust, simp_all)
  done

lemma tgt_notin_srcs:
  "is_arrow_tgt (arrow spc ind) inst \<Longrightarrow> (inst \<in> arrow_srcs (arrow spc ind)) = False"
  by(case_tac "arrow spc ind", case_tac[!] inst, simp_all)

lemma tgt_other_i_in_srcs:
  "\<lbrakk> is_arrow_tgt (arrow spc ind) inst; inst' = other_i inst \<rbrakk> \<Longrightarrow>
    (inst' \<in> arrow_srcs (arrow spc ind)) = True"
  by(case_tac "arrow spc ind", case_tac[!] inst, simp_all)

lemma tgt_other_i_in_srcsD:
  "is_arrow_tgt (arrow spc ind) inst \<Longrightarrow>
    other_i inst \<in> arrow_srcs (arrow spc ind)"
  by(case_tac "arrow spc ind", case_tac[!] inst, simp_all)

lemma comb_one_True:
  "comb_rel_writ (x, y) (True, v) = comb_rel_writ (x', y) (True, v)"
  "comb_rel_writ (True, y) (u, v) = comb_rel_writ (True, y) (u', v)"
  by(case_tac[!] x, case_tac[!] x', case_tac[!] y, case_tac[!] u, case_tac[!] u', case_tac[!] v,
    simp_all)

lemma bex_Un: "(\<exists>x \<in> A \<union> B. P x) = ((\<exists>x \<in> A. P x) \<or> (\<exists>x \<in> B. P x))"
  by (metis Un_iff)

lemma arrow_srcs_override_SBot:
  "arrow_srcs (override_on f (\<lambda>y. SBot) arr_del x) =
    (if x \<in> arr_del then {} else arrow_srcs (f x))"
  by(auto)

lemma override_triv_iff:
  "(f = override_on f g A) = (\<forall>x \<in> A. f x = g x)"
  "(override_on f g A = f) = (\<forall>x \<in> A. f x = g x)"
  apply(auto simp add: override_on_def split: split_if_asm dest: fun_cong)
  apply(smt)
  done

lemma override_on_with_K_LDead_le:
  "(override_on f (K LDead) A x \<le> f x) = True"
  by(auto simp add: override_on_def K_def)

lemma comb_rel_writ_is_OK:
  "(comb_rel_writ (x, y) (u, v) = LOK) = ((u \<or> x) \<and> y \<and> v)"
  "(LOK = comb_rel_writ (x, y) (u, v)) = ((u \<or> x) \<and> y \<and> v)"
  by(case_tac[!] x, case_tac[!] y, case_tac[!] u, case_tac[!] v, simp_all)

lemma aux1million: "(P \<and> (P \<longrightarrow> Q)) = (P \<and> Q)"
  by blast

lemma fun_map_override_on_simp:
    "\<lbrakk> x \<in> A; A \<subseteq> dom g \<rbrakk> \<Longrightarrow> fun_map_override_on f g A x = the (g x)"
  by(auto simp add: fun_map_override_on_def)

lemma double_override_simp:
  "override_on (override_on f g A) (h o override_on f g A) B
  = override_on f (\<lambda>x.
    if x \<in> A
    then
        if x \<in> B
        then (h o g) x
        else g x
    else
        if x \<in> B
        then (h o f) x
        else f x
  ) (A \<union> B)"
  by(auto simp add: override_on_def intro!: ext)

lemma min_CNormal_CDegraded:
  "min CNormal CDegraded = CDegraded"
  "min CDegraded CNormal  = CDegraded"
  by(simp_all add: min_def)

term "(maybe_mess_up_admin_silence_deg \<guillemotright> chooseN UNIV)"
lemma get_arrow_ref_NN:
  "sim_normN_ext abs5_abs4_strel
    (maybe_mess_up_admin_silence_deg \<guillemotright> chooseN UNIV)
    (get_arrow ind)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc v svc' spc')
  apply(simp_all only: get_arrow_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_arrow_defs)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def)
  apply(clarsimp_all simp add: degrade_admin_def fail_admin_def restart_mangle_sp_def brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac sva, case_tac svc)
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps all_conj abs_vol_mem.truncate_def)
  apply(metis)
  done
 
lemma get_arrow_ref_NE:
  "sim_normE_ext abs5_abs4_strel
    (maybe_mess_up_admin_silence_deg \<guillemotright> chooseN UNIV)
    (get_arrow ind)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc e svc' spc')
  apply(simp_all only: get_arrow_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_arrow_defs)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def)
  apply(clarsimp_all simp add: degrade_admin_def fail_admin_def restart_mangle_sp_def brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps all_conj)
  apply(case_tac sva)
  apply(clarsimp simp add: override_on_def K_def min_def abs_vol_mem.truncate_def)
  done
 
lemma get_arrow_ref:
  "sim_ext abs5_abs4_strel
    (maybe_mess_up_admin_silence_deg \<guillemotright> chooseN UNIV)
    (get_arrow ind)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!: get_arrow_ref_NN get_arrow_ref_NE del: subsetI)
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: get_arrow_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: brute_force get_arrow_defs)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def)
  apply(clarsimp_all simp add: degrade_admin_def fail_admin_def restart_mangle_sp_def brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps all_conj abs_vol_mem.truncate_def )
  apply(case_tac[!] svc', clarsimp)
  apply (metis abs_vol_mem.select_convs(1) abs_vol_mem.select_convs(2))
  apply(clarsimp simp add: override_on_def K_def min_def abs_vol_mem.truncate_def)
  by (metis abs_vol_mem.select_convs(1) abs_vol_mem.select_convs(2))

lemma sync_ref_NN:
  "from \<noteq> to \<Longrightarrow>
  sim_normN_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.read ind)
    (Abs4_CommonDefs.sync ind to from)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp add: sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: sync_defs brute_force is_sync_tgt_def)
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_ref_simps in_view_ref_simps)
  apply(clarsimp_all simp add: abs_reg_state_def in_view_def)
  apply(case_tac [!] "arrow spc ind")
  apply(case_tac[!] to, clarsimp_all simp add: comb_simps all_conj intro!:ext)
  apply (smt Pair_eq comb_rel_writ.simps(1) comb_rel_writ.simps(2) comb_simps(6)
    insertI1 insertI2 arrow_srcs.simps(3) arrow_srcs.simps(2) arrow_srcs.simps(1))+
  done

lemma sync_fail_ref_NE:
  "from \<noteq> to \<Longrightarrow>
  sim_normE_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.read ind)
    (Abs4_CommonDefs.sync_fail ind to from)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(clarsimp_all simp add: sync_defs brute_force)
  apply(clarsimp_all simp add: read_defs is_sync_tgt_def brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_def in_view_def)
  apply(iterative_simp)+
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] sva)
  apply(clarsimp_all simp add: comb_simps abs_vol_mem.truncate_def)
  apply(case_tac[!] to, clarsimp_all simp add: comb_simps image_def override_on_def tgt_notin_srcs
    comb_mono tgt_other_i_in_srcs intro!: ext)
  done

lemma sim_normE_ext_choice_conc:
  "sim_normE_ext R ma (mc [+] mc') Se = (sim_normE_ext R ma mc Se \<and> sim_normE_ext R ma mc' Se)"
  by(auto simp add: sim_normE_ext_def project_simps simN_wpc_choice_conc simS_wpc_choice_conc)

lemma x_in_fst_simp:
  "(\<exists>x\<in>A. y = fst x) = (\<exists>z. (y, z) \<in> A)"
  by(force)

lemma x_witness_simp:
  "P x \<Longrightarrow> (\<exists>x. P x) = True"
  by(force)

lemma INST_disj_exhaust: "P x \<Longrightarrow> (P Primary \<or> P Secondary) = True"
  by (metis (full_types) INST.exhaust)

lemma ex_vol_mem_assignment:
  "(\<exists>a. b = \<lparr>abs_card_mode = abs_card_mode a, abs_updated = abs_updated a\<rparr>) = True"
  "(\<exists>a. \<lparr>abs_card_mode = abs_card_mode a, abs_updated = abs_updated a\<rparr> = b) = True "
  apply (metis (full_types) abs_vol_mem.select_convs(1) 
    abs_vol_mem.select_convs(2) abs_vol_mem.surjective unit.exhaust)+
  done

lemma sync_ref_NE:
  "from \<noteq> to \<Longrightarrow>
  sim_normE_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.read ind)
    (Abs4_CommonDefs.sync ind to from)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sync_def sim_normE_ext_choice_conc intro!: sync_fail_ref_NE)
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac[!] sva spa svc spc vc svc' spc')
  apply(clarsimp_all simp add: sync_defs brute_force)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: read_defs is_sync_tgt_def brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps)
  apply(iterative_simp)+
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] sva)
  apply(clarsimp_all simp add: in_view_def abs_reg_state_def abs_vol_mem.truncate_def)
  apply(case_tac[!] to, clarsimp_all simp add: comb_simps image_def tgt_notin_srcs
    comb_mono tgt_other_i_in_srcs dest!: tgt_other_i_in_srcsD)
  apply(safe)
  apply(tactic {* distinct_subgoals_tac *})
  thm allE
  apply(rule_tac[!] x="fst ` unwr" in exI)
  apply(safe)
  apply(clarsimp_all simp add: image_def override_on_def comb_simps tgt_other_i_in_srcs
    x_in_fst_simp x_witness_simp INST_disj_exhaust ex_vol_mem_assignment
    min_def intro!: ext)
  apply(safe)
  apply(clarsimp_all simp add: image_def override_on_def comb_simps tgt_other_i_in_srcs
    x_in_fst_simp x_witness_simp INST_disj_exhaust
    min_def intro!: ext)
  apply(metis INST.exhaust)+
  done

lemma sync_proj_simps:
  "(Abs4_CommonDefs.sync_success ind to from)\<Down>NE = empty"
  "(Abs4_CommonDefs.sync_degrade ind to from)\<Down>NN = empty"
  "(Abs4_CommonDefs.sync_degrade_write ind to from)\<Down>NN = empty"
  "(Abs4_CommonDefs.sync_fail ind to from)\<Down>NN = empty"
  "(Abs4_CommonDefs.sync_success ind to from)\<Down>R = empty"
  "(Abs4_CommonDefs.sync_degrade ind to from)\<Down>R = empty"
  "(Abs4_CommonDefs.sync_degrade_write ind to from)\<Down>R = empty"
  "(Abs4_CommonDefs.sync_fail ind to from)\<Down>R = empty"
  "(Abs4_CommonDefs.sync_restart ind to from)\<Down>NN = empty"
  "(Abs4_CommonDefs.sync_restart ind to from)\<Down>NE = empty"
  by(auto simp add: Abs4_CommonDefs.sync_defs focused_fail'_def brute_force)

lemma INST_fun_eq_iff:
  "(f = f') = (f Primary = f' Primary \<and> f Secondary = f' Secondary)"
  apply(rule iffI, simp)
  apply(rule ext, case_tac x, simp_all)
  done

lemma abs5_read_proj_simps':
  "(Abs5_CommonDefs.read_success ind)\<Down>NE = empty"
  "(Abs5_CommonDefs.read_degrade ind)\<Down>NN = empty"
  "(Abs5_CommonDefs.read_fail ind)\<Down>NN = empty"
  "(Abs5_CommonDefs.read_success ind)\<Down>R = empty"
  "(Abs5_CommonDefs.read_degrade ind)\<Down>R = empty"
  "(Abs5_CommonDefs.read_fail ind)\<Down>R = empty"
  "(Abs5_CommonDefs.restart_mangle_sp)\<Down>NE = empty"
  "(Abs5_CommonDefs.restart_mangle_sp)\<Down>NN = empty"
  by(auto simp add: read_defs brute_force)

lemma sync_restart_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.restart_mangle_sp)
    (Abs4_CommonDefs.sync_restart ind to from)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_def sync_proj_simps simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(subst (asm) project_defs)
  apply(clarsimp_all simp add: sync_defs brute_force is_sync_tgt_def simp del: project_defs)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps abs_vol_mem.truncate_def)
  apply(safe)
  apply(case_tac[!] to, clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps
    ex_vol_mem_assignment
    is_arrow_tgt_2nd_simp arrow_srcs_override_SBot override_triv_iff K_beta override_on_with_K_LDead_le)
  done

lemma sync_restart_write_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.restart_mangle_sp)
    (Abs4_CommonDefs.sync_restart_write ind to from)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_def sync_proj_simps simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(subst (asm) project_defs)
  apply(clarsimp_all simp add: sync_defs brute_force is_sync_tgt_def simp del: project_defs)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps)
  apply(safe)
  apply(case_tac[!] to, clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps
    all_conj in_view_def abs_vol_mem.truncate_def
    is_arrow_tgt_2nd_simp arrow_srcs_override_SBot override_triv_iff K_beta override_on_with_K_LDead_le)
  apply(clarsimp_all simp add: abs_reg_state_def in_view_def comb_mono comb_one_True
    ex_vol_mem_assignment
    override_on_def min_def intro!: eq_refl)
  apply (smt comb_rel_writ.simps(1) comb_rel_writ.simps(2) comb_simps(7) order_refl)+
  done

lemma sync_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.read ind)
    (Abs4_CommonDefs.sync ind to from)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!: sync_ref_NN sync_ref_NE del: subsetI)
  apply(auto simp only: sync_def first_stage simS_wpc_choice_conc sync_proj_simps)
  apply(simp_all only: read_def first_stage abs5_read_proj_simps')
  apply(rule simS_wpc_choice_abs_I2)
  apply(rule simS_wpc_choice_abs_I1)
  apply(insert sync_restart_ref[simplified sim_ext_def])[1]
  apply(force)
  apply(rule simS_wpc_choice_abs_I2)
  apply(rule simS_wpc_choice_abs_I1)
  apply(insert sync_restart_write_ref[simplified sim_ext_def])
  apply(force)
  done
 
lemma read_proj_simps:
  "(Abs4_CommonDefs.read_success ind inst)\<Down>NE = empty"
  "(Abs4_CommonDefs.read_corrupt ind inst)\<Down>NN = empty"
  "(Abs4_CommonDefs.read_degrade ind inst)\<Down>NN = empty"
  "(Abs4_CommonDefs.read_degrade_reliable ind inst)\<Down>NN = empty"
  "(Abs4_CommonDefs.focused_adm_fail ind inst)\<Down>NN = empty"
  "(Abs4_CommonDefs.focused_fail' ind inst)\<Down>NN = empty"
  "(Abs4_CommonDefs.read_success ind inst)\<Down>R = empty"
  "(Abs4_CommonDefs.read_corrupt ind inst)\<Down>R = empty"
  "(Abs4_CommonDefs.read_degrade ind inst)\<Down>R = empty"
  "(Abs4_CommonDefs.read_degrade_reliable ind inst)\<Down>R = empty"
  "(Abs4_CommonDefs.focused_adm_fail ind inst)\<Down>R = empty"
  "(Abs4_CommonDefs.focused_fail' ind inst)\<Down>R = empty"
  by(auto simp add: Abs4_CommonDefs.read_defs focused_fail'_def brute_force)

lemma[no_throw]:
  "doesnt_throw_raw ((Abs4_CommonDefs.read_degrade ind inst)\<Down>NE) FailRead"
  "doesnt_throw_raw ((Abs4_CommonDefs.read_degrade_reliable ind inst)\<Down>NE) FailRead"
  "doesnt_throw_raw ((Abs4_CommonDefs.focused_fail' ind inst)\<Down>NE) FailRead"
  "\<forall>v. doesnt_throw_raw ((Abs4_CommonDefs.focused_fail' ind inst)\<Down>NE) (Degraded v)"
  "doesnt_throw_raw ((Abs4_CommonDefs.focused_adm_fail ind inst)\<Down>NE) FailRead"
  "\<forall>v. doesnt_throw_raw ((Abs4_CommonDefs.focused_adm_fail ind inst)\<Down>NE) (Degraded v)"
  "\<forall>v. doesnt_throw_raw ((Abs4_CommonDefs.read_corrupt ind inst)\<Down>NE) (Degraded v)"
  by(auto simp add: doesnt_throw_raw_def Abs4_CommonDefs.read_defs focused_fail'_def brute_force)

definition abs5_read_corrupt where
  "abs5_read_corrupt ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailRead), sv, sp')|sp' mode'.
        mode' \<in> {reg_state sp ind, LDead}
        \<and> sp' = sp\<lparr> reg_state := (reg_state sp)(ind := mode') \<rparr>
    }"
        
definition abs5_read_degrade_EU where
  "abs5_read_degrade_EU ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp' deg reg_state' reg_state'' a_st.
        ind \<in> deg
        \<and> reg_state' = override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg
        \<and> (reg_state'' = reg_state' \<or> reg_state'' = reg_state'(ind := LDead))
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> (a_st \<le> admin_state sp)
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := reg_state'',
            admin_state := a_st
        \<rparr>
  }"
 
definition abs5_read' where
  "abs5_read' ind \<equiv> read ind [+] abs5_read_corrupt ind [+] abs5_read_degrade_EU ind"

lemmas abs5_read'_defs =
  abs5_read'_def
  read_defs
  abs5_read_corrupt_def
  abs5_read_degrade_EU_def

lemma read_ref_NN:
  "sim_normN_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (abs5_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs4_CommonDefs.read_def)
  apply(simp_all add: read_proj_simps first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs
    focused_fail'_def Abs4_CommonDefs.sync_defs
    arrow_is_def brute_force min_def
    split: split_if_asm exceptions.split_asm)
  apply(simp_all (no_asm) add: abs5_read'_defs brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def comb_simps abs_reg_state_def)
  apply(safe)
  apply(subgoal_tac "(ind, inst) \<in> in_view spc")
  apply(case_tac[!] inst)
  apply(auto simp add: in_view_eqI)
  done

lemma read_ref_NE:
  "sim_normE_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (abs5_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: Abs4_CommonDefs.read_def)
  apply(simp_all add: read_proj_simps first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs
    focused_fail'_def Abs4_CommonDefs.sync_defs
    arrow_is_def brute_force min_def
    split: split_if_asm exceptions.split_asm)
  apply(simp_all (no_asm) add: abs5_read'_defs brute_force min_def)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(simp_all add: abs5_abs4_strel_def in_view_ref_simps comb_simps abs_reg_state_ref_simps all_conj)
  apply(case_tac[!] inst, clarsimp_all)
  apply(case_tac[!] sva)
  apply(safe)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp
    abs_vol_mem.truncate_def
  )
  done

definition view_not_on where
  "view_not_on ind inst \<equiv> {(svc, spc). (ind, inst) \<notin> in_view spc}"

lemma read_or_fail_ref_NN: 
  "sim_normN_ext (abs5_abs4_strel \<inter> UNIV \<times>
    (arrow_is v ind \<inter> (if inst \<in> arrow_srcs v then UNIV else {}) \<inter> view_not_on ind (other_i inst)))
    (read ind)
    (Abs4_CommonDefs.read_or_fail ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: read_or_fail_def)
  apply(simp_all add: first_stage read_proj_simps vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs brute_force
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: brute_force read_defs min_def)
  apply(case_tac[!] svc, clarsimp_all)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_ref_simps in_view_ref_simps
    view_not_on_def arrow_is_def)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_ref_simps in_view_ref_simps
    ex_override_simp)
  done

lemma read_or_fail_ref_NE: 
  "sim_normE_ext (abs5_abs4_strel \<inter> UNIV \<times>
    (arrow_is v ind \<inter> (if inst \<in> arrow_srcs v then UNIV else {}) \<inter> view_not_on ind (other_i inst)))
    (read ind)
    (Abs4_CommonDefs.read_or_fail ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: read_or_fail_def)
  apply(simp_all add: first_stage read_proj_simps vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_or_fail_defs
    brute_force
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: brute_force read_defs min_def)
  apply(safe)
  apply(case_tac[!] svc, clarsimp_all)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_ref_simps in_view_ref_simps
    view_not_on_def arrow_is_def )
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] "arrow spc ind")
  apply(simp_all)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp abs_vol_mem.truncate_def)
  apply (smt other_i.simps(1))
  defer
  apply (smt other_i.simps(2))
  apply(case_tac[!] "ind \<in> arr_del", simp_all)
  apply (smt other_i.simps(1) other_i.simps(2))+
  done

lemma read_or_fail_ref: 
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times>
    (arrow_is v ind \<inter> (if inst \<in> arrow_srcs v then UNIV else {}) \<inter> view_not_on ind (other_i inst)))

    (read ind)
    (Abs4_CommonDefs.read_or_fail ind inst)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!:
    read_or_fail_ref_NE read_or_fail_ref_NN del: subsetI split del: split_if)
  apply(auto simp add: simS_wpc_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: read_or_fail_def)
  apply(simp_all add: first_stage read_proj_simps vol_ops_project_simps
    if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_or_fail_defs
    brute_force
    split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: brute_force read_defs min_def)
  apply(case_tac[!] svc, clarsimp_all)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def abs_reg_state_ref_simps in_view_ref_simps
    view_not_on_def arrow_is_def )
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] inst)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp 
    abs_vol_mem.truncate_def ex_vol_mem_assignment)
  done

definition read_silence_deg where
  read_silence_deg_def[monad_def]:
  "read_silence_deg ind \<equiv>
    <try> read ind <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (get_exc_val_cnt v)
        | _ \<Rightarrow> throw e)"

declare read_or_fail_ref[ref_thm]
declare sim_ext_return[ref_thm]
declare sim_ext_throw[ref_thm]
declare
  sim_ext_bind_rt[ref_bind]
  sim_ext_try_catch_rt[ref_try_catch]
  sim_ext_exc_case_rt[ref_exc_case]
  sim_ext_if_rt[ref_if]
  sim_ext_if_full_rt[ref_if_full]

lemma read_or_fail_silence_deg_ref[ref_thm]:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times>(arrow_is SEq ind \<inter> view_not_on ind (other_i inst)))
    (read_silence_deg ind)
    (Abs4_StartupDefs.read_or_fail_silence_deg ind inst)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(atomize, erule_tac x=SEq in allE)
  apply(erule sim_ext_mono)
  apply(auto simp add: view_not_on_def arrow_is_def)
  done

definition abs5_2nd_read_corrupt where
  "abs5_2nd_read_corrupt ind \<equiv> throw FailRead"
        
definition abs5_2nd_read_degrade_EU where
  "abs5_2nd_read_degrade_EU ind \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded EU)), sv', sp')|sv' sp' deg a_st.
        ind \<in> deg
        \<and> (a_st \<noteq> LDead \<or> a_st = admin_state sp)
        \<and> a_st \<le> admin_state sp
        \<and> sv' = sv\<lparr> abs_card_mode := min (abs_card_mode sv) ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
            admin_state := a_st
        \<rparr>
  }"
 
definition abs5_2nd_read' where
  "abs5_2nd_read' ind \<equiv> read ind [+] abs5_2nd_read_corrupt ind [+] abs5_2nd_read_degrade_EU ind"

lemmas abs5_2nd_read'_defs =
  abs5_2nd_read'_def
  read_defs
  abs5_2nd_read_corrupt_def
  abs5_2nd_read_degrade_EU_def

definition other_in_view where
  "other_in_view ind inst \<equiv> {(svc, spc). (ind, other_i inst) \<in> in_view spc }"

lemma sec_read_ref_NN:
  "sim_normN_ext (abs5_abs4_strel \<inter> UNIV \<times> (arrow_is SEq ind \<inter> other_in_view ind inst))
    (abs5_2nd_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs4_CommonDefs.read_def)
  apply(simp_all add: read_proj_simps first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs
    focused_fail'_def Abs4_CommonDefs.sync_defs
    arrow_is_def other_in_view_def brute_force min_def
    split: split_if_asm exceptions.split_asm)
  apply(simp_all (no_asm) add: abs5_2nd_read'_defs brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def comb_simps abs_reg_state_def)
  apply (smt INST.exhaust in_view_ref_simps(6))
  done

lemma sec_read_ref_NE:
  "sim_normE_ext (abs5_abs4_strel \<inter> UNIV \<times> (arrow_is SEq ind \<inter> other_in_view ind inst))
    (abs5_2nd_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc ec svc' spc')
  apply(simp_all only: Abs4_CommonDefs.read_def)
  apply(simp_all add: read_proj_simps first_stage vol_ops_project_simps if_choice
    restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs
    focused_fail'_def Abs4_CommonDefs.sync_defs
    arrow_is_def other_in_view_def brute_force min_def
    split: split_if_asm exceptions.split_asm)
  apply(simp_all (no_asm) add: abs5_2nd_read'_defs brute_force min_def)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(simp_all add: abs5_abs4_strel_def in_view_ref_simps comb_simps abs_reg_state_ref_simps all_conj)
  apply(case_tac[!] inst, clarsimp_all)
  apply(safe)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp
    abs_vol_mem.truncate_def
  )
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply (smt in_view_ref_simps(6))+
  done

definition abs5_check_pub_same where
  "abs5_check_pub_same pind \<equiv>
       <try> do {
        abs5_read' pind;
        (<try> do {
            abs5_2nd_read' pind;
            return ()
        } <catch>
            (\<lambda>e. case e of
                FailRead \<Rightarrow> access_silence_deg pind 
                | Degraded _ \<Rightarrow> return ()
                | _ \<Rightarrow> throw e
            )
        )
   } <catch> (\<lambda>e. case e of
        FailRead \<Rightarrow> access_silence_deg pind 
        | Degraded v \<Rightarrow> (
            if v = EU
            then read_silence_deg pind \<guillemotright> return ()
            else return ()
        )
        | _ \<Rightarrow> throw e
    )
  "

definition access_silence_deg' where
  "access_silence_deg' \<equiv> access_silence_deg"

thm abs5_read'_defs

lemma abs5_read_proj_simps'':
  "(abs5_read_degrade_EU ind)\<Down>NN = empty"
  "(abs5_read_corrupt ind)\<Down>NN = empty"
  "(abs5_read_degrade_EU ind)\<Down>R = empty"
  "(abs5_read_corrupt ind)\<Down>R = empty"
  "(abs5_2nd_read_degrade_EU ind)\<Down>NN = empty"
  "(abs5_2nd_read_corrupt ind)\<Down>NN = empty"
  "(abs5_2nd_read_degrade_EU ind)\<Down>R = empty"
  "(abs5_2nd_read_corrupt ind)\<Down>R = empty"
  by(auto simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force)

lemmas abs5_read_proj_simps = 
  abs5_read_proj_simps'
  abs5_read_proj_simps''

lemma fun_forall_arg_eq:
  "(f = g) = (\<forall>x. f x = g x)"
  by(auto)

lemma double_min_LDegraded_override_simp:
  "override_on (override_on f (min LDegraded \<circ> f) A)
    (min LDegraded \<circ> override_on f (min LDegraded \<circ> f) A) B
  = override_on f (min LDegraded o f) (A \<union> B)"
  by(auto simp add: override_on_def)

lemma first_read_post_NN:
  "\<lbrace>arrow_is SEq ind\<rbrace> Abs4_CommonDefs.read ind inst
  \<lbrace>>N K (arrow_is SEq ind \<inter> other_in_view ind (other_i inst))\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(simp_all add: Abs4_CommonDefs.read_def first_stage read_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs arrow_is_def other_in_view_def
    brute_force in_view_def)
  apply(case_tac inst, simp_all)
  done

lemma first_read_post_NE:
  "\<lbrace>arrow_is SEq ind\<rbrace>
    Abs4_CommonDefs.read ind inst
  \<lbrace>@N \<lambda>e.
    if e = FailRead
    then is_sync_tgt ind inst
    else if e = Degraded EU
    then (arrow_is SEq ind \<inter> view_not_on ind inst)
    else UNIV
  \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_def)
  apply(simp_all add: Abs4_CommonDefs.read_def restart_mangle_sv_proj_simps first_stage read_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs arrow_is_def is_sync_tgt_def
    view_not_on_def in_view_ref_simps
    brute_force)
  done

lemma abs_state_ge_LDegraded:
  "\<lbrakk>(svc, spc) \<in> arrow_is SEq ind; (ind, inst) \<in> reliable spc\<rbrakk>
       \<Longrightarrow> (LDegraded \<le> abs_reg_state spc ind) = True"
  "(ind, inst) \<in> in_view spc
       \<Longrightarrow> (LDegraded \<le> abs_reg_state spc ind) = True"
  apply(auto simp add: abs_reg_state_def)
  apply(case_tac[!] inst, simp_all add: LDegraded_le_comb arrow_is_def in_view_def)
  done

lemma first_read_ref':
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (abs5_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!: read_ref_NN read_ref_NE)
  apply(auto simp add: simS_wpc_def Abs4_CommonDefs.read_def first_stage read_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs brute_force)
  apply(simp_all (no_asm) add: abs5_read'_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(simp_all add: abs5_abs4_strel_def in_view_ref_simps comb_simps abs_reg_state_ref_simps all_conj
    abs_state_ge_LDegraded)
  apply(case_tac[!] inst, clarsimp_all)
  apply(safe)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp
    abs_state_ge_LDegraded abs_vol_mem.truncate_def ex_vol_mem_assignment
  )
  apply (smt fun_upd_apply less_eq_log_state.simps(1) min_LDead(1) min_max.inf_le2
    o_apply order_refl override_on_apply_in override_on_apply_notin)+
  done

lemma first_read_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (abs5_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (arrow_is SEq ind \<inter> other_in_view ind (other_i inst))))
    (\<lambda>ea ec. force_same_and abs5_abs4_strel ea ec \<inter>
        UNIV \<times>
        (if ec = FailRead
        then is_sync_tgt ind inst
        else if ec = Degraded EU
        then (arrow_is SEq ind \<inter> view_not_on ind inst)
        else UNIV)
        
    )
    abs5_abs4_strel
  "
  apply(rule first_read_ref'[THEN sim_ext_mono,
      THEN sim_ext_add_NN_inv[OF true_postNN first_read_post_NN],
      THEN sim_ext_add_NE_inv[OF true_postNE first_read_post_NE],
       THEN sim_ext_mono_norm, THEN sim_ext_mono_exc])
  apply(auto simp add: K_def force_same_and_def) 
  done

lemma sec_read_ref':
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (arrow_is SEq ind \<inter> other_in_view ind inst))
    (abs5_2nd_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!: sec_read_ref_NN sec_read_ref_NE)
  apply(auto simp add: simS_wpc_def Abs4_CommonDefs.read_def first_stage read_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs4_CommonDefs.read_defs brute_force)
  apply(simp_all (no_asm) add: abs5_2nd_read'_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(simp_all add: abs5_abs4_strel_def in_view_ref_simps comb_simps abs_reg_state_ref_simps all_conj
    abs_state_ge_LDegraded)
  apply(case_tac[!] inst, clarsimp_all)
  apply(safe)
  apply(clarsimp_all simp add: ex_override_simp override_updated_simp
    abs_reg_state_ref_simps in_view_ref_simps double_override override_le_simp
    abs_state_ge_LDegraded abs_vol_mem.truncate_def ex_vol_mem_assignment
  )
  apply (smt fun_upd_apply less_eq_log_state.simps(1) min_LDead(1) min_max.inf_le2
    o_apply order_refl override_on_apply_in override_on_apply_notin)+
  done

lemma sec_read_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (arrow_is SEq ind \<inter> other_in_view ind inst))
    (abs5_2nd_read' ind)
    (Abs4_CommonDefs.read ind inst)
    (force_same_and abs5_abs4_strel)
    (\<lambda>ea ec. force_same_and abs5_abs4_strel ea ec \<inter>
        UNIV \<times>
        (if ec = FailRead
        then is_sync_tgt ind inst
        else UNIV)
        
    )
    abs5_abs4_strel
  "
  apply(rule sec_read_ref'[THEN sim_ext_mono
    , THEN sim_ext_add_NE_inv[OF true_postNE first_read_post_NE]
    , THEN sim_ext_mono, THEN sim_ext_mono_exc
    , where R2="abs5_abs4_strel \<inter> UNIV \<times> other_in_view ind inst"
    ]
  )
  apply(auto simp add: K_def force_same_and_def)
  done

lemma[no_throw]:
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) FailRead"
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) FailMode"
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((read_degrade ind)\<Down>NE) InvalidInput"
  "doesnt_throw_raw ((read_fail ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((read_fail ind)\<Down>NE) FailRead"
  "doesnt_throw_raw ((read_fail ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((read_fail ind)\<Down>NE) InvalidInput"
  "\<forall>v. doesnt_throw_raw ((read_fail ind)\<Down>NE) (Degraded v)"
  "doesnt_throw_raw ((read_fail ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) FailRead"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) FailMode"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) InvalidInput"
  "\<forall>v. doesnt_throw_raw ((read_success ind)\<Down>NE) (Degraded v)"
  "doesnt_throw_raw ((read_success ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) FailRead"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) FailMode"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((abs5_read_degrade_EU ind)\<Down>NE) InvalidInput"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) FailRead"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) FailMode"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((abs5_2nd_read_degrade_EU ind)\<Down>NE) InvalidInput"
  "doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) FailMode"
  "\<forall>v. doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) (Degraded v)"
  "doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) StartedExc"
  "doesnt_throw_raw ((abs5_2nd_read_corrupt ind)\<Down>NE) InvalidInput"
  "doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) FailLLRead"
  "doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) FailWrite"
  "doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) FailMode"
  "doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) InvalidInput"
  "\<forall>v. doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) (Degraded v)"
  "doesnt_throw_raw ((abs5_read_corrupt ind)\<Down>NE) StartedExc"
  by(auto simp add: doesnt_throw_raw_def abs5_read'_defs abs5_2nd_read'_defs brute_force)

lemma check_pub_same_inclusion_NN:
  "(v, sv', sp') \<in> ((abs5_check_pub_same ind)\<Down>NN) (sv, sp) \<Longrightarrow>
    (v, sv', sp') \<in> ((access_silence_deg ind)\<Down>NN) (sv, sp)"
  apply(simp only: access_silence_deg'_def[THEN symmetric])
  apply(simp_all only: abs5_check_pub_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_2nd_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: access_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_silence_deg_def)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: access_silence_deg'_def access_silence_deg_def access_def)
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] sv, case_tac[!] sp, clarsimp_all simp add: double_min_LDegraded_override_simp)
  apply (smt Un_iff linorder_linear min_max.inf_absorb1 min_max.inf_commute UnI1 min_absorb1
    min_max.le_inf_iff)+
  done

use "../abs2/trans_simproc.ML"

simproc_setup trans_simproc 
  ("r \<le> s")
  = {* K Trans_Simproc.trans_simproc *}

lemma check_pub_same_inclusion_NE:
  "(v, sv', sp') \<in> ((abs5_check_pub_same ind)\<Down>NE) (sv, sp) \<Longrightarrow>
    (v, sv', sp') \<in> ((access_silence_deg ind)\<Down>NE) (sv, sp)"
  apply(simp only: access_silence_deg'_def[THEN symmetric])
  apply(simp_all only: abs5_check_pub_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_2nd_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: access_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_silence_deg_def)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: access_silence_deg'_def access_silence_deg_def access_def)
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] sv, case_tac[!] sp, clarsimp_all simp add:
    double_min_LDegraded_override_simp split: split_if_asm)
  apply (smt min_LDead(1) min_max.inf_le1 min_max.inf_le2 o_def order_trans override_on_def)+
  done

lemma check_pub_same_inclusion_R:
  "(sv', sp') \<in> ((abs5_check_pub_same ind)\<Down>R) (sv, sp) \<Longrightarrow>
    (sv', sp') \<in> ((access_silence_deg ind)\<Down>R) (sv, sp)"
  apply(simp only: access_silence_deg'_def[THEN symmetric])
  apply(simp_all only: abs5_check_pub_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_2nd_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: access_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: access_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: read_silence_deg_def)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: access_silence_deg'_def access_silence_deg_def access_def)
  apply(clarsimp_all simp add: read_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(case_tac[!] sv, case_tac[!] sp, clarsimp_all simp add:
    double_min_LDegraded_override_simp split: split_if_asm)
  apply(clarsimp_all simp add: min_def override_triv_iff)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply ((smt min_LDead(1) min_max.inf_le1 min_max.inf_le2 o_def order_trans override_on_def
    order_refl UnI1)+)
  done

lemma check_pub_same_inclusion:
  "\<forall>s. (abs5_check_pub_same ind) s \<subseteq> (access_silence_deg ind) s"
  apply(intro allI proj_subset)
  apply(auto intro: check_pub_same_inclusion_NN check_pub_same_inclusion_NE check_pub_same_inclusion_R)
  done
 
text {* This operation never throws FailRead nor Degraded, but we need
  the lemma in order to get the same relation as for the first read *}
lemma just_sync_silence_deg_post_NE:
  "\<lbrace>P\<rbrace>
  just_sync_silence_deg to from ind
  \<lbrace>@N \<lambda>ec.
        if ec = FailRead
        then is_sync_tgt ind Primary
        else if ec = Degraded EU
        then arrow_is SEq ind \<inter> view_not_on ind Primary
        else UNIV\<rbrace>
  "
  apply(auto simp add: HoareNE_alt Hoare_def Box_def)
  apply(simp_all only: just_sync_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: just_sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: sync_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice sync_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: sync_defs brute_force split: split_if_asm exceptions.split_asm)
  done

declare access_silence_deg_def[monad_def]
declare access_def[monad_def]
declare sync_ref[ref_thm]

lemma just_sync_silence_deg_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.access_silence_deg ind)
    (Abs4_CommonDefs.just_sync_silence_deg to from ind)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *}, simp)

lemma sec_just_sync_silence_deg_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (Abs5_CommonDefs.access_silence_deg ind)
    (Abs4_CommonDefs.just_sync_silence_deg to from ind)
    (force_same_and abs5_abs4_strel)
    (\<lambda>ea ec. force_same_and abs5_abs4_strel ea ec \<inter> UNIV \<times>
        (if ec = FailRead
        then is_sync_tgt ind Primary
        else if ec = Degraded EU
        then arrow_is SEq ind \<inter> view_not_on ind Primary
        else UNIV)
    )
    abs5_abs4_strel
  "
  apply(rule just_sync_silence_deg_ref[
    THEN sim_ext_add_NE_inv[OF true_postNE just_sync_silence_deg_post_NE]
    , THEN sim_ext_mono_exc
    ]
  )
  apply(auto simp add: K_def)
  done

lemma check_pub_same_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (Abs5_CommonDefs.access_silence_deg ind)
    (check_pub_same ind)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(rule sim_ext_mono_abs_comp[OF _ check_pub_same_inclusion])
  apply(simp only: abs5_check_pub_same_def check_pub_same_def)
  apply(rule sim_ext_try_catchI)
  apply(rule sim_ext_bindI[OF first_read_ref])
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_try_catchI)
  apply(rule sim_ext_bindI[OF sec_read_ref])
  apply(intro allI)
  apply(rule sim_ext_return_diffI)
  apply(simp add: force_same_and_def)
  apply(intro allI)
  apply(auto simp add: force_same_and_diff' force_same_and_same split: exceptions.split
    intro!: sim_exts_empty_rel sim_ext_throw_diffI sim_ext_return_diffI
    intro!: sim_ext_bindI
    intro: sec_just_sync_silence_deg_ref
    just_sync_silence_deg_ref
    read_or_fail_silence_deg_ref[THEN sim_ext_mono]
  )
  apply(auto simp add: force_same_and_def)
  done

lemma get_arrow_post_NN:
  "\<lbrace>P\<rbrace> get_arrow ind \<lbrace>>N \<lambda>v. if v = SBot then {} else arrow_is v ind  \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def)
  apply(auto simp add: get_arrow_defs arrow_is_def brute_force)
  done

lemma get_arrow_ref':
  "sim_ext abs5_abs4_strel
    (maybe_mess_up_admin_silence_deg \<guillemotright> chooseN UNIV)
    (get_arrow ind)
    (\<lambda>va vc. force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (if vc = SBot then {} else arrow_is vc ind)) va vc)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(rule
    get_arrow_ref[THEN sim_ext_mono
      , THEN sim_ext_add_NN_inv[OF true_postNN get_arrow_post_NN]
      , THEN sim_ext_mono, THEN sim_ext_mono_norm, where R2=abs5_abs4_strel and I2=UNIV and J2=UNIV
    ]
  )
  by(auto simp add: force_same_and_def)

lemma maybe_mess_up_admin_ref:
  "sim_ext abs5_abs4_strel maybe_mess_up_admin_silence_deg
     (Abs4_CommonDefs.maybe_mess_up_admin ind) (force_same_and abs5_abs4_strel)
     (force_same_and abs5_abs4_strel) abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all only: Abs4_CommonDefs.maybe_mess_up_admin_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: get_arrow_defs brute_force split: split_if_asm)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def degrade_admin_def fail_admin_def
    restart_mangle_sp_def
    brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps all_conj)
  apply(clarsimp_all simp add: override_on_def min_def K_def 
    abs_vol_mem.truncate_def ex_vol_mem_assignment)
  apply(metis)
  done
 
declare maybe_mess_up_admin_ref[ref_thm]

lemma just_sync_silence_deg'_ref:
  "from \<noteq> to \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> UNIV \<times> is_sync_tgt ind to)
    (<try> read ind \<guillemotright>
        Abs5_CommonDefs.maybe_mess_up_admin_silence_deg
        <catch> (\<lambda>e. if \<exists>v. e = Degraded v then return () else throw e))
    (just_sync_silence_deg' to from ind)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(simp add: just_sync_silence_deg_def')
  apply(rule sim_ext_try_catchI)
  apply(rule sim_ext_bindI[OF sync_ref])
  apply(simp)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro: maybe_mess_up_admin_ref sim_exts_empty_rel)[1]
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro: sim_ext_return sim_ext_throw intro!: sim_exts_empty_rel)
  done
    
definition rhs where
  "rhs ind \<equiv> (access_silence_deg ind \<guillemotright> maybe_mess_up_admin_silence_deg)"

lemma access_silence_deg_inclusion:
  "\<forall>s.
    (<try> read ind \<guillemotright> Abs5_CommonDefs.maybe_mess_up_admin_silence_deg
    <catch> (\<lambda>e. if \<exists>v. e = Degraded v then return () else throw e)) s
    \<subseteq> rhs ind s"
  apply(intro allI proj_subset)
  apply(auto)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    abs5_read_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_defs degrade_admin_def fail_admin_def brute_force)
  apply(iterative_simp)+
  apply(case_tac[!] b, clarsimp_all)
  apply(simp_all add: rhs_def access_silence_deg_def access_def maybe_mess_up_admin_silence_deg_def
    read_defs fail_admin_def degrade_admin_def brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(metis)+
  done

lemma check_pub_region_ref:
  "sim abs5_abs4_strel
    (Abs5_StartupDefs.check_pub_region ind)
    (Abs4_StartupDefs.check_pub_region ind)
  "
  apply(simp only: sim_def)
  apply(simp only: Abs5_StartupDefs.check_pub_region_def Abs4_StartupDefs.check_pub_region_def)
  apply(rule sim_ext_bindI[OF get_arrow_ref'])
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF check_pub_same_ref])
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule maybe_mess_up_admin_ref)
  apply(simp add: just_sync_silence_deg_def')
  apply(rule sim_ext_mono_abs_comp[OF _ access_silence_deg_inclusion, simplified rhs_def])
  apply(rule sim_ext_try_catchI)
  apply(rule sim_ext_bindI[OF sync_ref, THEN sim_ext_mono])
  apply(clarsimp)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel maybe_mess_up_admin_ref)[1]
  apply(clarsimp simp add: is_sync_tgt_def arrow_is_def)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel sim_ext_return sim_ext_throw)[1]
  apply(simp add: just_sync_silence_deg_def')
  apply(rule sim_ext_mono_abs_comp[OF _ access_silence_deg_inclusion, simplified rhs_def])
  apply(rule sim_ext_try_catchI)
  apply(rule sim_ext_bindI[OF sync_ref, THEN sim_ext_mono])
  apply(clarsimp)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel maybe_mess_up_admin_ref)[1]
  apply(clarsimp simp add: is_sync_tgt_def arrow_is_def)
  apply(case_tac "arrow ba ind", simp_all)[1]
  apply(subst force_same_and_def)
  apply(auto intro: sim_ext_return sim_ext_throw sim_exts_empty_rel)
  done

definition abs5_fetch_same where
  "abs5_fetch_same pind \<equiv>
    <try> do {
        abs5_read' pind
    } <catch> (\<lambda>e. case e of
        FailRead \<Rightarrow> read pind 
        | Degraded v \<Rightarrow> (
            if v = EU
            then do {
                cnt \<leftarrow> read pind;
                throw (Degraded (EC cnt))
            }
            else throw e
        )
        | _ \<Rightarrow> throw e
    )
  "

lemma abs5_fetch_same_inclusion_NN:
  "(v, sv', sp') \<in> ((abs5_fetch_same ind)\<Down>NN) (sv, sp) \<Longrightarrow>
    (v, sv', sp') \<in> ((read ind)\<Down>NN) (sv, sp)"
  apply(simp_all only: abs5_fetch_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  done
 
lemma FailLLRead_exhaust:
  "\<lbrakk> e \<noteq> FailWrite; e \<noteq> FailRead; e \<noteq> FailMode;
  \<forall>c. e \<noteq> Degraded c; \<forall>c. e \<noteq> DegradedH c; e \<noteq> StartedExc; e \<noteq> InvalidInput
   \<rbrakk> \<Longrightarrow> (e = FailLLRead) = True"
  "\<lbrakk> e \<noteq> FailWrite; e \<noteq> FailRead; e \<noteq> FailMode;
  \<forall>c. e \<noteq> Degraded c; \<forall>c. e \<noteq> DegradedH c; e \<noteq> StartedExc; e \<noteq> InvalidInput
   \<rbrakk> \<Longrightarrow> (FailLLRead = e) = True"
  by(case_tac[!] e, auto)

lemma abs5_fetch_same_inclusion_NE:
  "(v, sv', sp') \<in> ((abs5_fetch_same ind)\<Down>NE) (sv, sp) \<Longrightarrow>
    (v, sv', sp') \<in> ((read ind)\<Down>NE) (sv, sp)"
  apply(simp_all only: abs5_fetch_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(case_tac[!] sp)
  apply(case_tac[!] sv, clarsimp_all simp add: FailMode_exhaust Degraded_exhaust)
  apply(iterative_simp split: split_if_asm)+
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: double_min_LDegraded_override_simp split: split_if_asm)
  apply((metis
    min_max.le_inf_iff o_apply override_on_apply_in override_on_apply_notin
    le_LDead min_max.le_infE o_def order_eq_refl override_on_def
    UnI1
  )+)
  done

lemma abs5_fetch_same_inclusion_R:
  "(sv', sp') \<in> ((abs5_fetch_same ind)\<Down>R) (sv, sp) \<Longrightarrow>
    (sv', sp') \<in> ((read ind)\<Down>R) (sv, sp)"
  apply(simp_all only: abs5_fetch_same_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_read'_def read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice
    abs5_read_proj_simps restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: abs5_read'_defs abs5_2nd_read'_defs brute_force min_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(case_tac[!] sp)
  apply(case_tac[!] sv, clarsimp_all simp add: FailMode_exhaust Degraded_exhaust)
  apply(iterative_simp split: split_if_asm)+
  apply(clarsimp_all simp add: override_on_ref_simps double_min_LDegraded_override_simp)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply((metis
    min_max.le_inf_iff o_apply override_on_apply_in override_on_apply_notin
    le_LDead min_max.le_infE o_def order_eq_refl override_on_def
    UnI1
  )+)
  done

lemma abs5_fetch_same_inclusion:
  "\<forall>s. (abs5_fetch_same ind) s \<subseteq> read ind s"
  apply(intro allI proj_subset)
  apply(auto intro: abs5_fetch_same_inclusion_NN abs5_fetch_same_inclusion_NE abs5_fetch_same_inclusion_R)
  done

lemma fetch_same_ref[ref_thm]:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> arrow_is SEq ind)
    (Abs5_CommonDefs.read ind)
    (Abs4_FetchDefs.fetch_same ind)
    (force_same_and abs5_abs4_strel)
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  apply(rule sim_ext_mono_abs_comp[OF _ abs5_fetch_same_inclusion])
  apply(simp only: abs5_fetch_same_def fetch_same_def)
  apply(rule sim_ext_try_catchI)
  apply(rule first_read_ref[THEN sim_ext_mono_norm])
  apply(clarsimp simp add: force_same_and_def)
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel sim_ext_return sim_ext_throw sync_ref
    intro!: sim_ext_bindI[OF read_or_fail_ref[THEN sim_ext_mono]]
    simp add: force_same_and_diff
    split: exceptions.split)
  apply(subst force_same_and_def, auto intro!: sim_exts_empty_rel sim_ext_throw)
  done

lemma vol_ops_refs[ref_thm]:
  "sim abs5_abs4_strel get_abs_mode get_abs_mode"
  "sim abs5_abs4_strel get_abs_updated get_abs_updated"
  "sim abs5_abs4_strel (set_abs_updated pinds) (set_abs_updated pinds)"
  by(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def
    abs5_abs4_strel_def brute_force abs_vol_mem.truncate_def)

declare sims_throw(1)[ref_thm]
declare sim_return[ref_thm]

lemma fetch_ref':
  "sim abs5_abs4_strel
    (Abs5_FetchDefs.fetch ind)
    (Abs4_FetchDefs.fetch ind)
  "
  apply(simp_all only: Abs5_FetchDefs.fetch_def Abs4_FetchDefs.fetch_def)
  apply(rule sim_bindI[OF vol_ops_refs(1)])
  apply(intro allI sim_bindI)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(subst add_chooseN[where A=UNIV]) 
  apply(simp)
  apply(subst bind_assoc[THEN sym])
  apply(simp only: sim_def)
  apply(rule sim_ext_bindI[OF get_arrow_ref'])
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(atomize, erule_tac x=abs5_abs4_strel in allE, erule_tac x=SR in allE, simp)
  apply(erule sim_ext_mono)
  apply(clarsimp simp add: arrow_is_def view_not_on_def in_view_def)
  apply(case_tac[!] vc, simp_all)
  apply(tactic {* (refinement_ins_tac false Ref_Tac.default_simp @{context}) 1 *})
  apply(atomize, erule_tac x=abs5_abs4_strel in allE, erule_tac x=SL in allE, simp)
  apply(erule sim_ext_mono)
  apply(clarsimp simp add: arrow_is_def view_not_on_def in_view_def)
  done

declare check_pub_region_ref[ref_thm]

lemma startup1_ref[ref_thm]:
  "sim abs5_abs4_strel
    Abs5_StartupDefs.startup1 
    Abs4_StartupDefs.startup1 
  "
  apply(simp add: sim_def sim_ext_def Abs5_StartupDefs.startup1_def Abs4_StartupDefs.startup1_def
    compose_project)
  apply(auto simp add: simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all add: Abs4_StartupDefs.startup1_NN_def Abs4_StartupDefs.startup1_NE_def)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: Abs5_StartupDefs.startup1_NN_def fail_admin_def brute_force) 
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs5_abs4_strel_def in_view_ref_simps abs_reg_state_ref_simps
    override_on_ref_simps)
  apply(clarsimp_all simp add: override_on_def abs_vol_mem.truncate_def)
  done

declare sim_chooseN[ref_thm]
  sim_mapM[ref_mapM]

lemma startup_ref':
  "sim abs5_abs4_strel
    Abs5_StartupDefs.startup 
    Abs4_StartupDefs.startup 
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

definition update_mem_rel
  :: "(('c abs5_vol_state \<times> 'c abs5_state) \<times> (abs_vol_mem \<times> 'c abs4_state))set"     
  where
  "update_mem_rel \<equiv> {((sva, spa), (svc, spc)).
    (\<forall>ind \<in> abs_updated sva.
        arrow spc ind \<noteq> SBot
        \<and> (is_arrow_tgt (arrow spc ind) Primary \<longrightarrow> (ind, Primary) \<in> reliable spc
            \<and> abs4_state.memory spc ind Primary = the (upd_mem sva ind))
        \<and> (is_arrow_tgt (arrow spc ind) Secondary \<longrightarrow> (ind, Secondary) \<in> reliable spc
            \<and> abs4_state.memory spc ind Secondary = the (upd_mem sva ind))
        \<and> (arrow spc ind = SEq \<longrightarrow> the (upd_mem sva ind) = memory spa ind)
    )
  }"

definition precommit_rel where
  "precommit_rel \<equiv> abs5_abs4_strel \<inter> update_mem_rel \<inter>
    sv_abs_mode_equals ACNormal \<times> (not_abs_updated_are_SEq \<inter> all_reliable_inv)"

definition notin_updated :: "pub_indx \<Rightarrow> (abs_vol_mem \<times> 'sp)set" where
  "notin_updated ind \<equiv> {(sv, sp). ind \<notin> abs_updated sv }"

lemma fun_upd_triv':
  "z = f x \<Longrightarrow> f(x := z) = f"
  by(auto)

lemma write_new_ref_NN:
  "ind \<in> dom upd_f \<Longrightarrow>
  sim_normN_ext (precommit_rel \<inter> UNIV \<times>  notin_updated ind)
    (Abs5_UpdateDefs.write_new' upd_f ind)
    (Abs4_UpdateDefs.write_new' upd_f ind)
    (force_same_and precommit_rel)
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc v svc' spc')
  apply(auto simp add: Abs5_UpdateDefs.write_new'_def Abs4_UpdateDefs.write_new'_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: write_new_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_new_defs brute_force min_def split: split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def all_reliable_inv_def
    not_abs_updated_are_SEq_def sv_abs_mode_equals_def
    notin_updated_def
    in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj)
  using [[simp_depth_limit=10]]
  apply(subgoal_tac[!] "arrow spc ind = SEq")
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj fun_upd_triv')
  apply(clarsimp_all simp add: abs_vol_mem.truncate_def)
  apply(safe)
  apply(clarsimp_all simp add: in_view_def abs_reg_state_def override_on_def intro!: ext)
  apply(erule_tac[!] x=ind in allE)+
  apply(clarsimp_all simp add: comb_one_True LDegraded_le_comb_x_True)
  done

lemma write_new_proj_simps:
  "(write_new_success ind inst cnt)\<Down>NE = empty"
  "(write_new_corrupt_restart ind x v)\<Down>NE = empty"
  by(auto simp add: write_new_defs brute_force)

lemma abs4_state_ref_simps:
  "reliable spc = A \<Longrightarrow> spc\<lparr> reliable := A \<rparr> = spc"
  by(simp)

lemma new_abs_reg_state_ref_simps:
  "\<lbrakk> inst' = other_i inst; (ind, inst') \<in> in_view spc \<rbrakk> \<Longrightarrow>
    abs_reg_state (spc\<lparr>arrow := (arrow spc)(ind := point_arrow_to inst) \<rparr>) =
    abs_reg_state spc
  "
  "\<lbrakk> (ind, Primary) \<in> reliable spc; arrow spc ind = SEq \<rbrakk> \<Longrightarrow>
    (abs_reg_state (spc\<lparr>arrow := (arrow spc)(ind := SR)\<rparr>)) = abs_reg_state spc"
  "\<lbrakk> (ind, Secondary) \<in> reliable spc; arrow spc ind = SEq \<rbrakk> \<Longrightarrow>
    (abs_reg_state (spc\<lparr>arrow := (arrow spc)(ind := SL)\<rparr>)) = abs_reg_state spc"
  by(case_tac inst, auto simp add: abs_reg_state_def in_view_def comb_one_True intro!: ext)

lemma in_view_point_arrow:
  "
  (ind', inst') \<in> in_view (spc\<lparr>arrow := (arrow spc)(ind := point_arrow_to inst)\<rparr>)
  = ((ind' \<noteq> ind \<and> (ind', inst') \<in> in_view spc)
    \<or> (ind' = ind \<and> (ind, inst') \<in> reliable spc \<and> inst' = other_i inst))
  "
  by(case_tac inst, case_tac[!] inst', auto simp add: in_view_def)

lemma new_in_view_ref_simps:
  "((ind, Primary) \<in> in_view (spc\<lparr>arrow := (arrow spc)(ind := SL)\<rparr>)) = False"
  "((ind, Secondary) \<in>  in_view (spc\<lparr>arrow := (arrow spc)(ind := SR)\<rparr>)) = False"
  "ind' \<noteq> ind \<Longrightarrow>
    ((ind', inst) \<in> in_view (spc\<lparr>arrow := (arrow spc)(ind := dir)\<rparr>))
    = ((ind', inst) \<in> in_view spc)
  "
  by(auto simp add: in_view_def)

lemma min_ACNormal_ACDegraded:
  "min ACNormal ACDegraded = ACDegraded"
  "min ACDegraded ACNormal  = ACDegraded"
  by(simp_all add: min_def split: split_if_asm)

lemma write_new_ref_NE:
  "\<lbrakk> ind \<in> dom upd_f; \<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind) \<rbrakk> \<Longrightarrow>
  sim_normE_ext (precommit_rel \<inter> UNIV \<times>  notin_updated ind)
    (Abs5_UpdateDefs.write_new' upd_f ind)
    (Abs4_UpdateDefs.write_new' upd_f ind)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc v e svc' spc')
  apply(auto simp add: Abs5_UpdateDefs.write_new'_def Abs4_UpdateDefs.write_new'_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: write_new_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    write_new_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_new_defs brute_force min_def split: split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def all_reliable_inv_def
    not_abs_updated_are_SEq_def sv_abs_mode_equals_def sv_abs_mode_neq_def
    in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj)
  using [[simp_depth_limit=10]]
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj fun_upd_triv'
    min_ACNormal_ACDegraded
  )
  apply(clarsimp_all simp add: abs_vol_mem.truncate_def)
  apply(subgoal_tac[!] "arrow spc ind = SEq")
  apply(clarsimp_all simp add: notin_updated_def)
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj
    new_abs_reg_state_ref_simps
    override_on_ref_simps ex_override_simp)
  apply(safe)
  apply(simp_all (asm_lr) add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj
    override_on_ref_simps ex_override_simp insert_UNIV fun_upd_triv' abs4_state_ref_simps
    new_abs_reg_state_ref_simps in_view_eqI in_view_point_arrow
  )
  (* TODO: setting a higher depth causes looping with terms of the @{term "False \<longrightarrow> P"} form *)
  using [[simp_depth_limit=2]]
  apply(simp_all add: new_in_view_ref_simps)
  using [[simp_depth_limit=10]]
  apply(clarsimp_all simp add: new_in_view_ref_simps)
  apply (metis UNIV_I in_view_ref_simps(6))+
  by (metis (full_types) domI leD the.simps)

lemma override_with_min_le:
  "override_on f (min c \<circ> f) A x \<le> (f :: 'a \<Rightarrow> 'b :: linorder) x"
  by(auto simp add: override_on_def min_def)

lemma write_new_ref:
  "\<lbrakk> ind \<in> dom upd_f; \<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind) \<rbrakk> \<Longrightarrow>
  sim_ext (precommit_rel \<inter> UNIV \<times>  notin_updated ind)
    (Abs5_UpdateDefs.write_new' upd_f ind)
    (Abs4_UpdateDefs.write_new' upd_f ind)
    (force_same_and precommit_rel)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(auto simp only: sim_ext_decompose intro!: write_new_ref_NN write_new_ref_NE)
  apply(auto simp add: simS_wpc_def force_same_and_def)
  apply(rename_tac v sva spa svc spc svc' spc')
  apply(auto simp add: Abs5_UpdateDefs.write_new'_def Abs4_UpdateDefs.write_new'_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(clarsimp_all simp add: write_new_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    write_new_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_new_defs brute_force min_def split: split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    log_state_simps
    ex_conj_distribs all_conj
  )
  using [[simp_depth_limit=0]]
  apply(clarsimp_all simp add: precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def all_reliable_inv_def
    not_abs_updated_are_SEq_def sv_abs_mode_equals_def
    in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj)
  using [[simp_depth_limit=10]]
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj fun_upd_triv'
    override_with_min_le
    min_ACNormal_ACDegraded
  )
  apply(clarsimp_all simp add: abs_vol_mem.truncate_def)
  apply(subgoal_tac[!] "arrow spc ind = SEq")
  apply(clarsimp_all simp add: notin_updated_def)
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj
    new_abs_reg_state_ref_simps insert_UNIV abs4_state_ref_simps fun_upd_triv'
    override_on_ref_simps ex_override_simp)
  using [[simp_depth_limit=2]]
  apply(simp_all only: new_in_view_ref_simps)
  apply(simp_all)
  using [[simp_depth_limit=10]]
  apply(simp_all add: new_in_view_ref_simps override_with_min_le in_view_point_arrow)
  apply(safe)
  apply(clarsimp_all simp add: in_view_ref_simps abs_reg_state_ref_simps ball_conj all_conj
    override_on_ref_simps ex_override_simp insert_UNIV fun_upd_triv' abs4_state_ref_simps
    new_abs_reg_state_ref_simps in_view_eqI in_view_point_arrow
    override_triv_iff ex_vol_mem_assignment
  )
  apply (metis UNIV_I in_view_ref_simps(6))+
  done

lemma write_new'_notin_updated_inv_NN:
  "x \<noteq> x' \<Longrightarrow>
  \<lbrace>notin_updated x\<rbrace> Abs4_UpdateDefs.write_new' upd_f x' \<lbrace>>N K (notin_updated x)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def notin_updated_def K_def)
  apply(simp add: Abs4_UpdateDefs.write_new'_def write_new_def first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_new_defs brute_force)
  done

lemma mapM'_write_new'_ref_NN:
  assumes
  sizes: "\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind)"
  and l_origin: "l \<in> order_set (dom upd_f)"
  shows "sim_ext (precommit_rel \<inter> UNIV \<times> abs_updated_equals {})
        (mapM' (Abs5_UpdateDefs.write_new' upd_f) l)
        (mapM' (Abs4_UpdateDefs.write_new' upd_f) l)
        (force_same_and precommit_rel)
        (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
        abs5_abs4_strel
  "
proof-
  have
    set: "set l = dom upd_f"
    and distinct: "distinct l"
    using l_origin
    by(auto simp add: order_set_set intro: order_set_distinct)
  hence
    "
    sim_ext (precommit_rel \<inter> UNIV \<times> (\<Inter>x \<in> set l. notin_updated x))
        (mapM' (Abs5_UpdateDefs.write_new' upd_f) l)
        (mapM' (Abs4_UpdateDefs.write_new' upd_f) l)
        (force_same_and precommit_rel)
        (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
        abs5_abs4_strel
    "
    using sizes
    apply -
    apply(rule sim_ext_mapM'_distinct)
    apply(auto intro!: distinct write_new_ref true_postNN write_new'_notin_updated_inv_NN
       simp add: set)
    done
  thus ?thesis
    apply -
    apply(erule sim_ext_mono)
    apply(clarsimp simp add: abs_updated_equals_def notin_updated_def)
    done
qed

lemma set_updated_pre_commit_ref:
  "
  sim_ext (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv))
     (set_abs_updated {}) (set_abs_updated {})
    (force_same_and (precommit_rel \<inter> UNIV \<times> abs_updated_equals {}))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(auto simp add: brute_force abs5_abs4_strel_def update_mem_rel_def sv_abs_mode_equals_def
    precommit_rel_def not_abs_updated_are_SEq_def all_SEq_def
    abs_updated_equals_def abs_vol_mem.truncate_def
    all_reliable_inv_def)
  done

lemma sim_ext_add_conc_skip:
  "(sim_ext R ma mc Sn Se Sr) = (sim_ext R ma (return x \<guillemotright> mc) Sn Se Sr)"
  by(simp add: bind_left_unit)

lemma clear_upd_mem_ref:
  "
  sim_ext (precommit_rel \<inter> UNIV \<times> abs_updated_equals {})
    clear_upd_mem
    (return ())
    (force_same_and (precommit_rel \<inter> UNIV \<times> abs_updated_equals {}))
    Se
    Sr
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(auto simp add: clear_upd_mem_def modifyRE_def
    precommit_rel_def abs_vol_mem.truncate_def
    brute_force abs5_abs4_strel_def update_mem_rel_def sv_abs_mode_equals_def
    abs_updated_equals_def all_reliable_inv_def)
  done

lemma sim_ext_chooseN:
  "sim_ext R
    (chooseN A)
    (chooseN A)
    (\<lambda>va vc. force_same_and R va vc \<inter> (if va \<in> A then UNIV else {})) Se Sr"
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def chooseN_def brute_force)

lemma maybe_mess_up_admin_inclusion:
  "maybe_mess_up_admin_silence_deg s
    \<subseteq>
    (
    (<try> Abs5_CommonDefs.read ind \<guillemotright>= (\<lambda>_. maybe_mess_up_admin_silence_deg)
        <catch> (\<lambda>e. if \<exists>v. e = Degraded v then return () else throw e))
    [+]
    <try> chooseN UNIV \<guillemotright>= (\<lambda>_. maybe_mess_up_admin_silence_deg)
        <catch> (\<lambda>e. if \<exists>v. e = Degraded v then return () else throw e)
    )
    s
  "
  apply(subst maybe_mess_up_admin_silence_deg_def)
  apply(clarsimp_all simp add: sef_chooseN intro!: choice_subsetI2)
  apply(erule rev_subsetD)
  apply(intro proj_subset)
  apply(clarsimp_all)
  apply(simp_all add: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all)
  apply(clarsimp_all simp add: maybe_mess_up_admin_silence_deg_def maybe_mess_up_admin_defs
    restart_mangle_sp_def read_defs brute_force min_def split: split_if_asm)
  done

lemma sync_by_arrow_silence_deg_ref[ref_thm]:
  "sim abs5_abs4_strel
    (Abs5_UpdateDefs.touch_one ind)
    (Abs4_UpdateDefs.sync_by_arrow_silence_deg ind)
  "
  apply(simp add: sim_def Abs4_UpdateDefs.sync_by_arrow_silence_deg_def touch_one_def)
  apply(subst add_chooseN[where A="UNIV :: sync_direction set"])
  apply(simp)
  apply(subst bind_assoc[THEN sym])
  apply(rule sim_ext_bindI[OF get_arrow_ref'])
  apply(subst force_same_and_def)
  apply(simp_all only: first_stage)
  apply(auto intro!: sim_exts_empty_rel
    just_sync_silence_deg'_ref[THEN sim_ext_mono, THEN sim_ext_mono_abs_comp] intro: choice_subsetI1
  )
  apply(clarsimp simp add: arrow_is_def is_sync_tgt_def)
  apply(clarsimp simp add: arrow_is_def is_sync_tgt_def)
  apply(rule sim_ext_mono_abs_comp[OF maybe_mess_up_admin_ref[simplified sim_def, THEN sim_ext_mono]])
  apply(clarsimp)
  apply(intro allI maybe_mess_up_admin_inclusion)
  apply(case_tac vc, simp_all)
  done

lemma rollback_ref:
  "sim abs5_abs4_strel
    Abs5_UpdateDefs.rollback
    Abs4_UpdateDefs.rollback
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma rollback_ref':
  "sim_ext abs5_abs4_strel
    Abs5_UpdateDefs.rollback
    Abs4_UpdateDefs.rollback
    Sn
    (force_same_and abs5_abs4_strel)
    abs5_abs4_strel
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})
 
lemma pre_commit_ref':
  "\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind) \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv))
    (Abs5_UpdateDefs.pre_commit upd_f) 
    (Abs4_UpdateDefs.pre_commit upd_f)
    (force_same_and precommit_rel)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(simp_all only: Abs5_UpdateDefs.pre_commit_def Abs4_UpdateDefs.pre_commit_def)
  apply(rule sim_ext_bindI[OF set_updated_pre_commit_ref])
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(subst sim_ext_add_conc_skip)
  apply(rule sim_ext_bindI[OF clear_upd_mem_ref])
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF sim_ext_chooseN])
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_try_catchI)
  apply(rule mapM'_write_new'_ref_NN)
  apply(auto split: exceptions.split simp add: force_same_and_diff' force_same_and_same
    intro: sim_exts_empty_rel sim_ext_throw sim_ext_return)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule rollback_ref'[THEN sim_ext_mono
    , THEN sim_ext_add_NE_inv[OF true_postNE rollback_no_mode_upgrade_NE]
    , THEN sim_ext_mono_exc
    ]
  )
  apply(auto simp add: force_same_and_def K_def)
  done
 
definition update_complete where
  "update_complete \<equiv> {(sva, spa). dom (upd_mem sva) = abs_updated sva }"

definition abs_updated_equals where
  "abs_updated_equals inds \<equiv> {(sv, sp). abs_updated sv = inds}"

definition upd_mem_equals where
  "upd_mem_equals f \<equiv> {(sv, sp). upd_mem sv = f}"

lemma rollback_post_NN:
  "\<lbrace>P\<rbrace> rollback \<lbrace>>N Q\<rbrace>"
  apply(auto simp add: HoareNN_return' K_beta
    rollback_def
    intro!:
    HoareNN_throw
    bind_ruleNN
    true_postNN
  )
  done

lemma pre_commit_upd_mem_post_NN:
  "\<lbrace>P\<rbrace> pre_commit upd_if \<lbrace>>N K update_complete\<rbrace>"
  apply(simp only: pre_commit_def)
  apply(rule bind_ruleNN[where Q="K (abs_updated_equals {})"])
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def brute_force abs_updated_equals_def)
  apply(clarsimp simp add: K_beta)
  apply(rule bind_ruleNN[where Q="K (abs_updated_equals {} \<inter> upd_mem_equals Map.empty)"])
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def clear_upd_mem_def modifyRE_def
    brute_force upd_mem_equals_def abs_updated_equals_def)
  apply(clarsimp simp add: K_beta)
  apply(rule conseqNN_rule[where Q=update_complete and R="K update_complete"])
  apply(clarsimp simp add: abs_updated_equals_def upd_mem_equals_def update_complete_def)
  apply(rule bind_ruleNN[OF chooseN_inv_NN])
  apply(clarsimp simp add: K_beta)
  apply(rule try_catch_ruleNN)
  apply(rule mapM'_invNN)
  apply(clarsimp simp add: HoareNN_alt Hoare_def Box_def modifyRE_def
    write_new'_def write_upd_mem_defs
    brute_force update_complete_def)
  apply(rule true_postNE)
  apply(auto split: exceptions.split intro: HoareNN_throw rollback_post_NN)
  done

lemma pre_commit_ref:
  "\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind) \<Longrightarrow>
  sim_ext (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv))
    (Abs5_UpdateDefs.pre_commit upd_f) 
    (Abs4_UpdateDefs.pre_commit upd_f)
    (force_same_and (precommit_rel \<inter> update_complete \<times> UNIV))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(rule pre_commit_ref'[THEN sim_ext_mono
    , THEN sim_ext_add_NN_inv[OF pre_commit_upd_mem_post_NN true_postNN]
    , THEN sim_ext_mono_norm
    ]
  )
  apply(auto simp add: K_beta force_same_and_def)
  done

lemma in_view_switch_arrow:
  "\<lbrakk> arr = arrow spc; \<forall>ind \<in> A. arr ind \<noteq> SBot; rel = reliable spc;
    \<forall>ind \<in> A. is_arrow_tgt (arr ind) Primary \<longrightarrow> (ind, Primary) \<in> rel;
    \<forall>ind \<in> A. is_arrow_tgt (arr ind) Secondary \<longrightarrow> (ind, Secondary) \<in> rel
  \<rbrakk> \<Longrightarrow>
    in_view (spc\<lparr>arrow := override_on arr (switch_arrow arr) A \<rparr>)
    = in_view spc - {(ind, inst). ind \<in> A \<and> is_arrow_tgt (arr ind) (other_i inst)}
        \<union> {(ind, inst). ind \<in> A \<and> is_arrow_tgt (arr ind) inst} 
  "
  apply(clarsimp simp add: in_view_def override_on_def, safe)
  apply(clarsimp_all split: split_if_asm sync_direction.split_asm sync_direction.split)
  apply(safe)
  apply(case_tac[!] "arrow spc a", simp_all)
  apply(case_tac[!] "b", simp_all)
  done

lemma in_view_switch_arrow':
  "\<lbrakk> arr = arrow spc; \<forall>ind \<in> A. arr ind \<noteq> SBot; rel = reliable spc;
    rel = UNIV
  \<rbrakk> \<Longrightarrow>
    in_view (spc\<lparr>arrow := override_on arr (switch_arrow arr) A \<rparr>)
    = in_view spc - {(ind, inst). ind \<in> A \<and> is_arrow_tgt (arr ind) (other_i inst)}
        \<union> {(ind, inst). ind \<in> A \<and> is_arrow_tgt (arr ind) inst} 
  "
  by(rule in_view_switch_arrow) (auto)

lemma abs_reg_state_switch_arrow:
  "\<lbrakk> arr = arrow spc; rel = reliable spc;
    \<forall>ind \<in> A. arr ind \<noteq> SBot;
    \<forall>ind \<in> A. is_arrow_tgt (arr ind) Primary \<longrightarrow> (ind, Primary) \<in> rel;
    \<forall>ind \<in> A. is_arrow_tgt (arr ind) Secondary \<longrightarrow> (ind, Secondary) \<in> rel
  \<rbrakk>\<Longrightarrow>
   abs_reg_state (spc\<lparr>arrow := override_on arr (switch_arrow arr) A\<rparr>)
   = override_on (abs_reg_state spc)
    (\<lambda>ind. if (ind, Primary) \<notin> writable spc \<or> (ind, Secondary) \<notin> writable spc
        then LDegraded
        else LOK
    )
    (A - {ind. ind \<in> A \<and> arr ind = SEq})
  "
  apply(clarsimp simp add: abs_reg_state_def override_on_def in_view_def intro!: ext, safe)
  apply(clarsimp_all split: split_if_asm sync_direction.split_asm sync_direction.split)
  done
 
lemma abs_reg_state_switch_arrow':
  "\<lbrakk> arr = arrow spc; rel = reliable spc;
    \<forall>ind \<in> A. arr ind \<noteq> SBot;
    rel = UNIV
  \<rbrakk>\<Longrightarrow>
   abs_reg_state (spc\<lparr>arrow := override_on arr (switch_arrow arr) A\<rparr>)
   = override_on (abs_reg_state spc)
    (\<lambda>ind. if (ind, Primary) \<notin> writable spc \<or> (ind, Secondary) \<notin> writable spc
        then LDegraded
        else LOK
    )
    (A - {ind. ind \<in> A \<and> arr ind = SEq})
  "
  by(rule abs_reg_state_switch_arrow, auto)

lemma switch_arrows_ref_NN:
  "sim_normN_ext
    (precommit_rel \<inter> update_complete \<times> abs_updated_equals inds)
    Abs5_UpdateDefs.copy_upd_mem
    (Abs4_UpdateDefs.switch_arrows inds)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv))
  "
  apply(auto simp add: sim_normN_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: switch_arrows_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs_updated_equals_def switch_arrows_defs brute_force all_conj)
  apply(clarsimp_all simp add: copy_upd_mem_defs brute_force)
  apply(clarsimp_all simp add: abs5_abs4_strel_def update_mem_rel_def update_complete_def
    in_view_ref_simps abs_reg_state_ref_simps override_on_ref_simps
    in_view_switch_arrow abs_reg_state_switch_arrow
    all_conj ball_conj
  )
  apply(safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(case_tac svc, clarsimp_all)
  apply(clarsimp_all simp add:
    precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def
    sv_abs_mode_equals_def all_reliable_inv_def
    in_view_ref_simps abs_reg_state_ref_simps override_on_ref_simps
    in_view_switch_arrow abs_reg_state_switch_arrow
    all_conj abs_vol_mem.truncate_def
    intro!: ext
  )
  apply(clarsimp_all simp add: fun_map_override_on_def override_on_def is_arrow_tgt_2nd_simp)
  apply (metis (full_types) sync_direction.exhaust)
  apply (metis (full_types) UNIV_I in_view_ref_simps(6) sync_direction.exhaust)
  apply(clarsimp_all simp add: in_view_def abs_reg_state_def)
  apply(safe)
  apply(clarsimp_all)
  apply (smt comb_simps(6) is_arrow_tgt.simps(1)
    is_arrow_tgt_2nd_simp(1) other_i.simps(1) other_i.simps(2)
    sync_direction.exhaust tgt_other_i_in_srcsD
    INST.exhaust comb_one_True(1) comb_rel_writ.simps(1) is_arrow_tgt.simps(1)
    is_arrow_tgt.simps(2) 
  )+
  done

lemma switch_arrows_ref_NE:
  "sim_normE_ext
    (precommit_rel \<inter> update_complete \<times> abs_updated_equals inds)
    Abs5_UpdateDefs.copy_upd_mem
    (Abs4_UpdateDefs.switch_arrows inds)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
  "
  apply(auto simp add: sim_normE_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc vc svc' spc')
  apply(simp_all only: switch_arrows_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs_updated_equals_def switch_arrows_defs brute_force all_conj)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: copy_upd_mem_defs brute_force)
  apply(clarsimp_all simp add:
    precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def update_complete_def
    sv_abs_mode_equals_def sv_abs_mode_neq_def all_reliable_inv_def abs_updated_equals_def
    in_view_ref_simps abs_reg_state_ref_simps override_on_ref_simps
    in_view_switch_arrow in_view_switch_arrow' abs_reg_state_switch_arrow
    all_conj ball_conj
  )
  apply(safe)
  apply(case_tac[!] spa, simp_all)
  apply(simp_all add: double_override_simp abs_vol_mem.truncate_def)
  apply(clarsimp_all simp add: min_ACNormal_ACDegraded fun_map_override_on_simp aux1million
    in_view_switch_arrow' abs_reg_state_switch_arrow')
  apply(metis)
  apply(metis)
  apply(clarsimp_all simp add: fun_map_override_on_def override_on_def is_arrow_tgt_2nd_simp min_def
    aux1million)[1]
  apply (metis (full_types) arrow_srcs.cases)
  apply(case_tac "arrow spc ind", simp_all add: comb_simps)
  apply(clarsimp_all simp add: fun_map_override_on_def override_on_def is_arrow_tgt_2nd_simp min_def
    in_view_ref_simps
    aux1million)[4]
  apply (metis UNIV_I in_view_eqI(1))
  apply (metis UNIV_I in_view_eqI(2) is_arrow_tgt.simps(1) other_i.simps(2) tgt_other_i_in_srcsD)
  (* \<exists>deg *)
  apply(rule_tac x=
    "(abs_updated sva -
    {ind \<in> abs_updated sva. arrow spc ind = SEq
        \<or> (ind, Primary) \<in> writable spc \<and> (ind, Secondary) \<in> writable spc }
    \<union> fst ` unwr)"
    in exI)
  apply(clarsimp_all simp add: override_on_def min_def intro!: ext)
  apply(safe)
  apply (metis LDegraded_le_comb abs_reg_state_def comb_rel_writ_is_OK(2))
  apply (metis UNIV_I abs_reg_state_ref_simps(6) is_arrow_tgt.simps(1)
    is_arrow_tgt.simps(2) sync_direction.exhaust tgt_other_i_in_srcsD)
  apply (metis UNIV_I abs_reg_state_ref_simps(6) is_arrow_tgt.simps(1)
    is_arrow_tgt.simps(2) sync_direction.exhaust tgt_other_i_in_srcsD)
  apply(clarsimp simp add: min_def)
  apply (metis UNIV_I abs_reg_state_ref_simps(6) is_arrow_tgt.simps(1) is_arrow_tgt.simps(2)
    sync_direction.exhaust tgt_other_i_in_srcsD)
  apply(clarsimp simp add: min_def abs_reg_state_def in_view_def comb_rel_writ_is_OK)
  apply(case_tac "arrow spc a", simp_all)
  done

lemma switch_arrows_ref:
  "sim_ext
    (precommit_rel \<inter> update_complete \<times> abs_updated_equals inds)
    Abs5_UpdateDefs.copy_upd_mem
    (Abs4_UpdateDefs.switch_arrows inds)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(auto simp add: sim_ext_decompose intro!: switch_arrows_ref_NN switch_arrows_ref_NE del: subsetI)
  apply(auto simp add: sim_ext_def simS_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp_all only: switch_arrows_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs_updated_equals_def switch_arrows_defs brute_force all_conj)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: copy_upd_mem_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: precommit_rel_def
    abs5_abs4_strel_def update_mem_rel_def update_complete_def
    sv_abs_mode_equals_def all_reliable_inv_def override_triv_iff
    in_view_ref_simps abs_reg_state_ref_simps override_on_ref_simps
    in_view_switch_arrow abs_reg_state_switch_arrow
    is_arrow_tgt_2nd_simp
    all_conj ball_conj
  )
  apply(clarsimp_all simp add: fun_map_override_on_def override_on_def abs_reg_state_def in_view_def
    comb_rel_writ_is_OK comb_simps)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj abs_vol_mem.truncate_def ex_vol_mem_assignment
  )
  apply(rule disjI2)
  apply(safe)
  apply(case_tac "arrow spc ind", clarsimp_all)
  apply(case_tac "arrow spc ind", clarsimp_all intro!: ext)
  apply (metis UNIV_I UNIV_inst arrow_srcs.simps(3))
  apply(case_tac "arrow spc ind", clarsimp_all simp add: comb_one_True intro!: ext)
  apply (metis (hide_lams, no_types) comb_one_True(1) comb_one_True(2))+
  done

lemma get_updated_do_commit_ref:
  "
  sim_ext
    (precommit_rel \<inter> update_complete \<times> UNIV)
    (chooseN UNIV)
    get_abs_updated
    (\<lambda>va inds. (force_same_and (precommit_rel \<inter> update_complete \<times> abs_updated_equals inds)) va inds)
    Se
    Sr
  "
  by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def chooseN_def
    precommit_rel_def
    brute_force sv_abs_mode_equals_def abs_updated_equals_def all_reliable_inv_def)

lemma not_ACNormal_imp_all_reliable:
  "sv_abs_mode_neq ACNormal \<subseteq> all_reliable_inv"
  by(auto simp add: sv_abs_mode_neq_def all_reliable_inv_def)

lemma do_commit_ref':
  "sim_ext (precommit_rel \<inter> update_complete \<times> UNIV)
    Abs5_UpdateDefs.do_commit
    Abs4_UpdateDefs.do_commit
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(subst sef_chooseN[where A="UNIV :: pub_indx set set", THEN sym]) 
  apply(simp_all only: Abs5_UpdateDefs.do_commit_def Abs4_UpdateDefs.do_commit_def)
  apply(simp)
  apply(rule sim_ext_bindI[OF get_updated_do_commit_ref])
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_try_catchI[OF switch_arrows_ref])
  apply(subst force_same_and_def)
  apply(auto simp add: force_same_and_same
    intro!: sim_exts_empty_rel sim_ext_exceptions_caseI sim_ext_throw sim_ext_return_diffI
    not_ACNormal_imp_all_reliable[THEN subsetD]
  )
  apply(rule rollback_ref[simplified sim_def, THEN sim_ext_mono
      , THEN sim_ext_add_NN_inv[OF true_postNN Abs4_Invariant_Lemmas.rollback_post_NN]
      , THEN sim_ext_add_NE_inv[OF true_postNE rollback_no_mode_upgrade_NE]
      , THEN sim_ext_mono, THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
    ]
  )
  apply(auto simp add: K_beta force_same_and_def)
  done

lemma do_commit_ref:
  "sim_ext (precommit_rel \<inter> update_complete \<times> UNIV)
    Abs5_UpdateDefs.do_commit
    Abs4_UpdateDefs.do_commit
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (not_abs_updated_are_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
  apply(rule do_commit_ref'[THEN sim_ext_mono
    , THEN sim_ext_add_NN_inv[OF true_postNN do_commit_preserves_SEq_NN]
    , THEN sim_ext_mono_norm, THEN sim_ext_mono
    , where R2="precommit_rel \<inter> update_complete \<times> UNIV" and I2=UNIV
    ]
  )
  apply(auto simp add: precommit_rel_def K_def force_same_and_def)
  done

lemma rollforward_ref':
  "sim abs5_abs4_strel
    Abs5_UpdateDefs.rollforward
    Abs4_UpdateDefs.rollforward
  "
  by(tactic {* (refinement_tac false Ref_Tac.default_simp @{context}) 1 *})

lemma rollforward_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (not_abs_updated_are_SEq \<inter> all_reliable_inv))
    Abs5_UpdateDefs.rollforward
    Abs4_UpdateDefs.rollforward
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
  "
proof-
  have
    "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv)
    Abs5_UpdateDefs.rollforward
    Abs4_UpdateDefs.rollforward
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal))
    abs5_abs4_strel
    "
    apply(rule rollforward_ref'[simplified sim_def, THEN sim_ext_mono
      , THEN sim_ext_add_NN_inv[OF true_postNN rollforward_all_reliable_inv_NN]
      , THEN sim_ext_add_NE_inv[OF true_postNE rollforward_mode_post_NE]
      , THEN sim_ext_mono_norm, THEN sim_ext_mono_exc
      ]
    )
    apply(auto simp add: force_same_and_def K_beta)
    done
  note r = this[THEN sim_ext_mono
      , THEN sim_ext_add_NN_inv[OF true_postNN rollforward_makes_all_SEq_NN]
      , THEN sim_ext_mono_norm, THEN sim_ext_mono
  ]
  show ?thesis
    apply(rule r[where R2="abs5_abs4_strel \<inter> UNIV \<times> all_reliable_inv" and I2=UNIV])
    apply(auto simp add: force_same_and_def K_def)
    done
qed
    
declare vol_ops_refs[ref_thm del]
lemma vol_ops_inv_refs[ref_thm]:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    get_abs_mode
    get_abs_mode
    (\<lambda>va vc. force_same_and (abs5_abs4_strel \<inter> sv_abs_mode_equals va \<times> (all_SEq \<inter> all_reliable_inv)) va vc)
    Se
    Sr
  "
  by(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def
    sv_abs_mode_equals_def abs5_abs4_strel_def brute_force abs_vol_mem.truncate_def)

(* TODO: this is a duplicate of an Abs3_Invariant_Lemmas theorem;
    check_sizes should go into PmmDefs and this into Pmm_Lemmas *)
lemma check_sizes_post_NN:
  "\<lbrace>P\<rbrace> check_sizes upd_f 
    \<lbrace>>N K (if (\<forall>ind \<in> dom upd_f. size (the (upd_f ind)) \<le> capacity (Public ind))
        then P
        else {}
    )\<rbrace>"
proof(auto simp add: check_sizes_def HoareNE_throw' HoareNN_return' K_beta
    intro!: bind_ruleNN
    HoareNE_return
    HoareNN_throw
    mapM'_invNN
    chooseN_ruleNN ex_falsumNN)
  fix cnt ind l
  assume 
    sz: "\<not> size cnt \<le> capacity (Public ind)" 
    and in_dom: "upd_f ind = Some cnt" 
    and l_def: "l \<in> order_set (dom upd_f)"
  thm mapM'_ruleNN
  have rewr:
    "K {} = K (\<Inter>ind' \<in> set l. if size (the (upd_f ind')) \<le> capacity (Public ind') then UNIV else {})"
    apply(auto simp add: K_def intro!: ext)
    apply(erule_tac x=ind in ballE)
    apply(auto simp add: sz in_dom order_set_set[OF l_def] split: split_if_asm)
    done
  show 
    "\<lbrace>P\<rbrace> mapM' 
        (\<lambda>pind. if capacity (Public pind) < size (the (upd_f pind)) 
            then throw InvalidInput else return ()) 
        l
    \<lbrace>>N K {}\<rbrace>"
    apply(subst rewr)
    apply(rule mapM'_ruleNN)
    apply(auto simp add: HoareNN_return' HoareNE_throw' K_beta
      intro!: HoareNN_throw)
    done
qed

lemma check_sizes_ref_ext:
  "sim_ext (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv))
    (PmmDefs.check_sizes upd_f) (PmmDefs.check_sizes upd_f)
    (force_same_and (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv)))
    abs5_abs4_strel
  "
  apply(auto simp add: PmmDefs.check_sizes_def
    split del: split_if)
  apply(rule sim_ext_bindI[where Tn="\<lambda>va vc. 
    if va = vc \<and> vc \<in> order_set (dom (upd_f)) 
        then (abs5_abs4_strel \<inter> sv_abs_mode_equals ACNormal \<times> (all_SEq \<inter> all_reliable_inv))
        else {}"])
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force chooseN_def)[1]
  apply(auto intro!: sim_exts_empty_rel sim_ext_mapM' split)
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force chooseN_def order_set_set)
  done

lemma Times_Int_distrib2: "A <*> (B Int C) = (A <*> B) Int (A <*> C)"
  by(auto)

lemma update_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (Abs5_UpdateDefs.update upd_f)
    (Abs4_UpdateDefs.update upd_f)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    abs5_abs4_strel
  "
  apply(simp only: Abs5_UpdateDefs.update_def Abs4_UpdateDefs.update_def throw_abs_mode_error_def
    first_stage)
  apply(rule sim_ext_bindI[OF vol_ops_inv_refs])
  apply(intro allI)
  apply(subst force_same_and_def)
  apply(auto simp add: first_stage force_same_and_same
    intro!: sim_exts_empty_rel sim_ext_throw_diffI)
  apply(case_tac vc, simp_all)
  apply(rule sim_ext_bindI)
  thm sim_ext_add_NN_inv[OF true_postNN check_sizes_post_NN]
  apply(rule check_sizes_ref_ext[THEN
    sim_ext_add_NN_inv[OF check_sizes_post_NN true_postNN ]
    , THEN sim_ext_mono_exc
    ])
  apply(clarsimp simp add: force_same_and_def)
  apply(auto simp add: force_same_and_same K_beta intro: sim_exts_empty_rel)
  thm pre_commit_ref
  apply(rule sim_ext_mono_exc[where 
    Se="force_same_and (abs5_abs4_strel \<inter> UNIV \<times> sv_abs_mode_neq ACNormal)"])
  apply(simp add: Int_assoc Times_Int_distrib2[THEN sym])
  apply(rule sim_ext_bindI[OF pre_commit_ref])
  apply(simp)
  apply(subst force_same_and_def)
  apply(auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF do_commit_ref])
  apply(subst force_same_and_def, auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF rollforward_ref])
  apply(subst force_same_and_def, auto intro: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF vol_ops_inv_refs])
  apply(subst force_same_and_def, auto intro: sim_exts_empty_rel)
  apply(auto intro!: sim_ext_return_diffI sim_ext_throw_diffI
    simp add: force_same_and_same)
  apply(auto simp add: abs5_abs4_strel_def sv_abs_mode_neq_def sv_abs_mode_equals_def 
    abs_vol_mem.truncate_def force_same_and_def all_SEq_def all_reliable_inv_def split: split_if_asm)
  done
 
lemma fetch_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (Abs5_FetchDefs.fetch ind)
    (Abs4_FetchDefs.fetch ind)
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    abs5_abs4_strel
  "
  apply(rule fetch_ref'[simplified sim_def, THEN sim_ext_mono
    , THEN sim_ext_add_NN_inv[OF true_postNN
        HoareNN_join[OF fetch_reliable_inv fetch_preserves_all_SEq_NN]]
    , THEN sim_ext_add_NE_inv[OF true_postNE fetch_mode_post_NE]
    , THEN sim_ext_mono, THEN sim_ext_mono_exc, THEN sim_ext_mono_norm
    ]
  )
  apply(auto simp add: K_def force_same_and_def)
  apply(auto simp add: sv_abs_mode_neq_def all_SEq_def all_reliable_inv_def)
  done

lemma startup_ref:
  "sim_ext abs5_abs4_strel
    Abs5_StartupDefs.startup
    Abs4_StartupDefs.startup
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    abs5_abs4_strel
  "
  apply(rule startup_ref'[simplified sim_def, THEN sim_ext_mono
    , THEN sim_ext_add_NE_inv[OF true_postNE
        HoareNE_join[OF startup_post_NE startup_all_SEq_post_NE]]
    , THEN sim_ext_add_NN_inv[OF true_postNN startup_post_NN]
    , THEN sim_ext_mono_norm, THEN sim_ext_mono_exc, THEN sim_ext_mono
    , where J'3="K (all_SEq \<inter> all_reliable_inv)"
    ]
  )
  apply(auto simp add: K_def force_same_and_def)
  done

lemma fetch_full_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (<tryR>
        Abs5_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (abs5_abs4_strel)
  "
  by(intro sim_ext_try_catchRI[OF fetch_ref] sim_ext_rec_tryRI startup_ref)

lemma update_full_ref:
  "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (<tryR>
        Abs5_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (abs5_abs4_strel)
  "
  by(intro sim_ext_try_catchRI[OF update_ref] sim_ext_rec_tryRI startup_ref)

lemma startup_full_ref:
  "sim_ext abs5_abs4_strel
    (<tryR>
        Abs5_StartupDefs.startup
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_StartupDefs.startup
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
    (abs5_abs4_strel)
  "
  by(intro sim_ext_try_catchRI[OF startup_ref] sim_ext_rec_tryRI startup_ref)

lemma abs5_startup_full_ref_N:
  "sim_norm (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (<tryR>
        Abs5_StartupDefs.startup
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_StartupDefs.startup
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
  "
  apply(unfold sim_norm_def)
  apply(rule sim_ext_imp_sim_norm_ext)
  apply(rule sim_ext_mono)
  apply(rule Abs5_Ref.startup_full_ref)
  apply(simp)
  done

lemma abs5_update_full_ref_N:
  "sim_norm (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (<tryR>
        Abs5_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
  "
  by(auto simp add: sim_norm_def intro!: sim_ext_imp_sim_norm_ext Abs5_Ref.update_full_ref)

lemma abs5_fetch_full_ref_N:
  "sim_norm (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
    (<tryR>
        Abs5_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
    (<tryR>
        Abs4_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs4_StartupDefs.startup)
    )
  "
  by(auto simp add: sim_norm_def intro!: sim_ext_imp_sim_norm_ext Abs5_Ref.fetch_full_ref)

end

