theory Abs5_SpecRef
  imports "../abs4/Abs4_Specification" Abs5_Specification Abs5_Ref "../SpecRef_Lemmas"
begin

fun abs4_abs5_pi
  :: "(abs_vol_mem \<times> 'c abs4_state) \<Rightarrow> (abs_card_mode \<times> 'c obs_abs_mem)" where
  "abs4_abs5_pi (svc, spc) = (abs_card_mode svc,
        \<lparr>
            obs_abs_mem.memory = if abs4_state.admin_state spc = LDead \<or> 
                    (\<exists>ind. (ind, Primary) \<notin> in_view spc \<and> (ind, Secondary) \<notin> in_view spc)
                then None
                else Some (\<lambda>ind. 
                    if (ind, Primary) \<in> in_view spc
                    then abs4_state.memory spc ind Primary
                    else abs4_state.memory spc ind Secondary
                )
            , obs_abs_mem.reg_state = abs_reg_state spc
            , obs_abs_mem.admin_state = abs4_state.admin_state spc
        \<rparr>
    )"
 
lemma abs_reg_state_LDead_simp:
  "(abs_reg_state spc ind = LDead) = ((ind, Primary) \<notin> in_view spc \<and> (ind, Secondary) \<notin> in_view spc)"
  by(auto simp add: abs_reg_state_def in_view_def comb_is_LDead)

lemma abs_reg_state_LDeadI:
  "\<lbrakk> (ind, Primary) \<notin> in_view spc; (ind, Secondary) \<notin> in_view spc \<rbrakk> \<Longrightarrow>
    abs_reg_state spc ind = LDead"
  by(auto simp add: abs_reg_state_def)

lemma ex_abs_mem:
  "(\<exists>s. P s) = 
  (\<exists> mem rs as. P \<lparr> abs_mem.memory = mem, abs_mem.reg_state = rs, abs_mem.admin_state = as \<rparr>)"
  apply(auto)
  by (metis abs_mem.cases)

theorem abs4_refines_abs5:
  "refines (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)) abs4_abs5_pi abs5_spec abs4_spec"
proof(intro refinesI)
  show "obs_consistent (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
        abs4_abs5_pi abs5_spec abs4_spec"
    apply(auto simp add: obs_consistent_def abs5_spec_def abs4_spec_def 
      abs5_abs4_strel_def obs_ts min_def all_SEq_def all_reliable_inv_def 
      abs_reg_state_LDead_simp abs_vol_mem.truncate_def intro!: ext)
    apply(case_tac[!] b, force)+
    done
next
  show "PO_refines (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)) abs5_spec abs4_spec"
  proof(rule refine_basic)
    show "init_trans abs4_spec `` init abs4_spec
        \<subseteq> (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)) `` 
            init_trans abs5_spec `` init abs5_spec
      "
      apply(simp add: abs4_spec_def abs4_ts_def abs5_spec_def abs5_ts_def TS.defs spec.defs)
      apply(rule sim_norm_init_inclusion)
      apply(rule Abs5_Ref.startup_ref[THEN sim_ext_imp_sim_norm_ext])
      apply(auto simp add: abs5_abs4_strel_def abs4_init_def all_conj 
        abs_vol_mem.truncate_def ex_vol_mem_assignment ex_abs_mem)
      apply(rule_tac x="(\<lambda>ind. 
                if (ind, Primary) \<in> in_view b 
                then abs4_state.memory b ind Primary
                else abs4_state.memory b ind Secondary
            )"
        in exI)
      apply(clarsimp simp add: abs_mem.defs in_view_def)
      apply(case_tac "arrow b ind", simp_all)
      done
  next
    have restart_ref:
      "sim_ext (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv))
      restart_mangle_sv restart_mangle_sv
      (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
      (force_same_and (abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)))
      abs5_abs4_strel"
      by(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force
        abs5_abs4_strel_def abs_vol_mem.truncate_def ex_vol_mem_assignment)
    show "{abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)} 
            TS.trans abs5_spec, TS.trans abs4_spec 
        {> abs5_abs4_strel \<inter> UNIV \<times> (all_SEq \<inter> all_reliable_inv)}"
      apply(simp add: abs5_spec_def abs5_ts_def abs4_spec_def abs4_ts_def spec.defs TS.defs 
        abs5_trans_def abs4_trans_def)
      apply(intro relhoare_UN_lockstep allI relhoare_set_fun_choice_lockstep)
      apply(auto intro!: sim_norm_ext_imp_PO_rhoare 
        Abs5_Ref.fetch_full_ref[THEN sim_ext_imp_sim_norm_ext]
        Abs5_Ref.update_full_ref[THEN sim_ext_imp_sim_norm_ext]
        sim_norm_ext_try_catchRI[OF restart_ref]
        Abs5_Ref.startup_ref[THEN sim_ext_rec_tryRI, THEN sim_ext_imp_sim_norm_ext]
        simp add: try_rec_tryR_def)
      done
  qed
qed

theorem abs4_implements_abs5:
  "implements abs4_abs5_pi abs5_spec abs4_spec"
  by(auto intro!: refinement_soundness abs4_refines_abs5)

end