theory Abs5_UpdateDefs
imports Abs5_CommonDefs
begin

definition write_new'
  :: "(pub_indx \<rightharpoonup> 'c) \<Rightarrow> pub_indx \<Rightarrow> (unit, 'c)abs5_monad" where
  "write_new' upd_f pind \<equiv> write_upd_mem pind ((the o upd_f) pind)"

definition touch_one
  :: "pub_indx \<Rightarrow> (unit, 'c)abs5_monad" where
  "touch_one ind \<equiv> do {
    maybe_mess_up_admin_silence_deg;
    <try> (read ind [+] chooseN UNIV)\<guillemotright>
        Abs5_CommonDefs.maybe_mess_up_admin_silence_deg
    <catch> (\<lambda>e. if \<exists>v. e = Degraded v then return () else throw e) 
  }"

definition touch_updated
  :: "(unit, 'c)abs5_monad" where
  touch_updated_def[monad_def]:
  "touch_updated \<equiv> do {
    abs_updated_s \<leftarrow> get_abs_updated;
    abs_updated \<leftarrow> chooseN (order_set abs_updated_s);
    mapM' touch_one abs_updated;
    set_abs_updated {}
  }"

definition rollback 
  :: "(unit, 'c)abs5_monad" where
  rollback_def[monad_def]:
  "rollback \<equiv> do {
    touch_updated;
    throw (Degraded (EB False))
  }"

definition clear_upd_mem :: "(unit, 'c)abs5_monad" where
  "clear_upd_mem \<equiv> modifyRE (\<lambda>(sv, sp). (sv\<lparr> upd_mem := Map.empty \<rparr>, sp))"

definition pre_commit
  :: "(pub_indx \<rightharpoonup> 'c) \<Rightarrow> (unit, 'c)abs5_monad" where
  pre_commit_def[monad_def]:
  "pre_commit upd_f \<equiv> do {
    set_abs_updated {};
    clear_upd_mem;
    ord \<leftarrow> chooseN (order_set (dom upd_f));
    <try>
        mapM' (write_new' upd_f) ord
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> rollback
        | _ \<Rightarrow> throw e)
  }"

definition copy_upd_mem_success
  :: "(unit, 'c)abs5_monad" where
  "copy_upd_mem_success \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv, sp')|sp'.
        admin_state sp = LOK
        \<and> sp' = sp\<lparr> memory := fun_map_override_on (memory sp) (upd_mem sv) (abs_updated sv) \<rparr>
    }"

definition copy_upd_mem_degrade
  :: "(unit, 'c)abs5_monad" where
  "copy_upd_mem_degrade \<equiv> \<lambda>(sv, sp).
    {(Some (Inl (Degraded (EB False))), sv', sp')|sp' sv' deg.
        LDegraded \<le> admin_state sp 
        \<and> sv' = sv\<lparr> abs_card_mode := ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
            admin_state := LDegraded
        \<rparr>
    }"

definition copy_upd_mem_degrade_write
  :: "(unit, 'c)abs5_monad" where
  "copy_upd_mem_degrade_write \<equiv> \<lambda>(sv, sp).
    {(Some (Inr ()), sv', sp')|sp' sv' deg.
        LDegraded \<le> admin_state sp 
        \<and> sv' = sv\<lparr> abs_card_mode := ACDegraded \<rparr>
        \<and> sp' = sp\<lparr>
            memory := fun_map_override_on (memory sp) (upd_mem sv) (abs_updated sv),
            reg_state := override_on (reg_state sp) ((min LDegraded) o (reg_state sp)) deg,
            admin_state := LDegraded
        \<rparr>
    }"

definition copy_upd_mem_fail
  :: "(unit, 'c)abs5_monad" where
  "copy_upd_mem_fail \<equiv> \<lambda>(sv, sp).
    {(Some (Inl FailMode), sv', sp')|sp' sv' reg_state'.
        sv' = sv\<lparr> abs_card_mode := ACFail \<rparr>
        \<and> (\<forall>ind. reg_state' ind \<le> reg_state sp ind)
        \<and> sp' = sp\<lparr>
            reg_state := reg_state',
            admin_state := LDead
        \<rparr>
    }"

definition copy_upd_mem :: "(unit, 'c)abs5_monad" where
  "copy_upd_mem \<equiv>
    copy_upd_mem_success
    [+] copy_upd_mem_degrade
    [+] (try_catch copy_upd_mem_degrade (K (throw (Degraded EU))))
    [+] (copy_upd_mem_degrade_write \<guillemotright> throw (Degraded (EB True)))
    [+] copy_upd_mem_fail
    [+] restart_mangle_sv
    [+] (copy_upd_mem_success \<guillemotright> restart_mangle_sv)
  "

lemmas copy_upd_mem_defs =
  copy_upd_mem_def
  copy_upd_mem_success_def
  copy_upd_mem_degrade_def
  copy_upd_mem_degrade_write_def
  copy_upd_mem_fail_def

definition do_commit :: "(unit, 'c)abs5_monad" where
  do_commit_def[monad_def]:
  "do_commit \<equiv> do {
    <try> do {
        copy_upd_mem
    } <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then return ()
            else rollback
        | _ \<Rightarrow> throw e
    )
  }
  "

definition rollforward where
  rollforward_def[monad_def]:
  "rollforward \<equiv> touch_updated"

definition update :: "(pub_indx \<rightharpoonup> 'c :: size) \<Rightarrow> (unit, 'c)abs5_monad" where
  update_def[monad_def]:
  "update upd_f \<equiv> do {
    throw_abs_mode_error (EB False);
    check_sizes upd_f;
    pre_commit upd_f;
    do_commit;
    rollforward;
    throw_abs_mode_error (EB True)
  }"

end
