theory Abs6_Lemmas
  imports 
  "../Pmm_Lemmas" "../Repeated_Restarts"
  Abs6_Specification
  uses ("../abs2/trans_simproc.ML")
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  restart_mangle_sp_def
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs

definition abs6_skip where
  "abs6_skip \<equiv>
    return ()
    [+] abs6_degrade
    [+] abs6_fail
    [+] restart_mangle_sp
  "

definition abs6_loud_skip where
  "abs6_loud_skip \<equiv>
    return ()
    [+] (abs6_degrade \<guillemotright> chooseN UNIV \<guillemotright>= (\<lambda>e. throw (Degraded e)))
    [+] abs6_fail
    [+] restart_mangle_sp
  "

lemmas abs6_skip_defs =
  abs6_skip_def
  abs6_loud_skip_def
  abs6_degrade_def
  abs6_fail_def

lemma abs6_loud_skip_idempotent':
  "(abs6_loud_skip \<guillemotright> abs6_loud_skip) (sv, sp) \<subseteq> abs6_loud_skip (sv, sp)"
proof(rule proj_subset, clarsimp_all del: subsetI)
  fix sv sp
  show
    "((abs6_loud_skip \<guillemotright>= (\<lambda>_. abs6_loud_skip))\<Down>R) (sv, sp) \<subseteq> (abs6_loud_skip\<Down>R) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all add: abs6_loud_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs6_degrade_def abs6_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    done
next
  fix sv sp
  show
    "((abs6_loud_skip \<guillemotright>= (\<lambda>_. abs6_loud_skip))\<Down>NN) (sv, sp) \<subseteq> (abs6_loud_skip\<Down>NN) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all add: abs6_loud_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs6_degrade_def abs6_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    done
next
  fix sv sp
  show "((abs6_loud_skip \<guillemotright>= (\<lambda>_. abs6_loud_skip))\<Down>NE) (sv, sp) \<subseteq> (abs6_loud_skip\<Down>NE) (sv, sp)"
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(simp_all add: abs6_loud_skip_def)
    apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      cong: exceptions.case_cong if_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(clarsimp_all simp add: abs6_degrade_def abs6_fail_def restart_mangle_sp_def brute_force
      split: exceptions.split_asm split_if_asm)
    done
qed
 
lemma abs6_loud_skip_idempotent:
  "(abs6_loud_skip \<guillemotright> abs6_loud_skip) = abs6_loud_skip"
  apply(rule ext, clarsimp, intro Set.equalityI abs6_loud_skip_idempotent')
  apply(rule_tac B="(return () \<guillemotright> abs6_loud_skip) (a, b)" in subset_trans)
  apply(simp add: bind_left_unit)
  apply(rule bind_mono1)
  apply(clarsimp simp add: abs6_loud_skip_def choice_def)
  done

lemma mapM'_abs6_loud_skip_inclusion:
  "\<forall>s. mapM' (\<lambda>_. abs6_loud_skip) l s \<subseteq>  abs6_loud_skip s"
  apply(intro allI)
  apply(induct l arbitrary: s)
  apply(clarsimp simp add: abs6_loud_skip_def choice_def)
  apply(case_tac s, clarsimp del: subsetI)
  apply(rename_tac l sv sp)
  apply(rule_tac B="(abs6_loud_skip \<guillemotright> abs6_loud_skip) (sv, sp)" in subset_trans)
  apply(rule ball_bind_mono2)
  apply(clarsimp_all simp add: abs6_loud_skip_idempotent)
  done

definition abs6_pre_commit where
  "abs6_pre_commit \<equiv>
    <try>
        abs6_loud_skip 
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> (abs6_skip \<guillemotright> throw (Degraded (EB False)))
        | _ \<Rightarrow> throw e)"

definition sv_is where
  "sv_is sv \<equiv> {(sv', sp). sv' = sv}"

definition sv_isnt where
  "sv_isnt sv \<equiv> {(sv', sp). sv' \<noteq> sv}"

lemma abs6_pre_commit_mode_post_NN:
  "\<lbrace>sv_is m\<rbrace> abs6_pre_commit \<lbrace>>N K (sv_is m)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def)
  apply(auto simp add: sv_is_def abs6_pre_commit_def abs6_skip_defs chooseN_def
    brute_force min_def split: split_if_asm exceptions.split_asm)
  done
 
definition abs6_copy_upd_mem where
  "abs6_copy_upd_mem upd_f \<equiv>
        (update_success upd_f \<guillemotright> (
            return ()
            [+] (abs6_degrade \<guillemotright> throw (Degraded (EB True)))
            [+] abs6_fail
            [+] restart_mangle_sp
        ))
        [+] (abs6_degrade \<guillemotright> (throw (Degraded (EB False)) [+] throw (Degraded EU)))
        [+] abs6_fail
        [+] restart_mangle_sp
  "

(* TODO: there are two redundant "do"'s in the corresponding Abs5_UpdateDefs definition *)
definition abs6_do_commit where
  abs6_do_commit_def[monad_def]:
  "abs6_do_commit upd_f \<equiv>
    <try>
        abs6_copy_upd_mem upd_f
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> 
            if v = EB True
            then return ()
            else (abs6_skip \<guillemotright> throw (Degraded (EB False)))
        | _ \<Rightarrow> throw e
    )"
 
lemma ACNormal_exhaust:
  "\<lbrakk> x \<noteq> ACFail; x \<noteq> ACDegraded \<rbrakk> \<Longrightarrow>
    x = ACNormal"
  by(case_tac[!] x, simp_all)
 
lemma throw_mode_error_post_NN:
  "\<lbrace>P\<rbrace> throw_mode_error val \<lbrace>>N K (sv_is ACNormal)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_def sv_is_def throw_mode_error_def
    sv_isnt_def
    brute_force)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
  apply(simp_all)
  done

definition abs6_update' where
  "abs6_update' upd_f \<equiv> do {
    throw_mode_error (EB False);
    check_sizes upd_f;
    abs6_pre_commit;
    abs6_do_commit upd_f;
    abs6_skip;
    throw_mode_error (EB True)
  }"

lemma abs6_update'_proj_simps:
  "abs6_degrade\<Down>R = empty"
  "abs6_degrade\<Down>NE = empty"
  "abs6_fail\<Down>R = empty"
  "abs6_fail\<Down>NN = empty"
  "restart_mangle_sp\<Down>NN = empty"
  "restart_mangle_sp\<Down>NE = empty"
  "(throw_mode_error val)\<Down>R = empty"
  "(update_success upd_f)\<Down>NE = empty"
  "(update_success upd_f)\<Down>R = empty"
  by(auto simp add: brute_force abs6_degrade_def abs6_fail_def throw_mode_error_def
    update_success_def
    split: split_if_asm)

declare disjCI[rule del]

use "../abs2/trans_simproc.ML"

simproc_setup trans_simproc 
  ("r \<le> s")
  = {* K Trans_Simproc.trans_simproc *}

lemma low_bound:
  "\<forall>ind. g ind \<le> override_on f (min LDegraded \<circ> f) deg ind \<Longrightarrow> \<forall>ind. g ind \<le> f ind"
  apply(auto simp add: override_on_def o_def)
  apply (smt min_max.le_infE)
  done

lemma abs6_update'_inclusion:
  "\<forall>(s :: abs_card_mode \<times> 'c :: size abs_mem).
    (abs6_update' upd_f) s \<subseteq> (update upd_f) s"
proof(intro proj_subset' allI)
  fix s
  show "((abs6_update' upd_f)\<Down>R) s \<subseteq> ((update upd_f)\<Down>R) s"
    apply(auto)
    apply(rename_tac sv sp)
    apply(simp_all add: abs6_update'_def update_def abs6_pre_commit_def abs6_loud_skip_def
      abs6_skip_def first_stage abs6_update'_proj_simps
      check_sizes_alt
      abs6_do_commit_def
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(simp_all)
    apply(simp_all add: abs6_copy_upd_mem_def 
      abs6_loud_skip_def abs6_skip_def first_stage abs6_update'_proj_simps
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(clarsimp_all simp add: brute_force throw_mode_error_def 
      check_sizes'_def
      abs6_fail_def abs6_degrade_def update_success_def
      split: split_if_asm exceptions.split_asm
    )
    apply(tactic {* distinct_subgoals_tac *})
    apply(case_tac[!] b, clarsimp_all)
    apply(clarsimp_all simp add: override_on_def low_bound)
    done

next
  {
    fix x y
    assume "min x y \<noteq> x"
    hence "min x y = y"
      by(auto simp add: min_def)
  } note min_not1=this
  {
    fix x y
    assume "min x y \<noteq> y"
    hence "min x y = x"
      by(auto simp add: min_def)
  } note min_not2=this

  fix s :: "abs_card_mode \<times> 'c abs_mem"
  show "((abs6_update' upd_f)\<Down>NN) s \<subseteq> ((update upd_f)\<Down>NN) s"
    apply(case_tac s)
    apply(auto)
    apply(rename_tac sv sp sv' sp')
    apply(simp_all add: abs6_update'_def update_def abs6_pre_commit_def abs6_loud_skip_def
      abs6_skip_def first_stage abs6_update'_proj_simps
      abs6_do_commit_def
      check_sizes_alt
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(simp_all add: abs6_copy_upd_mem_def 
      abs6_loud_skip_def abs6_skip_def first_stage abs6_update'_proj_simps
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(clarsimp_all simp add: brute_force throw_mode_error_def 
      check_sizes'_def
      abs6_fail_def abs6_degrade_def update_success_def
      split: split_if_asm exceptions.split_asm
    )
    apply(case_tac[!] sv, clarsimp_all simp add: min_def)
    done
next
  {
    fix f deg dega
    have 
      "override_on (override_on f (min LDegraded \<circ> f) deg)
               (min LDegraded \<circ> override_on f (min LDegraded \<circ> f) deg) dega
      = override_on f (min LDegraded o f) (deg \<union> dega)"
      by(auto simp add: override_on_def min_def split: split_if_asm intro!: ext)
  } note double_override_simp=this
  fix s :: "abs_card_mode \<times> 'c abs_mem"
  show "((abs6_update' upd_f)\<Down>NE) s \<subseteq> ((update upd_f)\<Down>NE) s"
    apply(case_tac s)
    apply(auto)
    apply(rename_tac sv sp e sv' sp')
    apply(simp_all add: abs6_update'_def update_def abs6_pre_commit_def abs6_loud_skip_def
      check_sizes_alt
      abs6_skip_def first_stage abs6_update'_proj_simps
      abs6_do_commit_def
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(simp_all add: abs6_copy_upd_mem_def 
      abs6_loud_skip_def abs6_skip_def first_stage abs6_update'_proj_simps
      cong: if_cong exceptions.case_cong)
    apply(simp_all only: choice_def Un_iff, safe)
    apply(tactic {* distinct_subgoals_tac *})
    apply(clarsimp_all simp add: brute_force throw_mode_error_def
      check_sizes'_def
      abs6_fail_def abs6_degrade_def update_success_def
      split: split_if_asm exceptions.split_asm
    )
    apply(iterative_simp)+
    apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
    apply(case_tac[!] sp)
    apply(clarsimp_all simp add: min_def double_override_simp dest!: low_bound)
    apply(tactic {* distinct_subgoals_tac *})
    apply(smt sup_eq_bot_iff)+
    done
qed

lemma repeat_simp': "\<forall>s. ((try_catchR h h')\<Down>N) s \<subseteq> (h'\<Down>N) s
  \<Longrightarrow> ((rec_tryR h)\<Down>N) s \<subseteq> (h'\<Down>N) s"
  by(force simp add: projectN_def intro: repeat_simp)

lemma projectN_subsetI:
  "\<lbrakk> (m\<Down>NN) s \<subseteq> (m'\<Down>NN) s; (m\<Down>NE) s \<subseteq> (m'\<Down>NE) s \<rbrakk> \<Longrightarrow> (m\<Down>N) s \<subseteq> (m'\<Down>N) s"
  apply(auto simp add: project_defs)
  apply(case_tac a, auto)
  done

lemma startup_proj_simps:
  "((startup :: ('a, 'c)abs6_monad)\<Down>R) = ((restart_mangle_sp :: ('a, 'c)abs6_monad)\<Down>R)"
  "(startup\<Down>NN) = empty"
  apply(auto simp add: startup_def first_stage intro!: ext)
  apply(auto simp add: choice_def startup_defs project_defs)
  apply(auto simp add: restart_mangle_sp_def)
  done

lemma bound_LDead_from_below:
  "\<lbrakk> \<forall>ind. g ind \<le> f ind; LDead \<notin> range g\<rbrakk>
  \<Longrightarrow> (LDead \<notin> range f = True)"
  apply(auto simp add: image_def)
  by (metis bot_log_state_def le_bot)

lemma LOK_le:
  "(LOK \<le> x) = (x = LOK)"
  by(case_tac x, auto)

lemma ex_equal_overrides:
  "\<lbrakk> LDead \<notin> range g; \<forall>ind. g ind \<le> f ind\<rbrakk> \<Longrightarrow>
    \<exists>dega. override_on g (\<lambda>y. LDegraded) deg = override_on f (\<lambda>y. LDegraded) dega"
  apply(frule bound_LDead_from_below)
  apply(assumption)
  apply(rule_tac x="deg \<union> {x. g x = LDegraded}" in exI)
  apply(auto simp add: override_on_def intro!: ext)
  by (metis (full_types) LOK_le log_state.exhaust rangeI)

lemma startup_reduction':
  "((<tryR> startup <catchR> startup)\<Down>N) s \<subseteq> (startup\<Down>N) s"
  apply(intro projectN_subsetI)
  apply(case_tac[!] s)
  apply(clarsimp_all)
  apply(rename_tac[!] sv sp v sv' sp')
  apply(simp_all add: first_stage bindS_empty startup_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: startup_def)
  apply(simp_all add: first_stage bindS_empty startup_proj_simps)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: startup_defs brute_force)
  apply(iterative_simp)+
  apply(case_tac[!] sp, clarsimp_all simp add: bound_LDead_from_below LOK_le ex_equal_overrides)
  apply(rule_tac x="deg \<union> {x. reg_state' x = LDegraded}" in exI)
  apply(clarsimp simp add: override_on_def intro!: ext)
  apply (metis (full_types) LOK_le log_state.exhaust rangeI)
  apply (metis (lifting) order_trans)+
  done


lemma startup_reduction:
  "((rec_tryR startup)\<Down>N) s \<subseteq> (startup\<Down>N) s"
  by(intro repeat_simp' allI startup_reduction')

subsubsection {* API properties *}

definition memory_eq where
  "memory_eq mem \<equiv> {(sv, sp). memory sp = mem }"

lemma transaction_property_NN:
  "\<lbrace>memory_eq mem\<rbrace>
    (<tryR> update upd_f <catchR> startup)
  \<lbrace>>N K (memory_eq (fun_map_override_on mem upd_f (dom upd_f)))\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def
    startup_defs
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

lemma transaction_property_NE:
  "\<lbrace>memory_eq mem\<rbrace>
    (<tryR> update upd_f <catchR> startup)
  \<lbrace>@N K
    (memory_eq (fun_map_override_on mem upd_f (dom upd_f))
     \<union> memory_eq mem)
  \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    startup_defs check_sizes'_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  done

lemma fetch_correctness_NN:
  "\<lbrace>memory_eq mem\<rbrace> 
    (<tryR> fetch ind <catchR> startup)
  \<lbrace>>N \<lambda>cnt. if cnt = mem ind then UNIV else {}\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

lemma fetch_correctness_NE:
  "\<lbrace>memory_eq mem\<rbrace> 
    (<tryR> fetch ind <catchR> startup)
  \<lbrace>@N \<lambda>e. 
    if e = FailMode \<or> e = StartedExc \<or> e = Degraded EU \<or> (\<exists>cnt. e = Degraded (EC cnt) \<and> cnt = mem ind)
    then UNIV
    else {}
  \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    abs6_update'_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

lemma fetch_correctness_no_restart_NE:
  "\<lbrace>memory_eq mem\<rbrace> 
    fetch ind
  \<lbrace>@N \<lambda>e. 
    if e = FailMode \<or> (\<exists>cnt. e = Degraded (EC cnt) \<and> cnt = mem ind)
    then UNIV
    else {}
  \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    abs6_update'_proj_simps
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

lemma fetch_no_mem_upd_NN:
  "\<lbrace>memory_eq mem\<rbrace> 
    (<tryR> fetch ind <catchR> startup)
  \<lbrace>>N K (memory_eq mem)\<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

lemma fetch_no_mem_upd_NE:
  "\<lbrace>memory_eq mem\<rbrace> 
    (<tryR> fetch ind <catchR> startup)
  \<lbrace>@N K (memory_eq mem)\<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  done

definition correct_mode where
  "correct_mode \<equiv> {(sv, sp). 
    (sv = ACNormal \<longrightarrow> LDead \<notin> range (reg_state sp) \<and> admin_state sp = LOK)
    \<and> (sv = ACDegraded \<longrightarrow> (
        (LDead \<notin> range (reg_state sp) \<and> admin_state sp \<noteq> LDead)
        \<and> (LDegraded \<in> range (reg_state sp) \<or> admin_state sp = LDegraded)))
    \<and> (sv = ACFail \<longrightarrow> (LDead \<in> range (reg_state sp) \<or> admin_state sp = LDead))
  }"

lemma startup_post_NN:
  "\<lbrace>P\<rbrace> startup \<lbrace>>N Q\<rbrace>"
  by(auto simp add: HoareNN_alt Hoare_def Box_def startup_defs brute_force)

lemma startup_correct_mode_NN:
  "\<lbrace>P\<rbrace> startup \<lbrace>>N K correct_mode \<rbrace>"
  by(rule startup_post_NN)

lemma LDead_override_LDegraded:
  "(LDead = override_on f (\<lambda>_. LDegraded) A x) \<Longrightarrow> (LDead \<in> range f)"
  "LDead \<notin> range f \<Longrightarrow>       
    LDead \<noteq> override_on f (min LDegraded \<circ> f) A xa"
  "LDead \<notin> range f \<Longrightarrow>       
    LDead \<notin> range (override_on f (min LDegraded \<circ> f) A)"
  by(auto simp add: override_on_def image_def min_def split: split_if_asm)

lemma startup_correct_mode_NE:
  "\<lbrace>P\<rbrace> startup \<lbrace>@N K correct_mode \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(auto simp add: throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force correct_mode_def 
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply (metis override_on_apply_in rangeI)
  done

lemma startup_correct_mode_N:
  "\<lbrace>UNIV\<rbrace> startup \<lbrace>>R K correct_mode \<rbrace>"
  by(auto simp add: decompose_HoareN intro!: startup_correct_mode_NN startup_correct_mode_NE)

lemma fetch_correct_mode_NN:
  "\<lbrace> correct_mode \<rbrace> fetch ind \<lbrace>>N K correct_mode \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(auto simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force correct_mode_def 
    split: split_if_asm exceptions.split_asm)
  done

lemma LDegraded_in_override_LDegraded:
  assumes "LDead \<notin> range f" "x \<in> deg"
  shows "LDegraded \<in> range (override_on f (min LDegraded \<circ> f) deg)"
proof-
  have "f x \<noteq> LDead" using `LDead \<notin> range f`
    apply(auto simp add: override_on_def image_def min_def)
    apply(metis)
    done
  hence "LDegraded = override_on f (min LDegraded \<circ> f) deg x" using `x \<in> deg`
    by(case_tac "f x", auto simp add: override_on_def min_def)
  thus ?thesis
    apply -
    apply(erule range_eqI)
    done
qed

lemma fetch_correct_mode_NE:
  "\<lbrace> correct_mode \<rbrace> fetch ind \<lbrace>@N K correct_mode \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' e sv sp)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force correct_mode_def 
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(auto dest: LDead_override_LDegraded)
  apply(auto dest!: LDegraded_in_override_LDegraded)
  apply(case_tac[!] sv', simp_all)
  done

lemma fetch_correct_mode_N: 
  "\<lbrace> correct_mode \<rbrace> fetch ind \<lbrace>>R K correct_mode \<rbrace>"
  by(auto simp add: decompose_HoareN intro!: fetch_correct_mode_NN fetch_correct_mode_NE)

lemma HoareN_try_catchR_reestablish:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>>R Q\<rbrace>; \<lbrace>UNIV\<rbrace> h \<lbrace>>R Q\<rbrace> \<rbrakk> \<Longrightarrow> 
    \<lbrace>P\<rbrace> <tryR> m <catchR> h \<lbrace>>R Q\<rbrace>"
  by(force simp add: Hoare_alt Hoare_def Box_def project_defs try_catchR_some_decompose)

lemma Hoare_try_catchR_inv:
  "\<lbrakk> \<lbrace>P\<rbrace> m \<lbrace>> K P\<rbrace>; \<lbrace>P\<rbrace> h \<lbrace>> K P\<rbrace> \<rbrakk> \<Longrightarrow> 
    \<lbrace>P\<rbrace> <tryR> m <catchR> h \<lbrace>> K P\<rbrace>"
  apply(auto simp add: Hoare_alt Hoare_def Box_def project_defs K_def unfold_monad)
  apply(case_tac aa, force+)
  done

lemma fetch_call_correct_mode_N: 
  "\<lbrace> correct_mode \<rbrace> <tryR> fetch ind <catchR> startup \<lbrace>>R K correct_mode \<rbrace>"
  by(intro HoareN_try_catchR_reestablish fetch_correct_mode_N startup_correct_mode_N)

lemma update_correct_mode_NN:
  "\<lbrace> correct_mode \<rbrace> update upd_f \<lbrace>>N K correct_mode \<rbrace>"
  apply(auto simp add: HoareNN_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def
    startup_defs brute_force correct_mode_def 
    split: split_if_asm exceptions.split_asm)
  done

lemma update_correct_mode_NE:
  "\<lbrace> correct_mode \<rbrace> update upd_f \<lbrace>@N K correct_mode \<rbrace>"
  apply(auto simp add: HoareNE_alt Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' e sv sp)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def
    startup_defs brute_force correct_mode_def 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp simp add: min_def)+
  apply(auto dest: LDead_override_LDegraded)
  apply(auto dest!: LDegraded_in_override_LDegraded)
  apply(case_tac[!] sv', simp_all)
  done

lemma update_correct_mode_N: 
  "\<lbrace> correct_mode \<rbrace> update ind \<lbrace>>R K correct_mode \<rbrace>"
  by(auto simp add: decompose_HoareN intro!: update_correct_mode_NN update_correct_mode_NE)

lemma update_call_correct_mode_N: 
  "\<lbrace> correct_mode \<rbrace> <tryR> update upd_f <catchR> startup \<lbrace>>R K correct_mode \<rbrace>"
  by(intro HoareN_try_catchR_reestablish update_correct_mode_N startup_correct_mode_N)

lemma restart_mangle_sv_HoareN:
  "\<lbrace>P\<rbrace> restart_mangle_sv \<lbrace>>R Q\<rbrace>"
  by(auto simp add: Hoare_alt Hoare_def Box_def brute_force)

lemma idle_restart_correct_mode_N: 
  "\<lbrace> correct_mode \<rbrace> <tryR> restart_mangle_sv <catchR> startup \<lbrace>>R K correct_mode \<rbrace>"
  by(intro HoareN_try_catchR_reestablish startup_correct_mode_N restart_mangle_sv_HoareN)

definition abs6_obs_correct_mode where
  "abs6_obs_correct_mode \<equiv> {(sv, sp). 
    (sv = ACNormal \<longrightarrow> LDead \<notin> range (obs_abs_mem.reg_state sp) \<and> obs_abs_mem.admin_state sp = LOK)
    \<and> (sv = ACDegraded \<longrightarrow> (
        (LDead \<notin> range (obs_abs_mem.reg_state sp) \<and> obs_abs_mem.admin_state sp \<noteq> LDead)
        \<and> (LDegraded \<in> range (obs_abs_mem.reg_state sp) \<or> obs_abs_mem.admin_state sp = LDegraded)))
    \<and> (sv = ACFail \<longrightarrow> (LDead \<in> range (obs_abs_mem.reg_state sp) \<or> obs_abs_mem.admin_state sp = LDead))
  }"

lemma abs6_obs_correct_mode: "oreach abs6_spec \<subseteq> abs6_obs_correct_mode"
proof(unfold abs6_spec_def, rule internal_to_external[where I=Abs6_Lemmas.correct_mode])
  show "reach abs6_ts \<subseteq> correct_mode" 
    apply(safe)
    apply(erule reach.induct)
    apply(auto simp add: abs6_ts_def abs6_trans_def set_fun_into_rel_def projectSN_def projectS_def
      choice_def)
    apply(insert startup_correct_mode_N, auto simp add: Hoare_alt Hoare_def Box_def K_def)[1]
    apply(insert update_call_correct_mode_N, auto simp add: Hoare_alt Hoare_def Box_def K_def)[1]
    apply(drule_tac x=upd_f in meta_spec)
    apply(force)
    apply(insert fetch_call_correct_mode_N, auto simp add: Hoare_alt Hoare_def Box_def K_def)[1]
    apply(drule_tac x=ind in meta_spec)
    apply(force)
    apply(insert idle_restart_correct_mode_N, auto simp add: Hoare_alt Hoare_def Box_def K_def)[1]
    done
next
  show "abs6_obs ` correct_mode \<subseteq> abs6_obs_correct_mode"
    by(force simp add: correct_mode_def abs6_obs_correct_mode_def image_def)
qed

definition obs_mem_in where
  "obs_mem_in S \<equiv> {(sv, sp). obs_abs_mem.memory sp \<in> S}"

definition no_failed where
  "no_failed \<equiv> {(sv, sp). admin_state sp \<noteq> LDead \<and> (\<forall>ind. reg_state sp ind \<noteq> LDead)}"

lemma startup_no_revival:
  "\<lbrace>- no_failed\<rbrace> startup \<lbrace>> K (- no_failed)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' v sv sp)
  apply(clarsimp_all simp add: throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force no_failed_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply ((metis antisym_conv less_eq_log_state.simps(1) less_eq_log_state.simps(2)
    min_def o_eq_dest_lhs override_on_def rangeI)+)
  done

lemma update_no_revival:
  "\<lbrace>-no_failed\<rbrace> update upd_f \<lbrace>> K (-no_failed) \<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' v sv sp)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: update_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    check_sizes'_def
    startup_defs min_def brute_force no_failed_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
  apply(simp_all)
  apply ((metis antisym_conv less_eq_log_state.simps(1) less_eq_log_state.simps(2)
    min_def o_eq_dest_lhs override_on_def rangeI)+)
  done

lemma update_call_no_revival:
  "\<lbrace>-no_failed\<rbrace> <tryR> update upd_f <catchR> startup \<lbrace>> K (-no_failed) \<rbrace>"
  by(intro Hoare_try_catchR_inv update_no_revival startup_no_revival)

lemma fetch_no_revival:
  "\<lbrace>- no_failed\<rbrace> fetch ind \<lbrace>> K (- no_failed)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' v sv sp)
  apply(simp_all only: fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: read_success_def fail_if_failed_def throw_mode_error_def abs6_skip_defs
    startup_defs min_def brute_force no_failed_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply(auto dest: LDead_override_LDegraded)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
  apply(simp_all)
  apply ((metis antisym_conv less_eq_log_state.simps(1) less_eq_log_state.simps(2)
    min_def o_eq_dest_lhs override_on_def rangeI)+)
  done

lemma fetch_call_no_revival:
  "\<lbrace>-no_failed\<rbrace> 
    (<tryR> fetch ind <catchR> startup)
  \<lbrace>> K (-no_failed)\<rbrace>"
  by(intro Hoare_try_catchR_inv fetch_no_revival startup_no_revival)

lemma restart_mangle_sp_no_revival:
  "\<lbrace>-no_failed\<rbrace> 
    restart_mangle_sp
  \<lbrace>> K (-no_failed)\<rbrace>"
  apply(auto simp add: Hoare_def Box_def K_beta first_stage)
  apply(rename_tac sv' sp' v sv sp)
  apply(clarsimp_all simp add: min_def brute_force no_failed_def
    dest: LDead_override_LDegraded 
    split: split_if_asm exceptions.split_asm)
  apply(iterative_simp)+
  apply (metis bot_log_state_def le_bot)+
  done

lemma idle_restart_no_revival:
  "\<lbrace>-no_failed\<rbrace> 
    <tryR> restart_mangle_sp <catchR> startup
  \<lbrace>> K (-no_failed)\<rbrace>"
  by(intro Hoare_try_catchR_inv startup_no_revival restart_mangle_sp_no_revival)

lemma obs_mem_in_None_failed:
  "abs6_obs -` obs_mem_in {None} = -no_failed"
  by(auto simp add: no_failed_def obs_mem_in_def split: split_if_asm)

lemma Hoare_imp_HoareN: "\<lbrace> I \<rbrace> m \<lbrace>> K I \<rbrace> \<Longrightarrow> \<lbrace> I \<rbrace> m \<lbrace>>R K I\<rbrace>"
  by(auto simp add: project_defs Hoare_alt Hoare_def Box_def K_def)

lemma transaction_None:
  "\<lbrace>abs6_obs -` obs_mem_in {None}\<rbrace> <tryR> update upd_f <catchR> startup \<lbrace>>R 
    K (abs6_obs -` (obs_mem_in {None}))\<rbrace>"
  by(auto simp add: obs_mem_in_None_failed update_call_no_revival[THEN Hoare_imp_HoareN])

definition memory_in where
  "memory_in S \<equiv> {(sv, sp). memory sp \<in> S }"

lemma obs_mem_in_Some: "abs6_obs -` obs_mem_in {Some mem} = no_failed \<inter> memory_eq mem"
  by(auto simp add: obs_mem_in_def no_failed_def memory_eq_def split: split_if_asm)

lemma transaction_Some_normal:
  "\<lbrace>abs6_obs -` obs_mem_in {Some mem}\<rbrace> 
    <tryR> update upd_f <catchR> startup 
  \<lbrace>>N K (abs6_obs -` obs_mem_in {Some (fun_map_override_on mem upd_f (dom upd_f))}) \<rbrace>
  "
  apply(auto simp add: obs_mem_in_Some)
  apply(rule HoareNN_Int)
  prefer 2
  apply(rule transaction_property_NN)
  apply(auto simp add: Hoare_alt Hoare_def Box_def K_beta first_stage)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: update_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice 
    cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(tactic {* distinct_subgoals_tac *})
  apply(clarsimp_all simp add: update_success_def throw_mode_error_def abs6_skip_defs
    startup_defs check_sizes'_def
    brute_force memory_eq_def split: split_if_asm exceptions.split_asm)
  apply(clarsimp_all simp add: no_failed_def)
  done

lemma transaction_Some_exceptional:
  "\<lbrace>abs6_obs -` obs_mem_in {Some mem}\<rbrace> 
    <tryR> update upd_f <catchR> startup 
  \<lbrace>@N K (abs6_obs -` obs_mem_in 
    {None, 
    Some mem, 
    Some (fun_map_override_on mem upd_f (dom upd_f))
    })
  \<rbrace>"
  apply(rule conseqNE_rule[OF _ transaction_property_NE[where mem=mem]])
  apply(auto simp add: obs_mem_in_def memory_eq_def K_def split: split_if_asm)
  done

end

