theory Abs6_Ref
imports
  "../Composition"
  "../Pmm_Refinement"
  "../Misc_Lemmas"
  "../Pmm_Lemmas"
  "../abs5/Abs5_CommonDefs"
  "../abs5/Abs5_Lemmas"
  "../abs5/Abs5_UpdateDefs"
  "../abs5/Abs5_FetchDefs"
  "../abs5/Abs5_StartupDefs"
  Abs6_CommonDefs
  Abs6_UpdateDefs
  Abs6_FetchDefs
  Abs6_StartupDefs
  Abs6_Lemmas
begin

lemmas brute_force =
  unfold_monad
  project_defs
  restart_mangle_sv_raw
  restart_mangle_sp_def
  getters_setters
  modify_def
  choose_def
  put_def
  vol_ops_defs
  force_same_and_def

definition abs6_abs5_strel :: "((abs_card_mode \<times> 'c abs_mem) \<times> 
  ('c abs5_vol_state \<times> 'c abs5_state))set" where
  "abs6_abs5_strel \<equiv> {((cmode, spa), (svc, spc)).
    cmode = abs_card_mode svc
    \<and> spa = abs_mem.truncate spc
  }"

lemma sim_mono_conc:
  "\<lbrakk> sim R ma mc'; \<forall>s. mc s \<subseteq> mc' s \<rbrakk> \<Longrightarrow> sim R ma mc"
  by(auto simp add: sim_def sim_ext_def del: subsetI
    elim!: simN_wpc_monotone_comp_subset order_trans[OF _ simS_wpc_monotone_comp_conc]
    dest: projectNN_mono projectNE_mono projectR_mono
  )

definition abs5_startup2' where
  "abs5_startup2' \<equiv> abs5_check_all \<guillemotright> throw_abs_mode_error EU \<guillemotright> throw StartedExc"

lemma startup_inclusion:
  "\<forall>s. Abs5_StartupDefs.startup s \<subseteq> (Abs5_StartupDefs.startup1 \<guillemotright> abs5_startup2') s"
  apply(simp add: Abs5_StartupDefs.startup_def)
  apply(intro allI ball_bind_mono2)
  apply(safe)
  apply(simp add: startup1_def compose_project)
  apply(erule_tac[!] rev_subsetD)
  apply(simp_all only: startup2_def abs5_startup2'_def)
  apply(intro bind_mono1 check_all_inclusion)
  apply(auto simp add: startup1_NN_def brute_force post_check_admin_def
    admin_not_LDead_def sv_abs_mode_neq_def)
  done

declare disjCI[rule del]

lemma override_with_LDegraded:
  "\<lbrakk> \<forall>ind. g ind \<le> f ind; LDead \<notin> range g\<rbrakk> \<Longrightarrow> (\<exists>deg. g = override_on f (\<lambda>y. LDegraded) deg)"
  apply(auto simp add: image_def override_on_def)
  apply(rule_tac x="g -` {LDegraded}" in exI)
  apply(rule ext, clarsimp)
  apply(case_tac "g a", auto)
  apply(metis)
  by (metis top_le top_log_state_def)

lemma double_override_simp:
  "override_on (override_on f g A) g B = override_on f g (A \<union> B)"
  by(auto simp add: override_on_def)

lemma log_state_simps:
  "\<lbrakk> x \<le> LDegraded; x \<noteq> LDead \<rbrakk> \<Longrightarrow>  x = LDegraded"
  "\<lbrakk> x \<le> LDegraded; x \<noteq> LDead \<rbrakk> \<Longrightarrow>  (LDegraded = x) = True"
  by(case_tac[!] x, simp_all)

lemma startup_ref:
  "sim abs6_abs5_strel
    Abs6_StartupDefs.startup
    Abs5_StartupDefs.startup
  "
  apply(rule sim_mono_conc[OF _ startup_inclusion])
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: startup1_def compose_project)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: abs5_startup2'_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: startup1_NN_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: abs5_check_all_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: startup1_NN_def brute_force fail_admin_def abs5_check_all_defs
    throw_abs_mode_error_def 
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: Abs6_StartupDefs.startup_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs6_abs5_strel_def)
  apply(case_tac[!] ba, clarsimp_all simp add: abs_mem.truncate_def ex_arg range_eqI[OF sym]
    bound_LDead_from_below override_with_LDegraded double_override_simp)
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms override_with_LDegraded} *})
  apply(clarsimp_all simp add: double_override_simp log_state_simps)
  apply (smt order_refl order_trans override_with_LDegraded)+
  done

lemma abs_card_mode_exhaust:
  "\<lbrakk>cm \<noteq> ACNormal; cm \<noteq> ACDegraded; cm \<noteq> ACFail\<rbrakk>
  \<Longrightarrow> False"
  by(case_tac cm, simp_all)

lemma fetch_ref_N:
  "sim_norm abs6_abs5_strel
    (Abs6_FetchDefs.fetch ind) 
    (Abs5_FetchDefs.fetch ind)
  "
  apply(auto simp add: sim_norm_def sim_norm_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(rename_tac[!] sva spa svc spc vc svc' spc')
  apply(simp_all only: Abs5_FetchDefs.fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: startup1_NN_def brute_force fail_admin_def read_defs
    throw_abs_mode_error_def degrade_admin_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
  apply(clarsimp_all simp add: min_def split: split_if_asm)
  apply(clarsimp_all simp add: Abs6_FetchDefs.fetch_def fail_if_failed_def
    Abs6_CommonDefs.throw_mode_error_def
    abs6_degrade_def abs6_fail_def
    read_success_def brute_force min_def)
  apply(safe)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs6_abs5_strel_def sv_isnt_def)
  apply(case_tac[!] spc, clarsimp_all simp add: abs_mem.truncate_def ex_arg range_eqI[OF sym]
    bound_LDead_from_below override_with_LDegraded double_override_simp min_def)
  apply(clarsimp_all simp add: double_override_simp log_state_simps override_on_with_min_le
    ex_override)
  apply(smt equals0D order_trans override_on_with_min_le ex_override override_on_emptyset
    abs_card_mode_exhaust less_eq_abs_card_mode.simps(6))+
  done

lemma fetch_ref:
  "sim abs6_abs5_strel
    (Abs6_FetchDefs.fetch ind) 
    (Abs5_FetchDefs.fetch ind)
  "
  apply(auto simp add: sim_def sim_ext_def del: subsetI intro:
    fetch_ref_N[simplified sim_norm_def sim_norm_ext_def Int_subset_iff, THEN conjunct1]
    fetch_ref_N[simplified sim_norm_def sim_norm_ext_def Int_subset_iff, THEN conjunct2]
  )
  apply(auto simp add: simS_wpc_def)
  apply(rename_tac[!] sva spa svc spc svc' spc')
  apply(simp_all only: Abs5_FetchDefs.fetch_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: maybe_mess_up_admin_silence_deg_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps maybe_mess_up_admin_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(simp_all only: read_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps read_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: startup1_NN_def brute_force fail_admin_def read_defs
    throw_abs_mode_error_def degrade_admin_def
    split: exceptions.split_asm split_if_asm)
  apply(iterative_simp)+
  apply(tactic {* Pmm_Tacs.full_dest_tac @{context} @{thms ACNormal_exhaust} *})
  apply(clarsimp_all simp add: Abs6_FetchDefs.fetch_def fail_if_failed_def
    Abs6_CommonDefs.throw_mode_error_def
    abs6_degrade_def abs6_fail_def
    read_success_def brute_force min_def)
  apply(safe)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs6_abs5_strel_def)
  apply(case_tac[!] spc, clarsimp_all simp add: abs_mem.truncate_def ex_arg range_eqI[OF sym]
    bound_LDead_from_below override_with_LDegraded double_override_simp min_def)
  apply(clarsimp_all simp add: double_override_simp log_state_simps override_on_with_min_le
    ex_override)
  apply(smt equals0D order_trans override_on_with_min_le ex_override override_on_emptyset
    abs_card_mode_exhaust less_eq_abs_card_mode.simps(6))+
  done

lemma skip_ref:
  "sim abs6_abs5_strel abs6_skip abs5_skip"
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all only: abs5_skip_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: abs5_skip_def abs5_degrade_def abs5_fail_def brute_force
    Abs5_CommonDefs.restart_mangle_sp_def)
  apply(iterative_simp)+
  apply(clarsimp_all simp add: abs6_skip_defs brute_force)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs6_abs5_strel_def abs_mem.truncate_def restart_mangle_sp_def)
  apply(metis)+
  done

lemma write_new'_ref_NN:
  "sim_normN abs6_abs5_strel
    abs6_loud_skip
    (Abs5_UpdateDefs.write_new' upd_f ind)
  "
  apply(auto simp add: sim_normN_def sim_normN_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp only: write_new'_def write_upd_mem_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  apply(case_tac spc)
  apply(clarsimp_all simp add: abs6_skip_defs brute_force abs6_abs5_strel_def abs_mem.truncate_def)
  done

lemma write_new'_ref_NE:
  "sim_normE abs6_abs5_strel
    abs6_loud_skip
    (Abs5_UpdateDefs.write_new' upd_f ind)
  "
  apply(auto simp add: sim_normE_def sim_normE_ext_def simN_wpc_def force_same_and_def)
  apply(rename_tac sva spa svc spc e svc' spc')
  apply(simp only: write_new'_def write_upd_mem_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  apply(case_tac[!] spc)
  apply(clarsimp_all simp add: abs6_skip_defs brute_force abs6_abs5_strel_def abs_mem.truncate_def
    chooseN_def)
  apply (metis equals0D range_eqI)+
  done

lemma write_new'_ref:
  "sim abs6_abs5_strel
    abs6_loud_skip
    (Abs5_UpdateDefs.write_new' upd_f ind)
  "
  apply(auto simp add: sim_through_simN_simS
    intro:
    write_new'_ref_NN[simplified sim_normN_through_simN]
    write_new'_ref_NE[simplified sim_normE_through_simN]
  )
  apply(auto simp add: simS_def simS_wpc_def)
  apply(rename_tac sva spa svc spc svc' spc')
  apply(simp only: write_new'_def write_upd_mem_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: write_upd_mem_defs brute_force)
  apply(case_tac[!] spc)
  apply(clarsimp_all simp add: abs6_skip_defs brute_force abs6_abs5_strel_def abs_mem.truncate_def
    chooseN_def)
  done

lemma set_updated_ref:
  "sim abs6_abs5_strel (return ()) (set_abs_updated u)"
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(auto simp add: brute_force abs6_abs5_strel_def)
  done

lemma clear_upd_mem_ref:
  "sim abs6_abs5_strel (return ()) clear_upd_mem"
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(auto simp add: brute_force clear_upd_mem_def modifyRE_def abs6_abs5_strel_def
    abs_mem.truncate_def)
  done

lemma chooseN_bind_subset:
  "(chooseN A \<guillemotright> m) s \<subseteq> m s"
  by(auto simp add: chooseN_def brute_force)

lemma sim_mono_abs:
  "\<lbrakk>sim R ma' mc; \<forall>s. ma' s \<subseteq> ma s\<rbrakk> \<Longrightarrow> sim R ma mc"
  by(auto simp add: sim_def sim_ext_def del: subsetI
    elim!: order_trans[OF _ simN_wpc_monotone_comp_abs] order_trans[OF _ simS_wpc_monotone_comp_abs]
    dest: projectNN_mono projectNE_mono projectR_mono
  )

lemma rollback_ref[ref_thm]:
  "sim abs6_abs5_strel
    (abs6_skip \<guillemotright> throw (Degraded (EB False)))
    Abs5_UpdateDefs.rollback
  "
  apply(auto simp add: K_beta rollback_def intro!: sim_bindI)
  apply(rule sim_mono_conc[OF _ touch_updated_inclusion])
  apply(subst bind_right_unit[THEN sym])
  apply(rule sim_bindI[OF skip_ref])
  apply(auto intro: sims_throw set_updated_ref)
  done
 
lemma pre_commit_ref':
  "sim abs6_abs5_strel
    abs6_pre_commit
    (Abs5_UpdateDefs.pre_commit upd_f)"
  apply(simp add: Abs5_UpdateDefs.pre_commit_def)
  apply(subst bind_left_unit[where a="()", THEN sym])
  apply(rule sim_bindI[OF set_updated_ref])
  apply(clarsimp)
  apply(subst bind_left_unit[where a="()", THEN sym])
  apply(rule sim_bindI[OF clear_upd_mem_ref])
  apply(clarsimp)
  apply(rule sim_mono_abs[OF _ allI, OF _ chooseN_bind_subset[where A="order_set (dom upd_f)"]])
  apply(rule sim_bindI)
  apply(rule sim_chooseN)
  apply(clarsimp simp add: abs6_pre_commit_def)
  apply(rule sim_try_catchI)
  apply(rule sim_mono_abs[OF _ mapM'_abs6_loud_skip_inclusion])
  apply(rule sim_mapM')
  apply(clarsimp intro!: write_new'_ref)
  apply(auto split: exceptions.split intro: sims_throw rollback_ref)
  done

lemma pre_commit_ref:
  "sim_ext (abs6_abs5_strel \<inter> sv_is m \<times> UNIV)
    abs6_pre_commit
    (Abs5_UpdateDefs.pre_commit upd_f)
    (force_same_and (abs6_abs5_strel \<inter> sv_is m \<times> (upd_mem_eq upd_f \<inter> dom_upd_mem_is_updated)))
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(rule pre_commit_ref'[simplified sim_def, THEN sim_ext_mono
    , THEN sim_ext_add_NN_inv[OF abs6_pre_commit_mode_post_NN
        HoareNN_join_post[OF pre_commit_post_NN pre_commit_updated_post_NN]]
    , THEN sim_ext_mono, THEN sim_ext_mono_norm
    ]
  )
  apply(auto simp add: K_def force_same_and_def)
  done

lemma copy_upd_mem_ref[ref_thm]:
  "sim_ext (abs6_abs5_strel \<inter> sv_is ACNormal \<times> (upd_mem_eq upd_f \<inter> dom_upd_mem_is_updated))
    (abs6_copy_upd_mem upd_f)
    Abs5_UpdateDefs.copy_upd_mem
    (force_same_and abs6_abs5_strel)
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def force_same_and_def)
  apply(simp_all add: copy_upd_mem_def)
  apply(simp_all add: first_stage vol_ops_project_simps if_choice restart_mangle_sv_proj_simps
      restart_mangle_sp_proj_simps
      cong: exceptions.case_cong if_cong)
  apply(simp_all only: choice_def Un_iff, safe)
  apply(clarsimp_all simp add: copy_upd_mem_defs brute_force)
  apply(clarsimp_all simp add: abs6_copy_upd_mem_def abs6_skip_defs brute_force
    update_success_def)
  apply(simp_all add:
    HOL.conj_disj_distribR HOL.conj_disj_distribL ex_disj_distrib conj_assoc 
    disj_commute disj_left_commute disj_absorb disj_left_absorb
    conj_commute conj_left_commute conj_absorb conj_left_absorb
    ex_conj_distribs all_conj
  )
  apply(clarsimp_all simp add: abs6_abs5_strel_def abs_mem.truncate_def min_def
    dom_upd_mem_is_updated_def upd_mem_eq_def sv_is_def)
  apply(tactic {* distinct_subgoals_tac *})
  apply(metis)
  done

lemma do_commit_ref':
  "sim_ext (abs6_abs5_strel \<inter> sv_is ACNormal \<times> (upd_mem_eq upd_f \<inter> dom_upd_mem_is_updated))
    (abs6_do_commit upd_f)
    Abs5_UpdateDefs.do_commit
    (force_same_and abs6_abs5_strel)
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(auto simp add: abs6_do_commit_def Abs5_UpdateDefs.do_commit_def
    force_same_and_diff' force_same_and_same
    intro!: sim_ext_try_catchI copy_upd_mem_ref rollback_ref sim_exts_empty_rel
    sim_ext_throw sim_ext_return
    split: exceptions.split
  )
  apply(subst force_same_and_def)
  apply(auto intro!: rollback_ref[simplified sim_def] sim_exts_empty_rel)
  done

lemma do_commit_ref:
  "sim_ext (abs6_abs5_strel \<inter> sv_is ACNormal \<times> (upd_mem_eq upd_f \<inter> dom_upd_mem_is_updated))
    (abs6_do_commit upd_f)
    Abs5_UpdateDefs.do_commit
    (force_same_and abs6_abs5_strel)
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(rule do_commit_ref'[THEN sim_ext_mono_norm])
  apply(auto simp add: force_same_and_def K_def)
  done

lemma rollforward_ref':
  "sim abs6_abs5_strel abs6_skip rollforward"
  apply(simp add: rollforward_def)
  apply(rule sim_mono_conc[OF _ touch_updated_inclusion])
  apply(subst bind_right_unit[THEN sym])
  apply(rule sim_bindI[OF skip_ref])
  apply(auto intro: sims_throw set_updated_ref)
  done

lemma rollforward_ref:
  "sim_ext abs6_abs5_strel
    abs6_skip
    rollforward
    (force_same_and abs6_abs5_strel)
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(rule rollforward_ref'[simplified sim_def])
  done
 
lemma check_sizes_ref:
  "sim R
    (PmmDefs.check_sizes upd_f) (PmmDefs.check_sizes upd_f)
  "
  by(auto simp add: PmmDefs.check_sizes_def
    intro!: sims_all_intros)

lemma sim_ext_mono_restart:
  "\<lbrakk> sim_ext R ma mc Sn Se Sr; Sr \<subseteq> Sr' \<rbrakk> \<Longrightarrow> sim_ext R ma mc Sn Se Sr'"
  by(force simp add: sim_ext_def simS_wpc_def projectR_def)

lemma update_ref':
  "sim_ext abs6_abs5_strel
    (abs6_update' upd_f)
    (Abs5_UpdateDefs.update upd_f)
    (force_same_and abs6_abs5_strel)
    (force_same_and abs6_abs5_strel)
    abs6_abs5_strel
  "
  apply(simp add: abs6_update'_def Abs5_UpdateDefs.update_def)
  apply(rule sim_ext_bindI[where Tn="force_same_and (abs6_abs5_strel \<inter> sv_is ACNormal \<times> UNIV)"])
  defer
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_bindI)
  apply(rule check_sizes_ref[unfolded sim_def, THEN sim_ext_mono_exc, THEN sim_ext_mono_restart])
  apply(auto simp add: force_same_and_def)[2]
  apply(auto simp add: force_same_and_same)
  apply(rule sim_ext_bindI[OF pre_commit_ref])
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF do_commit_ref])
  apply(subst force_same_and_def)
  apply(auto intro!: sim_exts_empty_rel)
  apply(rule sim_ext_bindI[OF rollforward_ref])
  apply(auto simp add: sim_def sim_ext_def simN_wpc_def simS_wpc_def abs6_abs5_strel_def
    sv_isnt_def sv_is_def abs_mem.truncate_def
    throw_mode_error_def throw_abs_mode_error_def brute_force)
  apply(metis abs_card_mode.exhaust)+
  done
 
lemma simS_wpc_monotone_comp_abs':
  "\<forall>s \<in> fst ` (simS_wpc ma mc S). ma s \<subseteq> ma' s \<Longrightarrow> simS_wpc ma mc S \<subseteq> simS_wpc ma' mc S"
  apply(auto simp add: simS_wpc_def)
  by (metis (lifting) set_rev_mp) 

lemma simN_wpc_monotone_comp_abs':
  "\<forall>s \<in> fst ` (simN_wpc ma mc S). ma s \<subseteq> ma' s \<Longrightarrow> simN_wpc ma mc S \<subseteq> simN_wpc ma' mc S"
  apply(auto simp add: simN_wpc_def)
  by (metis (lifting) set_rev_mp) 

lemma sim_ext_mono_abs_comp':
  "\<lbrakk>sim_ext R ma mc Sn Se Sr; \<forall>s \<in> fst ` R. ma s \<subseteq> ma' s\<rbrakk> \<Longrightarrow> sim_ext R ma' mc Sn Se Sr"
  apply(auto simp add: sim_ext_def simN_wpc_def simS_wpc_def brute_force del: subsetI)
  apply(clarsimp_all)
  apply (smt empty_iff fst_conv in_mono mem_Collect_eq prod.simps(2))+
  done

lemma sim_mono_abs':
  "\<lbrakk>sim R ma' mc; \<forall>s \<in> fst ` R. ma' s \<subseteq> ma s\<rbrakk> \<Longrightarrow> sim R ma mc"
  apply(auto simp add: sim_def sim_ext_def del: subsetI)
  apply(clarsimp_all simp add: simN_wpc_def simS_wpc_def brute_force)
  apply (smt empty_iff fst_conv in_mono mem_Collect_eq prod.simps(2))+
  done

lemma update_ref:
  "sim abs6_abs5_strel
    (Abs6_UpdateDefs.update upd_f)
    (Abs5_UpdateDefs.update upd_f)
  "
  apply(unfold sim_def)
  apply(rule sim_ext_mono_abs_comp'[OF update_ref'])
  apply(auto intro!: abs6_update'_inclusion[rule_format, THEN subsetD])
  done

lemma fetch_full_ref:
  "sim abs6_abs5_strel
    (<tryR>
        Abs6_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs6_StartupDefs.startup)
    )
    (<tryR>
        Abs5_FetchDefs.fetch ind
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
  "
  by(intro sim_try_catchRI[OF fetch_ref] sim_rec_tryRI startup_ref)

lemma update_full_ref:
  "sim abs6_abs5_strel
    (<tryR>
        Abs6_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs6_StartupDefs.startup)
    )
    (<tryR>
        Abs5_UpdateDefs.update upd_f
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
  "
  by(intro sim_try_catchRI[OF update_ref] sim_rec_tryRI startup_ref)

lemma startup_full_ref:
  "sim abs6_abs5_strel
    (<tryR>
        Abs6_StartupDefs.startup
    <catchR>
        rec_tryR (Abs6_StartupDefs.startup)
    )
    (<tryR>
        Abs5_StartupDefs.startup
    <catchR>
        rec_tryR (Abs5_StartupDefs.startup)
    )
  "
  by(intro sim_try_catchRI[OF startup_ref] sim_rec_tryRI startup_ref)
 
lemma sim_norm_ext_mono_abs_comp:
  "\<lbrakk> sim_norm_ext R ma mc Sn Se; \<forall>s. (ma\<Down>N) s \<subseteq> (ma'\<Down>N) s \<rbrakk> \<Longrightarrow> sim_norm_ext R ma' mc Sn Se"
  apply(auto simp add: sim_norm_ext_def del: subsetI)
  apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs])
  apply(force, force simp add: project_defs)
  apply(rule subset_trans[OF _ simN_wpc_monotone_comp_abs])
  apply(force, force simp add: project_defs)
  done

lemma startup_ref_reduction:
  "sim R
    (<tryR>
        ma
    <catchR>
        rec_tryR (Abs6_StartupDefs.startup)
    )
    mc
  \<Longrightarrow>
  sim_norm R
    (<tryR>
        ma
    <catchR>
        (Abs6_StartupDefs.startup)
    )
    mc
  "
  apply(unfold sim_def sim_norm_def)
  apply(drule sim_ext_imp_sim_norm_ext)
  apply(erule sim_norm_ext_mono_abs_comp)
  apply(intro allI)
  apply(simp only: first_stage choice_def)
  apply(intro Un_mono[OF order_refl] bindS_mono2 allI startup_reduction)
  done

lemmas startup_reduced_ref = startup_full_ref[THEN startup_ref_reduction]

lemmas fetch_reduced_ref = fetch_full_ref[THEN startup_ref_reduction]

lemmas update_reduced_ref = update_full_ref[THEN startup_ref_reduction]

lemma abs6_startup_full_ref_N:
  "sim_norm abs6_abs5_strel
    Abs6_StartupDefs.startup
    (<tryR> 
        Abs5_StartupDefs.startup 
    <catchR> rec_tryR Abs5_StartupDefs.startup)
  "
  apply(unfold sim_norm_def)
  apply(rule sim_norm_ext_mono_abs_comp)
  apply(rule sim_norm_ext_mono)
  apply(rule startup_reduced_ref[simplified sim_norm_def])
  apply(simp)
  apply(rule allI)
  apply(rule startup_reduction')
  done

lemmas abs6_update_full_ref_N = update_reduced_ref[folded sim_norm_def]
lemmas abs6_fetch_full_ref_N = fetch_reduced_ref[folded sim_norm_def]

end