theory CommonDefs
  imports LLInterface
begin

text {*
  Definitions used throughout the PMM operations
*}

definition get_hash_table :: "INST \<Rightarrow> ('c htable, 'sv, 'c :: contents_cl)pmm_monad" where
  get_hash_table_def[monad_def]:
  "get_hash_table inst \<equiv> do {
    cnt \<leftarrow> ll_read (Admin HASH_REG) inst;
    (if ok_hash_cnt cnt
        then return (parse_htable (fst (parse_hinst cnt)))
        else throw FailRead
    )
  }"

definition get_hash_table_or_fail where
  get_hash_table_or_fail_def[monad_def]:
  "get_hash_table_or_fail inst \<equiv>
    <try> get_hash_table inst <catch> K (do { set_failed; throw FailMode })"

definition write_hash_table :: 
  "INST \<Rightarrow> 'c htable \<Rightarrow> (unit, 'sv, 'c :: contents_cl)pmm_monad" where
  write_hash_table_def[monad_def]:
  "write_hash_table inst ht \<equiv> write_f (Admin HASH_REG) inst (htable_to_hinst_cnt ht)"

definition get_hashes :: "('c htable, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  get_hashes_def[monad_def]:
  "get_hashes \<equiv>
    <try> 
        (get_hash_table Secondary) 
    <catch> (\<lambda>e. case e of
        FailLLRead \<Rightarrow> do {
                ht \<leftarrow> get_hash_table_or_fail Primary;
                set_degraded_h;
                throw (Degraded EU)
            }
        | FailRead \<Rightarrow> do {
                ht \<leftarrow> get_hash_table_or_fail Primary;
                (<try>
                    write_hash_table Secondary ht
                <catch> K (do { set_degraded_h; throw (Degraded EU) })
                );
                return ht
            }
        | _ \<Rightarrow> throw e
    )"

text {*

  This is the version we want to use in DegradedH mode; no syncing
  is attempted, we just read the first correct hash region instance
  we can get our hands on

  *}

definition get_hashes_DegradedH :: "('c htable, pmm_vol_mem, 'c :: contents_cl)pmm_monad"
  where
  get_hashes_DegradedH_def[monad_def]:
  "get_hashes_DegradedH \<equiv>
    <try> 
        (get_hash_table Secondary) 
    <catch> K (get_hash_table_or_fail Primary)
    "

definition read_DegradedH :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)pmm_monad" where
   read_DegradedH_def[monad_def]: 
  "read_DegradedH ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> get_hashes_DegradedH;
    (if \<not> matches_htable ht ind inst cnt
        then throw FailRead
        else return cnt)
  }"

definition read_DegradedH_or_fail
  :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)pmm_monad" where
   read_DegradedH_or_fail_def[monad_def]: 
  "read_DegradedH_or_fail ind inst \<equiv>
  <try> read_DegradedH ind inst <catch> K (do { set_failed; throw FailMode })"

definition read :: "indx \<Rightarrow> INST \<Rightarrow> ('c :: contents_cl, pmm_vol_mem, 'c)pmm_monad" where
  read_def[monad_def]:
  "read ind inst \<equiv> do {
    cnt \<leftarrow> (
        <try>
            ll_read ind inst
        <catch> K (do { set_degraded; throw (Degraded EU) })
    );
    ht \<leftarrow> (
        <try>
            get_hashes
        <catch> (\<lambda>e. case e of
            Degraded _ \<Rightarrow> do {
                ht \<leftarrow> get_hashes_DegradedH;
                (if matches_htable ht ind inst cnt
                    then throw (Degraded (EC cnt))
                    else throw (Degraded EU)
                )
            }
            | _ \<Rightarrow> throw e
        )
    );
    (if \<not>matches_htable ht ind inst cnt
        then throw FailRead
        else return cnt)
  }"

text {*
  Read a public region instance, when the other instance is known
  to be faulty.
*}
definition read_or_fail :: "indx \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  read_or_fail_def[monad_def]:
  "read_or_fail pind inst \<equiv> 
    <try>
        read pind inst
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do { set_failed; throw FailMode }
            else throw (Degraded v)
        | _ \<Rightarrow> do { set_failed; throw FailMode }
    )
  "

definition read_or_fail_silence_deg
  :: "indx \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)pmm_monad"
  where read_or_fail_silence_deg_def[monad_def]:
  "read_or_fail_silence_deg ind inst \<equiv>
    <try> read_or_fail ind inst <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (get_exc_val_cnt v)
        | _ \<Rightarrow> throw e
    )
  "
 
text {*
  In case of a failed write to the hashes, we check whether the other
  hash instance is still good. This is not strictly necessary, still
  it should be OK (otherwise we would be in DegradedH and would not
  attempt using @{term "write_hashes"} at all! But we would need
  an invariant to prove this, and this check makes things simpler

  If we encounter errors while writing the hashes, we return a
  bool-carrying exception, where the bool value determines whether
  the new hash value was written or not

*}
definition write_hashes ::
  "'c htable \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  write_hashes_def[monad_def]:
  "write_hashes ht \<equiv> do {
    (<try> 
        write_hash_table Secondary ht
    <catch> (\<lambda>e. (case e of
        FailWrite \<Rightarrow> do { get_hash_table_or_fail Primary; set_degraded_h; throw (Degraded (EB False)) }
        | _ \<Rightarrow> throw e))
    );
    <try>
        write_hash_table Primary ht
    <catch> (\<lambda>e. (case e of
        FailWrite \<Rightarrow> do { set_degraded_h; throw (Degraded (EB True)) }
        | _ \<Rightarrow> throw e))
  }"

definition update_instance_hash ::
  "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, pmm_vol_mem, 'c)pmm_monad" where
  update_instance_hash_def[monad_def]:
  "update_instance_hash ind inst cnt \<equiv> do {
        ht \<leftarrow> get_hashes;
        write_hashes (update_htable ht ind inst cnt)
  }
  "

definition write_instance ::
  "indx \<Rightarrow> INST \<Rightarrow> 'c \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  write_instance_def[monad_def]:
  "write_instance ind inst cnt \<equiv> do {
    mode \<leftarrow> get_mode;
    (if mode = CDegradedH
    then throw (Degraded (EB False))
    else return());
    (<try>
        write_f ind inst cnt
    <catch> (\<lambda>e. case e of
        InvalidInput \<Rightarrow> throw InvalidInput
        | FailWrite \<Rightarrow> do {
            set_degraded;
            (<try> 
                read ind inst
            <catch> (\<lambda>e. case e of
                Degraded v \<Rightarrow>
                    if v = EU
                    then throw (Degraded (EB False))
                    else throw (Degraded (EB True))
                | FailRead \<Rightarrow> throw (Degraded (EB False))
                | _ \<Rightarrow> throw e)
            );
            throw (Degraded (EB True))
        }
        | _ \<Rightarrow> throw e
    ));
    <try>
        update_instance_hash ind inst cnt
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then throw (Degraded v)
            else
                do {
                      cnt \<leftarrow> (
                          <try>
                              read_DegradedH ind inst
                          <catch> (\<lambda>e. case e of
                              FailMode \<Rightarrow> throw FailMode
                              | _ \<Rightarrow> throw (Degraded (EB False))
                          )
                      );
                      throw (Degraded (EB True))
                  }
        | _ \<Rightarrow> throw e
    )
  }"

definition read_ptr_instance_DegradedH 
  :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  "read_ptr_instance_DegradedH inst \<equiv> do {
    cnt \<leftarrow> read_DegradedH (Admin PTR_REG) inst;
    return (parse_ptr cnt)
  }"

definition write_ptr_instance :: 
  "INST \<Rightarrow> pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  write_ptr_instance_def[monad_def]:
  "write_ptr_instance inst ptr_set \<equiv> do {
    mode \<leftarrow> get_mode;
    (if mode = CDegradedH
    then throw (Degraded (EB False))
    else return());
    let cnt = serialize_ptr ptr_set;
    (<try>
        write_f (Admin PTR_REG) inst cnt
    <catch> K (do { set_degraded; throw (Degraded (EB False)) })
    );
    <try>
        update_instance_hash (Admin PTR_REG) inst cnt
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EB True
            then throw (Degraded v)
            else
                do {
                    (<try>
                        read_ptr_instance_DegradedH inst
                    <catch> (\<lambda>e. case e of
                        FailMode \<Rightarrow> throw FailMode
                        | _ \<Rightarrow> throw (Degraded (EB False))
                    ));
                    throw (Degraded (EB True))
                  }
        | _ \<Rightarrow> throw e
    )
  }"

definition sync :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> ('c, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  sync_def[monad_def]:
  "sync ind to from \<equiv> do {
    cnt \<leftarrow> (
        <try>
            read_or_fail (Public ind) from
        <catch> (\<lambda>e. case e of
            Degraded v  \<Rightarrow> throw (Degraded (EC (get_exc_val_cnt v)))
            | _ \<Rightarrow> throw e
        )
    );
    (<try>
        write_instance (Public ind) to cnt
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> throw (Degraded (EC cnt))
        | _ \<Rightarrow> throw e
    ));
    return cnt
  }"

definition read_ptr_instance where
  read_ptr_instance_def[monad_def]:
  "read_ptr_instance inst \<equiv> do {
    cnt \<leftarrow> read (Admin PTR_REG) inst;
    return (parse_ptr cnt)
  }"

definition read_ptr_instance_or_fail 
  :: "INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)pmm_monad" 
  where
  read_ptr_instance_or_fail_def[monad_def]:
  "read_ptr_instance_or_fail inst \<equiv> do {
    <try>
        read_ptr_instance inst
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do { set_failed; throw FailMode }
            else throw (Degraded v)
        | _ \<Rightarrow> do { set_failed; throw FailMode }
    )
  }"

definition sync_ptr :: "INST \<Rightarrow> INST \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  sync_ptr_def[monad_def]:
  "sync_ptr to from \<equiv> do {
    ptr_set \<leftarrow> (
        <try>
            read_ptr_instance_or_fail from
        <catch> (\<lambda>e. case e of
            Degraded v  \<Rightarrow> throw (Degraded (EC (get_exc_val_cnt v)))
            | _ \<Rightarrow> throw e
        )
    );
    (<try>
        write_ptr_instance to ptr_set
    <catch> (\<lambda>e. case e of 
        Degraded _ \<Rightarrow> do { set_degraded_p; throw (Degraded (EC (serialize_ptr ptr_set))) }
        | _ \<Rightarrow> throw e
    ));
    return ptr_set
  }"

definition just_sync :: "pub_indx \<Rightarrow> INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  just_sync_def[monad_def]:
  "just_sync pind to from \<equiv> sync pind to from \<guillemotright> return ()"

definition just_sync_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad"
  where just_sync_silence_deg_def[monad_def]:
  "just_sync_silence_deg to from ind \<equiv>
    <try> 
        just_sync ind to from
    <catch> (\<lambda>e. if (\<exists>v. e = Degraded v) then return () else throw e)
  "

definition just_sync_ptr :: "INST \<Rightarrow> INST \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  just_sync_ptr_def[monad_def]:
  "just_sync_ptr to from \<equiv> sync_ptr to from \<guillemotright> return ()"

definition read_ptr_instance_or_fail_silence_deg where
  read_ptr_instance_or_fail_silence_deg_def[monad_def]:
  "read_ptr_instance_or_fail_silence_deg inst \<equiv>
    <try> read_ptr_instance_or_fail inst <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow> return (parse_ptr (get_exc_val_cnt v))
        | _ \<Rightarrow> throw e
    )
  "

text {*
  Get the value written in the PTR. It is safe to call this even
  when the PTR itself is degraded, since the syncing attempt below
  cannot make things worse (i.e. there is no danger of a restart
  interrupting a write and thus destroying the only remaining correct
  copy of the PTR). However, a Degraded exception will be thrown.
*}

(* TODO: cnt seems to be unused in the FailLLRead hander *)
definition get_ptr :: "bool \<Rightarrow> (pub_indx set, pmm_vol_mem, 'c :: contents_cl) pmm_monad" where
  get_ptr_def[monad_def]:
  "get_ptr silence_deg \<equiv>
    <try> 
        read_ptr_instance Secondary
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            if v = EU
            then do {
                ptr_set \<leftarrow> read_ptr_instance_or_fail_silence_deg Primary;
                set_degraded_p;
                if silence_deg
                then return ptr_set
                else throw (Degraded EU)
            }
            else
                if silence_deg
                then return (parse_ptr (get_exc_val_cnt v))
                else throw (Degraded EU)
        | FailRead \<Rightarrow> 
                    <try>
                        sync_ptr Secondary Primary
                    <catch> (\<lambda>e. case e of
                        Degraded v \<Rightarrow>
                            if silence_deg
                            then return (parse_ptr (get_exc_val_cnt v))
                            else throw (Degraded EU)
                        | _ \<Rightarrow> throw e
                    )
        | _ \<Rightarrow> throw e
    )
  "
  
text {*
  Writing the PTR.

  Similar to @{term "write_hashes"}, in case of a failure we check
  that the other instance is still valid, which will make proofs
  easier later on, even if it is not strictly necessary.

  Also, we try to make sure that we don't end up with two different,
  but individually correct versions of the PTR. Note that @{term
  "write_hashes"} and, consequently, @{term "write_instance"} throw
  a bool-carrying DegradedH exception, signalling whether the write
  to the hash region was carried out before detecting the error.

  Finally, if a degradation in the PTR is detected, we still signal
  whether the write went through, through a boolean value wrapped
  in the exception
*}

definition write_ptr :: "pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  write_ptr_def[monad_def]:
  "write_ptr ptr_set \<equiv> 
  do {
    mode \<leftarrow> get_mode;
    (if mode \<in> {CFail, CDegradedH, CDegradedP}
    then throw (Degraded (EB False))
    else return ());
    (<try>
        write_ptr_instance Secondary ptr_set
    <catch> (\<lambda>e. case e of
        Degraded v \<Rightarrow>
            do {
                (if v = EB True
                then <try> write_f (Admin PTR_REG) Primary (serialize_ptr ptr_set) <catch> K (return ())
                else read_ptr_instance_or_fail_silence_deg Primary \<guillemotright> return ()
                );
                set_degraded_p;
                throw (Degraded (if v = EB True then v else EB False))
            }
        | _ \<Rightarrow> throw e
    ));
    (<try>
        write_ptr_instance Primary ptr_set
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> do { set_degraded_p; throw (Degraded (EB True)) }
        | _ \<Rightarrow> throw e
    ))
  }
  "

definition remove_from_ptr :: "pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  remove_from_ptr_def[monad_def]:
  "remove_from_ptr ind \<equiv> do {
    ptr \<leftarrow> get_ptr False;
    let ptr' = ptr - {ind};
    write_ptr ptr'
  }"

text {* This definition serves just to facilitate the abs2/4
  refinement proof; after get_ptr, we know that at least one PTR is
  good, so we can prove we don't "bring the PTR back from the dead" in
  that abstraction, without needing invariants (such as "write_ptr
  is never called when the PTR is degraded" *}

definition set_ptr :: "pub_indx set \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  set_ptr_def[monad_def]:
  "set_ptr ptr \<equiv> do {
    get_ptr False;
    write_ptr ptr
  }"

definition get_hashes_silence_deg
  :: "('c htable, pmm_vol_mem, 'c :: contents_cl)pmm_monad" where
  get_hashes_silence_deg_def[monad_def]:
  "get_hashes_silence_deg \<equiv>
    <try>
        get_hashes
    <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> get_hashes_DegradedH
        | _ \<Rightarrow> throw e
    )
  "

definition get_sync_direction
  :: "pub_indx \<Rightarrow> (sync_direction, pmm_vol_mem, 'c :: contents_cl)pmm_monad"
  where
  get_sync_direction_def[monad_def]:
  "get_sync_direction pind \<equiv> do {
    ht \<leftarrow> get_hashes_silence_deg;
    if (pub_hashes ht (pind, Primary) = pub_hashes ht (pind, Secondary))
    then (return SEq)
    else do {
        ptr_set \<leftarrow> get_ptr True;
        (if pind \<in> ptr_set
        then return SL
        else return SR)
    }
  }"

definition remove_from_ptr_silence_deg where
  remove_from_ptr_silence_deg_def[monad_def]:
  "remove_from_ptr_silence_deg ind \<equiv>
        <try>
            remove_from_ptr ind
        <catch> (\<lambda>e. if e = FailMode then throw FailMode else return ())
  "

definition sync_and_remove_from_ptr_silence_deg
  :: "INST \<Rightarrow> INST \<Rightarrow> pub_indx \<Rightarrow> (unit, pmm_vol_mem, 'c :: contents_cl)pmm_monad"
  where sync_and_remove_from_ptr_silence_deg_def[monad_def]:
  "sync_and_remove_from_ptr_silence_deg to from ind \<equiv>
    <try> do {
        sync ind to from;
        remove_from_ptr_silence_deg ind
    } <catch> (\<lambda>e. case e of
        Degraded _ \<Rightarrow> return ()
        | _ \<Rightarrow> throw e
    )
  "

lemmas procedure_defs =
  get_hash_table_def
  write_hash_table_def
  get_hashes_def
  get_hashes_DegradedH_def
  write_hashes_def
  update_instance_hash_def
  write_instance_def
  read_def
  sync_def
  just_sync_def
  read_or_fail_def
  get_ptr_def
  write_ptr_def
  set_degraded_def
  set_degraded_h_def
  throw_mode_error_def

end
