theory LLInterface_Lemmas
  imports LLInterface "../Pmm_Tacs"
begin

lemmas brute_force =
  unfold_monad
  project_defs
  write_f_defs
  ll_read_def 
  ll_read_success_def 
  ll_read_failure_def 
  ll_read_corrupt_def
  restart_mangle_sv_raw

definition write_f_alt :: "indx \<Rightarrow> INST \<Rightarrow> 'c :: contents_cl \<Rightarrow> (unit, 'sv, 'c)pmm_monad" where
  "write_f_alt ind inst cnt \<equiv> write_success ind inst cnt
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_fail_instance ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_fail_readonly ind inst cnt <catch> K restart_mangle_sv)"

declare disjCI[iff del]

lemma write_f_alt_L:
  "write_f ind inst cnt s \<subseteq> (write_success ind inst cnt
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_fail_instance ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_fail_readonly ind inst cnt <catch> K restart_mangle_sv)) s"
  apply(fold write_f_alt_def)
  apply(auto simp only: write_f_def choice_distrib monad_laws)
  apply(auto simp only: choice_def)
  apply(clarsimp_all simp add: brute_force)
  apply(clarsimp_all simp add: write_f_alt_def brute_force HOL.conj_disj_distribR
    insert_commute fun_upd_twist)
  apply(metis)+
  done
  
lemma write_f_alt_R:
  "(write_success ind inst cnt
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_fail_instance ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_fail_readonly ind inst cnt <catch> K restart_mangle_sv)) s
  \<subseteq> write_f ind inst cnt s"
  apply(auto simp only: choice_distrib monad_laws)
  apply(auto simp only: choice_def)
  apply(unfold write_f_def)
  apply(simp_all only: choice_distrib monad_laws)
  apply(clarsimp_all simp add: brute_force HOL.conj_disj_distribR)
  by smt

lemma write_f_alt:
  "write_f ind inst cnt = write_success ind inst cnt
    [+] write_fail_instance ind inst cnt
    [+] write_corrupt_restart ind inst cnt
    [+] write_fail_readonly ind inst cnt
    [+] write_too_large ind inst cnt
    [+] restart_mangle_sv
    [+] do { write_success ind inst cnt; restart_mangle_sv }
    [+] (<try> write_fail_instance ind inst cnt <catch> K restart_mangle_sv)
    [+] (<try> write_fail_readonly ind inst cnt <catch> K restart_mangle_sv)"
  by(intro ext equalityI write_f_alt_L write_f_alt_R)

definition ll_read_alt where
  "ll_read_alt ind inst = 
    ll_read_success ind inst 
    [+] ll_read_failure ind inst 
    [+] ll_read_corrupt ind inst
    [+] restart_mangle_sv
    [+] do { ll_read_success ind inst; restart_mangle_sv }
    [+] (<try> ll_read_failure ind inst <catch> K restart_mangle_sv)
    [+] do { ll_read_corrupt ind inst; restart_mangle_sv }
  "

lemma ll_read_alt_L:
  "ll_read ind inst s \<subseteq> (
    ll_read_success ind inst 
    [+] ll_read_failure ind inst 
    [+] ll_read_corrupt ind inst
    [+] restart_mangle_sv
    [+] do { ll_read_success ind inst; restart_mangle_sv }
    [+] (<try> ll_read_failure ind inst <catch> K restart_mangle_sv)
    [+] do { ll_read_corrupt ind inst; restart_mangle_sv }
    ) s
  "
  apply(fold ll_read_alt_def)
  apply(auto simp only: ll_read_def choice_distrib monad_laws)
  apply(auto simp only: choice_def)
  apply(clarsimp_all simp add: brute_force)
  apply(clarsimp_all simp add: ll_read_alt_def brute_force HOL.conj_disj_distribR
    insert_commute fun_upd_twist)
  apply(metis)
  done
  
lemma ll_read_alt_R:
  "(
    ll_read_success ind inst 
    [+] ll_read_failure ind inst 
    [+] ll_read_corrupt ind inst
    [+] restart_mangle_sv
    [+] do { ll_read_success ind inst; restart_mangle_sv }
    [+] (<try> ll_read_failure ind inst <catch> K restart_mangle_sv)
    [+] do { ll_read_corrupt ind inst; restart_mangle_sv }
    ) s
    \<subseteq>
    ll_read ind inst s
  "
  apply(auto simp only: choice_distrib monad_laws)
  apply(auto simp only: choice_def)
  apply(unfold ll_read_def)
  apply(simp_all only: choice_distrib monad_laws)
  apply(clarsimp_all simp add: choice_def)
  apply(clarsimp_all simp add: brute_force HOL.conj_disj_distribR insert_commute)
  apply(metis)
  done

lemma ll_read_alt:
  "ll_read ind inst = 
    ll_read_success ind inst 
    [+] ll_read_failure ind inst 
    [+] ll_read_corrupt ind inst
    [+] restart_mangle_sv
    [+] do { ll_read_success ind inst; restart_mangle_sv }
    [+] (<try> ll_read_failure ind inst <catch> K restart_mangle_sv)
    [+] do { ll_read_corrupt ind inst; restart_mangle_sv }
  "

  by(intro ext equalityI ll_read_alt_L ll_read_alt_R)


lemma ll_read_simps:
  "((ll_read_failure ind inst)\<Down>NN) = empty"
  "((ll_read_failure ind inst)\<Down>R) = empty"
  "((ll_read_success ind inst)\<Down>NE) = empty"
  "((ll_read_success ind inst)\<Down>R) = empty"
  "((ll_read_corrupt ind inst)\<Down>NE) = empty"
  "((ll_read_corrupt ind inst)\<Down>R) = empty"
  by(auto simp add: project_defs unfold_monad ll_read_defs split_if prod_case_beta)

end
