fun find_subterms ctxt t =     
    Monad_Defs.get_term_def ctxt t
                            |> map_opt (fn thm => Term.add_consts (Thm.rhs_of thm |> term_of) [])
                            |> map_opt (map Const)
                            
fun filter_monad_defs ctxt extra = let
    val test1 = (Monad_Defs.get_term_def ctxt #> is_some)
    fun test2 t = member (op =) extra (const_name t)
in
    filter (fn t => test1 t orelse test2 t)
end
                                    
fun call_graph ctxt extra t = let
    fun get_call_graph' [] = []
      | get_call_graph' (t::ts) = 
        find_subterms ctxt t 
                      |> (flat o the_list)
                      |> filter_monad_defs ctxt extra
                      |> (fn subts => 
                             (const_name t, map const_name subts) 
                             :: get_call_graph' (distinct (op =) (ts @ subts))
                         )
in
    get_call_graph' [t]
end

fun remove_th_name s = space_explode "." s |> flip nth 1
                                     
fun call_graph_stripped ctxt extra t = 
    call_graph ctxt extra t 
               |> map (fn (t, tl) => (remove_th_name t, map remove_th_name tl))
               
fun call_graph_dot ctxt extra t = let
    val data = call_graph_stripped ctxt extra t
    val edges = map (fn (t, ts) => map (fn t' => t ^ " -> " ^ t') ts) data 
                    |> flat
                    |> distinct (op =)
    val wrap_final = enclose "digraph {\n" "\n}"
in
    space_implode ";\n" edges
                  |> wrap_final
end
                            
fun write_graph f_name ctxt extra t = let
    val outs = TextIO.openOut f_name
in 
    TextIO.output (outs, call_graph_dot ctxt extra t);
    TextIO.closeOut outs
end
    
    
