signature MONAD_DEFS =
sig
  val get: Proof.context -> thm list
  val get_term_def: Proof.context -> term -> thm option
  val add_thm: thm -> Context.generic -> Context.generic
  val del_thm: thm -> Context.generic -> Context.generic
  val default_simp_thms: thm list (* theorems which are used to preprocess the monad definitions *)
  val add: attribute
  val del: attribute
end

structure Monad_Defs : MONAD_DEFS =
struct

structure MonadDefs = Generic_Data
(
  type T = (term * thm) Item_Net.T;
  val empty : T = Item_Net.init (op aconv o pairself fst) (single o fst);
  val extend = I;
  fun merge tabs : T = Item_Net.merge tabs;
)

val find_fun_name_meta_eq = head_of o term_of o Thm.lhs_of;

val obj_lhs_of = fst o HOLogic.dest_eq o HOLogic.dest_Trueprop o prop_of;
val find_fun_name_obj_eq = head_of o obj_lhs_of;

val default_simp_thms = [@{thm K_def}, @{thm Let_def}, @{thm Product_Type.split}]
(*
val prep_thm = Simplifier.rewrite_rule default_simp_thms
 *)
val prep_thm = Simplifier.simplify (HOL_basic_ss addsimps default_simp_thms)

fun thm_to_item thm =
    (find_fun_name_meta_eq thm, prep_thm thm)
    handle TERM _ => (find_fun_name_obj_eq thm, thm RS eq_reflection |> prep_thm);

(*
fun thm_to_item thm =
    (Thm.lhs_of thm |> term_of, thm)
    handle TERM _ => (obj_lhs_of thm, thm RS eq_reflection);
    *)

val add_thm = MonadDefs.map o Item_Net.update o thm_to_item;
val del_thm = MonadDefs.map o Item_Net.remove o thm_to_item;

val all_thms = map snd o Item_Net.content o MonadDefs.get;

val get = all_thms o Context.Proof;

val get_full = MonadDefs.get o Context.Proof;

val add = Thm.declaration_attribute add_thm;
val del = Thm.declaration_attribute del_thm;

fun get_term_def ctxt t =
    (Item_Net.retrieve (get_full ctxt) t
    |> the_single |> snd |> SOME) handle Empty => NONE;

end
