(*

The traditional hodgepodge of somewhat useful functions, made globally
available.

The first half of the functions is ripped straight from the Isabelle
programming tutorial.

*)

infix bind_opt apply_opt

signature PMM_LIBRARY = sig

    val pretty_term: Proof.context -> term -> Pretty.T
    val pwriteln: Pretty.T -> unit
    val pretty_terms: Proof.context -> term list -> Pretty.T
    val pretty_cterm: Proof.context -> cterm -> Pretty.T
    val pretty_cterms: Proof.context -> cterm list -> Pretty.T
    val pretty_thm_no_vars: Proof.context -> thm -> Pretty.T
    val pretty_typ: Proof.context -> typ -> Pretty.T
    val pretty_typs: Proof.context -> typ list -> Pretty.T
    val pretty_thm: Proof.context -> thm -> Pretty.T
    val pretty_thms: Proof.context -> thm list -> Pretty.T
    val pretty_thms_no_vars: Proof.context -> thm list -> Pretty.T
    val pretty_tenv: Proof.context -> Envir.tenv -> Pretty.T 
    val pretty_tyenv: Proof.context -> Type.tyenv -> Pretty.T 

    val pprint: Pretty.T -> unit

    val split_hd: 'a list -> 'a * 'a list
    val bind_opt: 'a option * ('a -> 'b option) -> 'b option
    val map_opt: ('a -> 'b) -> 'a option -> 'b option
    val apply_opt: ('a -> 'b) option * 'a option -> 'b option
    val flip: ('a -> 'b -> 'c) -> 'b -> 'a -> 'c
    val merge_eq: ''a list * ''a list -> ''a list
    val all_pairs: 'a list -> 'b list -> ('a * 'b) list
    val take_somes: 'a option list -> 'a list
    val last: 'a list -> 'a
    val divide_list_by: ('a -> bool) -> 'a list -> 'a list * 'a list

    val is_comb: term -> bool
    val is_Abs: term -> bool
    val get_arg: int -> term -> term 
    val get_fun_term: term -> term
    val const_name: term -> string 
    val term_constr: term -> string
    val strip_all: term -> (string * typ) list * term
    val apply_typed: Proof.context -> term -> term -> term
    val apply_typed': Proof.context -> term -> term -> term * (typ -> typ)
    val apply_typed_list: Proof.context -> term * term list -> term
    val eta_expand: term -> term
    val mk_abs_free: term -> term list * term
    val mk_abs_var: term -> term list * term
    val dest_binop: term -> term * term

end

structure Pmm_Library : PMM_LIBRARY =
struct

val pretty_term = Syntax.pretty_term

val pwriteln = Pretty.writeln

fun pretty_terms ctxt trms =
  Pretty.block (Pretty.commas (map (pretty_term ctxt) trms)) 

fun pretty_cterm ctxt ctrm =
  pretty_term ctxt (term_of ctrm)

fun pretty_cterms ctxt ctrms =
  Pretty.block (Pretty.commas (map (pretty_cterm ctxt) ctrms))

fun pretty_thm_no_vars ctxt thm =
  let
      val ctxt' = Config.put show_question_marks false ctxt
  in
      pretty_term ctxt' (prop_of thm)
  end


val pretty_typ = Syntax.pretty_typ
fun pretty_typs ctxt tys =
  Pretty.block (Pretty.commas (map (pretty_typ ctxt) tys))

fun pretty_thm ctxt thm =
  pretty_term ctxt (prop_of thm)

fun pretty_thms ctxt thms =
  Pretty.block (Pretty.commas (map (pretty_thm ctxt) thms))

fun pretty_thms_no_vars ctxt thms =
  Pretty.block (Pretty.commas (map (pretty_thm_no_vars ctxt) thms))

fun pretty_helper aux env =
  env   |> Vartab.dest
        |> map aux
        |> map (fn (s1, s2) => Pretty.block [s1, Pretty.str " := ", s2])
        |> Pretty.enum "," "[" "]"

fun pretty_tenv ctxt env = let
  fun get_trms (v, (T, t)) = (Var (v, T), t)
  val print = pairself (pretty_term ctxt)
in
  pretty_helper (print o get_trms) env
end

fun pretty_tyenv ctxt tyenv = let
    fun get_typ (x, (S, ty)) = (TVar (x, S), ty)
    val print = pairself (pretty_typ ctxt)
in
    pretty_helper (print o get_typ) tyenv
end


(* Homebrew stuff *)

(* Pretty printing *)
fun pprint prt = tracing (Pretty.string_of prt)


(* Functions on ML types such as option, list, pair *)
                 
fun NONE bind_opt _        = NONE
  | (SOME x) bind_opt f    = f x

fun map_opt f (SOME x) = SOME (f x)
  | map_opt _ _ = NONE

fun fo apply_opt  xo = fo bind_opt (fn f => xo bind_opt (SOME o f))

fun last [x] = x
|   last (_::xs) = last xs
  
fun split_hd (x::xs) = (x,xs)

fun flip f a b = f b a

fun merge_eq l = merge (fn (x, y) => x = y) l

fun all_pairs [] _ = []
  |   all_pairs (x :: xs) ys = map (pair x) ys @ all_pairs xs ys

fun take_somes l = let
    fun aux (SOME x :: xs, ss) = aux (xs, ss o (cons x))
      | aux (NONE :: xs, ss) = aux (xs, ss)
      | aux ([], ss) = ss []
in
    aux (l, I)
end

(* The first resulting list contains elements satisfying the predicate,
 the second one those not satisfying the predicate
 *)
fun divide_list_by p = let
    fun aux (yess, nos) _ [] = (yess [], nos [])
      | aux (yess, nos) p (x::xs) = if p x
                                    then aux (yess o cons x, nos) p xs
                                    else aux (yess, nos o cons x) p xs
in
    aux (I, I) p
end
                   
(* Functions on Isabelle terms *)             

fun is_comb (_ $ _) = true
  | is_comb _       = false

fun is_Abs (Abs _) = true
  | is_Abs _ = false

fun get_arg 0 (_ $ r) = r
  | get_arg n (l $ _) = get_arg (n-1) l

fun get_fun_term (l $ _) = get_fun_term l
  | get_fun_term t = t

val const_name = fst o Term.dest_Const

fun term_constr (Const _) = "Const"
  | term_constr (Free _) = "Free"
  | term_constr (Var _) = "Var"
  | term_constr (Bound _) = "Bound"
  | term_constr (Abs _) = "Abs" 
  | term_constr (_ $ _) = "App"

fun strip_all (t : term) : ((string * typ) list * term) =
    if Logic.is_all t
    then Logic.dest_all t ||> strip_all |> (fn (v, (vs,t')) => (v::vs, t'))
    else ([], t)
 
(* Creates a term t $ t', instantiating the polymorphic types in t appropriately *)
fun apply_typed ctxt t t' =
let
  val thy = Proof_Context.theory_of ctxt

  fun mk_subst pat t = 
  let
      val subst' = Sign.typ_match thy (pat, t) Vartab.empty
  in
      Envir.subst_type subst'
  end
  
  val arg_type = Term.type_of t
                  |> binder_types
                  |> hd
  val subst = mk_subst arg_type (type_of t')
in
  (map_types subst t) $ t'
end

(* Creates a term t $ t', instantiating the polymorphic types in both
 t and t' appropriately *)
fun apply_typed' ctxt t t' =
let
  val thy = Proof_Context.theory_of ctxt

  val maxidx = maxidx_of_typs [type_of t, type_of t']
  fun mk_unif T T' = 
  let
      val subst' = Sign.typ_unify thy (T, T') (Vartab.empty, maxidx) |> fst
  in
      Envir.subst_type subst'
  end
  
  val arg_type = Term.type_of t
                  |> binder_types
                  |> hd
  val unif = mk_unif arg_type (type_of t')
in
  (map_types unif (t $ t'), unif)
end


(* Like list_comb, but takes types into account *)
fun apply_typed_list ctxt = Library.foldl (uncurry (apply_typed ctxt))

fun eta_expand t = let
    val fvs = Term.variant_frees t [("x", fastype_of t |> domain_type)]
                                 |> map Free

    val t' = list_comb (t, fvs)
in
    Term.lambda (hd fvs) t'
end

fun mk_abs_free t = let
    val (frees, body) = strip_abs t |>> map Free
in
    (frees, subst_bounds (rev frees, body))
end

fun mk_abs_var t = let
    val (vars, body) = strip_abs t |>> map (Var o (fn (s, T) => ((s, 0), T)))
in
    (vars, subst_bounds (rev vars, body))
end


fun list_to_pair [x, y] = (x, y)
val dest_binop = list_to_pair o snd o strip_comb

val term_pat_setup =
let
  fun term_pat (ctxt, str) =
      str 
      |> Proof_Context.read_term_pattern ctxt
      |> ML_Syntax.print_term
      |> ML_Syntax.atomic

  val parser = Args.context -- Scan.lift Args.name_source
in
  ML_Antiquote.inline @{binding "term_pat"} (parser >> term_pat)
end

val _ = (Context.>> o Context.map_theory) term_pat_setup

val type_pat_setup =
let
  val parser = Args.context -- Scan.lift Args.name_source
  
  fun typ_pat (ctxt, str) =
  let
      val ctxt' = Proof_Context.set_mode Proof_Context.mode_schematic ctxt
  in
      str
      |> Syntax.read_typ ctxt'
      |> ML_Syntax.print_typ
      |> ML_Syntax.atomic
  end
in
  ML_Antiquote.inline @{binding "typ_pat"} (parser >> typ_pat)
end

val _ = (Context.>> o Context.map_theory) type_pat_setup

end

open Pmm_Library;
