infix 0 RSf; 

signature REF_TAC =
sig
    type prem_simp = (Proof.context -> thm list -> thm list)
    val refinement_tac: bool ->  prem_simp ->
                        Proof.context -> int -> tactic
    val refinement_noendsimp_tac: bool ->  prem_simp ->
                        Proof.context -> int -> tactic
    val refinement_ins_tac: bool -> prem_simp ->
                            Proof.context -> int -> tactic
    val refinement_gen_tac: bool -> prem_simp ->
        (thm list -> tactic) -> Proof.context -> int -> tactic
    val refinement_force_tac: bool -> prem_simp ->
                            Proof.context -> int -> tactic
(*    val refinement_tac': bool -> Proof.context -> int -> tactic *)
    val build_sim_thms: bool -> (Proof.context -> thm list -> thm list) ->
                        Proof.context -> term -> term -> (thm list * Proof.context)
    val default_simp: prem_simp
    val instantiate_redundant_subsets: thm -> thm

    val ref_tac_debug_match: bool Config.T
    val ref_tac_store_non_gen: bool Config.T
    val default_exc_case_decomposer: Proof.context -> term * term -> (term * term) list
    val put_exc_case_decomposer: (Proof.context -> term * term -> (term * term) list) -> Context.generic -> Context.generic
end

structure Ref_Tac : REF_TAC =
struct

type prem_simp = (Proof.context -> thm list -> thm list)

val free_name = fst o dest_Free
val var_name = fst o fst o dest_Var

fun ap_fn_pair_pair (f, g) (x, y) = (f x, g y)

val my_nr_prems = prop_of
                      #> Term.strip_all_body
                      #> Logic.count_prems

fun strip_Alls (Const (@{const_name All}, _) $ Abs (x, T, t)) = strip_Alls t |>> cons (x, T)
  | strip_Alls t = ([], t)

fun is_obj_imp (Const (@{const_name "implies"}, _) $ _ $ _) = true
  | is_obj_imp _ = false

fun is_funtype (Type ("fun", [_, _])) = true
  | is_funtype _ = false;

fun beta_contraction_rule thm = 
    Thm.equal_elim (Thm.beta_conversion true (cprop_of thm)) thm;

val list_elem_typ = dest_Type #> snd #> hd

(* Return the first "working" RS with a theorem from the list *)
fun thm RSf [] = raise THM ("RS_first: no match found for", 1, [thm])
|   thm RSf (thm'::thms) = (thm RS thm') handle THM _ => thm RSf thms

val distinct_prems =
    Tactic.distinct_subgoals_tac #> Seq.pull #> the #> fst

fun focus thm = let
    val ct = cprop_of thm
    val params = strip_all_vars (term_of ct) |> map Free
    val cparams = map (cterm_of (theory_of_cterm ct)) params
    val goal = forall_elim_list cparams thm |> cprop_of
in
    (cparams, Drule.strip_imp_prems goal, Drule.strip_imp_concl goal)
end

(* Create an index permutation which moves elements satisfying the
 predicate to the front (counting from 0) *)
fun perm_by_pred p l = let
    val cnt = filter p l |> length
in
    fold_map (fn x =>
              fn (cur, next_y, next_n) =>
                 if p x
                 then (next_y, (cur + 1, next_y + 1, next_n))
                 else (next_n, (cur + 1, next_y, next_n + 1))
             )
             l
             (0, 0, cnt)
             |> fst
end

(* Find an inverse of a permutation (useful for combining with perm_by_pred
and Drule.rearrange_prems *)
fun inv_perm l = let
    val cnt = length l
    fun aux cpos ps = if cpos < cnt
                      then find_index (equal cpos) ps :: aux (cpos + 1) ps
                      else []
in
    aux 0 l
end
    
val sim_type =
    concl_of
        #> Term.strip_all_body
        #> Logic.strip_imp_concl
        #> HOLogic.dest_Trueprop
        #> head_of
        #> const_name

val sim_ext_name = @{const_name "sim_ext"}

fun is_monad_binop (Const (@{const_name bind}, _)) = true
  | is_monad_binop (Const (@{const_name bindN}, _)) = true
  | is_monad_binop (Const (@{const_name try_catch}, _)) = true
  | is_monad_binop (Const (@{const_name try_catchR}, _)) = true
  | is_monad_binop _ = false

val same_const = (op =) o pairself const_name

fun RLN_full _ [] = []
  | RLN_full thas thbs = let
        val max_prems = map Thm.nprems_of thbs |> foldl1 Int.max
        val all_pos = 1 upto max_prems
        fun add_thms thms1 thms2 = fold (Thm.add_thm) thms1 thms2
    in
        fold (fn i => fn thms => add_thms (thas RLN (i, thbs)) thms) all_pos []
    end

fun get_exc_cases ctxt t = let

    val exception_tyname = @{type_name "exceptions"}
    val exception_constrs =
        Datatype_Data.get_constrs (Proof_Context.theory_of ctxt) exception_tyname
                                  |> the
                                  |> map Term.Const
in
    strip_comb t
               |> snd
               |> take (length exception_constrs)
end

fun default_exc_case_decomposer ctxt (ta, tc) =
    (get_exc_cases ctxt ta) ~~ (get_exc_cases ctxt tc)


structure Data = Generic_Data (
    type T = Proof.context -> (term * term) -> (term * term) list
    val empty = default_exc_case_decomposer
    val extend = I
    val merge = snd
)

val put_exc_case_decomposer = Data.put

fun sim_prems_to_front thm = let
    fun is_sim_prem t = let
        val strip_obj_All = let
            fun rem_Abs (Abs (x, T, t)) = (x, T, t)
            fun aux ((Const (@{const_name "All"}, _)) $ t) = rem_Abs t |> Term.dest_abs |> snd |> aux
              |   aux u = u
        in
            aux
        end
        val sims = [@{const_name "sim"}, @{const_name "sim_norm"}, @{const_name "sim_ext"},
                    @{const_name "sim_norm_ext"}]
    in
        (t |> HOLogic.dest_Trueprop |> strip_obj_All |> head_of |> const_name
           |> member (op =) sims
        ) handle _ => false
    end
in
    thm
        |> prems_of
        |> perm_by_pred is_sim_prem
        |> inv_perm
        |> flip Drule.rearrange_prems thm
end

fun compose_binop_thms trace ctxt binop thms1 thms2 = let
    val cond_tracing = if trace then tracing else K ()
    val cond_pprint = if trace then pprint else K ()
    val _ = cond_tracing "Composing:";
    val _ = cond_pprint (pretty_thms ctxt thms1);
    val _ = cond_pprint (pretty_thms ctxt thms2);

    fun aux @{const_name "bind"} = let
        val _ = cond_tracing "...over bind";
        val _ = if null thms1
                then
                    cond_tracing "compose_binop_thms failed: thms1 are null"
                else ()
        val s1 = (thms1 RL (Ref_Thms.get_bind_thms ctxt)) |> map sim_prems_to_front

        val _ = cond_tracing "Step 1:";
        val _ = cond_pprint (pretty_thms ctxt s1);

        val _ = if null thms2 then
                    cond_tracing "compose_binop_thms failed: thms2 are null"
                    else ()
      
        val s2 = thms2 RL s1;
        val _ = cond_tracing "Step 2:";
        val _ = cond_pprint (pretty_thms ctxt s2);
    in 
        s2
    end
                                   
      | aux @{const_name "try_catch"} = let
            val _ = cond_tracing "...over try_catch";
            val _ = if null thms1
                    then
                        cond_tracing "compose_binop_thms (try_catch) failed: thms1 are null"
                    else ()

            val s1 = (thms1 RL (Ref_Thms.get_try_catch_thms ctxt)) |> map sim_prems_to_front

            val _ = cond_tracing "Step 1:";
            val _ = cond_pprint (pretty_thms ctxt s1);

            val _ = if null thms2
                    then cond_tracing "compose_binop_thms (try_catch) failed: thms2 are null"
                    else ()

            val s2 = thms2 RL s1;
            val _ = cond_tracing "Step 2:";
            val _ = cond_pprint (pretty_thms ctxt s2);
        in
            s2
        end

      | aux _ = []
in
    aux binop
end

(* Unlike Logic.unvarify_types_global, this shouldn't explode if the
term already contains a TFree type *)
fun safe_unvarify_types_global t = let
    val tvars = Term.add_tvars t []
    fun replace_names ts new_names = (new_names ~~ map (snd o fst) ts) ~~ map snd ts
    fun drop_indices ((n, _), S) = (n, S)
in
    (fold_map Name.variant (map (fst o fst) tvars) (fold_types Term.declare_typ_names t Name.context)
              |> fst)
        |> replace_names tvars
        |> map drop_indices
        |> map TFree
        |> (curry (op ~~)) (map fst tvars)
        |> flip subst_TVars t
end
                   
fun mk_abs_var t = let
    val (vars, body) =
        t
            |> safe_unvarify_types_global
            |> Logic.varify_types_global
            |> strip_abs
            |>> map Free
            |>> map Logic.unvarify_types_global
            |>> map Logic.varify_global
in
    (vars, subst_bounds (rev vars, body))
end

fun variant_TVars vars t = let
    val t_tvars = Term.add_tvars t [] |> map (fst o fst)
    val xs = Name.variant_list t_tvars (map (fst o fst) vars)
in
    (xs ~~ map (snd o fst) vars) ~~ map snd vars
end

(* Find a substitution which maps TVars in t to ones not present in
t' *)
fun variant_TVar_subst trace t t' = let
    val cond_tracing = if trace then tracing else K ()
    val cond_pprint = if trace then pprint else K ()

    val tvars = Term.add_tvars t []
    (*
     val _ = cond_tracing "tvars are";
     val _ = cond_tracing (PolyML.makestring tvars);

     val _ = cond_tracing "target term is:";
     val _ = cond_tracing (PolyML.makestring t');
     *)

    val new_tvars = variant_TVars tvars t' |> map TVar
(*
 val _ = cond_tracing "new_tvars are";
 val _ = cond_tracing (PolyML.makestring new_tvars);
 *)
in
    subst_TVars (map fst tvars ~~ new_tvars)
end

fun make_terms (tenv : Envir.tenv) = let
    fun aux (x, (T, t)) = (Var (x, T), t)
in
    map aux (Vartab.dest tenv)
end
                                     
fun make_types (tyenv : Type.tyenv) = let
    fun aux (x, (S, ty)) = (TVar (x, S), ty)
in
    map aux (Vartab.dest tyenv)
end
                                      

local
fun prep_trm thy (x, (T, t)) = let
    val t' = cterm_of thy t
in
    (cterm_of thy (Var (x, T)), t')
end

fun prep_ty thy (x, (S, ty)) =
    (ctyp_of thy (TVar (x, S)), ctyp_of thy ty)

in
fun produce_inst_param thy (tyenv, tenv) = let
    val tyenv' = Vartab.dest tyenv
    val tenv' = Vartab.dest tenv
in
    (map (prep_ty thy) tyenv', map (prep_trm thy) tenv')
end
end;

fun dest_env env = (Envir.type_env env, Envir.term_env env)

fun ho_matcher_inst i thy pat trm = let
    val univ = Unify.matchers thy [(pat, trm)]
    val env = nth (Seq.list_of univ) i
in
    produce_inst_param thy (dest_env env)
end

fun ho_unify_inst i thy pat trm = let
    val init = Envir.empty 0
    val univ = Unify.unifiers (thy, init, [(pat, trm)])
    val env = nth (Seq.list_of univ) i |> fst
in
    produce_inst_param thy (dest_env env)
end

fun make_env_simple (tyenv : Type.tyenv, tenv : Envir.tenv) = let
    val maximum = Library.foldl1 Int.max

    val max_tenv = 
        make_terms tenv
                   |> map (Int.max o pairself maxidx_of_term)
                   |> maximum

    val max_tyenv =
        make_types tyenv
                   |> map (Int.max o pairself maxidx_of_typ)
                   |> maximum
        
    val maxidx = Int.max (max_tenv, max_tyenv)
in
    Envir.Envir {maxidx=maxidx, tenv=tenv, tyenv=tyenv}
end
                                                                  
local
    fun simpl_variant_var ts names = let
        val vars = fold Term.add_vars ts []
    in
        Name.variant_list (map (fst o fst) vars) names
    end

    fun make_fresh_vars (names_types : (string * typ) list) ts = let
        val new_fnames = simpl_variant_var ts (map fst names_types)
        val new_indnames = map (fn x => (x, 0)) new_fnames
    in
        map Var (new_indnames ~~ map snd names_types)
    end
in

fun varify_indnames ts (indnames_types : (indexname * typ) list) = let
    val vars = make_fresh_vars (indnames_types |> map (apfst fst)) ts
in
    (indnames_types ~~ vars)
end


fun varify_indnames_subst ts indnames_types =
    varify_indnames ts indnames_types
                    |> map (apfst fst)
                    |> subst_Vars

fun varify_fvars_subst ctxt names_types ts = let
    val vars = make_fresh_vars names_types ts
in
    subst_free (map Free names_types ~~ vars)
end

end;


(* Find the match which can instantiate the given simulation theorem
 to the given monad operations *)
fun match_sim_thm_with matcher ctxt thm ma mc = let
    val (ma', mc') = Ref_Thms.extract_sim_targets_thm thm
    val both_conc = HOLogic.mk_prod (ma', mc')
    val both_ts = HOLogic.mk_prod (ma, mc)
in
    matcher ctxt both_conc both_ts
end

val (ref_tac_debug_match, ref_tac_debug_setup) = 
    Attrib.config_bool @{binding "ref_tac_debug_match"} (K false)
val _ = (Context.>> o Context.map_theory) ref_tac_debug_setup

val (ref_tac_store_non_gen, ref_tac_store_non_gen_setup) = 
    Attrib.config_bool @{binding "ref_tac_store_non_gen"} (K false)
val _ = (Context.>> o Context.map_theory) ref_tac_store_non_gen_setup


fun match ctxt pat obj = let
    val thy = Proof_Context.theory_of ctxt
in 
    (
     Thm.first_order_match (cterm_of thy pat, cterm_of thy obj)
    )
    handle Pattern.MATCH => (
           if Config.get ctxt ref_tac_debug_match
           then (
             tracing "Match failed; debugging";
             tracing "Matching:";
             pprint (pretty_terms ctxt [pat, obj]);
             Debug_Match.first_order_match ctxt thy (pat, obj) (Vartab.empty, Vartab.empty);
             tracing "Survived?"
               )
           else ();
           raise Pattern.MATCH
    )
end

val match_sim_thm = match_sim_thm_with match

local 

    fun remove_dummy_names get_name = let
        val is_dummy_name = try Name.dest_internal #> is_some
    in
        filter (not o is_dummy_name o get_name)
    end
                             
in

exception NO_ABS;

fun inst_sim_thm ctxt thm ma mc =
    match_sim_thm ctxt thm ma mc
                  |> flip instantiate_normalize thm


fun abs_over trace ctxt frees ma mc thm = let
    val cond_tracing = if trace then tracing else K ()
    val cond_pprint = if trace then pprint else K ()
in
    if not (null frees)
    then (
        match_sim_thm ctxt thm ma mc
                      |> flip Drule.instantiate_normalize thm
                      |> (fn thm' =>
                             forall_intr_list
                                 (map (cterm_of (theory_of_thm thm')) (rev frees))
                                 thm'
                         )
        ) handle Pattern.MATCH => (
                 cond_tracing "abs_over: Match failed";
                 cond_tracing "thm:";
                 cond_pprint (pretty_thm ctxt thm);
                 cond_tracing "terms:";
                 cond_pprint (pretty_terms ctxt [ma, mc]);
                 raise NO_ABS
                 )
    else raise NO_ABS
end
end;

fun fix_thm_vars thm = forall_intr_vars thm
                                        |> ` focus
                                        |> (fn ((cparams, _, _), thm') =>
                                               forall_elim_list cparams thm')
    
(* Caution: do not use with theorems which do NOT contain at least one meta-level connective;
   otherwise, the simplifier loops forever
   TODO: no simplifier here, does this still hold? :)
*)
val atomize = Conv.fconv_rule Object_Logic.atomize

val atomize_prems = Conv.fconv_rule Object_Logic.atomize_prems

fun pull_imp_out_safe thm = let
    fun pull thm = thm RSf [@{thm All_impE'}, @{thm All2_impE'}]

    fun aux (unv_thm, thm) = let
        (*
        val _ = tracing "pull_imp_out_safe/aux: current theorem";
        val _ = pprint (pretty_thm @{context} thm)
         *)
    in
        ((unv_thm, thm)
             |> pairself pull
             |> aux
    ) handle THM _ => (unv_thm, thm)
    end
in
    aux (fix_thm_vars thm, thm)
        |> snd
end

fun compose_exc_thms trace ctxt assm_thms = let
    val cond_tracing = if trace then tracing else K ()
    val cond_pprint = if trace then pprint else K ()

    val _ = if exists null assm_thms
            then
                (cond_tracing "compose_exc_thms: exists a null thm";
                 cond_tracing "with index: ";
                 cond_tracing (find_index null assm_thms |> PolyML.makestring)
                 )
            else (
                cond_tracing "compose_exc_thms: composing";
                fold (fn thms => fn i => (
                         cond_tracing (PolyML.makestring i);
                         cond_pprint (pretty_thms ctxt thms);
                         i+1
                         )
                     ) assm_thms 1;
                ()
                )
    (* val _ = cond_pprint (Pretty.block (map (pretty_thms ctxt) assm_thms)) *)
    val nr_thms = length assm_thms
    val numbered = rev (1 upto nr_thms ~~ assm_thms)
    val sims_excs = Ref_Thms.get_exc_case_thms ctxt
                    (*
                    @{thms sims_exceptions_caseI
                          sim_ext_exceptions_caseI
                          sim_norm_ext_exceptions_caseI
                          sim_norm_exceptions_case_bindS_I
                     }
                     *)
in
    if null assm_thms orelse exists null assm_thms
    then []
    else Library.foldl (fn (thms, (i, assms)) => assms RLN (i, thms)) 
                       (sims_excs, numbered)
         |> tap (K (cond_tracing "compose_exc: resulting theorems"))
         |> tap (pretty_thms ctxt #> cond_pprint)
end

local

  fun same_monad_ops t1 t2 =
    forall is_monad_binop [t1, t2]
    andalso same_const (t1, t2)

  fun is_exceptions_case t = (head_of t |> const_name |> equal @{const_name exceptions_case})
      handle (TERM _) => false
  fun dest_if (Const (@{const_name "If"}, _) $ c $ tb $ fb) = (c, tb, fb)
  fun is_if t = (head_of t |> const_name |> equal @{const_name "If"})
      handle (TERM _) => false

  fun is_choice t = (head_of t |> const_name |> equal @{const_name "choice"})
      handle (TERM _) => false

  fun is_mapM t = (head_of t |> const_name |> equal @{const_name "mapM'"}) handle (TERM _) => false

  fun abs_body (Abs (_, _, t)) = t
    | abs_body t = raise (TERM ("Not an Abs", [t]))

  fun is_lambda_bindS t = (abs_body t |> head_of |> const_name |> equal @{const_name "bindS"})
      handle (TERM _) => false

  fun sim_fold_defs ctxt = Ref_Thms.get_sim_defs ctxt |> map Thm.symmetric
                      (*  @{thms sims_defs}  *)
  
in

fun build_sim_thms trace simp ctxt ma mc = let
    val cond_tracing = if trace then tracing else K ()
    val cond_pprint = if trace then pprint else K ()

    val _ = (cond_tracing "Searching for refinement for";
             (* pprint (pretty_term ctxt ma); *)
             cond_pprint (pretty_terms ctxt [ma, mc]))


    fun distinct_thms thms = fold Thm.add_thm thms []

     (* A heuristic: given two theorems with equivalent conclusions,
      keep only the one with fewer premises *)

    fun filter_by_prem_nrs thms = let
        fun eq_concls (th, th') =
            Pattern.matches (theory_of_thm th) (concl_of th, concl_of th')
            andalso Pattern.matches (theory_of_thm th') (concl_of th', concl_of th)
    in
        thms
            |> tap (K (cond_tracing "filter_by_prem_nrs: before partition"))
            |> tap (cond_pprint o pretty_thms ctxt)
            |> partition_eq (eq_concls)
            (*
            |> tap (K (cond_tracing "filter_by_prem_nrs: after partition"))
            |> tap (cond_pprint o pretty_thms ctxt o flat)
            *)
            |> map (sort_distinct (int_ord o pairself my_nr_prems))
            |> map hd
            |> tap (K (cond_tracing "filter_by_prem_nrs: result"))
            |> tap (cond_pprint o pretty_thms ctxt)
    end

    fun finalize_thms (thms, ctxt) =
        (simp ctxt thms
              |> map (distinct_prems o beta_contraction_rule o Drule.eta_contraction_rule)
              |> map (Raw_Simplifier.rewrite_rule (sim_fold_defs ctxt))
              |> filter (fn thm =>
                            not (member (op =) (prems_of thm)
                                        (HOLogic.mk_Trueprop @{const "False"})
                                       )
                        )
              |> distinct_thms
              |> filter_by_prem_nrs
       , ctxt)

    fun store_thms thms ctxt = (
        cond_tracing "Storing thms:";
        cond_pprint (pretty_thms ctxt thms);
        Context.proof_map (Ref_Thms.add_thms thms) ctxt
    )

    fun process_fun' ctxt ma mc adjust_ma adjust_mc = let
        fun expand t = (not (is_Abs t)? eta_expand) t

        val _ = cond_tracing "process_fun; doing: "
        val _ = cond_pprint (pretty_terms ctxt [ma, mc])

        val (a_frees, ma') = mk_abs_free (expand ma)
        val (c_frees, mc') = mk_abs_free (expand mc)
                              
        val all_frees = merge_eq (a_frees, c_frees)
 
        val free_subst = varify_fvars_subst ctxt (map dest_Free all_frees)
                                            [ma', mc']

        val (ma'', mc'') = pairself free_subst (ma', mc')

        fun do_abs ctxt thm = (
            abs_over trace ctxt all_frees (adjust_ma ma') (adjust_mc mc') thm 
        ) handle NO_ABS => thm
    in
        build_sim_thms trace simp ctxt (adjust_ma ma'') (adjust_mc mc'')
                       |> (fn (thms, ctxt') => (map (do_abs ctxt') thms, ctxt'))
                       |> finalize_thms
                       |>> tap (K (cond_tracing "process_fun' result:"))
                       |>> tap (cond_pprint o pretty_thms ctxt)
    end

    fun process_fun ctxt ma mc =
        process_fun' ctxt ma mc I I

    fun process_fun_push_prems' ctxt ma mc adjust_ma adjust_mc = let
        fun push_prems_into_rel thm = let
            val concl =
                concl_of thm
                         |> strip_all_body
                         |> Logic.strip_imp_concl
                         |> HOLogic.dest_Trueprop
                         |> strip_Alls
                         |> snd

        in
            if is_obj_imp concl
            then
                thm
                    |> tap (K (cond_tracing "push_prems_into_rel: current theorem"))
                    |> tap (fn thm => (cond_pprint (pretty_thm ctxt thm)))
                    |> Simplifier.simplify (HOL_basic_ss addsimps (
                                            [(@{thm imp_conjL} RS eq_reflection)
                                                 |> Thm.symmetric] 
                                            @ @{thms push_sim_exts_premise}
                                            @ Ref_Thms.get_sim_defs ctxt (* @{thms sims_defs} *)
                                           ))
                    |> tap (K (cond_tracing "push_prems_into_rel: after rewrite"))
                    |> tap (fn thm => (cond_pprint (pretty_thm ctxt thm)))

            else (
                cond_tracing "push_prems_into_rel: not an obj_imp:";
                cond_pprint (pretty_term ctxt concl);
                thm
                )
        end

        fun reorder_prems ctxt thm = let
            val (cparams, cprems, _) = focus thm
            val params = map term_of cparams |> map dest_Free

            fun refers_to_params ct =
                ct |> term_of |> flip Term.add_frees [] |> exists (member (op =) params)

            val thm_no_prems = fold (fn ct => fn thm' => Thm.forall_elim ct thm') cparams thm
        in
            cprems
                |> perm_by_pred (not o refers_to_params)
                |> inv_perm
                (*
                 |> tap (K (cond_tracing "reorder_prems: permuting"))
                    |> tap (K (cond_pprint (pretty_thm ctxt thm)))
                    |> tap (fn perm => (cond_tracing (PolyML.makestring perm)))
                    |> tap (K (cond_tracing "reorder_prems: thm_no_prems"))
                    |> tap (K (cond_pprint (pretty_thm ctxt thm_no_prems)))
                 *)
                |> flip Drule.rearrange_prems thm_no_prems
                |> Drule.forall_intr_list cparams
        end
    in
        process_fun' ctxt ma mc adjust_ma adjust_mc
                    |>> map (reorder_prems ctxt)
                    |>> map atomize
                    |>> map pull_imp_out_safe
                    |>> map push_prems_into_rel
    end

    fun process_fun_push_prems ctxt ma mc = process_fun_push_prems' ctxt ma mc I I

    fun process_same_binop ma mc = 
      let
          val _ = cond_tracing "Same monad op";
          val (ma1, ma2) = dest_binop ma
          val (mc1, mc2) = dest_binop mc

          val (thms1, ctxt1) = build_sim_thms trace simp ctxt ma1 mc1
          val op_name = const_name (head_of ma)
     (*
          val push_thm = @{thms push_sim_ext_premise}
      *)

          (*
          fun push_sim_ext_prems thm = case sim_type thm of
                           sim_ext_name =>
                           ((
                            cond_tracing "push_sim_ext: pushing";
                            atomize_prems
                                thm
                                |> Raw_Simplifier.rewrite_rule push_thm
                                |> tap (K (cond_tracing "push_sim_ext_prems: result"))
                                |> tap (cond_pprint o pretty_thm ctxt)
                            ) handle e => (cond_tracing "exception thrown";
                                          cond_tracing (PolyML.makestring e); thm))
                         | _ => thm
           *)
      in
          if null thms1
          then ([], ctxt1)
          else
              process_fun_push_prems ctxt1 ma2 mc2
                          |> (fn (thms2, ctxt') =>
                                 (compose_binop_thms trace ctxt' op_name thms1 thms2, ctxt'))
                          |>> tap (fn thms => if null thms
                                              then
                                                  (cond_tracing "Failed to compose thms for";
                                                   cond_pprint (pretty_terms ctxt [ma1, ma2])
                                                  )
                                              else ()
                                  )

                                                    
      end

     fun process_exceptions_case' ma mc adjust_ma adjust_mc = let
        val _ = cond_tracing "In process_exceptions_case"
        fun no_abs ma mc ctxt = build_sim_thms trace simp ctxt (adjust_ma ma) (adjust_mc mc)
        fun with_abs ma mc ctxt = process_fun_push_prems' ctxt ma mc adjust_ma adjust_mc
        fun choose_fun (ma, mc) = if is_Abs ma then with_abs ma mc else no_abs ma mc
        val exc_decomposer = Data.get (Context.Proof ctxt)
        (*
        val mas = get_exceptions_cases ctxt ma
        val mcs = get_exceptions_cases ctxt mc
        *)
        val cases = exc_decomposer ctxt (ma, mc)
                  
        val tacs = map choose_fun cases

        val (thms, ctxt') = fold_map (fn tac => fn ctxt' =>
            tac ctxt') tacs ctxt

        val _ = if exists null thms
                then
                    (cond_tracing "process_exceptions_case failed: no theorem could be derived for:";
                    find_index null thms
                               |> nth cases
                               |> (fn (ma, mc) => (adjust_ma ma, adjust_mc mc))
                               |> HOLogic.mk_prod
                               |> pretty_term ctxt'
                               |> cond_pprint
                   )
                else ()
                   
        val res_thms = compose_exc_thms trace ctxt' thms
    in
        (res_thms, ctxt')
    end

     
    fun process_exceptions_case ma mc = process_exceptions_case' ma mc I I

    fun process_if' ma mc adjust_ma adjust_mc = let
        val _ = cond_tracing "in process_if";
        val (ca, tba, tfa) = dest_if ma
        val (cc, tbc, tfc) = dest_if mc
        fun pair_to_list (x,y) = [x,y]
        val (t_thms, ctxt') = build_sim_thms trace simp ctxt (adjust_ma tba) (adjust_mc tbc)
        val (f_thms, ctxt'') = build_sim_thms trace simp ctxt' (adjust_ma tfa) (adjust_mc tfc)
        val pairs = all_pairs t_thms f_thms |> map pair_to_list
        fun comb_t thm = try (fn thm' => thm OF thm')
        val if_thms = if (ca aconv cc)
                      then Ref_Thms.get_if_thms ctxt
                      else Ref_Thms.get_if_full_thms ctxt
    in
        let
            
            val res_thms =
                map (fn f => map f pairs) (map comb_t if_thms)
                    |> flat
                    |> take_somes
            val _ = cond_tracing "process_if: result of combining"
            val _ = cond_pprint (pretty_thms ctxt'' t_thms)
            val _ = cond_pprint (pretty_thms ctxt'' f_thms)
            val _ = cond_tracing "process_if: is"
            val _ = cond_pprint (pretty_thms ctxt'' res_thms)
                    
        in
            (res_thms, ctxt'')
        end
    end
        (*
        else let
                val if_thms = Ref_Thms.get_if_full_thms ctxt
                val res_thms =
                    (*
                    map (comb_t @{thm sim_ifI_full}) pairs
                    @ map (comb_t @{thm sim_ext_ifI_full}) pairs
                          *)
                    (*
                    @ map (comb_t @{thm sim_ext_ifI'}) pairs
                    @ map (comb_t @{thm sim_ext_ifI''}) pairs
                     *)
                    (*
                    @ map (comb_t @{thm sim_ext_if_no_cond}) pairs
                     *)
                        |> take_somes
                val _ = cond_tracing "process_if: result of combining"
                val _ = cond_pprint (pretty_thms ctxt'' t_thms)
                val _ = cond_pprint (pretty_thms ctxt'' f_thms)
                val _ = cond_tracing "process_if: is"
                val _ = cond_pprint (pretty_thms ctxt'' res_thms)
            in
                (res_thms, ctxt'')
            end
         *)
                                        
    fun process_if ma mc = process_if' ma mc I I

     (* TODO: this could be done in compose_binop *)
    fun process_choice' ma mc adjust_ma adjust_mc = let
        val _ = cond_tracing "process_choice";

        val (ma1, ma2) = dest_binop ma
        val (mc1, mc2) = dest_binop mc

        fun pair_to_list (x,y) = [x,y]
        val (thms1, ctxt') = build_sim_thms trace simp ctxt (adjust_ma ma1) (adjust_mc mc1)

        val _ = if null thms1
                then
                    cond_tracing "compose_binop_thms (choice) failed: thms1 are null"
                else ()

        val s1 = thms1 RL Ref_Thms.get_choice_thms ctxt
        val _ = cond_tracing "Step 1:";
        val _ = cond_pprint (pretty_thms ctxt s1);

        val (thms2, ctxt'') = build_sim_thms trace simp ctxt' (adjust_ma ma2) (adjust_mc mc2)

        val _ = if null thms2
                then cond_tracing "compose_binop_thms (choice) failed: thms2 are null"
                else ()

        val s2 = RLN_full thms2  s1;
        val _ = cond_tracing "Step 2 (choice):";
        val _ = cond_pprint (pretty_thms ctxt s2);
    in
        (s2, ctxt'')
    end

    fun process_choice ma mc = process_choice' ma mc I I

    fun dest_mapM ma mc = let
        val (ma', la) = dest_binop ma
        val (mc', _) = dest_binop mc
    in
        (ma', mc', la)
    end
        
    fun process_mapM' ctxt ma mc adjust_ma adjust_mc mapM_thms = let
        val thy = Proof_Context.theory_of ctxt

        val (ma', mc', la) = dest_mapM ma mc

        val (thms, ctxt') = process_fun' ctxt ma' mc' adjust_ma adjust_mc
        
        (* Thm.no_prems doesn't "look" under !! *)
        val no_prems = my_nr_prems #> (fn x => x = 0)

        val (wo_prems, w_prems) = divide_list_by no_prems thms
     (*
        val w_prems' = map (forall_intr_list [elem_cterm]) w_prems
      *)

        val _ = cond_tracing "process_mapM': wo_prems"
        val _ = cond_pprint (pretty_thms ctxt wo_prems)

        val _ = cond_tracing "process_mapM': w_prems"
        val _ = cond_pprint (pretty_thms ctxt w_prems)
        
        val res_thms1 = (map atomize wo_prems) RL @{thms All_imp_ball}
        val _ =
            cond_tracing "res_thms1"
        val _ = cond_pprint (pretty_thms ctxt res_thms1)

        fun pull_imps_out thm = let
            val _ = cond_tracing "pulling imps out from:"
            val _ = cond_pprint (pretty_thm ctxt thm)

            fun aux thm' = (
                aux (thm' RSf [@{thm ball_impE'}, @{thm ball_impE}])
            ) handle (THM _) => thm'
        in
            (* caution: |> binds stronger than RS/RSf *)
            ((thm RSf [@{thm ball_decompose_All'}, @{thm ball_decompose_All}]) 
                |> aux
            ) handle (THM _) => thm
        end

        val res_thms2 = map (pull_imps_out o atomize) w_prems
                        
        val _ =
            cond_tracing "res_thms2"
        val _ = cond_pprint (pretty_thms ctxt res_thms2)

        val res = ((res_thms1 @ res_thms2) RL mapM_thms
                 , ctxt')
        val _ = 
            cond_tracing "mapM' res"
        val _ = cond_pprint (pretty_thms ctxt (fst res))
    in
        res
    end

    fun process_mapM ma mc =
        process_mapM' ctxt ma mc I I (Ref_Thms.get_mapM_thms ctxt)
                                      (*@{thms sim_mapM(1) sim_mapM(2) sim_norm_ext_mapM'} *)

    fun get_defs ma mc = let
        val ha = head_of ma
        val hc = head_of mc

        val ma_def = Monad_Defs.get_term_def ctxt ha |> map_opt Drule.eta_contraction_rule
        val mc_def = Monad_Defs.get_term_def ctxt hc |> map_opt Drule.eta_contraction_rule

        val _ = if is_none ma_def orelse is_none mc_def
                then (cond_tracing "Couldn't find the monadic definitions for";
                      cond_pprint (pretty_terms ctxt [ha, hc]))
                else ()

        fun rename_conflicting_vars (thm1, thm2) = let
            fun rename (t1, t2) = 
                (* order important! t1 vars will end up first in the list *)
                fold Term.add_vars [t2, t1] [] 
                     |> partition_eq ((op =) o pairself fst)
                     |> map (distinct ((op =) o pairself snd))
                     |> filter (fn l => length l > 1)
                     |> tap (K (cond_tracing "rename_conflicting_vars: conflicting"))
                     |> tap (fn tll => flat tll |> map Var |> cond_pprint o pretty_terms ctxt)
                     |> map (flip nth 1) (* take the t2 variables *)
                     |> varify_indnames [t1, t2]
                     |> map (apfst Var)
                     |> map (pairself (cterm_of (theory_of_thm thm2)))
        in
            rename (prop_of thm1, prop_of thm2)
                   |> flip cterm_instantiate thm2
                   |> pair thm1
        end

        val term_defs = (map_opt pair ma_def apply_opt mc_def)
                            |> map_opt rename_conflicting_vars

    in
        term_defs
    end

    fun rewr_thm_folding_defs ma_def mc_def adjust_ma adjust_mc thm = let
        val ma_rhs = Thm.rhs_of ma_def |> term_of |> adjust_ma
        val mc_rhs = Thm.rhs_of mc_def |> term_of |> adjust_mc

        val (ma_rhs', mc_rhs') = Ref_Thms.extract_sim_targets_thm thm

        val (inst_ma_def, inst_mc_def, inst_thm) = (
            (*
             cond_tracing "rewr_thm_folding_defs; matching:";
             cond_pprint (pretty_terms ctxt [ma_rhs, ma_rhs']);
             cond_tracing "rewr_thm_folding_defs; also matching:";
             cond_pprint (pretty_terms ctxt [mc_rhs, mc_rhs']);
             *)
            let
                val match = match_sim_thm ctxt thm ma_rhs mc_rhs
                val inst = instantiate_normalize match
            in
                (inst ma_def, inst mc_def, inst thm)
            end
        ) handle Pattern.MATCH => (
                 cond_tracing "rewr_thm_folding_defs: no match";
                 (ma_def, mc_def, thm)
                 )

        val (ma_def_sym, mc_def_sym) = pairself Thm.symmetric (inst_ma_def, inst_mc_def)
        val res = Raw_Simplifier.rewrite_rule
                      (* Use add_thm to avoid duplication when ma and mc are the same *)
                      (Thm.add_thm ma_def_sym [mc_def_sym])
                       inst_thm
    in
        if prop_of res = prop_of thm
        then (
            cond_tracing "Failed to revert the definitions: ";
            cond_pprint (pretty_thms ctxt
                                     [ma_def_sym, mc_def_sym]);
            cond_tracing "in";
            cond_pprint (pretty_thm ctxt inst_thm))
        else ();
        res
    end

    fun try_process_defs' ma mc adjust_ma adjust_mc = let

        val _ = cond_tracing "in try_process_defs"
        val _ = cond_pprint (pretty_terms ctxt [ma, mc])
        val thy = Proof_Context.theory_of ctxt

        val term_defs = get_defs ma mc
        val terms_w_defs = term_defs
                           |> map_opt (fn (ma_def, mc_def) => ((ma, ma_def), (mc, mc_def)))

        val prep_thms =
            ` (pairself (Thm.rhs_of #> term_of))
              #>> (fn (ma_rhs, mc_rhs) => build_sim_thms trace simp ctxt (adjust_ma ma_rhs) (adjust_mc mc_rhs))
              #> (fn ((sim_thms, ctxt'), (ma_def, mc_def)) =>
                     (map (rewr_thm_folding_defs ma_def mc_def adjust_ma adjust_mc) sim_thms, ctxt'))
              #> (fn (thms, ctxt') => (thms, store_thms thms ctxt'))
              (* Instantiating will throw a Pattern.MATCH if the
                constant definitions use different variable names,
                hence the try; notice that the theorems
                are already stored anyway above *)
              #> (fn (thms, ctxt') => (
                     map (try (fn thm => (inst_sim_thm ctxt' thm (adjust_ma ma) (adjust_mc mc)))) thms, ctxt'
                     )
                 )
              #>> take_somes


        fun rewrite_term_using_def (t, t_def) =
            Raw_Simplifier.rewrite_term thy [t_def] [] t

        val gen_frees = Thm.forall_intr_frees #> gen_all

        fun non_generalized_ref ctxt =
            ` (pairself rewrite_term_using_def)
              ##> pairself snd
              #> (fn ((ma', mc'), (ma_def, mc_def)) =>
                    build_sim_thms trace simp ctxt (adjust_ma ma') (adjust_mc mc')
                           |>> map (rewr_thm_folding_defs ma_def mc_def adjust_ma adjust_mc))
              #> (fn (thms, ctxt') => (thms, 
                    if Config.get ctxt' ref_tac_store_non_gen
                    then store_thms (map gen_frees thms) ctxt'
                    else ctxt'))

    in
        term_defs
            |> map_opt prep_thms
            |> the_default ([], ctxt)
            |> (fn (thms, ctxt) =>
                   if null thms
                   then map_opt (non_generalized_ref ctxt) terms_w_defs
                                |> the_default ([], ctxt)
                   else (thms, ctxt)
               )

    end                                                    

    fun process_lambda_bindS ma mc = let
        val _ = (
            cond_tracing "process_lambda_bindS: relating";
            cond_pprint (pretty_terms ctxt [ma, mc])
        )

        val dest_projectR_bindS = abs_body #> dest_binop #>> snd o dest_comb o fst o dest_comb
        val (ra, ha) = dest_projectR_bindS ma
        val (rc, hc) = dest_projectR_bindS mc

        val (thms_r, ctxt') = build_sim_thms trace simp ctxt ra rc
        val (thms_h, ctxt'') = build_sim_thms trace simp ctxt' ha hc

        fun make_bindS_with ctxt h m = let
            val projectR = @{term "projectR"} |> Logic.varify_global
            val bindS = @{term "bindS"} |> Logic.varify_global

            val mR = apply_typed ctxt projectR m
            val fv = Term.variant_frees mR [("s", fastype_of mR |> domain_type)]
                                         |> hd
            val mRs = mR $ Free fv
            val bound = apply_typed_list ctxt (bindS, [mRs, h])
            val fv' = dest_binop bound |> fst |> dest_comb |> snd
            val res = Term.lambda fv' bound
        in
            res
        end

        fun process_same_binop_bindS ctxt ra rc ha hc = let
          val _ = cond_tracing "bindS/Same monad op";
          val _ = cond_pprint (pretty_terms ctxt [ra, rc])
          val (ra1, ra2) = dest_binop ra
          val (rc1, rc2) = dest_binop rc
          val op_name = head_of ra |> const_name

          val (thms1, ctxt') = build_sim_thms trace simp ctxt ra1 rc1
          val (thms_r1, ctxt'') =
              build_sim_thms trace simp ctxt' (make_bindS_with ctxt' ha ra1) (make_bindS_with ctxt' hc rc1)
          val _ = if null thms_r1 then (cond_tracing "bindS/Same monad op failed: thms_r1 are null") else ()

          val (thms_r2, ctxt''') = process_fun_push_prems' ctxt'' ra2 rc2
                                                 (make_bindS_with ctxt'' ha)
                                                 (make_bindS_with ctxt'' hc)


          fun aux @{const_name "bind"} = let
              val _ = cond_tracing "bindS/aux: bind"
              val _ = cond_tracing "thms1"
              val _ = cond_pprint (pretty_thms ctxt thms1)
              val _ = if null thms1 then (cond_tracing "failed: thms1 are null") else ()
              val step1 = (thms1 RL (Ref_Thms.get_bind_bindS_thms ctxt))
                              |> map sim_prems_to_front
                          (* @{thms sim_norm_bind_bindS sim_norm_bind_bindS'} *)
              val _ = cond_tracing "step1"
              val _ = cond_pprint (pretty_thms ctxt step1)
              val _ = cond_tracing "thms_r1"
              val _ = cond_pprint (pretty_thms ctxt thms_r1)
              val _ = if null thms_r1 then (cond_tracing "failed: thms_r1 are null") else ()
              val step2 = (thms_r1 RL step1)
                              |> map sim_prems_to_front
              val _ = cond_tracing "step2"
              val _ = cond_pprint (pretty_thms ctxt step2)
              val _ = cond_tracing "thms_r2"
              val _ = cond_pprint (pretty_thms ctxt thms_r2)
              val _ = if null thms_r2 then (cond_tracing "failed: thms_r2 are null") else ()
              val step3 = thms_r2 RL step2
              val _ = cond_tracing "step3"
              val _ = cond_pprint (pretty_thms ctxt step3)
          in
              step3
          end
            | aux @{const_name "try_catch"} = let
                  val _ = cond_tracing "bindS/aux: try_catch"
                  val _ = cond_tracing "thms1"
                  val _ = cond_pprint (pretty_thms ctxt thms1)
                  val _ = if null thms1 then (cond_tracing "failed: thms1 are null") else ()
                  val step1 = (thms1 RL (Ref_Thms.get_try_catch_bindS_thms ctxt))
                                    |> map sim_prems_to_front
                              (* @{thms sim_norm_try_catch_bindS} *)
                  val _ = cond_tracing "step1"
                  val _ = cond_pprint (pretty_thms ctxt step1)
                  val _ = cond_tracing "thms_r1"
                  val _ = cond_pprint (pretty_thms ctxt thms_r1)
                  val _ = if null thms_r1 then (cond_tracing "failed: thms_r1 are null") else ()
                  val step2 = (thms_r1 RL step1) |> map sim_prems_to_front
                  val _ = cond_tracing "step2"
                  val _ = cond_pprint (pretty_thms ctxt step2)
                  val _ = cond_tracing "thms_r2"
                  val _ = cond_pprint (pretty_thms ctxt thms_r2)
                  val _ = if null thms_r2 then (cond_tracing "failed: thms_r2 are null") else ()
                  val step3 = thms_r2 RL step2
                  val _ = cond_tracing "step3"
                  val _ = cond_pprint (pretty_thms ctxt step3)
              in
                  step3
              end
            | aux op_name = let
                  val _ = cond_tracing ("bindS/aux: Unknown binder: " ^ op_name)
              in
                  []
              end
                          
        in
            (aux op_name, ctxt''')
        end

        fun process_exceptions_case_bindS ctxt ra rc ha hc =
            process_exceptions_case' ra rc (make_bindS_with ctxt'' ha) (make_bindS_with ctxt'' hc)

        fun process_if_bindS ctxt ra rc ha hc =
            process_if' ra rc (make_bindS_with ctxt'' ha) (make_bindS_with ctxt'' hc)

        fun process_choice_bindS ctxt ra rc ha hc =
            process_choice' ra rc (make_bindS_with ctxt'' ha) (make_bindS_with ctxt'' hc)

        fun process_mapM_bindS ctxt ra rc ha hc = let
            (* TODO: this will probably not work if process_fun returns theorems with assumptions *)
        in
            process_mapM' ctxt'' ra rc
                          I I (Ref_Thms.get_mapM_bindS_thms ctxt) (* @{thms sim_norm_mapM_bindS} *)
                                   |> (fn (thms, ctxt) =>
                                          process_mapM' ctxt ra rc
                                            (make_bindS_with ctxt'' ha)
                                            (make_bindS_with ctxt'' hc)
                                            thms
                                      )
        end

        fun try_process_defs_bindS ra rc ha hc =
            try_process_defs' ra rc (make_bindS_with ctxt'' ha) (make_bindS_with ctxt'' hc)

    in
        if exists (fn thm =>
                      member (op =)
                             [@{const_name "sim"}, @{const_name "sim_ext"}]
                             (sim_type thm))
                  thms_r
        then (
            cond_tracing "process_lambda_bindS: found a sim for ma/mc";
            let
                val _ = cond_tracing "thms_r";
                val _ = cond_pprint (pretty_thms ctxt'' thms_r)
                val s1 = (thms_r RL (Ref_Thms.get_bindS_thms ctxt''))
                val _ = cond_tracing "Step 1:"
                val _ = cond_pprint (pretty_thms ctxt'' s1)
                val s2 = thms_h RL s1
                val _ = cond_tracing "Step 2:"
                val _ = cond_pprint (pretty_thms ctxt'' s2)
                val res =
                    (s2, ctxt'')
                    (* @{thms sim_norm_bindS}*)
            in
                (cond_tracing "process_lambda_bindS result:";
                 cond_pprint (pretty_thms ctxt'' (fst res));
                 res
                 )
            end
            )
        else (
            cond_tracing "process_lambda_bindS: no sim for ra/rc";
            if same_monad_ops (head_of ra) (head_of rc)
            then process_same_binop_bindS ctxt'' ra rc ha hc
            else if forall is_exceptions_case [head_of ra, head_of rc]
            then process_exceptions_case_bindS ctxt'' ra rc ha hc
            else if forall is_if [head_of ra, head_of rc]
            then process_if_bindS ctxt'' ra rc ha hc
            else if forall is_choice [head_of ra, head_of rc]
            then process_choice_bindS ctxt' ra rc ha hc
            else if forall is_mapM [head_of ra, head_of rc]
            then process_mapM_bindS ctxt'' ra rc ha hc
            else (
                cond_tracing "process_lambda_bindS: not same binop_bindS";
                cond_pprint (pretty_terms ctxt [ra, rc]);
                try_process_defs_bindS ra rc ha hc
                )
            )

    end

    fun try_process_defs ma mc = try_process_defs' ma mc I I                         

    fun continue_search ma mc = let
       val ha = head_of ma
       val hc = head_of mc
       val _ = cond_tracing "No established refinement theorems found; continuing search"
    in
        (if same_monad_ops ha hc
            then process_same_binop ma mc
        else if forall is_exceptions_case [ha, hc]
            then process_exceptions_case ma mc
        else if forall is_if [ha, hc]
            then process_if ma mc
        else if forall is_choice [ha, hc]
            then process_choice ma mc
        else if forall is_mapM [ha, hc]
            then process_mapM ma mc
        else if forall is_lambda_bindS [ha, hc]
            then process_lambda_bindS ma mc
        else try_process_defs ma mc
        )
            |> finalize_thms
            |> tap (fn (thms, _) =>
                       if null thms 
                       then (
                           cond_tracing "continue_search failed for";
                           cond_pprint (pretty_terms ctxt [ma, mc])
                           )
                       else ()
                   )

             (*
        else    
            (cond_tracing "Not same monad op"; [])
            *)
    end

     (* Instantiate a list of simulation theorems so that it matches
     the given abstract/concrete terms *)

    fun instantiate_sim_thms ma mc thms =
        thms
            |> tap (K (cond_tracing "instantiate_sim_thms: before instantiating"))
            |> tap (cond_pprint o pretty_thms ctxt)
            |> map (fn thm => (
                       (* cond_tracing "trying to instantiate";
                        cond_pprint (pretty_thm ctxt thm);
                        *)
                       ((inst_sim_thm ctxt thm ma mc) |> SOME)
                       handle TYPE e => (
                        
                              cond_tracing "instantiate_sim_thms: Caught an exception";
                              cond_tracing (PolyML.makestring e);
                              if trace
                              then raise (TYPE e)
                              else NONE
                              )
                            | _ => NONE
                                                            
                       (* try (inst_sim_thm ctxt thm ma) mc *)
                   ))
            |> take_somes
            |> tap (K (cond_tracing "instantiate_sim_thms: result"))
            |> tap (cond_pprint o pretty_thms ctxt)

    val thms = Ref_Thms.get_term_ref ctxt mc
                                      |> instantiate_sim_thms ma mc
                                      |> filter_by_prem_nrs

in
  if not (null thms)
    then (thms, ctxt)
    else continue_search ma mc
end

end;

fun mk_ref_tac trace simp tac {context, concl,...} = let
  val cond_tracing = if trace then tracing else K ()
  val cond_pprint = if trace then pprint else K ()
  val ctxt = context

  val concl' = term_of concl |> HOLogic.dest_Trueprop

  val _ = (
      cond_tracing "Solving goal";
      cond_pprint (pretty_term ctxt concl')
  )

  val (thms, ctxt') = (
      Ref_Thms.extract_sim_targets concl'
      |> uncurry (build_sim_thms trace simp ctxt)
      |>> sort (int_ord o pairself my_nr_prems)
  ) handle TERM _ => ([], ctxt)

  val _ = (
      cond_tracing "mk_ref_tac: Resulting theorems: ";
      cond_pprint (pretty_thms ctxt thms)
  )

in
  tac thms
end

fun iterateN 0 _ x = x
  | iterateN n f x = iterateN (n-1) f (f x)

fun is_Trueprop (Const ("HOL.Trueprop", _) $ _) = true
  | is_Trueprop _ = false

fun instantiate_redundant_subsets thm = 
    let
        val thy = theory_of_thm thm
        fun does_match spat obj =
            Pattern.matches thy
                                 (spat |> Syntax.read_term_global thy |> Logic.varify_global, obj)

        val cprems = cprems_of thm
        val subset_cprems = cprems
                               |> filter (is_Trueprop o term_of)
                               |> map Thm.dest_arg
                               |> filter (does_match "(x :: 'a :: order) <= y" o term_of)

        val target_prems =  subset_cprems
                                |> filter (is_Var o term_of o fst o Thm.dest_binop)

        fun swap_binop ct = let
            val bop = Thm.dest_fun2 ct
            val (lhs, rhs) = Thm.dest_binop ct
        in
            Thm.mk_binop bop rhs lhs
        end

        val antisym_prems = filter (fn ct => member (op aconvc) subset_cprems (swap_binop ct)) target_prems
                            
     (*
        val _ = if not (null antisym_prems)
                then (tracing "Instantiating:";
                      pprint (pretty_cterms @{context} (map (fst o Thm.dest_binop) antisym_prems));
                      tracing "With";
                      pprint (pretty_cterms @{context} (map (snd o Thm.dest_binop) antisym_prems))
                     )
                else ()
      *)

    in
        cterm_instantiate (map Thm.dest_binop antisym_prems) thm
    end

    
    
fun default_simp ctxt thms = let
    val simp_thms = Ref_Thms.get_simp_thms ctxt
        (*
         @{thms
                          UNIV_def[THEN sym] Set.empty_def[THEN sym]
                          Int_UNIV_right Int_UNIV_left Int_empty_left Int_empty_right empty_subsetI
                          Int_Collects Collect_subset Set.Int_Collect Pmm_Refinement.Int_Collect
                          mem_Collect_eq
                          subset_Collect subset_bCollect empty_iff Int_iff imp_conj_same imp_conj_simps
                          conj_commute conj_assoc conj_left_absorb iso_tuple_UNIV_I
                          HOL.simp_thms HOL.ex_simps
                          HOL.True_implies_equals
                          Product_Type.split
                          hash_ptr_diff triv_precond bool_into_set_def
                          indx.simps
                          comm_same_alter
                          force_same_and_diff force_same_and_subset force_same_and_same
                          if_True
                          subset_refl_eq
                          imp_ex[THEN sym]
                          Int_lower1_eq Int_lower2_eq}

         *)
     (*
    fun RS_first_prem thma thmb = let
        val nr_prems = Thm.nprems_of thmb
        fun aux 1 = thma RS thmb
          | aux n = (thma RSN (n, thmb)) handle e => aux (n-1)
    in
        aux (Int.max (nr_prems, 1))
    end

    fun elim_true thm = RS_first_prem @{thm TrueI} thm

    fun try_elim_true thm = the_default thm (try elim_true thm)
      *)
in
    map
        (
             (Simplifier.asm_full_simplify (
                  HOL_basic_ss addsimps simp_thms
                               |> Raw_Simplifier.add_cong @{thm Collect_cong}
                 ))
             #> instantiate_redundant_subsets
             #> Simplifier.full_simplify (HOL_basic_ss addsimps simp_thms)

        )
        thms
        (* |> map try_elim_true *)
end

(*
fun refinement_tac trace ctxt i st = let
  val cond_tracing = if trace then tracing else K ()
  val cond_pprint = if trace then pprint else K ()


    val ({concl = conc, ...}, _) = Subgoal.focus ctxt i st
  in
    conc |> term_of |> HOLogic.dest_Trueprop
  end

  val (ta, tc) = Ref_Thms.extract_sim_targets concl

  val thms = build_sim_thms trace ctxt ta tc

  val _ = (cond_tracing "Resulting theorems: "; cond_pprint (pretty_thms ctxt thms))

in
  resolve_tac thms i st
end
*)

fun refinement_noendsimp_tac trace simp =
    Subgoal.FOCUS (mk_ref_tac trace simp (fn thms => resolve_tac thms 1))

(*
 TODO: this proves all the goals, but then dies
with "proved a different theorem" for Abs2_Ref.get_raw_hashes_ref
     *)
fun refinement_tac trace simp ctxt i =
    (Subgoal.FOCUS (mk_ref_tac trace simp
                               (fn thms => FIRST' [eresolve_tac thms, resolve_tac thms] 1)) ctxt i)
    THEN (TRY (PRIMSEQ (Seq.of_list o simp ctxt o single)))

    

fun refinement_ins_tac trace simp =
    Subgoal.FOCUS (mk_ref_tac trace simp (fn thms => Method.insert_tac thms 1))

fun refinement_gen_tac trace simp tac =
    Subgoal.FOCUS (mk_ref_tac trace simp (fn thms => tac thms))
    
fun refinement_force_tac trace simp ctxt =
    Subgoal.FOCUS
        (mk_ref_tac trace simp
                    (fn thms =>
                        CHANGED (force_tac (ctxt addIs thms) 1)
                    )
        ) ctxt

end

open Ref_Tac
